/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file algorithms.hpp
 *
 * @brief
 *         Linear algebra operations on `HMatrix`es and `HVector`s.
 *
 */

#ifndef BSC_KLAMMLER_ALGORITHMS_HPP
#define BSC_KLAMMLER_ALGORITHMS_HPP

#include <cmath>                      // std::sqrt
#include <functional>                 // std::plus, std::minus
#include <memory>                     // std::unique_ptr

#ifdef __GNUC__
#  pragma GCC diagnostic push
#  pragma GCC diagnostic ignored "-Wunused-local-typedefs"
#endif
#include <Eigen/Eigenvalues>          // Eigen::EigenSolver
#ifdef __GNUC__
#  pragma GCC diagnostic pop
#endif

#include "arranged_matrix.hpp"        // bsc_klammler::ArrangedMatrix
#include "block_matrix.hpp"           // bsc_klammler::BlockMatrix
#include "full_matrix.hpp"            // bsc_klammler::FullMatrix
#include "full_vector.hpp"            // bsc_klammler::FullVector
#include "inefficient_algorithm.hpp"  // BSC_KLAMMLER_INEFFICIENT_ALGORITHM
#include "matrix_view.hpp"            // bsc_klammler::MatrixView
#include "nest.hpp"                   // bsc_klammler::nest::
#include "not_implemented.hpp"        // NotImplementedError
#include "rectangle.hpp"              // bsc_klammler::Rectangle


namespace bsc_klammler
{

  /**
   * @brief
   *         Implementation details of header file `algorithms.hpp`.
   *
   */
  namespace algorithms_hpp
  {

    // Declarations for template functions that will be needed soon but can
    // only be defined near the end of this file.

    /**
     * @brief
     *         Polymorphic as-efficient-as-possible implementation of a
     *         matrix-matrix product.
     *
     * @param left
     *         left-hand operand
     *
     * @param right
     *         right-hand operand
     *
     * @returns
     *         the matrix product of the two `HMatrix`es
     *
     * @throws std::invalid_argument
     *         if the matrices have incompatible dimensions
     *
     */
    template<typename T>
    std::unique_ptr<HMatrix<T>>
    poly_prod(const HMatrix<T>& left, const HMatrix<T>& right);

    /**
     * @brief
     *         Most inefficient fallback implementation of a matrix-matrix
     *         product.
     *
     * @param left
     *         left-hand operand
     *
     * @param right
     *         right-hand operand
     *
     * @returns
     *         the matrix product of the two `HMatrix`es
     *
     * @throws std::invalid_argument
     *         if the matrices have incompatible dimensions
     *
     */
    template<typename T>
    FullMatrix<T>
    dumb_prod(const HMatrix<T>& left, const HMatrix<T>& right);

    /**
     * @brief
     *         Adds or subtracts (part of) a `HMatrix` (a `FullMatrix`,
     *         preferrably) to or from an other `HMatrix`.
     *
     * This function handles addition and subtraction equally.  For the sake of
     * brevity, we'll only talk about addition here.
     *
     * While there are no restrictions on the <code><var>augend</var></code>,
     * the user of this function is responsible for preparing the
     * <code><var>addend</var></code> to be -- if all possible -- a single
     * `FullMatrix`.  Otherwise, the function will not be able to perform
     * efficiently.  (The result will still be correct, however.)  Usually,
     * this means that addition of an arbitrary `HMatrix` <var>B</var> to
     * another arbitrary `HMatrix` <var>A</var> will involve two steps:
     *
     * 1. Find all atomic `FullMatrix`es <var>B</var><sub>1</sub>, &hellip;,
     *    <var>B</var><sub><var>n</var></sub> that compose <var>B</var>.
     * 2. For each block <var>B</var><sub><var>i</var></sub>, call this
     *    function with <var>A</var> as the augend and
     *    <var>B</var><sub><var>i</var></sub> as the addend.
     *
     * @param <Op>
     *         selects the operation: add (<code>'+'</code>) or subtract
     *         (<code>'-'</code>)
     *
     * @param <T>
     *         type of the matrix elements
     *
     * @param augend
     *         augend (or minuend) to be updated
     *
     * @param aug_north
     *         first row in the augend that is to be updated
     *
     * @param aug_west
     *         first column in the augend that is to be updated
     *
     * @param addend
     *         addend (or subtrahend) to be added (or subtracted) to the augend
     *         (or minuend)
     *
     * @param add_trans
     *         whether the addend is to be transposed before doing anything
     *
     * @param add_north
     *         first row in the addend to be taken
     *
     * @param add_west
     *         first column in the addend to be taken
     *
     * @param add_south
     *         first row in the addend to be *not* taken
     *
     * @param add_east
     *         first column in the addend to be *not* taken
     *
     */
    template<char Op, typename T>
    void
    take_delta(HMatrix<T>& augend,
               const typename HMatrix<T>::index_type aug_north,
               const typename HMatrix<T>::index_type aug_west,
               const HMatrix<T>& addend,
               const bool add_trans,
               const typename HMatrix<T>::index_type add_north,
               const typename HMatrix<T>::index_type add_west,
               const typename HMatrix<T>::index_type add_south,
               const typename HMatrix<T>::index_type add_east);

    // End of declarations.

    /**
     * @brief
     *         Functor to enable selection of NeSt only for certain types.
     *
     * @param <T>
     *         type for matrix entries
     *
     */
    template<typename T>
    struct inverter
    {

      /**
       * @brief
       *         Inverts the given matrix without parameter validation *not*
       *         using NeSt.
       *
       * @param
       *         regular square matrix to invert
       *
       * @param spd
       *         whether the matrix is symmetric and positive definite
       *         (ignored)
       *
       * @returns
       *         the inverted matrix
       *
       */
      static FullMatrix<T>
      invert(const FullMatrix<T>& matrix,
             const bool spd BSC_KLAMMLER_ATTRIBUTE_UNUSED)
      {
        return FullMatrix<T> {matrix.getRawMatrix().inverse()};
      }
    };

    /**
     * @brief
     *         Specialized functor to enable selection of NeSt only for certain
     *         types.
     *
     */
    template<>
    struct inverter<double>
    {
      /** @brief Thereshold above which NeSt may be used. */
      static constexpr typename HMatrix<double>::index_type small_size {32};

      /**
       * @brief
       *         Inverts the given matrix without parameter validation
       *         *evantually* using NeSt.
       *
       * @param
       *         regular square matrix to invert
       *
       * @param spd
       *         whether the matrix is symmetric and positive definite
       *
       * @returns
       *         the inverted matrix
       *
       */
      static FullMatrix<double>
      invert(const FullMatrix<double>& matrix, const bool spd)
      {
        // The checks are ordered by their cost.  The first is a compile-time
        // constant, the second a virtual function call and the third requires
        // a mutex be acquired.
        if (BSC_KLAMMLER_WITH_NEST
            && BSC_KLAMMLER_PARALLEL >= BSC_KLAMMLER_PARALLEL_DATA)
          {
            if (matrix.size() > inverter::small_size)
              {
                if (enable_data_level_multithreading() > 1)
                  {
                    FullMatrix<double> inverse {matrix.rows(), matrix.cols()};
                    bsc_klammler::nest::invert(matrix.getRawMatrix().data(),
                                               inverse.getRawMatrix().data(),
                                               matrix.size(), spd);
                    return inverse;
                  }
              }
          }
        return FullMatrix<double> {matrix.getRawMatrix().inverse()};
      }
    };

    /**
     * @brief
     *         Argument checking for matrix products.
     *
     * If the two `HMatrix`es can be multiplied with each other, this function
     * has no effect.  Otherwise, a `std::invalid_argument` exception is
     * thrown.  Calls to this function should be wrapped into
     * `if (BSC_KLAMMLER_CHECK_ARGS) { check_dims_prod(left, right); }`
     * clauses.
     *
     * @param <T>
     *         type of the matrix elements
     *
     * @param left
     *         left-hand operand
     *
     * @param right
     *         right-hand operand
     *
     * @throws std::invalid_argument
     *         if <code><var>left</var>.cols() !=
     *         <var>right</var>.rows()</code>
     *
     */
    template<typename T>
    void
    check_dims_prod(const HMatrix<T>& left, const HMatrix<T>& right)
    {
      if (left.cols() != right.rows())
        HMatrix<T>::throwMismatchedDimensions("multiply", left, "with", right);
    }

    /**
     * @brief
     *         Generic implementation for operator `*=` and `/=` for
     *         `HVector`s.
     *
     * @param <Op>
     *         selects the operation: multiply (<code>'*'</code>) or divide
     *         (<code>'/'</code>)
     *
     * @param <T>
     *         type of the vector elements
     *
     * @param self
     *         &ldquo;left-hand&rdquo; operand that will be modified
     *
     * @param factor
     *         factor to multiply with or divide the vector elements by
     *
     */
    template<char Op, typename T>
    void
    mul_div_to(HVector<T>& self, const T factor)
    {
      static_assert(Op == '*' || Op == '/',
                    "template parameter 'Op' must be '*' or '/'");
      using value_type = typename HVector<T>::value_type;
      using index_type = typename HVector<T>::index_type;
      using full_vector = FullVector<value_type>;
      auto full_self_ptr = dynamic_cast<full_vector *>(&self);
      if (full_self_ptr != nullptr)
        {
          switch (Op)
            {
            case '*':
              full_self_ptr->getRawVector() *= factor;
              break;
            case '/':
              full_self_ptr->getRawVector() /= factor;
              break;
            default:
              BSC_KLAMMLER_NOT_REACHED;
            }
        }
      else
        {
          BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
          for (index_type i = 0; i < self.size(); ++i)
            {
              value_type v {self.get(i)};
              if (v != HVector<T>::zero())
                {
                  switch (Op)
                    {
                    case '*': v *= factor; break;
                    case '/': v /= factor; break;
                    default: BSC_KLAMMLER_NOT_REACHED;
                    }
                  self.set(v, i);
                }
            }
        }
    }

    /**
     * @brief
     *         Generic implementation for operator `+=` and `-=` for
     *         `HVector`s.
     *
     * @param <Op>
     *         selects the operation: add (<code>'+'</code>) or subtract
     *         (<code>'-'</code>)
     *
     * @param <T>
     *         type of the vector elements
     *
     * @param self
     *         &ldquo;left-hand&rdquo; operand that will be modified
     *
     * @param other
     *         &ldquo;right-hand&rdquo; operand that will be added / subtracted
     *
     */
    template<char Op, typename T>
    void
    plus_minus_to(HVector<T>& self, const HVector<T>& other)
    {
      static_assert(Op == '+' || Op == '-',
                    "template parameter 'Op' must be '+' or '-'");
      using value_type = typename HVector<T>::value_type;
      using index_type = typename HVector<T>::index_type;
      using full_vector = FullVector<value_type>;
      auto full_self_ptr = dynamic_cast<full_vector *>(&self);
      auto full_other_ptr = dynamic_cast<const full_vector *>(&other);
      if (full_self_ptr != nullptr && full_other_ptr != nullptr)
        {
          switch (Op)
            {
            case '+':
              full_self_ptr->getRawVector() += full_other_ptr->getRawVector();
              break;
            case '-':
              full_self_ptr->getRawVector() -= full_other_ptr->getRawVector();
              break;
            default:
              BSC_KLAMMLER_NOT_REACHED;
            }
        }
      else
        {
          BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
          for (index_type i = 0; i < other.size(); ++i)
            {
              const value_type delta {other.get(i)};
              if (delta != HVector<T>::zero())
                {
                  value_type v {self.get(i)};
                  switch (Op)
                    {
                    case '+': v += delta; break;
                    case '-': v -= delta; break;
                    default: BSC_KLAMMLER_NOT_REACHED;
                    }
                  self.set(v, i);
                }
            }
        }
    }

    /**
     * @brief
     *         Generic implementation for operator `*=` and `/=` for
     *         `HMatrix`es.
     *
     * @param <Op>
     *         selects the operation: multiply (<code>'*'</code>) or divide
     *         (<code>'/'</code>)
     *
     * @param <T>
     *         type of the matrix elements
     *
     * @param self
     *         &ldquo;left-hand&rdquo; operand that will be modified
     *
     * @param factor
     *         factor to multiply with or divide by the matrix elements
     *
     */
    template<char Op, typename T>
    void
    mul_div_to(HMatrix<T>& self, const T factor)
    {
      static_assert(Op == '*' || Op == '/',
                    "template parameter 'Op' must be '*' or '/'");
      using h_matrix = HMatrix<T>;
      using full_matrix = FullMatrix<T>;
      using value_type = typename h_matrix::value_type;
      using index_type = typename h_matrix::index_type;
      using active_visitor = typename h_matrix::ActiveVisitor;
      struct Visitor : public active_visitor
      {
        const value_type v_factor;

        Visitor(const value_type f) : v_factor {f}
        {
        }

        virtual void
        visitRegion(const index_type global_north BSC_KLAMMLER_ATTRIBUTE_UNUSED,
                    const index_type global_west BSC_KLAMMLER_ATTRIBUTE_UNUSED,
                    const bool global_trans,
                    h_matrix& matrix,
                    const index_type north,
                    const index_type west,
                    const index_type south,
                    const index_type east) final override
        {
          if (auto full_ptr = dynamic_cast<full_matrix *>(&matrix))
            {
              const index_type height {south - north};
              const index_type width {east - west};
              switch ((global_trans ? 'T' : 0) + Op)
                {
                case '*':
                  full_ptr->getRawMatrix().block(north, west, height, width)
                    *= this->v_factor;
                  break;
                case '/':
                  full_ptr->getRawMatrix().block(north, west, height, width)
                    /= this->v_factor;
                  break;
                case 'T' + '*':
                  full_ptr->getRawMatrix().block(west, north, width, height)
                    *= this->v_factor;
                  break;
                case 'T' + '/':
                  full_ptr->getRawMatrix().block(west, north, width, height)
                    /= this->v_factor;
                  break;
                default:
                  BSC_KLAMMLER_NOT_REACHED;
                }
            }
          else
            {
              BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
              const index_type i0 {!global_trans ? north : west};
              const index_type i1 {!global_trans ? south : east};
              const index_type j0 {!global_trans ? west  : north};
              const index_type j1 {!global_trans ? east  : south};
              for (index_type j = j0; j < j1; ++j)
                {
                  for (index_type i = i0; i < i1; ++i)
                    {
                      value_type v {matrix.get(i, j)};
                      if (v != h_matrix::zero())
                        {
                          switch (Op)
                            {
                            case '*': v *= this->v_factor; break;
                            case '/': v /= this->v_factor; break;
                            default: BSC_KLAMMLER_NOT_REACHED;
                            }
                          matrix.set(v, i, j);
                        }
                    }
                }
            }
        }
      };
      Visitor vtor {factor};
      pass_over_region(self, 0, 0, false, vtor);
    }

    /**
     * @brief
     *         Generic implementation for operator `+=` and `-=` for
     *         `HMatrix`es.
     *
     * @param <Op>
     *         selects the operation: addition (<code>'+'</code>) or
     *         subtraction (<code>'-'</code>)
     *
     * @param <T>
     *         type of the matrix elements
     *
     * @param self
     *         &ldquo;left-hand&rdquo; operand that will be modified
     *
     * @param other
     *         &ldquo;right-hand&rdquo; operand that will be added / subtracted
     *
     */
    template<char Op, typename T>
    void
    plus_minus_to(HMatrix<T>& self, const HMatrix<T>& other)
    {
      static_assert(Op == '+' || Op == '-',
                    "template parameter 'Op' must be '+' or '-'");
      using h_matrix = HMatrix<T>;
      using index_type = typename h_matrix::index_type;
      using passive_visitor = typename h_matrix::PassiveVisitor;
      struct Visitor : public passive_visitor
      {
        h_matrix *const augend_ptr;

        Visitor(h_matrix *const aug_ptr) : augend_ptr {aug_ptr}
        {
          assert(this->augend_ptr != nullptr);
        }

        virtual void
        visitRegion(const index_type global_north,
                    const index_type global_west,
                    const bool global_trans,
                    const h_matrix& addend,
                    const index_type north,
                    const index_type west,
                    const index_type south,
                    const index_type east) final override
        {
          take_delta<Op>(*this->augend_ptr,
                         global_north, global_west,
                         addend,
                         global_trans,
                         north, west, south, east);
        }
      };
      Visitor vtor {&self};
      pass_over_region(other, 0, 0, false, vtor);
    }
  }  // namespace algorithms_hpp

  /**
   * @brief
   *         Multiplies a `HVector` with a scalar factor.
   *
   * @param self
   *         `HVector` to be scaled
   *
   * @param factor
   *         factor to scale the vector elements with
   *
   * @returns
   *         a reference to the `HVector`
   *
   */
  template<typename T>
  HVector<T>&
  operator*=(HVector<T>& self, const T factor)
  {
    algorithms_hpp::mul_div_to<'*'>(self, factor);
    return self;
  }

  /**
   * @brief
   *         Divides a `HVector` by a scalar factor.
   *
   * @param self
   *         `HVector` to be divided
   *
   * @param factor
   *         inverse factor to scale the vector elements with
   *
   * @returns
   *         a reference to the `HVector`
   *
   */
  template<typename T>
  HVector<T>&
  operator/=(HVector<T>& self, const T factor)
  {
    algorithms_hpp::mul_div_to<'/'>(self, factor);
    return self;
  }

  /**
   * @brief
   *         Adds a `HVector` to another one.
   *
   * @param self
   *         augend
   *
   * @param other
   *         addend
   *
   * @returns
   *         a reference to the augend
   *
   * @throws std::invalid_argument
   *         if the addend has the wrong dimensions
   *
   */
  template<typename T>
  HVector<T>&
  operator+=(HVector<T>& self, const HVector<T>& other)
  {
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (other.size() != self.size())
          HVector<T>::throwMismatchedDimensions("add", other, "to", self);
      }
    algorithms_hpp::plus_minus_to<'+'>(self, other);
    return self;
  }

  /**
   * @brief
   *         Subtracts a `HVector` from another one.
   *
   * @param self
   *         minuend
   *
   * @param other
   *         subtrahend
   *
   * @returns
   *         a reference to the minuend
   *
   * @throws std::invalid_argument
   *         if the subtrahend has the wrong dimensions
   *
   */
  template<typename T>
  HVector<T>&
  operator-=(HVector<T>& self, const HVector<T>& other)
  {
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (other.size() != self.size())
          HVector<T>::throwMismatchedDimensions("subtract", other, "from", self);
      }
    algorithms_hpp::plus_minus_to<'-'>(self, other);
    return self;
  }

  /**
   * @brief
   *         Multiplies a `HMatrix` with a scalar factor.
   *
   * @param factor
   *         factor to scale the matrix elements with
   *
   * @returns
   *         a reference to the `HMatrix`
   *
   */
  template<typename T>
  HMatrix<T>&
  operator*=(HMatrix<T>& self, const T factor)
  {
    algorithms_hpp::mul_div_to<'*'>(self, factor);
    return self;
  }

  /**
   * @brief
   *         Divides a `HMatrix` by a scalar factor.
   *
   * @param factor
   *         inverse factor to scale the matrix elements with
   *
   * @returns
   *         a reference to the `HMatrix`
   *
   */
  template<typename T>
  HMatrix<T>&
  operator/=(HMatrix<T>& self, const T factor)
  {
    algorithms_hpp::mul_div_to<'/'>(self, factor);
    return self;
  }

  /**
   * @brief
   *         Adds a `HMatrix` to another one.
   *
   * @param self
   *         augend
   *
   * @param other
   *         addend
   *
   * @returns
   *         a reference to the augend
   *
   * @throws std::invalid_argument
   *         if the addend has the wrong dimensions
   *
   */
  template<typename T>
  HMatrix<T>&
  operator+=(HMatrix<T>& self, const HMatrix<T>& other)
  {
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (other.rows() != self.rows() || other.cols() != self.cols())
          HMatrix<T>::throwMismatchedDimensions("add", other, "to", self);
      }
    algorithms_hpp::plus_minus_to<'+'>(self, other);
    return self;
  }

  /**
   * @brief
   *         Subtracts a `HMatrix` from another one.
   *
   * @param self
   *         minuend
   *
   * @param other
   *         subtrahend
   *
   * @returns
   *         a reference to the minuend
   *
   * @throws std::invalid_argument
   *         if the matrix has the wrong dimensions
   *
   */
  template<typename T>
  HMatrix<T>&
  operator-=(HMatrix<T>& self, const HMatrix<T>& other)
  {
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (other.rows() != self.rows() || other.cols() != self.cols())
          HMatrix<T>::throwMismatchedDimensions("subtract", other, "from", self);
      }
    algorithms_hpp::plus_minus_to<'-'>(self, other);
    return self;
  }

  /**
   * @brief
   *         Adds two `HMatrix`es and returns the result in a new `FullMatrix`.
   *
   * @param left
   *         left-hand operand <var>A</var> &isin;
   *          <var>K</var><sup><var>n</var>&times;<var>m</var></sup>
   *
   * @param right
   *         right-hand operand <var>B</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>m</var></sup>
   *
   * @returns
   *         sum <var>A</var> + <var>B</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>m</var></sup> of the two
   *         matrices
   *
   * @throws std::invalid_argument
   *         if the two matrices have different dimensions
   *
   */
  template<typename T>
  FullMatrix<T>
  operator+(const HMatrix<T>& left, const HMatrix<T>& right)
  {
    BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (left.rows() != right.rows() || left.cols() != right.cols())
          HMatrix<T>::throwMismatchedDimensions("add", left, "and", right);
      }
    {
      FullMatrix<T> sum {left};
      sum += right;
      return sum;
    }
  }

  /**
   * @brief
   *         Subtracts two `HMatrix`es and returns the result in a new
   *         `FullMatrix`.
   *
   * @param left
   *         left-hand operand <var>A</var> &isin;
   *          <var>K</var><sup><var>n</var>&times;<var>m</var></sup>
   *
   * @param right
   *         right-hand operand <var>B</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>m</var></sup>
   *
   * @returns
   *         difference <var>A</var> &minus; <var>B</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>m</var></sup> of the two
   *         matrices
   *
   * @throws std::invalid_argument
   *         if the two matrices have different dimensions
   *
   */
  template<typename T>
  FullMatrix<T>
  operator-(const HMatrix<T>& left, const HMatrix<T>& right)
  {
    BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (left.rows() != right.rows() || left.cols() != right.cols())
          HMatrix<T>::throwMismatchedDimensions("subtract", left, "and", right);
      }
    {
      FullMatrix<T> diff {left};
      diff -= right;
      return diff;
    }
  }

  /**
   * @brief
   *         Multiplies two `HMatrix`es and returns the result in a new
   *         `FullMatrix`.
   *
   * @param left
   *         left-hand operand <var>A</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>k</var></sup>
   *
   * @param right
   *         right-hand operand <var>B</var> &isin;
   *         <var>K</var><sup><var>k</var>&times;<var>m</var></sup>
   *
   * @returns
   *         matrix product <var>A</var>&sdot;<var>B</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>m</var></sup> of the two
   *         matrices
   *
   * @throws std::invalid_argument
   *         if the number of columns in <code><var>left</var></code> is
   *         different from the number of rows in <code><var>right</var></code>
   *
   */
  template<typename T>
  FullMatrix<T>
  operator*(const HMatrix<T>& left, const HMatrix<T>& right)
  {
    if (BSC_KLAMMLER_CHECK_ARGS)
      algorithms_hpp::check_dims_prod(left, right);
    std::unique_ptr<HMatrix<T>> prod_uptr {
      algorithms_hpp::poly_prod(left, right)
    };
    assert(prod_uptr);
    if (auto full_ptr = dynamic_cast<FullMatrix<T> *>(prod_uptr.get()))
      {
        const FullMatrix<T> prod {std::move(*full_ptr)};
        assert(prod_uptr->size() == 0);
        return prod;
      }
    else
      {
        // This cannot happn with the current implementatin of 'poly_prod'.
        BSC_KLAMMLER_NOT_REACHED;
      }
  }

  /**
   * @brief
   *         Multiplies a `HMatrix` and an `ArrangedMatrix`es and returns the
   *         product in a new `ArrangedMatrix`.
   *
   * The computation will be most efficient if the right-hand operand
   * `isRowLinear`.
   *
   * @param left
   *         left-hand operand <var>A</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>k</var></sup>
   *
   * @param right
   *         right-hand operand <var>B</var> &isin;
   *         <var>K</var><sup><var>k</var>&times;<var>m</var></sup>
   *
   * @returns
   *         matrix product <var>A</var>&sdot;<var>B</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>m</var></sup> of the two
   *         matrices
   *
   * @throws std::invalid_argument
   *         if the number of columns in <code><var>left</var></code> is
   *         different from the number of rows in <code><var>right</var></code>
   *
   */
  template<typename T>
  ArrangedMatrix<T>
  operator*(const HMatrix<T>& left, const ArrangedMatrix<T>& right)
  {
    if (BSC_KLAMMLER_CHECK_ARGS)
      algorithms_hpp::check_dims_prod(left, right);
    ArrangedMatrix<T> prod {left.rows(), right.cols()};
    if (right.isRowLinear())
      {
        for (auto tile : right.tiles())
          prod.addTile(algorithms_hpp::poly_prod(left, *tile.getMatrix()),
                       0,
                       tile.west());
      }
    else
      {
        BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
        prod.addTile(algorithms_hpp::poly_prod(left, right), 0, 0);
      }
    assert(prod.isRowLinear());
    return prod;
  }

  /**
   * @brief
   *         Multiplies an `ArrangedMatrix` and a `HMatrix` and returns the
   *         product in a new `ArrangedMatrix`.
   *
   * The computation will be most efficient if the left-hand operand
   * `isColumnLinear()`.
   *
   * @param left
   *         left-hand operand <var>A</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>k</var></sup>
   *
   * @param right
   *         right-hand operand <var>B</var> &isin;
   *         <var>K</var><sup><var>k</var>&times;<var>m</var></sup>
   *
   * @returns
   *         matrix product <var>A</var>&sdot;<var>B</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>m</var></sup> of the two
   *         matrices
   *
   * @throws std::invalid_argument
   *         if the number of columns in <code><var>left</var></code> is
   *         different from the number of rows in <code><var>right</var></code>
   *
   */
  template<typename T>
  ArrangedMatrix<T>
  operator*(const ArrangedMatrix<T>& left, const HMatrix<T>& right)
  {
    if (BSC_KLAMMLER_CHECK_ARGS)
      algorithms_hpp::check_dims_prod(left, right);
    ArrangedMatrix<T> prod {left.rows(), right.cols()};
    if (left.isColumnLinear())
      {
        for (auto tile : left.tiles())
          prod.addTile(algorithms_hpp::poly_prod(*tile.getMatrix(), right),
                       tile.north(),
                       0);
      }
    else
      {
        BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
        prod.addTile(algorithms_hpp::poly_prod(left, right), 0, 0);
      }
    assert(prod.isColumnLinear());
    return prod;
  }

  /**
   * @brief
   *         Multiplies two `ArrangedMatrix`es and returns the product in a new
   *         `ArrangedMatrix`.
   *
   * The computation will be most efficient if the left-hand operand
   * `isColumnLinear()` and the right-hand one `isRowLinear()`.
   *
   * @param left
   *         left-hand operand <var>A</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>k</var></sup>
   *
   * @param right
   *         right-hand operand <var>B</var> &isin;
   *         <var>K</var><sup><var>k</var>&times;<var>m</var></sup>
   *
   * @returns
   *         matrix product <var>A</var>&sdot;<var>B</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>m</var></sup> of the two
   *         matrices
   *
   * @throws std::invalid_argument
   *         if the number of columns in <code><var>left</var></code> is
   *         different from the number of rows in <code><var>right</var></code>
   *
   */
  template<typename T>
  ArrangedMatrix<T>
  operator*(const ArrangedMatrix<T>& left, const ArrangedMatrix<T>& right)
  {
    if (BSC_KLAMMLER_CHECK_ARGS)
      algorithms_hpp::check_dims_prod(left, right);
    ArrangedMatrix<T> prod {left.rows(), right.cols()};
    if (left.isColumnLinear() && right.isRowLinear())
      {
        for (const auto& lhs : left.tiles())
          {
            for (const auto& rhs : right.tiles())
              {
                assert(lhs.west() == 0);
                assert(rhs.north() == 0);
                assert(lhs.cols() == rhs.rows());
                auto p_uptr = algorithms_hpp::poly_prod(
                    *lhs.getMatrix(), *rhs.getMatrix());
                prod.addTile(std::move(p_uptr), lhs.north(), rhs.west());
              }
          }
      }
    else
      {
        BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
        prod.addTile(algorithms_hpp::poly_prod(left, right), 0, 0);
      }
    return prod;
  }

  /**
   * @brief
   *         Computes the dot-product of two `HVector`s.
   *
   * @param left
   *         left-hand operand <var>u</var> &isin;
   *         <var>K</var><sup><var>n</var></sup>
   *
   * @param right
   *         right-hand operand <var>v</var> &isin;
   *         <var>K</var><sup><var>n</var></sup>
   *
   * @returns
   *         dot-product <var>u</var>&sdot;<var>v</var> &isin; <var>K</var>
   *         of the two vectors
   *
   * @throws std::invalid_argument
   *         if the two vectors have different or zero dimensions
   *
   */
  template<typename T>
  T
  operator*(const HVector<T>& left, const HVector<T>& right)
  {
    using index_type = typename HMatrix<T>::index_type;
    using value_type = typename HMatrix<T>::value_type;
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (left.size() != right.size())
          {
            std::ostringstream oss {};
            oss << "cannot compute dot product of a "
                << left.size()
                << " and a "
                << right.size()
                << " vector";
            throw std::invalid_argument {oss.str()};
          }
        if (left.size() == 0)
          {
            throw std::invalid_argument {
              "cannot compute dot product of 0 vector"
            };
          }
      }
    auto left_full_ptr = dynamic_cast<const FullVector<T> *>(&left);
    auto right_full_ptr = dynamic_cast<const FullVector<T> *>(&right);
    if (left_full_ptr != nullptr && right_full_ptr != nullptr)
      {
        return left_full_ptr->getRawVector().dot(right_full_ptr->getRawVector());
      }
    else
    {
      BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
      value_type acc {};
      for (index_type i = 0; i < left.size(); ++i)
        {
          acc += left.get(i) * right.get(i);
        }
      return acc;
    }
  }

  /**
   * @brief
   *         Multiplies a `HVector` from the left with a `HMatrix` and returns
   *         the result in a new `FullVector`.
   *
   * @param left
   *         left-hand operand <var>A</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>k</var></sup>
   *
   * @param right
   *         right-hand operand <var>v</var> &isin;
   *         <var>K</var><sup><var>n</var></sup>
   *
   * @returns
   *         product <var>A</var>&sdot;<var>v</var> &isin;
   *         <var>K</var><sup>n</var></sup> of the matrix and the vector
   *
   * @throws std::invalid_argument
   *         if the vector's dimension is different from the number of columns
   *         in the matrix
   *
   */
  template<typename T>
  FullVector<T>
  operator*(const HMatrix<T>& left, const HVector<T>& right)
  {
    using index_type = typename HMatrix<T>::index_type;
    using value_type = typename HMatrix<T>::value_type;
    using raw_index_type = typename FullMatrix<T>::raw_index_type;
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (left.cols() != right.size())
          {
            std::ostringstream oss {};
            oss << "cannot multiply a "
                << "(" << left.rows() << ", " << left.cols() << ")"
                << " matrix with a "
                << right.size()
                << " vector";
            throw std::invalid_argument {oss.str()};
          }
      }
    struct Visitor : HMatrix<T>::PassiveVisitor
    {
      FullVector<T> prod;
      const HVector<T>& vector;

      Visitor(const HVector<T>& v, const index_type n) : prod {n}, vector(v)
      // Round parenthesis for initialization of 'vector' are not a typo.  See
      // https://stackoverflow.com/questions/25561387/spurious-warning-about-binding-temporary-to-reference-member-in-constructor/25595974
      {
      }

      virtual void
      visitRegion(const index_type global_north,
                  const index_type global_west,
                  const bool global_trans,
                  const HMatrix<T>& matrix,
                  const index_type north,
                  const index_type west,
                  const index_type south,
                  const index_type east) final override
      {
        auto full_vec_ptr = dynamic_cast<const FullVector<T> *>(&this->vector);
        auto full_mat_ptr = dynamic_cast<const FullMatrix<T> *>(&matrix);
        if (full_vec_ptr != nullptr && full_mat_ptr != nullptr)
          {
            const Rectangle<raw_index_type> local_rect {
              static_cast<raw_index_type>(north),
              static_cast<raw_index_type>(west),
              static_cast<raw_index_type>(south),
              static_cast<raw_index_type>(east),
            };
            const Rectangle<raw_index_type> global_rect {
              static_cast<raw_index_type>(global_north),
              static_cast<raw_index_type>(global_west),
              static_cast<raw_index_type>(global_north) + local_rect.height(),
              static_cast<raw_index_type>(global_west) + local_rect.width(),
            };
            if (!global_trans)
              {
                this->prod.getRawVector().block(
                      global_rect.north(), 0, global_rect.height(), 1)
                  += full_mat_ptr->getRawMatrix().block(
                         local_rect.north(), local_rect.west(),
                         local_rect.height(), local_rect.width())
                  * full_vec_ptr->getRawVector().block(
                        global_rect.west(), 0, global_rect.width(), 1);
              }
            else
              {
                this->prod.getRawVector().block(
                    global_rect.north(), 0, global_rect.height(), 1)
                  += full_mat_ptr->getRawMatrix().transpose().block(
                         local_rect.north(), local_rect.west(),
                         local_rect.height(), local_rect.width())
                  * full_vec_ptr->getRawVector().block(
                        global_rect.west(), 0, global_rect.width(), 1);
              }
          }
        else
          {
            BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
            for (index_type i = north; i < south; ++i)
              {
                const index_type vi {global_north + i - north};
                value_type acc {};
                for (index_type j = west; j < east; ++j)
                  {
                    const index_type vj {global_west + j - west};
                    if (!global_trans)
                      acc += matrix.get(i, j) * this->vector.get(vj);
                    else
                      acc += matrix.get(j, i) * this->vector.get(vj);
                  }
                this->prod.set(this->prod.get(vi) + acc, vi);
              }
          }
      }
    };
    Visitor vtor {right, left.rows()};
    pass_over_region(left, 0, 0, false, vtor);
    return std::move(vtor.prod);
  }

  /**
   * @brief
   *         Multiplies a `HVector` from the right with a `HMatrix` and returns
   *         the result in a new `FullVector`.
   *
   * @param left
   *         left-hand operand <var>v</var> &isin;
   *         <var>K</var><sup><var>n</var></sup>
   *
   * @param right
   *         right-hand operand <var>A</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>k</var></sup>
   *
   * @returns
   *         product <var>v</var><sup>T</sup>&sdot;<var>A</var> &isin;
   *         <var>K</var><sup>n</var></sup> of the vector and the matrix
   *
   * @throws std::invalid_argument
   *         if the vector's dimension is different from the number of rows in
   *         the matrix
   *
   */
  template<typename T>
  FullVector<T>
  operator*(const HVector<T>& left, const HMatrix<T>& right)
  {
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (left.size() != right.rows())
          {
            std::ostringstream oss {};
            oss << "cannot multiply a "
                << left.size()
                << " vector with a "
                << "(" << right.rows() << ", " << right.cols() << ")"
                << " matrix";
            throw std::invalid_argument {oss.str()};
          }
      }
    // This shortcut is correct since we do not distinguish between row- and
    // column-vectors.  Hence, v A = (v A)^T = A^T v^T = A^T v.
    return ConstMatrixView<T>::makeTransposedView(right) * left;
  }

  /**
   * @brief
   *         Adds two `HVector`s and returns the result in a new `HVector`.
   *
   * @param left
   *         left-hand operand <var>u</var> &isin;
   *         <var>K</var><sup><var>n</var></sup>
   *
   * @param right
   *         right-hand operand <var>v</var> &isin;
   *         <var>K</var><sup><var>n</var></sup>
   *
   * @returns
   *         sum <var>u</var> + <var>v</var> &isin;
   *         <var>K</var><sup><var>n</var></sup> of the two vectors
   *
   * @throws std::invalid_argument
   *         if the two vectors have different dimensions
   *
   */
  template<typename T>
  FullVector<T>
  operator+(const HVector<T>& left, const HVector<T>& right)
  {
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (left.size() != right.size())
          HVector<T>::throwMismatchedDimensions("add", left, "and", right);
      }
    {
      FullVector<T> sum {left};
      sum += right;
      return sum;
    }
  }

  /**
   * @brief
   *         Subtracts two `HVector`s and returns the result in a new
   *         `HVector`.
   *
   * @param left
   *         left-hand operand <var>u</var> &isin;
   *         <var>K</var><sup><var>n</var></sup>
   *
   * @param right
   *         right-hand operand <var>v</var> &isin;
   *         <var>K</var><sup><var>n</var></sup>
   *
   * @returns
   *         difference <var>u</var> &minus; <var>v</var> &isin;
   *         <var>K</var><sup><var>n</var></sup> of the two vectors
   *
   * @throws std::invalid_argument
   *         if the two vectors have different dimensions
   *
   */
  template<typename T>
  FullVector<T>
  operator-(const HVector<T>& left, const HVector<T>& right)
  {
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (left.size() != right.size())
          HVector<T>::throwMismatchedDimensions("subtract", left, "and", right);
      }
    {
      FullVector<T> diff {left};
      diff -= right;
      return diff;
    }
  }

  /**
   * @brief
   *         Returns the 2-norm of a `HVector`.
   *
   * @param vec
   *         vector <var>v</var> &isin; <var>K</var><sup><var>n</var></sup>
   *
   * @returns
   *         2-norm |<var>v</var>| &isin; &real;<sup>+</sup> of the vector
   *
   * @throws std::invalid_argument
   *         if the vector has zero dimension
   *
   */
  template<typename T>
  T
  norm(const HVector<T>& vec)
  {
    using value_type = typename HMatrix<T>::value_type;
    value_type dot {vec * vec};
    assert(dot >= static_cast<T>(0));
    return std::sqrt(dot);
  }

  /**
   * @brief
   *         Computes the inverse of a regular square `HMatrix` and returns the
   *         reult in a new `FullMatrix`.
   *
   * If the matrix is not regular, undefined behaviour will result.
   *
   * @param matrix
   *         regular square matrix <var>A</var> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>n</var></sup>
   *
   * @returns
   *         inverse matrix <var>A</var><sup>&minus;1</sup> &isin;
   *         <var>K</var><sup><var>n</var>&times;<var>n</var></sup>
   *
   * @throws std::invalid_argument
   *         if the matrix is not square
   *
   */
  template<typename T>
  FullMatrix<T>
  invert(const HMatrix<T>& matrix, const bool spd = false)
  {
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (matrix.rows() != matrix.cols())
          throw std::invalid_argument {"matrix is not square"};
      }
    if (auto fm_ptr = dynamic_cast<const FullMatrix<T> *const>(&matrix))
      {
        return algorithms_hpp::inverter<T>::invert(*fm_ptr, spd);
      }
    else
      {
        BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
        return invert(FullMatrix<T> {matrix}, spd);
      }
  }

  /**
   * @brief
   *         Checks whether a `HMatrix` is all-zero.
   *
   * A matrix is all-zero if and only if it has no entries that are different
   * from zero.
   *
   * This check is looking at the actual data.  It does not use pre-existing
   * knowledge about the structure of the matrix.  Therefore, it is
   * computationally expensive.
   *
   * @param <T>
   *         type of matrix entries
   *
   * @param matrix
   *         matrix to check
   *
   * @returns
   *         whether the matrix is all-zero
   *
   */
  template<typename T>
  bool
  is_zero(const HMatrix<T>& matrix)
  {
    BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
    using index_type = typename HMatrix<T>::index_type;
    const index_type rows {matrix.rows()};
    const index_type cols {matrix.cols()};
    for (index_type j = 0; j < cols; ++j)
      for (index_type i = 0; i < rows; ++i)
        if (matrix.get(i, j) != HMatrix<T>::zero()) { return false; }
    return true;
  }

  /**
   * @brief
   *         Checks whether a `HMatrix` is (strictly) lower-triangular.
   *
   * A matrix is *lower-triangular* if and only if it has no entries above the
   * south-west diagonal that are different from zero.
   *
   * A matrix is *strictly lower-triangular* if and only if it has no entries
   * on or above the south-west diagonal that are different from zero.
   *
   * The *south-west diagonal* of a square matrix is the main diagonal.  For a
   * non-square matrix, it is the main diagonal of the largest possible square
   * sub-matrix located at the south-west corner of the matrix.
   *
   * This check is looking at the actual data.  It does not use pre-existing
   * knowledge about the structure of the matrix.  Therefore, it is
   * computationally expensive.
   *
   * @param <T>
   *         type of matrix entries
   *
   * @param matrix
   *         matrix to check
   *
   * @param strict
   *         whether to check for strict lower-triangularity
   *
   * @returns
   *         whether the matrix is (strictly) lower-triangular
   *
   */
  template<typename T>
  bool
  is_lower_triangular(const HMatrix<T>& matrix, const bool strict = true)
  {
    BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
    using index_type = typename HMatrix<T>::index_type;
    const index_type rows {matrix.rows()};
    const index_type cols {matrix.cols()};
    if (rows > cols)
      {
        const ConstMatrixView<T> top {&matrix, 0, 0, rows - cols, cols};
        const ConstMatrixView<T> bottom {&matrix, rows - cols, 0, cols, cols};
        return is_lower_triangular(bottom, strict) && is_zero(top);
      }
    else if (rows < cols)
      {
        const ConstMatrixView<T> left {&matrix, 0, 0, rows, rows};
        const ConstMatrixView<T> right {&matrix, 0, rows, rows, cols - rows};
        return is_lower_triangular(left, strict) && is_zero(right);
      }
    else
      {
        for (index_type j = 0; j < cols; ++j)
          for (index_type i = 0; strict ? (i <= j) : (i < j); ++i)
            if (matrix.get(i, j) != HMatrix<T>::zero()) { return false; }
        return true;
      }
  }

  /**
   * @brief
   *         Checks whether a `HMatrix` is (strictly) upper-triangular.
   *
   * A matrix is *upper-triangular* if and only if it has no entries below the
   * north-east diagonal that are different from zero.
   *
   * A matrix is *strictly upper-triangular* if and only if it has no entries
   * on or below the north-east diagonal that are different from zero.
   *
   * The *north-east diagonal* of a square matrix is the main diagonal.  For a
   * non-square matrix, it is the main diagonal of the largest possible square
   * sub-matrix located at the north-east corner of the matrix.
   *
   * This check is looking at the actual data.  It does not use pre-existing
   * knowledge about the structure of the matrix.  Therefore, it is
   * computationally expensive.
   *
   * @param <T>
   *         type of matrix entries
   *
   * @param matrix
   *         matrix to check
   *
   * @param strict
   *         whether to check for strict upper-triangularity
   *
   * @returns
   *         whether the matrix is (strictly) upper-triangular
   *
   */
  template<typename T>
  bool
  is_upper_triangular(const HMatrix<T>& matrix, const bool strict = true)
  {
    BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
    using index_type = typename HMatrix<T>::index_type;
    const index_type rows {matrix.rows()};
    const index_type cols {matrix.cols()};
    if (rows > cols)
      {
        const ConstMatrixView<T> top {&matrix, 0, 0, cols, cols};
        const ConstMatrixView<T> bottom {&matrix, cols, 0, rows - cols, cols};
        return is_upper_triangular(top, strict) && is_zero(bottom);
      }
    else if (rows < cols)
      {
        const ConstMatrixView<T> left {&matrix, 0, 0, rows, cols - rows};
        const ConstMatrixView<T> right {&matrix, 0, cols - rows, rows, rows};
        return is_upper_triangular(right, strict) && is_zero(left);
      }
    else
      {
        for (index_type j = 0; j < cols; ++j)
          for (index_type i = strict ? j : j + 1; i < rows; ++i)
            if (matrix.get(i, j) != HMatrix<T>::zero()) { return false; }
        return true;
      }
  }

  /**
   * @brief
   *         Checks whether a `HMatrix` is symmetric.
   *
   * A square (<var>n</var>, <var>n</var>) matrix <var>M</var> is *symmetric*
   * if and only if <var>M</var><sub><var>ij</var></sub> =
   * <var>M</var><sub><var>ji</var></sub> for all <var>i</var>, <var>j</var>
   * &isin; 0, 1, &hellip; <var>n</var> &minus; 1.
   *
   * This check is looking at the actual data.  It does not use pre-existing
   * knowledge about the structure of the matrix.  Therefore, it is
   * computationally expensive.
   *
   * @param <T>
   *         type of matrix entries
   *
   * @param matrix
   *         matrix to check
   *
   * @returns
   *         whether the matrix is symmetric
   *
   * @throws std::invalid_argument
   *         if the matrix is not square
   *
   */
  template<typename T>
  bool
  is_symmetric(const HMatrix<T>& matrix)
  {
    BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
    using index_type = typename HMatrix<T>::index_type;
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (matrix.rows() != matrix.cols())
          throw std::invalid_argument {"matrix is not square"};
      }
    for (index_type j = 0; j < matrix.size(); ++j)
      for (index_type i = 0; i < j; ++i)
        if (matrix.get(i, j) != matrix.get(j, i)) { return false; }
    return true;
  }

  /**
   * @brief
   *         Checks whether a `HMatrix` is positive definite.
   *
   * A square (<var>n</var>, <var>n</var>) matrix <var>M</var> is *positive
   * definite* if and only if all of its eigenvalues are real and positive.
   *
   * This check is looking at the actual data.  It does not use pre-existing
   * knowledge about the structure of the matrix.  Therefore, it is
   * computationally *very* expensive as it computes all eigenvalues of the
   * matrix.  If you want to test whether a matrix is both, symmetric and
   * positive definite (spd), apply the checks in this order.
   *
   * @param <T>
   *         type of matrix entries
   *
   * @param matrix
   *         matrix to check
   *
   * @returns
   *         whether the matrix is positive definite
   *
   * @throws std::invalid_argument
   *         if the matrix is not square
   *
   */
  template<typename T>
  bool
  is_positive_definite(const HMatrix<T>& matrix)
  {
    using full_matrix = FullMatrix<T>;
    using raw_matrix = typename full_matrix::raw_matrix_type;
    using raw_index_type = typename full_matrix::raw_index_type;
    BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (matrix.rows() != matrix.cols())
          throw std::invalid_argument {"matrix is not square"};
      }
    if (matrix.size() > 0)
      {
        Eigen::EigenSolver<raw_matrix> eigensolver {};
        if (auto full_ptr = dynamic_cast<const full_matrix *const>(&matrix))
          {
            eigensolver.compute(full_ptr->getRawMatrix());
          }
        else
          {
            const full_matrix full {matrix};
            eigensolver.compute(full.getRawMatrix());
          }
        {
          const auto eigenvalues = eigensolver.eigenvalues();
          for (raw_index_type i = 0; i < eigenvalues.size(); ++i)
            {
              const auto iter = eigenvalues(i);
              if (iter.imag() != 0.0)
                return false;
              if (iter.real() <= 0.0)
                return false;
            }
        }
      }
    return true;
  }

  namespace algorithms_hpp
  {

    /**
     * @brief
     *         Visitor that allows to access the first block (region) it was
     *         passed over.
     *
     * After having been passed over a region of a `HMatrix`, an instance of
     * this class will provide access to the very first sub-region it was
     * passed over via its `anchor` and `rectangle` data members.  The former
     * stores a pointer to the region's matrix and its north west coordinates
     * (that is, those of the region, which need not coincide with that of the
     * matrix) in the &ldquo;global context&rdquo;.  The latter holds the
     * indices defining the region in the aforementioned matrix.
     *
     * @param <T>
     *         type of the matrix elements
     *
     * @param <M>
     *         whether the extracted matrix needs to be mutable
     *
     */
    template<typename T, bool M>
    struct FirstBlockExtractor : public std::conditional<M, typename HMatrix<T>::ActiveVisitor, typename HMatrix<T>::PassiveVisitor>::type
    {

      /** @brief Type for indices. */
      using index_type = typename HMatrix<T>::index_type;

      /** @brief Type for elements. */
      using value_type = typename HMatrix<T>::value_type;

      /**
       * @brief
       *         Number of blocks (regions) that were shown to this visitor.
       *
       * The `anchor`, `rectangle` and `trans` members will only be valid if
       * this counter is at least one and only useful if it is at most one.
       *
       */
      int block_count {0};

      /**
       * @brief
       *         Pointer to the first region's matrix and the `global_north`
       *         and `global_west` coordinates (of the region, not the matrix).
       *
       */
      MatrixAnchor<T, M> anchor {};

      /**
       * @brief
       *         Relevant rectangle in the extracted matrix.
       *
       */
      Rectangle<index_type> rectangle {};

      /**
       * @brief
       *         Indicates whether the matrix pointed to by the `anchor` needs
       *         to be transposed before doing anything.
       *
       */
      bool trans {};

      /**
       * @brief
       *         If called for the first time, sets the `anchor`, `rectangle`
       *         and `trans` members.  Always increments the `block_count`.
       *
       */
      virtual void
      visitRegion(const index_type global_north,
                  const index_type global_west,
                  const bool global_trans,
                  typename const_if<HMatrix<T>, !M>::type& matrix,
                  const index_type north,
                  const index_type west,
                  const index_type south,
                  const index_type east) final override
        {
          if (this->block_count++)
            return;
          this->anchor = MatrixAnchor<T, M> {&matrix, global_north, global_west};
          this->rectangle = Rectangle<index_type> {north, west, south, east};
          this->trans = global_trans;
        }
    };

    /**
     * @brief
     *         Most inefficient fallback implementation to add or subtract
     *         (part of) a `HMatrix` to an other `HMatrix`.
     *
     * All parameters have the same meaning as for the `take_delta()` function.
     *
     * Don't call this function directly; it is only needed by the
     * implementation of the `take_delta()` function.
     *
     */
    template<char Op, typename T>
    void
    dumb_take_delta(HMatrix<T>& augend,
                    const typename HMatrix<T>::index_type aug_north,
                    const typename HMatrix<T>::index_type aug_west,
                    const HMatrix<T>& addend,
                    const bool add_trans,
                    const typename HMatrix<T>::index_type add_north,
                    const typename HMatrix<T>::index_type add_west,
                    const typename HMatrix<T>::index_type add_south,
                    const typename HMatrix<T>::index_type add_east)
    {
      static_assert(Op == '+' || Op == '-',
                    "template parameter 'Op' must be '+' or '-'");
      using index_type = typename HMatrix<T>::index_type;
      using value_type = typename HMatrix<T>::value_type;
      BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
      for (index_type j = add_west; j < add_east; ++j)
        {
          for (index_type i = add_north; i < add_south; ++i)
            {
              const index_type aug_row {aug_north + i - add_north};
              const index_type aug_col {aug_west  + j - add_west};
              const index_type add_row {!add_trans ? i : j};
              const index_type add_col {!add_trans ? j : i};
              const value_type diff {addend.get(add_row, add_col)};
              if (diff != HMatrix<T>::zero())
                {
                  value_type v {augend.get(aug_row, aug_col)};
                  switch (Op)
                    {
                    case '+': v += diff; break;
                    case '-': v -= diff; break;
                    default: BSC_KLAMMLER_NOT_REACHED;
                    }
                  augend.set(v, aug_row, aug_col);
                }
            }
        }
    }

    template<char Op, typename T>
    void
    fast_take_delta(ArrangedMatrix<T>& augend,
                    const typename HMatrix<T>::index_type aug_north,
                    const typename HMatrix<T>::index_type aug_west,
                    const HMatrix<T>& addend,
                    const bool add_trans,
                    const typename HMatrix<T>::index_type add_north,
                    const typename HMatrix<T>::index_type add_west,
                    const typename HMatrix<T>::index_type add_south,
                    const typename HMatrix<T>::index_type add_east)
    {
      static_assert(Op == '+' || Op == '-',
                    "template parameter 'Op' must be '+' or '-'");
      if (add_south - add_north == 0 && add_east - add_west == 0)
        return;
      std::unique_ptr<HMatrix<T>> new_tile {};
      if (!add_trans
          && add_north == 0             && add_west == 0
          && add_south == addend.rows() && add_east == addend.cols())
        {
          try
            {
              new_tile = addend.clone();
            }
          catch (const NotImplementedError&)
            {
              assert(!new_tile);
            }
        }
      if (!new_tile)
        {
          const ConstMatrixView<T> add {
            &addend,
            add_north,
            add_west,
            add_south - add_north,
            add_east - add_west,
            add_trans,
          };
          new_tile.reset(new FullMatrix<T> {add});
        }
      assert(new_tile);
      if (Op == '-')
        *new_tile *= static_cast<T>(-1);
      augend.addTile(std::move(new_tile), aug_north, aug_west);
    }

    // See DocString at the declaration near the beginning of this file.
    template<char Op, typename T>
    void
    take_delta(HMatrix<T>& augend,
               const typename HMatrix<T>::index_type aug_north,
               const typename HMatrix<T>::index_type aug_west,
               const HMatrix<T>& addend,
               const bool add_trans,
               const typename HMatrix<T>::index_type add_north,
               const typename HMatrix<T>::index_type add_west,
               const typename HMatrix<T>::index_type add_south,
               const typename HMatrix<T>::index_type add_east)
    {
      static_assert(Op == '+' || Op == '-',
                    "template parameter 'Op' must be '+' or '-'");
      using index_type = typename HMatrix<T>::index_type;
      using value_type = typename HMatrix<T>::value_type;
      using full_matrix = FullMatrix<value_type>;
      using rectangle_t = Rectangle<index_type>;
      const rectangle_t add {
        add_north, add_west, add_south, add_east
      };
      const rectangle_t aug {
        aug_north, aug_west, aug_north + add.height(), aug_west + add.width()
      };
      assert(add.width() == aug.width());
      assert(add.height() == aug.height());
      // We need a nested scope we can break out of so that we don't jump over
      // initializations.
      {
        FirstBlockExtractor<value_type, true> fbe {};
        pass_over_region(augend, add.north(), add.west(), false, fbe, aug);
        if (fbe.block_count == 0)
          {
            if (auto arr_aug_ptr = dynamic_cast<ArrangedMatrix<T> *>(&augend))
              return fast_take_delta<Op>(
                  *arr_aug_ptr, aug_north, aug_west,
                  addend, add_trans, add_north, add_west, add_south, add_east);
          }
        if (fbe.block_count != 1)
          goto dumb;
        assert(fbe.anchor.getMatrix() != nullptr);
        auto full_aug_ptr = dynamic_cast<full_matrix *>(fbe.anchor.getMatrix());
        if (full_aug_ptr == nullptr)
          goto dumb;
        auto full_add_ptr = dynamic_cast<const full_matrix *>(&addend);
        if (full_add_ptr == nullptr)
          goto dumb;
        if (fbe.rectangle.height() != aug.height())
          goto dumb;
        if (fbe.rectangle.width() != aug.width())
          goto dumb;
        switch (Op + (lxor(add_trans, fbe.trans) ? 'T' : 0))
          {
          case '+':
            full_aug_ptr->getRawMatrix().block(
                fbe.rectangle.north(), fbe.rectangle.west(),
                fbe.rectangle.height(), fbe.rectangle.width())
              += full_add_ptr->getRawMatrix().block(
                     add.north(), add.west(), add.height(), add.width());
            break;
          case '-':
            full_aug_ptr->getRawMatrix().block(
                fbe.rectangle.north(), fbe.rectangle.west(),
                fbe.rectangle.height(), fbe.rectangle.width())
              -= full_add_ptr->getRawMatrix().block(
                     add.north(), add.west(), add.height(), add.width());
            break;
          case 'T' + '+':
            full_aug_ptr->getRawMatrix().block(
                fbe.rectangle.north(), fbe.rectangle.west(),
                fbe.rectangle.height(), fbe.rectangle.width())
              += full_add_ptr->getRawMatrix().transpose().block(
                     add.north(), add.west(), add.height(), add.width());
            break;
          case 'T' + '-':
            full_aug_ptr->getRawMatrix().block(
                fbe.rectangle.north(), fbe.rectangle.west(),
                fbe.rectangle.height(), fbe.rectangle.width())
              -= full_add_ptr->getRawMatrix().transpose().block(
                     add.north(), add.west(), add.height(), add.width());
            break;
          default:
            BSC_KLAMMLER_NOT_REACHED;
          }
        return;
      }
    dumb:
      return dumb_take_delta<Op>(
          augend, aug_north, aug_west,
          addend, add_trans, add_north, add_west, add_south, add_east);
    }

    /**
     * @brief
     *         Extracts a `FullMatrix` out of a polymorphic `HMatrix` that is
     *         sufficiently simple.
     *
     * By &ldquo;sufficiently simple&rdquo; we mean that the `HMatrix` must
     * be representible by a region of a `FullMatrix` it contains.
     *
     * Examples of `HMatrix`es that are sufficiently simple:
     *
     * - `FullMatrix`es
     * - `MatrixView`s of sufficiently simple `HMatrix`es
     * - `ArrangedMatrix`es with exactly one tile that is a sufficiently simple
     *   `HMatrix` at offset (0, 0)
     *
     * Examples of `HMatrix`es that are not simple enough:
     *
     * - empty `MatrixView`s or `ArrangedMatrix`es
     * - `ArrangedMatrix` with more than one tile
     * - `ArrangedMatrix` with a tile at an offset different from (0, 0)
     *
     * Note that creating a `MatrixView` of a region of an `ArrangedMatrix`
     * with more tile that only overlaps with one tile makes the resulting
     * `MatrixView` sufficiently simple again.
     *
     * If the extraction succeeds, a pointer to the `FullMatrix` is returned
     * and <code>*<var>rect_ptr</var></code> is set to the rectangle of said
     * `FullMatrix` that holds the data visible in the simple `HMatrix`.
     * <code>*<var>trans_ptr</var></code> is set to indicate whether the
     * contents of the returned `FullMatrix` need to be transposed.  If so, the
     * rectangle is measured in the already transposed matrix.
     *
     * @param simple_matrix
     *         sufficiently simple `HMatrix` to have its only `FullMatrix`
     *         extracted
     *
     * @param rect_ptr
     *         address to store the rectangle visible from the `FullMatrix` at
     *
     * @param trans_ptr
     *         address to store the flag whether the `FullMatrix` needs to be
     *         transposed
     *
     * @returns
     *         a pointer to the `FullMatrix` or a `nullptr` if extraction was
     *         not possible
     *
     */
    template<typename T>
    const FullMatrix<T> *
    extract_full(const HMatrix<T>& simple_matrix,
                 Rectangle<typename HMatrix<T>::index_type> *const rect_ptr,
                 bool *const trans_ptr)
    {
      assert(rect_ptr != nullptr);
      assert(trans_ptr != nullptr);
      FirstBlockExtractor<T, false> vtor {};
      pass_over_region(simple_matrix, 0, 0, false, vtor);
      if (vtor.block_count == 1
          && vtor.anchor.north() == 0
          && vtor.anchor.west() == 0)
        {
          if (auto full_ptr = dynamic_cast<const FullMatrix<T> *>(vtor.anchor.getMatrix()))
            {
              *rect_ptr = vtor.rectangle;
              *trans_ptr = vtor.trans;
              return full_ptr;
            }
        }
      return nullptr;
    }

    // See DocString at the declaration near the beginning of this file.
    template<typename T>
    std::unique_ptr<HMatrix<T>>
    poly_prod(const HMatrix<T>& left, const HMatrix<T>& right)
    {
      using index_type = typename HMatrix<T>::index_type;
      if (BSC_KLAMMLER_CHECK_ARGS)
        algorithms_hpp::check_dims_prod(left, right);
      std::unique_ptr<HMatrix<T>> prod_uptr {};
      if (!prod_uptr)
        {
          // Handle the trivial case first (and correctly).
          if (left.rows() == 0 || right.cols() == 0)
            prod_uptr.reset(new FullMatrix<T> {left.rows(), right.cols()});
        }
      if (!prod_uptr)
        {
          // See if we can extract two 'FullMatrix'es and perfom fast dense
          // multiplication.
          Rectangle<index_type> rect_left {};
          Rectangle<index_type> rect_right {};
          bool trans_left {};
          bool trans_right {};
          auto full_left_ptr = extract_full(left, &rect_left, &trans_left);
          auto full_right_ptr = extract_full(right, &rect_right, &trans_right);
          if (full_left_ptr != nullptr && full_right_ptr != nullptr)
            {
              // We'd rather not introduce temporaries so we cannot save the
              // (transposed) block expressions into a variable.  Hence the
              // somewhat verbose case selection.
              switch ((trans_left ? 'L' : 0) + (trans_right ? 'R' : 0))
                {
                case 0:
                  prod_uptr.reset(new FullMatrix<T> {
                      full_left_ptr->getRawMatrix().block(
                          rect_left.north(), rect_left.west(),
                          rect_left.height(), rect_left.width())
                        *
                      full_right_ptr->getRawMatrix().block(
                          rect_right.north(), rect_right.west(),
                          rect_right.height(), rect_right.width())
                  });
                  break;
                case 'L':
                  prod_uptr.reset(new FullMatrix<T> {
                      full_left_ptr->getRawMatrix().transpose().block(
                          rect_left.north(), rect_left.west(),
                          rect_left.height(), rect_left.width())
                        *
                      full_right_ptr->getRawMatrix().block(
                          rect_right.north(), rect_right.west(),
                          rect_right.height(), rect_right.width())
                  });
                  break;
                case 'R':
                  prod_uptr.reset(new FullMatrix<T> {
                      full_left_ptr->getRawMatrix().block(
                          rect_left.north(), rect_left.west(),
                          rect_left.height(), rect_left.width())
                        *
                      full_right_ptr->getRawMatrix().transpose().block(
                          rect_right.north(), rect_right.west(),
                          rect_right.height(), rect_right.width())
                  });
                  break;
                case 'L' + 'R':
                  prod_uptr.reset(new FullMatrix<T> {
                      full_left_ptr->getRawMatrix().transpose().block(
                          rect_left.north(), rect_left.west(),
                          rect_left.height(), rect_left.width())
                        *
                      full_right_ptr->getRawMatrix().transpose().block(
                          rect_right.north(), rect_right.west(),
                          rect_right.height(), rect_right.width())
                  });
                  break;
                default:
                  BSC_KLAMMLER_NOT_REACHED;
                }
              assert(prod_uptr->rows() == left.rows());
              assert(prod_uptr->cols() == right.cols());
            }
        }
      if (!prod_uptr)
        {
          // Damn it, we cannot handle this efficiently!
          BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
          // std::clog << BSC_KLAMMLER_LABEL
          //           << "left is a"
          //           << " (" << left.rows() << ", " << left.cols() << ") "
          //           << typeid(left).name() << " matrix" << BSC_KLAMMLER_EOL;
          // std::clog << BSC_KLAMMLER_LABEL
          //           << "right is a"
          //           << " (" << right.rows() << ", " << right.cols() << ") "
          //           << typeid(right).name() << " matrix" << BSC_KLAMMLER_EOL;
          prod_uptr.reset(new FullMatrix<T> {dumb_prod(left, right)});
        }
      return prod_uptr;
    }

    // See DocString at the declaration near the beginning of this file.
    template<typename T>
    FullMatrix<T>
    dumb_prod(const HMatrix<T>& left, const HMatrix<T>& right)
    {
      using index_type = typename HMatrix<T>::index_type;
      using value_type = typename HMatrix<T>::value_type;
      BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
      if (BSC_KLAMMLER_CHECK_ARGS)
        algorithms_hpp::check_dims_prod(left, right);
      FullMatrix<T> prod {left.rows(), right.cols()};
      for (index_type j = 0; j < prod.cols(); ++j)
        {
          for (index_type i = 0; i < prod.rows(); ++i)
            {
              value_type acc {};
              for (index_type k = 0; k < left.cols(); ++k)
                {
                  acc += left.get(i, k) * right.get(k, j);
                }
              prod.set(acc, i, j);
            }
        }
      return prod;
    }
  }  // namespace algorithms_hpp
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_ALGORITHMS_HPP
