/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file arranged_matrix.hpp
 *
 * @brief
 *         Header file providing the `ArrangedMatrix` template class.
 *
 */

#ifndef BSC_KLAMMLER_ARRANGED_MATRIX_HPP
#define BSC_KLAMMLER_ARRANGED_MATRIX_HPP

#include <algorithm>            // std::sort
#include <cassert>              // assert
#include <cstddef>              // std::size_t
#include <functional>           // std::function
#include <memory>               // std::unique_ptr
#include <sstream>              // std::ostringstream
#include <stdexcept>            // std::invalid_argument
#include <utility>              // std::swap
#include <vector>               // std::vector

#include "h_matrix.hpp"         // bsc_klammler::HMatrix
#include "macros.hpp"           // (package-specific macros)
#include "matrix_anchor.hpp"    // bsc_klammler::MatrixAnchor
#include "rectangle.hpp"        // bsc_klammler::Rectangle


namespace bsc_klammler
{


  /**
   * @brief
   *         A matrix composed of arbitrairly arranged non-zero blocks.
   *
   * An `ArrangedMatrix` consists of a *frame* (all zero) that defines the
   * matrix size and an arbitrary number of *tiles* freely placed inside the
   * frame.  The tiles itself may be any `HMatrix` type.
   *
   * An `ArrangedMatrix` takes full ownership of its tiles which must be
   * allocated on the free store using operator `new` (perhaps indirectly via
   * `HMatrix::clone`).  &ldquo;Hide&rdquo; the actual matrix inside a
   * `MatrixView` if you wish to avoid a deep copy be made (and know exactly
   * what you are doing).
   *
   * There are three special cases of `ArrangedMatrix`es identified as
   * *row-linear*, *column-linear* and *diagonal*.
   *
   * A *row-linear* `ArrangedMatrix` has only tiles that fill the complete
   * height of the matrix.  It might look like this:
   *
   *     +----+---+----------+---+--------+---+
   *     |    | A |          | B |        | C |
   *     +----+---+----------+---+--------+---+
   *
   * A *column-linear* matrix has only tiles that fill the complete width of
   * the matrix.  It might look like this:
   *
   *     +-------+
   *     |       |
   *     |   A   |
   *     |       |
   *     +-------+
   *     |       |
   *     |       |
   *     +-------+
   *     |   B   |
   *     +-------+
   *     |       |
   *     |       |
   *     +-------+
   *
   * A *diagonal* matrix is square and has only tiles anchored at the main
   * diagonal.  It might look like this:
   *
   *     +---+----------------+
   *     | A |                |
   *     +---+-------+        |
   *     |   |       |        |
   *     |   |   B   |        |
   *     |   |       |        |
   *     |   +-------+        |
   *     |                    |
   *     |              +---+ |
   *     |              | C | |
   *     |              +---+ |
   *     +--------------------+
   *
   * These forms allow for more efficient matrix operations.
   *
   * An `ArrangedMatrix` is row-linear if and only if its transposed is
   * column-linear and vice versa.  An empty matrix is both row- and
   * column-linear -- and, if it is square, also diagonal -- in a trivial sense
   * as is a (square) `ArrangedMatrix` with only one tile filling its entire
   * size.  `ArrangedMatrix`es start out row- and column-linear (and maybe
   * diagonal) and might loose either property as tiles continue to be added.
   *
   * Unlike most other `HMatrix` classes, `ArrangedMatrix`es can dynamically
   * grow as tiles are added.  However, this feature is disabled by default.
   *
   * @param <T>
   *         type of the matrix entries
   *
   */
  template<typename T>
  class ArrangedMatrix : public HMatrix<T>
  {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Type of the super-class. */
    using super_type = HMatrix<T>;

    /** @brief Type for `ArrangedMatrix` elements. */
    using value_type = typename super_type::value_type;

    /** @brief Type for `ArrangedMatrix` indices. */
    using index_type = typename super_type::index_type;

    /** @brief Convenience type alias for mutable `MatrixAnchor`s. */
    using matrix_anchor = MatrixAnchor<value_type, true>;

    /** @brief Convenience type alias for const `MatrixAnchor`s. */
    using const_matrix_anchor = MatrixAnchor<value_type, false>;

    /** @brief Imported type for passive visitors. */
    using PassiveVisitor = typename super_type::PassiveVisitor;

    /** @brief Imported type for active visitors. */
    using ActiveVisitor = typename super_type::ActiveVisitor;

  private:

    /** @brief Convenience alias for `struct region_pass`. */
    using region_pass = typename super_type::region_pass;


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief Number of rows in this `ArrangedMatrix`. */
    index_type rows_ {0};

    /** @brief Number of columns in this `ArrangedMatrix`. */
    index_type cols_ {0};

    /** @brief Whether this `ArrangedMatrix` is row-linear. */
    bool rowLinear_ {true};

    /** @brief Whether this `ArrangedMatrix` is column-linear. */
    bool colLinear_ {true};

    /** @brief Whether this `ArrangedMatrix` is diagonal. */
    bool diagonal_ {true};

    /** @brief The tiles that are arranged inside this `ArrangedMatrix`. */
    std::vector<matrix_anchor> tiles_ {};


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    /**
     * @brief
     *         Default Constructor: Creates a new empty `ArrangedMatrix`.
     *
     */
    ArrangedMatrix() noexcept
    {
      assert(this->rows_ == 0);
      assert(this->cols_ == 0);
      assert(this->tiles_.empty());
      assert(this->classInvariantsHold_());
    }

    /**
     * @brief
     *         Constructor: Creates a new `ArrangedMatrix` with the given
     *         dimensions and no tiles.
     *
     * @param rows
     *         number of rows
     *
     * @param cols
     *         number of columns
     *
     * @throws std::invalid_argument
     *         if argument checking is enabled and the matrix dimensions are
     *         ill-defined
     *
     */
    ArrangedMatrix(const index_type rows, const index_type cols) :
      rows_ {rows}, cols_ {cols}, diagonal_ {rows == cols}
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        {
          super_type::ctorCheckSize(rows, cols);
        }
      assert(this->classInvariantsHold_());
    };

    /**
     * @brief
     *         Copy constructor: Creates a new `ArrangedMatrix` that is a deep
     *         clone of an existing one.
     *
     * @param other
     *         `ArrangedMatrix` to be cloned
     *
     */
    ArrangedMatrix(const ArrangedMatrix& other) :
      rows_ {other.rows_},
      cols_ {other.cols_},
      rowLinear_ {other.rowLinear_},
      colLinear_ {other.colLinear_},
      diagonal_ {other.diagonal_}
    {
      assert(this->tiles_.size() == 0);
      this->tiles_.reserve(other.tiles_.size());
      for (const auto& iter : other.tiles_)
        {
          this->tiles_.emplace_back(
              iter.getMatrix()->clone().release(),
              iter.north(),
              iter.west());
        }
      assert(this->tiles_.size() == other.tiles_.size());
      assert(this->classInvariantsHold_());
    }

    /**
     * @brief
     *         Move constructor: Creates a new `ArrangedMatrix` by canibalizing
     *         an existing one.
     *
     * @param other
     *         `ArrangedMatrix` to be moved away from
     *
     */
    ArrangedMatrix(ArrangedMatrix&& other) noexcept : ArrangedMatrix {}
    {
      swap(*this, other);
    }

    /**
     * @brief
     *         Destructor: Disposes of this `ArrangedMatrix` and releases all
     *         resources.
     *
     */
    virtual ~ArrangedMatrix() noexcept
    {
      assert(this->classInvariantsHold_());
      for (const auto& iter : this->tiles_)
        delete iter.getMatrix();
    }


    /* ------------- */
    /*   Operators   */
    /* ------------- */

    /**
     * @brief
     *         Copy-assignment operator: Makes this `ArrangedMatrix` become a
     *         deep clone of an existing one.
     *
     * Data previously owned by this `ArrangedMatrix` (if any) is freed.
     *
     * @param other
     *         `ArrangedMatrix` to be cloned into this one
     *
     */
    ArrangedMatrix&
    operator=(const ArrangedMatrix& other)
    {
      ArrangedMatrix temp {other};
      swap(*this, temp);
      return *this;
    }

    /**
     * @brief
     *         Move-assignment operator: Moves all resources from an existing
     *         `ArrangedMatrix` -- that is canibalized during the process --
     *         into this one.
     *
     * Data previously owned by this `ArrangedMatrix` (if any) is freed.
     *
     * @param other
     *         `ArrangedMatrix` to be moved away from
     *
     */
    ArrangedMatrix&
    operator=(ArrangedMatrix&& other) noexcept
    {
      ArrangedMatrix temp {};  // noexcept
      swap(*this, temp);
      swap(*this, other);
      return *this;
    }

    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  public:

    /**
     * @brief
     *         Adds a tile to this `ArrangedMatrix`, transferring full
     *         ownership.
     *
     * @param matrix_uptr
     *         data for the tile, allocated an the free store
     *
     * @param off_row
     *         row index of the tile's (0, 0) element in the frame
     *
     * @param off_col
     *         column index of the tile's (0, 0) element in the frame
     *
     * @param resize
     *         whether to adjust the size of the frame to accomodate the new
     *         tile
     *
     * @throws std::invalid_argument
     *         if argument checking is enabled and the tile would overlap with
     *         an existing, or not fit the current frame without resizing
     *         and <code><var>resize</var> == false</code> or is a `nullptr`
     *
     */
    void
    addTile(std::unique_ptr<super_type> matrix_uptr,
            const index_type off_row,
            const index_type off_col,
            const bool resize = false)
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        {
          if (!matrix_uptr)
            throw std::invalid_argument {"matrix is a nullptr"};
            this->checkTile_(
                const_matrix_anchor {matrix_uptr.get(), off_row, off_col},
                resize);
        }
      // Only 'get()' but don't 'release()' the pointer yet.
      const matrix_anchor tile {matrix_uptr.get(), off_row, off_col};
      if (resize)
        {
          if (tile.south() > this->rows_)
            this->rows_ = tile.south();
          if (tile.east() > this->cols_)
            this->cols_ = tile.east();
          if (this->rows_ != this->cols_)
            this->diagonal_ = false;
        }
      // Don't waste work & space for empty matrices.
      // (Still consider them for the size, however!)
      if (tile.rows() > 0 && tile.cols() > 0)
        {
          if (tile.north() != 0 || tile.south() != this->rows_)
            this->rowLinear_ = false;
          if (tile.west() != 0 || tile.east() != this->cols_)
            this->colLinear_ = false;
          if (tile.rows() != tile.cols() || tile.north() != tile.west())
            this->diagonal_ = false;
          this->tiles_.push_back(tile);
          matrix_uptr.release();  //   <-------- here -------+
        }  //                                                |
      assert(this->classInvariantsHold_());  //              |
      // 'matrix_uptr' goes out of scope.  If we did not 'release()' the
      // matrix, it will be destroyed now.
    }

    /**
     * @brief
     *         Adds a tile to this `ArrangedMatrix`.
     *
     * The `HMatrix` will be `clone()`d and the clone added.
     *
     * @param matrix
     *         data for the tile
     *
     * @param off_row
     *         row index of the tile's (0, 0) element in the frame
     *
     * @param off_col
     *         column index of the tile's (0, 0) element in the frame
     *
     * @param resize
     *         whether to adjust the size of the frame to accomodate the new
     *         tile
     *
     * @throws std::invalid_argument
     *         if argument checking is enabled and the tile would overlap with
     *         an existing one or not fit the current frame without resizing
     *         and <code><var>resize</var> == false</code>
     *
     */
    void
    addTile(const super_type& matrix,
            const index_type off_row,
            const index_type off_col,
            const bool resize = false)
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        {
          // This performs the (rather lightweight) parameter validation twice
          // but does not waste (potentially very expensive) 'clone()'s on
          // invalid arguments.
          this->checkTile_(
              const_matrix_anchor {&matrix, off_row, off_col},
              resize);
        }
      this->addTile(matrix.clone(), off_row, off_col, resize);
    }

    // void
    // absorb(ArrangedMatrix&& other,
    //        const std::ptrdiff_t off_rows,
    //        const std::ptrdiff_t off_cols)
    // {
    //   std::vector<matrix_anchor> other_tiles {};
    //   try
    //     {
    //       other_tiles.swap(other.tiles_);
    //       other.rows_ = 0;
    //       other.cols_ = 0;
    //       other.rowLineear_ = true;
    //       other.colLineear_ = true;
    //       other.diagonal_ = true;
    //       assert(other.classInvariantsHold_());
    //       while (!other_tiles.empty())
    //         {
    //           // Lines marked with a '*' won't throw.
    //           const matrix_anchor tile {other_tiles.back()};  // (*)
    //           std::unique_ptr<super_type> holder {tile.getMatrix()};  // (*)
    //           other_tiles.pop_back();  // (*)
    //           this->addTile(std::move(holder),
    //                         tile.north() + off_rows,
    //                         tile.south() + off_cols);
    //           assert(holder.get() == nullptr);
    //         }
    //     }
    //   catch (const std::exception&)
    //     {
    //       for (const matrix_anchor tile : other_tiles)
    //         delete tile.getMatrix();
    //       throw;
    //     }
    //   assert(other_tiles.empty());
    // }

    /**
     * @brief
     *         Returns the current frame height, ie the number of rows in this
     *         `ArrangedMatrix`.
     *
     * @returns
     *         number of rows
     *
     */
    virtual index_type
    rows() const noexcept final override
    {
      return this->rows_;
    };

    /**
     * @brief
     *         Returns the current frame width, ie the number of columns in
     *         this `ArrangedMatrix`.
     *
     * @returns
     *         number of columns
     *
     */
    virtual index_type
    cols() const noexcept final override
    {
      return this->cols_;
    };

    /**
     * @brief
     *         Creates a deep clone of this `ArrangedMatrix`, allocated on the
     *         free store.
     *
     * The caller of this function obtains full ownership of the clone and is
     * responsible for `delete`ing it when it is no longer needed.
     *
     * @returns
     *         deep clone allocated on the free store
     *
     */
    virtual std::unique_ptr<super_type>
    clone() const final override
    {
      return std::unique_ptr<super_type> {new ArrangedMatrix {*this}};
    }

    /**
     * @brief
     *         Returns the value of the `ArrangedMatrix` element
     *         (<code><var>row</var></code>, <code><var>col</var></code>).
     *
     * @param row
     *         row index of the element
     *
     * @param col
     *         column index of the element
     *
     * @returns
     *         element's value
     *
     * @throws std::invalid_argument
     *         if the index is out of range
     */
    virtual value_type
    get(const index_type row, const index_type col) const final override
    {
      index_type i {static_cast<index_type>(-1)};
      index_type j {static_cast<index_type>(-1)};
      const super_type *const matrix_ptr {this->findIndex_(row, col, &i, &j)};
      return (matrix_ptr != nullptr)
        ? matrix_ptr->get(i, j)
        : super_type::zero();
    }

    /**
     * @brief
     *         Assigns <code><var>val</var></code> to the `ArrangedMatrix`
     *         element (<code><var>row</var></code>,
     *         <code><var>col</var></code>).
     *
     * @param val
     *         value to assign to the element
     *
     * @param row
     *         row index of the element
     *
     * @param col
     *         if the index is out of range or not inside a tile
     *
     */
    virtual void
    set(const value_type val,
        const index_type row,
        const index_type col) final override
    {
      index_type i {static_cast<index_type>(-1)};
      index_type j {static_cast<index_type>(-1)};
      super_type *const matrix_ptr {this->findIndex_(row, col, &i, &j)};
      if (BSC_KLAMMLER_CHECK_ARGS && matrix_ptr == nullptr)
        throw std::invalid_argument {"invalid attempt to write outside tile"};
      matrix_ptr->set(val, i, j);
    }

    // {@inheritDoc}
    virtual void
    passOverRegion(const index_type global_north,
                   const index_type global_west,
                   const bool global_trans,
                   PassiveVisitor& visitor,
                   const index_type north,
                   const index_type west,
                   const index_type south,
                   const index_type east) const final override
    {
      const region_pass rp {
        global_north, global_west, global_trans, north, west, south, east
      };
      ArrangedMatrix::applyVisitor_(*this, visitor, rp);
    }

    // {@inheritDoc}
    virtual void
    passOverRegion(const index_type global_north,
                   const index_type global_west,
                   const bool global_trans,
                   ActiveVisitor& visitor,
                   const index_type north,
                   const index_type west,
                   const index_type south,
                   const index_type east) final override
    {
      const region_pass rp {
        global_north, global_west, global_trans, north, west, south, east
      };
      ArrangedMatrix::applyVisitor_(*this, visitor, rp);
    }

    /**
     * @brief
     *         Returns whether this `ArrangedMatrix` is row-linear.
     *
     * @returns
     *         whether this `ArrangedMatrix` is row-linear
     *
     */
    bool
    isRowLinear() const
    {
      return this->rowLinear_;
    }

    /**
     * @brief
     *         Returns whether this `ArrangedMatrix` is column-linear.
     *
     * @returns
     *         whether this `ArrangedMatrix` is column-linear
     *
     */
    bool
    isColumnLinear() const
    {
      return this->colLinear_;
    }

    /**
     * @brief
     *         Returns whether this `ArrangedMatrix` is diagonal.
     *
     * @returns
     *         whether this `ArrangedMatrix` is diagonal
     *
     */
    bool
    isDiagonal() const
    {
      return this->diagonal_;
    }

    const std::vector<matrix_anchor>&
    tiles() const noexcept
    {
      return this->tiles_;
    }

    /**
     * @brief
     *         Sorts the tiles in this `ArrangedMatrix` west to east and
     *         then north to south.
     *
     */
    void
    sortTiles()
    {
      const auto cmp = [](const matrix_anchor ma1, const matrix_anchor ma2){
        return ma1.west() < ma2.west() || ma1.north() < ma2.north();
      };
      std::sort(this->tiles_.begin(), this->tiles_.end(), cmp);
    }

    using extractor_func = std::function<std::unique_ptr<super_type>(const super_type&, const Rectangle<index_type>, const bool)>;

    static ArrangedMatrix
    makeArrangement(const super_type& matrix,
                    const index_type north,
                    const index_type west,
                    const index_type south,
                    const index_type east,
                    const extractor_func& extract)
    {
      struct Visiotr : public PassiveVisitor
      {
        ArrangedMatrix am;
        const extractor_func& ext;

        Visiotr(const index_type r,
                const index_type c,
                const extractor_func& e) : am {r, c}, ext(e)
        // The round parenthesis for the initialization of the 'ext' member are
        // needed on some compilers (eg GCC < 4.9) due to a now-fixed issue in
        // the C++11 standard that would cause a initializer-list to create a
        // temporary copy.  See:
        // https://stackoverflow.com/questions/25561387/spurious-warning-about-binding-temporary-to-reference-member-in-constructor/25595974
        // We do not provide a "bug test macro" for this one since the
        // workaround is (except, perhaps, for stylistic reasons) equally good.
        // (GCC 4.9 creates identical assembly code.)
        {
        }

        virtual void
        visitRegion(const index_type vr_global_north,
                    const index_type vr_global_west,
                    const bool vr_global_trans,
                    const super_type& vr_matrix,
                    const index_type vr_north,
                    const index_type vr_west,
                    const index_type vr_south,
                    const index_type vr_east)
        {
          const Rectangle<index_type> rect {
            vr_north, vr_west, vr_south, vr_east
          };
          this->am.addTile(this->ext(vr_matrix, rect, vr_global_trans),
                           vr_global_north,
                           vr_global_west);
        }
      };
      Visiotr vtor {south - north, east- west, extract};
      matrix.passOverRegion(
          0, 0, false, vtor,
          north, west, south, east);
      return std::move(vtor.am);
    }

    static ArrangedMatrix
    makeArrangement(const super_type& matrix,
                    const extractor_func& extract)
    {
      return ArrangedMatrix::makeArrangement(
          matrix, 0, 0, matrix.rows(), matrix.cols(), extract);
    }

  private:

    /**
     * @brief
     *         Checks whether an attempt to insert a new tile would be valid.
     *
     * If the attempt is valid, this function does nothing.  Otherwise, an
     * exception is thrown.
     *
     * @param tile
     *         candidate to be added
     *
     * @param resize
     *         whether to adjust the size of the frame to accomodate the new
     *         tile
     *
     * @throws std::invalid_argument
     *         if the insertion would be invalid, that is, the tile overlaps
     *         an existing one or would not fit the matrix without disabled
     *         resizing
     *
     */
    void
    checkTile_(const const_matrix_anchor tile, const bool resize) const
    {
      if (tile.north() < 0 || tile.west() < 0)
        {
          std::ostringstream oss {};
          oss << "invalid anchor "
              << "(" << tile.north() << ", " << tile.west() << ")";
          throw std::invalid_argument {oss.str()};
        }
      if (!resize)
        {
          if (tile.south() > this->rows() || tile.east() > this->cols())
            {
              std::ostringstream oss {};
              oss << "cannot insert a "
                  << "(" << tile.rows() << ", " << tile.cols() << ")"
                  << " tile from "
                  << "(" << tile.north() << ", " << tile.west() << ")"
                  << " to "
                  << "(" << tile.south() << ", " << tile.east() << ")"
                  << " into a "
                  << "(" << this->rows() << ", " << this->cols() << ")"
                  << " matrix without resizing";
              throw std::invalid_argument {oss.str()};
            }
        }
      for (const auto& iter : this->tiles_)
        {
          if (tile.getMatrix() == iter.getMatrix())
            throw std::invalid_argument {"tile aliases existing tile"};
          if (overlap(tile, iter))
            {
              std::ostringstream oss {};
              oss << "(" << tile.rows() << ", " << tile.cols() << ")"
                  << " tile from "
                  << "(" << tile.north() << ", " << tile.west() << ")"
                  << " to "
                  << "(" << tile.south() << ", " << tile.east() << ")"
                  << " would overlap with existing "
                  << "(" << iter.rows() << ", " << iter.cols() << ")"
                  << " tile from "
                  << "(" << iter.north() << ", " << iter.west() << ")"
                  << "  to "
                  << "(" << iter.south() << ", " << iter.east() << ")";
              throw std::invalid_argument {oss.str()};
            };
        }
    }

    /**
     * @brief
     *         Maps a pair of indices to the corresponding tile, if such tile
     *         exists.
     *
     * On success, a pointer to a matrix is returned such that the element
     * <code>(*<var>i_ptr</var>, *<var>j_ptr</var>)</code> in that matrix is
     * the subscripted element.  If the index is valid but does not fall into a
     * tile, a `nullptr` is returned and
     * <code>(*<var>i_ptr</var>, *<var>j_ptr</var>)</code> is undefined.
     *
     * @param row
     *         row index for the whole `ArrangedMatrix`
     *
     * @param col
     *         column index for the whole `ArrangedMatrix`
     *
     * @param i_ptr
     *         address to store the mapped row index at
     *
     * @param j_ptr
     *         address to store the mapped column index at
     *
     * @returns
     *         a pointer to the subscripted tile or a `nullptr` if no such tile
     *         exists
     *
     * @throws std::invalid_argument
     *         if argument checking is enabled and the index is out of range
     *
     */
    super_type *
    findIndex_(const index_type row,
               const index_type col,
               index_type *const i_ptr,
               index_type *const j_ptr) const
    {
      if (BSC_KLAMMLER_CHECK_ARGS) { this->checkIndex(row, col); }
      for (const auto& tile : this->tiles_)
        {
          assert(tile.south() > tile.north());
          assert(tile.east() > tile.west());
          if (row >= tile.north()
              && row < tile.south()
              && col >= tile.west()
              && col < tile.east())
            {
              *i_ptr = row - tile.north();
              *j_ptr = col - tile.west();
              return tile.getMatrix();
            }
        }
      *i_ptr = static_cast<index_type>(-1);
      *j_ptr = static_cast<index_type>(-1);
      return nullptr;
    }

    /**
     * @brief
     *         Mapps a region for this whole `ArrangedMatrix` to an equivalent
     *         list of non-zero regions for our tiles.
     *
     * If a tile does not overlap with the region, the corresponding entry in
     * the returned vector will be an empty region.
     *
     * @param rp
     *         region for this whole `ArrangedMatrix`
     *
     * @returns
     *         regions for the tiles in the same order as found in
     *         `this->tiles_`
     *
     * @throws std::invalid_argument
     *         if the region is invalid
     *
     */
    // TODO: Could we do better than returning a fixed-length list with
    //       potential dummy-entris?  We could return a map but that does not
    //       sound too brilliant either, does it?
    std::vector<region_pass>
    mapRegion_(const region_pass rp) const
    {
      using rectangle_t = Rectangle<index_type>;
      if (BSC_KLAMMLER_CHECK_ARGS)
        this->checkRegion(rp.rp_north, rp.rp_west, rp.rp_south, rp.rp_east,
                          rp.rp_global_trans);
      std::vector<region_pass> mapped_regions {};
      const rectangle_t whole_rectangle {
        rp.rp_north, rp.rp_west, rp.rp_south, rp.rp_east,
      };
      mapped_regions.reserve(this->tiles_.size());
      for (const auto tile : this->tiles_)
        {
          const rectangle_t r {
            !rp.rp_global_trans ? tile.north() : tile.west(),
            !rp.rp_global_trans ? tile.west()  : tile.north(),
            !rp.rp_global_trans ? tile.south() : tile.east(),
            !rp.rp_global_trans ? tile.east()  : tile.south(),
          };
          if (overlap(whole_rectangle, r))
            {
              const index_type r_global_north {
                rp.rp_global_north + r.north() - rp.rp_north
              };
              const index_type r_global_west {
                rp.rp_global_west + r.west() - rp.rp_west
              };
              mapped_regions.emplace_back(
                rp.rp_north > r.north() ? rp.rp_global_north : r_global_north,
                rp.rp_west  > r.west()  ? rp.rp_global_west  : r_global_west,
                rp.rp_global_trans,
                rp.rp_north > r.north() ? rp.rp_north - r.north() : 0,
                rp.rp_west  > r.west()  ? rp.rp_west  - r.west()  : 0,
                rp.rp_south < r.south() ? rp.rp_south - r.north() : r.height(),
                rp.rp_east  < r.east()  ? rp.rp_east  - r.west()  : r.width());
            }
          else
            {
              mapped_regions.emplace_back();
            }
        }
      assert(mapped_regions.size() == this->tiles_.size());
      return mapped_regions;
    }

    /**
     * @brief
     *         Checks whether the class invariants for this object hold.
     *
     * The *only* intended use for this function is inside a run-time assertion
     * like
     *
     *     assert(this->classInvariantsHold_());
     *
     * This can be used at the end of constructors or assignment operators to
     * help avoid redundant logic.  Using its return value inside an assertion
     * rather than putting the assertion inside this function allows zero
     * run-time overhead if assertions are disabled.
     *
     * This function is not suitable for parameter validation; if it returns
     * `false`, the program should be aborted.
     *
     * @returns
     *         whether this class' invariants hold for this object
     *
     */
    bool
    classInvariantsHold_() const noexcept
    {
      bool row_linear {true};
      bool col_linear {true};
      bool diagonal {this->rows_ == this->cols_};
      for (const auto& tile : this->tiles_)
        {
          if (tile.getMatrix() == nullptr)
            return false;
          if (tile.rows() == 0 || tile.cols() == 0)
            return false;
          if (tile.north() < 0 || tile.west() < 0)
            return false;
          if (tile.south() > this->rows_ || tile.east() > this->cols_)
            return false;
          if (tile.north() != 0 || tile.south() != this->rows_)
            row_linear = false;
          if (tile.west() != 0 || tile.east() != this->cols_)
            col_linear = false;
          if (tile.rows() != tile.cols() || tile.north() != tile.west())
            diagonal = false;
        }
      if (row_linear != this->rowLinear_)
        return false;
      if (col_linear != this->colLinear_)
        return false;
      if (diagonal != this->diagonal_)
        return false;
      for (std::size_t i = 0; i < this->tiles_.size(); ++i)
        {
          for (std::size_t j = 0; j < i; ++j)
            {
              const matrix_anchor tile1 {this->tiles_.at(i)};
              const matrix_anchor tile2 {this->tiles_.at(j)};
              if (tile1.getMatrix() == tile2.getMatrix())
                return false;
              if (overlap(tile1, tile2))
                return false;
            }
        }
      return true;
    }

    /**
     * @brief
     *         Applies a visitor (either kind) to this `ArrangedMatrix`.
     *
     * This template generalizes away the differences between `ActiveVisitor`s
     * and `PassiveVisitor`s to avoid code duplication between the two virtual
     * `passOverRegion` member functions.
     *
     * @param <V>
     *         type of the visitor (*must* be derived from `PassiveVisitor` or
     *         `ActiveVisitor`)
     *
     * @param <C>
     *         whether the matrix is not modified by the visitor (automatically
     *         determined from the type of the visitor)
     *
     * @param self
     *         reference to the value of the `this` pointer
     *
     * @param visitor
     *         visitor to apply
     *
     * @param params
     *         parameters given to the `passOverRegion` function
     *
     */
    template<typename V, bool C = !std::is_base_of<ActiveVisitor, V>::value>
    static void applyVisitor_(typename const_if<ArrangedMatrix, C>::type& self,
                              V& visitor,
                              const region_pass params)
    {
      static_assert(std::is_base_of<PassiveVisitor, V>::value ||
                    std::is_base_of<ActiveVisitor, V>::value,
                    "'V' must be derived from 'HMatrix::PassiveVisitor' or"
                    " 'HMatrix::ActiveVisitor'");
      const auto mapped_regions = self.mapRegion_(params);
      assert(mapped_regions.size() == self.tiles_.size());
      for (std::size_t i = 0; i < mapped_regions.size(); ++i)
        {
          const region_pass rp = mapped_regions.at(i);
          if (rp.rp_south > rp.rp_north && rp.rp_east > rp.rp_west)
            {
              self.tiles_.at(i).getMatrix()->passOverRegion(
                  rp.rp_global_north,
                  rp.rp_global_west,
                  rp.rp_global_trans,
                  visitor,
                  rp.rp_north,
                  rp.rp_west,
                  rp.rp_south,
                  rp.rp_east);
            }
        }
    }


    /* ----------- */
    /*   Friends   */
    /* ----------- */

  public:

    /**
     * @brief
     *         Swaps data between two `ArrangedMatrix`es.
     *
     * @param first
     *         first matrix to have its data swapped
     *
     * @param second
     *         second matrix to have its data swapped
     *
     */
    friend void
    swap(ArrangedMatrix& first, ArrangedMatrix& second) noexcept
    {
      using std::swap;
      swap(first.rows_, second.rows_);
      swap(first.cols_, second.cols_);
      swap(first.rowLinear_, second.rowLinear_);
      swap(first.colLinear_, second.colLinear_);
      swap(first.diagonal_, second.diagonal_);
      swap(first.tiles_, second.tiles_);
      assert(first.classInvariantsHold_());
      assert(second.classInvariantsHold_());
    }
  };  // class ArrangedMatrix
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_ARRANGED_MATRIX_HPP
