/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file block_ldlt_preconditioner.hpp
 *
 * @brief
 *         Header file providing the `BlockLDLTPreconditioner` class template.
 *
 */

#ifndef BSC_KLAMMLER_BLOCK_LDLT_PRECONDITIONER_HPP
#define BSC_KLAMMLER_BLOCK_LDLT_PRECONDITIONER_HPP

#include <cassert>   // assert
#include <iterator>  // std::next

#include "inefficient_algorithm.hpp"         // BSC_KLAMMLER_INEFFICINT_ALGORITHM
#include "numeric_costs.hpp"                 // bsc_klammler::numeric_cost::*
#include "recursive_spd_preconditioner.hpp"  // bsc_klammler::RecursiveSPDPreconditioner


namespace bsc_klammler
{
  /**
   * @brief
   *         A `Preconditioner` for symmetric and positive definite (spd)
   *         linear systems <var>M</var>&sdot;<var>x</var> = <var>b</var> that
   *         computes a block
   *         <var>L</var>&sdot;<var>D</var>&sdot;<var>L</var><sup>T</sup>
   *         factorization.
   *
   * The coefficient matrix <var>M</var> is factorized into a block lower
   * triangular matrix <var>L</var> and a block diagonal matrix <var>D</var>
   * such that <var>M</var> =
   * <var>L</var>&sdot;<var>D</var>&sdot;<var>L</var><sup>T</sup>.  The
   * preconditioner is stored internally as the matrices <var>L</var> and
   * <var>D</var><sup>&minus;1</sup> which allows for efficient solving.
   *
   * @param <T>
   *         type of the variables in the linear system
   *
   */
  template<typename T>
  class BlockLDLTPreconditioner : public RecursiveSPDPreconditioner<T, true>
  {

    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  private:

    /**
     * @brief
     *         Convenince type alias for
     *         `RecursiveSPDPreconditioner<T, true>`.
     *
     */
    using super_type = RecursiveSPDPreconditioner<T, true>;

  public:

    /** @brief Type of the variables in the linear system. */
    using value_type = typename super_type::value_type;

    /** @brief Type for the dimensionality of the linear system. */
    using index_type = typename super_type::index_type;

    /** @brief Type for counting FLOPs. */
    using flops_type = typename super_type::flops_type;

  private:

    /** @brief Convenience type alias for `HMatrix`es. */
    using h_matrix = typename super_type::h_matrix;

    /** @brief Convenience type alias for `FullMatrix`es. */
    using full_matrix = typename super_type::full_matrix;

    /** @brief Convenience type alias for `BlockMatrix`es. */
    using block_matrix = typename super_type::block_matrix;

    /** @brief Convenience type alias for `ArrangedMatrix`es. */
    using arranged_matrix = typename super_type::arranged_matrix;

    /** @brief Convenience type alias for `MutableMatrixView`s. */
    using mutable_matrix_view = typename super_type::mutable_matrix_view;

    /** @brief Convenience type alias for `ConstMatrixView`s. */
    using const_matrix_view = typename super_type::const_matrix_view;

    /** @brief Convenience type alias for `HVector`s. */
    using h_vector = typename super_type::h_vector;

    /** @brief Convenience type alias for `FullVector`s. */
    using full_vector = typename super_type::full_vector;


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief Inverse diagonal factor <var>D</var><sup>&minus;1</sup>. */
    arranged_matrix factorDInv_ {};

    /** @brief Mutex used for locking `factorDInv_`. */
    std::mutex mutexDInv_ {};


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    BlockLDLTPreconditioner(const index_type size) :
      super_type {size}, factorDInv_ {size, size}
    {
      // do nothing
    }


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  protected:

    // {@inheritDoc}
    virtual std::string
    getName(std::vector<std::string>&) const final override
    {
      return "Block-LDLT";
    }

    // {@inheritDoc}
    virtual arranged_matrix
    decomposeImpl(const full_matrix& m11,
                  const arranged_matrix& m12,
                  const arranged_matrix& m21,
                  const h_matrix& m22 BSC_KLAMMLER_ATTRIBUTE_UNUSED,
                  const index_type offset) final override
    {
      full_matrix inverse {invert(m11)};
      arranged_matrix update {};
      if (BSC_KLAMMLER_COUNT_FLOPS)
        {
          namespace nc = numeric_cost;
          const flops_type sz {static_cast<flops_type>(m11.size())};
          const flops_type ntrs {
            static_cast<flops_type>(non_zero_count(m21) / m21.cols())
          };
          this->addInitFLOPs(nc::matrix_inversion_spd(sz) +
                             nc::symmetric_matrix_matrix_product(sz, ntrs) +
                             nc::matrix_matrix_product(ntrs, sz, ntrs));
        }
      if (m21.rows() > 0)
        {
          arranged_matrix factor {m21 * inverse};
          update = {factor * m12};
          std::unique_ptr<h_matrix> factor_uptr {
            new arranged_matrix {std::move(factor)}
          };
          {
            const auto lck = this->lockL();
            this->getFactorL().addTile(std::move(factor_uptr),
                                       offset + m11.rows(),
                                       offset);
          }
          assert(!factor_uptr);
          assert(factor.size() == 0);
        }
      std::unique_ptr<h_matrix> inverse_uptr {
        new full_matrix {std::move(inverse)}
      };
      {
        const auto lck = this->lockDInv();
        this->getFactorDInv().addTile(std::move(inverse_uptr),
                                      offset,
                                      offset);
      }
      assert(!inverse_uptr);
      assert(inverse.size() == 0);
      return update;
    }

    /**
     * @brief
     *         Solves the linear system
     *         <var>P</var>&sdot;<var>y</var>=<var>c</var> where <var>P</var>
     *         is the preconditioner.
     *
     * The system is solved by performing a forward-substitution on
     * <var>L</var>, then a multiplication with <var>D</var><sup>&minus;1</sup>
     * and finally a backward-substitution on <var>L</var><sup>T</sup>.
     *
     * @param step0
     *         right-side to solve the system for
     *
     * @returns
     *         an (approximate) solution to the linear system
     *
     */
    virtual full_vector
    solveImplImpl(const h_vector& step0) const final override
    {
      // solve  L D L^T x = b  ./-> x
      // via    L d = b        ./-> d       { forward substitution }
      // and    D c = d        ./-> c       { multiplication with inverse }
      // and    L^T x = c      ./-> x       { backward substitution }
      const bool add_flops {BSC_KLAMMLER_COUNT_FLOPS && !this->hasSolved()};
      full_vector step1 {this->substituteForward_(step0, add_flops)};
      full_vector step2 {this->solveForD_(step1, add_flops)};
      full_vector step3 {this->substituteBackward_(step2, add_flops)};
      return step3;
    }

  private:

    // Albeit being private, these functions do not have an underscore
    // suffixed to their name to be consistent with the ones inherited for the
    // factor 'L'.

    /**
     * @brief
     *         Returns an exclusive lock for the factor
     *         <var>D</var><sup>&minus;1</sup>.
     *
     * @returns
     *         lock for <var>D</var><sup>&minus;1</sup>
     *
     */
    std::unique_lock<std::mutex>
    lockDInv()
    {
      return std::unique_lock<std::mutex> {this->mutexDInv_};
    }

    /**
     * @brief
     *         Returns a reference to the factor
     *         <var>D</var><sup>&minus;1</sup>.
     *
     * @returns
     *         reference to the factor <var>D</var><sup>&minus;1</sup>
     *
     */
    arranged_matrix&
    getFactorDInv() noexcept
    {
      return this->factorDInv_;
    }

    /**
     * @brief
     *         Returns a `const` reference to the factor
     *         <var>D</var><sup>&minus;1</sup>.
     *
     * @returns
     *         `const` reference to the factor <var>D</var><sup>&minus;1</sup>
     *
     */
    const arranged_matrix&
    getFactorDInv() const noexcept
    {
      return this->factorDInv_;
    }

    /**
     * @brief
     *         Solves the linear system <var>L</var>&sdot;<var>d</var> =
     *         <var>b</var> via forward substitution.
     *
     * @param vec
     *         right-hand side to solve for
     *
     * @param add_flops
     *         whether to update the FLOP count (if enabled at compile-time)
     *
     * @returns
     *         (approximate) solution
     *
     */
    full_vector
    substituteForward_(const h_vector& vec, const bool add_flops) const
    {
      assert(vec.size() == this->size());
      full_vector res {vec};
      index_type top {-1};
      index_type bottom {-1};
      const auto beg = this->getFactorL().tiles().cbegin();
      const auto end = this->getFactorL().tiles().cend();
      for (auto itr = beg; itr != end; ++itr)
        {
          assert(itr->north() == itr->east());
          assert(itr->south() == this->size());
          assert(itr == beg || bottom == itr->north());
          top = itr->north();
          if (std::next(itr) != end)
            bottom = std::next(itr)->north();
          else
            bottom = this->size();
          assert(top >= 0);
          assert(bottom <= this->size());
          assert(top < bottom);
          for (const auto& tile : this->getFactorL().tiles())
            {
              if (tile.north() >= bottom)
                break;
              const full_vector piece_v {
                res.getRawVector().block(tile.west(), 0, tile.cols(), 1)
              };
              const const_matrix_view piece_m {
                tile.getMatrix(),
                top - tile.north(), 0,
                bottom - top, tile.cols(),
              };
              const full_vector update {piece_m * piece_v};
              res.getRawVector().block(top, 0, bottom - top, 1)
                -= update.getRawVector();
              if (BSC_KLAMMLER_COUNT_FLOPS && add_flops)
                {
                  namespace nc = numeric_cost;
                  const flops_type cols {
                    static_cast<flops_type>(piece_m.cols())
                  };
                  const flops_type rows {
                    static_cast<flops_type>(non_zero_count(piece_m)) / cols
                  };
                  this->addSolveFLOPs(nc::matrix_vector_product(rows, cols));
                }
            }
        }
      assert(this->checkSubstitution(vec, res, false));
      return res;
    }

    /**
     * @brief
     *         Solves the linear system
     *         <var>L</var><sup>T</sup>&sdot;<var>x</var> = <var>c</var> via
     *         backward substitution.
     *
     * @param vec
     *         right-hand side to solve for
     *
     * @param add_flops
     *         whether to update the FLOP count (if enabled at compile-time)
     *
     * @returns
     *         (approximate) solution
     *
     */
    full_vector
    substituteBackward_(const h_vector& vec, const bool add_flops) const
    {
      assert(vec.size() == this->size());
      full_vector res {vec};
      const auto beg = this->getFactorL().tiles().crbegin();
      const auto end = this->getFactorL().tiles().crend();
      for (auto itr = beg; itr != end; ++itr)
        {
          assert(itr->north() == itr->east());   // sic
          assert(itr->south() == this->size());   // sic
          const index_type top = itr->west();     // sic
          const index_type bottom = itr->east();  // sic
          assert(top >= 0);
          assert(bottom <= this->size());
          assert(top < bottom);
          const full_vector piece_v {
            res.getRawVector().block(itr->north(), 0, itr->rows(), 1)
          };
          const const_matrix_view piece_m {
            const_matrix_view::makeTransposedView(*itr->getMatrix())
          };
          const full_vector update {piece_m * piece_v};
          res.getRawVector().block(top, 0, bottom - top, 1)
            -= update.getRawVector();
          if (BSC_KLAMMLER_COUNT_FLOPS && add_flops)
            {
              namespace nc = numeric_cost;
              const flops_type cols {
                static_cast<flops_type>(piece_m.cols())
              };
              const flops_type rows {
                static_cast<flops_type>(non_zero_count(piece_m)) / cols
              };
              this->addSolveFLOPs(nc::matrix_vector_product(rows, cols));
            }
        }
      assert(this->checkSubstitution(vec, res, true));
      return res;
    }

    /**
     * @brief
     *         Solves the linear system <var>D</var>&sdot;<var>c</var> =
     *         <var>d</var> via multiplication with
     *         <var>D</var><sup>&minus;1</sup>.
     *
     * @param vec
     *         right-hand side to solve for
     *
     * @param add_flops
     *         whether to update the FLOP count (if enabled at compile-time)
     *
     * @returns
     *         (approximate) solution
     *
     */
    full_vector
    solveForD_(const h_vector& vec, const bool add_flops) const
    {
      assert(vec.size() == this->size());
      const full_vector res {this->getFactorDInv() * vec};
      if (BSC_KLAMMLER_COUNT_FLOPS && add_flops)
        {
          namespace nc = numeric_cost;
          for (const auto& tile : this->getFactorDInv().tiles())
            {
              const flops_type sz {
                static_cast<flops_type>(tile.getMatrix()->size())
              };
              this->addSolveFLOPs(nc::symmetric_matrix_vector_product(sz));
            }
        }
      return res;
    }

  };  // class BlockLDLTPreconditioner
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_BLOCK_LDLT_PRECONDITIONER_HPP
