/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file block_llt_preconditioner.hpp
 *
 * @brief
 *         Header file providing the `BlockLLTPreconditioner` class template.
 *
 */

#ifndef BSC_KLAMMLER_BLOCK_LLT_PRECONDITIONER_HPP
#define BSC_KLAMMLER_BLOCK_LLT_PRECONDITIONER_HPP

#include <cassert>   // assert
#include <iterator>  // std::advance, std::distance, std::prev

#include "numeric_costs.hpp"                 // bsc_klammler::numeric_cost::*
#include "recursive_spd_preconditioner.hpp"  // bsc_klammler::RecursiveSPDPreconditioner
#include "utility.hpp"                       // bsc_klammler::Rectangle


namespace bsc_klammler
{
  /**
   * @brief
   *         A `Preconditioner` for symmetric and positive definite (spd)
   *         linear systems <var>M</var>&sdot;<var>x</var> = <var>b</var> that
   *         computes a block <var>L</var>&sdot;<var>L</var><sup>T</sup>
   *         factorization.
   *
   * The coefficient matrix <var>M</var> is factorized into a lower
   * triangular matrix <var>L</var> such that <var>M</var> =
   * <var>L</var>&sdot;<var>D</var>&sdot;<var>L</var><sup>T</sup>.  The
   * preconditioner is stored internally as the matrix <var>L</var>.
   *
   * @param <T>
   *         type of the variables in the linear system
   *
   */
  template<typename T>
  class BlockLLTPreconditioner : public RecursiveSPDPreconditioner<T, false>
  {

    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  private:

    /**
     * @brief
     *         Convenince type alias for
     *         `RecursiveSPDPreconditioner<T, false>`.
     *
     */
    using super_type = RecursiveSPDPreconditioner<T, false>;

  public:

    /** @brief Type of the variables in the linear system. */
    using value_type = typename super_type::value_type;

    /** @brief Type for the dimensionality of the linear system. */
    using index_type = typename super_type::index_type;

    /** @brief Type for counting FLOPs. */
    using flops_type = typename super_type::flops_type;

  private:

    /** @brief Convenience type alias for `HMatrix`es. */
    using h_matrix = typename super_type::h_matrix;

    /** @brief Convenience type alias for `FullMatrix`es. */
    using full_matrix = typename super_type::full_matrix;

    /** @brief Convenience type alias for `BlockMatrix`es. */
    using block_matrix = typename super_type::block_matrix;

    /** @brief Convenience type alias for `ArrangedMatrix`es. */
    using arranged_matrix = typename super_type::arranged_matrix;

    /** @brief Convenience type alias for `MutableMatrixView`s. */
    using mutable_matrix_view = typename super_type::mutable_matrix_view;

    /** @brief Convenience type alias for `ConstMatrixView`s. */
    using const_matrix_view = typename super_type::const_matrix_view;

    /** @brief Convenience type alias for `HVector`s. */
    using h_vector = typename super_type::h_vector;

    /** @brief Convenience type alias for `FullVector`s. */
    using full_vector = typename super_type::full_vector;

    /** @brief Type of the raw dense matrix implementation. */
    using raw_matrix = typename full_matrix::raw_matrix_type;

    /** @brief Type of the raw dense vector implementation. */
    using raw_vector = typename full_vector::raw_vector_type;

    /** @brief Index type for the raw dense matrix / vector implementation. */
    using raw_index_type = typename full_matrix::raw_index_type;

    /** @brief Internal representation of the factorization. */
    using llt_factorization = Eigen::LLT<raw_matrix>;

    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    BlockLLTPreconditioner(const index_type size) : super_type {size}
    {
      // do nothing
    }


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  protected:

    // {@inheritDoc}
    virtual std::string
    getName(std::vector<std::string>&) const final override
    {
      return "Block-LLT";
    }

    // {@inheritDoc}
    virtual arranged_matrix
    decomposeImpl(const full_matrix& m11,
                  const arranged_matrix& m12 BSC_KLAMMLER_ATTRIBUTE_UNUSED,
                  const arranged_matrix& m21,
                  const h_matrix& m22 BSC_KLAMMLER_ATTRIBUTE_UNUSED,
                  const index_type offset) final override
    {
      llt_factorization factorization {};
      factorization.compute(m11.getRawMatrix());
      full_matrix l11 {factorization.matrixL()};
      arranged_matrix l21 {
        this->triangularSolveL21_(factorization.matrixU(), m21)
      };
      const arranged_matrix u12 {
        arranged_matrix::makeArrangement(
            const_matrix_view::makeTransposedView(l21),
            const_matrix_view::makeView)
      };
      assert(l21.isColumnLinear());
      assert(u12.isRowLinear());
      const arranged_matrix update {l21 * u12};
      if (BSC_KLAMMLER_COUNT_FLOPS)
        {
          namespace nc = numeric_cost;
          // We solve 'ntrs' linear systems where 'ntrs' -- since
          // 'm21.isColumnLinear()' can be computed by counting the non-zero
          // entries in 'm21' and dividing by its number of columns.  We apply
          // the same reasoning to get the cost of the product 'l21 * u12'.
          const flops_type sz {static_cast<flops_type>(m11.size())};
          const flops_type ntrs {
            static_cast<flops_type>(non_zero_count(m21) / m21.cols())
          };
          this->addInitFLOPs(nc::llt_factorization(sz) +
                             ntrs * nc::backward_substitution(sz) +
                             nc::matrix_matrix_product(ntrs, sz, ntrs));
        }
      {
        const auto lck = this->lockL();
        this->getFactorL().addTile(
            std::unique_ptr<h_matrix> {new full_matrix {std::move(l11)}},
            offset, offset);
        this->getFactorL().addTile(
            std::unique_ptr<h_matrix> {new arranged_matrix {std::move(l21)}},
            offset + m11.rows(), offset);
      }
      assert(l11.size() == 0);
      assert(l21.size() == 0);
      return update;
    }

    /**
     * @brief
     *         Solves the linear system
     *         <var>P</var>&sdot;<var>y</var>=<var>c</var> where <var>P</var>
     *         is the preconditioner.
     *
     * The system is solved by performing a forward-substitution on
     * <var>L</var> and a backward-substitution on <var>L</var><sup>T</sup>.
     *
     * @param step0
     *         right-side to solve the system for
     *
     * @returns
     *         a solution to the linear system
     *
     */
    virtual full_vector
    solveImplImpl(const h_vector& step0) const final override
    {
      // solve  L L^T x = b  ./-> x
      // via    L c = b      ./-> c       { forward substitution }
      // and    L^T x = c    ./-> x       { backward substitution }
      const bool add_flops {BSC_KLAMMLER_COUNT_FLOPS && !this->hasSolved()};
      full_vector step1 {this->substituteForward_(step0, add_flops)};
      full_vector step2 {this->substituteBackward_(step1, add_flops)};
      return step2;
    }

  private:

    /**
     * @brief
     *         Solves the multiple linear systems <var>L</var><sub>21</sub>
     *         &sdot; <var>U</var><sub>11</sub> = <var>M</var><sub>21</sub> for
     *         <var>L</var><sub>21</sub>.
     *
     * The matrix <var>U</var><sub>11</sub> is an upper triagonal square matrix
     * found by factorizing <var>M</var><sub>11</sub>.  Its type is whatever
     * the Eigen library returns.
     *
     * This function does not update the FLOP count.  It is the caller's
     * responsibility to do so.
     *
     * @param <U>
     *         type for the upper triagonal matrix <var>U</var><sub>11</sub>
     *
     * @param u11
     *         the upper triagonal matrix <var>U</var><sub>11</sub>
     *
     * @param m12
     *         the supposedly row linear matrix <var>M</var><sub>21</sub>
     *
     * @returns
     *         the matrix <var>L</var><sub>21</sub>
     *
     */
    template<typename U>
    arranged_matrix
    triangularSolveL21_(const U& u11, const arranged_matrix& m21)
    {
      struct Visitor : public h_matrix::PassiveVisitor
      {

        // I don't understand why these awkward completely redundant typedefs
        // are needed.
        using h_matrix = HMatrix<value_type>;
        using full_matrix = FullMatrix<value_type>;
        using index_type = typename full_matrix::index_type;
        using raw_index_type = typename full_matrix::raw_index_type;
        using rectangle = Rectangle<index_type>;
        using raw_rectangle = Rectangle<raw_index_type>;

        const U *const u11_ptr_;
        arranged_matrix *const l21_ptr_;

        Visitor(const U *const u11_ptr,
                arranged_matrix *const l21_ptr) :
          u11_ptr_ {u11_ptr}, l21_ptr_ {l21_ptr}
        {
          assert(this->u11_ptr_ != nullptr);
          assert(this->l21_ptr_ != nullptr);
        }

        virtual void
        visitRegion(const index_type global_north,
                    const index_type global_west,
                    const bool global_trans,
                    const h_matrix& matrix,
                    const index_type north,
                    const index_type west,
                    const index_type south,
                    const index_type east) final override
        {
          const rectangle rect {north, west, south, east};
          const raw_rectangle raw_rect {
            static_cast<raw_index_type>(north),
            static_cast<raw_index_type>(west),
            static_cast<raw_index_type>(south),
            static_cast<raw_index_type>(east),
          };
          assert(rect.width() == static_cast<index_type>(this->u11_ptr_->cols()));
          if (auto full_ptr = dynamic_cast<const full_matrix *>(&matrix))
            {
              if (!global_trans)
                {
                  this->l21_ptr_->addTile(
                      std::unique_ptr<h_matrix> {new full_matrix {
                          this->u11_ptr_->transpose().solve(
                              full_ptr->getRawMatrix().block(
                                  raw_rect.north(), raw_rect.west(),
                                  raw_rect.height(), raw_rect.width()
                              ).transpose()).transpose()
                            }}, global_north, global_west);
                }
              else
                {
                  this->l21_ptr_->addTile(
                      std::unique_ptr<h_matrix> {new full_matrix {
                          this->u11_ptr_->transpose().solve(
                              full_ptr->getRawMatrix().transpose().block(
                                  raw_rect.north(), raw_rect.west(),
                                  raw_rect.height(), raw_rect.width()
                              ).transpose()).transpose()
                            }}, global_north, global_west);
                }
            }
          else
            {
              BSC_KLAMMLER_NOT_REACHED;
            }
        }
      };
      assert(m21.isColumnLinear());
      arranged_matrix l21 {m21.rows(), m21.cols()};
      Visitor vtor {&u11, &l21};
      pass_over_region(m21, 0, 0, false, vtor);
      return l21;
    }

    /**
     * @brief
     *         Solves the linear system <var>L</var>&sdot;<var>d</var> =
     *         <var>b</var> via forward substitution.
     *
     * @param vec
     *         right-hand side to solve for
     *
     * @param add_flops
     *         whether to update the FLOP count (if enabled at compile-time)
     *
     * @returns
     *         (approximate) solution
     *
     */
    full_vector
    substituteForward_(const h_vector& vec, const bool add_flops) const
    {
      assert(vec.size() == this->size());
      full_vector res {vec};
      const auto beg = this->getFactorL().tiles().cbegin();
      const auto end = this->getFactorL().tiles().cend();
      for (auto itr = beg; std::distance(itr, end) > 0; std::advance(itr, 2))
        {
          const index_type top = itr->north();
          const index_type bottom = itr->south();
          assert(itr->north() == itr->west());
          assert(itr->getMatrix()->rows() == itr->getMatrix()->cols());
          assert(is_lower_triangular(*itr->getMatrix(), false));
          assert((itr != beg) == (itr->west() > 0));
          for (const auto& tile : this->getFactorL().tiles())
            {
              if (tile.east() > tile.north())
                continue;  // triangular piece
              if (tile.north() >= bottom)
                break;
              const full_vector piece_v {
                res.getRawVector().block(tile.west(), 0, tile.cols(), 1)
              };
              const const_matrix_view piece_m {
                tile.getMatrix(),
                top - tile.north(), 0,
                bottom - top, tile.cols(),
              };
              const full_vector update {piece_m * piece_v};
              res.getRawVector().block(top, 0, bottom - top, 1)
                -= update.getRawVector();
              if (BSC_KLAMMLER_COUNT_FLOPS && add_flops)
                {
                  namespace nc = numeric_cost;
                  const flops_type rows {
                    static_cast<flops_type>(piece_m.rows())
                  };
                  const flops_type cols {
                    static_cast<flops_type>(non_zero_count(piece_m)) / rows
                  };
                  this->addSolveFLOPs(nc::matrix_vector_product(rows, cols));
                }
            }
          if (auto full_ptr = dynamic_cast<const full_matrix *>(itr->getMatrix()))
            {
              const auto upper = full_ptr->getRawMatrix()
                .template triangularView<Eigen::Lower>();
              auto rhs = res.getRawVector().block(top, 0, bottom - top, 1);
              rhs = upper.solve(rhs);
              if (BSC_KLAMMLER_COUNT_FLOPS && add_flops)
                {
                  namespace nc = numeric_cost;
                  const flops_type sz {static_cast<flops_type>(rhs.size())};
                  this->addSolveFLOPs(nc::forward_substitution(sz));
                }
            }
          else
            {
              // We know that this can't happen since we've only put
              // 'FullMatrix'es into 'L'.
              BSC_KLAMMLER_NOT_REACHED;
            }
        }
      assert(this->checkSubstitution(vec, res, false));
      return res;
    }

    /**
     * @brief
     *         Solves the linear system
     *         <var>L</var><sup>T</sup>&sdot;<var>x</var> = <var>c</var> via
     *         backward substitution.
     *
     * @param vec
     *         right-hand side to solve for
     *
     * @param add_flops
     *         whether to update the FLOP count (if enabled at compile-time)
     *
     * @returns
     *         (approximate) solution
     *
     */
    full_vector
    substituteBackward_(const h_vector& vec, const bool add_flops) const
    {
      assert(vec.size() == this->size());
      full_vector res {vec};
      const auto beg = this->getFactorL().tiles().crbegin();
      const auto end = this->getFactorL().tiles().crend();
      for (auto itr = beg; std::distance(itr, end) > 0; std::advance(itr, 2))
        {
          const index_type top = itr->north();
          const index_type bottom = itr->south();
          assert(itr->north() == itr->west());
          assert(itr->south() == itr->east());
          assert(is_lower_triangular(*itr->getMatrix(), false));
          assert((itr == beg) == (itr->east() == this->size()));
          if (std::distance(beg, std::prev(itr)) > 0)
            {
              auto pred = std::prev(itr);
              assert(itr->north() == pred->west());  // sic
              assert(itr->east() == pred->north());  // sic
              const full_vector piece_v {
                res.getRawVector().block(itr->south(), 0, pred->rows(), 1)
              };
              const const_matrix_view piece_m {
                const_matrix_view::makeTransposedView(*pred->getMatrix())
              };
              const full_vector update {piece_m * piece_v};
              res.getRawVector().block(top, 0, bottom - top, 1)
                -= update.getRawVector();
              if (BSC_KLAMMLER_COUNT_FLOPS && add_flops)
                {
                  namespace nc = numeric_cost;
                  const flops_type cols {
                    static_cast<flops_type>(piece_m.cols())
                  };
                  const flops_type rows {
                    static_cast<flops_type>(non_zero_count(piece_m)) / cols
                  };
                  this->addSolveFLOPs(nc::matrix_vector_product(rows, cols));
                }
            }
          if (auto full_ptr = dynamic_cast<const full_matrix *>(itr->getMatrix()))
            {
              const auto upper = full_ptr->getRawMatrix().transpose()
                .template triangularView<Eigen::Upper>();
              auto rhs = res.getRawVector().block(top, 0, bottom - top, 1);
              rhs = upper.solve(rhs);
              if (BSC_KLAMMLER_COUNT_FLOPS && add_flops)
                {
                  namespace nc = numeric_cost;
                  const flops_type sz {static_cast<flops_type>(rhs.size())};
                  this->addSolveFLOPs(nc::backward_substitution(sz));
                }
            }
          else
            {
              // We know that this can't happen since we've only put
              // 'FullMatrix'es into 'L'.
              BSC_KLAMMLER_NOT_REACHED;
            }
        }
      assert(this->checkSubstitution(vec, res, true));
      return res;
    }
  };  // class BlockLLTPreconditioner
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_BLOCK_LLT_PRECONDITIONER_HPP
