/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file block_matrix.hpp
 *
 * @brief
 *         Header file providing the `BlockMatrix` template class.
 *
 */

#ifndef BSC_KLAMMLER_BLOCK_MATRIX_HPP
#define BSC_KLAMMLER_BLOCK_MATRIX_HPP

#include "h_matrix.hpp"     // HMatrix

#include <array>            // std::array
#include <cassert>          // assert
#include <cstddef>          // std::size_t
#include <map>              // std::map
#include <memory>           // std::unique_ptr
#include <sstream>          // std::ostringstream
#include <stdexcept>        // std::invalid_argument
#include <utility>          // std::move

#include "not_reached.hpp"  // BSC_KLAMMLER_NOT_REACHED
#include "rectangle.hpp"    // bsc_klammler::Rectangle
#include "utility.hpp"      // bsc_klammler::const_if


namespace bsc_klammler
{


  /**
   * @brief
   *         A block matrix consisting of nine sub-blocks two of wich are
   *         all-zero.
   *
   * The layout of the matrix is as follows (not proportional due to the
   * limitations of plain character art):
   *
   * <pre>
   *     +------------+-------------+------+
   *     |            |             |      |
   *     |            |             |      |
   *     |     A      |             |  X   |
   *     |            |             |      |
   *     |            |             |      |
   *     +------------+-------------+------+
   *     |            |             |      |
   *     |            |             |      |
   *     |            |             |      |
   *     |            |      B      |  Y   |
   *     |            |             |      |
   *     |            |             |      |
   *     +------------+-------------+------+
   *     |            |             |      |
   *     |     U      |      V      |  C   |
   *     |            |             |      |
   *     +------------+-------------+------+
   * </pre>
   *
   * <var>A</var>, <var>B</var>, <var>C</var> are square matrices of possibly
   * different dimensions and <var>U</var>, <var>V</var>, <var>X</var>,
   * <var>Y</var> are rectangular matrices with appropriate dimensions.  The
   * off-diagonal blocks with no label are all-zeros.  We will use these labels
   * consistently throughout this documentation and in the source code.
   *
   * An object of this class stores (and holds full ownership of) pointers to
   * the seven sub-matrices which may be any `HMatrix`.  The zero blocks are of
   * course not stored explicitely.  &ldquo;Zero&rdquo; in this context means
   * whatever the default constructor of <code><var>T</var></code> produces.
   *
   * @param <T>
   *         type of the matrix elements
   *
   */
  template<typename T>
  class BlockMatrix : public HMatrix<T>
  {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Type of the super-class. */
    using super_type = HMatrix<T>;

    /** @brief Type for elements. */
    using value_type = typename super_type::value_type;

    /** @brief Type for indices. */
    using index_type = typename super_type::index_type;

    /** @brief Imported type for passive visitors. */
    using PassiveVisitor = typename super_type::PassiveVisitor;

    /** @brief Imported type for active visitors. */
    using ActiveVisitor = typename super_type::ActiveVisitor;

  private:

    /** @brief Convenience alias for `struct region_pass`. */
    using region_pass = typename super_type::region_pass;


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief Dimension of the square matrix block <var>A</var>. */
    index_type nA_ {};

    /** @brief Dimension of the square matrix block <var>B</var>. */
    index_type nB_ {};

    /** @brief Dimension of the square matrix block <var>C</var>. */
    index_type nC_ {};

    /** @brief Diagonal square matrix block <var>A</var>. */
    std::unique_ptr<super_type> aUptr_ {};

    /** @brief Diagonal square matrix block <var>B</var>. */
    std::unique_ptr<super_type> bUptr_ {};

    /** @brief Diagonal square matrix block <var>C</var>. */
    std::unique_ptr<super_type> cUptr_ {};

    /** @brief Off-diagonal rectangular matrix block <var>U</var>. */
    std::unique_ptr<super_type> uUptr_ {};

    /** @brief Off-diagonal rectangular matrix block <var>V</var>. */
    std::unique_ptr<super_type> vUptr_ {};

    /** @brief Off-diagonal rectangular matrix block <var>X</var>. */
    std::unique_ptr<super_type> xUptr_ {};

    /** @brief Off-diagonal rectangular matrix block <var>Y</var>. */
    std::unique_ptr<super_type> yUptr_ {};


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    /**
     * @brief
     *         Default constructor: Creates an &ldquo;empty&rdquo;
     *         (0, 0) `BlockMatrix` that is not really useful for much.
     *
     */
    BlockMatrix() noexcept
    {
      assert(this->size() == 0);
      assert(this->classInvariantsHold_());
    }

    /**
     * @brief
     *         Constructor: Creates a new `BlockMatrix` from the given building
     *         blocks.
     *
     * See the DocString for this class for the meaning of the blocks and the
     * requirements for the matrix block dimensions.
     *
     * @param a_uptr
     *         diagonal
     *         (<var>n<sub>A</sub></var>, <var>n<sub>A</sub></var>)
     *         square matrix block
     *
     * @param b_uptr
     *         diagonal
     *         (<var>n<sub>B</sub></var>, <var>n<sub>B</sub></var>)
     *         square matrix block
     *
     * @param c_uptr
     *         diagonal
     *         (<var>n<sub>C</sub></var>, <var>n<sub>C</sub></var>)
     *         square matrix block
     *
     * @param u_uptr
     *         off-diagonal
     *         (<var>n<sub>C</sub></var>, <var>n<sub>U</sub></var>)
     *         rectangular matrix block
     *
     * @param v_uptr
     *         off-diagonal
     *         (<var>n<sub>C</sub></var>, <var>n<sub>V</sub></var>)
     *         rectangular matrix block
     *
     * @param x_uptr
     *         off-diagonal
     *         (<var>n<sub>A</sub></var>, <var>n<sub>X</sub></var>)
     *         rectangular matrix block
     *
     * @param y_uptr
     *         off-diagonal
     *         (<var>n<sub>B</sub></var>, <var>n<sub>Y</sub></var>)
     *         rectangular matrix block
     *
     * @throws std::invalid_argument
     *         if a `nullptr`, an aliased pointer or matrices with incompatible
     *         dimensions are passed
     *
     */
    BlockMatrix(std::unique_ptr<super_type> a_uptr,
                std::unique_ptr<super_type> b_uptr,
                std::unique_ptr<super_type> c_uptr,
                std::unique_ptr<super_type> u_uptr,
                std::unique_ptr<super_type> v_uptr,
                std::unique_ptr<super_type> x_uptr,
                std::unique_ptr<super_type> y_uptr)
    {
      // Do not use an initialization before we have validated the parameters.
      if (BSC_KLAMMLER_CHECK_ARGS)
        {
          const std::map<char, const super_type *> blocks {
            {'A', a_uptr.get()},
            {'B', b_uptr.get()},
            {'C', c_uptr.get()},
            {'U', u_uptr.get()},
            {'V', v_uptr.get()},
            {'X', x_uptr.get()},
            {'Y', y_uptr.get()},
          };
          BlockMatrix::checkNotNull_(blocks);
          BlockMatrix::checkNotAliased_(blocks);
          BlockMatrix::checkDimensions_(blocks);
        }
      this->nA_ = a_uptr->size();
      this->nB_ = b_uptr->size();
      this->nC_ = c_uptr->size();
      this->aUptr_ = std::move(a_uptr);
      this->bUptr_ = std::move(b_uptr);
      this->cUptr_ = std::move(c_uptr);
      this->uUptr_ = std::move(u_uptr);
      this->vUptr_ = std::move(v_uptr);
      this->xUptr_ = std::move(x_uptr);
      this->yUptr_ = std::move(y_uptr);
      assert(this->classInvariantsHold_());
    }

    /**
     * @brief
     *         Constructor: Creates a new `BlockMatrix` from `clone()`s of the
     *         given building blocks.
     *
     * See the DocString for this class for the meaning of the blocks and the
     * requirements for the matrix block dimensions.
     *
     * Invoking this constructor as `%BlockMatrix(a, b, c, u, v, x, y)` is
     * equivalent to invoking `%BlockMatrix(a.clone(), &hellip;, y.clone())`.
     *
     * @param a
     *         diagonal
     *         (<var>n<sub>A</sub></var>, <var>n<sub>A</sub></var>)
     *         square matrix block
     *
     * @param b
     *         diagonal
     *         (<var>n<sub>B</sub></var>, <var>n<sub>B</sub></var>)
     *         square matrix block
     *
     * @param c
     *         diagonal (<var>n<sub>C</sub></var>, <var>n<sub>C</sub></var>)
     *         square matrix block
     *
     * @param u
     *         off-diagonal
     *         (<var>n<sub>C</sub></var>, <var>n<sub>U</sub></var>)
     *         rectangular matrix block
     *
     * @param v
     *         off-diagonal
     *         (<var>n<sub>C</sub></var>, <var>n<sub>V</sub></var>)
     *         rectangular matrix block
     *
     * @param x
     *         off-diagonal
     *         (<var>n<sub>A</sub></var>, <var>n<sub>X</sub></var>)
     *         rectangular matrix block
     *
     * @param y
     *         off-diagonal
     *         (<var>n<sub>B</sub></var>, <var>n<sub>Y</sub></var>)
     *         rectangular matrix block
     *
     * @throws std::invalid_argument
     *         if matrices with incompatible dimensions are passed
     *
     */
    BlockMatrix(const super_type& a,
                const super_type& b,
                const super_type& c,
                const super_type& u,
                const super_type& v,
                const super_type& x,
                const super_type& y) :
      BlockMatrix {
        a.clone(), b.clone(), c.clone(),
        u.clone(), v.clone(), x.clone(), y.clone()}
    {
    }

    /**
     * @brief
     *         Copy constructor: Creates a new `BlockMatrix` that is a (deep)
     *         clone of existing one.
     *
     * @param other
     *         `BlockMatrix` to be cloned
     *
     */
    BlockMatrix(const BlockMatrix& other) :
      nA_ {other.nA_},
      nB_ {other.nB_},
      nC_ {other.nC_},
      aUptr_ {other.aUptr_ ? other.aUptr_->clone() : nullptr},
      bUptr_ {other.bUptr_ ? other.bUptr_->clone() : nullptr},
      cUptr_ {other.cUptr_ ? other.cUptr_->clone() : nullptr},
      uUptr_ {other.uUptr_ ? other.uUptr_->clone() : nullptr},
      vUptr_ {other.vUptr_ ? other.vUptr_->clone() : nullptr},
      xUptr_ {other.xUptr_ ? other.xUptr_->clone() : nullptr},
      yUptr_ {other.yUptr_ ? other.yUptr_->clone() : nullptr}
    {
      assert(this->classInvariantsHold_());
    }

    /**
     * @brief
     *         Move constructor: Creates a new `BlockMatrix` by canibalizing an
     *         existing one.
     *
     * @param other
     *         `BlockMatrix` to be moved away from
     *
     */
    BlockMatrix(BlockMatrix&& other) noexcept : BlockMatrix {}
    {
      swap(*this, other);
      assert(other.rows() == 0);
      assert(other.cols() == 0);
    }

    // No explicit d'tor is needed.


    /* ------------- */
    /*   Operators   */
    /* ------------- */

  public:

    /**
     * @brief
     *         Copy-assignment operator: Makes this `BlockMatrix` become a
     *         clone of an existing one.
     *
     * The existing data in this `BlockMatrix` is lost and the associated
     * memory released while new memory is allocated to hold the new data.
     *
     * @param other
     *         `BlockMatrix` to be cloned
     *
     * @returns
     *         reference to this `BlockMatrix`
     *
     */
    BlockMatrix&
    operator=(const BlockMatrix& other)
    {
      BlockMatrix temp {other};
      swap(*this, temp);
      return *this;
    }

    /**
     * @brief
     *         Move-Assignment operator: Moves all resouces from an existing
     *         `BlockMatrix` -- which is canibalized during the process -- into
     *         this one.
     *
     * The existing data (if any) in this `BlockMatrix` is lost and the
     * associated memory released.
     *
     * @param other
     *         the `BlockMatrix` to be moved away from
     *
     * @returns
     *         a reference to this `BlockMatrix`
     *
     */
    BlockMatrix&
    operator=(BlockMatrix&& other) noexcept
    {
      BlockMatrix temp {};  // default c'tor is noexcept
      swap(*this, temp);
      swap(*this, other);
      return *this;
    }


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  public:

    /**
     * @brief
     *         Returns the number of rows in this `BlockMatrix`.
     *
     * The number of rows of a `BlockMatrix` is the sum of the number of rows
     * of the matrix blocks <var>A</var>, <var>B</var> and <var>C</var>.  Since
     * `BlockMatrix`es are square by definition, `BlockMatrix::rows` is
     * equivalent to `BlockMatrix::cols`.
     *
     * @returns
     *         number of rows
     *
     */
    virtual index_type
    rows() const noexcept final override
    {
      return this->nA_ + this->nB_ + this->nC_;
    }

    /**
     * @brief
     *         Returns the number of cols in this `BlockMatrix`.
     *
     * The number of columns of a `BlockMatrix` is the sum of the number of
     * columns of the matrix blocks <var>A</var>, <var>B</var> and
     * <var>C</var>.  Since `BlockMatrix`es are square by definition,
     * `BlockMatrix::cols` is equivalent to `BlockMatrix::rows`.
     *
     * @returns
     *         number of columns
     *
     */
    virtual index_type
    cols() const noexcept final override
    {
      return this->nA_ + this->nB_ + this->nC_;
    }

    /**
     * @brief
     *         Returns the value of the matrix element
     *         (<code><var>row</var></code>, <code><var>col</var></code>).
     *
     * If the index points into a zero-block, a default-constructed `T` is
     * returned.
     *
     * @param row
     *         row index of the element
     *
     * @param col
     *         column index of the element
     *
     * @returns
     *         element's value
     *
     */
    virtual value_type
    get(const index_type row, const index_type col) const final override
    {
      index_type i {static_cast<index_type>(-1)};
      index_type j {static_cast<index_type>(-1)};
      const super_type *const block_ptr {
        this->findIndex_(row, col, &i, &j)
      };
      return block_ptr ? block_ptr->get(i, j) : super_type::zero();
    }

    /**
     * @brief
     *         Assigns <code><var>val</var></code> to the matrix element
     *         (<code><var>row</var></code>, <code><var>col</var></code>).
     *
     * Assigning to an element inside a zero-block is not allowed and an
     * attempt to do so will throw an exception.
     *
     * @param val
     *         value to assign to the element
     *
     * @param row
     *         row index of the element
     *
     * @param col
     *         column index of the element
     *
     */
    virtual void
    set(const value_type val,
        const index_type row,
        const index_type col) final override
    {
      index_type i {static_cast<index_type>(-1)};
      index_type j {static_cast<index_type>(-1)};
      super_type *const block_ptr {this->findIndex_(row, col, &i, &j)};
      if (BSC_KLAMMLER_CHECK_ARGS && block_ptr == nullptr)
        throw std::invalid_argument {"invalid attempt to write outside block"};
      block_ptr->set(val, i, j);
    }

    /**
     * Returns a clone of this `BlockMatrix`.
     *
     * The caller of this function obtains full ownership of the clone and is
     * responsible for `delete`ing it when done with it.  This clone is really
     * deep and descends recursively into the matrix blocks this `BlockMatrix`
     * is built up from.
     *
     * @returns
     *         a newly created clone of this `BlockMatrix` allocated on the
     *         free store
     *
     */
    virtual std::unique_ptr<super_type>
    clone() const final override
    {
      return std::unique_ptr<super_type> {new BlockMatrix {*this}};
    }

    // {@inheritDoc}
    virtual void
    passOverRegion(const index_type global_north,
                   const index_type global_west,
                   const bool global_trans,
                   PassiveVisitor& visitor,
                   const index_type north,
                   const index_type west,
                   const index_type south,
                   const index_type east) const final override
    {
      const region_pass rp {
        global_north, global_west, global_trans, north, west, south, east
      };
      BlockMatrix::applyVisitor_(*this, visitor, rp);
    }

    // {@inheritDoc}
    virtual void
    passOverRegion(const index_type global_north,
                   const index_type global_west,
                   const bool global_trans,
                   ActiveVisitor& visitor,
                   const index_type north,
                   const index_type west,
                   const index_type south,
                   const index_type east) final override
    {
      const region_pass rp {
        global_north, global_west, global_trans, north, west, south, east
      };
      BlockMatrix::applyVisitor_(*this, visitor, rp);
    }

    /**
     * Returns a pointer to the diagonal square matrix block <var>A</var>; no
     * copy is made nor is ownership transferred.
     *
     * @returns
     *         pointer to the diagonal square matrix block <var>A</var>
     *
     */
    super_type *
    getBlockA() const noexcept
    {
      return this->aUptr_.get();
    }

    /**
     * Returns a pointer to the diagonal square matrix block <var>B</var>; no
     * copy is made nor is ownership transferred.
     *
     * @returns
     *         pointer to the diagonal square matrix block <var>B</var>
     *
     */
    super_type *
    getBlockB() const noexcept
    {
      return this->bUptr_.get();
    }

    /**
     * Returns a pointer to the diagonal square matrix block <var>C</var>; no
     * copy is made nor is ownership transferred.
     *
     * @returns
     *         pointer to the diagonal square matrix block <var>C</var>
     *
     */
    super_type *
    getBlockC() const noexcept
    {
      return this->cUptr_.get();
    }

    /**
     * Returns a pointer to the off-diagonal rectangular matrix block
     * <var>u</var>; no copy is made nor is ownership transferred.
     *
     * @returns
     *         pointer to the off-diagonal rectangular matrix block
     *         <var>u</var>
     *
     */
    super_type *
    getBlockU() const noexcept
    {
      return this->uUptr_.get();
    }

    /**
     * Returns a pointer to the off-diagonal rectangular matrix block
     * <var>V</var>; no copy is made nor is ownership transferred.
     *
     * @returns
     *         pointer to the off-diagonal rectangular matrix block
     *         <var>V</var>
     *
     */
    super_type *
    getBlockV() const noexcept
    {
      return this->vUptr_.get();
    }

    /**
     * Returns a pointer to the off-diagonal rectangular matrix block
     * <var>X</var>; no copy is made nor is ownership transferred.
     *
     * @returns
     *         pointer to the off-diagonal rectangular matrix block
     *         <var>X</var>
     *
     */
    super_type *
    getBlockX() const noexcept
    {
      return this->xUptr_.get();
    }

    /**
     * Returns a pointer to the off-diagonal rectangular matrix block
     * <var>Y</var>; no copy is made nor is ownership transferred.
     *
     * @returns
     *         pointer to the off-diagonal rectangular matrix block
     *         <var>Y</var>
     *
     */
    super_type *
    getBlockY() const noexcept
    {
      return this->yUptr_.get();
    }

  private:

    /**
     * @brief
     *         Maps an index to the appropriate matrix block.
     *
     * On success, a pointer to the matrix block -- or `nullptr` if the index
     * falls into a zero-block -- is returned and <code><var>i_ptr</var></code>
     * and <code><var>j_ptr</var></code> are set to the index of the indexed
     * element in that block.  If the index is out of bound,
     * <code><var>i_ptr</var></code> and <code><var>j_ptr</var></code> will be
     * invalid and an exception is thrown, unless `BSC_KLAMMLER_CHECK_ARGS` is
     * `#define`d to 0.
     *
     * @param row
     *         row index to be mapped
     *
     * @param col
     *         column index to be mapped
     *
     * @param i_ptr
     *         address of a variable to store the mapped row index
     *
     * @param j_ptr
     *         address of a variable to store the mapped column index
     *
     * @returns
     *         a pointer to the matrix block or a `nullptr` if a zero-block is
     *         indexed
     *
     * @throws std::invalid_argument
     *         if the index is out of bounds
     *
     */
    super_type *
    findIndex_(const index_type row,
               const index_type col,
               index_type *const i_ptr,
               index_type *const j_ptr) const
    {
      if (BSC_KLAMMLER_CHECK_ARGS) { this->checkIndex(row, col); }
      if (row < this->nA_)
        {
          *i_ptr = row;
          if (col < this->nA_)
            {
              *j_ptr = col;
              return this->aUptr_.get();
            }
          else if (col < this->nA_ + this->nB_)
            {
              *j_ptr = col - this->nA_;
              return nullptr;
            }
          else if (col < this->nA_ + this->nB_ + this->nC_)
            {
              *j_ptr = col - this->nA_ - this->nB_;
              return this->xUptr_.get();
            }
        }
      else if (row < this->nA_ + this->nB_)
        {
          *i_ptr = row - this->nA_;
          if (col < this->nA_)
            {
              *j_ptr = col;
              return nullptr;
            }
          else if (col < this->nA_ + this->nB_)
            {
              *j_ptr = col - this->nA_;
              return this->bUptr_.get();
            }
          else if (col < this->nA_ + this->nB_ + this->nC_)
            {
              *j_ptr = col - this->nA_ - this->nB_;
              return this->yUptr_.get();
            }
        }
      else if (row < this->nA_ + this->nB_ + this->nC_)
        {
          *i_ptr = row - this->nA_ - this->nB_;
          if (col < this->nA_)
            {
              *j_ptr = col;
              return this->uUptr_.get();
            }
          else if (col < this->nA_ + this->nB_)
            {
              *j_ptr = col - this->nA_;
              return this->vUptr_.get();
            }
          else if (col < this->nA_ + this->nB_ + this->nC_)
            {
              *j_ptr = col - this->nA_ - this->nB_;
              return this->cUptr_.get();
            }
        }
      BSC_KLAMMLER_NOT_REACHED;
    }

    /**
     * @brief
     *         Mapps a region for this whole `BlockMatrix` to an equivalent
     *         list of non-zero regions for our blocks.
     *
     * If a block does not overlap with the region, the corresponding entry in
     * the returned array will be an empty region.
     *
     * @param rp
     *         region for this whole `BlockMatrix`
     *
     * @returns
     *         regions for the blocks <var>A</var>, <var>B</var>, <var>C</var>,
     *         <var>U</var>, <var>V</var>, <var>X</var>, <var>Y</var> -- in
     *         this order
     *
     * @throws std::invalid_argument
     *         if the region is invalid
     *
     */
    std::array<region_pass, 7>
    mapRegion_(const region_pass rp) const
    {
      using rectangle_t = Rectangle<index_type>;
      if (BSC_KLAMMLER_CHECK_ARGS)
        this->checkRegion(rp.rp_north, rp.rp_west, rp.rp_south, rp.rp_east,
                          rp.rp_global_trans);
      const index_type d00 {0};
      const index_type d11 {d00 + this->nA_};
      const index_type d22 {d11 + this->nB_};
      const index_type d33 {d22 + this->nC_};
      const rectangle_t whole_rectangle {
        rp.rp_north, rp.rp_west, rp.rp_south, rp.rp_east,
      };
      std::array<rectangle_t, 7> non_zero_rectangles = {
        rectangle_t {d00, d00, d11, d11},  // A or A^T
        rectangle_t {d11, d11, d22, d22},  // B or B^T
        rectangle_t {d22, d22, d33, d33},  // C or C^T
        rectangle_t {d22, d00, d33, d11},  // U or X^T
        rectangle_t {d22, d11, d33, d22},  // V or Y^T
        rectangle_t {d00, d22, d11, d33},  // X or U^T
        rectangle_t {d11, d22, d22, d33},  // Y or V^T
      };
      std::array<region_pass, 7> mapped_regions;
      for (std::size_t i = 0; i < non_zero_rectangles.size(); ++i)
        {
          const rectangle_t r {non_zero_rectangles.at(i)};
          if (overlap(whole_rectangle, r))
            {
              const index_type r_global_north {
                rp.rp_global_north + r.north() - rp.rp_north
              };
              const index_type r_global_west {
                rp.rp_global_west + r.west() - rp.rp_west
              };
              mapped_regions.at(i) = region_pass {
                rp.rp_north > r.north() ? rp.rp_global_north : r_global_north,
                rp.rp_west  > r.west()  ? rp.rp_global_west  : r_global_west,
                rp.rp_global_trans,
                rp.rp_north > r.north() ? rp.rp_north - r.north() : 0,
                rp.rp_west  > r.west()  ? rp.rp_west  - r.west()  : 0,
                rp.rp_south < r.south() ? rp.rp_south - r.north() : r.height(),
                rp.rp_east  < r.east()  ? rp.rp_east  - r.west()  : r.width(),
              };
              assert(mapped_regions.at(i).rp_north <= mapped_regions.at(i).rp_south);
              assert(mapped_regions.at(i).rp_west <= mapped_regions.at(i).rp_east);
              assert(mapped_regions.at(i).rp_south <= r.height());
              assert(mapped_regions.at(i).rp_east <= r.width());
            }
          else
            {
              assert(mapped_regions.at(i).rp_north == 0);
              assert(mapped_regions.at(i).rp_west == 0);
              assert(mapped_regions.at(i).rp_south == 0);
              assert(mapped_regions.at(i).rp_east == 0);
            }
        }
      return mapped_regions;
    }

    /**
     * @brief
     *         Checks whether the class invariants for this object hold.
     *
     * The *only* intended use for this function is inside a run-time assertion
     * like
     *
     *     assert(this->classInvariantsHold_());
     *
     * This can be used at the end of constructors or assignment operators to
     * help avoid redundant logic.  Using its return value inside an assertion
     * rather than putting the assertion inside this function allows zero
     * run-time overhead if assertions are disabled.
     *
     * This function is not suitable for parameter validation; if it returns
     * `false`, the program should be aborted.
     *
     * @returns
     *         whether this class' invariants hold for this object
     *
     */
    bool
    classInvariantsHold_() const noexcept
    {
      bool all_null {true};
      bool none_null {true};
      const std::array<super_type *, 7> pointers = {
        this->aUptr_.get(),
        this->bUptr_.get(),
        this->cUptr_.get(),
        this->uUptr_.get(),
        this->vUptr_.get(),
        this->xUptr_.get(),
        this->yUptr_.get(),
      };
      for (const auto iter : pointers)
        {
          if (iter == nullptr)
            none_null = false;
          else
            all_null = false;
        }
      if (all_null && none_null)
        {
          BSC_KLAMMLER_NOT_REACHED;
        }
      else if (all_null)
        {
          if (this->nA_ != 0 || this->nB_ != 0 || this->nC_ != 0)
            return false;
        }
      else if (none_null)
        {
          // check dimensions
          if (this->aUptr_->size() != this->nA_)
            return false;
          if (this->bUptr_->size() != this->nB_)
            return false;
          if (this->cUptr_->size() != this->nC_)
            return false;
          if (this->uUptr_->rows() != this->nC_
              || this->uUptr_->cols() != this->nA_)
            return false;
          if (this->vUptr_->rows() != this->nC_
              || this->vUptr_->cols() != this->nB_)
            return false;
          if (this->xUptr_->rows() != this->nA_
              || this->xUptr_->cols() != this->nC_)
            return false;
          if (this->yUptr_->rows() != this->nB_
              || this->yUptr_->cols() != this->nC_)
            return false;
          // check for aliased pointers
          for (std::size_t i = 0; i < pointers.size(); ++i)
            for (std::size_t j = 0; j < i; ++j)
              if (pointers.at(i) == pointers.at(j))
                return false;
        }
      else
        {
          return false;
        }
      return true;
    }

    /**
     * @brief
     *         Applies a visitor (either kind) to this `BlockMatrix`.
     *
     * This template generalizes away the differences between `ActiveVisitor`s
     * and `PassiveVisitor`s to avoid code duplication between the two virtual
     * `passOverRegion` member functions.
     *
     * @param <V>
     *         type of the visitor (*must* be derived from `PassiveVisitor` or
     *         `ActiveVisitor`)
     *
     * @param <C>
     *         whether the matrix is not modified by the visitor (automatically
     *         determined from the type of the visitor)
     *
     * @param self
     *         reference to the value of the `this` pointer
     *
     * @param visitor
     *         visitor to apply
     *
     * @param params
     *         parameters given to the `passOverRegion` function
     *
     */
    template<typename V, bool C = !std::is_base_of<ActiveVisitor, V>::value>
    static void applyVisitor_(typename const_if<BlockMatrix, C>::type& self,
                              V& visitor,
                              const region_pass params)
    {
      // TODO: What if a visitor implements both, the active and the passive
      //       interface?  The problem here is not HOW to do this, but WHAT to
      //       do in this case.  Currently, we will give priority to the active
      //       interface (by using '!std::is_base_of<ActiveVisitor, V>::value'
      //       instead of 'std::is_base_of<PassiveVisitor, V>::value').  Is
      //       this good?  It looks more compatible to the general "non-const
      //       by defaul" spirit of C++.
      static_assert(std::is_base_of<PassiveVisitor, V>::value ||
                    std::is_base_of<ActiveVisitor, V>::value,
                    "'V' must be derived from 'HMatrix::PassiveVisitor' or"
                    " 'HMatrix::ActiveVisitor'");
      constexpr std::size_t nblocks {7};
      typename const_if<std::array<super_type *const, nblocks>, C>::type pointers = {
        self.aUptr_.get(),
        self.bUptr_.get(),
        self.cUptr_.get(),
        !params.rp_global_trans ? self.uUptr_.get() : self.xUptr_.get(),
        !params.rp_global_trans ? self.vUptr_.get() : self.yUptr_.get(),
        !params.rp_global_trans ? self.xUptr_.get() : self.uUptr_.get(),
        !params.rp_global_trans ? self.yUptr_.get() : self.vUptr_.get(),
      };
      const auto mapped_regions = self.mapRegion_(params);
      static_assert(mapped_regions.size() == nblocks, "wrong size()");
      for (std::size_t i = 0; i < nblocks; ++i)
        {
          typename const_if<super_type *const, C>::type block_ptr {pointers.at(i)};
          const region_pass rp = mapped_regions.at(i);
          if (rp.rp_south > rp.rp_north && rp.rp_east > rp.rp_west)
            {
              assert(block_ptr != nullptr);
              block_ptr->passOverRegion(rp.rp_global_north,
                                        rp.rp_global_west,
                                        rp.rp_global_trans,
                                        visitor,
                                        rp.rp_north,
                                        rp.rp_west,
                                        rp.rp_south,
                                        rp.rp_east);
            }
        }
    }

    /**
     * @brief
     *         Checks that none of the pointers is a `nullptr`.
     *
     * If everything is okay, this function does nothing.  Else, an
     * `std::invalid_argument` exception is thrown.
     *
     * <code><var>blocks</var></code> must be a mapping from the uppercase
     * characters `'A'`, `'B'`, `'C'`, `'U'`, `'V'`, `'X'`, `'Y'` to the
     * respective pointers.
     *
     * @param blocks
     *         symbol-to-pointer mapping of the matrix blocks
     *
     * @throws std::invalid_argument
     *         if the map contains a `nullptr`
     *
     */
    static void
    checkNotNull_(const std::map<char, const super_type *>& blocks)
    {
      for (const auto& iter : blocks)
        {
          if (iter.second == nullptr)
            {
              std::ostringstream oss {};
              oss << "sub-matrix " << iter.first << " is a nullptr";
              throw std::invalid_argument {oss.str()};
            }
        }
    }

    /**
     * @brief
     *         Checks that no pair of pointers are mutual aliases.
     *
     * If everything is okay, this function does nothing.  Else, an
     * `std::invalid_argument` exception is thrown.
     *
     * <code><var>blocks</var></code> must be a mapping from the uppercase
     * characters `'A'`, `'B'`, `'C'`, `'U'`, `'V'`, `'X'`, `'Y'` to the
     * respective pointers.
     *
     * @param blocks
     *         symbol-to-pointer mapping of the matrix blocks
     *
     * @throws std::invalid_argument
     *         if the map contains aliased pointers
     *
     */
    static void
    checkNotAliased_(const std::map<char, const super_type *>& blocks)
    {
      // Since we're now using 'std::unique_ptr's almost exclusively, this
      // cannot really happen without a more fundamental memory corruption in
      // the background.  However, an explicit check does not hurt.  If nothing
      // else, it might give the user a helpful error message just moments
      // before the program crashes in a segfault anyway.
      for (const auto& iter1 : blocks)
        {
          for (const auto& iter2 : blocks)
            {
              if (iter1.first == iter2.first)
                {
                  assert(iter1.second == iter2.second);
                  continue;
                }
              if (iter1.second == iter2.second)
                {
                  std::ostringstream oss {};
                  oss << "sub-matrix " << iter1.first
                      << " (" << iter1.second << ")"
                      << " aliases sub-matrix " << iter2.first
                      << " (" << iter2.second << ")";
                  throw std::invalid_argument {oss.str()};
                }
            }
        }
    }

    /**
     * @brief
     *         Checks that the matrix blocks have appropriate dimensions.
     *
     * If everything is okay, this function does nothing.  Else, an
     * `std::invalid_argument` exception is thrown.
     *
     * <code><var>blocks</var></code> must be a mapping from the uppercase
     * characters `'A'`, `'B'`, `'C'`, `'U'`, `'V'`, `'X'`, `'Y'` to the
     * respective pointers.
     *
     * @param blocks
     *         symbol-to-pointer mapping of the matrix blocks
     *
     * @throws std::invalid_argument
     *         if the map contains aliased pointers
     *
     */
    static void
    checkDimensions_(const std::map<char, const super_type *>& blocks)
    {
      for (const auto& iter : blocks)
        {
          if (iter.first == 'A' || iter.first == 'B' || iter.first == 'C')
            {
              if (iter.second->rows() != iter.second->cols())
                {
                  std::ostringstream oss {};
                  oss << "sub-matrix "
                      << iter.first
                      << " is not square";
                  throw std::invalid_argument {oss.str()};
                }
            }
        }
      if (blocks.at('U')->rows() != blocks.at('C')->rows()
          || blocks.at('U')->cols() != blocks.at('A')->cols())
        throw std::invalid_argument {"sub-matrix U has wrong dimensions"};
      if (blocks.at('V')->rows() != blocks.at('C')->rows()
          || blocks.at('V')->cols() != blocks.at('B')->cols())
        throw std::invalid_argument {"sub-matrix V has wrong dimensions"};
      if (blocks.at('X')->rows() != blocks.at('A')->rows()
          || blocks.at('X')->cols() != blocks.at('C')->cols())
        throw std::invalid_argument {"sub-matrix X has wrong dimensions"};
      if (blocks.at('Y')->rows() != blocks.at('B')->rows()
          || blocks.at('Y')->cols() != blocks.at('C')->cols())
        throw std::invalid_argument {"sub-matrix Y has wrong dimensions"};
    }


    /* ----------- */
    /*   Friends   */
    /* ----------- */

  public:

    /**
     * Swaps data between two `BlockMatrix`es.
     *
     * @param first
     *         the first `BlockMatrix` to have its data exchanged
     *
     * @param second
     *         the second `BlockMatrix` to have its data exchanged
     *
     */
    friend void
    swap(BlockMatrix& first, BlockMatrix& second) noexcept
    {
      using std::swap;
      swap(first.nA_, second.nA_);
      swap(first.nB_, second.nB_);
      swap(first.nC_, second.nC_);
      swap(first.aUptr_, second.aUptr_);
      swap(first.bUptr_, second.bUptr_);
      swap(first.cUptr_, second.cUptr_);
      swap(first.uUptr_, second.uUptr_);
      swap(first.vUptr_, second.vUptr_);
      swap(first.xUptr_, second.xUptr_);
      swap(first.yUptr_, second.yUptr_);
      assert(first.classInvariantsHold_());
      assert(second.classInvariantsHold_());
    }
  };  // class BlockMatrix
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_BLOCK_MATRIX_HPP
