/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


#ifndef BSC_KLAMMLER_CONTROL_FLOW_ERROR_HPP
#define BSC_KLAMMLER_CONTROL_FLOW_ERROR_HPP


#include <sstream>    // std::ostringstream
#include <stdexcept>  // std::logic_error
#include <string>     // std::string


namespace bsc_klammler
{


  /**
   * @brief
   *         Exception to indicate that the control flow has reached a point it
   *         rather not have.
   *
   * This class is really meant to be subclassed and not used directly.
   *
   */
  class ControlFlowError : public std::logic_error
  {


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief Source file name the exception is thrown from. */
    std::string file_ {};

    /** @brief Line number the exception is thrown from. */
    int line_ {};


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  protected:

    /**
     * @brief
     *         Constructor: Creates a new exception.
     *
     * @param file
     *         source file name the exception is thrown from
     *
     * @param line
     *         line number the exception is thrown from
     *
     * @param message
     *         additional error message
     *
     */
    ControlFlowError(const std::string& file,
                     const int line,
                     const std::string& message) :
      std::logic_error {ControlFlowError::formatMsg_(file, line, message)},
      file_ {file},
      line_ {line}
    {
      // nothing to do
    }


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  public:

    /**
     * @brief
     *         Returns the souce file name the exception pretends to be thrown
     *         from.
     *
     * @returns
     *         source file name
     *
     */
    std::string
    getFile() const
    {
      return this->file_;
    }

    /**
     * @brief
     *         Returns the line number the exception pretends to be thrown
     *         from.
     *
     * @returns
     *         line number
     *
     */
    int
    getLine() const
    {
      return this->line_;
    }

  private:

    /**
     * @brief
     *         Format an error message with a file name and line number.
     *
     * @param file
     *         source file name
     *
     * @param line
     *         line number
     *
     * @param message
     *         additional error message
     *
     * @returns
     *         formatted error message
     *
     */
    static std::string formatMsg_(const std::string& file,
                                  const int line,
                                  const std::string& message)
    {
      std::ostringstream oss {};
      oss << file << ":" << line << ": " << message;
      return oss.str();
    }
  };  // class ControlFlowError
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_CONTROL_FLOW_ERROR_HPP
