/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file full_matrix.hpp
 *
 * @brief
 *         Header file providing the `FullMatrix` template class.
 *
 */

#ifndef BSC_KLAMMLER_FULL_MATRIX_HPP
#define BSC_KLAMMLER_FULL_MATRIX_HPP

#include "h_matrix.hpp"     // bsc_klammler::HMatrix
#include "rectangle.hpp"    // bsc_klammler::Rectangle

#ifdef __GNUC__
#  pragma GCC diagnostic push
#  pragma GCC diagnostic ignored "-Wunused-local-typedefs"
#endif
#include <Eigen/Dense>      // Eigen::Matrix
#ifdef __GNUC__
#  pragma GCC diagnostic pop
#endif



namespace bsc_klammler
{
  /**
   * @brief
   *         Matrix class for general dense matrices.
   *
   * This class is a wrapper for another matrix data structure that holds the
   * actual data.  It is accessible via the `getMatrix` function member.
   *
   * @param <T>
   *         type of the matrix elements
   *
   */
  template<typename T>
  class FullMatrix : public HMatrix<T>
  {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Type of the super-class. */
    using super_type = HMatrix<T>;

    /** @brief Type for elements. */
    using value_type = typename super_type::value_type;

    /** @brief Type for indices. */
    using index_type = typename super_type::index_type;

    /** @brief Type of the actual dense matrix implementation. */
    using raw_matrix_type = typename Eigen::Matrix<value_type, Eigen::Dynamic, Eigen::Dynamic>;

    /** @brief Type for subscripting the actual dense matrix implementation. */
    using raw_index_type = typename raw_matrix_type::Index;

    /** @brief Imported type for passive visitors. */
    using PassiveVisitor = typename super_type::PassiveVisitor;

    /** @brief Imported type for active visitors. */
    using ActiveVisitor = typename super_type::ActiveVisitor;


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief The wrapped dense matrix implementation. */
    raw_matrix_type matrix_ {};


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    /**
     * @brief
     *         Default constructor: Creates a new empty (0, 0) `FullMatrix`.
     *
     */
    FullMatrix() noexcept
    {
      // Do nothing -- all members are already initialized correctly.
      assert(this->matrix_.rows() == 0);
      assert(this->matrix_.cols() == 0);
    }

    /**
     * @brief
     *         Constructor: Creates a new value-initialized
     *         (<code><var>rows</var></code>, <code><var>cols</var></code>)
     *         `FullMatrix`.
     *
     * @param rows
     *         number of rows
     *
     * @param cols
     *         number of columns
     *
     * @throws std::invalid_argument
     *         if one (but not both) dimension(s) are zero
     *
     */
    explicit FullMatrix(const index_type rows, const index_type cols)
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        super_type::ctorCheckSize(rows, cols);
      this->matrix_.resize(rows, cols);
      // If a type requires initialization (by C++ semantics) then Eigen will
      // already do it and doing it again would be wasted work.  However,
      // primitive types are not value-initialized by Eigen, so we must do
      // this.  At least, this is what our interface promises right now.
      if (!Eigen::NumTraits<value_type>::RequireInitialization)
        this->matrix_.fill(super_type::zero());
    }

    /**
     * @brief
     *         Constructor: Creates a new `FullMatrix` from an initializer
     *         list.
     *
     * The list is interpreted as list of row-vectors.  It goes without saying
     * that all vectors must have the same length.
     *
     * A typical initialization for a (3, 4) `FullMatrix` might look like this:
     *
     *     FullMatrix<double> M1 {
     *         {11.0, 12.0, 13.0, 14.0},
     *         {21.0, 22.0, 23.0, 24.0},
     *         {31.0, 32.0, 33.0, 34.0},
     *     };
     *
     * These initializers are invalid and produce a runtime error:
     *
     *     FullMatrix<double> M2 {{}};                 // error: (1, 0) matrix ill-defined
     *     FullMatrix<double> M3 {{1.0}, {2.0, 3.0}};  // error: non-rectangular matrix
     *
     * This initializer is valid but doesn't call this constructor.  Instead,
     * the default constructor is called and creates a (0, 0) `FullMatrix` as
     * expected.
     *
     *     Matrix<double> M4 {};
     *
     * @param values
     *         values for the `FullMatrix`
     *
     * @throws std::invalid_argument
     *         if the initializer list does not define a valid matrix
     *
     */
    FullMatrix(const std::initializer_list<std::initializer_list<value_type>> values)
    {
      // determine dimensions
      index_type rows {static_cast<index_type>(values.size())};
      index_type cols {0};
      for (const auto& row : values)
        {
          if (!cols)
            cols = row.size();
          else if (!BSC_KLAMMLER_CHECK_ARGS)
            break;
          else if (cols != static_cast<index_type>(row.size()))
            throw std::invalid_argument {"jagged initializer list"};
        }
      if (BSC_KLAMMLER_CHECK_ARGS)
        super_type::ctorCheckSize(rows, cols);
      // copy data
      if (rows == 0 && cols == 0)
        {
          // We have a (n = 0, m = 0) matrix.
          //
          // This case cannot be reached under normal conditions since
          // 'FullMatrix {}' will call the default c'tor.  However, one can
          // trick the compiler into calling this c'tor by using an explicit
          // 'std::initialzer_list' object.
          assert(this->rows() == rows);
          assert(this->cols() == cols);
        }
      else
        {
          this->matrix_.resize(rows, cols);
          index_type row {0};
          for (const auto& row_iter : values)
            {
              index_type col {0};
              for (const auto& col_iter : row_iter)
                {
                  this->matrix_(row, col) = col_iter;
                  ++col;
                }
              assert(col == this->cols());
              ++row;
            }
          assert(row == this->rows());
        }
    }

    /**
     * @brief
     *         Copy constructor: Creates a new `FullMatrix` that is a (deep)
     *         clone of an existing one.
     *
     * @param other
     *         `FullMatrix` to be cloned
     *
     */
    FullMatrix(const FullMatrix& other) : matrix_ {other.matrix_}
    {
      assert(this->rows() == other.rows());
      assert(this->cols() == other.cols());
    }

    /**
     * @brief
     *         Move constructor: Creates a new `FullMatrix` by canibalizing an
     *         existing one.
     *
     * @param other
     *         `FullMatrix` to be moved away from
     *
     */
    FullMatrix(FullMatrix&& other) noexcept : FullMatrix {}
    {
      swap(*this, other);
      assert(other.rows() == 0);
      assert(other.cols() == 0);
    }

    /**
     * @brief
     *         Alternative copy constructor: Creates a new `FullMatrix` by
     *         copying data from an existing `HMatrix`.
     *
     * This constructor must copy the data element-by-element and is thus
     * rather inefficient.  It it marked as `explicit` to avoid unitended
     * invocation.
     *
     * @param data
     *         matrix data to be copied
     *
     */
    explicit FullMatrix(const super_type& other)
    {
      // Note: There is no move-overload -- ie 'FullMatrix(HMatrix&&)' -- for
      //       this c'tor since it would give no performance advantage.  Since
      //       we cannot assume anything about the internal layout of the other
      //       'HMatrix', we would have no choice but implement a copy loop
      //       like for this one.
      struct Visitor : public PassiveVisitor
      {
        raw_matrix_type *const raw_ptr;

        Visitor(raw_matrix_type *const rp) noexcept : raw_ptr {rp}
        {
          assert(this->raw_ptr != nullptr);
        }

        virtual void
        visitRegion(const index_type global_north,
                    const index_type global_west,
                    const bool global_trans,
                    const super_type& matrix,
                    const index_type north,
                    const index_type west,
                    const index_type south,
                    const index_type east) final override
        {
          if (auto full_ptr = dynamic_cast<const FullMatrix *>(&matrix))
            {
              const Rectangle<raw_index_type> src {
                static_cast<raw_index_type>(north),
                static_cast<raw_index_type>(west),
                static_cast<raw_index_type>(south),
                static_cast<raw_index_type>(east),
              };
              const Rectangle<raw_index_type> dst {
                static_cast<raw_index_type>(global_north),
                static_cast<raw_index_type>(global_west),
                static_cast<raw_index_type>(global_north) + src.height(),
                static_cast<raw_index_type>(global_west) + src.width(),
              };
              if (!global_trans)
                this->raw_ptr->block(
                    dst.north(), dst.west(), dst.height(), dst.width())
                  = full_ptr->matrix_.block(
                        src.north(), src.west(), src.height(), src.width());
              else
                this->raw_ptr->block(
                    dst.north(), dst.west(), dst.height(), dst.width())
                  = full_ptr->matrix_.transpose().block(
                        src.north(), src.west(), src.height(), src.width());
            }
          else
            {
              BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
              for (index_type j = west; j < east; ++j)
                {
                  for (index_type i = north; i < south; ++i)
                    {
                      const raw_index_type gri {
                        static_cast<raw_index_type>(global_north + i - north)
                      };
                      const raw_index_type grj {
                        static_cast<raw_index_type>(global_west + j - west)
                      };
                      if (!global_trans)
                        (*this->raw_ptr)(gri, grj) = matrix.get(i, j);
                      else
                        (*this->raw_ptr)(gri, grj) = matrix.get(j, i);
                    }
                }
            }
        }
      };
      const raw_index_type rows_r {static_cast<raw_index_type>(other.rows())};
      const raw_index_type cols_r {static_cast<raw_index_type>(other.cols())};
      this->matrix_.resize(rows_r, cols_r);
      this->matrix_.fill(super_type::zero());
      Visitor vtor {&this->matrix_};
      pass_over_region(other, 0, 0, false, vtor);
    }

    /**
     * @brief
     *         Alternative copy constructor: Wraps a copy of a raw dense matrix
     *         into a new `FullMatrix`.
     *
     * @param data
     *         raw matrix data to be copied into the new `FullMatrix`
     *
     */
    FullMatrix(const raw_matrix_type& data) : matrix_ {data}
    {
      assert(this->rows() == static_cast<index_type>(data.rows()));
      assert(this->cols() == static_cast<index_type>(data.cols()));
    }

    /**
     * @brief
     *         Alternative move constructor: Wraps a raw dense matrix, which is
     *         canibalized during the process, into a new `FullMatrix`.
     *
     * @param data
     *         raw matrix data to be moved into the new `FullMatrix`
     *
     */
    FullMatrix(raw_matrix_type&& data) noexcept
    {
      // Eigen is not yet C++11 aware so we cannot 'std::move(data)' directly.
      this->matrix_.swap(data);
      assert(data.rows() == 0);
      assert(data.cols() == 0);
    }

    // No explicit d'tor is needed.


    /* ------------- */
    /*   Operators   */
    /* ------------- */

  public:

    /**
     * @brief
     *         Copy-assignment operator: Makes this `FullMatrix` become a clone
     *         of an existing one.
     *
     * The existing data in this `FullMatrix` is lost and the associated memory
     * released while new memory is allocated to hold the new data.
     *
     * @param other
     *         `FullMatrix` to be cloned
     *
     * @returns
     *         reference to this `FullMatrix`
     *
     */
    FullMatrix&
    operator=(const FullMatrix& other)
    {
      FullMatrix temp {other};
      swap(*this, temp);
      return *this;
    }

    /**
     * @brief
     *         Move-Assignment operator: Moves all resouces from an existing
     *         `FullMatrix` -- which is canibalized during the process -- into
     *         this one.
     *
     * The existing data (if any) in this `FullMatrix` is lost and the
     * associated memory released.
     *
     * @param other
     *         the `FullMatrix` to be moved away from
     *
     * @returns
     *         a reference to this `FullMatrix`
     *
     */
    FullMatrix&
    operator=(FullMatrix&& other) noexcept
    {
      FullMatrix temp {};  // default c'tor is noexcept
      swap(*this, temp);
      swap(*this, other);
      return *this;
    }


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  public:

    // {@InheritDoc}
    virtual index_type
    rows() const noexcept final override
    {
      const raw_index_type r {this->matrix_.rows()};
      assert(r >= 0);
      return static_cast<index_type>(r);
    }

    // {@InheritDoc}
    virtual index_type
    cols() const noexcept final override
    {
      const raw_index_type c {this->matrix_.cols()};
      assert(c >= 0);
      return static_cast<index_type>(c);
    }

    // {@InheritDoc}
    virtual value_type
    get(const index_type row, const index_type col) const final override
    {
      if (BSC_KLAMMLER_CHECK_ARGS) { this->checkIndex(row, col); }
      return this->matrix_(row, col);
    }

    // {@InheritDoc}
    virtual void
    set(const value_type val,
        const index_type row,
        const index_type col) final override
    {
      if (BSC_KLAMMLER_CHECK_ARGS) { this->checkIndex(row, col); }
      this->matrix_(row, col) = val;
    }

    // {@InheritDoc}
    virtual std::unique_ptr<super_type>
    clone() const final override
    {
      return std::unique_ptr<super_type> {new FullMatrix {*this}};
    }

    // {@InheritDoc}
    virtual void
    passOverRegion(const index_type global_north,
                   const index_type global_west,
                   const bool global_trans,
                   PassiveVisitor& visitor,
                   const index_type north,
                   const index_type west,
                   const index_type south,
                   const index_type east) const final override
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        this->checkRegion(north, west, south, east, global_trans);
      visitor.visitRegion(global_north, global_west, global_trans, *this,
                          north, west, south, east);
    }

    // {@InheritDoc}
    virtual void
    passOverRegion(const index_type global_north,
                   const index_type global_west,
                   const bool global_trans,
                   ActiveVisitor& visitor,
                   const index_type north,
                   const index_type west,
                   const index_type south,
                   const index_type east) final override
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        this->checkRegion(north, west, south, east, global_trans);
      visitor.visitRegion(global_north, global_west, global_trans, *this,
                          north, west, south, east);
    }

    /**
     * @brief
     *         Returns a reference to the internal dense matrix representation.
     *
     * @returns
     *         internal dense matrix representation
     *
     */
    raw_matrix_type&
    getRawMatrix() noexcept
    {
      return this->matrix_;
    }

    /**
     * @brief
     *         Returns a read-only reference to the internal dense matrix
     *         representation.
     *
     * @returns
     *         internal dense matrix representation
     *
     */
    const raw_matrix_type&
    getRawMatrix() const noexcept
    {
      return this->matrix_;
    }


    /* ----------- */
    /*   Friends   */
    /* ----------- */

  public:

    /**
     * Swaps data between two `FullMatrix`es.
     *
     * @param first
     *         first `FullMatrix` to have its data exchanged
     *
     * @param second
     *         second `FullMatrix` to have its data exchanged
     *
     */
    friend void
    swap(FullMatrix& first, FullMatrix& second) noexcept
    {
      // 'swap' is not a friend but a member function of 'Eigen::Matrix'.
      first.matrix_.swap(second.matrix_);
    }
  };  // class FullMatrix
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_FULL_MATRIX_HPP
