/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file full_vector.hpp
 *
 * @brief
 *         Header file providing the `FullVector` template class.
 *
 */

#ifndef BSC_KLAMMLER_FULL_VECTOR_HPP
#define BSC_KLAMMLER_FULL_VECTOR_HPP

#include "h_vector.hpp"  // bsc_klammler::HVector

#include <cassert>       // assert
#include <type_traits>   // std::enable_if
#include <typeinfo>      // typeid

#ifdef __GNUC__
#  pragma GCC diagnostic push
#  pragma GCC diagnostic ignored "-Wunused-local-typedefs"
#endif
#include <Eigen/Dense>   // Eigen::Matrix
#ifdef __GNUC__
#  pragma GCC diagnostic pop
#endif


namespace bsc_klammler
{


  /**
   * @brief
   *         Vector class for general dense vectors.
   *
   * @param <T>
   *         type of the vector elements
   *
   */
  template<typename T>
  class FullVector : public HVector<T>
  {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Type of the super-class. */
    using super_type = HVector<T>;

    /** @brief Type for elements. */
    using value_type = typename super_type::value_type;

    /** @brief Type for indices. */
    using index_type = typename super_type::index_type;

    /** @brief Type of the actual dense vector implementation. */
    using raw_vector_type = typename Eigen::Matrix<value_type, Eigen::Dynamic, 1>;

    /** @brief Type for subscripting the actual dense vector implementation. */
    using raw_index_type = typename raw_vector_type::Index;


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief The wrapped full vector implementation. */
    raw_vector_type vector_ {};


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    /**
     * @brief
     *         Default constructor: Creates a new empty 0-dimensional
     *         `FullVector`.
     *
     */
    FullVector() noexcept
    {
      // Do nothing -- all members are already initialized correctly.
      assert(this->vector_.size() == 0);
    }

    /**
     * @brief
     *         Constructor: Creates a new value-initialized
     *         <code><var>size</var></code>-dimensional `FullVector`.
     *
     * @param size
     *         size for the new `FullVector`
     *
     */
    explicit FullVector(const index_type size)
    {
      if (BSC_KLAMMLER_CHECK_ARGS) { super_type::ctorCheckSize(size); }
      this->vector_.resize(size);
      // If a type requires initialization (by C++ semantics) then Eigen will
      // already do it and doing it again would be wasted work.  However,
      // primitive types are not value-initialized by Eigen, so we must do
      // this.  At least, this is what our interface promises right now.
      if (!Eigen::NumTraits<value_type>::RequireInitialization)
        this->vector_.fill(super_type::zero());
    }

    /**
     * @brief
     *         Constructor: Creates a new `FullVector` from an
     *         `initializer_list` of its elements.
     *
     * To avoid confusion of lists with one element with the constructor that
     * takes the vector size, the list must be nested, that is, enclosed in
     * double braces:
     *
     *     FullVector<int> v1 {3};            // creates [0, 0, 0]
     *     FullVector<int> v2 {{3}};          // creates [3]
     *     FullVector<int> v2 {{1, 2}, {3}};  // error
     *
     * @param values
     *         size for the new `FullVector`
     *
     */
    FullVector(const std::initializer_list<std::initializer_list<value_type>> values)
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        {
          if (values.size() != 1)
            throw std::invalid_argument {
              "please use an initializer list of the form '{{1, 2, 3, ...}}'"
            };
        }
      {
        const auto elements = *values.begin();
        const index_type size {static_cast<index_type>(elements.size())};
        raw_index_type k {0};
        if (BSC_KLAMMLER_CHECK_ARGS) { super_type::ctorCheckSize(size); }
        this->vector_.resize(static_cast<raw_index_type>(size));
        for (const auto v : elements) { this->vector_(k++) = v; }
      }
    }

    /**
     * @brief
     *         Copy constructor: Creates a new `FullVector` from an existing
     *         one.
     *
     * @param other
     *         `FullVector` to be cloned
     *
     */
    FullVector(const FullVector& other) : vector_ {other.vector_}
    {
      // nothing to do
    }

    /**
     * @brief
     *         Move constructor: Creates a new `FullVector` by canibalizing an
     *         existing one.
     *
     * @param other
     *         `FullVector` to be moved away from
     *
     */
    FullVector(FullVector&& other) noexcept : FullVector {}
    {
      swap(*this, other);
    }

    /**
     * @brief
     *         Alternative copy constructor: Creates a new `FullVector` from an
     *         existing `HVector`.
     *
     * @param other
     *         `HVector` to be cloned
     *
     */
    explicit FullVector(const super_type& other)
    {
      if (auto full_ptr = dynamic_cast<const FullVector *>(&other))
        {
          this->vector_ = full_ptr->vector_;
        }
      else
        {
          BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
          const raw_index_type size_r {
            static_cast<raw_index_type>(other.size())
          };
          this->vector_.resize(size_r);
          for (index_type i = 0; i < other.size(); ++i)
            {
              const raw_index_type i_r {static_cast<raw_index_type>(i)};
              this->vector_(i_r) = other.get(i);
            }
        }
    }

    /**
     * @brief
     *         Alternative copy constructor: Wraps a copy of a raw dense vector
     *         into a new `FullVector`.
     *
     * @param data
     *         raw vector data to be copied into the new `FullVector`
     *
     */
    FullVector(const raw_vector_type& data) : vector_ {data}
    {
      assert(this->size() == static_cast<index_type>(data.size()));
    }

    /**
     * @brief
     *         Alternative move constructor: Wraps a raw dense vector, which is
     *         canibalized during the process, into a new `FullVector`.
     *
     * @param data
     *         raw vector data to be moved into the new `FullVector`
     *
     */
    FullVector(raw_vector_type&& data) noexcept
    {
      // Eigen is not yet C++11 aware so we cannot 'std::move(data)' directly.
      this->vector_.swap(data);
      assert(data.size() == 0);
    }

    // No explicit d'tor is needed.


    /* ------------- */
    /*   Operators   */
    /* ------------- */

  public:
    /**
     * @brief
     *         Copy-assignment operator: Makes this `FullVector` become a clone
     *         of an existing one.
     *
     * The existing data in this `FullVector` is lost and the associated memory
     * released while new memory is allocated to hold the new data.
     *
     * @param other
     *         `FullVector` to be cloned
     *
     * @returns
     *         a reference to this `FullVector`
     *
     */
    FullVector&
    operator=(const FullVector& other)
    {
      FullVector temp {other};
      swap(*this, temp);
      return *this;
    }

    /**
     * @brief
     *         Move-Assignment operator: Moves all resouces from an existing
     *         `FullVector` into this one.
     *
     * The existing data (if any) in this `FullVector` is lost and the
     * associated memory released.
     *
     * @param other
     *         the `FullVector` to be moved away from
     *
     * @returns
     *         a reference to this `FullVector`
     *
     */
    FullVector&
    operator=(FullVector&& other) noexcept
    {
      FullVector temp {};
      swap(*this, temp);
      swap(*this, other);
      return *this;
    }


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  public:

    // {@InheritDoc}
    virtual index_type
    size() const noexcept final override
    {
      const raw_index_type s {this->vector_.size()};
      assert(s >= 0);
      return static_cast<index_type>(s);
    }

    // {@InheritDoc}
    virtual value_type
    get(const index_type idx) const final override
    {
      if (BSC_KLAMMLER_CHECK_ARGS) { this->checkIndex(idx); }
      return this->vector_(static_cast<raw_index_type>(idx));
    }

    // {@InheritDoc}
    virtual void
    set(const value_type val, const index_type idx) final override
    {
      if (BSC_KLAMMLER_CHECK_ARGS) { this->checkIndex(idx); }
      this->vector_(static_cast<raw_index_type>(idx)) = val;
    }

    // {@InheritDoc}
    virtual std::unique_ptr<super_type>
    clone() const final override
    {
      return std::unique_ptr<super_type> {new FullVector {*this}};
    }

    /**
     * @brief
     *         Returns a reference to the internal dense vector representation.
     *
     * @returns
     *         internal dense vector representation
     *
     */
    raw_vector_type&
    getRawVector() noexcept
    {
      return this->vector_;
    }

    /**
     * @brief
     *         Returns a read-only reference to the internal dense vector
     *         representation.
     *
     * @returns
     *         internal dense vector representation
     *
     */
    const raw_vector_type&
    getRawVector() const noexcept
    {
      return this->vector_;
    }


    /* ----------- */
    /*   Friends   */
    /* ----------- */

  public:

    /**
     * @brief
     *         Swaps data between two `FullVector`s.
     *
     * @param first
     *         first `FullVector` to have its data exchanged
     *
     * @param second
     *         second `FullVector` to have its data exchanged
     *
     */
    friend void
    swap(FullVector& first, FullVector& second) noexcept
    {
      // 'swap' is not a friend but a member function of 'Eigen::Matrix'.
      first.vector_.swap(second.vector_);
    }
  };  // class FullVector
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_FULL_VECTOR_HPP
