/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file h_matrix.hpp
 *
 * @brief
 *         Header file providing the abstract `HMatrix` template class.
 *
 */

#ifndef BSC_KLAMMLER_H_MATRIX_HPP
#define BSC_KLAMMLER_H_MATRIX_HPP

#include <cassert>                    // assert
#include <iomanip>                    // std::setw
#include <limits>                     // std::numeric_limits
#include <memory>                     // std::unique_ptr
#include <ostream>                    // std::ostream
#include <sstream>                    // std::ostringstream
#include <stdexcept>                  // std::invalid_argument

#include "inefficient_algorithm.hpp"  // BSC_KLAMMLER_INEFFICIENT_ALGORITHM
#include "macros.hpp"                 // (package-specific macros)
#include "rectangle.hpp"              // bsc_klammler::Rectangle


namespace bsc_klammler
{


  /**
   * @brief
   *         Abstract base class for hirachical matrices.
   *
   * There can be any number of implementations of a `HMatrix` optimized for
   * the special structure of a particular (sub-)matrix.  All these
   * implementations must provide at least the interface defined by this class.
   *
   * The `rows()` and `cols()` and the `get()` and `set()` virtual function
   * members provide high-level access to the matrix data, possibly at the
   * price of efficiency.  They completely hide any details about the structure
   * of the matrix from the user.  There is also a low-level data interface
   * provided via the visitor pattern that allows access to the matrix'
   * structure for the implementation of efficient operations without breaking
   * encapsulation.
   *
   * @param <T>
   *         type of the matrix elements
   *
   */
  template<typename T>
  class HMatrix {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Type for `HMatrix` elements. */
    typedef T value_type;

    /** @brief Type for `HMatrix` indices. */
    typedef BSC_KLAMMLER_INDEX_TYPE index_type;

    /**
     * @brief
     *         Interface for &ldquo;passive&rdquo; visitors that do not modify
     *         the visited `HMatrix`.
     *
     */
    class PassiveVisitor
    {
    public:

      /** @brief Type for matrix values. */
      using value_type = typename HMatrix::value_type;

      /** @brief Type for matrix indices. */
      using index_type = typename HMatrix::index_type;

      /** @brief Virtual default destructor. */
      virtual ~PassiveVisitor() noexcept = default;

      /**
       * @brief
       *         Visits a region of a `HMatrix`.  It is up to the
       *         implementation to define what this means.
       *
       * @param global_north
       *         row index of <code><var>north</var></code> in some bigger
       *         context
       *
       * @param global_west
       *         column index of <code><var>west</var></code> in some bigger
       *         context
       *
       * @param global_trans
       *         whether the visited `HMatrix` is to be (mentally) transposed
       *         before doing anything
       *
       * @param matrix
       *         current matrix to be visited
       *
       * @param north
       *         first row in the (eventually transposed) current matrix to be
       *         visited
       *
       * @param west
       *         first column in the (eventually transposed) current matrix to
       *         be visited
       *
       * @param south
       *         first row in the (eventually transposed) current matrix to be
       *         not visited
       *
       * @param east
       *         first column in the (eventually transposed) current matrix to
       *         be not visited
       *
       */
      virtual void
      visitRegion(const index_type global_north,
                  const index_type global_west,
                  const bool global_trans,
                  const HMatrix& matrix,
                  const index_type north,
                  const index_type west,
                  const index_type south,
                  const index_type east) = 0;
    };  // class PassiveVisitor

    /**
     * @brief
     *         Interface for &ldquo;active&rdquo; visitors that do modify the
     *         visited `HMatrix`.
     *
     */
    class ActiveVisitor
    {
    public:

      /** @brief Type for matrix values. */
      using value_type = typename HMatrix::value_type;

      /** @brief Type for matrix indices. */
      using index_type = typename HMatrix::index_type;

      /** @brief Virtual default destructor. */
      virtual ~ActiveVisitor() noexcept = default;

      /**
       * @brief
       *         Visits a region of a `HMatrix`.  It is up to the
       *         implementation to define what this means.
       *
       * @param global_north
       *         row index of <code><var>north</var></code> in some bigger
       *         context
       *
       * @param global_west
       *         column index of <code><var>west</var></code> in some bigger
       *         context
       *
       * @param global_trans
       *         whether the visited `HMatrix` is to be (mentally) transposed
       *         before doing anything
       *
       * @param matrix
       *         current matrix to be visited
       *
       * @param north
       *         first row in the (eventually transposed) current matrix to be
       *         visited
       *
       * @param west
       *         first column in the (eventually transposed) current matrix to
       *         be visited
       *
       * @param south
       *         first row in the (eventually transposed) current matrix to be
       *         not visited
       *
       * @param east
       *         first column in the (eventually transposed) current matrix to
       *         be not visited
       *
       */
      virtual void
      visitRegion(const index_type global_north,
                  const index_type global_west,
                  const bool global_trans,
                  /* !const */ HMatrix& matrix,
                  const index_type north,
                  const index_type west,
                  const index_type south,
                  const index_type east) = 0;
    };  // class ActiveVisitor

  protected:

    /**
     * @brief
     *         Data transfer object to pass arguments for visitors around.
     *
     * See the documentation for `HMatrix:passOverRegion()` for the meaning of
     * the individual data members.
     *
     */
    struct region_pass
    {
      index_type rp_global_north {};
      index_type rp_global_west {};
      bool rp_global_trans {};
      index_type rp_north {};
      index_type rp_west {};
      index_type rp_south {};
      index_type rp_east {};

      // The following pile of code is only needed to give this type its obvios
      // default because of a "bug in the standard".

      constexpr region_pass() noexcept
      {
        // do nothing
      }

      constexpr region_pass(const index_type gn,
                            const index_type gw,
                            const bool gt,
                            const index_type n,
                            const index_type w,
                            const index_type s,
                            const index_type e) noexcept :
        rp_global_north {gn},
        rp_global_west {gw},
        rp_global_trans {gt},
        rp_north {n},
        rp_west {w},
        rp_south {s},
        rp_east {e}
      {
        // do nothing
      }

      region_pass(const region_pass&) noexcept = default;
      region_pass& operator=(const region_pass&) noexcept = default;
      region_pass(region_pass&&) noexcept = default;
      region_pass& operator=(region_pass&&) noexcept = default;
    };


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    /**
     * @brief
     *         Virtual destructor: Destroys the `HMatrix` and releases its
     *         associated memory.
     *
     */
    virtual ~HMatrix() noexcept = default;


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  public:

    /**
     * @brief
     *         Returns the number of rows in this `HMatrix`.
     *
     * @returns
     *         number of rows
     *
     */
    virtual index_type
    rows() const noexcept = 0;

    /**
     * @brief
     *         Returns the number of columns in this `HMatrix`.
     *
     * @returns
     *         number of columns
     *
     */
    virtual index_type
    cols() const noexcept = 0;

    /**
     * @brief
     *         Asserts that the `HMatrix` is square and returns its size.
     *
     * It is a fatal error to call this function on a matrix for which it is
     * not asserted that it is square.
     *
     * Rationale: When dealing with a `HMatrix` that is known to be square, one
     * is in a dilemma whether to call `rows()` or `cols()` on it.  And, by the
     * way, should we not use an `assert` to make sure that the other dimension
     * is really the same as we expected it?  This function does exactly that.
     *
     * @returns
     *         size of square `HMatrix`
     *
     */
    inline index_type
    size() const noexcept
    {
      const index_type n {this->rows()};
      assert(n == this->cols());
      return n;
    }

    /**
     * @brief
     *         Returns the value of the `HMatrix` element
     *         (<code><var>row</var></code>, <code><var>col</var></code>).
     *
     * @param row
     *         row index of the element
     *
     * @param col
     *         column index of the element
     *
     * @returns
     *         element's value
     *
     * @throws std::invalid_argument
     *         if argument checking is enabled and the index is out of range
     *
     */
    virtual value_type
    get(const index_type row BSC_KLAMMLER_ATTRIBUTE_UNUSED,
        const index_type col BSC_KLAMMLER_ATTRIBUTE_UNUSED) const
      BSC_KLAMMLER_PURE_VIRTUAL;

    /**
     * @brief
     *         Assigns <code><var>val</var></code> to the `HMatrix` element
     *         (<code><var>row</var></code>, <code><var>col</var></code>).
     *
     * @param val
     *         value to assign to the element
     *
     * @param row
     *         row index of the element
     *
     * @param col
     *         column index of the element
     *
     * @throws std::invalid_argument
     *         if argument checking is enabled and the index is out of range or
     *         does not permit writing to
     *
     */
    virtual void
    set(const value_type val BSC_KLAMMLER_ATTRIBUTE_UNUSED,
        const index_type row BSC_KLAMMLER_ATTRIBUTE_UNUSED,
        const index_type col BSC_KLAMMLER_ATTRIBUTE_UNUSED)
      BSC_KLAMMLER_PURE_VIRTUAL;

    /**
     * @brief
     *         Returns a polymorphic clone of this `HMatrix`.
     *
     * This function can be used to create a deep-copy of a polymorphic
     * `HMatrix`.  Its semantics are that of the copy constructor except that
     * it is polymorphic and allocates the new `HMatrix` on the free store
     * using operator `new`.  The caller of this function obtains full
     * ownership of the clone.
     *
     * This function should be implemented by derived classes like this:
     *
     *     template<typename T>
     *     class XYZMatrix : public HMatrix<T>
     *     {
     *     public:
     *       XYZMatrix(const XYZMatrix& other);  // copy constructor
     *       virtual std::unique_ptr<HMatrix<T>> clone() const override;
     *       // ...
     *     };
     *
     *     std::unique_ptr<HMatrix<T>> XYZMatrix::clone() const
     *     {
     *       return std::unique_ptr<HMatrix<T>> {new XYZMatrix {*this}};
     *     }
     *
     *
     * @returns
     *         a newly created clone of this `HMatrix` allocated on the free
     *         store
     *
     */
    virtual std::unique_ptr<HMatrix<T>>
    clone() const = 0;

    /**
     * @brief
     *         Accepts a `PassiveVisitor` and passes it over a specified region
     *         of this `HMatrix`.
     *
     * This implementation simply calls the `PassiveVisitor::visitRegion()`
     * function of the visitor with the exact same arguments and itself as the
     * matrix.  Derived classes are encouraged to optimize this to only pass
     * the visitor over regions that are not known a priori to be all-zero.
     *
     * @param global_north
     *         row index of <code><var>north</var></code> in some bigger
     *         context
     *
     * @param global_west
     *         column index of <code><var>west</var></code> in some bigger
     *         context
     *
     * @param global_trans
     *         whether this `HMatrix` is to be (mentally) transposed before
     *         doing anything
     *
     * @param visitor
     *         `PassiveVisitor` to accept
     *
     * @param north
     *         first row to be visited in this (eventually transposed)
     *         `HMatrix`
     *
     * @param west
     *         first column to be visited in this (eventually transposed)
     *         `HMatrix`
     *
     * @param south
     *         first row to be not visited in this (eventually transposed)
     *         `HMatrix`
     *
     * @param east
     *         first column to be not visited in this (eventually transposed)
     *         `HMatrix`
     *
     */
    virtual void
    passOverRegion(const index_type global_north,
                   const index_type global_west,
                   const bool global_trans,
                   PassiveVisitor& visitor,
                   const index_type north,
                   const index_type west,
                   const index_type south,
                   const index_type east) const = 0;

    /**
     * @brief
     *         Accepts an `ActiveVisitor` and passes it over a specified region
     *         of this `HMatrix`.
     *
     * This implementation simply calls the `ActiveVisitor::visitRegion()`
     * function of the visitor with the exact same arguments and itself as the
     * matrix.  Derived classes are encouraged to optimize this to only pass
     * the visitor over regions that are not known a priori to be all-zero.
     *
     * @param global_north
     *         row index of <code><var>north</var></code> in some bigger
     *         context
     *
     * @param global_west
     *         column index of <code><var>west</var></code> in some bigger
     *         context
     *
     * @param global_trans
     *         whether this `HMatrix` is to be (mentally) transposed before
     *         doing anything
     *
     * @param visitor
     *         `PassiveVisitor` to accept
     *
     * @param north
     *         first row to be visited in this (eventually transposed)
     *         `HMatrix`
     *
     * @param west
     *         first column to be visited in this (eventually transposed)
     *         `HMatrix`
     *
     * @param south
     *         first row to be not visited in this (eventually transposed)
     *         `HMatrix`
     *
     * @param east
     *         first column to be not visited in this (eventually transposed)
     *         `HMatrix`
     *
     */
    virtual void
    passOverRegion(const index_type global_north,
                   const index_type global_west,
                   const bool global_trans,
                   ActiveVisitor& visitor,
                   const index_type north,
                   const index_type west,
                   const index_type south,
                   const index_type east) = 0;

    /**
     * @brief
     *         Returns the value of zero elemets.
     *
     * @returns
     *         zero
     *
     */
    static constexpr value_type
    zero()
    {
      return value_type {};
    }

    /**
     * @brief
     *         Unconditionally throws an exception complaining about a
     *         dimansion mismatch between two `HMatrix`es.
     *
     * The error message will be: <code>"cannot <var>verb</var> a
     * (<var>n</var>, <var>m</var>) <var>preposition</var> a (<var>k</var>,
     * <var>l</var>) matrix"</code> where
     * <code><var>n</var> = <var>first</var>.rows()</code>,
     * <code><var>m</var> = <var>first</var>.cols()</code>,
     * <code><var>k</var> = <var>second</var>.rows()</code> and
     * <code><var>l</var> = <var>second</var>.cols()</code>.
     *
     * Example:
     *
     *     template<typename T>
     *     void
     *     fake_multiply(const HMatrix<T>& lhs, const HMatrix<T> rhs)
     *     {
     *       if (BSC_KLAMMLER_CHECK_ARGS && lhs.cols() != rhs.rows())
     *         {
     *           HMatrix<T>::throwMismatchedDimensions(
     *               "multiply", lhs, "and", rhs);
     *           // Will not be reached today.
     *         }
     *     }
     *
     * @param verb
     *         word or phrase to be used as the verb in the error message
     *
     * @param first
     *         first (usually left-hand) operand
     *
     * @param preposition
     *         word or phrase to be used as the preposition in the error
     *         message (I hope this is grammatically correct.)
     *
     * @param second
     *         second (usually right-hand) operand
     *
     * @returns
     *         never
     *
     * @throws std::invalid_argument
     *         always
     *
     */
    [[noreturn]] static void
    throwMismatchedDimensions(const std::string& verb,
                              const HMatrix& first,
                              const std::string& preposition,
                              const HMatrix& second)
    {
      std::ostringstream oss {};
      oss << "cannot " << verb << " a "
          << "(" << first.rows() << ", " << first.cols() << ")"
          << " " << preposition << " a "
          << "(" << second.rows() << ", " << second.cols() << ")"
          << " matrix";
      throw std::invalid_argument {oss.str()};
    }

  protected:

    /**
     * @brief
     *         Checks a pair of subscription indices for this `HMatrix` and
     *         throws an exception if it is out of range.
     *
     * This function can be used by accessor functions of derived classes.  If
     * the index is valid, this function has no effect.  This function does not
     * check whether an index might be inappropriate for writing in a special
     * subclass (eg with zero blocks).  Calls to this function should be
     * wrapped in `if (BSC_KLAMMLER_CHECK_ARGS) { checkIndex(...); }`
     * clauses.
     *
     * @param row
     *         row index to check
     *
     * @param col
     *         column index to check
     *
     * @throws std::invalid_argument
     *         if the index is out of range
     *
     */
    void
    checkIndex(const index_type row, const index_type col) const
    {
      if (row < 0 || row >= this->rows() || col < 0 || col >= this->cols())
        {
          std::ostringstream oss {};
          oss << "invalid index "
              << "(" << row << ", " << col << ")"
              << " for a "
              << "(" << this->rows() << ", " << this->cols() << ")"
              << " matrix";
          throw std::invalid_argument {oss.str()};
        }
    }

    /**
     * @brief
     *         Checks a quadruple of indices defining a region for this
     *         `HMatrix` and throws an exception if it is out of range.
     *
     * This function can be used by implementations `passOverRegion` in derived
     * classes.  If the region is valid, this function has no effect.  Calls to
     * this function should be wrapped in
     * `if (BSC_KLAMMLER_CHECK_ARGS) { checkRegion(...); }` clauses.
     *
     * @param north
     *         index of the region's first row
     *
     * @param west
     *         index of the region's first column
     *
     * @param south
     *         index of the (hypothetical) first row below the region
     *
     * @param east
     *         index of the (hypothetical) first column right of the region
     *
     * @param trans
     *         whether the indices are for the transposed matrix
     *
     * @throws std::invalid_argument
     *         if the region is out of range
     *
     */
    void
    checkRegion(const index_type north,
                const index_type west,
                const index_type south,
                const index_type east,
                const bool trans) const
    {
      const index_type height {trans ? this->cols() : this->rows()};
      const index_type width {trans ? this->rows() : this->cols()};
      if (north < 0)
        throw std::invalid_argument {"north < 0"};
      if (west < 0)
        throw std::invalid_argument {"west < 0"};
      if (south < north)
        throw std::invalid_argument {"south < north"};
      if (east < west)
        throw std::invalid_argument {"east < west"};
      if (south > north && north >= height)
        throw std::invalid_argument {"south > north && north >= rows"};
      if (south > height)
        throw std::invalid_argument {"south > rows"};
      if (east > west && west >= width)
        throw std::invalid_argument {"east > west && west >= cols"};
      if (east > width)
        throw std::invalid_argument {"east > cols"};
    }

    /**
     * @brief
     *         Checks whether a designated size is acceptable and throws an
     *         exception if it is not.
     *
     * This function can be called by constructors of derived classes to
     * validate their arguments.  If the size is valid, this function has no
     * effect.  Calls to it should be wrapped in
     * `if (BSC_KLAMMLER_CHECK_ARGS) { ctorCheckSize(rows, cols); }`
     * clauses.
     *
     * @param size
     *         designated size to check
     *
     * @throws std::invalid_argument
     *         if the size is inacceptable
     *
     */
    static void
    ctorCheckSize(const index_type rows, const index_type cols)
    {
      constexpr index_type max_size {
        std::numeric_limits<index_type>::max() >> 2
      };
      if (rows < 0)
        throw std::invalid_argument {"rows < 0"};
      if (cols < 0)
        throw std::invalid_argument {"cols < 0"};
      if (rows > max_size)
        throw std::invalid_argument {"rows > max_size"};
      if (cols > max_size)
        throw std::invalid_argument {"cols > max_size"};
      /* This is "now" allowed. */
      // if ((rows == 0 && cols != 0) || (rows != 0 && cols == 0))
      //   {
      //     std::ostringstream oss {};
      //     oss << "(" << rows << ", " << cols << ") matrix is ill-defined";
      //     throw std::invalid_argument {oss.str()};
      //   }
    }
  };  // class HMatrix

  /**
   * @brief
   *         Writes an informal human-readable representation of a `HMatrix` to
   *         an output stream.
   *
   * The output might consist of several lines and will have a terminating
   * newline.
   *
   * This is a debugging feature.  There is no extraction operator defined that
   * would deserialize a `HMatrix` written by this operator, nor is any promise
   * made that such an operator could be written without loosing information
   * about the matrix.
   *
   * @param <T>
   *         type of the matrix elements
   *
   * @param os
   *         stream to write to
   *
   * @param matrix
   *         matrix to write
   *
   */
  template<typename T>
  std::ostream&
  operator<<(std::ostream& os, const HMatrix<T>& matrix)
  {
    using index_type = typename HMatrix<T>::index_type;
    for (index_type i = 0; i < matrix.rows(); ++i)
      {
        for (index_type j = 0; j < matrix.cols(); ++j)
          {
            os << std::setw(12) << matrix.get(i, j);
            if (!os) { return os; }
          }
        os << BSC_KLAMMLER_EOL;
        if (!os) { return os; }
      }
    return os;
  }

  /**
   * @brief
   *         Convenience wrapper function for the common case of passing a
   *         visitor over an entire `HMatrix`.
   *
   * @param <M>
   *         type of the matrix to pass the visitor over (correctly
   *         `const`-qualified descendant of `HMatrix<T>`)
   *
   * @param <V>
   *         type of the visitor to pass over the matrix (descendant of
   *         `HMatrix<T>::ActiveVisitor` or `HMatrix<T>::PassiveVisitor`)
   *
   * @param matrix
   *         `HMatrix` to pass the visitor over
   *
   * @param global_north
   *         parameter forwarded to the `HMatrix`' `passOverRegion` function
   *         member's parameter with the same name and meaning
   *
   * @param global_west
   *         parameter forwarded to the `HMatrix`' `passOverRegion` function
   *         member's parameter with the same name and meaning
   *
   * @param global_trans
   *         parameter forwarded to the `HMatrix`' `passOverRegion` function
   *         member's parameter with the same name and meaning
   *
   * @param visitor
   *         visitor to pass over the `HMatrix`
   *
   */
  template<typename M, typename V>
  inline void
  pass_over_region(M& matrix,
                   const typename M::index_type global_north,
                   const typename M::index_type global_west,
                   const bool global_trans,
                   V& visitor)
  {
    matrix.passOverRegion(global_north, global_west, global_trans, visitor,
                          0,
                          0,
                          !global_trans ? matrix.rows() : matrix.cols(),
                          !global_trans ? matrix.cols() : matrix.rows());
  }

  /**
   * @brief
   *         Convenience wrapper function to pass a visitor over a region of
   *         `HMatrix` conveniently specified by a `Rectangle`.
   *
   * @param <M>
   *         type of the matrix to pass the visitor over (correctly
   *         `const`-qualified descendant of `HMatrix<T>`)
   *
   * @param <V>
   *         type of the visitor to pass over the matrix (descendant of
   *         `HMatrix<T>::ActiveVisitor` or `HMatrix<T>::PassiveVisitor`)
   *
   * @param matrix
   *         `HMatrix` to pass the visitor over
   *
   * @param global_north
   *         parameter forwarded to the `HMatrix`' `passOverRegion` function
   *         member's parameter with the same name and meaning
   *
   * @param global_west
   *         parameter forwarded to the `HMatrix`' `passOverRegion` function
   *         member's parameter with the same name and meaning
   *
   * @param global_trans
   *         parameter forwarded to the `HMatrix`' `passOverRegion` function
   *         member's parameter with the same name and meaning
   *
   * @param visitor
   *         visitor to pass over the `HMatrix`
   *
   * @param rect
   *         packed parameters forwarded to the `HMatrix`' `passOverRegion`
   *         function member's last four parameters with the same meaning
   *
   */
  template<typename M, typename V>
  inline void
  pass_over_region(M& matrix,
                   const typename M::index_type global_north,
                   const typename M::index_type global_west,
                   const bool global_trans,
                   V& visitor,
                   const Rectangle<typename M::index_type> rect)
  {
    matrix.passOverRegion(global_north, global_west, global_trans, visitor,
                          rect.north(), rect.west(), rect.south(), rect.east());
  }
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_H_MATRIX_HPP
