/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file h_vector.hpp
 *
 * @brief
 *         Header file providing the abstract `HVector` template class.
 *
 */

#ifndef BSC_KLAMMLER_H_VECTOR_HPP
#define BSC_KLAMMLER_H_VECTOR_HPP

#include <iomanip>                    // std::setw
#include <limits>                     // std::numeric_limits
#include <memory>                     // std::unique_ptr
#include <ostream>                    // std::ostream
#include <sstream>                    // std::ostringstream
#include <stdexcept>                  // std::invalid_argument

#include "macros.hpp"                 // (package-specific macros)
#include "inefficient_algorithm.hpp"  // BSC_KLAMMLER_INEFFICIENT_ALGORITHM


namespace bsc_klammler
{


  /**
   * @brief
   *         Abstract base class for hirachical vectors.
   *
   * @param <T>
   *         type of the vector elements
   *
   */
  template<typename T>
  class HVector
  {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Type for `HVector` elements. */
    typedef T value_type;

    /** @brief Type for `HVector` indices. */
    typedef BSC_KLAMMLER_INDEX_TYPE index_type;


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    /**
     * @brief
     *         Virtual destructor: Destroys the `HVector` and releases its
     *         associated memory.
     *
     */
    virtual ~HVector() noexcept = default;


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  public:

    /**
     * @brief
     *         Returns the dimension (number of elements) for this `HVector`.
     *
     * @returns
     *         number of elements
     *
     */
    virtual index_type
    size() const noexcept = 0;

    /**
     * @brief
     *         Returns the value of the <code><var>i</var></code>-th vector
     *         element.
     *
     * @param idx
     *         index of the element
     *
     * @returns
     *         element's value
     *
     */
    virtual value_type
    get(const index_type idx) const = 0;

    /**
     * @brief
     *         Assigns <code><var>val</var></code> to the
     *         <code><var>i</var></code>-th vector element.
     *
     * @param val
     *         value to assign to the element
     *
     * @param idx
     *         index of the element
     *
     */
    virtual void
    set(const value_type val, const index_type idx) = 0;

    /**
     * @brief
     *         Returns a polymorphic clone of this `HVector`.
     *
     * This function can be used to create a deep-copy of a polymorphic
     * `HVector`.  Its semantics are that of the copy constructor except that
     * it is polymorphic and allocates the new `HVector` on the free store
     * using operator `new`.  The caller of this function obtains full
     * ownership of the clone.
     *
     * This function should be implemented by derived classes like this:
     *
     *     template<typename T>
     *     class XYZVector : public HVector<T>
     *     {
     *     public:
     *       XYZVector(const XYZVector& other);  // copy constructor
     *       virtual std::unique_ptr<HVector<T>> clone() const override;
     *       // ...
     *     };
     *
     *     std::unique_ptr<HVector<T>> XYZVector::clone() const
     *     {
     *       return std::unique_ptr<HVector<T>> {new XYZVector {*this}};
     *     }
     *
     *
     * @returns
     *         a newly created clone of this `HVector` allocated on the free
     *         store
     *
     */
    virtual std::unique_ptr<HVector>
    clone() const = 0;

    /**
     * @brief
     *         Returns the value of zero elemets.
     *
     * @returns
     *         zero
     *
     */
    static constexpr value_type
    zero() noexcept
    {
      return value_type {};
    }

    /**
     * @brief
     *         Unconditionally throws an exception complaining about a
     *         dimansion mismatch between two `HVector`s.
     *
     * The error message will be: <code>"cannot <var>verb</var> a <var>n</var>
     * <var>preposition</var> a <var>m</var> vector"</code> where
     * <code><var>n</var> = <var>first</var>.size()</code> and
     * <code><var>m</var> = <var>second</var>.size()</code>.
     *
     * Example:
     *
     *     template<typename T>
     *     void
     *     fake_add(const HVector<T>& lhs, const HVector<T> rhs)
     *     {
     *       if (BSC_KLAMMLER_CHECK_ARGS && lhs.size() != rhs.size())
     *         {
     *           HVector<T>::throwMismatchedDimensions(
     *               "add", lhs, "and", rhs);
     *           // Will not be reached today.
     *         }
     *     }
     *
     * @param verb
     *         word or phrase to be used as the verb in the error message
     *
     * @param first
     *         first (usually left-hand) operand
     *
     * @param preposition
     *         word or phrase to be used as the preposition in the error
     *         message (I hope this is grammatically correct.)
     *
     * @param second
     *         second (usually right-hand) operand
     *
     * @returns
     *         never
     *
     * @throws std::invalid_argument
     *         always
     *
     */
    [[noreturn]] static void
    throwMismatchedDimensions(const std::string& verb,
                              const HVector& first,
                              const std::string& preposition,
                              const HVector& second)
    {
      std::ostringstream oss {};
      oss << "cannot "
          << verb
          << " a "
          << first.size()
          << " "
          << preposition
          << " a "
          << second.size()
          << " vector";
      throw std::invalid_argument {oss.str()};
    }

  protected:

    /**
     * @brief
     *         Checks a subscription index for this `HVector` and throws an
     *         exception if it is out of range.
     *
     * This function can be used by accessor functions of derived classes.  If
     * the index is valid, this function has no effect.  This function does not
     * check whether an index might be inappropriate for writing in a special
     * subclass (eg with zero blocks).  Calls to this function should be
     * wrapped in `if (BSC_KLAMMLER_CHECK_ARGS) { checkIndex(idx); }`
     * clauses.
     *
     * @param idx
     *         subscription index to check
     *
     * @throws std::invalid_argument
     *         if the index is out of range
     *
     */
    void
    checkIndex(const index_type idx) const
    {
      if (idx < 0 || idx >= this->size())
        {
          std::ostringstream oss {};
          oss << "invalid index "
              << idx
              << " for a "
              << this->size()
              << " vector";
          throw std::invalid_argument {oss.str()};
        }
    }

    /**
     * @brief
     *         Checks whether a designated size is acceptable and throws an
     *         exception if it is not.
     *
     * This function can be called by constructors of derived classes to
     * validate their arguments.  If the size is valid, this function has no
     * effect.  Calls to it should be wrapped in
     * `if (BSC_KLAMMLER_CHECK_ARGS) { ctorCheckSize(idx); }` clauses.
     *
     * @param size
     *         designated size to check
     *
     * @throws std::invalid_argument
     *         if the size is inacceptable
     *
     */
    static void
    ctorCheckSize(const index_type size)
    {
      constexpr index_type max_size {
        std::numeric_limits<index_type>::max() >> 2
      };
      if (size < 0)
        throw std::invalid_argument {"size < 0"};
      if (size > max_size)
        throw std::invalid_argument {"size > max_size"};
    }
  };  // class HVector

  /**
   * @brief
   *         Writes an informal human-readable representation of a `HVector` to
   *         an output stream.
   *
   * The output might consist of several lines and will have a terminating
   * newline.
   *
   * This is a debugging feature.  There is no extraction operator defined that
   * would deserialize a `HVector` written by this operator, nor is any promise
   * made that such an operator could be written without loosing information
   * about the vector.
   *
   * @param <T>
   *         type of the vector elements
   *
   * @param os
   *         stream to write to
   *
   * @param vector
   *         vector to write
   *
   */
  template<typename T>
  std::ostream&
  operator<<(std::ostream& os, const HVector<T>& vector)
  {
    using index_type = typename HVector<T>::index_type;
    for (index_type i = 0; i < vector.size(); ++i)
      {
        os << std::setw(12) << vector.get(i) << BSC_KLAMMLER_EOL;
        if (!os) { return os; }
      }
    return os;
  }
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_H_VECTOR_HPP
