/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file inefficient_algorithm.hpp
 *
 * @brief
 *         Header file providing the `InefficientAlgorithm` class and the
 *         `BSC_KLAMMLER_INEFFICIENT_ALGORITHM` convenience macro.
 *
 */

#ifndef BSC_KLAMMLER_INEFFICIENT_ALGORITHM_HPP
#define BSC_KLAMMLER_INEFFICIENT_ALGORITHM_HPP

#include <iostream>                // std::cerr
#include <stdexcept>               // std::invalid_argument
#include <string>                  // std::string

#include "macros.hpp"              // (package-specific macros)
#include "control_flow_error.hpp"  // bsc_klammler::ControlFlowError
#include "not_reached.hpp"         // BSC_KLAMMLER_NOT_REACHED


/**
 * @def BSC_KLAMMLER_INEFFICIENT_ALGORITHM
 *
 * @brief
 *         Convenience macro to instantiate an `InefficientAlgorithm` object
 *         with the current source file name and line number.
 *
 */
#define BSC_KLAMMLER_INEFFICIENT_ALGORITHM                                    \
  ::bsc_klammler::InefficientAlgorithm {__FILE__, __LINE__}


namespace bsc_klammler
{

  /**
   * @brief
   *         Guard object to warn about inefficient algorithms.
   *
   * To mark an algorithm as inefficient, place an instance of this class with
   * automatic storage duration into its body at a place where it is guaranteed
   * to be instantiated (usually as the very first statement).  Every time the
   * algorithm is called, epending on the current global policy, this will have
   * no effect, warn the user about an inefficient algorithm be used or throw
   * an exception.
   *
   * The current policy is a static property of this class and gan be queried
   * and changed via the `getPolicy()` and `setPolicy()` static function
   * members.  These functions are not thread safe and are not intended to be
   * called repetitively during program execution.  Instead, the policy should
   * be set once at program startup, before any threads are created.
   *
   * Example:
   *
   *     // library code:
   *
   *     template<typename T>
   *     void
   *     sort(T *const array, std::size_t n)
   *     {
   *       InefficientAlgorithm {"algorithm.cpp", 42};
   *       bool sorted;
   *       do
   *         {
   *           sorted = true;
   *           for (std::size_t i = 1; i < n; ++i)
   *             {
   *               if (array[i - 1] > array[i])
   *                 {
   *                   std::swap(array[i - 1], array[i]);
   *                   sorted = false;
   *                 }
   *             }
   *         } while (!sorted);
   *     }
   *
   *     // client code:
   *
   *     int
   *     main(int argc, char * * argv)
   *     {
   *       sort(argv, argc);  // will sort happily
   *       InefficientAlgorithm::setPolicy(WARN);
   *       sort(argv, argc);  // will sort and print a warning
   *       InefficientAlgorithm::setPolicy(THROW);
   *       sort(argv, argc);  // will not sort but throw an exception
   *     }
   *
   * Almost certainly, you will want to parametrize the `InefficientAlgorithm`
   * object with the values of the magic `__FILE__` and `__LINE__` macros.  If
   * so, rather use the `BSC_KLAMMLER_INEFFICIENT_ALGORITHM` convenience macro
   * instead of instantiating the class yourself.
   *
   */
  class InefficientAlgorithm
  {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /**
     * @brief
     *         Policies for dealing with an inefficient algorithm be invoked.
     *
     */
    enum class Policies
    {
      // NB: The hardcoded enum values are used in the unit test code.
      IGNORE = 'i',  /*!< do nothing */
      WARN   = 'w',  /*!< print a warning message to standard error output */
      THROW  = 't',  /*!< throw a `ControlFlowError` */
    };

    /**
     * @brief
     *         Parse a string to its corresponding `Policy` value.
     *
     * @param name
     *         string to parse
     *
     * @returns
     *         the parsed value
     *
     * @throws std::invalid_argument
     *         if the string does not name a policy
     *
     */
    static InefficientAlgorithm::Policies
    policyFromString(const std::string& name)
    {
      if (name == "IGNORE")
        return InefficientAlgorithm::Policies::IGNORE;
      else if (name == "WARN")
        return InefficientAlgorithm::Policies::WARN;
      else if (name == "THROW")
        return InefficientAlgorithm::Policies::THROW;
      else
        throw std::invalid_argument {"unknown policy name"};
    }

    /**
     * @brief
     *         Returns the name of a `Policy`.
     *
     * @param policy
     *         policy to format
     *
     * @returns
     *         name of the policy
     *
     * @throws std::invalid_argument
     *         if the policy has an invalid value
     *
     */
    static std::string
    policyToString(const InefficientAlgorithm::Policies policy)
    {
      switch (policy)
        {
        case InefficientAlgorithm::Policies::IGNORE:
          return "IGNORE";
        case InefficientAlgorithm::Policies::WARN:
          return "WARN";
        case InefficientAlgorithm::Policies::THROW:
          return "THROW";
        default:
          throw std::invalid_argument {"invalid policy"};
        }
    }

    /**
     * @brief
     *         Format a policy into an output stream.
     *
     * @param os
     *         stream to write to
     *
     * @param policy
     *         policy to format
     *
     * @returns
     *         a reference to the stream
     *
     * @throws std::invalid_argument
     *         if the policy has an invalid value
     *
     */
    friend std::ostream&
    operator<<(std::ostream& os, const InefficientAlgorithm::Policies policy)
    {
      return os << InefficientAlgorithm::policyToString(policy);
    }

    /**
     * @brief
     *         Exception that is thrown if an exception is thrown.
     *
     */
    class Error : public ControlFlowError
    {

    public:

      /**
       * @brief
       *         Constructor: Creates a new exception.
       *
       * @param file
       *         source file name with the inefficient algorithm
       *
       * @param line
       *         line number with the inefficient algorithm
       *
       * @param message
       *         additional free-text message
       *
       */
      Error(const std::string& file,
            const int line,
            const std::string& message = "inefficient algorithm used") :
        ControlFlowError {file, line, message}
      {
        // do nothing
      }
    };  // class Error


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief Source file name with the inefficient algorithm. */
    const std::string file_;

    /** @brief Line number with the inefficient algorithm. */
    const int line_;


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    /**
     * @brief
     *         Constructor: Creates a new `InefficientAlgorithm` guard object
     *         and performs the associated action.
     *
     * @param file
     *         source file name with the inefficient algorithm
     *
     * @param line
     *         line number with the inefficient algorithm
     *
     * @throws InefficientAlgorithm::Error
     *         if the current global policy is `THROW`
     *
     */
    InefficientAlgorithm(const std::string& file,
                         const int line) :
      file_ {file}, line_ {line}
    {
      const auto policy = InefficientAlgorithm::getPolicy();
      if (policy == InefficientAlgorithm::Policies::IGNORE)
        {
          return;
        }
      else
        {
          std::ostringstream oss {};
          oss << "inefficient algorithm used (policy = " << policy << ")";
          const InefficientAlgorithm::Error err {
            this->file_,
            this->line_,
            oss.str()
          };
          if (policy == InefficientAlgorithm::Policies::WARN)
            {
#ifdef PACKAGE  // quasi standard, defined by GNU Autoconf
              std::cerr << PACKAGE << ": ";
#endif
              std::cerr << "warning: " << err.what() << std::endl;
            }
          else if (policy == InefficientAlgorithm::Policies::THROW)
            {
              throw err;
            }
          else
            {
              BSC_KLAMMLER_NOT_REACHED;
            }
        }
    }


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

    /**
     * @brief
     *         Returns the current global policy.
     *
     * @returns
     *         current global policy
     *
     */
    static InefficientAlgorithm::Policies
    getPolicy() noexcept
    {
      return InefficientAlgorithm::policy_();
    }

    /**
     * @brief
     *         Changes the current global policy.
     *
     * @param policy
     *         new global policy
     *
     * @throws std::invalid_argument
     *         if the new policy is not valid
     *
     */
    static void
    setPolicy(const InefficientAlgorithm::Policies policy)
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        {
          switch (policy)
            {
            case InefficientAlgorithm::Policies::IGNORE:
              break;
            case InefficientAlgorithm::Policies::WARN:
              break;
            case InefficientAlgorithm::Policies::THROW:
              break;
            default:
              throw std::invalid_argument {"invalid policy"};
            }
        }
      InefficientAlgorithm::policy_() = policy;
    }

  private:

    /**
     * @brief
     *         Returns a reference to the static global policy object.
     *
     * @returns
     *         reference to the current global policy
     *
     */
    static InefficientAlgorithm::Policies&
    policy_() noexcept
    {
      static InefficientAlgorithm::Policies policy {
        InefficientAlgorithm::Policies::IGNORE
      };
      return policy;
    }
  };  // class InefficientAlgorithm
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_INEFFICIENT_ALGORITHM_HPP
