/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file inverse_preconditioner.hpp
 *
 * @brief
 *         Header file providing the `InversePreconditioner` template class.
 *
 */

#ifndef BSC_KLAMMLER_INVERSE_PRECONDITIONER_HPP
#define BSC_KLAMMLER_INVERSE_PRECONDITIONER_HPP

#include "algorithms.hpp"      // (linear algebra operations)
#include "full_matrix.hpp"     // bsc_klammler::FullMatrix
#include "numeric_costs.hpp"   // bsc_klammler::numeric_cost::*
#include "preconditioner.hpp"  // bsc_klammler::Preconditioner


namespace bsc_klammler
{


  /**
   * @brief
   *         `Preconditioner` that computes an explicit inverse of the
   *         coefficient matrix and stores it as a dense matrix.
   *
   * @param <T>
   *         type of the variables in the linear system
   *
   */
  template<typename T>
  class InversePreconditioner : public Preconditioner<T>
  {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Type of the variables in the linear system. */
    using value_type = typename Preconditioner<T>::value_type;

    /** @brief Type for the dimensionality of the linear system. */
    using index_type = typename Preconditioner<T>::index_type;

    /** @brief Type for counting FLOPs. */
    using flops_type = typename Preconditioner<T>::flops_type;


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief Inverse of the coefficient matrix. */
    FullMatrix<value_type> inverse_ {};

    /** @brief Whether to assume the matrix is spd. */
    const bool spd_;


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    /**
     * @brief
     *         Default constructor: Creates a new uninitialized
     *         `InversePreconditioner`.
     *
     * @param size
     *         size of the linear system
     *
     * @param spd
     *         whether to assume that the coefficient matrix is symmtric and
     *         positive definite
     *
     */
    InversePreconditioner(const index_type size, const bool spd = false) :
      Preconditioner<T> {size}, spd_ {spd}
    {
      // do nothing
    }


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  public:

    /**
     * @brief
     *         Tells that `InversePreconditioner`s cannot take advantage of the
     *         special structure of hirachical matrices.
     *
     * @returns
     *         `false`
     *
     */
    virtual bool
    isRecursive() const noexcept final override
    {
      return false;
    }

    /**
     * @brief
     *         Returns the flag that was passed to the constructor.
     *
     * @returns
     *         whehter this `InversePreconditioner` was configured to only
     *         handle symmetric and positive definite matrices
     *
     */
    virtual bool
    onlySPD() const noexcept final override
    {
      return this->spd_;
    }

  protected:

    // {@inheritDoc}
    virtual std::string
    getName(std::vector<std::string>&) const final override
    {
      return "Inverse";
    }

    /**
     * @brief
     *         Computes the inverse coefficient matrix
     *         <var>M</var><sup>&minus;1</sup>.
     *
     * @param cofficients
     *         coefficient matrix for the linear system
     *
     */
    virtual void
    initImpl(const HMatrix<value_type>& coefficients) final override
    {
      if (BSC_KLAMMLER_PARALLEL >= BSC_KLAMMLER_PARALLEL_DATA)
        enable_data_level_multithreading(this->getNumCPUs());
      else
        assert(enable_data_level_multithreading() == 1);
      this->inverse_ = invert(coefficients, this->spd_);
      if (BSC_KLAMMLER_COUNT_FLOPS)
        {
          namespace nc = numeric_cost;
          const flops_type sz {static_cast<flops_type>(this->size())};
          if (this->spd_)
            this->addInitFLOPs(nc::matrix_inversion_spd(sz));
          else
            this->addInitFLOPs(nc::matrix_inversion(sz));
        }
    }

    /**
     * @brief
     *         Solves the linear system
     *         <var>P</var>&sdot;<var>y</var>=<var>c</var> where <var>P</var>
     *         is the preconditioner.
     *
     * The system is solved by multiplying with
     * <var>M</var><sup>&minus;1</sup>.
     *
     * @param right
     *         right-side to solve the system for
     *
     * @returns
     *         an (approximate) solution to the linear system
     *
     */
    virtual FullVector<value_type>
    solveImpl(const HVector<value_type>& right) const final override
    {
      if (BSC_KLAMMLER_COUNT_FLOPS)
        {
          namespace nc = numeric_cost;
          const flops_type sz {static_cast<flops_type>(this->size())};
          if (this->spd_)
            this->addSolveFLOPs(nc::symmetric_matrix_vector_product(sz));
          else
            this->addSolveFLOPs(nc::matrix_vector_product(sz, sz));
        }
      return this->inverse_ * right;
    }
  };  // class InversePreconditioner
}  // bsc_klammler bsc_klammler

#endif  // ifndef BSC_KLAMMLER_INVERSE_PRECONDITIONER_HPP
