/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file llt_preconditioner.hpp
 *
 * @brief
 *         Header file providing the `LLTPreconditioner` class template.
 *
 */

#ifndef BSC_KLAMMLER_LLT_PRECONDITIONER_HPP
#define BSC_KLAMMLER_LLT_PRECONDITIONER_HPP

#include "full_matrix.hpp"     // bsc_klammler::FullMatrix
#include "full_vector.hpp"     // bsc_klammler::FullVector
#include "numeric_costs.hpp"   // bsc_klammler::numeric_cost::*
#include "preconditioner.hpp"  // bsc_klammler::Preconditioner
#include "utility.hpp"         // bsc_klammler::enable_data_level_multithreading


namespace bsc_klammler
{


  /**
   * @brief
   *         A `Preconditioner` for symmetric and positive definite (spd)
   *         linear systems <var>M</var>&sdot;<var>x</var> = <var>b</var> that
   *         computes a Cholesky <var>L</var>&sdot;<var>L</var><sup>T</sup>
   *         factorization.
   *
   * @param <T>
   *         type of the variables in the linear system
   *
   */
  template<typename T>
  class LLTPreconditioner : public Preconditioner<T>
  {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Type of the variables in the linear system. */
    using value_type = typename Preconditioner<T>::value_type;

    /** @brief Type for the dimensionality of the linear system. */
    using index_type = typename Preconditioner<T>::index_type;

    /** @brief Type for counting FLOPs. */
    using flops_type = typename Preconditioner<T>::flops_type;

  private:

    /** @brief `FullMatrix` type. */
    using full_matrix = FullMatrix<value_type>;

    /** @brief `FullVector` type. */
    using full_vector = FullVector<value_type>;

    /** @brief Type of the raw dense matrix implementation. */
    using raw_matrix = typename full_matrix::raw_matrix_type;

    /** @brief Type of the raw dense vector implementation. */
    using raw_vector = typename full_vector::raw_vector_type;

    /** @brief Internal representation of the factorization. */
    using llt_factorization = Eigen::LLT<raw_matrix>;


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief Cholesky factorization of the coefficient matrix. */
    llt_factorization factorization_ {};


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    /**
     * @brief
     *         Default constructor: Creates a new uninitialized
     *         `LLTPreconditioner`.
     *
     */
    LLTPreconditioner(const index_type size) : Preconditioner<T> {size}
    {
      // do nothing
    }


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  public:

    /**
     * @brief
     *         Tells that `LLTPreconditioner`s cannot take advantage of the
     *         special structure of hirachical matrices.
     *
     * @returns
     *         `false`
     *
     */
    virtual bool
    isRecursive() const noexcept final override
    {
      return false;
    }

    /**
     * @brief
     *         Tells that `LLTPreconditioner`s can only solve symmetric and
     *         positive definite (spd) linear systems.
     *
     * @returns
     *         `true`
     *
     */
    virtual bool
    onlySPD() const noexcept final override
    {
      return true;
    }

  protected:

    // {@inheritDoc}
    virtual std::string
    getName(std::vector<std::string>&) const final override
    {
      return "Cholesky";
    }

    /**
     * @brief
     *         Computes the factor <var>L</var>.
     *
     * @param cofficients
     *         coefficient matrix for the linear system
     *
     */
    virtual void
    initImpl(const HMatrix<value_type>& coefficients) final override
    {
      if (BSC_KLAMMLER_PARALLEL >= BSC_KLAMMLER_PARALLEL_DATA)
        enable_data_level_multithreading(this->getNumCPUs());
      else
        assert(enable_data_level_multithreading() == 1);
      if (auto dense_ptr = dynamic_cast<const full_matrix *const>(&coefficients))
        {
          this->factorization_.compute(dense_ptr->getRawMatrix());
        }
      else
        {
          const full_matrix dense {coefficients};
          this->factorization_.compute(dense.getRawMatrix());
        }
      if (BSC_KLAMMLER_COUNT_FLOPS)
        {
          namespace nc = numeric_cost;
          const flops_type sz {static_cast<flops_type>(this->size())};
          this->addInitFLOPs(nc::llt_factorization(sz));
        }
    }

    /**
     * @brief
     *         Solves the linear system
     *         <var>P</var>&sdot;<var>y</var>=<var>c</var> where <var>P</var>
     *         is the preconditioner.
     *
     * The system is solved by performing a forward-substitution on
     * <var>L</var> and a backward-substitution on <var>L</var><sup>T</sup>.
     *
     * @param right
     *         right-side to solve the system for
     *
     * @returns
     *         an (approximate) solution to the linear system
     *
     */
    virtual FullVector<value_type>
    solveImpl(const HVector<value_type>& right) const final override
    {
      if (BSC_KLAMMLER_COUNT_FLOPS)
        {
          namespace nc = numeric_cost;
          const flops_type sz {static_cast<flops_type>(this->size())};
          this->addSolveFLOPs(nc::forward_substitution(sz) +
                              nc::backward_substitution(sz));
        }
      if (auto dense_ptr = dynamic_cast<const full_vector *const>(&right))
        {
          return full_vector {
            this->factorization_.solve(dense_ptr->getRawVector())
          };
        }
      else
        {
          const full_vector dense {right};
          return full_vector {
            this->factorization_.solve(dense.getRawVector())
          };
        }
    }
  };  // class LLTPreconditioner
}  // bsc_klammler bsc_klammler

#endif  // ifndef BSC_KLAMMLER_LLT_PRECONDITIONER_HPP
