/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file logger.hpp
 *
 * @brief
 *         Header file providing the abstract `Logger` template class.
 *
 */

#ifndef BSC_KLAMMLER_LOGGER_HPP
#define BSC_KLAMMLER_LOGGER_HPP

#include <string>              // std::string

#include "h_matrix.hpp"        // bsc_klammler::HMatrix
#include "preconditioner.hpp"  // bsc_klammler::Preconditioner
#include "timer.hpp"           // bsc_klammler::Timer

namespace bsc_klammler
{

  /**
   * @brief
   *         Simple logger interface for reporting benchmark data to the user.
   *
   * Unfortunately, to support polymorphism, this class (rather than the
   * individual function members) needs to be a template, so it can only be
   * used with one floating point type.
   *
   * This class is not stateless although there is no guard against calling
   * functions in nonsensical order.  Here is an informal pattern that matches
   * the intended use:
   *
   *     template<typename T>
   *     void logging(Logger<T>& logger)
   *     {
   *       const HMatrix<T>& coeffs = get_coefficient_matrix();
   *       logger.logBefore(coeffs);
   *       for (Preconditioner<T>& precond : get_preconditioners())
   *         {
   *           logger.logPreconditioner(precond);
   *           if (!do_we_like_it(precond))
   *             {
   *               logger.logSkipping("we don't like it");
   *               continue;
   *             }
   *           Timer timer_init {};
   *           precond.init(coeffs, &timer_init);
   *           logger.logInit(precond, coeffs, timer_init.getSeconds());
   *           for (const HVector<T>& rhs : get_right_hand_sides())
   *             {
   *               Timer timer_solve {};
   *               FullVector<T> solved = precond.solve(rhs, &timer_solve);
   *               auto relres = compute_rel_residual(coeffs, rhs, solved);
   *               logger.logSolve(precond, relres, timer_solve.getSeconds());
   *             }
   *         }
   *       logger.logAfter();
   *     }
   *
   * @param <T>
   *         coefficient type for linear systems
   *
   */
  template<typename T>
  class Logger
  {
  public:

    /** @brief Coefficient type for linear systems. */
    using value_type = typename HMatrix<T>::value_type;

    /** @brief Index type for linear systems. */
    using index_type = typename HMatrix<T>::index_type;

    /** @brief Type for counting FLOPs. */
    using flops_type = typename Preconditioner<T>::flops_type;

    /**
     * @brief
     *         Virtual default destructor.
     *
     */
    virtual ~Logger() noexcept = default;

    /**
     * @brief
     *         Initializes a new logging session.
     *
     * @param coeffs
     *         coefficient matrix that will be decomposed
     *
     */
    virtual void
    logBefore(const HMatrix<T>& coeffs) = 0;

    /**
     * @brief
     *         Finalizes a logging sssion.
     *
     */
    virtual void
    logAfter() = 0;

    /**
     * @brief
     *         Indicates that a new `Preconditioner` will be used now.
     *
     * A `Logger` may print a message directly or cache some information about
     * the `Preconditioner` to be used when invoking later functions.
     *
     * @param precond
     *         `Preconditioner` that will now be used
     *
     */
    virtual void
    logPreconditioner(const Preconditioner<value_type>& precond) = 0;

    /**
     * @brief
     *         Logs statistics about a decomposition.
     *
     * @param precond
     *         `Preconditioner` that was initialized
     *
     * @param elapsed
     *         seconds elapsed while decomposing
     *
     */
    virtual void
    logInit(const Preconditioner<value_type>& precond,
            const Timer::time_delta_type elapsed) = 0;

    /**
     * @brief
     *         Logs statistics about solving a single linear system.
     *
     * @param precond
     *         `Preconditioner` that solved the linear system
     *
     * @param elapsed
     *         seconds elapsed while solving
     *
     * @param relative_residual
     *         relative residual left by the computed solution
     *
     */
    virtual void
    logSolve(const Preconditioner<value_type>& precond,
             const Timer::time_delta_type elapsed,
             const value_type relative_residual) = 0;

    /**
     * @brief
     *         Logs the fact that the last `Preconditioner` will be skipped.
     *
     * @param why
     *         reason for skipping the `Preconditioner`
     *
     */
    virtual void
    logSkipping(const std::string& why) = 0;
  };

}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_LOGGER_HPP
