/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file macros.hpp
 *
 * @brief
 *         Global macros that can be used in header files.
 *
 * All macros are only defined conditionally if they are not already defined.
 * The documented definitions are the fallback values and usually empty.  In
 * order to get the most out of this package, a user should `#define` (some of)
 * these macros to values appropriate for her platform prior to including the
 * first header file from this package.
 *
 * Our own implementation files (`*.cpp`) do this by including `global.hpp`
 * (instead of including this file directly) which will define better values by
 * using the results of Autoconf.
 *
 * Header files (`*.hpp`) must not `#include "global.hpp"` since it includes
 * `config.h` and defines all sorts of stuff.  This will lead to better
 * targeted code but pollutes the global namespace in a way that is
 * inacceptable for header files.  This header file on the other hand is very
 * clean and only defines macros in the `BSC_KLAMMLER_*` namespace.
 *
 */

#ifndef BSC_KLAMMLER_MACROS_HPP
#define BSC_KLAMMLER_MACROS_HPP

/**
 * @def BSC_KLAMMLER_DEBUG
 *
 * @brief
 *         Define to `false` to disable run-time checks in favour of
 *         performance; `true` by default unless `NDEBUG` is defined.
 *
 * The standard `assert.h` header only checks whether the `NDEBUG` macro is
 * defined, so it cannot / should not be defined to something false.
 * Therefore, it cannot be used as a boolean expression as in
 *
 *     if (!NDEBUG)
 *       check_something();
 *
 * which -- according to the GNU coding standards -- is the preferred form over
 * using
 *
 *     #ifndef NDEBUG
 *     check_something();
 *     #endif
 *
 * (Somewhat silly) example usage:
 *
 *     double
 *     largest(const double * numbers, size_t n)
 *     {
 *       const double max = numbers[n - 1];  // assume numbers are sorted
 *       if (BSC_KLAMMLER_DEBUG)
 *         {
 *           for (size_t i = 0; i < n; ++i)
 *             assert(items[i] <= max);
 *         }
 *       return max;
 *     }
 *
 */
#ifndef BSC_KLAMMLER_DEBUG
#  ifdef NDEBUG
#    if !(NDEBUG)
#      warning "Defining NDEBUG to something false is ambiguous"
#    endif
#    define BSC_KLAMMLER_DEBUG false
#  else
#    define BSC_KLAMMLER_DEBUG true
#  endif
#endif
static_assert(
    (!BSC_KLAMMLER_DEBUG) || (!!BSC_KLAMMLER_DEBUG),
    "Please '#define BSC_KLAMMLER_DEBUG' to a boolean literal");


/**
 * @def BSC_KLAMMLER_CHECK_ARGS
 *
 * @brief
 *         Define to a boolean / integral literal that indicates whether
 *         argument checking is to be performed.
 *
 * If undefined, this macro will be defined to `BSC_KLAMMLER_DEBUG`.
 *
 * Example usage:
 *
 *     void
 *     eat(const std::string& what)
 *     {
 *       if (BSC_KLAMMLER_CHECK_ARGS)
 *         {
 *           if (what == "toxic mushroom")
 *             throw std::invalid_argument {"Are you serious?"};
 *         }
 *       std::cout << "Eating a " << what << "..." << std::endl;
 *     }
 *
 */
#ifndef BSC_KLAMMLER_CHECK_ARGS
#  define BSC_KLAMMLER_CHECK_ARGS BSC_KLAMMLER_DEBUG
#endif
static_assert(
    (!BSC_KLAMMLER_CHECK_ARGS) || (!!BSC_KLAMMLER_CHECK_ARGS),
    "Please '#define BSC_KLAMMLER_CHECK_ARGS' to a boolean literal");


/**
 * @def BSC_KLAMMLER_COUNT_FLOPS
 *
 * @brief
 *         Define to a boolean / integral literal that indicates whether
 *         counting the (theoretical) number of floating point operations
 *         (FLOPS) is to be performed.
 *
 * If undefined, this macro will be defined to `true`.
 *
 */
#ifndef BSC_KLAMMLER_COUNT_FLOPS
#  define BSC_KLAMMLER_COUNT_FLOPS true
#endif
static_assert(
    (!BSC_KLAMMLER_COUNT_FLOPS) || (!!BSC_KLAMMLER_COUNT_FLOPS),
    "Please '#define BSC_KLAMMLER_COUNT_FLOPS' to a boolean literal");


/**
 * @def BSC_KLAMMLER_PARALLEL_NONE
 *
 * @brief
 *         Macro constant referring to parallelization level 0
 *         (no parallelization).
 *
 */
#define BSC_KLAMMLER_PARALLEL_NONE 0

/**
 * @def BSC_KLAMMLER_PARALLEL_STRUCTURAL
 *
 * @brief
 *         Macro constant referring to parallelization level 1
 *         (structural, ie recursive descet).
 *
 */
#define BSC_KLAMMLER_PARALLEL_STRUCTURAL 1

/**
 * @def BSC_KLAMMLER_PARALLEL_EXPRESSION
 *
 * @brief
 *         Macro constant referring to parallelization level 2
 *         (expression-level, eg sums of arranged matrices).
 *
 */
#define BSC_KLAMMLER_PARALLEL_EXPRESSION 2

/**
 * @def BSC_KLAMMLER_PARALLEL_DATA
 *
 * @brief
 *         Macro constant referring to parallelization level 3
 *         (data-level, eg single dense matrix products).
 *
 */
#define BSC_KLAMMLER_PARALLEL_DATA 3

/**
 * @def BSC_KLAMMLER_PARALLEL
 *
 * @brief
 *         Define to the desired parallelization level (default: 0).
 *
 * There are four levels of parallelism defined:
 *
 * 0. `BSC_KLAMMLER_PARALLEL_NONE`
 * 1. `BSC_KLAMMLER_PARALLEL_STRUCTURAL`
 * 2. `BSC_KLAMMLER_PARALLEL_EXPRESSION` (not implemented)
 * 3. `BSC_KLAMMLER_PARALLEL_DATA`
 *
 * Level-3 enables OpenMP.
 *
 */
#ifndef BSC_KLAMMLER_PARALLEL
#  define BSC_KLAMMLER_PARALLEL BSC_KLAMMLER_PARALLEL_NONE
#endif
static_assert(
    BSC_KLAMMLER_PARALLEL >= BSC_KLAMMLER_PARALLEL_NONE
    && BSC_KLAMMLER_PARALLEL <= BSC_KLAMMLER_PARALLEL_DATA,
    "Please #define BSC_KLAMMLER_PARALLEL to an integer between 0 and 3");

#if BSC_KLAMMLER_PARALLEL == BSC_KLAMMLER_PARALLEL_EXPRESSION
#  warning "BSC_KLAMMLER_PARALLEL_EXPRESSION: not implemented"
#endif

#if BSC_KLAMMLER_PARALLEL < BSC_KLAMMLER_PARALLEL_DATA
#  ifdef _OPENMP
#    warning "Using OpenMP but data-level parallelization not enabled"
#  endif
#  ifndef EIGEN_DONT_PARALLELIZE
#    define EIGEN_DONT_PARALLELIZE
#  endif
#else
#  ifndef _OPENMP
#    warning "Data-level parallelization enabled but not using OpenMP"
#  endif
#  ifdef EIGEN_DONT_PARALLELIZE
#    warning "Data-level parallelization enabled but EIGEN_DONT_PARALLELIZE defined"
#  endif
#endif


/**
 * @def BSC_KLAMMLER_WITH_NEST
 *
 * @brief
 *         Define to `true` to use the NeSt algorithm for dense matrix
 *         inversion and to `false` otherwise.
 *
 * If you want to use the NeSt algorithm, you'll need its source code, which
 * unfortunately cannot be part of this project, in the directory `nest`.  The
 * default is `false`.
 *
 */
#ifndef BSC_KLAMMLER_WITH_NEST
#  define BSC_KLAMMLER_WITH_NEST false
#endif

static_assert(
    (!BSC_KLAMMLER_WITH_NEST) || (!!BSC_KLAMMLER_WITH_NEST),
    "Please '#define BSC_KLAMMLER_WITH_NEST' to a boolean literal");


/**
 * @def BSC_KLAMMLER_INDEX_TYPE
 *
 * @brief
 *         Define to an integer type suitable for specifying matrix and vector
 *         dimensions and subscripts.
 *
 * The default is `std::size_t` as defined in the standard header `cstddef`.
 * This header is `#include`d automatically.  This macro shall not be used in
 * normal code but only to `typedef` more high-level type aliases.
 *
 */
#ifndef BSC_KLAMMLER_INDEX_TYPE
#include <cstddef>
#define BSC_KLAMMLER_INDEX_TYPE std::ptrdiff_t
#endif
#include <limits>
static_assert(
    std::numeric_limits<BSC_KLAMMLER_INDEX_TYPE>::is_integer
    && std::numeric_limits<BSC_KLAMMLER_INDEX_TYPE>::is_signed,
    "Please '#define BSC_KLAMMLER_INDEX_TYPE' to a signed integer type");


/**
 * @def BSC_KLAMMLER_RESTRICT
 *
 * @brief
 *         Define to the `restrict` &ldquo;keyword&rdquo; if your compiler
 *         supports it.
 *
 * This macro can be used to qualify pointers as `restrict`ed which may lead to
 * more efficient code and eventually warnings on compilers that support it.
 *
 * Example usage:
 *
 *     void
 *     copy_chars(const char *BSC_KLAMMLER_RESTRICT src,
 *                char *BSC_KLAMMLER_RESTRICT dst)
 *     {
 *       // The compiler may exploit that 'src' and 'dst' are not aliased here.
 *       do
 *         *dst++ = *src++;
 *       while (*src);
 *       *dst = 0;
 *     }
 *
 *     void
 *     buggy()
 *     {
 *       char * data = "hello, world";
 *       copy_chars(data, data);  // hope for a compiler warning
 *     }
 *
 * If using GCC:
 *
 *     #define BSC_KLAMMLER_RESTRICT __restrict__
 *
 */
#ifndef BSC_KLAMMLER_RESTRICT
#  define BSC_KLAMMLER_RESTRICT /* empty */
#endif


/**
 * @def BSC_KLAMMLER_ATTRIBUTE_UNUSED
 *
 * @brief
 *         Define to a function parameter annotation for unused parameters if
 *         your compiler supports it.
 *
 * This macro can then be applied to function arguments that are required to
 * match a certain signature but are not used inside the function's body.  On
 * compilers that support it, this supresses a warning (`-Wunused-parameter`
 * for GCC).  Prefer using this macro over tricks like casting to void.
 *
 * Example usage:
 *
 *     int
 *     f(int a, int b BSC_KLAMMLER_ATTRIBUTE_UNUSED)
 *     {
 *         // parameter 'b' is not used here
 *         return a;
 *     }
 *
 * If using GCC:
 *
 *     #define BSC_KLAMMLER_ATTRIBUTE_UNUSED __attribute__((unused))
 *
 */
#ifndef BSC_KLAMMLER_ATTRIBUTE_UNUSED
#  define BSC_KLAMMLER_ATTRIBUTE_UNUSED /* empty */
#endif


/**
 * @def BSC_KLAMMLER_ATTRIBUTE_DEPRECATED
 *
 * @brief
 *         Define to a function annotation for deprecated functions if your
 *         compiler supports it.
 *
 * On compilers that support it, this produces a warning if the deprecated
 * function is used (`-Wdeprecated` for GCC).
 *
 * Example usage:
 *
 *     int
 *     good_one(int first, int second);
 *
 *     int BSC_KLAMMLER_ATTRIBUTE_DEPRECATED
 *     bad_one(int second, int first)
 *     {
 *       return good_one(first, second);
 *     }
 *
 *     void
 *     not_up_to_date()
 *     {
 *       bad_one(14, 92);  // compiler warning
 *     }
 *
 * If using GCC:
 *
 *     #define BSC_KLAMMLER_ATTRIBUTE_DEPRECATED __attribute__((deprecated))
 *
 */
#ifndef BSC_KLAMMLER_ATTRIBUTE_DEPRECATED
#  define BSC_KLAMMLER_ATTRIBUTE_DEPRECATED /* empty */
#endif


/**
 * @def BSC_KLAMMLER_ATTRIBUTE_WARNING
 *
 * @brief
 *         Define to a function annotation for functions to warn about if not
 *         all calls get eliminated by dead-code analysis if your compiler
 *         supports it.
 *
 * Example usage:
 *
 *     int
 *     eliminate_this() BSC_KLAMMLER_ATTRIBUTE_WARNING("not eliminated")
 *     {
 *       return -1;
 *     }
 *
 *     void
 *     good
 *     {
 *       constexpr bool predicate {sizeof(double) > sizeof(float)};
 *       if (!predicate)
 *         eliminate_this();  // ok: dead code, will be eliminated
 *     }
 *
 *     void
 *     bad(int a, int b)
 *     {
 *       if (a > b)
 *         eliminate_this();  // warning: cannot be eliminated
 *     }
 *
 * If using GCC:
 *
 *     #define BSC_KLAMMLER_ATTRIBUTE_WARNING(M) __attribute__((warning(M)))
 *
 */
#ifndef BSC_KLAMMLER_ATTRIBUTE_WARNING
#  define BSC_KLAMMLER_ATTRIBUTE_WARNING(M) /* empty */
#endif


/**
 * @def BSC_KLAMMLER_PURE_VIRTUAL
 *
 * @brief
 *         Macro to make a function member pure virtual with a fallback for
 *         buggy compilers.
 *
 * Use this macro to declare a a purely virtual function member in templated
 * types if a GCC bug will cause troubles with local types.  It is reccommended
 * to only use this hack if needed and not by default.
 *
 * Example usage:
 *
 *     template<typename T>
 *     class VirtualBase
 *     {
 *     public:
 *
 *       virtual T
 *       getIt() BSC_KLAMMLER_PURE_VIRTUAL;
 *
 *       T
 *       getItTwice()
 *       {
 *         return this->getIt() + this->getIt();
 *       }
 *     };
 *
 *     int
 *     f()
 *     {
 *       class LocalDerived : public VirtualBase<int>
 *       {
 *       public:
 *
 *         virtual int
 *         getIt() override
 *         {
 *           return 42;
 *         }
 *       };
 *       return LocalDerived {}.getItTwice();
 *     }
 *
 *     void
 *     g()
 *     {
 *       VirtualBase vb {};  // error: type is abstract
 *       vb.getIt();         // error: it should have never come to this
 *     }
 *
 * In this example, `f` will be fine even if the compiler suffers from the
 * mentioned Gcc bug.  `g` clearly is an error and will not compile if the
 * compiler is free from the bug.  If the compiler bug is present, it will
 * compile and the consecutive call to `VirtualBase::getIt` will be a runtime
 * error.
 *
 * @see
 *         - https://gcc.gnu.org/bugzilla/show_bug.cgi?id=51048
 *         - https://stackoverflow.com/questions/18994041/templates-inner-structs-local-types-and-pure-virtual-functions-oh-my
 *
 */
#if BSC_KLAMMLER_GCC_BUG_51048
#  define BSC_KLAMMLER_PURE_VIRTUAL { throw __func__; }
#else
#  define BSC_KLAMMLER_PURE_VIRTUAL = 0
#endif


// TODO: Not sure where to put these...
#define BSC_KLAMMLER_EOL '\n'
#define BSC_KLAMMLER_LABEL __FILE__ << ":" << __LINE__ << ": "
#define BSC_KLAMMLER_PICKET_FENCE                                             \
  std::cout << BSC_KLAMMLER_LABEL << "PICKET FENCE" << BSC_KLAMMLER_EOL

#endif  // ifndef BSC_KLAMMLER_MACROS_HPP
