/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


#include "global.hpp"   // (package configuration)

#include <array>        // std::array
#include <cassert>      // assert
#include <cstdlib>      // std::getenv, EXIT_SUCCESS, EXIT_FAILURE
#include <fstream>      // std::ofstream
#include <iomanip>      // std::setw
#include <iostream>     // std::cout, ...
#include <sstream>      // std::stringstream
#include <string>       // std::string
#include <tuple>        // std::tuple
#include <type_traits>  // std::enable_if, std::is_floating_point, ...
#include <vector>       // std::vector

#include <boost/format.hpp>               // boost::format
#include <boost/program_options.hpp>

#include "algorithms.hpp"                 // (linear algebra operations)
#include "block_ldlt_preconditioner.hpp"  // bsc_klammler::BlockLDLTPreconditioner
#include "block_llt_preconditioner.hpp"   // bsc_klammler::BlockLLTPreconditioner
#include "block_matrix.hpp"               // bsc_klammler::BlockMatrix
#include "full_matrix.hpp"                // bsc_klammler::FullMatrix
#include "full_vector.hpp"                // bsc_klammler::FullVector
#include "inverse_preconditioner.hpp"     // bsc_klammler::InversePreconditioner
#include "llt_preconditioner.hpp"         // bsc_klammler::LLTPreconditioner
#include "not_reached.hpp"                // BSC_KLAMMLER_NOT_REACHED
#include "random_problems.hpp"            // bsc_klammler::get_random_*
#include "simple_logger.hpp"              // bsc_klammler::SimpleLogger
#include "sql_logger.hpp"                 // bsc_klammler::SqlLogger
#include "timer.hpp"                      // bsc_klammler::Timer

#include <Eigen/Core>                     // Eigen::initParallel, ...


#define STRINGIZE_IMPL(X) #X
#define STRINGIZE(X) STRINGIZE_IMPL(X)

namespace /* anonymous */
{

  using namespace bsc_klammler;

  template<typename T>
  std::tuple<Timer::time_delta_type, Timer::time_delta_type>
  demo(const unsigned recursions,
       const typename HMatrix<T>::index_type block_min,
       const typename HMatrix<T>::index_type block_max,
       const bool spd,
       const unsigned num_cpus,
       /* !const */ Logger<T>& logger)
  {
    using value_type = T;
    using full_vector = FullVector<value_type>;
    using block_matrix = BlockMatrix<value_type>;
    using full_matrix = FullMatrix<value_type>;
    using preconditioner_type = Preconditioner<value_type>;
    using precond_uptr_type = std::unique_ptr<preconditioner_type>;
    Timer timer_setup {};
    Timer::time_delta_type computing_total {};
    timer_setup.start();
    constexpr unsigned max_non_recursive {3};
    constexpr unsigned solve_sample_size {3};
    if (BSC_KLAMMLER_DEBUG && recursions > max_non_recursive)
      throw std::invalid_argument {"problem is too big for a debug build"};
    const block_matrix coeffs {
      get_random_block_matrix<value_type>(recursions,
                                          block_min,
                                          block_max,
                                          spd)
    };
    const full_matrix coeffs_full {
      recursions <= max_non_recursive
        ? full_matrix {coeffs}
        : full_matrix {}
    };
    std::array<full_vector, solve_sample_size> right_sides;
    for (unsigned i = 0; i < solve_sample_size; ++i)
      {
        right_sides.at(i) = get_random_vector<value_type>(coeffs.size());
      }
    assert(!spd || is_symmetric(coeffs));
    assert(!spd || is_positive_definite(coeffs));
    constexpr std::size_t nprec {4};
    std::array<precond_uptr_type, nprec> preconditioners = {
      precond_uptr_type {
        new LLTPreconditioner<value_type> {coeffs.size()}
      },
      precond_uptr_type {
        new InversePreconditioner<value_type> {coeffs.size(),spd}
      },
      precond_uptr_type {
        new BlockLDLTPreconditioner<value_type> {coeffs.size()}
      },
      precond_uptr_type {
        new BlockLLTPreconditioner<value_type> {coeffs.size()}
      },
    };
    timer_setup.stop();
    logger.logBefore(coeffs);
    for (auto& iter : preconditioners)
      {
        // Move the pointer out of the array to make sure we release the memory
        // for the decomposition so the next preconditioner can have it again.
        precond_uptr_type precond_uptr {std::move(iter)};
        Timer timer_init {};
        precond_uptr->setNumCPUs(num_cpus);
        logger.logPreconditioner(*precond_uptr);
        if (!spd && precond_uptr->onlySPD())
          {
            logger.logSkipping(
                "only applicable to symmetric and positive definite matrices");
            continue;
          }
        if (recursions > max_non_recursive && !precond_uptr->isRecursive())
          {
            logger.logSkipping(
                "not applying a non-recursive algorithm to a huge recursive"
                " matrix");
            continue;
          }
        if (precond_uptr->isRecursive())
          precond_uptr->init(coeffs, &timer_init);
        else
          precond_uptr->init(coeffs_full, &timer_init);
        logger.logInit(*precond_uptr, timer_init.getSeconds());
        computing_total += timer_init.getSeconds();
        for (const auto& right : right_sides)
          {
            Timer timer_solve {};
            const full_vector solved {precond_uptr->solve(right, &timer_solve)};
            const full_vector residual {coeffs * solved - right};
            const value_type relative_residual {norm(residual) / norm(right)};
            logger.logSolve(
                *precond_uptr, timer_solve.getSeconds(), relative_residual);
            computing_total += timer_solve.getSeconds();
          }
      }
    logger.logAfter();
    return std::make_tuple(timer_setup.getSeconds(), computing_total);
  }

  /**
   * @brief
   *         Returns the version information of Eigen formatted as a string.
   *
   * The returned string is one English sentence starting with a capital letter
   * and ending with a period but no line break.
   *
   * @returns
   *         Eigen version information
   *
   */
  std::string
  get_eigen_version_string()
  {
    std::ostringstream oss {};
    oss << "Using Eigen "
        << EIGEN_WORLD_VERSION
        << "."
        << EIGEN_MAJOR_VERSION
        << "."
        << EIGEN_MINOR_VERSION
#ifdef EIGEN_USE_MKL_ALL
        << " linked to Intel MKL"
#endif
        << " for general dense linear algebra.";
    return oss.str();
  }

  /**
   * @brief
   *         Returns an informal string naming the parallelization level and
   *         the number of CPUs used.
   *
   * The returned string is one English sentence starting with a capital letter
   * and ending with a period but no line break.
   *
   * @param ncpus
   *         number of CPUs that will be used (0 means unknown)
   *
   * @returns
   *         informal description of parallelization level
   *
   */
  std::string
  get_parallel_level_string(const unsigned ncpus)
  {
    std::ostringstream oss {};
    oss << "Level-" << BSC_KLAMMLER_PARALLEL << " (";
    switch (BSC_KLAMMLER_PARALLEL)
      {
      case BSC_KLAMMLER_PARALLEL_NONE:
        oss << "none";
        break;
      case BSC_KLAMMLER_PARALLEL_STRUCTURAL:
        oss << "structural";
        break;
      case BSC_KLAMMLER_PARALLEL_EXPRESSION:
        oss << "expression-level";
        break;
      case BSC_KLAMMLER_PARALLEL_DATA:
        oss << "data-level";
        break;
      default:
        oss << "?";
        break;
      }
    oss << ") parallelization";
    if (BSC_KLAMMLER_PARALLEL)
      {
        if (ncpus)
          oss << " using up to " << ncpus << " CPUs";
        else
          oss << " using all available CPUs";
      }
    oss << " enabled.";
    return oss.str();
  }

  /**
   * @brief
   *         Returns an informal string telling the compiler version, comile
   *         date and time and the host architecture type.
   *
   * The returned string is one English sentence starting with a capital letter
   * and ending with a period but no line break.
   *
   * @returns
   *         informal description of compiler and host
   *
   */
  std::string
  get_compiler_string()
  {
    return
      "Compiled by " CXX_COMPILER_STRING
      " on " __DATE__ " " __TIME__
      " for " CANONICAL_HOST ".";
  }

  /**
   * @brief
   *         Returns an informal string telling the floating point type.
   *
   * The returned string is one English sentence starting with a capital letter
   * and ending with a period but no line break.
   *
   * @returns
   *         informal name of the floating point type
   *
   */
  std::string
  get_floating_point_type_string()
  {
    return
      "Using floating point type "
      "'" STRINGIZE(BSC_KLAMMLER_FLOATING_POINT_TYPE) "'"
      ".";
  }

  /**
   * @brief
   *         Prints the program banner.
   *
   * Unless the exception bit is set on the stream, errors will be ignored.
   * The stream will be flushed after the banner was printed.
   *
   * @param os
   *         stream to write to
   *
   * @param ncpus
   *         number of CPUs that will be used
   *
   */
  void
  print_banner(std::ostream& os, const unsigned ncpus)
  {
    std::vector<std::string> lines {};
    lines.push_back(PACKAGE_STRING);
    lines.push_back("Copyright (C) " PACKAGE_YEAR " " PACKAGE_AUTHOR);
    lines.push_back("Please report bugs to <" PACKAGE_BUGREPORT ">.");
    lines.push_back(get_compiler_string());
    lines.push_back(get_eigen_version_string());
    if (BSC_KLAMMLER_WITH_NEST)
      lines.push_back("Using NeSt for parallel matrix inversion.");
    lines.push_back(get_floating_point_type_string());
    lines.push_back(get_parallel_level_string(ncpus));
    if (BSC_KLAMMLER_DEBUG)
      lines.push_back("Warning: This is a debug build. "
                      " Do not use it for performance measurements!");
    for (const auto& line : lines)
      os << line << BSC_KLAMMLER_EOL;
    os << std::flush;
  }

  /**
   * @brief
   *         Prints an execution time summary.
   *
   * Unless the exception bit is set on the stream, errors will be ignored.
   * The stream will be flushed after the banner was printed.
   *
   * @param os
   *         stream to write to
   *
   * @param setup
   *         total seconds of random problem generation time
   *
   * @param computing
   *         total seconds of usful computation time
   *
   * @param total
   *         total seconds of program execution time (well, almost&hellip;)
   *
   */
  void
  print_total_stats(std::ostream& os,
                    const Timer::time_delta_type setup,
                    const Timer::time_delta_type computing,
                    const Timer::time_delta_type total)
  {
    os << BSC_KLAMMLER_EOL;
    os << boost::format("%-32s%15.06f s      (%5.01f %%)")
      % "Problem generation overhead:"
      % setup
      % (100.0 * setup / total)
       << BSC_KLAMMLER_EOL;
    os << boost::format("%-32s%15.06f s      (%5.01f %%)")
      % "Total (useful) computation time:"
      % computing
      % (100.0 * computing / total)
       << BSC_KLAMMLER_EOL;
    os << boost::format("%-32s%15.06f s      (%5.01f %%)")
      % "Total program execution time:"
      % total
      % (100.0 * total / total)
       << BSC_KLAMMLER_EOL;
    os << std::flush;
  }

  struct CommandLineArgs
  {
    unsigned int recursions {3};
    unsigned int block_min {BSC_KLAMMLER_DEBUG ? 16 : 128};
    unsigned int block_max {BSC_KLAMMLER_DEBUG ? 32 : 512};
    bool spd {false};
    unsigned ncpus {0};  // use what is available
    bool sql {false};
    InefficientAlgorithm::Policies policy {InefficientAlgorithm::Policies::IGNORE};
  };

  /**
   * @brief
   *         Parses the command line arguments.
   *
   * If the arguments are invalid, an exception is thrown.  If the `--help` or
   * `--version` option was presnt, `false` is returned to indicate the program
   * shall exit immediately.  Otherwise `true` is returned.
   *
   * @param cla
   *         default parameters to update in-place with the user's choices
   *
   * @param args
   *         command line arguments (excluding the program name `argv[0]`)
   *
   * @returns
   *         `false` if the program shall exit, else `true`
   *
   * @throws std::exception
   *         if the arguments are invalid
   *
   */
  bool
  parse_args(CommandLineArgs& cla, const std::vector<std::string> &args)
  {
    namespace po = boost::program_options;
    po::options_description desc {"Options"};
    desc.add_options()
      ("help", "show this message and exit")
      ("version", "show program version and exit")
      ("recursions", po::value<unsigned>(&cla.recursions),
       "recursion depth in the generated random matrix")
      ("block-size-min", po::value<unsigned>(&cla.block_min),
       "lower bound for the smallest block size in the generated random"
       " matrix")
      ("block-size-max", po::value<unsigned>(&cla.block_max),
       "upper bound for the smallest block size in the generated random"
       " matrix")
      ("spd", po::bool_switch(&cla.spd),
       "create a symmetric and positive definite random matrix")
      ("cpus", po::value<unsigned>(&cla.ncpus), "number of CPUs to use")
      ("sql", po::bool_switch(&cla.sql),
       "produce output in machine-readable SQL syntax")
      ("policy", po::value<std::string>(), "policy for dealing with"
       " accidentially invoked inefficient algorithms (IGNORE, WARN, THROW)");
    po::variables_map vm;
    po::command_line_parser parser {args};
    parser.options(desc);
    po::store(parser.run(), vm);
    if (vm.count("help"))
      {
        std::cout << desc << std::endl;
        return false;
      }
    if (vm.count("version"))
      {
        print_banner(std::cout, 0);
        return false;
      }
    po::notify(vm);
    if (vm.count("policy"))
      cla.policy = InefficientAlgorithm::policyFromString(
          vm["policy"].as<std::string>());
    if (vm.count("block-size-min") && !vm.count("block-size-max"))
      cla.block_max = cla.block_min;
    if (!vm.count("block-size-min") && vm.count("block-size-max"))
      cla.block_min = cla.block_max;
    if (cla.block_min > cla.block_max)
      {
        assert(vm.count("block-size-min") && vm.count("block-size-max"));
        throw std::invalid_argument {"block-size-min > block-size-max"};
      }
    return true;
  }

  /**
   * @brief
   *         Sets up the multi-threading environment.
   *
   * @param ncpus
   *         number of CPUs to use (0 means auto-detect)
   *
   * @returns
   *         number of CPUs to be used (either the argument or the guess)
   *
   */
  unsigned
  setup_multi_threading(const unsigned ncpus)
  {
    if (!BSC_KLAMMLER_PARALLEL && ncpus > 1)
      {
        throw std::invalid_argument {
          "compiled without concurrency support: cannot use more than 1 CPU"
        };
      }
    const unsigned actual_num_cpus {
      ncpus
        ? ncpus
        : (BSC_KLAMMLER_PARALLEL ? std::thread::hardware_concurrency() : 1)
    };
    if (BSC_KLAMMLER_PARALLEL)
      {
        Eigen::initParallel();
      }
    if (BSC_KLAMMLER_PARALLEL >= BSC_KLAMMLER_PARALLEL_DATA)
      {
        enable_data_level_multithreading(actual_num_cpus);
      }
    return actual_num_cpus;
  }
}  // namespace /* anonymous */


int
main(int argc, char * * argv)
{
  typedef BSC_KLAMMLER_FLOATING_POINT_TYPE value_type;
  try
    {
      Timer timer_total {};
      CommandLineArgs cla {};
      std::unique_ptr<bsc_klammler::Logger<value_type>> logger_uptr {};
      timer_total.start();
      std::cin.exceptions(std::istream::failbit);
      std::cout.exceptions(std::ostream::failbit);
      std::cerr.exceptions(std::ostream::failbit);
      std::clog.exceptions(std::ostream::failbit);
      {
        std::vector<std::string> args {};
        assert(argc > 0);
        args.reserve(argc - 1);
        for (std::size_t i = 1; i < static_cast<std::size_t>(argc); ++i)
          args.push_back(argv[i]);
        if (!parse_args(cla, args))
          return EXIT_SUCCESS;
      }
      bsc_klammler::InefficientAlgorithm::setPolicy(cla.policy);
      cla.ncpus = setup_multi_threading(cla.ncpus);
      print_banner(std::cerr, cla.ncpus);
      std::cerr << BSC_KLAMMLER_EOL;
      if (cla.sql)
        logger_uptr.reset(new bsc_klammler::SQLLogger<value_type> {
            cla.recursions, cla.block_min, cla.block_max, cla.spd});
      else
        logger_uptr.reset(new bsc_klammler::SimpleLogger<value_type> {
            cla.ncpus});
      const auto timing = demo<value_type>(
          cla.recursions, cla.block_min, cla.block_max, cla.spd, cla.ncpus,
          *logger_uptr);
      timer_total.stop();
      print_total_stats(std::cerr,
                        std::get<0>(timing), std::get<1>(timing),
                        timer_total.getSeconds());
      return EXIT_SUCCESS;
    }
  catch (const std::exception& e)
    {
      std::cerr << PACKAGE << ": error: " << e.what() << std::endl;
      return EXIT_FAILURE;
    }
}
