/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file matrix_anchor.hpp
 *
 * @brief
 *         Header file providing the `MatrixAnchor` template class.
 *
 */

#ifndef BSC_KLAMMLER_MATRIX_ANCHOR_HPP
#define BSC_KLAMMLER_MATRIX_ANCHOR_HPP

#include <cassert>        // assert
#include <stdexcept>      // std::invalid_argument

#include "h_matrix.hpp"   // HMatrix
#include "macros.hpp"     // (package-specific macros)
#include "rectangle.hpp"  // bsc_klammler::Rectangle
#include "utility.hpp"    // bsc_klammler::const_if


namespace bsc_klammler
{


  /**
   * @brief
   *         Triple of two coordinates and a pointer to an `HMatrix`.
   *
   * This is an immutable POD type.  There is no ownership taken for the
   * pointed-to `HMatrix`.
   *
   *     (0, 0)  origin of outer reference frame
   *     +------------------------------------->
   *     |
   *     |
   *     |      (row, col)  anchor of matrix
   *     |      +--------------+
   *     |      |w    north    |e
   *     |      |e             |a
   *     |      |s             |s
   *     |      |t             |t
   *     |      +--------------+
   *     |            south
   *     |
   *     V
   *
   * @param <T>
   *         type of the matrix entries
   *
   * @param <M>
   *         whether the anchored matrix is mutable
   *
   */
  template<typename T, bool M>
  class MatrixAnchor
  {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Convenience type alias for the `HMatrix` class. */
    using h_matrix = HMatrix<T>;

    /** @brief Correctly `const`-qualified type of the `HMatrix` class. */
    using qualified_h_matrix = typename const_if<h_matrix, !M>::type;

    /** @brief Type for elements. */
    using value_type = typename h_matrix::value_type;

    /** @brief Type for indices. */
    using index_type = typename h_matrix::index_type;


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief Pointer to the matrix data. */
    qualified_h_matrix * matrixPtr_ {};

    /** @brief Row index of the top-left element. */
    index_type row_ {};

    /** @brief Column index of the top-left element. */
    index_type col_ {};


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    /**
     * @brief
     *         Default constructor: Creates an empty anchor.
     *
     */
    MatrixAnchor() noexcept
    {
      assert(this->getMatrix() == nullptr);
      assert(this->rows() == 0);
      assert(this->cols() == 0);
      assert(this->north() == 0);
      assert(this->west() == 0);
      assert(this->south() == 0);
      assert(this->east() == 0);
    }

    /**
     * @brief
     *         Constructor: Creates a new `MatrixAnchor`.
     *
     * @param matrix_ptr
     *         address of matrix
     *
     * @param row
     *         row index of the (0, 0) element
     *
     * @param col
     *         column index of the (0, 0) element
     *
     * @throws std::invalid_argument
     *         if a `nullptr` is passed
     *
     */
    MatrixAnchor(qualified_h_matrix *const matrix_ptr,
                 const index_type row,
                 const index_type col) :
      matrixPtr_ {matrix_ptr}, row_ {row}, col_ {col}
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        {
          if (matrix_ptr == nullptr)
            throw std::invalid_argument {"matrix_ptr == nullptr"};
        }
    }

    /** @brief Default copy constructor. */
    MatrixAnchor(const MatrixAnchor&) noexcept = default;

    /** @brief Default move constructor. */
    MatrixAnchor(MatrixAnchor&&) noexcept = default;


    /* ------------- */
    /*   Operators   */
    /* ------------- */

    /** @brief Default copy-assignment operator. */
    MatrixAnchor&
    operator=(const MatrixAnchor&) noexcept = default;

    /** @brief Default move-assignment operator. */
    MatrixAnchor&
    operator=(MatrixAnchor&&) noexcept = default;



    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

    /**
     * @brief
     *         Returns the number of rows in the anchored `HMatrix`.
     *
     * @returns
     *         number of rows
     *
     */
    index_type
    rows() const noexcept
    {
      return this->matrixPtr_
        ? this->matrixPtr_->rows()
        : 0;
    }

    /**
     * @brief
     *         Returns the number of columns in the anchored `HMatrix`.
     *
     * @returns
     *         number of columns
     *
     */
    index_type
    cols() const noexcept
    {
      return this->matrixPtr_
        ? this->matrixPtr_->cols()
        : 0;
    }

    /**
     * @brief
     *         Returns the row index of the element of the anchored `HMatrix`.
     *
     * @returns
     *         row index of (0, 0) the element
     *
     */
    index_type
    north() const noexcept
    {
      return this->row_;
    }

    /**
     * @brief
     *         Returns the index the row below of the anchored `HMatrix`.
     *
     * @returns
     *         `north() + rows()`
     *
     */
    index_type
    south() const noexcept
    {
      return this->matrixPtr_
        ? this->row_ + this->matrixPtr_->rows()
        : 0;
    }

    /**
     * @brief
     *         Returns the column index of the anchored `HMatrix`.
     *
     * @returns
     *         column index of (0, 0) the element
     *
     */
    index_type
    west() const noexcept
    {
      return this->col_;
    }

    /**
     * @brief
     *         Returns the index of the column right of the anchored `HMatrix`.
     *
     * @returns
     *         `west() + cols()`
     *
     */
    index_type
    east() const noexcept
    {
      return this->matrixPtr_
        ? this->col_ + this->matrixPtr_->cols()
        : 0;
    }

    /**
     * @brief
     *         Returns a pointer to the anchored `HMatrix`.
     *
     * @returns
     *         pointer to the anchored `HMatrix`
     *
     */
    qualified_h_matrix *
    getMatrix() const noexcept
    {
      return this->matrixPtr_;
    }
  };  // class MatrixAnchor


  /* ------------------ */
  /*   Free Functions   */
  /* ------------------ */

  /**
   * @brief
   *         Checks whether two the anchored `HMatrix`es overlap.
   *
   * Two anchored matrices <var>A</var> and <var>B</var> are considered
   * overlapping if and only if there are two pairs of valid indices
   * (<var>i</var>, <var>j</var>) and (<var>k</var>, <var>l</var>) such that
   * <var>A<sub>ij</sub></var> and <var>A<sub>kl</sub></var> address the same
   * index in the outer reference frame.  Note that this means that an empty
   * matrix does not overlap anything in a trivial sense.
   *
   * @param <T>
   *         type of matrix elements
   *
   * @param <M1>
   *         whether the first `MatrixAnchor` is mutable
   *
   * @param <M2>
   *         whether the second `MatrixAnchor` is mutable
   *
   * @param first
   *         first `MatrixAnchor`
   *
   * @param second
   *         second `MatrixAnchor`
   *
   * @returns
   *         whether the two `MatrixAnchor`s overlap with at least one
   *         element
   *
   */
  template<typename T, bool M1, bool M2>
  bool
  overlap(const MatrixAnchor<T, M1>& first,
          const MatrixAnchor<T, M2>& second) noexcept
  {
    using index_type = typename MatrixAnchor<T, M1>::index_type;
    const Rectangle<index_type> r1 {
      first.north(), first.west(), first.south(), first.east()
    };
    const Rectangle<index_type> r2 {
      second.north(), second.west(), second.south(), second.east()
    };
    return overlap(r1, r2);
  }
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_MATRIX_ANCHOR_HPP
