/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file matrix_view.hpp
 *
 * @brief
 *         Header file providing the `MutableMatrixView` and `ConstMatrixView`
 *         template classes.
 *
 */

#ifndef BSC_KLAMMLER_MATRIX_VIEW_HPP
#define BSC_KLAMMLER_MATRIX_VIEW_HPP

#include "h_matrix.hpp"         // bsc_klammler::HMatrix

#include <stdexcept>            // std::invalid_argument
#include <utility>              // std::swap

#include "macros.hpp"           // (package-specific macros)
#include "not_implemented.hpp"  // BSC_KLAMMLER_NOT_IMPLEMENTED
#include "not_reached.hpp"      // BSC_KLAMMLER_NOT_REACHED
#include "rectangle.hpp"        // bsc_klammler::Rectangle
#include "utility.hpp"          // bsc_klammler::const_if


namespace bsc_klammler
{


  /**
   * @brief
   *         A (possibly partial) view of a (possibly transposed) `HMatrix`.
   *
   * View objects do not take ownership of the viewed matrix and are therefore
   * lightweight objects.  However, this means that care must be taken not to
   * create views that outlive the viewed matrix.
   *
   *     MatrixView<double, true>
   *     buggy()
   *     {
   *       FullMatrix<double> * inner_ptr = new FullMatrix<double> {};
   *       MatrixView<double, true> outer {inner_ptr, 0, 0, 2, 2};
   *       return outer;
   *       // 'outer' does not own 'inner_ptr' which is therefore leaked.
   *     }
   *
   *     MatrixView<double, true>
   *     also_buggy()
   *     {
   *       FullMatrix<double> inner {{11.0, 12.0}, {21.0, 22.0}};
   *       MatrixView<double, true> outer {&inner, 0, 0, 2, 2};
   *       return outer;
   *       // 'inner' goes out of scope -- 'outer' contains a dangling pointer.
   *     }
   *
   * It is also important to know that the `clone` function member is amputated
   * and must never be called.  Non-mutable `MatrixView`s also amputate the
   * `set` function member and don't accept `ActiveVisitor`s.
   *
   * @param <T>
   *         type of the matrix entries
   *
   * @param <M>
   *         wheter the view is mutable (ie not constant)
   *
   */
  template<typename T, bool M>
  class MatrixView : public HMatrix<T>
  {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Type of the super-class. */
    using super_type = HMatrix<T>;

    /** @brief Correctly `const`-qualified type of the super-class. */
    using qualified_super_type = typename const_if<super_type, !M>::type;

    /** @brief Type for elements. */
    using value_type = typename super_type::value_type;

    /** @brief Type for indices. */
    using index_type = typename super_type::index_type;

    /** @brief Imported type for passive visitors. */
    using PassiveVisitor = typename super_type::PassiveVisitor;

    /** @brief Imported type for active visitors. */
    using ActiveVisitor = typename super_type::ActiveVisitor;

    struct quarters
    {
      MatrixView north_west;
      MatrixView north_east;
      MatrixView south_west;
      MatrixView south_east;
    };

  private:

    /** @brief Convenience alias for `struct region_pass`. */
    using region_pass = typename super_type::region_pass;


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief Pointer to the viewed matrix (may be a `nullptr`). */
    qualified_super_type * thatPtr_ {nullptr};

    /** @brief Whether the viewed matrix is to be transposed. */
    bool transposeThat_ {false};

    /**
     * @brief
     *         Row index of the (0, 0) element in the (eventually transposed)
     *         matrix being viewed.
     *
     */
    index_type offRows_ {0};

    /**
     * @brief
     *         Column index of the (0, 0) element in the (eventually
     *         transposed) matrix being viewed.
     *
     */
    index_type offCols_ {0};

    /** @brief Number of rows in this view. */
    index_type rows_ {0};

    /** @brief Number of columns in this view. */
    index_type cols_ {0};


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    /**
     * Default constructor: Creates a new empty view.
     *
     * The created &ldquo;view&rdquo; does not view anything and is rarely
     * useful.
     *
     */
    MatrixView() noexcept
    {
      assert(this->classInvariantsHold_());
    }

    /**
     * @brief
     *         Constructor: Creates a new matrix view without taking ownership
     *         of the viewed matrix or its data.
     *
     * If transposing the viewed matrix is selected
     * (<code><var>transposed</var> == true</code>), the offsets are measured
     * in the already transposed matrix.
     *
     * In general, a `MatrixView` should not outlive its viewed matrix.  The
     * destructor of this class is guarenteed to succeed even if the viewed
     * matrix is already destroyed but invoking any other (non-static) member
     * function will result in undefined behaviour.
     *
     * @param matrix_ptr
     *         pointer to the matrix to be viewed
     *
     * @param off_rows
     *         row index (shift) of the (0, 0) element in the (transposed)
     *         <code><var>matrix_ptr</var></code>
     *
     * @param off_cols
     *         column index (shift) of the (0, 0) element in the (transposed)
     *         <code><var>matrix_ptr</var></code>
     *
     * @param rows
     *         number of rows to view
     *
     * @param cols
     *         number of columns to view
     *
     * @param transposed
     *         whether the viewed matrix is to be transposed (on the fly)
     *
     * @throws std::invalid_argument
     *         if <code><var>matrix_ptr</var></code> is a `nullptr` or the view
     *         would not fit into it
     *
     */
    MatrixView(qualified_super_type *const matrix_ptr,
               const index_type off_rows,
               const index_type off_cols,
               const index_type rows,
               const index_type cols,
               const bool transposed = false) :
      thatPtr_ {matrix_ptr},
      transposeThat_ {transposed},
      offRows_ {off_rows},
      offCols_ {off_cols},
      rows_ {rows},
      cols_ {cols}
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        {
          constexpr auto msg_nullptr = "matrix is a nullptr";
          constexpr auto msg_nofit_vert = "view does not fit vertically";
          constexpr auto msg_nofit_horz = "view does not fit horizontally";
          if (matrix_ptr == nullptr)
            throw std::invalid_argument {msg_nullptr};
          super_type::ctorCheckSize(rows, cols);
          if (!transposed)
            {
              if (off_rows + rows > matrix_ptr->rows())
                throw std::invalid_argument {msg_nofit_vert};
              if (off_cols + cols > matrix_ptr->cols())
                throw std::invalid_argument {msg_nofit_horz};
            }
          else
            {
              if (off_rows + rows > matrix_ptr->cols())
                throw std::invalid_argument {msg_nofit_vert};
              if (off_cols + cols > matrix_ptr->rows())
                throw std::invalid_argument {msg_nofit_horz};
            }
        }
      assert(this->classInvariantsHold_());
    }

    /**
     * @brief
     *         Copy constructor: Creates a new `MatrixView` that is a shallow
     *         copy of an existing one.
     *
     * The two `MatrixView`s view the same data that is owned by none of them.
     *
     * @param other
     *         `MatrixView` to be copied
     *
     */
    MatrixView(const MatrixView& other) noexcept :
      thatPtr_ {other.thatPtr_},
      transposeThat_ {other.transposeThat_},
      offRows_ {other.offRows_},
      offCols_ {other.offCols_},
      rows_ {other.rows_},
      cols_ {other.cols_}
    {
      assert(this->classInvariantsHold_());
    }

    /**
     * @brief
     *         Move constructor: Creates a new `MatrixView` by canibalizing an
     *         existing one.
     *
     * @param other
     *         `MatrixView` to be moved away from
     *
     */
    MatrixView(MatrixView&& other) noexcept :
      MatrixView {}
    {
      swap(*this, other);
    }

    // No explicit d'tor is needed.


    /* ------------- */
    /*   Operators   */
    /* ------------- */

    /**
     * @brief
     *         Copy-assignment operator: Makes this `MatrixView` become a
     *         shallow copy of an existing one.
     *
     * The two `MatrixView`s view the same data that is owned by none of them.
     * Since this `MatrixView` didn't own its previous data (if any) either, it
     * is not released.
     *
     * @param other
     *         `MatrixView` to be copied
     *
     * @returns
     *         reference to this `MatrixView`
     *
     */
    MatrixView&
    operator=(const MatrixView& other) noexcept
    {
      MatrixView temp {other};  // copy c'tor is noexcept
      swap(*this, temp);
      return *this;
    }

    /**
     * @brief
     *         Move-assignment operator: Makes this `MatrixView` become
     *         whatever another one (which is canibalized during the process)
     *         used to be.
     *
     * @param other
     *         `MatrixView` to be moved away from
     *
     * @returns
     *         reference to this `MatrixView`
     *
     */
    MatrixView&
    operator=(MatrixView&& other) noexcept
    {
      MatrixView temp {};  // default c'tor is noexcept
      swap(*this, temp);
      swap(*this, other);
      return *this;
    }


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  public:

    // {@InheritDoc}
    virtual index_type
    rows() const noexcept final override
    {
      return this->rows_;
    }

    // {@InheritDoc}
    virtual index_type
    cols() const noexcept final override
    {
      return this->cols_;
    }

    // {@InheritDoc}
    virtual value_type
    get(const index_type row,
        const index_type col) const final override
    {
      index_type i {static_cast<index_type>(-1)};
      index_type j {static_cast<index_type>(-1)};
      this->findIndex_(row, col, &i, &j);
      return this->thatPtr_->get(i, j);
    }

    // {@InheritDoc}
    virtual void
    set(const value_type val,
        const index_type row,
        const index_type col) final override
    {
      // Don't check preconditions or map indices if we are not mutable anyway.
      index_type i {static_cast<index_type>(-1)};
      index_type j {static_cast<index_type>(-1)};
      if (MatrixView::isMutable())
        this->findIndex_(row, col, &i, &j);
      MatrixView::applySet_(*this->thatPtr_, val, i, j);
    }

    // {@InheritDoc}
    virtual void
    passOverRegion(const index_type global_north,
                   const index_type global_west,
                   const bool global_trans,
                   PassiveVisitor& visitor,
                   const index_type north,
                   const index_type west,
                   const index_type south,
                   const index_type east) const final override
    {
      const region_pass params_in {
        global_north, global_west, global_trans, north, west, south, east
      };
      const region_pass params_out = this->mapRegion_(params_in);
      if (this->thatPtr_ != nullptr)
        {
          this->thatPtr_->passOverRegion(params_out.rp_global_north,
                                         params_out.rp_global_west,
                                         params_out.rp_global_trans,
                                         visitor,
                                         params_out.rp_north,
                                         params_out.rp_west,
                                         params_out.rp_south,
                                         params_out.rp_east);
        }
      else
        {
          assert(params_out.rp_south - params_out.rp_north == 0);
          assert(params_out.rp_east - params_out.rp_west == 0);
        }
    }

    // {@InheritDoc}
    virtual void
    passOverRegion(const index_type global_north,
                   const index_type global_west,
                   const bool global_trans,
                   ActiveVisitor& visitor,
                   const index_type north,
                   const index_type west,
                   const index_type south,
                   const index_type east) final override
    {
      const region_pass params_in {
        global_north, global_west, global_trans, north, west, south, east
      };
      const region_pass params_out = this->mapRegion_(params_in);
      if (this->thatPtr_ != nullptr)
        {
          MatrixView::applyActiveVisitor_(*this->thatPtr_, visitor, params_out);
        }
      else
        {
          assert(params_out.rp_south - params_out.rp_north == 0);
          assert(params_out.rp_east - params_out.rp_west == 0);
        }
    }

    index_type
    getRowOffset() const noexcept
    {
      return this->offRows_;
    }

    index_type
    getColOffset() const noexcept
    {
      return this->offCols_;
    }

    /**
     * @brief
     *         Returns a pointer to the viwed matrix.
     *
     * If the `MatrixView` is non-mutable, a `const` pointer is returned.
     *
     * @returns
     *         pointer to the viewed matrix
     *
     */
    qualified_super_type *
    getViewedMatrix() noexcept
    {
      return this->thatPtr_;
    }

    /**
     * @brief
     *         Returns a `const` pointer to the viwed matrix.
     *
     * @returns
     *         `const` pointer to the viewed matrix
     *
     */
    const super_type *
    getViewedMatrix() const noexcept
    {
      return this->thatPtr_;
    }

    /**
     * @brief
     *         Amputated function member -- never call!
     *
     * While `MatrixView`s are copyable, they cannot be cloned.  Allowing to
     * `clone` a `MatrixView` would be dangerous since it would have to violate
     * at least one contract.  If the clone where deep, the contract of a
     * &ldquo;view&rdquo; would be violated; if it were shallow, that of a
     * &ldquo;clone&rdquo;.
     *
     * @returns
     *         never
     *
     * @throws NotImplementedError
     *         always
     *
     */
    virtual std::unique_ptr<super_type>
    clone() const final override
    {
      throw NotImplementedError {
        __FILE__,
        __LINE__,
        "Sorry, you cannot 'clone()' a 'MatrixView'"
        " (amputated virtual function member)."
      };
    }

    bool
    isTransposed() const noexcept
    {
      return this->transposeThat_;
    }

    bool
    isTrivial() const noexcept
    {
      if (this->thatPtr_ == nullptr)
        return true;
      if (this->offRows_ != 0 || this->offCols_ != 0)
        return false;
      if (!this->transposeThat_)
        {
          if (this->rows_ != this->thatPtr_->rows())
            return false;
          if (this->cols_ != this->thatPtr_->cols())
            return false;
        }
      else
        {
          if (this->rows_ != this->thatPtr_->cols())
            return false;
          if (this->cols_ != this->thatPtr_->rows())
            return false;
        }
      return true;
    }

    /**
     * @brief
     *         Returns whether this class is mutable.
     *
     * @returns
     *         whether this class is mutable
     *
     */
    static constexpr bool
    isMutable() noexcept
    {
      return M;  // template parameter
    }

    static
    std::unique_ptr<super_type>
    makeView(qualified_super_type& matrix,
             const Rectangle<index_type> rect,
             const bool trans)
    {
      return std::unique_ptr<super_type> {new MatrixView {
          &matrix, rect.north(), rect.west(), rect.height(), rect.width(), trans
      }};
    }

    /**
     * @brief
     *         Creates four views of the quaters of a `HMatrix`.
     *
     * @param matrix
     *         matrix to split
     *
     * @param row
     *         row index of the first row in the two south-views
     *
     * @param col
     *         column index of the first column in the two east-views
     *
     */
    static quarters
    makeQuarters(qualified_super_type& matrix,
                 const index_type row,
                 const index_type col)
    {
      const index_type rows {matrix.rows()};
      const index_type cols {matrix.cols()};
      return quarters {
        MatrixView {&matrix, 0,   0,   row,        col},
        MatrixView {&matrix, 0,   col, row,        cols - col},
        MatrixView {&matrix, row, 0,   rows - row, col},
        MatrixView {&matrix, row, col, rows - row, cols - col},
      };
    }

    /**
     * @brief
     *         Creates four views of the quaters of a `HMatrix`.
     *
     * This is a convenience overload for
     * `MatrixView::makeQuarters(matrix, offset, offset)`.
     *
     * @param matrix
     *         matrix to split
     *
     * @param offset
     *         row index of the first row in the two south-views and column
     *         indext of the first column in the two east-views
     *
     */
    static quarters
    makeQuarters(qualified_super_type& matrix, const index_type offset)
    {
      return MatrixView::makeQuarters(matrix, offset, offset);
    }

    /**
     * @brief
     *         Factory function: Creates a transposed view of an entire matrix.
     *
     * This is a convenience function for invoking the constructor as
     * <code>MatrixView(&<var>matrix</var>, 0, 0, <var>matrix</var>.cols(),
     * <var>matrix</var>.rows(), true)</code>.
     *
     * @param matrix
     *         matrix to transpose
     *
     */
    static MatrixView
    makeTransposedView(qualified_super_type& matrix)
    {
      return MatrixView {
        &matrix,
        0,
        0,
        matrix.cols(),
        matrix.rows(),
        true
      };
    }

  private:

    /**
     * @brief
     *         Translates a pair of indices for this `MatrixView` to a pair of
     *         indices for the viewed `HMatrix`.
     *
     * After a successful return from this function, <code>(<var>row</var>,
     * <var>col</var>)</code> is mapped to <code>(*<var>i_ptr</var>,
     * *<var>j_ptr</var>)</code>.  On error, an exception is thrown and the
     * contents of the provided addresses are undefined.
     *
     * @param row
     *         row index for this matrix
     *
     * @param col
     *         column index for this matrix
     *
     * @param i_ptr
     *         address to store the row index for the viewed matrix at
     *
     * @param j_ptr
     *         address to store the column index for the viewed matrix at
     *
     * @throws std::invalid_argument
     *         if the index is not valid
     *
     */
    void
    findIndex_(const index_type row,
               const index_type col,
               index_type *const i_ptr,
               index_type *const j_ptr) const
    {
      if (BSC_KLAMMLER_CHECK_ARGS) { this->checkIndex(row, col); }
      if (!this->transposeThat_)
        {
          *i_ptr = this->offRows_ + row;
          *j_ptr = this->offCols_ + col;
        }
      else
        {
          *j_ptr = this->offRows_ + row;
          *i_ptr = this->offCols_ + col;
        }
    }

    /**
     * @brief
     *         Maps a region for this `MatrixView` to the equivalent region of
     *         the viewed matrix.
     *
     * @param params
     *         region for this `MatrixView`
     *
     * @returns
     *         region for the viewed `HMatrix`
     *
     * @throws std::invalid_argument
     *         if the region is invalid for this `MatrixView`
     *
     */
    region_pass
    mapRegion_(const region_pass params) const
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        this->checkRegion(params.rp_north,
                          params.rp_west,
                          params.rp_south,
                          params.rp_east,
                          params.rp_global_trans);
      const index_type ns_diff {
        params.rp_global_trans
          ? this->offCols_
          : this->offRows_
      };
      const index_type we_diff {
        params.rp_global_trans
          ? this->offRows_
          : this->offCols_
      };
      return region_pass {
        params.rp_global_north,
        params.rp_global_west,
        lxor(params.rp_global_trans, this->transposeThat_),
        params.rp_north + ns_diff,
        params.rp_west + we_diff,
        params.rp_south + ns_diff,
        params.rp_east + we_diff,
      };
    }

    /**
     * @brief
     *         Checks whether the class invariants for this object hold.
     *
     * The *only* intended use for this function is inside a run-time assertion
     * like
     *
     *     assert(this->classInvariantsHold_());
     *
     * This can be used at the end of constructors or assignment operators to
     * help avoid redundant logic.  Using its return value inside an assertion
     * rather than putting the assertion inside this function allows zero
     * run-time overhead if assertions are disabled.
     *
     * This function is not suitable for parameter validation; if it returns
     * `false`, the program should be aborted.
     *
     * @returns
     *         whether this class' invariants hold for this object
     *
     */
    bool
    classInvariantsHold_() const noexcept
    {
      if (this->thatPtr_ == nullptr)
        {
          if (this->size() != 0)
            return false;
        }
      else
        {
          if (!this->transposeThat_)
            {
              if (this->offRows_ + this->rows_ > this->thatPtr_->rows())
                return false;
              if (this->offCols_ + this->cols_ > this->thatPtr_->cols())
                return false;
            }
          else
            {
              if (this->offRows_ + this->rows_ > this->thatPtr_->cols())
                return false;
              if (this->offCols_ + this->cols_ > this->thatPtr_->rows())
                return false;
            }
        }
      return true;
    }

    /**
     * @brief
     *         Invokes <code>matrix.<var>set</var>(<var>v</var>, <var>i</var>,
     *         <var>j</var>)</code>.
     *
     * @param matrix
     *         matrix to set the element
     *
     * @param v
     *         value to set
     *
     * @param i
     *         row index (including any mappings and transformations)
     *
     * @param j
     *         row index (including any mappings and transformations)
     *
     * @throws std::invalid_argument
     *         if the `set` is rejected by the matrix
     *
     */
    inline static void
    applySet_(super_type& matrix,
              const value_type v,
              const index_type i,
              const index_type j)
    {
      matrix.set(v, i, j);
    }

    /**
     * @brief
     *         Unconditionally throws a `NotImplementedError`.  All arguments
     *         are ignored.
     *
     * @throws NotImplementedError
     *         always
     *
     */
    inline static void
    applySet_(const super_type&,
              const value_type,
              const index_type,
              const index_type)
    {
      throw NotImplementedError {
        __FILE__,
        __LINE__,
        "Sorry, you cannot 'set()' a value in a 'ConstMatrixView'"
        " (amputated virtual function member)."
      };
    }

    /**
     * @brief
     *         Applies -- all region mapping done beforehand -- an
     *         `ActiveVisitor` to the viewed matrix.
     *
     * @param matrix
     *         matrix to apply the visitor to
     *
     * @param visitor
     *         visitor to apply
     *
     * @param params
     *         parameters to call `matrix.passOverRegion` with
     *
     */
    inline static void
    applyActiveVisitor_(/* !const */ super_type& matrix,
                        /* !const */ ActiveVisitor& visitor,
                        const region_pass params)
    {
      matrix.passOverRegion(params.rp_global_north,
                            params.rp_global_west,
                            params.rp_global_trans,
                            visitor,
                            params.rp_north,
                            params.rp_west,
                            params.rp_south,
                            params.rp_east);
    }

    /**
     * @brief
     *         Unconditionally throws a `NotImplementedError`.  All arguments
     *         are ignored.
     *
     * @throws NotImplementedError
     *         always
     *
     */
    inline static void
    applyActiveVisitor_(const super_type&,
                        /* !const */ ActiveVisitor&,
                        const region_pass)
    {
      throw NotImplementedError {
        __FILE__,
        __LINE__,
        "Sorry, you cannot pass an 'ActiveVisitor' over a 'ConstMatrixView'"
        " (amputated virtual function member)."
      };
    }


    /* ----------- */
    /*   Friends   */
    /* ----------- */

  public:

    /**
     * @brief
     *         Exchanges all data between two `MatrixView`s.
     *
     * @param first
     *         first `MatrixView` to have its data exchanged
     *
     * @param second
     *         second `MatrixView` to have its data exchanged
     *
     */
    friend void
    swap(MatrixView& first, MatrixView& second) noexcept
    {
      using std::swap;
      swap(first.thatPtr_, second.thatPtr_);
      swap(first.transposeThat_, second.transposeThat_);
      swap(first.offRows_, second.offRows_);
      swap(first.offCols_, second.offCols_);
      swap(first.rows_, second.rows_);
      swap(first.cols_, second.cols_);
      assert(first.classInvariantsHold_());
      assert(second.classInvariantsHold_());
    }
  };  // class MatrixView

  /**
   * @brief
   *         An immutable view of a matrix.
   *
   */
  template<typename T>
  using ConstMatrixView = MatrixView<T, false>;

  /**
   * @brief
   *         A mutable view of a matrix.
   *
   */
  template<typename T>
  using MutableMatrixView = MatrixView<T, true>;

  template<typename T, bool M>
  ConstMatrixView<T>
  make_const(const MatrixView<T, M>& view)
  {
    if (view.rows() == 0 || view.cols() == 0)
      return ConstMatrixView<T> {};
    else
      return ConstMatrixView<T> {
        view.getViewedMatrix(),
        view.getRowOffset(),
        view.getColOffset(),
        view.rows(),
        view.cols(),
      };
  }
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_MATRIX_VIEW_HPP
