/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


#include "global.hpp"
#include "nest.hpp"

#include <cassert>
#include <string>

#define BLAS_LIBRARY 2
#include <nest/matrix.hpp>


namespace bsc_klammler
{
  namespace nest
  {
    namespace /* anonymous */
    {
      /**
       * @brief
       *         Returns the inversion strategy for a given matrix size.
       *
       * @param size
       *         size of the matrix to invert
       *
       * @returns
       *         inversion strategy in the format understood by NeSt
       *
       */
      inline std::string
      get_nest_strategy(const std::size_t size)
      {
        // TODO: Come up with a smarter way to determine the strategy.
        constexpr std::size_t threshold {1024UL};
        if (size < threshold)
          return "L";
        return "SSSSN";
      }
    }  // namespace /* anonymous */

    void
    invert(const double * data_in,
           double * data_out,
           const std::size_t size,
           const bool spd)
    {
      using nest_matrix_type = matrix<double>;
      using nest_index_type = matrix_base<double>::size_type;
      nest_matrix_type nest_matrix {static_cast<nest_index_type>(size)};
#if BSC_KLAMMLER_PARALLEL >= BSC_KLAMMLER_PARALLEL_DATA && defined(_OPENMP)
#pragma omp parallel for
#endif
      for (std::size_t k = 0; k < size * size; ++k)
        {
          const nest_index_type i {static_cast<nest_index_type>(k % size)};
          const nest_index_type j {static_cast<nest_index_type>(k / size)};
          nest_matrix.elem(i, j) = data_in[k];
        }
      const std::string strategy {get_nest_strategy(size)};
      nest_matrix_type nest_inverse {nest_matrix.invert(spd, strategy)};
#if BSC_KLAMMLER_PARALLEL >= BSC_KLAMMLER_PARALLEL_DATA && defined(_OPENMP)
#pragma omp parallel for
#endif
      for (std::size_t k = 0; k < size * size; ++k)
        {
          const nest_index_type i {static_cast<nest_index_type>(k % size)};
          const nest_index_type j {static_cast<nest_index_type>(k / size)};
          data_out[k] = nest_inverse.elem(i, j);
        }
    }
  }  // namespace nest
}  // namespace bsc_klammler
