/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file non_zero_counter.hpp
 *
 * @brief
 *         Header file providing the `NonZeroCounter` template class.
 *
 */

#ifndef BSC_KLAMMLER_NON_ZERO_COUNTER_HPP
#define BSC_KLAMMLER_NON_ZERO_COUNTER_HPP

#include "rectangle.hpp"  // bsc_klammler::Rectangle

namespace bsc_klammler
{

  /**
   * @brief
   *         A `PassiveVisitor` that extracts the number of non-zero entries
   *         from a `HMatrix`.
   *
   * By *NonZeroCounter entries* we mean elements for which we don't know *a
   * priori* they can only be zero.  This is a structural property of a matrix,
   * not an actual data lookup.
   *
   * To determine the number of non-zero entries in a (region of) a `HMatrix`,
   * create a *new* `NonZeroCounter` and pass it over the region, then query
   * its `count()` member function.  It is important to throw the
   * `NonZeroCounter` away and create a new one for counting the non.zero
   * elements of another `HMatrix` except if the sum of both is wanted.
   *
   * @param <T>
   *         type of the matrix entries
   *
   */
  template<typename T>
  class NonZeroCounter : public HMatrix<T>::PassiveVisitor
  {
  public:

    /** @brief Convenience type alias for `PassiveVisitor<T>`. */
    using super_type = typename HMatrix<T>::PassiveVisitor;

    /** @brief Type for matrix values. */
    using value_type = typename super_type::value_type;

    /** @brief Type for matrix indices. */
    using index_type = typename super_type::index_type;

    /** @brief Convenience type alias for `HMatrix<T>`. */
    using h_matrix = HMatrix<T>;

  private:

    /** @brief Non-zero element count (so far). */
    index_type count_ {0};

  public:

    /**
     * @brief
     *         Default constructor: Creates a new `NonZeroCounter` with an
     *         initial count of 0.
     *
     */
    NonZeroCounter() noexcept
    {
      // do nothing
    }

    /**
     * @brief
     *         Returns the non-zero element count of whatever regions this
     *         `NonZeroCounter` was passed over thus far.
     *
     * @returns
     *         number of a priori known non-zero entries
     *
     */
    index_type count() const noexcept
    {
      return this->count_;
    }

    // {@inheritDoc}
    virtual void
    visitRegion(const index_type global_north BSC_KLAMMLER_ATTRIBUTE_UNUSED,
                const index_type global_west BSC_KLAMMLER_ATTRIBUTE_UNUSED,
                const bool global_trans BSC_KLAMMLER_ATTRIBUTE_UNUSED,
                const h_matrix& matrix BSC_KLAMMLER_ATTRIBUTE_UNUSED,
                const index_type north,
                const index_type west,
                const index_type south,
                const index_type east) final override
    {
      const Rectangle<index_type> rect {north, west, south, east};
      this->count_ += rect.area();
    }
  };  // class NonZeroCounter

  /**
   * @brief
   *         Convenience function to count the non-zero entries of an entire
   *         `HMatrix`.
   *
   * @param matrix
   *         `HMatrix` to have its non-zero elements counted
   *
   * @returns
   *         number of a priori known non-zero entries
   *
   */
  template<typename T>
  typename HMatrix<T>::index_type
  non_zero_count(const HMatrix<T>& matrix)
  {
    NonZeroCounter<T> nzc {};
    pass_over_region(matrix, 0, 0, false, nzc);
    return nzc.count();
  }
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_NON_ZERO_COUNTER_HPP
