/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file not_reached.hpp
 *
 * @brief
 *         Header file providing the `NotReachedError` exception class and the
 *         `BSC_KLAMMLER_NOT_REACHED` convenience macro.
 *
 */

#ifndef BSC_KLAMMLER_NOT_REACHED_HPP
#define BSC_KLAMMLER_NOT_REACHED_HPP

#include "control_flow_error.hpp"  // bsc_klammler::ControlFlowError



/**
 * @def BSC_KLAMMLER_NOT_REACHED
 *
 * @brief
 *         Macro statement to place at code points the control flow should
 *         never reach.
 *
 * This macro will unconditionally throw a `NotReachedError` with the current
 * soucre file name and line number.  Use it as a more expressive and safer
 * alternative for &ldquo;not reached&rdquo;-style comments or ad-hoc
 * `assert(false)` solutions.
 *
 * Example:
 *
 *     enum class Signums {NEGATIVE, ZERO, POSITIVE};
 *
 *     Signums
 *     iclassify(const int n)
 *     {
 *       if (n == 0)
 *         return Signums::ZERO;
 *       else if (n < 0)
 *         return Signums::NEGATIVE:
 *       else if (n > 0)
 *         return Signums::POSITIVE;
 *       else
 *         BSC_KLAMMLER_NOT_REACHED;
 *     }
 *
 */
#define BSC_KLAMMLER_NOT_REACHED                                              \
  throw ::bsc_klammler::NotReachedError {__FILE__, __LINE__}


namespace bsc_klammler
{


  /**
   * @brief
   *         Exception to indicate that some functionality is not (yet)
   *         reached.
   *
   */
  class NotReachedError : public ControlFlowError
  {

  public:

    /**
     * @brief
     *         Constructor: Creates a new exception.
     *
     * @param file
     *         source file name the exception is thrown from
     *
     * @param line
     *         line number the exception is thrown from
     *
     * @param message
     *         additional error message
     *
     */
    NotReachedError(const std::string& file,
                    const int line,
                    const std::string& message = "the impossible has happened") :
      ControlFlowError {file, line, message}
    {
      // do nothing
    }
  };  // NotReachedError
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_NOT_REACHED_HPP
