/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file numeric_costs.hpp
 *
 * @brief
 *         Cost of various numeric computations.
 *
 */

#ifndef BSC_KLAMMLER_NUMERIC_COSTS_HPP
#define BSC_KLAMMLER_NUMERIC_COSTS_HPP

namespace bsc_klammler
{

  /**
   * @brief
   *         Cost of various numeric computations.
   *
   */
  namespace numeric_cost
  {
    /**
     * @brief
     *         Cost of general matrix-matrix product.
     *
     * This function returns the numeric cost (in FLOPs) of &ndash; given
     * matrices <var>A</var> and <var>B</var> &ndash; computing the product
     * <var>A</var> &sdot; <var>B</var> = <var>C</var> with <var>A</var> &isin;
     * <var>K</var><sup><var>n</var> &times; <var>k</var></sup>, <var>B</var>
     * &isin; <var>K</var><sup><var>k</var> &times; <var>m</var></sup> and
     * <var>C</var> &isin; <var>K</var><sup><var>n</var> &times;
     * <var>m</var></sup>.
     *
     * @param <T>
     *         type for measuring dimensions and counting FLOPs
     *
     * @param n
     *         rows in <var>A</var>
     *
     * @param k
     *         columns in <var>A</var> / rows in <var>B</var>
     *
     * @param m
     *         columns in <var>B</var>
     *
     * @returns
     *         number of FLOPs
     *
     */
    template<typename T>
    constexpr T
    matrix_matrix_product(const T n, const T k, const T m) noexcept
    {
      return n * k * m;
    }

    /**
     * @brief
     *         Cost of symmetric matrix-matrix product.
     *
     * This function returns the numeric cost (in FLOPs) of &ndash; given a
     * symmetric matrix <var>A</var> and a matrix <var>B</var> &ndash;
     * computing the product <var>A</var> &sdot; <var>B</var> = <var>C</var>
     * with <var>A</var> &isin; <var>K</var><sup><var>n</var> &times;
     * <var>n</var></sup>, <var>B</var> &isin; <var>K</var><sup><var>n</var>
     * &times; <var>m</var></sup> and <var>C</var> &isin;
     * <var>K</var><sup><var>n</var> &times; <var>m</var></sup>.
     *
     * The cost will be the same if the operands are exchanged (ie <var>B</var>
     * &sdot; <var>A</var>).
     *
     * @param <T>
     *         type for measuring dimensions and counting FLOPs
     *
     * @param n
     *         rows / columns in <var>A</var>, rows in <var>B</var>
     *
     * @param m
     *         columns in <var>B</var>
     *
     * @returns
     *         number of FLOPs
     *
     */
    template<typename T>
    constexpr T
    symmetric_matrix_matrix_product(const T n, const T m) noexcept
    {
      // TODO: Is this correct?
      return n * n * m;
    }

    /**
     * @brief
     *         Cost of general matrix-vector product.
     *
     * This function returns the numeric cost (in FLOPs) of &ndash; given a
     * matrix <var>A</var> and a vector <var>u</var> &ndash; computing the
     * product <var>A</var> &sdot; <var>u</var> = <var>w</var> with
     * <var>A</var> &isin; <var>K</var><sup><var>n</var> &times;
     * <var>m</var></sup> and <var>u</var>, <var>w</var> &isin;
     * <var>K</var><sup><var>m</var></sup>.
     *
     * @param <T>
     *         type for measuring dimensions and counting FLOPs
     *
     * @param n
     *         rows in <var>A</var>
     *
     * @param m
     *         columns in <var>A</var> / rows in <var>u</var>
     *
     * @returns
     *         number of FLOPs
     *
     */
    template<typename T>
    constexpr T
    matrix_vector_product(const T n, const T m) noexcept
    {
      return n * m;
    }

    /**
     * @brief
     *         Cost of symmetric matrix-vector product.
     *
     * This function returns the numeric cost (in FLOPs) of &ndash; given a
     * symmetric matrix <var>A</var> and a vector <var>u</var> &ndash;
     * computing the product <var>A</var> &sdot; <var>u</var> = <var>w</var>
     * with <var>A</var> &isin; <var>K</var><sup><var>n</var> &times;
     * <var>n</var></sup> and <var>u</var>, <var>w</var> &isin;
     * <var>K</var><sup><var>n</var></sup>.
     *
     * @param <T>
     *         type for measuring dimensions and counting FLOPs
     *
     * @param n
     *         rows in <var>A</var>
     *
     * @param m
     *         columns in <var>A</var> / rows in <var>u</var>
     *
     * @returns
     *         number of FLOPs
     *
     */
    template<typename T>
    constexpr T
    symmetric_matrix_vector_product(const T n) noexcept
    {
      // TODO: Is this correct?
      return n * n;
    }

    /**
     * @brief
     *         Cost of Cholesky factorization.
     *
     * This function returns the numeric cost (in FLOPs) of &ndash; given a
     * symmetric and positive definite matrix <var>A</var> &ndash; computing
     * the Cholesky factorization <var>A</var> = <var>L</var> &sdot;
     * <var>L</var><sup>T</sup> with <var>A</var>, <var>L</var> &isin;
     * <var>K</var><sup><var>n</var> &times; <var>n</var></sup> where
     * <var>L</var> is lower triangular.
     *
     * @param <T>
     *         type for measuring dimensions and counting FLOPs
     *
     * @param n
     *         rows / columns in <var>A</var>
     *
     * @returns
     *         number of FLOPs
     *
     */
    template<typename T>
    constexpr T
    llt_factorization(const T n) noexcept
    {
      return n * n * n / 6;
    }

    /**
     * @brief
     *         Cost of forward substitution.
     *
     * This function returns the numeric cost (in FLOPs) of &ndash; given a
     * lower triangular matrix <var>L</var> and a vector <var>b</var> &ndash;
     * solving the linear system <var>L</var> &sdot; <var>y</var> =
     * <var>b</var> with <var>L</var> &isin; <var>K</var><sup><var>n</var>
     * &times; <var>n</var></sup> and <var>y</var>, <var>b</var> &isin;
     * <var>K</var><sup><var>n</var></sup> via forward substitution.
     *
     * @param <T>
     *         type for measuring dimensions and counting FLOPs
     *
     * @param n
     *         rows / columns in <var>L</var>, elements in <var>y</var> and
     *         <var>b</var>
     *
     * @returns
     *         number of FLOPs
     *
     */
    template<typename T>
    constexpr T
    forward_substitution(const T n) noexcept
    {
      return n * n / 2;
    }

    /**
     * @brief
     *         Cost of backward substitution.
     *
     * This function returns the numeric cost (in FLOPs) of &ndash; given an
     * upper triangular matrix <var>U</var> and a vector <var>c</var> &ndash;
     * solving the linear system <var>U</var> &sdot; <var>x</var> =
     * <var>c</var> with matrix <var>U</var> &isin;
     * <var>K</var><sup><var>n</var> &times; <var>n</var></sup> and
     * <var>x</var>, <var>c</var> &isin; <var>K</var><sup><var>n</var></sup>
     * via backward substitution.
     *
     * @param <T>
     *         type for measuring dimensions and counting FLOPs
     *
     * @param n
     *         rows / columns in <var>U</var>, elements in <var>x</var> and
     *         <var>c</var>
     *
     * @returns
     *         number of FLOPs
     *
     */
    template<typename T>
    constexpr T
    backward_substitution(const T n) noexcept
    {
      return n * n / 2;
    }

    /**
     * @brief
     *         Cost of general matrix inversion.
     *
     * This function returns the numeric cost (in FLOPs) of &ndash; given a
     * regular matrix <var>A</var> &ndash; computing the inverse
     * <var>A</var><sup>&minus;1</var>.
     *
     * @param <T>
     *         type for measuring dimensions and counting FLOPs
     *
     * @param n
     *         rows / columns in <var>A</var>
     *
     * @returns
     *         number of FLOPs
     *
     */
    template<typename T>
    constexpr T
    matrix_inversion(const T n) noexcept
    {
      // TODO: Is this correct?
      return n * n * n;
    }

    /**
     * @brief
     *         Cost of symmetric matrix inversion.
     *
     * This function returns the numeric cost (in FLOPs) of &ndash; given a
     * symmetric and positive definite matrix <var>A</var> &ndash; computing
     * the inverse <var>A</var><sup>&minus;1</var>.
     *
     * @param <T>
     *         type for measuring dimensions and counting FLOPs
     *
     * @param n
     *         rows / columns in <var>A</var>
     *
     * @returns
     *         number of FLOPs
     *
     */
    template<typename T>
    constexpr T
    matrix_inversion_spd(const T n) noexcept
    {
      // TODO: Is this correct?
      return n * n * n / 2;
    }
  }  // namespace numeric_cost
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_NUMERIC_COSTS_HPP
