/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file preconditioner.hpp
 *
 * @brief
 *         Header file providing the abstract `Preconditioner` template class.
 *
 */

#ifndef BSC_KLAMMLER_PRECONDITIONER_HPP
#define BSC_KLAMMLER_PRECONDITIONER_HPP

#include <atomic>           // std::atomic
#include <stdexcept>        // std::invalid_argument, std::logic_error
#include <string>           // std::string
#include <type_traits>      // std::is_floating_point
#include <utility>          // std::swap
#include <vector>           // std::vector

#include "full_vector.hpp"  // bsc_klammler::FullVector
#include "h_matrix.hpp"     // bsc_klammler::HMatrix
#include "macros.hpp"       // (package-specific macros)
#include "timer.hpp"        // bsc_klammler::Timer


namespace bsc_klammler
{

  /**
   * @brief
   *         Abstract base class for preconditioners.
   *
   * A *preconditioner* for a linear system <var>M</var>&sdot;<var>x</var> =
   * <var>b</var> is a matrix <var>P</var> such that
   * <var>P</var>&sdot;<var>M</var>&sdot;<var>x</var> =
   * <var>P</var>&sdot;<var>b</var> is well conditioned.  In addition, the
   * system <var>P</var>&sdot;<var>y</var> = <var>c</var> should be easy to
   * solve.  (That is, *much* easier than the original system.)
   *
   * Implementations of this class should provide the following functionality:
   *
   * - A default constructor that does no computing.
   * - An implementation for the `initImpl()` function member to computing the
   *   preconditioner <var>P</var>.  (Usually, there will not be an explicit
   *   representation of <var>P</var> but rather an implicit one (eg given by a
   *   <var>LU</var> factorization) that allows for efficient solving.)
   * - An implementation for the `solveImpl()` member function to solve linear
   *   systems <var>P</var>&sdot;<var>y</var> = <var>c</var>.
   *
   * @param <T>
   *         type of the variables in the linear system (must be a floating
   *         point type)
   *
   */
  template<typename T>
  class Preconditioner
  {

    static_assert(std::is_floating_point<T>::value,
                  "This 'Preconditioner' is only meaningful for real systems.");

    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Type for the dimensionality of the linear system. */
    typedef BSC_KLAMMLER_INDEX_TYPE index_type;

    /** @brief Type of the variables in the linear system. */
    typedef T value_type;

    /** @brief Type for counting FLOPs. */
    typedef std::size_t flops_type;

    /**
     * @brief
     *         Exception that is thrown if the `init()` and `solve()` functions
     *         are called in an invalid sequence.
     *
     */
    class InvalidStateError : public std::logic_error
    {
    public:

      /**
       * @brief
       *         Constructor: Creates a new exception.
       *
       * @param msg
       *         informal error message explaining why the operation was
       *         disallowed in the current state
       *
       */
      InvalidStateError(const std::string& msg) : std::logic_error {msg}
      {
        // do nothing
      }
    };  // class InvalidStateError


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief Dimensionality of the linear system. */
    const index_type size_ {0};

    /** @brief Number of CPUs to use. */
    unsigned numCPUs_ {1};

    /** @brief Whether this `Preconditioner` has already been `init()`ialized. */
    mutable std::atomic<bool> isInitialized_ {false};

    /** @brief Whether this `Preconditioner` has already `solve()`d something. */
    mutable std::atomic<bool> hasSolved_ {false};

    /** @brief Theoretical number of FLOPs for decomposing. */
    mutable std::atomic<flops_type> flopsInit_ {0};

    /** @brief Theoretical number of FLOPs for solving. */
    mutable std::atomic<flops_type> flopsSolve_ {0};


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  protected:

    Preconditioner(const index_type size) : size_ {size}
    {
      if (size < 0)
        throw std::invalid_argument {"size < 0"};
    }

  public:

    /**
     * @brief
     *         Virtual destructor.
     *
     */
    virtual ~Preconditioner() noexcept = default;

    /** @brief Deleted copy constructor. */
    Preconditioner(const Preconditioner&) = delete;

    /** @brief Deleted move constructor. */
    Preconditioner(Preconditioner&&) = delete;


    /* ------------- */
    /*   Operators   */
    /* ------------- */

  private:

    /** @brief Deleted copy-asignment operator. */
    Preconditioner&
    operator=(const Preconditioner&) = delete;

    /** @brief Deleted move-asignment operator. */
    Preconditioner&
    operator=(Preconditioner&&) = delete;


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  public:

    /**
     * @brief
     *         Returns an informal name of the `Preconditioner` with some
     *         atributes.
     *
     * @returns
     *         informal name with attributes
     *
     */
    std::string
    getLongName() const
    {
      std::vector<std::string> attributes {};
      const std::string name {this->getName(attributes)};
      std::ostringstream oss {};
      if (this->isRecursive())
        attributes.push_back("recursive");
      else
        attributes.push_back("dense");
      if (this->onlySPD())
        attributes.push_back("spd-only");
      oss << name;
      if (!attributes.empty())
        {
          oss << " (";
          for (std::size_t i = 0; i < attributes.size(); ++i)
            oss << (i ? ", " : "") << attributes.at(i);
          oss << ")";
        }
      return oss.str();
    }

    /**
     * @brief
     *         Returns a short name of the `Preconditioner` that is a single
     *         word.
     *
     * @returns
     *         short name
     *
     */
    std::string
    getShortName() const
    {
      std::vector<std::string> attributes {};
      const std::string name {this->getName(attributes)};
      assert(name.find(' ') == std::string::npos);
      // TODO: Better assertions?
      return name;
    }

    /**
     * @brief
     *         Tells whether this `Preconditioner` supports only symetric and
     *         positive definite (spd) coefficient matrices.
     *
     * @returns
     *         whether only spd matrices are supported
     *
     */
    virtual bool
    onlySPD() const noexcept = 0;

    /**
     * @brief
     *         Returns whether this `Preconditioner` can take advantage of the
     *         special structure of hirachical matrices.
     *
     * If this function returns `false`, it will be more efficient to pass
     * dense matrices and vectors to the `Preconditioner`.
     *
     * @returns
     *         whether the structure of hirachical matrices is exploited
     *
     */
    virtual bool
    isRecursive() const noexcept = 0;

    /**
     * @brief
     *         Initializes the preconditioner.
     *
     * This function must be called at most once on a `Preconditioner` object.
     *
     * @param cofficients
     *         coefficient matrix for the linear system
     *
     * @param timer_ptr
     *         optional `Timer` to `start()` before and `stop()` after the
     *         actual computational work
     *
     * @throws std::invalid_argument
     *         if the coefficient matrix is not square, empty or otherwise
     *         misbehaved
     *
     * @throws Preconditioner::InvalidStateError
     *         if `init` has been called before
     *
     */
    void
    init(const HMatrix<value_type>& coefficients,
         Timer *const timer_ptr = nullptr)
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        {
          if (this->isInitialized_.load())
            throw Preconditioner::InvalidStateError {
              "Preconditioner::init called more than once"
            };
          if (coefficients.rows() != coefficients.cols())
            throw std::invalid_argument {"coefficient matrix is non-square"};
          if (coefficients.size() != this->size_)
            throw std::invalid_argument {"coefficient matrix has wrong size"};
        }
      {
        const Timer::Handle th {timer_ptr};
        this->initImpl(coefficients);
      }
      this->isInitialized_.store(true);
    }

    /**
     * @brief
     *         Solves the linear system
     *         <var>P</var>&sdot;<var>y</var>=<var>c</var> where <var>P</var>
     *         is the preconditioner.
     *
     * This function is only valid to call on `init()`ialized
     * `Preconditioner`s.
     *
     * @param right
     *         right-side to solve the system for
     *
     * @param timer_ptr
     *         optional `Timer` to `start()` before and `stop()` after the
     *         actual computational work
     *
     * @returns
     *         an (approximate) solution to the linear system (fast, if
     *         possible)
     *
     * @throws std::invalid_argument
     *         if the vector has the wrong dimension
     *
     * @throws Preconditioner::InvalidStateError
     *         if `init` has not been called yet
     *
     */
    FullVector<value_type>
    solve(const HVector<value_type>& right,
          Timer *const timer_ptr = nullptr) const
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        {
          if (!this->isInitialized_.load())
            throw Preconditioner::InvalidStateError {
              "Preconditioner::solve called before Preconditioner::init"
            };
          if (right.size() != this->size())
            throw std::invalid_argument {"right.size() is not the right size"};
        }
      FullVector<value_type> solution {};
      {
        const Timer::Handle th {timer_ptr};
        solution = this->solveImpl(right);
      }
      assert(solution.size() == this->size());
      this->hasSolved_.store(true);
      return solution;
    }

    /**
     * @brief
     *         Returns the dimensionality of the linear system.
     *
     * @returns
     *         dimensionality of the linear system
     *
     */
    index_type
    size() const noexcept
    {
      return this->size_;
    }

    /**
     * @brief
     *         Sets the number of CPUs to use.
     *
     * The `Preconditioner` implementation is advised to use *up to* that
     * number of CPUs concurrently.  This is not a hard limit and the
     * implementation is free to ignore it.  <code><var>n</var></code> need not
     * be a power of two, though it will be generally useful.
     *
     * It goes without saying that this value should be set *before* any
     * computation is requested from the `Preconditioner`.
     *
     * @param n
     *         number of CPUs
     *
     * @throws std::invalid_argument
     *         if <code><var>n</var></code> is lass than 1
     *
     */
    void
    setNumCPUs(const unsigned n) noexcept
    {
      if (BSC_KLAMMLER_CHECK_ARGS && n < 1)
        throw std::invalid_argument {"need at least 1 CPU"};
      this->numCPUs_ = n;
    }

    /**
     * @brief
     *         Returns the number of available CPUs.
     *
     * This is the value set via `setNumCPUs()`, not an original guess.
     *
     * @returns
     *         number of CPUs
     *
     */
    unsigned
    getNumCPUs() const noexcept
    {
      return this->numCPUs_;
    }

    /**
     * @brief
     *         Returns the theoretical number of floating point operations
     *         (FLOPs) needed to compute the decomposition for this
     *         `Preconditioner`.
     *
     * This number depends on the coefficient matrix and is computed on-the-fly
     * alongside with the decomposition.  Therefore, the function must only be
     * called after `init()` completed successfully.
     *
     * If counting FLOPs is disabled, this function always returns 0.
     *
     * @returns
     *         theoretical number of FLOPs for computing the decomposition
     *
     */
    flops_type
    getInitFLOPs() const noexcept
    {
      const flops_type flops {this->flopsInit_.load()};
      if (BSC_KLAMMLER_CHECK_ARGS && !this->isInitialized_.load())
        throw Preconditioner::InvalidStateError {
          "Preconditioner::getInitFLOPs called before Preconditioner::init"
        };
      assert(!!BSC_KLAMMLER_COUNT_FLOPS == !!flops);
      return flops;
    }

    /**
     * @brief
     *         Returns the theoretical number of floating point operations
     *         (FLOPs) needed to solve a single linera system using this
     *         `Preconditioner`.
     *
     * This number depends on the coefficient matrix and is computed on-the-fly
     * alongside with solving the first linear system.  Therefore, the function
     * must only be called after `solve()` has once completed successfully.
     *
     * If counting FLOPs is disabled, this function always returns 0.
     *
     * @returns
     *         theoretical number of FLOPs for solving a linear system
     *
     */
    flops_type
    getSolveFLOPs() const noexcept
    {
      const flops_type flops {this->flopsSolve_.load()};
      if (BSC_KLAMMLER_CHECK_ARGS && !this->hasSolved_.load())
        throw Preconditioner::InvalidStateError {
          "Preconditioner::getSolveFLOPs called before Preconditioner::solve"
        };
      assert(!!BSC_KLAMMLER_COUNT_FLOPS == !!flops);
      return flops;
    }

  protected:

    /**
     * @brief
     *         Initializes the preconditioner.
     *
     * Derived classes provide the functionality of the `init()` function by
     * overriding this function member.  They may safely assume that the
     * `Preconditioner` object is in a valid state to call this function and
     * that the parameters are already validated.
     *
     * @param cofficients
     *         coefficient matrix for the linear system
     *
     */
    virtual void
    initImpl(const HMatrix<value_type>& coefficients) = 0;

    /**
     * @brief
     *         Solves the linear system
     *         <var>P</var>&sdot;<var>y</var>=<var>c</var> where <var>P</var>
     *         is the preconditioner.
     *
     * Derived classes provide the functionality of the `solve()` function by
     * overriding this function member.  They may safely assume that the
     * `Preconditioner` object is in a valid state to call this function and
     * that the parameters are already validated.
     *
     * @param right
     *         right-side to solve the system for
     *
     * @returns
     *         an (approximate) solution to the linear system (fast, if
     *         possible)
     *
     */
    virtual FullVector<value_type>
    solveImpl(const HVector<value_type>& right) const = 0;

    /**
     * @brief
     *         Returns the name and eventually other attributes of this
     *         `Preconditioner`.
     *
     * Implementing classes shall override this function to return their
     * (short) name.  At the discretion of the implementing class, it can add
     * zero or more attributes to the provided vector.  An atribute can b any
     * short tag that further describes the `Preconditioner`.  The base class
     * will always add the attributes `"dense"` or `"recursive"` and
     * `"spd-only"` depending on what the `isRecursive()` and `onlySPD()`
     * function members return.
     *
     * @param attributes
     *         destination for further optional tags
     *
     * @returns
     *         name of the `Preconditioner`
     *
     */
    virtual std::string
    getName(std::vector<std::string>& attributes) const = 0;

    /**
     * @brief
     *         Returns whether `init()` was already called and completed
     *         successfully.
     *
     * If this function is enterd concurrently with `init()` it might return
     * `true` before `init()` actually returns.  However, it will not return
     * `true` unless `init()` has completed all of its work successfully.
     *
     * @returns
     *         whether this `Preconditioner` was `init()`ialized successfully
     *
     */
    bool
    isInitialized() const noexcept
    {
      return this->isInitialized_.load();
    }

    /**
     * @brief
     *         Returns whether `solve()` was already called and completed
     *         successfully.
     *
     * If this function is enterd concurrently with `solve()` it might return
     * `true` before `solve()` actually returns.  However, it will not return
     * `true` unless `solve()` has completed all of its work successfully.
     *
     * @returns
     *         whether this `Preconditioner` has already `solve()`d somthing
     *         successfully
     *
     */
    bool
    hasSolved() const noexcept
    {
      return this->hasSolved_.load();
    }

    /**
     * @brief
     *         Atomically increments the FLOP count for the `init` function by
     *         the given value.
     *
     * This function is declared `const` so it can be called from functions
     * that do cause FLOPs but don't mutate other member data.  This is done by
     * declaring the internal FLOP counter `mutable`.
     *
     * @param flops
     *         number of additional FLOPs
     *
     * @throws std::invalid_argument
     *         if compiled without `BSC_KLAMMLER_COUNT_FLOPS` or if passed a
     *         negative argument
     *
     */
    void
    addInitFLOPs(const flops_type flops) const
    {
      if (!BSC_KLAMMLER_COUNT_FLOPS)
        throw std::invalid_argument {"we are not counting FLOPs"};
      if (BSC_KLAMMLER_CHECK_ARGS && flops < 0)
        throw std::invalid_argument {"flops < 0"};
      this->flopsInit_.fetch_add(flops, std::memory_order_relaxed);
    }

    /**
     * @brief
     *         Atomically increments the FLOP count for the `solve` function by
     *         the given value.
     *
     * This function is declared `const` so it can be called from functions
     * that do cause FLOPs but don't mutate other member data.  This is done by
     * declaring the internal FLOP counter `mutable`.
     *
     * @param flops
     *         number of additional FLOPs
     *
     * @throws std::invalid_argument
     *         if compiled without `BSC_KLAMMLER_COUNT_FLOPS` or if passed a
     *         negative argument
     *
     */
    void
    addSolveFLOPs(const flops_type flops) const
    {
      if (!BSC_KLAMMLER_COUNT_FLOPS)
        throw std::invalid_argument {"we are not counting FLOPs"};
      if (BSC_KLAMMLER_CHECK_ARGS && flops < 0)
        throw std::invalid_argument {"flops < 0"};
      this->flopsSolve_.fetch_add(flops, std::memory_order_relaxed);
    }
  };  // class Preconditioner
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_PRECONDITIONER_HPP
