/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file random.hpp
 *
 * @brief
 *         Low-level pseudo random operations.
 *
 * None of the functions provided by this header file are thread-safe.
 *
 */

#ifndef BSC_KLAMMLER_RANDOM_HPP
#define BSC_KLAMMLER_RANDOM_HPP

#include <random>  // (pseudo random numbers)


namespace bsc_klammler
{
  namespace random_hpp
  {

    std::default_random_engine&
    get_random_engine()
    {
      static std::default_random_engine rndengine {0};
      static bool initialized {false};
      if (!initialized)
        {
          std::random_device rnddev {};
          rndengine.seed(rnddev());
        }
      return rndengine;
    }
  }  // namespace random_hpp

  template<typename T>
  T
  random_real(const T min, const T max)
  {
    std::uniform_real_distribution<T> rnddist {min, max};
    return rnddist(random_hpp::get_random_engine());
  }

  template<typename T>
  T
  random_real()
  {
    std::exponential_distribution<T> rnd_mag_dist {static_cast<T>(1.0)};
    std::bernoulli_distribution rnd_sgn_dist {};
    const T magnitute {rnd_mag_dist(random_hpp::get_random_engine())};
    const T signum {
      rnd_sgn_dist(random_hpp::get_random_engine())
        ? static_cast<T>(-1.0)
        : static_cast<T>(+1.0)
    };
    return signum * magnitute;
  }

  template<typename T>
  T
  random_int(const T min, const T max)
  {
    std::uniform_int_distribution<T> rnddist {min, max};
    return rnddist(random_hpp::get_random_engine());
  }

  template<typename T>
  T
  random_int()
  {
    const T min {std::numeric_limits<T>::min()};
    const T max {std::numeric_limits<T>::max()};
    return random_int<T>(min, max);
  }

  bool
  random_bool(const double p = 0.5)
  {
    std::bernoulli_distribution rnddist {p};
    return rnddist(random_hpp::get_random_engine());
  }
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_RANDOM_HPP
