/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file random_problems.hpp
 *
 * @brief
 *         Generate random linear systems.
 *
 */

#ifndef BSC_KLAMMLER_RANDOM_PROBLEMS_HPP
#define BSC_KLAMMLER_RANDOM_PROBLEMS_HPP

#include <utility>            // std::move

#include "block_matrix.hpp"   // bsc_klammler::BlockMatrix
#include "full_matrix.hpp"    // bsc_klammler::FullMatrix
#include "full_vector.hpp"    // bsc_klammler::FullVector
#include "matrix_view.hpp"    // bsc_klammler::ConstMatrixView
#include "not_reached.hpp"    // BSC_KLAMMLER_NOT_REACHED
#include "random.hpp"         // bsc_klammler::random_int, ...


namespace bsc_klammler
{
  namespace random_problems_hpp
  {
    /**
     * @brief
     *         Generates a random `HMatrix` allocated on the free store.
     *
     * If the type `T` has support for special values (NaNs, infinities, etc),
     * is made sure that the generated vector contains none of those.
     *
     * @param recursions
     *         depth of the `BlockMatrix`
     *
     * @param min_bs
     *         minimum size of a block (must be at least 2)
     *
     * @param min_bs
     *         maximum size of a block
     *         (must be at least <code><var>min_bs</var></code>)
     *
     * @param spd
     *         whether to fill the matrix such that it is symmetric and
     *         positive definite
     *
     * @returns
     *         random matrix
     *
     * @throws std::invalid_argument
     *         if <code><var>min_bs</var> &lt; 2 || <var>max_bs</var> &lt;
     *         <var>min_bs</var></code>
     *
     */
    template<typename T>
    std::unique_ptr<HMatrix<T>>
    get_random_h_matrix(const unsigned recursions,
                        const typename HMatrix<T>::index_type min_bs,
                        const typename HMatrix<T>::index_type max_bs,
                        const bool symmetric)
    {
      typedef T value_type;
      using h_matrix = HMatrix<value_type>;
      using full_matrix = FullMatrix<value_type>;
      using block_matrix = BlockMatrix<value_type>;
      using index_type = typename h_matrix::index_type;
      if (BSC_KLAMMLER_CHECK_ARGS)
        {
          if (min_bs < 2)
            throw std::invalid_argument {"min_bs < 2"};
          if (max_bs < min_bs)
            throw std::invalid_argument {"max_bs < min_bs"};
        }
      if (!recursions)
        {
          const index_type n {random_int<index_type>(min_bs, max_bs)};
          return std::unique_ptr<h_matrix> {new full_matrix {n, n}};
        }
      else
        {
          const auto scale4c = [min_bs](index_type bs)->index_type{
            return std::max(std::min(min_bs, static_cast<index_type>(8)),
                            static_cast<index_type>(2.0 * std::sqrt(bs)));
          };
          std::unique_ptr<h_matrix> a_uptr {get_random_h_matrix<value_type>(
              recursions - 1, min_bs, max_bs, symmetric)
          };
          std::unique_ptr<h_matrix> b_uptr {get_random_h_matrix<value_type>(
              recursions - 1, min_bs, max_bs, symmetric)
          };
          std::unique_ptr<h_matrix> c_uptr {get_random_h_matrix<value_type>(
              0, scale4c(min_bs), scale4c(max_bs), symmetric)
          };
          std::unique_ptr<h_matrix> u_uptr {
            new full_matrix {c_uptr->rows(), a_uptr->cols()}
          };
          std::unique_ptr<h_matrix> v_uptr {
            new full_matrix {c_uptr->rows(), b_uptr->cols()}
          };
          std::unique_ptr<h_matrix> x_uptr {
            new full_matrix {a_uptr->rows(), c_uptr->cols()}
          };
          std::unique_ptr<h_matrix> y_uptr {
            new full_matrix {b_uptr->rows(), c_uptr->cols()}
          };
          return std::unique_ptr<h_matrix> {new block_matrix {
              std::move(a_uptr),
              std::move(b_uptr),
              std::move(c_uptr),
              std::move(u_uptr),
              std::move(v_uptr),
              std::move(x_uptr),
              std::move(y_uptr)
          }};
        }
    }

    template<typename T>
    void
    make_xy_uv_trans(HMatrix<T>& matrix)
    {
      typedef T value_type;
      using block_matrix = BlockMatrix<value_type>;
      using const_matrix_view = ConstMatrixView<value_type>;
      if (auto block_ptr = dynamic_cast<block_matrix *>(&matrix))
        {
          const const_matrix_view u_trans {
            const_matrix_view::makeTransposedView(*block_ptr->getBlockU())
          };
          const const_matrix_view v_trans {
            const_matrix_view::makeTransposedView(*block_ptr->getBlockV())
          };
          make_xy_uv_trans(*block_ptr->getBlockA());
          make_xy_uv_trans(*block_ptr->getBlockB());
          // We know that the X and Y blocks only contain zeros, so there
          // is no need to clear them beforehand.
          *block_ptr->getBlockX() += u_trans;
          *block_ptr->getBlockY() += v_trans;
        }
    }

    template<typename T>
    void
    assign_diag(HMatrix<T>& matrix,
                const FullVector<T>& diag,
                const typename HMatrix<T>::index_type begin,
                const typename HMatrix<T>::index_type end)
    {
      typedef T value_type;
      using full_matrix = FullMatrix<value_type>;
      using block_matrix = BlockMatrix<value_type>;
      using index_type = typename full_matrix::index_type;
      using raw_index_type = typename full_matrix::raw_index_type;
      assert(matrix.size() == end - begin);
      if (auto block_ptr = dynamic_cast<block_matrix *>(&matrix))
        {
          index_type begin_b {begin + block_ptr->getBlockA()->size()};
          index_type end_b {begin_b + block_ptr->getBlockB()->size()};
          assign_diag(*block_ptr->getBlockA(), diag, begin, begin_b);
          assign_diag(*block_ptr->getBlockB(), diag, begin_b, end_b);
          assign_diag(*block_ptr->getBlockC(), diag, end_b, end);
        }
      else if (auto full_ptr = dynamic_cast<full_matrix *>(&matrix))
        {
          const raw_index_type rb {static_cast<raw_index_type>(begin)};
          const raw_index_type re {static_cast<raw_index_type>(end)};
          full_ptr->getRawMatrix().diagonal() = diag.getRawVector()
            .block(rb, 0, re - rb, 1);
        }
      else
        {
          BSC_KLAMMLER_NOT_REACHED;
        }
    }

    template<typename T>
    void
    fill_h_matrix(HMatrix<T>& matrix, const bool spd)
    {
      typedef T value_type;
      using h_matrix = HMatrix<value_type>;
      using full_matrix = FullMatrix<value_type>;
      using full_vector = FullVector<value_type>;
      using active_visitor = typename h_matrix::ActiveVisitor;
      using index_type = typename h_matrix::index_type;
      using raw_index_type = typename full_matrix::raw_index_type;
      struct Visitor : public active_visitor
      {
        bool symmetric;
        full_vector one;
        full_vector rowSums;

        Visitor(const bool s, const index_type n) :
          symmetric {s}, one {n}, rowSums {n}
        {
          this->one.getRawVector().setOnes();
        }

        virtual void
        visitRegion(const index_type global_north,
                    const index_type global_west,
                    const bool global_trans BSC_KLAMMLER_ATTRIBUTE_UNUSED,
                    h_matrix& matrix,
                    const index_type north,
                    const index_type west,
                    const index_type south,
                    const index_type east) final override
        {
          // We know this can't happen since this is an internal
          // function and we've only put in untransposed FullMatrix'es.
          assert(!global_trans);
          if (auto full_ptr = dynamic_cast<full_matrix *>(&matrix))
            {
              const raw_index_type rn {static_cast<raw_index_type>(north)};
              const raw_index_type rw {static_cast<raw_index_type>(west)};
              const raw_index_type rs {static_cast<raw_index_type>(south)};
              const raw_index_type re {static_cast<raw_index_type>(east)};
              const raw_index_type rr {rs - rn};
              const raw_index_type rc {re - rw};
              const raw_index_type rgn {
                static_cast<raw_index_type>(global_north)
              };
              const raw_index_type rgw {
                static_cast<raw_index_type>(global_west)
              };
              const raw_index_type rgs {rgn + rr};
              const raw_index_type rge {rgw + rc};
              // We have no business in filling a block above the main diagonal
              // we are filling a symmetric matrix.
              if (this->symmetric && global_west > global_north)
                return;
              full_ptr->getRawMatrix().block(rn, rw, rr, rc).setRandom();
              // If we are not filling a symmetric matrix, this isn't going to
              // become a positive definite one either, so we have no business
              // in computing the row sum.
              if (!this->symmetric)
                return;
              if (global_north == global_west)
                {
                  // This is a diagonal block.  Make it symmetric and write 1s
                  // all along the main diagonal.
                  auto lower = full_ptr->getRawMatrix()
                    .block(rn, rw, rr, rc)
                    .template triangularView<Eigen::Lower>();
                  auto upper = full_ptr->getRawMatrix()
                    .block(rn, rw, rr, rc)
                    .template triangularView<Eigen::Upper>();
                  upper = lower.transpose();
                  full_ptr->getRawMatrix()
                    .block(rn, rw, rr, rc).diagonal().setOnes();
                }
              else
                {
                  // This is an off-diagonal block.  Add row-sums for the
                  // transposed other part that we won't write.
                  this->rowSums.getRawVector().block(rgw, 0, rc, 1)
                    += full_ptr->getRawMatrix().block(rn, rw, rr, rc)
                    .transpose().cwiseAbs().rowwise().sum();
                }
              // Add row-sums for this block.
              this->rowSums.getRawVector().block(rgn, 0, rr, 1)
                += full_ptr->getRawMatrix().block(rn, rw, rr, rc)
                .cwiseAbs().rowwise().sum();
            }
        }
      };
      Visitor vtor {spd, spd ? matrix.size() : 0};
      pass_over_region(matrix, 0, 0, false, vtor);
      if (spd)
        {
          make_xy_uv_trans(matrix);
          assign_diag(matrix, vtor.rowSums, 0, matrix.size());
        }
    }
  }  // namespace random_problems_hpp


  /**
   * @brief
   *         Generates a random `BlockMatrix`.
   *
   * If the type `T` has support for special values (NaNs, infinities, etc),
   * is made sure that the generated vector contains none of those.
   *
   * @param recursions
   *         depth of the `BlockMatrix` (must be at least 1)
   *
   * @param min_bs
   *         minimum size of a block (must be at least 2)
   *
   * @param min_bs
   *         maximum size of a block
   *         (must be at least <code><var>min_bs</var></code>)
   *
   * @param spd
   *         whether to fill the matrix such that it is symmetric and positive
   *         definite
   *
   * @returns
   *         random matrix
   *
   * @throws std::invalid_argument
   *         if <code><var>recursions</var> &lt; 1 || <var>min_bs</var> &lt; 2
   *         || <var>max_bs</var> &lt; <var>min_bs</var></code>
   *
   */
  template<typename T>
  BlockMatrix<T>
  get_random_block_matrix(const unsigned recursions,
                          const typename HMatrix<T>::index_type min_bs,
                          const typename HMatrix<T>::index_type max_bs,
                          const bool spd)
  {
    using random_problems_hpp::fill_h_matrix;
    using random_problems_hpp::get_random_h_matrix;
    typedef T value_type;
    using block_matrix = BlockMatrix<value_type>;
    if (BSC_KLAMMLER_CHECK_ARGS)
      {
        if (recursions < 1)
          throw std::invalid_argument {"recursions < 1"};
        if (min_bs < 2)
          throw std::invalid_argument {"min_bs < 2"};
        if (max_bs < min_bs)
          throw std::invalid_argument {"max_bs < min_bs"};
      }
    auto poly_uptr = get_random_h_matrix<value_type>(recursions, min_bs, max_bs, spd);
    fill_h_matrix(*poly_uptr, spd);
    auto block_on_heap = dynamic_cast<block_matrix *const>(poly_uptr.get());
    assert(block_on_heap != nullptr);  // we did require 'recursions >= 1'
    block_matrix block_on_stack {std::move(*block_on_heap)};
    assert(poly_uptr->size() == 0);  // did we really move?
    // TODO: Assert no NaNs and infinities are in the matrix.
    return block_on_stack;
  }


  /**
   * @brief
   *         Generates a random vector of given size.
   *
   * The vector will be filled with random values.
   *
   * If the type `T` has support for special values (NaNs, infinities, etc),
   * is made sure that the generated vector contains none of those.
   *
   * @param size
   *         desired size of the vector
   *
   * @returns
   *         random vector
   *
   */
  template<typename T>
  FullVector<T>
  get_random_vector(const typename FullVector<T>::index_type size)
  {
    typedef T value_type;
    using full_vector = FullVector<value_type>;
    full_vector vector {size};
    vector.getRawVector().setRandom();
    // TODO: Assert no NaNs and infinities are in the vector.
    return vector;
  }
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_RANDOM_PROBLEMS_HPP
