/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file rectangle.hpp
 *
 * @brief
 *         Header file providing the `Rectangle` template class and associated
 *         functions.
 *
 */

#ifndef BSC_KLAMMLER_RECTANGLE_HPP
#define BSC_KLAMMLER_RECTANGLE_HPP

#include <type_traits>  // std::is_integral

namespace bsc_klammler
{

  template<typename T>
  class Rectangle
  {

    static_assert(std::is_integral<T>::value,
                  "'Rectangle's can only have integral coordinates");

  public:

    typedef T index_type;

  private:

    index_type north_ {};
    index_type west_ {};
    index_type south_ {};
    index_type east_ {};

  public:

    Rectangle() noexcept
    {
      // do nothing
    }

    Rectangle(const index_type north,
              const index_type west,
              const index_type south,
              const index_type east) :
      north_ {north}, west_ {west}, south_ {south}, east_ {east}
    {
      if (BSC_KLAMMLER_CHECK_ARGS)
        {
          if (south < north)
            throw std::invalid_argument {"south < north"};
          if (east < west)
            throw std::invalid_argument {"east < west"};
        }
    }

    Rectangle(const Rectangle&) noexcept = default;

    Rectangle(Rectangle&&) noexcept = default;

    Rectangle&
    operator=(const Rectangle&) noexcept = default;

    Rectangle&
    operator=(Rectangle&&) noexcept = default;

    index_type
    height() const noexcept
    {
      return this->south_ - this->north_;
    }

    index_type
    width() const noexcept
    {
      return this->east_ - this->west_;
    }

    index_type
    north() const noexcept
    {
      return this->north_;
    }

    index_type
    west() const noexcept
    {
      return this->west_;
    }

    index_type
    south() const noexcept
    {
      return this->south_;
    }

    index_type
    east() const noexcept
    {
      return this->east_;
    }

    index_type
    area() const noexcept
    {
      return this->height() * this->width();
    }
  };  // class Rectangle

  template<typename T>
  bool
  overlap(const Rectangle<T> first,
          const Rectangle<T> second) noexcept
  {
    if (first.height() == 0 || first.width() == 0)
      return false;
    if (second.height() == 0 || second.width() == 0)
      return false;
    if (first.east() <= second.west())
      return false;
    if (first.west() >= second.east())
      return false;
    if (first.south() <= second.north())
      return false;
    if (first.north() >= second.south())
      return false;
    return true;
  }
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_RECTANGLE_HPP
