/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file recursive_spd_preconditioner.hpp
 *
 * @brief
 *         Header file providing the `RecursiveSPDPreconditioner` template
 *         class.
 *
 */

#ifndef BSC_KLAMMLER_RECURSIVE_SPD_PRECONDITIONER_HPP
#define BSC_KLAMMLER_RECURSIVE_SPD_PRECONDITIONER_HPP

#include <atomic>                     // std::atomic, ...
#include <cassert>                    // assert
#include <cmath>                      // std::sqrt
#include <condition_variable>         // std::condition_variable
#include <limits>                     // std::numeric_limits
#include <mutex>                      // std::mutex, std::unique_lock
#include <thread>                     // std::thread

#include "algorithms.hpp"             // (linear algebra operations)
#include "arranged_matrix.hpp"        // bsc_klammler::ArrangedMatrix
#include "block_matrix.hpp"           // bsc_klammler::BlockMatrix
#include "full_matrix.hpp"            // bsc_klammler::FullMatrix
#include "full_vector.hpp"            // bsc_klammler::FullVector
#include "inefficient_algorithm.hpp"  // BSC_KLAMMLER_INEFFICINT_ALGORITHM
#include "macros.hpp"                 // (package-specific macros)
#include "matrix_view.hpp"            // bsc_klammler::MatrixView
#include "non_zero_counter.hpp"       // bsc_klammler::non_zero_count
#include "preconditioner.hpp"         // bsc_klammler::Preconditioner


namespace bsc_klammler
{

  /**
   * @brief
   *         Implementation details of the `recursive_spd_preconditioner.hpp`
   *         header file.
   *
   */
  namespace recursive_spd_preconditioner_hpp
  {
    /**
     * @brief
     *         A thread-safe counter.
     *
     * @param <T>
     *         type for the counter value
     *
     */
    template<typename T>
    class AtomicCounter
    {

      static_assert(std::is_integral<T>::value,
                    "The 'T' in 'AtomicCounter<T>' must be an integral type");

    public:

      /** @brief type for the counter value */
      typedef T counter_type;

    private:

      /** @brief Current counter value. */
      std::atomic<counter_type> value_ {0};

    public:

      /**
       * @brief
       *         RAII-style handle for incrementing and safely decrementing an
       *         `AtomicCounter`.
       *
       */
      class Handle
      {

      private:

        /** @brief Pointer to the managed `AtomicCounter` or a `nullptr`. */
        AtomicCounter *const acPtr_;

      public:

        /**
         * @brief
         *         Constructor: Creates a new `Handle` for the given
         *         `AtomicCounter` or a dummy object if passed a `nullptr`.
         *
         * @param ptr
         *         pointer to the `AtomicCounter` to manage
         *
         */
        Handle(AtomicCounter *const ac_ptr) noexcept : acPtr_ {ac_ptr}
        {
          if (this->acPtr_)
            this->acPtr_->increment();
        }

        /**
         * @brief
         *         Destructor: Decrements the `AtomicCounter` managed by this
         *         `Handle`, if any.
         *
         */
        ~Handle() noexcept
        {
          if (this->acPtr_)
            this->acPtr_->decrement();
        }
      };

    public:

      /**
       * @brief
       *         Constructor: Creates a new `AtomicCounter` initialized to 0.
       *
       * Warning: The construction process is *not* atomic!
       *
       */
      AtomicCounter() noexcept
      {
        assert(!this->current());
      }

    private:

      // Explicit deletion of these is required for older GCCs and makes our
      // intent clear regardlss of the compiler.

      /** @brief Deleted copy constructor. */
      AtomicCounter(const AtomicCounter&) = delete;

      /** @brief Deleted move constructor. */
      AtomicCounter(AtomicCounter&&) = delete;

      /** @brief Deleted copy-asignment operator. */
      AtomicCounter&
      operator=(const AtomicCounter&) = delete;

      /** @brief Deleted move-asignment operator. */
      AtomicCounter&
      operator=(AtomicCounter&&) = delete;

    public:

      /**
       * @brief
       *         Atomically returns the current counter value.
       *
       * @returns
       *         current counter value
       *
       */
      counter_type
      current() const noexcept
      {
        return std::atomic_load(&this->value_);
      }

      /**
       * @brief
       *         Atomically increments the counter.
       *
       * @returns
       *         the *new* counter value
       *
       */
      counter_type
      increment() noexcept
      {
        constexpr counter_type one {static_cast<counter_type>(1)};
        const counter_type prev {std::atomic_fetch_add(&this->value_, one)};
        return prev + one;
      }

      /**
       * @brief
       *         Atomically decrements the counter.
       *
       * @returns
       *         the *new* counter value
       *
       */
      counter_type
      decrement() noexcept
      {
        constexpr counter_type one {static_cast<counter_type>(1)};
        const counter_type prev {std::atomic_fetch_sub(&this->value_, one)};
        return prev - one;
      }

      /**
       * @brief
       *         Returns a `Handle` for this `AtomicCounter` or a dummy
       *         `Handle` based on the value of the
       *         <code><var>predicate</var></code>.
       *
       * @param predicate
       *         whether to return an actual `Handle`, not a dummy object
       *
       * @returns
       *         a `Handle` for this `AtomicCounter` or a dummy
       *
       */
      AtomicCounter::Handle
      handle(const bool predicate) noexcept
      {
        return AtomicCounter::Handle {predicate ? this : nullptr};
      }
    };  // class AtomicCounter
  }  // namespace recursive_spd_preconditioner_hpp

  /**
   * @brief
   *         A template class (in both, the OO and the C++ sense) for recursive
   *         `Preconditioner`s for symmetric and positive definite (spd) linear
   *         systems.
   *
   * @param <T>
   *         type of the variables in the linear system
   *
   * @param <I>
   *         whether the factor <var>L</var> is stored without the (implicits)
   *         1s along the main diagonal
   *
   */
  template<typename T, bool I>
  class RecursiveSPDPreconditioner : public Preconditioner<T>
  {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Type of the variables in the linear system. */
    using value_type = typename Preconditioner<T>::value_type;

    /** @brief Type for the dimensionality of the linear system. */
    using index_type = typename Preconditioner<T>::index_type;

    /** @brief Type for counting FLOPs. */
    using flops_type = typename Preconditioner<T>::flops_type;

  protected:

    /** @brief Convenience type alias for `HMatrix`es. */
    using h_matrix = HMatrix<value_type>;

    /** @brief Convenience type alias for `FullMatrix`es. */
    using full_matrix = FullMatrix<value_type>;

    /** @brief Convenience type alias for `BlockMatrix`es. */
    using block_matrix = BlockMatrix<value_type>;

    /** @brief Convenience type alias for `ArrangedMatrix`es. */
    using arranged_matrix = ArrangedMatrix<value_type>;

    /** @brief Convenience type alias for mutabl `MutableMatrixView`s. */
    using mutable_matrix_view = MutableMatrixView<value_type>;

    /** @brief Convenience type alias for `ConstMatrixView`s. */
    using const_matrix_view = ConstMatrixView<value_type>;

    /** @brief Convenience type alias for `HVector`s. */
    using h_vector = HVector<value_type>;

    /** @brief Convenience type alias for `FullVector`s. */
    using full_vector = FullVector<value_type>;


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief (Temporary) clone of the coefficient matrix to be modified. */
    std::weak_ptr<h_matrix> matrixWptr_ {};

    /** @brief Lower triangular factor <var>L</var>. */
    arranged_matrix factorL_ {};

    /** @brief Mutex used for locking `factorL_`. */
    std::mutex mutexL_ {};

    /** @brief Mutex used for &ldquo;locking all CPUs&rdquo;. */
    std::mutex mutexAllCPUs_ {};

    /** @brief COndition variable used to signal all CPUs are ready. */
    std::condition_variable conditionAllCPUs_ {};

    /** @brief Once-flag for enabling multithreading for dense operations. */
    std::once_flag enableMPOnceFlag_ {};

    /** @brief Number of nodes below the &ldquo;Single CPU&rdquo; level. */
    recursive_spd_preconditioner_hpp::AtomicCounter<unsigned long> singleCPUNodes_ {};


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  protected:

    /**
     * @brief
     *         Constructor: Creates a new `RecursiveSPDPreconditioner` base.
     *
     * @param size
     *         size of the linear system
     *
     */
    RecursiveSPDPreconditioner(const index_type size) :
      Preconditioner<T> {size}, factorL_ {size, size}
    {
      // do nothing
    }


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

  public:

    /**
     * @brief
     *         Tells that this `Preconditioner` takes advantage of the special
     *         structure of hirachical matrices.
     *
     * @returns
     *         `true`
     *
     */
    virtual bool
    isRecursive() const noexcept final override
    {
      return true;
    }

    /**
     * @brief
     *         Tells that this `Preconditioner` can only solve symmetric and
     *         positive definite (spd) linear systems.
     *
     * @returns
     *         `true`
     *
     */
    virtual bool
    onlySPD() const noexcept final override
    {
      return true;
    }

  protected:

    // {@inheritDoc}
    virtual void
    initImpl(const HMatrix<value_type>& matrix) final override
    {
      // Parameter validation happened in superclass.
      std::shared_ptr<h_matrix> matrix_sptr {matrix.clone()};
      this->matrixWptr_ = matrix_sptr;
      enable_data_level_multithreading(1);
      const arranged_matrix last_diff {
        this->recursiveDecompose_(*matrix_sptr.get(), 0, 0)
      };
      assert(last_diff.size() == 0);
      this->getFactorL().sortTiles();
      assert(is_lower_triangular(
          this->getFactorL(),
          RecursiveSPDPreconditioner::hasImplicit1sOnL()));
      assert(!this->matrixWptr_.expired());
    }

    // {@inheritDoc}
    virtual full_vector
    solveImpl(const h_vector& right) const final override
    {
      // Parameter validation happened in superclass.
      assert(this->matrixWptr_.expired());
      return this->solveImplImpl(right);
    }

    /**
     * @brief
     *         Solves the linear system
     *         <var>P</var>&sdot;<var>y</var>=<var>c</var> where <var>P</var>
     *         is the preconditioner.
     *
     * The `solveImpl()` function member inherited from `Preconditioner` is
     * implemented to delegate to this function.  The actual solving logic is
     * to be implemented by derived classes but will usually consist of a
     * forward substitution on <var>L</var> and a backward substitution on
     * <var>L</var><sup>T</sup> but maybe more and maybe with optimizations.
     *
     * @param right
     *         right-side to solve the system for
     *
     * @returns
     *         an (approximate) solution to the linear system
     *
     */
    virtual full_vector
    solveImplImpl(const h_vector& right) const = 0;

  private:

    /**
     * @brief
     *         Decomposes the sub-matrix <code><var>m</var></code> recursively.
     *
     * @param m
     *         sub-matrix to decompose
     *
     * @param offset
     *         offset of <code><var>m</var></code> in the coefficient matrix
     *
     * @param level
     *         recursion depth (only count entering this function)
     *
     * @returns
     *         those parts of the updates to the coefficient matrix that would
     *         create a data race
     *
     */
    arranged_matrix
    recursiveDecompose_(const h_matrix& m,
                        const index_type offset,
                        const unsigned level)
    {
      const auto count_push = this->singleCPUNodes_.handle(
          this->isBelowSingleCPULevel_(level));
      if (auto bm_ptr = dynamic_cast<const block_matrix *>(&m))
        return this->decomposeBlock_(*bm_ptr, offset, level);
      else
        return this->decomposeLeaf_(m, offset, level);
    }

    /**
     * @brief
     *         Decomposes the sub-matrix <code><var>m</var></code> (wich is
     *         itself a `BlockMatrix`) recursively.
     *
     * @param m
     *         sub-matrix to decompose
     *
     * @param offset
     *         offset of <code><var>m</var></code> in the coefficient matrix
     *
     * @param level
     *         recursion depth (only counting entering `recursiveDecompose_`)
     *
     * @returns
     *         those parts of the update to the coefficient matrix that would
     *         create a data race
     *
     */
    // NB: The FLOP counting for this function /is complete/.  We must not
    // count "adding" the updates to the coefficient matrix as a separate
    // operation since it was already counted when these updates were computed.
    // Remember that a 1 FLOP = 1 multiplication + 1 addition.  The way we are
    // deferring the addition is an implementation detail, not an algorithmical
    // neccessety.  By the way, our "adding" doesn't really add anything but
    // merely moves a few pointers...
    arranged_matrix
    decomposeBlock_(const block_matrix& m,
                    const index_type offset,
                    const unsigned level)
    {
      constexpr bool do_the_cpu_lock_thing {
        BSC_KLAMMLER_PARALLEL >= BSC_KLAMMLER_PARALLEL_DATA
      };
      std::unique_lock<std::mutex> all_cpus_lock {
        this->mutexAllCPUs_, std::defer_lock
      };
      const index_type n {this->size()};
      const index_type n_m {m.size()};
      const index_type n_a {m.getBlockA()->size()};
      const index_type n_b {m.getBlockB()->size()};
      const index_type n_c {m.getBlockC()->size()};
      const index_type off_a {offset};
      const index_type off_b {off_a + n_a};
      const index_type off_c {off_b + n_b};
      arranged_matrix delta_a {};
      arranged_matrix delta_b {};
      assert(n_m == n_a + n_b + n_c);
      this->recursiveDispatch_(delta_a, *m.getBlockA(), off_a,
                               delta_b, *m.getBlockB(), off_b,
                               level);
      if (do_the_cpu_lock_thing)
        {
          if (!this->isBelowSingleCPULevel_(level))
            {
              all_cpus_lock.lock();
              while (this->singleCPUNodes_.current())
                this->conditionAllCPUs_.wait(all_cpus_lock);
            }
          assert(all_cpus_lock.owns_lock() != this->isBelowSingleCPULevel_(level));
        }
      // The remainder of this function deals with handling the updates.
      arranged_matrix delta_m {n - offset - n_m, n - offset - n_m};
      std::shared_ptr<h_matrix> matrix_sptr {this->matrixWptr_.lock()};
      mutable_matrix_view view_a {
        matrix_sptr.get(), off_b, off_b, delta_a.rows(), delta_a.cols()
      };
      mutable_matrix_view view_b {
        matrix_sptr.get(), off_c, off_c, delta_b.rows(), delta_b.cols()
      };
      auto src_a = const_matrix_view::makeQuarters(delta_a, n_b + n_c);
      auto src_b = const_matrix_view::makeQuarters(delta_b, n_c, n_c);
      auto dst_a = mutable_matrix_view::makeQuarters(view_a, n_b + n_c);
      auto dst_b = mutable_matrix_view::makeQuarters(view_b, n_c, n_c);
      assert(delta_a.size() == n - offset - n_a);
      assert(delta_b.size() == n - offset - n_a - n_b);
      assert(delta_m.size() == n - offset - n_a - n_b - n_c);
      assert(view_a.size() == delta_a.size());
      assert(view_b.size() == delta_b.size());
      // These updates are safe to write...
      dst_a.north_west -= src_a.north_west;
      dst_a.north_east -= src_a.north_east;
      dst_a.south_west -= src_a.south_west;
      dst_b.north_west -= src_b.north_west;
      dst_b.north_east -= src_b.north_east;
      dst_b.south_west -= src_b.south_west;
      // ...but these are not.
      delta_m += src_a.south_east;
      delta_m += src_b.south_east;
      if (n_c > 0)
        {
          const arranged_matrix delta_c {
            this->decomposeLeaf_(*m.getBlockC(), off_c, level)
          };
          assert(delta_c.size() == n - offset - n_m);
          assert(delta_c.size() == delta_m.size());
          delta_m += delta_c;
        }
      else
        {
          assert(delta_m.size() == 0);
        }
      if (do_the_cpu_lock_thing)
        {
          if (!this->isBelowSingleCPULevel_(level))
            {
              all_cpus_lock.unlock();
              this->conditionAllCPUs_.notify_one();
            }
        }
      return delta_m;
    }

    /**
     * @brief
     *         Decomposes the sub-matrix <code><var>m</var></code>.  This
     *         function does not recurse.
     *
     * @param m
     *         sub-matrix to decompose
     *
     * @param offset
     *         offset of <code><var>m</var></code> in the coefficient matrix
     *
     * @param level
     *         recursion depth (only counting entering `recursiveDecompose_`)
     *
     * @returns
     *         update to the coefficient matrix
     *
     */
    arranged_matrix
    decomposeLeaf_(const h_matrix& m,
                   const index_type offset,
                   const unsigned level BSC_KLAMMLER_ATTRIBUTE_UNUSED)
    {
      std::shared_ptr<h_matrix> matrix_sptr {this->matrixWptr_.lock()};
      const arranged_matrix m12 {
        arranged_matrix::makeArrangement(
            *matrix_sptr,
            offset,
            offset + m.cols(),
            offset + m.rows(),
            matrix_sptr->cols(),
            const_matrix_view::makeView)
      };
      const arranged_matrix m21 {
        arranged_matrix::makeArrangement(
            *matrix_sptr,
            offset + m.rows(),
            offset,
            matrix_sptr->rows(),
            offset + m.cols(),
            const_matrix_view::makeView)
      };
      const const_matrix_view m22 {
        matrix_sptr.get(),
        offset + m.rows(),
        offset + m.cols(),
        matrix_sptr->rows() - offset - m.rows(),
        matrix_sptr->cols() - offset - m.cols(),
      };
      assert(m.rows() == m12.rows());
      assert(m.cols() == m21.cols());
      assert(m12.cols() == m22.cols());
      assert(m21.rows() == m22.rows());
      assert(m.size() + m22.size() + offset == matrix_sptr->size());
      assert(m12.isRowLinear());
      assert(m21.isColumnLinear());
      if (auto full_ptr = dynamic_cast<const full_matrix *>(&m))
        {
          return this->decomposeImpl(*full_ptr, m12, m21, m22, offset);
        }
      else
        {
          BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
          const full_matrix full_copy {m};
          return this->decomposeImpl(full_copy, m12, m21, m22, offset);
        }
    }

    /**
     * @brief
     *         Dispatches the recursive calls for decomposing two sub-blocks.
     *
     * @param delta_a
     *         destination for the update obtained from recursively decomposing
     *         the sub-block <var>A</var>
     *
     * @param block_a
     *         sub-block <var>A</var>
     *
     * @param offset_a
     *         offset of sub-block <var>A</var> in the coefficient matrix
     *
     * @param delta_b
     *         destination for the update obtained from recursively decomposing
     *         the sub-block <var>B</var>
     *
     * @param block_b
     *         sub-block <var>B</var>
     *
     * @param offset_b
     *         offset of sub-block <var>B</var> in the coefficient matrix
     *
     * @param level
     *         recursion depth (only counting entering `recursiveDecompose_`)
     *
     */
    void
    recursiveDispatch_(/* !const */ arranged_matrix& delta_a,
                       const h_matrix& block_a,
                       const index_type offset_a,
                       /* !const */ arranged_matrix& delta_b,
                       const h_matrix& block_b,
                       const index_type offset_b,
                       const unsigned level)
    {
      constexpr bool never_spawn_threads {
        BSC_KLAMMLER_PARALLEL < BSC_KLAMMLER_PARALLEL_STRUCTURAL
      };
      constexpr bool never_continue_on_all_cpus {
        BSC_KLAMMLER_PARALLEL < BSC_KLAMMLER_PARALLEL_DATA
      };
      auto task_a = [&delta_a, &block_a, offset_a, level, this](){
        arranged_matrix tmp {recursiveDecompose_(block_a, offset_a, level + 1)};
        swap(delta_a, tmp);
      };
      auto task_b = [&delta_b, &block_b, offset_b, level, this](){
        arranged_matrix tmp {recursiveDecompose_(block_b, offset_b, level + 1)};
        swap(delta_b, tmp);
      };
      if (never_spawn_threads || this->isBelowSingleCPULevel_(level))
        {
          task_a();
          task_b();
        }
      else
        {
          ThreadGuard guarded_threads[] = {
            std::thread {task_a},
            std::thread {task_b},
          };
          for (auto& iter : guarded_threads)
            iter.thread().join();
          if (!never_continue_on_all_cpus)
            {
              std::call_once(this->enableMPOnceFlag_,
                             enable_data_level_multithreading,
                             this->getNumCPUs());
            }
        }
    }

    /**
     * @brief
     *         Checks whether <code><var>level</var></code> is blow the
     *         &ldquo;Single CPU&rdquo; level.
     *
     * Below the &ldquo;Single CPU&rdquo; level, all work should be done
     * sequentially on a single CPU and recursion shall not fork.  Above it,
     * recursion shall fork and combining the results should synchronize and
     * then use all CPUs.
     *
     * @param level
     *         recursion depth (only counting entering `recursiveDecompose_`)
     *
     * @returns
     *         whether the <code><var>level</var></code> is below the
     *         &ldquo;Single CPU&rdquo; level
     *
     */
    bool
    isBelowSingleCPULevel_(const unsigned level) const noexcept
    {
      return (1UL << level) >= static_cast<unsigned long>(this->getNumCPUs());
    }

  protected:

    /**
     * @brief
     *         Returns whether the factor <var>L</var> is stored without the
     *         (implicit) 1s along the main diagonal.
     *
     * This function simply returns the template parameter `I`.
     *
     * @returns
     *         whether the factor <var>L</var> is stored without the (implicit)
     *         1s along the main diagonal
     *
     */
    static constexpr bool
    hasImplicit1sOnL() noexcept
    {
      return I;
    }

    /**
     * @brief
     *         Returns an exclusive lock for the factor <var>L</var>.
     *
     * @returns
     *         lock for <var>L</var>
     *
     */
    std::unique_lock<std::mutex>
    lockL()
    {
      return std::unique_lock<std::mutex> {this->mutexL_};
    }

    /**
     * @brief
     *         Returns a reference to the factor <var>L</var>.
     *
     * While in the concurrent section, the matrix needs to be locked before
     * accessed.
     *
     *     do_something_before();
     *     {
     *       const auto lck = this->lockL();
     *       std::cout << this->getFactorL().size() << std::endl;
     *     }
     *     do_something_after();
     *
     * @returns
     *         reference to the factor <var>L</var>
     *
     */
    arranged_matrix&
    getFactorL() noexcept
    {
      return this->factorL_;
    }

    /**
     * @brief
     *         Returns a `const` reference to the factor <var>L</var>.
     *
     *     do_something_before();
     *     {
     *       const auto lck = this->lockL();
     *       std::cout << this->getFactorL().size() << std::endl;
     *     }
     *     do_something_after();
     *
     * @returns
     *         `const` reference to the factor <var>L</var>
     *
     */
    const arranged_matrix&
    getFactorL() const noexcept
    {
      return this->factorL_;
    }

    /**
     * @brief
     *         Work-horse for decomposing a full sub-matrix.
     *
     * In the following illustration, suppose we have already decomposed the
     * &ldquo;hatched&rdquo; part of the coefficient matrix.  Then this
     * function will be called with referencees to <var>M</var><sub>11</sub>,
     * <var>M</var><sub>12</sub>, <var>M</var><sub>21</sub> and
     * <var>M</var><sub>22</sub> and the offset of the sub-matrix inside the
     * coefficient matrix (which is the same from the left and from the top).
     *
     *     +---------------------------------+
     *     |:::::::::::::::::::::::::::::::::|
     *     |:::::::::::::::::::::::::::::::::|
     *     | offset :::::::::::::::::::::::::|
     *     |<------>+-----+------------------+
     *     |::::::::| M11 | M12              |
     *     |::::::::|     |                  |
     *     |::::::::+-----+------------------+
     *     |::::::::| M21 | M22              |
     *     |::::::::|     |                  |
     *     |::::::::|     |                  |
     *     |::::::::|     |                  |
     *     |::::::::|     |                  |
     *     |::::::::|     |                  |
     *     |::::::::|     |                  |
     *     |::::::::|     |                  |
     *     +--------+-----+------------------+
     *
     * This function ought to compute a decomposition of
     * <var>M</var><sub>11</sub> and update <var>L</var> accordingly, then
     * return an update for the coefficient matrix (to be subtracted from
     * <var>M</var><sub>22</sub>).  What this means exactly may be defined
     * differently by implementing classes.
     *
     * This is the only function responsible for the FLOP bookeeping.
     *
     * @param m11
     *         full sub-matrix to decompose
     *
     * @param m12
     *         sub-matrix right of <var>M</var><sub>11</sub>
     *
     * @param m21
     *         sub-matrix below of <var>M</var><sub>11</sub>
     *
     * @param m22
     *         remaining sub-matrix (might be a `MatrixView`)
     *
     * @param offset
     *         offset of <var>M</var><sub>11</sub> in the coefficient matrix
     *
     * @returns
     *         update to the coefficient matrix
     *
     */
    virtual arranged_matrix
    decomposeImpl(const full_matrix& m11,
                  const arranged_matrix& m12,
                  const arranged_matrix& m21,
                  const h_matrix& m22,
                  const index_type offset) = 0;

    /**
     * @brief
     *         Checks whether the error of a solution found by substitution is
     *         &ldquo;small&rdquo;.
     *
     * @param vec
     *         right-hand side the system was solved for
     *
     * @param res
     *         solution that was found
     *
     * @param trans
     *         whether the system was solved for <var>L</var> (`false`) or
     *         <var>L</var><sup>T</sup> (`true`)
     *
     * @param tol
     *         tolerated relative residual
     *         (defaults to &epsilon;<sup>1/2</sup>)
     *
     * @returns
     *         whether the actual reletive residual is smaller than the maximum
     *         tolerated
     *
     */
    bool
    checkSubstitution(const h_vector& vec,
                      const h_vector& res,
                      const bool trans,
                      const value_type tol = std::sqrt(std::numeric_limits<value_type>::epsilon())) const
    {
      BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
      full_matrix lower {const_matrix_view {
          &this->getFactorL(), 0, 0, this->size(), this->size(), trans
      }};
      if (this->hasImplicit1sOnL())
        {
          for (index_type i = 0; i < lower.size(); ++i)
            {
              assert(lower.get(i, i) == h_matrix::zero());
              lower.set(static_cast<value_type>(1), i, i);
            }
        }
      const full_vector residual {lower * res - vec};
      const value_type rel_residual {norm(residual) / norm(vec)};
      assert(tol > 0);
      return rel_residual < tol;
    }
  };  // class RecursiveSPDPreconditioner
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_RECURSIVE_SPD_PRECONDITIONER_HPP
