/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file simple_logger.hpp
 *
 * @brief
 *         Header file providing the `SimpleLogger` template class.
 *
 */

#ifndef BSC_KLAMMLER_SIMPLE_LOGGER_HPP
#define BSC_KLAMMLER_SIMPLE_LOGGER_HPP

#include <cmath>                 // std::sqrt
#include <cstdlib>               // std::getenv
#include <cstring>               // std::strlen
#include <fstream>               // std::ofstream
#include <iomanip>               // std::setw, ...
#include <iostream>              // std::cout, std::endl, ...
#include <limits>                // std::numeric_limits

#include "full_matrix.hpp"       // bsc_klammler::FullMatrix
#include "logger.hpp"            // bsc_klammler::Logger
#include "macros.hpp"            // (package-specific macros)
#include "non_zero_counter.hpp"  // bsc_klammler::non_zero_count
#include "numeric_costs.hpp"     // bsc_klammler::numeric_cost::*


namespace bsc_klammler
{

  /**
   * @brief
   *         Simple `Logger` class that prints human-readable messages to the
   *         standard output stream.
   *
   * @param <T>
   *         coefficient type for linear systems
   *
   */
  template<typename T>
  class SimpleLogger : public Logger<T>
  {
  public:

    /** @brief Coefficient type for linear systems. */
    using value_type = typename Logger<T>::value_type;

    /** @brief Index type for linear systems. */
    using index_type = typename Logger<T>::index_type;

    /** @brief Type for counting FLOPs. */
    using flops_type = typename Logger<T>::flops_type;

  private:

    static constexpr value_type solveTol_ {
      std::sqrt(std::numeric_limits<value_type>::epsilon())
    };

    static constexpr unsigned fieldWidthFirst_ {16};

    static constexpr unsigned fieldWidth_ {16};

    static constexpr auto indent_ = " ";

    unsigned numCPUs_ {1};

  public:

    SimpleLogger(const unsigned ncpus) noexcept : numCPUs_ {ncpus}
    {
    }

    // {@inheritDoc}
    virtual void
    logBefore(const HMatrix<T>& coeffs) final override
    {
      const auto elems = coeffs.rows() * coeffs.cols();
      const auto nzc = non_zero_count(coeffs);
      const auto sparse100 = 100.0 * nzc / elems;
      assert(0.0 <= sparse100 && sparse100 <= 100.0);
      std::cout << "Using a random "
                << "(" << coeffs.rows() << ", " << coeffs.cols() << ")"
                << " matrix with "
                << nzc << " (" << std::setprecision(2) << sparse100 << " %)"
                << " non-zero entries."
                << std::endl;
      std::cout << BSC_KLAMMLER_EOL;
      this->toprule_();
      this->log_(false)
        .action("action")
        .any("time")
        .any("CPUs")
        .any("FLOPs")
        .any("efficiency")
        .any("rel error");
      this->log_(false)
        .action("")
        .any("s")
        .any("#")
        .any("#")
        .any("FPS / CPU")
        .any("1");
      this->midrule_();
      std::cout << std::flush;
      this->calibrate_();
    }

    // {@inheritDoc}
    virtual void
    logAfter() final override
    {
      this->bottomrule_();
      std::cout << std::flush;
    }

    // {@inheritDoc}
    virtual void
    logPreconditioner(const Preconditioner<value_type>& precond) final override
    {
      this->midrule_();
      std::cout << SimpleLogger::indent_
                << precond.getLongName()
                << BSC_KLAMMLER_EOL;
      this->midrule_();
      std::cout << std::flush;
    }

    // {@inheritDoc}
    virtual void
    logInit(const Preconditioner<value_type>& precond,
            const Timer::time_delta_type elapsed) final override
    {
      const auto ncpus = precond.getNumCPUs();
      const auto flops = precond.getInitFLOPs();
      const auto handle = this->log_();
      handle.action("decomposing").time(elapsed).cpus(ncpus);
      if (BSC_KLAMMLER_COUNT_FLOPS)
        handle.flops(flops).efficiency(flops, elapsed, ncpus);
      else
        handle.skip().skip();
      handle.skip();
    }

    // {@inheritDoc}
    virtual void
    logSolve(const Preconditioner<value_type>& precond,
             const Timer::time_delta_type elapsed,
             const value_type relative_residual) final override
    {
      const auto ncpus = precond.getNumCPUs();
      const auto flops = precond.getSolveFLOPs();
      const auto handle = this->log_();
      handle.action("solving").time(elapsed).cpus(ncpus);
      if (BSC_KLAMMLER_COUNT_FLOPS)
        handle.flops(flops).efficiency(flops, elapsed, ncpus);
      else
        handle.skip().skip();
      handle.error(relative_residual);
      if (relative_residual >= SimpleLogger::solveTol_)
        handle.any("check this!");
    }

    // {@inheritDoc}
    virtual void
    logSkipping(const std::string& why) final override
    {
      std::cout << SimpleLogger::indent_
                << "skipped: "
                << why
                << std::endl;
    }

  private:

    struct Handle
    {

      const SimpleLogger *const slPtr_;
      const bool flush_;

      Handle(const SimpleLogger *const sl_ptr, const bool flush) noexcept :
        slPtr_ {sl_ptr}, flush_ {flush}
      {
        std::cout << SimpleLogger::indent_;
      }

      ~Handle() noexcept
      {
        std::cout << BSC_KLAMMLER_EOL;
        if (this->flush_)
          std::cout << std::flush;
      }

      const Handle&
      action(const std::string& action) const
      {
        std::cout << std::setw(SimpleLogger::fieldWidthFirst_)
                  << std::left
                  << action;
        return *this;
      }

      const Handle&
      time(const Timer::time_delta_type seconds) const
      {
        std::cout << std::setw(SimpleLogger::fieldWidth_)
                  << std::right
                  << std::setprecision(6)
                  << std::fixed
                  << seconds;
        return *this;
      }

      const Handle&
      cpus(const unsigned ncpus) const
      {
        std::cout << std::setw(SimpleLogger::fieldWidth_)
                  << std::right
                  << std::fixed
                  << ncpus;
        return *this;
      }

      const Handle&
      flops(const flops_type flops) const
      {
        std::cout << std::setw(SimpleLogger::fieldWidth_)
                  << std::right
                  << std::setprecision(3)
                  << std::scientific
                  << flops;
        return *this;
      }

      const Handle&
      efficiency(const flops_type flops,
                 const Timer::time_delta_type seconds,
                 const unsigned ncpus) const
      {
        std::cout << std::setw(SimpleLogger::fieldWidth_)
                  << std::right
                  << std::setprecision(3)
                  << std::scientific
                  << (flops / (seconds * ncpus));
        return *this;
      }

      const Handle&
      error(const value_type relative_error) const
      {
        std::cout << std::setw(SimpleLogger::fieldWidth_)
                  << std::right
                  << std::setprecision(3)
                  << std::scientific
                  << relative_error;
        return *this;
      }

      const Handle&
      any(const std::string& message) const
      {
        std::cout << std::setw(SimpleLogger::fieldWidth_)
                  << std::right
                  << message;
        return *this;
      }

      const Handle&
      skip() const
      {
        return this->any("n/a");
      }
    };  // struct Handle

    const SimpleLogger::Handle
    log_(const bool flush = true) noexcept
    {
      return SimpleLogger::Handle {this, flush};
    }

    void
    toprule_() const
    {
      this->rule_('=');
    }

    void
    midrule_() const
    {
      this->rule_('-');
    }

    void
    bottomrule_() const
    {
      this->rule_('=');
    }

    void
    rule_(const char dash) const
    {
      const std::size_t width {
        SimpleLogger::fieldWidthFirst_
          + 5 * SimpleLogger::fieldWidth_
          + 2 * std::strlen(SimpleLogger::indent_)
      };
      for (std::size_t length = 0; length < width; ++length)
        std::cout << dash;
      std::cout << BSC_KLAMMLER_EOL;
    }

    void
    calibrate_()
    {
      Timer timer {};
      constexpr std::size_t size {BSC_KLAMMLER_DEBUG ? 256UL : 4096UL};
      FullMatrix<value_type> sample {size, size};
      FullMatrix<value_type> squared {};
      sample.getRawMatrix().setRandom();
      timer.start();
      squared = sample * sample;
      timer.stop();
      const auto flops = numeric_cost::matrix_matrix_product(size, size, size);
      const auto seconds = timer.getSeconds();
      const auto ncpus = enable_data_level_multithreading();
      this->log_()
        .action("calibrating")
        .time(seconds)
        .cpus(ncpus)
        .flops(flops)
        .efficiency(flops, seconds, ncpus)
        .skip();
      if (const auto envval = std::getenv("BSC_KLAMMLER_DUMP_GARBAGE"))
        {
          // This should never be executed but we keep it here so computation
          // of the squared matrix cannot be eliminated.
          std::ofstream sink {envval};
          sink << sample << squared << BSC_KLAMMLER_EOL;
          sink.close();
        }
    }
  };  // class SimpleLogger
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_SIMPLE_LOGGER_HPP
