/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file sql_logger.hpp
 *
 * @brief
 *         Header file providing the `SQLLogger` template class.
 *
 */

#ifndef BSC_KLAMMLER_SQL_LOGGER_HPP
#define BSC_KLAMMLER_SQL_LOGGER_HPP

#include "logger.hpp"            // bsc_klammler::Logger

#include <chrono>                // std::chrono::system_clock
#include <functional>            // std::hash
#include <iomanip>               // std::setprecision, ...
#include <iostream>              // std::cout, ...
#include <sstream>               // std::ostringstream
#include <type_traits>           // std::is_integral, ...

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "non_zero_counter.hpp"  // bsc_klammler::non_zero_count


namespace bsc_klammler
{

  /**
   * @brief
   *         Implementation details for the `sql_logger.hpp` file.
   *
   */
  namespace sql_logger_hpp
  {
    /**
     * @brief
     *         Returns the combined hashes of two objects.
     *
     * @param <T1>
     *         type of the first object
     *
     * @param <T1>
     *         type of the second object
     *
     * @param first
     *         first value to hash
     *
     * @param second
     *         second value to hash
     *
     * @returns
     *         the combined hashs of <code><var>first</var></code> and
     *         <code><var>second</var></code>
     *
     */
    template<typename T1, typename T2>
    std::size_t
    combined_hash(const T1& first, const T2& second)
    {
      // I'm somewhat surprised that this isn't the default behavior of
      // 'std::hash<std::pair<T1, T2>>' but apparently, it isn't.
      const std::hash<T1> hasher1 {};
      const std::hash<T2> hasher2 {};
      return hasher1(first) ^ hasher2(second);
    }
  }  // namespace sql_logger_hpp

  /**
   * @brief
   *         A `Logger` that produces simple SQL `INSERT` statements.
   *
   * See the file `schema.sql` that is distributed with the source code of this
   * project for the meaning of the field values.
   *
   * @param <T>
   *         coefficient type for linear systems
   *
   */
  template<typename T>
  class SQLLogger : public Logger<T>
  {
  public:

    /** @brief Coefficient type for linear systems. */
    using value_type = typename Logger<T>::value_type;

    /** @brief Index type for linear systems. */
    using index_type = typename Logger<T>::index_type;

    /** @brief Type for counting FLOPs. */
    using flops_type = typename Logger<T>::flops_type;

  private:

    std::size_t problemID_ {};
    std::string hostName_ {};
    unsigned recursions_;
    index_type blockSizeMin_;
    index_type blockSizeMax_;
    bool spd_;

  public:

    SQLLogger(const unsigned recursions,
              const index_type block_min,
              const index_type block_max,
              const bool spd) :
      recursions_ {recursions},
      blockSizeMin_ {block_min},
      blockSizeMax_ {block_max},
      spd_ {spd}
    {
#if HAVE_UNISTD_H
      char name[HOST_NAME_MAX + 1];
      if (gethostname(name, sizeof(name)) == 0)
        this->hostName_ = name;
#endif
    }

    // {@inheritDoc}
    virtual void
    logBefore(const HMatrix<T>& coeffs) final override
    {
      this->problemID_ = this->nextProblemID_();
      std::cout << "INSERT INTO `Problems` VALUES ("
                << SQLLogger::formatID_(this->problemID_) << ", "
                << SQLLogger::formatInteger_(coeffs.size()) << ", "
                << SQLLogger::formatBoolean_(this->spd_) << ", "
                << SQLLogger::formatInteger_(this->recursions_) << ", "
                << SQLLogger::formatInteger_(this->blockSizeMin_) << ", "
                << SQLLogger::formatInteger_(this->blockSizeMax_) << ", "
                << SQLLogger::formatInteger_(non_zero_count(coeffs)) << ");"
                << BSC_KLAMMLER_EOL;
    }

    // {@inheritDoc}
    virtual void
    logAfter() final override
    {
      // do nothing
    }

    virtual void
    logPreconditioner(const Preconditioner<value_type>&) final override
    {
    }

    virtual void
    logInit(const Preconditioner<value_type>& precond,
            const Timer::time_delta_type elapsed) final override
    {
      std::cout << "INSERT INTO `Decompositions` VALUES ("
                << SQLLogger::formatID_(this->problemID_) << ", "
                << SQLLogger::formatText_(precond.getShortName()) << ", "
                << SQLLogger::formatText_(this->hostName_) << ", "
                << SQLLogger::formatInteger_(BSC_KLAMMLER_PARALLEL) << ", "
                << SQLLogger::formatInteger_(precond.getNumCPUs()) << ", "
                << SQLLogger::formatInteger_(precond.getInitFLOPs()) << ", "
                << SQLLogger::formatReal_(elapsed) << ");"
                << BSC_KLAMMLER_EOL;
    }

    virtual void
    logSolve(const Preconditioner<value_type>& precond,
             const Timer::time_delta_type elapsed,
             const value_type relative_residual) final override
    {
      std::cout << "INSERT INTO `Solvings` VALUES ("
                << SQLLogger::formatID_(this->problemID_) << ", "
                << SQLLogger::formatText_(precond.getShortName()) << ", "
                << SQLLogger::formatText_(this->hostName_) << ", "
                << SQLLogger::formatInteger_(BSC_KLAMMLER_PARALLEL) << ", "
                << SQLLogger::formatInteger_(precond.getNumCPUs()) << ", "
                << SQLLogger::formatInteger_(precond.getSolveFLOPs()) << ", "
                << SQLLogger::formatReal_(elapsed) << ", "
                << SQLLogger::formatReal_(relative_residual) << ");"
                << BSC_KLAMMLER_EOL;
    }

    virtual void
    logSkipping(const std::string&)
    {
      // do nothing
    }

  private:

    std::size_t
    nextProblemID_() const noexcept
    {
      const auto ticks = std::chrono::system_clock::now().time_since_epoch();
      const auto nanos = std::chrono::duration_cast<std::chrono::nanoseconds>(ticks);
      return sql_logger_hpp::combined_hash(this->hostName_, nanos.count());
    }

    std::string
    formatText_(const std::string& value) const
    {
      std::ostringstream oss {};
      if (value.empty())
        oss << "NULL";
      else if (value.find("'") != std::string::npos)
        throw std::invalid_argument {"apostrophe in SQL string value"};
      else
        oss << "'" << value << "'";
      return oss.str();
    }

    template<typename V>
    static std::string
    formatBoolean_(const V value)
    {
      static_assert(std::is_integral<V>::value, "not a boolean type");
      return value ? "1" : "0";
    }

    template<typename V>
    static std::string
    formatInteger_(const V value)
    {
      static_assert(std::is_integral<V>::value, "not an integer type");
      std::ostringstream oss {};
      oss << value;
      return oss.str();
    }

    template<typename V>
    static std::string
    formatReal_(const V value)
    {
      static_assert(std::is_floating_point<V>::value, "not a real type");
      constexpr auto precision = std::numeric_limits<V>::max_digits10;
      std::ostringstream oss {};
      oss << std::scientific << std::setprecision(precision) << value;
      return oss.str();
    }

    std::string
    formatID_(const std::size_t value) const
    {
      std::ostringstream oss {};
      oss << "'"
          << std::setw((std::numeric_limits<std::size_t>::digits + 3) / 4)
          << std::setfill('0')
          << std::hex
          << value
          << "'";
      return oss.str();
    }
  };  // class SQLLogger
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_SQL_LOGGER_HPP
