/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file common_test_preconditioner.hpp
 *
 * @brief
 *         Unit tests common for `Preconditioner` implementations.
 *
 * To use this file, set up a compilation unit as for a usual test module,
 * define a template like
 *
 *     template<typename T>
 *     using preconditioner_t = bsc_klammler::XYZPreconditioner<T>;
 *
 * to test the `XYZPreconditioner` and then `#include` this file.  Feel free to
 * add any other unit tests specific to the `XYZPreconditioner` after that.
 *
 * This file facilitates code sharing but it is not a header file.  It shoule
 * better be given a `.cpp` suffix but this would cazse Automake to assume
 * wrong things.
 *
 */

#include <array>                // std::array
#include <limits>               // std::numeric_limits
#include <vector>               // std::vector

#include "../algorithms.hpp"    // (linear algebra operations)
#include "../block_matrix.hpp"  // bsc_klammler::BlockMatrix
#include "../full_matrix.hpp"   // bsc_klammler::FullMatrix
#include "../full_vector.hpp"   // bsc_klammler::FullVector
#include "mock_h_matrix.hpp"    // bsc_klammler::test::MockHMatrix
#include "mock_h_vector.hpp"    // bsc_klammler::test::MockHVector
#include "util.hpp"             // MY_CHECK_VECTOR_CLOSE, ...

#ifndef TOL_DBL
#define TOL_DBL 1.0E-12
#endif

/** @brief Tolerance for considering two `double`s close (in percent). */
static constexpr double tol_double_percent {100 * TOL_DBL};


/**
 * @brief
 *         Elementary unit test for any `Preconditioner`.
 *
 * This test checks that:
 *
 * - tiny dense systems are solved correctly
 *
 */
BOOST_AUTO_TEST_CASE(general_full)
{
  typedef double value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using full_vector = bsc_klammler::FullVector<value_type>;
  struct solving_exercise
  {
    full_vector lhs;
    full_vector rhs;
  };
  struct problem
  {
    bool spd;
    full_matrix coeffs;
    std::vector<solving_exercise> exercises;
  };
  std::vector<problem> problems {
    problem {
      true,
      full_matrix {{8.0}},
      std::vector<solving_exercise> {
        solving_exercise {
          full_vector {{7.0}},
          full_vector {{56.0}},
        },
        solving_exercise {
          full_vector {{8.0}},
          full_vector {{64.0}},
        },
      },
    },
    problem {
      true,
      full_matrix {
        {1.0, 0.0, 0.0},
        {0.0, 2.0, 0.0},
        {0.0, 0.0, 3.0},
      },
      std::vector<solving_exercise> {
        solving_exercise {
          full_vector {{0.5, 0.5, 0.5}},
          full_vector {{0.5, 1.0, 1.5}},
        },
      },
    },
    problem {
      true,
      // Symmetric, positive definite random (5, 5) matrix with condition
      // number 11.848, generated with GNU Octave.
      full_matrix {
        {
          1.475054782111704,
          1.002072332422560,
          0.5042971730315465,
          0.8184405767582001,
          0.8330819569755075,
        },
        {
          1.002072332422560,
          2.164621794131814,
          1.499833050023571,
          1.179048818311109,
          1.181187492223209,
        },
        {
          0.5042971730315465,
          1.499833050023571,
          2.512462745859203,
          1.179490539567045,
          0.9867362199135012,
        },
        {
          0.8184405767582001,
          1.179048818311109,
          1.179490539567045,
          1.854752830172606,
          0.8786617292182324,
        },
        {
          0.8330819569755075,
          1.181187492223209,
          0.9867362199135012,
          0.8786617292182324,
          1.428668171845887,
        },
      },
      std::vector<solving_exercise> {
        solving_exercise {
          // Solution computed with GNU Octave.
          full_vector {{
            -0.3220284919078402,
            -0.02364461406390893,
             0.1798755919171163,
            -0.2724831956331226,
             0.8002804586781518,
          }},
          // Random vector generated with GNU Octave.
          full_vector {{
             0.03569537882600675,
             0.5199161469810790,
             0.7223440550747466,
             0.1185090534508559,
             0.7851995769976035,
          }},
        },
        solving_exercise {
          // Solution computed with GNU Octave.
          full_vector {{
            -0.5898218867919559,
             0.7172482412366772,
            -0.6873990659233671,
             0.4950387320689246,
             0.4082358981294520,
          }},
          // Random vector generated with GNU Octave.
          full_vector {{
             0.2473153639391905,
             0.9964211121560490,
             0.03795721065366735,
             0.8290315804206515,
             0.6957583463830843,
          }},
        },
        solving_exercise {
          // Solution computed with GNU Octave.
          full_vector {{
             0.9576765249583218,
             0.08228024500706066,
             0.4511760731350171,
            -0.1963414513688052,
            -0.7045419297183968,
          }},
          // Random vector generated with GNU Octave.
          full_vector {{
             0.9749679328229497,
             0.7507632751558057,
             0.8131433457002976,
             0.4297527701427573,
             0.1611288528453601,
          }},
        },
      },
    },
    problem {
      false,
      // Random (10, 10) matrix with condition number 90.833, generated with
      // GNU Octave.
      full_matrix {
        {
          0.7915622189414213,
          0.3383525698980906,
          0.2501876167219056,
          0.1110090686839106,
          0.5180874088996015,
          0.9172006630558834,
          0.4066781786087655,
          0.1209541320785232,
          0.7596875431396022,
          0.3174392651582412,
        },
        {
          0.8226431968335353,
          0.8747749373028254,
          0.6031275104035867,
          0.7687337494211183,
          0.4996571273859724,
          0.6367431466787480,
          0.9468434611997921,
          0.2072252267314980,
          0.4167812462526041,
          0.7118077894433775,
        },
        {
          0.8312759432557721,
          0.2743383732528109,
          0.5211249940118302,
          0.3120155468919227,
          0.9653052123582968,
          0.7579237076273699,
          0.8769048036244340,
          0.6063415663487119,
          0.7577804930965166,
          0.1000525810998783,
        },
        {
          0.2414902504002504,
          0.8273787569462671,
          0.926114767550726,
          0.4684277451769871,
          0.4150447497259460,
          0.4101355653862532,
          0.9344721554075347,
          0.1969527458688365,
          0.06674246832929635,
          0.5330745057299077,
        },
        {
          0.8391475951912536,
          0.4022974683414283,
          0.9181554653855498,
          0.3911239570504705,
          0.5089985396187467,
          0.4011208302425039,
          0.9268605518552089,
          0.9536477677570597,
          0.2998738984699825,
          0.3612886315413420,
        },
        {
          0.9897592310781166,
          0.3393094162905535,
          0.9825927035855724,
          0.3580086750805201,
          0.001124254590347085,
          0.6049101548428010,
          0.1188433031968976,
          0.2804096173230811,
          0.2972396768809685,
          0.5463379765343115,
        },
        {
          0.8050508593889806,
          0.9328272870223818,
          0.5453265086610247,
          0.7657451940442849,
          0.2734764894850318,
          0.2146323217240854,
          0.6936711699384589,
          0.9934108133949349,
          0.7757393547952836,
          0.6593338872764702,
        },
        {
          0.1515745038906526,
          0.0313332935573029,
          0.4600269202533880,
          0.6957462501091271,
          0.3142430323440550,
          0.5006972232162530,
          0.3620581435922581,
          0.7778308334067905,
          0.8471097489002626,
          0.09929577956425155,
        },
        {
          0.001097000816536786,
          0.9290932108749256,
          0.8679955679041212,
          0.4128420606517924,
          0.8734897875813079,
          0.03512300286321883,
          0.9588333010058483,
          0.7275212087937788,
          0.8540905445690439,
          0.5856824310192494,
        },
        {
          0.5432060028414891,
          0.3968223220385780,
          0.6536015867992678,
          0.7025991380433262,
          0.3435930740956861,
          0.02886310331795421,
          0.5926486175605952,
          0.4764334050932742,
          0.5893346595990423,
          0.4891551309603742,
        },
      },
      std::vector<solving_exercise> {
        solving_exercise {
          // Solution computed with GNU Octave.
          full_vector {{
            -0.5892877665829502,
            -0.2712824350621680,
            -0.8350262830671038,
             1.393466512506772,
             3.438115828573026,
             1.054766211038443,
            -2.910519082090422,
             1.645030612893326,
            -2.340259450473622,
             2.645230641004007,
          }},
          // Random vector generated with GNU Octave.
          full_vector {{
             0.2133667017368179,
             0.7275924060774560,
             0.4900711379043075,
             0.2303212705832944,
             0.4728266587039814,
             0.5099434808938569,
             0.5947496721141037,
             0.6020478549145685,
             0.5945586621260027,
             0.1908394616923327,
          }},
        },
        solving_exercise {
          // Solution computed with GNU Octave.
          full_vector {{
             0.8844625913080864,
             0.8979038677663199,
             0.2923451493692558,
            -0.9282940405982280,
            -1.537875752531668,
            -0.3702365059660587,
             1.153999802431111,
            -0.3711819108055861,
             1.667579422185511,
            -2.045395263553595,
          }},
          // Random vector generated with GNU Octave.
          full_vector {{
             0.8796331959102186,
             0.2264416985832718,
             0.9250346778347702,
             0.02849342264663184,
             0.5541788604026949,
             0.3205581866241415,
             0.8749447186884813,
             0.3208046341522103,
             0.4121620047665538,
             0.3258424045147791,
          }},
        },
        solving_exercise {
          // Solution computed with GNU Octave.
          full_vector {{
             0.7921872964054859,
            -0.1397719241678254,
             0.06780516889775534,
            -0.04731067532944067,
            -1.199309231330109,
            -0.6812458949357713,
             1.521141829858586,
            -1.036786525473258,
             1.634690495293535,
            -0.8588107604474813,
          }},
          // Random vector generated with GNU Octave.
          full_vector {{
             0.8077445692405328,
             0.7963573287713570,
             0.8247830231201106,
             0.2076846412420474,
             0.3696546025018904,
             0.2796416510595840,
             0.7609837663836835,
             0.4397825932179103,
             0.4362394435596423,
             0.9050271298116717,
          }},
        },
      },
    },
  };
  for (const auto& problem : problems)
    {
      preconditioner_t<value_type> precond {problem.coeffs.size()};
      if (precond.onlySPD() && !problem.spd)
        continue;
      precond.init(problem.coeffs);
      for (const auto& iter : problem.exercises)
        {
          const full_vector actual {precond.solve(iter.rhs)};
          MY_CHECK_VECTOR_CLOSE(iter.lhs, actual, tol_double_percent);
        }
    }
}


/**
 * @brief
 *         Elementary unit test for any `Preconditioner`.
 *
 * This test checks that:
 *
 * - tiny symmetric hirachical systems are solved correctly
 *
 */
BOOST_AUTO_TEST_CASE(general_h_symmetric)
{
  typedef double value_type;
  using block_matrix = bsc_klammler::BlockMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using full_vector = bsc_klammler::FullVector<value_type>;
  constexpr value_type eps {std::numeric_limits<value_type>::epsilon()};
  constexpr value_type tol {std::sqrt(eps)};
  const block_matrix coeffs = {
    full_matrix {
      {11.0,  1.2},
      { 1.2, 22.0},
    },
    full_matrix {
      {33.0,  3.4,  3.5},
      { 3.4, 44.0,  4.5},
      { 3.5,  4.5, 55.0},
    },
    full_matrix {
      {66.0,  6.7},
      { 6.7, 77.0},
    },
    full_matrix {
      { 1.6,  2.6},
      { 1.7,  2.7},
    },
    full_matrix {
      { 3.6,  4.6,  5.6},
      { 3.7,  4.7,  5.7},
    },
    full_matrix {
      { 1.6,  1.7},
      { 2.6,  2.7},
    },
    full_matrix {
      { 3.6,  3.7},
      { 4.6,  4.7},
      { 5.6,  5.7},
    },
  };
  preconditioner_t<value_type> precond {coeffs.size()};
  precond.init(coeffs);
  for (int i = 0; i < 10; ++i)
    {
      full_vector right {coeffs.size()};
      right.getRawVector().setRandom();
      const full_vector residual {coeffs * precond.solve(right) - right};
      BOOST_CHECK(norm(residual) / norm(right) < tol);
    }
}


/**
 * @brief
 *         Elementary unit test for any `Preconditioner`.
 *
 * This test checks that:
 *
 * - any `HMatrix` and `HVector` class is working
 *
 */
BOOST_AUTO_TEST_CASE(general_any)
{
  typedef double value_type;
  using index_type = bsc_klammler::HMatrix<value_type>::index_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using mock_matrix = bsc_klammler::test::MockHMatrix<value_type>;
  using mock_vector = bsc_klammler::test::MockHVector<value_type>;
  const full_matrix full_a {{3.0, 1.0}, {1.0, 4.0}};
  const full_vector full_x {{2.0, 5.0}};
  const full_vector full_b {{11.0, 22.0}};
  const mock_matrix coeffs {};
  const mock_vector right {};
  coeffs.mockRows([&full_a]()->index_type{ return full_a.rows(); });
  coeffs.mockCols([&full_a]()->index_type{ return full_a.cols(); });
  right.mockSize([&full_b]()->index_type{ return full_b.size(); });
  coeffs.mockGet([&full_a](const index_type i, const index_type j)->value_type{
      return full_a.get(i, j);
    });
  right.mockGet([&full_b](const index_type i)->value_type{
      return full_b.get(i);
    });
  coeffs.mockClone([&full_a](void){ return full_a.clone(); });
  right.mockClone([&full_b](void){ return full_b.clone(); });
  {
    preconditioner_t<value_type> precond {coeffs.size()};
    precond.init(coeffs);
    MY_CHECK_VECTOR_CLOSE(full_x, precond.solve(right), tol_double_percent);
    MY_CHECK_VECTOR_CLOSE(full_x, precond.solve(full_b), tol_double_percent);
  }
}
