/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file foo.hpp
 *
 * @brief
 *         Header file providing the fickle `Foo` class.
 *
 */

#ifndef BSC_KLAMMLER_TEST_FOO_HPP
#define BSC_KLAMMLER_TEST_FOO_HPP

#include <iostream>    // std::ostream
#include <stdexcept>   // std::runtime_error

#include <Eigen/Core>  // Eigen::NumTraits


namespace bsc_klammler
{
  namespace test
  {
    /**
     * @brief
     *         Dummy type than can be activated and deactivated.
     *
     * If deactivated, construction and assignment fails.
     *
     */
    class Foo
    {
    private:

      /** Arbitrary payload to make the type none-empty. */
      int value_;

    public:

      Foo() : value_ {0}
      {
        if (!Foo::isActive_())
          throw Foo::NotActive {"Foo::Foo(const Foo&)"};
      }

      Foo(const int value) : value_ {value}
      {
        if (!Foo::isActive_())
          throw Foo::NotActive {"Foo::Foo(const Foo&)"};
      }

      Foo(const Foo& other) : value_ {other.value_}
      {
        if (!Foo::isActive_())
          throw Foo::NotActive {"Foo::Foo(const Foo&)"};
      };

      Foo& operator=(const Foo& other)
      {
        if (!Foo::isActive_())
          throw Foo::NotActive {"Foo::operator=(const Foo&)"};
        this->value_ = other.value_;
        return *this;
      }

      Foo(Foo&& other) : value_ {other.value_}
      {
        if (!Foo::isActive_())
          throw Foo::NotActive {"Foo::Foo(Foo&&)"};
        other.value_ = ~other.value_;  // invalidate old value
      }

      Foo& operator=(Foo&& other)
      {
        if (!Foo::isActive_())
          throw Foo::NotActive {"Foo::operator=(Foo&&)"};
        this->value_ = other.value_;
        other.value_ = ~other.value_;  // invalidate old value
        return *this;
      }

      int
      value() const noexcept
      {
        return this->value_;
      }

      Foo
      operator~() const noexcept
      {
        return Foo {~this->value_};
      }

      friend bool
      operator==(const Foo& us, const Foo& them) noexcept
      {
        return us.value_ == them.value_;
      }

      friend bool
      operator!=(const Foo& us, const Foo& them) noexcept
      {
        return us.value_ != them.value_;
      }

      friend std::ostream&
      operator<<(std::ostream& ostr, const Foo& foo)
      {
        ostr << "Foo { value : " << foo.value_ << "}";
        return ostr;
      }

      /**
       * @brief
       *         Activates this type.
       *
       * After calling this function, construction and assignment, both by
       * copying and moving, of this type will succeed.  Calling this function
       * more than once has no (additional) effect.
       *
       */
      static void activate() noexcept
      {
        Foo::isActive_() = true;
      }

      /**
       * @brief
       *         Deactivates this type.
       *
       * After calling this function, construction and assignment, both by
       * copying and moving, of this type will fail unconditionally by throwing
       * a `Foo::NotActive` exception.  Calling this function more than once
       * has no (additional) effect.
       *
       */
      static void deactivate() noexcept
      {
        Foo::isActive_() = false;
      }

    private:

      /**
       * @brief
       *         Returns a reference to the static flag that toggles whether
       *         this type is active.
       *
       * @returns
       *         reference to the static flag
       *
       */
      static bool&
      isActive_() noexcept
      {
        static bool active {};
        return active;
      }

    public:

      /**
       * @brief
       *
       *         Exception that is thrown if a deactivated operation is
       *         attempted.
       *
       */
      class NotActive : std::runtime_error
      {
      public:
        NotActive(const char *const msg) : std::runtime_error {msg}
        {
        }
      };
    };
  }  // namespace test
}  // namespace bsc_klammler

// Tell Eigen about 'Foo'; see:
// http://eigen.tuxfamily.org/dox/TopicCustomizingEigen.html#user_defined_scalars

namespace Eigen
{
  template<>
  struct NumTraits<bsc_klammler::test::Foo>
  {
    typedef double Real;
    typedef double NonInteger;
    typedef double Nested;
    enum
      {
        IsComplex             = 0,
        IsInteger             = 1,
        ReadCost              = 8,
        AddCost               = 8,
        MulCost               = 8,
        IsSigned              = 1,
        RequireInitialization = 1,
      };
    static inline Real epsilon() { return 1.0; }
    static inline Real dummy_epsilon() { return 0.0; }
  };
}  // namespace Eigen

#endif  // ifndef BSC_KLAMMLER_TEST_FOO_HPP
