/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file mock_h_matrix.hpp
 *
 * @brief
 *         Header file providing the `MockHMatrix` template class.
 *
 */

#ifndef BSC_KLAMMLER_TEST_MOCK_H_MATRIX_HPP
#define BSC_KLAMMLER_TEST_MOCK_H_MATRIX_HPP

#include "../h_matrix.hpp"         // bsc_klammler::HMatrix
#include "../not_implemented.hpp"  // BSC_KLAMMLER_NOT_IMPLEMENTED

#include <functional>              // std::function


namespace bsc_klammler {
  namespace test {

    /**
     * @brief
     *         Mockup for testing `HMatrx`es.
     *
     * This class overrides all purely virtual function members in `HMatrix` to
     * throw exceptions unconditionally unless a different action is mocked.
     * In addition, it makes the static protected member functions from
     * `HMatrix` publicly accessible.
     *
     * @param <T>
     *         type of the matrix entries
     *
     */
    template<typename T>
    class MockHMatrix : public HMatrix<T>
    {

    public:

      /** @brief Type of the super-class. */
      using super_type = HMatrix<T>;

      /** @brief Type for elements. */
      using value_type = typename super_type::value_type;

      /** @brief Type for indices. */
      using index_type = typename super_type::index_type;

      /** @brief Convenince type for mocking the `rows` function. */
      using rows_mock_type = std::function<index_type(void)>;

      /** @brief Convenince type for mocking the `cols` function. */
      using cols_mock_type = std::function<index_type(void)>;

      /** @brief Convenince type for mocking the `get` function. */
      using get_mock_type = std::function<value_type(index_type, index_type)>;

      /** @brief Convenince type for mocking the `set` function. */
      using set_mock_type = std::function<void(value_type,index_type, index_type)>;

      /** @brief Convenince type for mocking the `clone` function. */
      using clone_mock_type = std::function<std::unique_ptr<super_type>(void)>;

    private:

      /** @brief Mock for the `rows` function, if any. */
      mutable rows_mock_type mockRows_ {};

      /** @brief Mock for the `cols` function, if any. */
      mutable cols_mock_type mockCols_ {};

      /** @brief Mock for the `get` function, if any. */
      mutable get_mock_type mockGet_ {};

      /** @brief Mock for the `set` function, if any. */
      mutable set_mock_type mockSet_ {};

      /** @brief Mock for the `clone` function, if any. */
      mutable clone_mock_type mockClone_ {};

    public:

      /**
       * @brief
       *         Default constructor: Creates a new mockup.
       *
       */
      MockHMatrix() noexcept
      {
        // do nothing
      }

      /**
       * @brief
       *         Mocks the `rows` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockRows(const rows_mock_type& mockup) const noexcept
      {
        this->mockRows_ = mockup;
      }

      /**
       * @brief
       *         Mocks the `cols` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockCols(const cols_mock_type& mockup) const noexcept
      {
        this->mockCols_ = mockup;
      }

      /**
       * @brief
       *         Mocks the `get` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockGet(const get_mock_type& mockup) const noexcept
      {
        this->mockGet_ = mockup;
      }

      /**
       * @brief
       *         Mocks the `set` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockSet(const set_mock_type& mockup) const noexcept
      {
        this->mockSet_ = mockup;
      }

      /**
       * @brief
       *         Mocks the `clone` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockClone(const clone_mock_type& mockup) const noexcept
      {
        this->mockClone_ = mockup;
      }

      // {@InheritDoc}
      virtual index_type
      rows() const noexcept final override
      {
        if (this->mockRows_) { return this->mockRows_(); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      // {@InheritDoc}
      virtual index_type
      cols() const noexcept final override
      {
        if (this->mockCols_) { return this->mockCols_(); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      // {@InheritDoc}
      virtual value_type
      get(const index_type row, const index_type col) const final override
      {
        if (this->mockGet_) { return this->mockGet_(row, col); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      // {@InheritDoc}
      virtual void
      set(const value_type val,
          const index_type row,
          const index_type col) final override
      {
        if (this->mockSet_) { return this->mockSet_(val, row, col); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      // {@InheritDoc}
      virtual std::unique_ptr<super_type>
      clone() const final override
      {
        if (this->mockClone_) { return this->mockClone_(); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      virtual void
      passOverRegion(const index_type global_north,
                     const index_type global_west,
                     const bool global_trans,
                     typename super_type::PassiveVisitor& visitor,
                     const index_type north,
                     const index_type west,
                     const index_type south,
                     const index_type east) const
      {
        visitor.visitRegion(global_north, global_west, global_trans, *this,
                            north, west, south, east);
      }

      virtual void
      passOverRegion(const index_type global_north,
                     const index_type global_west,
                     const bool global_trans,
                     typename super_type::ActiveVisitor& visitor,
                     const index_type north,
                     const index_type west,
                     const index_type south,
                     const index_type east)
      {
        visitor.visitRegion(global_north, global_west, global_trans, *this,
                            north, west, south, east);
      }

      // {@InheritDoc}
      using super_type::throwMismatchedDimensions;

      // {@InheritDoc}
      using super_type::checkIndex;

      // {@InheritDoc}
      using super_type::checkRegion;

      // {@InheritDoc}
      using super_type::ctorCheckSize;
    };
  }  // namespace test
}  // namespace bsc_klammler

#endif  // #ifndef BSC_KLAMMLER_TEST_MOCK_H_MATRIX_HPP
