/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file mock_h_vector.hpp
 *
 * @brief
 *         Header file providing the `MockHVector` template class.
 *
 */

#ifndef BSC_KLAMMLER_TEST_MOCK_H_VECTOR_HPP
#define BSC_KLAMMLER_TEST_MOCK_H_VECTOR_HPP

#include "../not_implemented.hpp"  // BSC_KLAMMLER_NOT_IMPLEMENTED
#include "../h_vector.hpp"         // bsc_klammler::HVector

#include <functional>              // std::function


namespace bsc_klammler {
  namespace test {

    /**
     * @brief
     *         Mockup for testing `HMatrx`es.
     *
     * This class overrides all purely virtual function members in `HVector` to
     * throw exceptions unconditionally unless a different action is mocked.
     * In addition, it makes the static protected member functions from
     * `HVector` publicly accessible.
     *
     * @param <T>
     *         type of the vector entries
     *
     */
    template<typename T>
    class MockHVector : public HVector<T>
    {
    public:

      /** @brief Type of the super-class. */
      using super_type = HVector<T>;

      /** @brief Type for elements. */
      using value_type = typename super_type::value_type;

      /** @brief Type for indices. */
      using index_type = typename super_type::index_type;

      /** @brief Convenince type for mocking the `size` function. */
      using size_mock_type = std::function<index_type(void)>;

      /** @brief Convenince type for mocking the `get` function. */
      using get_mock_type = std::function<value_type(index_type)>;

      /** @brief Convenince type for mocking the `set` function. */
      using set_mock_type = std::function<void(value_type, index_type)>;

      /** @brief Convenince type for mocking the `clone` function. */
      using clone_mock_type = std::function<std::unique_ptr<super_type>(void)>;

    private:

      /** @brief Mock for the `size` function, if any. */
      mutable size_mock_type mockSize_ {};

      /** @brief Mock for the `get` function, if any. */
      mutable get_mock_type mockGet_ {};

      /** @brief Mock for the `set` function, if any. */
      mutable set_mock_type mockSet_ {};

      /** @brief Mock for the `clone` function, if any. */
      mutable clone_mock_type mockClone_ {};

    public:

      /**
       * @brief
       *         Default constructor: Creates a new mockup.
       *
       */
      MockHVector() noexcept
      {
        // do nothing
      }

      /**
       * @brief
       *         Mocks the `size` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockSize(const size_mock_type& mockup) const noexcept
      {
        this->mockSize_ = mockup;
      }

      /**
       * @brief
       *         Mocks the `get` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockGet(const get_mock_type& mockup) const noexcept
      {
        this->mockGet_ = mockup;
      }

      /**
       * @brief
       *         Mocks the `set` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockSet(const set_mock_type& mockup) const noexcept
      {
        this->mockSet_ = mockup;
      }

      /**
       * @brief
       *         Mocks the `clone` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockClone(const clone_mock_type& mockup) const noexcept
      {
        this->mockClone_ = mockup;
      }

      // {@InheritDoc}
      virtual index_type
      size() const noexcept final override
      {
        if (this->mockSize_) { return this->mockSize_(); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      // {@InheritDoc}
      virtual value_type
      get(const index_type idx) const final override
      {
        if (this->mockGet_) { return this->mockGet_(idx); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      // {@InheritDoc}
      virtual void
      set(const value_type val, const index_type idx) final override
      {
        if (this->mockSet_) { return this->mockSet_(val, idx); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      // {@InheritDoc}
      virtual std::unique_ptr<super_type>
      clone() const final override
      {
        if (this->mockClone_) { return this->mockClone_(); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      // {@InheritDoc}
      using super_type::throwMismatchedDimensions;

      // {@InheritDoc}
      using super_type::checkIndex;

      // {@InheritDoc}
      using super_type::ctorCheckSize;
    };
  }  // namespace test
}  // namespace bsc_klammler

#endif  // #ifndef BSC_KLAMMLER_TEST_MOCK_H_VECTOR_HPP
