/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file mock_h_matrix.hpp
 *
 * @brief
 *         Header file providing the `MockPreconditioner` template class.
 *
 */

#ifndef BSC_KLAMMLER_MOCK_PRECONDITIONER_HPP
#define BSC_KLAMMLER_MOCK_PRECONDITIONER_HPP

#include "../not_implemented.hpp"  // BSC_KLAMMLER_NOT_IMPLEMENTED
#include "../preconditioner.hpp"   // bsc_klammler::Preconditioner

#include <functional>              // std::function


namespace bsc_klammler {
  namespace test {

    template<typename T>
    class MockPreconditioner : public Preconditioner<T>
    {
    public:

      /** @brief Type of the super-class. */
      using super_type = Preconditioner<T>;

      /** @brief Type for elements. */
      using value_type = typename super_type::value_type;

      /** @brief Type for indices. */
      using index_type = typename super_type::index_type;

      /** @brief Convenince type for mocking the `getName` function. */
      using name_mock_type = std::function<std::string(std::vector<std::string>&)>;

      /** @brief Convenince type for mocking the `isRecursive` function. */
      using recursive_mock_type = std::function<bool(void)>;

      /** @brief Convenince type for mocking the `onlySPD` function. */
      using spd_mock_type = std::function<bool(void)>;

      /** @brief Convenince type for mocking the `initImpl` function. */
      using init_mock_type = std::function<void(const HMatrix<value_type>&)>;

      /** @brief Convenince type for mocking the `solveImpl` function. */
      using solve_mock_type = std::function<FullVector<value_type>(const HVector<value_type>&)>;

    private:

      /** @brief Mock for the `getName` function, if any. */
      mutable name_mock_type mockName_ {};

      /** @brief Mock for the `isRecursive` function, if any. */
      mutable recursive_mock_type mockRecursive_ {};

      /** @brief Mock for the `onlySPD` function, if any. */
      mutable spd_mock_type mockSpd_ {};

      /** @brief Mock for the `initImpl` function, if any. */
      mutable init_mock_type mockInit_ {};

      /** @brief Mock for the `solveImpl` function, if any. */
      mutable solve_mock_type mockSolve_ {};

  public:

      /**
       * @brief
       *         Default constructor: Creates a new mockup.
       *
       */
      MockPreconditioner(const index_type size) : Preconditioner<T> {size}
      {
        // do nothing
      }

      /**
       * @brief
       *         Mocks the `getName` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockName(const name_mock_type& mockup) const noexcept
      {
        this->mockName_ = mockup;
      }

      /**
       * @brief
       *         Mocks the `isRecursive` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockRecursive(const recursive_mock_type& mockup) const noexcept
      {
        this->mockRecursive_ = mockup;
      }

      /**
       * @brief
       *         Mocks the `onlySPD` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockSpd(const spd_mock_type& mockup) const noexcept
      {
        this->mockSpd_ = mockup;
      }

      /**
       * @brief
       *         Mocks the `initImpl` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockInit(const init_mock_type& mockup) const noexcept
      {
        this->mockInit_ = mockup;
      }

      /**
       * @brief
       *         Mocks the `solveImpl` function.
       *
       * A previous mock, if any, is replaced.
       *
       * @param mockup
       *         mock action to register
       *
       */
      void
      mockSolve(const solve_mock_type& mockup) const noexcept
      {
        this->mockSolve_ = mockup;
      }

      // {@InheritDoc}
      virtual std::string
      getName(std::vector<std::string>& attributes) const final override
      {
        if (this->mockName_) { return this->mockName_(attributes); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      // {@InheritDoc}
      virtual bool
      isRecursive() const noexcept final override
      {
        if (this->mockRecursive_) { return this->mockRecursive_(); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      // {@InheritDoc}
      virtual bool
      onlySPD() const noexcept final override
      {
        if (this->mockSpd_) { return this->mockSpd_(); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      // {@InheritDoc}
      virtual void
      initImpl(const HMatrix<value_type>& coefficients) final override
      {
        if (this->mockInit_) { return this->mockInit_(coefficients); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }

      // {@InheritDoc}
      virtual FullVector<value_type>
      solveImpl(const HVector<value_type>& right) const final override
      {
        if (this->mockSolve_) { return this->mockSolve_(right); }
        BSC_KLAMMLER_NOT_IMPLEMENTED;
      }
    };  // class MockPreconditioner
  }  // namespace test
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_MOCK_PRECONDITIONER_HPP
