/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_algorithms.cpp
 *
 * @brief
 *         Unit tests for the `algorithms.hpp` header file.
 *
 */

// Please don't use C++11 initialization syntax for random sizes as in
//
//     const index_type n {random_int<index_type>(14, 92)};
//
// in this file.  It turns out that some older versions of GCC which we still
// need to support screw up on it.  Sometimes.  But it is not feasible to keep
// track which line works and which doesn't so we rather not use it at all.
//
// Instead, please use traditional initialization like this:
//
//     const index_type = random_int<index_type>(14, 92);


#include "../global.hpp"                 // (project configuration)
#include "../algorithms.hpp"             // (algorithms to test)

#include <array>                         // std::array
#include <cmath>                         // std::sqrt
#include <limits>                        // std::numeric_limits
#include <stdexcept>                     // std::invalid_argument
#include <type_traits>                   // std::type_traits, std::enable_if
#include <utility>                       // std::pair

#include <boost/mpl/list.hpp>            // boost::mpl::list

#include "../arranged_matrix.hpp"        // bsc_klammler::ArrangedMatrix
#include "../full_matrix.hpp"            // bsc_klammler::FullMatrix
#include "../full_vector.hpp"            // bsc_klammler::FullVector
#include "../inefficient_algorithm.hpp"  // bsc_klammler::InefficientAlgorithm
#include "../matrix_view.hpp"            // bsc_klammler::ConstMatrixView
#include "../random.hpp"                 // bsc_klammler::random_real
#include "mock_h_matrix.hpp"             // bsc_klammler::test::MockHMatrix
#include "mock_h_vector.hpp"             // bsc_klammler::test::MockHVector
#include "util.hpp"                      // MY_CHECK_MATRIX_CLOSE, ...

#define BOOST_TEST_MODULE algorithms
#include "module.hpp"

#ifndef TOL_DBL
#define TOL_DBL 1.0E-12
#endif

#ifndef TOL_FLT
#define TOL_FLT 1.0E-6f
#endif

using bsc_klammler::random_int;
using bsc_klammler::random_real;


using mpl_plus_minus = boost::mpl::list<std::integral_constant<char, '+'>,
                                        std::integral_constant<char, '-'>>;

using mpl_reals = boost::mpl::list<float, double>;

namespace /* anonymous */
{

  /**
   * @brief
   *         ?
   *
   */
  template<typename T>
  struct tol;

  template<>
  struct tol<double>
  {
    /** @brief Tolerance for considering two `double`s close. */
    static constexpr double value {TOL_DBL};

    /** @brief Tolerance for considering two `double`s close (in percent). */
    static constexpr double percent() noexcept { return 100.0 * tol::value; }
  };

  template<>
  struct tol<float>
  {
    /** @brief Tolerance for considering two `float`s close. */
    static constexpr float value {TOL_FLT};

    /** @brief Tolerance for considering two `float`s close (in percent). */
    static constexpr double percent() noexcept { return 100.0 * tol::value; }
  };

  /**
   * @brief
   *         Guard object that will -- at destruction time -- restore the
   *         `InefficientAlgorithm` policy that was active at construction
   *         time.
   *
   */
  class PolicyGuard
  {
  private:
    using inefficient_algorithm = bsc_klammler::InefficientAlgorithm;
    using policy_type = inefficient_algorithm::Policies;

    const policy_type policy_;

  public:
    PolicyGuard(const policy_type policy) noexcept :
      policy_ {inefficient_algorithm::getPolicy()}
    {
      inefficient_algorithm::setPolicy(policy);
    }

    ~PolicyGuard() noexcept
    {
      inefficient_algorithm::setPolicy(this->policy_);
    }
  };

  /**
   * @brief
   *         Dummy data type that is immutable and has only two states: *zero*
   *         and *non-zero*.  The only supported operators are `==` and `!=`.
   *
   */
  // README: Do not add more operators!  Thus type is used to check that only
  //         '==' and '!=' are used to compare zero.
  class Bit
  {
  private:
    bool zero_;

  public:
    Bit() noexcept : zero_ {true}
    {
    }

    template<typename T>
    Bit(const T& predicate) noexcept : zero_ {predicate ? false : true}
    {
    }

    friend bool
    operator==(const Bit& first, const Bit& second) noexcept
    {
      return first.zero_ == second.zero_;
    }

    friend bool
    operator!=(const Bit& first, const Bit& second) noexcept
    {
      return first.zero_ != second.zero_;
    }
  };

  /**
   * @brief
   *         Helper function template to generate a random integer that is
   *         different from another one.
   *
   * @param <T>
   *         type of the integer
   *
   * @param min
   *         minimum value to generate (inclusive)
   *
   * @param max
   *         maximum value to generate (inclusive)
   *
   * @param taboo
   *         value not to generate
   *
   * @returns
   *         random value <code><var>rnd</var></code> such that
   *         <code><var>rnd</var> &gt;= <var>min</var> &amp;&amp;
   *         <var>rnd</var> &lt;= <var>max</var> &amp;&amp;
   *         <var>rnd</var></code> != <code><var>taboo</var></code>
   *
   */
  template<typename T>
  T
  other_random_int_(const T min, const T max, const T taboo)
  {
    T rnd;
    do { rnd = random_int<T>(min, max); } while (rnd == taboo);
    return rnd;
  }

  /**
   * @brief
   *         Helper template to generate a list of `HMatrix`es that can't be
   *         addes or subtracted.
   *
   * @param <M>
   *         `HMatrix` type to generate (*not* the `value_type`)
   *
   * @returns
   *         a `std::vector` of `std::pair`s of matrices of type `M` that
   *         cannot be added or subtracted due to dimension mismatch
   *
   */
  template<typename M, typename = typename std::enable_if<std::is_base_of<bsc_klammler::HMatrix<typename M::value_type>, M>::value>::type>
  std::vector<std::pair<M, M> >
  get_mismatched_matrices_()
  {
    using index_type = typename M::index_type;
    std::vector<std::pair<M, M>> matrices {};
    {
      // two arbitrary matrices that don't fit
      const index_type rows1 = random_int<index_type>(1, 32);
      const index_type cols1 = random_int<index_type>(1, 32);
      index_type rows2;
      index_type cols2;
      do
        {
          rows2 = random_int<index_type>(1, 32);
          cols2 = random_int<index_type>(1, 32);
        }
      while (rows1 == rows2 && cols1 == cols2);
      matrices.push_back(std::make_pair(M {rows1, cols1}, M {rows2, cols2}));
    }
    {
      // a non-square matrix and its transpose
      const index_type rows = random_int<index_type>(1, 32);
      index_type cols;
      do { cols = random_int<index_type>(1, 32); } while (rows == cols);
      matrices.push_back(std::make_pair(M {rows, cols}, M {cols, rows}));
    }
    return matrices;
  }
}  // namespace /* anonymous */


/**
 * @brief
 *         Unit test for the `*=` operator.
 *
 * This test checks that:
 *
 * - the scaling is applied correctly
 * - the result is exact
 * - the return value is a reference to the updated `HVector`
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_assign)
{
  typedef double value_type;
  using full_vector = bsc_klammler::FullVector<value_type>;
  {
    // empty vector
    const full_vector expected {};
    full_vector actual {};
    actual *= 12.0;
    MY_CHECK_VECTOR_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(&actual, &(actual *= 12.0));
  }
  {
    // zero vector
    const full_vector expected {25};
    full_vector actual {25};
    actual *= 13.0;
    MY_CHECK_VECTOR_EQUAL(expected, actual);
  }
  {
    // exact results
    full_vector actual {{0.0, 0.5, 1.0, 1.5}};
    const full_vector expected {{0.0 * 0.1, 0.5 * 0.1, 1.0 * 0.1, 1.5 * 0.1}};
    actual *= 0.1;
    MY_CHECK_VECTOR_EQUAL(expected, actual);
  }
  {
    // scaling with 0 and 1
    const full_vector before {{0.0, 0.5, 1.0, 1.5}};
    const full_vector zero {before.size()};
    full_vector actual0 {before};
    full_vector actual1 {before};
    actual0 *= 0.0;
    actual1 *= 1.0;
    MY_CHECK_VECTOR_EQUAL(zero, actual0);
    MY_CHECK_VECTOR_EQUAL(before, actual1);
  }
  {
    // random data
    // Test data generated with GNU Octave.
    constexpr value_type factor {0.9378907252393873};
    const full_vector before {{
        0.1405119048664128,
        0.5211624268952875,
        0.7511839291476352,
        0.8240084859329092,
        0.1052276514555485,
        0.1253811143971581,
        0.9741607350812191,
        0.02338344884940307,
    }};
    const full_vector expected {{
        0.1317848123599277,
        0.4887934065283404,
        0.7045284400964481,
        0.7728299164750257,
        0.09869203833888186,
        0.1175937843132732,
        0.9136563183250593,
        0.02193111979996476,
    }};
    full_vector actual1 {before};
    full_vector actual2 {before};
    actual1 *= factor;
    MY_CHECK_VECTOR_CLOSE(expected, actual1, tol<double>::percent());
    MY_CHECK_VECTOR_CLOSE(expected, actual2 *= factor, tol<double>::percent());
  }
}


/**
 * @brief
 *         Unit test for the `*=` operator.
 *
 * This test checks that:
 *
 * - the scaling is applied correctly
 * - the result is exact
 * - the return value is a reference to the updated `HVector`
 *
 */
BOOST_AUTO_TEST_CASE(operator_div_assign)
{
  typedef double value_type;
  using full_vector = bsc_klammler::FullVector<value_type>;
  {
    // empty vector
    const full_vector expected {};
    full_vector actual {};
    actual /= 12.0;
    MY_CHECK_VECTOR_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(&actual, &(actual /= 12.0));
  }
  {
    // zero vector
    const full_vector expected {25};
    full_vector actual {25};
    actual /= 13.0;
    MY_CHECK_VECTOR_EQUAL(expected, actual);
  }
  {
    // exact results
    full_vector actual {{0.0, 0.5, 1.0, 1.5}};
    const full_vector expected {{0.0 / 0.1, 0.5 / 0.1, 1.0 / 0.1, 1.5 / 0.1}};
    actual /= 0.1;
    MY_CHECK_VECTOR_EQUAL(expected, actual);
  }
  {
    // dividing by 1
    const full_vector before {{0.0, 0.5, 1.0, 1.5}};
    full_vector actual {before};
    actual /= 1.0;
    MY_CHECK_VECTOR_EQUAL(before, actual);
  }
  {
    // random data
    // Test data generated with GNU Octave.
    constexpr value_type factor {0.7394486667109449};
    const full_vector before {{
        0.6797569617608550,
        0.3406906560398762,
        0.5125423231448824,
        0.8636934705722334,
        0.6555828353322463,
        0.8728765835746278,
        0.08696832683893178,
        0.6645709235238364,
    }};
    const full_vector expected {{
        0.9192753903856536,
        0.4607360475139707,
        0.6931411823685637,
        1.168023568713603,
        0.8865832948868074,
        1.180442433491926,
        0.1176123925218033,
        0.8987384161226993,
    }};
    full_vector actual1 {before};
    full_vector actual2 {before};
    actual1 /= factor;
    MY_CHECK_VECTOR_CLOSE(expected, actual1, tol<double>::percent());
    MY_CHECK_VECTOR_CLOSE(expected, actual2 /= factor, tol<double>::percent());
  }
}


/**
 * @brief
 *         Unit test for the `+=` operator.
 *
 * This test checks that:
 *
 * - the adding is applied correctly
 * - the result is exact
 * - the return value is a reference to the updated vector
 * - a vector can be added to itself
 *
 */
BOOST_AUTO_TEST_CASE(operator_plus_assign)
{
  typedef double value_type;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using index_type = full_vector::index_type;
  {
    // empty vectors
    const full_vector update {};
    const full_vector expected {};
    full_vector actual {};
    actual += update;
    MY_CHECK_VECTOR_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(&actual, &(actual += update));
  }
  {
    // zero vectors
    const index_type size = random_int<index_type>(1, 128);
    const full_vector update {size};
    const full_vector expected {size};
    full_vector actual {size};
    actual += update;
    MY_CHECK_VECTOR_EQUAL(expected, actual);
  }
  {
    // exact results
    const full_vector before {{0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5}};
    const full_vector update {{0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7}};
    const full_vector expected {{
        0.0 + 0.0,
        0.5 + 0.1,
        1.0 + 0.2,
        1.5 + 0.3,
        2.0 + 0.4,
        2.5 + 0.5,
        3.0 + 0.6,
        3.5 + 0.7,
    }};
    full_vector actual1 {before};
    full_vector actual2 {before};
    actual1 += update;
    MY_CHECK_VECTOR_EQUAL(expected, actual1);
    MY_CHECK_VECTOR_EQUAL(expected, actual2 += update);
  }
  {
    // random
    // Test data generated with GNU Octave.
    const full_vector expected {{
        0.9520031801493716,
        0.4163331938197876,
        0.9619145012574248,
        1.211139811820217,
        0.8302426641712415,
        0.5256246990449754,
        0.8125856284119608,
    }};
    const full_vector update {{
        0.6027512365221558,
        0.03640569510159168,
        0.4737320964738114,
        0.3389990947142272,
        0.3943875595943225,
        0.008472839568807112,
        0.4390855119405524,
    }};
    full_vector actual {{
        0.3492519436272158,
        0.3799274987181959,
        0.4881824047836135,
        0.8721407171059893,
        0.4358551045769190,
        0.5171518594761684,
        0.3735001164714084,
    }};
    full_vector actual1 {actual};
    full_vector actual2 {actual};
    actual1 += update;
    MY_CHECK_VECTOR_CLOSE(expected, actual1, tol<double>::percent());
    MY_CHECK_VECTOR_CLOSE(expected, actual2 += update, tol<double>::percent());
  }
  {
    // self
    const full_vector expected {{0.0, 2.0, 4.0, 6.0, 8.0}};
    full_vector actual {{0.0, 1.0, 2.0, 3.0, 4.0}};
    actual += actual;
    MY_CHECK_VECTOR_EQUAL(expected, actual);
  }
}


/**
 * @brief
 *         Unit test for the `-=` operator.
 *
 * This test checks that:
 *
 * - the subtracting is applied correctly
 * - the result is exact
 * - the return value is a reference to the updated vector
 * - a vector can be subtracted from itself
 *
 */
BOOST_AUTO_TEST_CASE(operator_minus_assign)
{
  typedef double value_type;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using index_type = full_vector::index_type;
  {
    // empty vectors
    const full_vector update {};
    const full_vector expected {};
    full_vector actual {};
    actual -= update;
    MY_CHECK_VECTOR_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(&actual, &(actual -= update));
  }
  {
    // zero vectors
    const index_type size = random_int<index_type>(1, 128);
    const full_vector update {size};
    const full_vector expected {size};
    full_vector actual {size};
    actual -= update;
    MY_CHECK_VECTOR_EQUAL(expected, actual);
  }
  {
    // exact results
    const full_vector before {{0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5}};
    const full_vector update {{0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7}};
    const full_vector expected {{
        0.0 - 0.0,
        0.5 - 0.1,
        1.0 - 0.2,
        1.5 - 0.3,
        2.0 - 0.4,
        2.5 - 0.5,
        3.0 - 0.6,
        3.5 - 0.7,
    }};
    full_vector actual1 {before};
    full_vector actual2 {before};
    actual1 -= update;
    MY_CHECK_VECTOR_EQUAL(expected, actual1);
    MY_CHECK_VECTOR_EQUAL(expected, actual2 -= update);
  }
  {
    // random
    // Test data generated with GNU Octave.
    const full_vector expected {{
         0.4611709193480474,
         0.1026466598767539,
        -0.1156733897546993,
        -0.5372627744261385,
        -0.7583967226368733,
        -0.7246879177920473,
         0.4868186438138388,
    }};
    const full_vector update {{
         0.1486912134345201,
         0.7618971231329780,
         0.1958141030374570,
         0.8835765879908306,
         0.8919568242434394,
         0.8235727509377414,
         0.1101807706291301,
    }};
    full_vector actual {{
         0.6098621327825675,
         0.8645437830097319,
         0.08014071328275767,
         0.3463138135646921,
         0.1335601016065661,
         0.09888483314569411,
         0.5969994144429689,
    }};
    full_vector actual1 {actual};
    full_vector actual2 {actual};
    actual1 -= update;
    MY_CHECK_VECTOR_CLOSE(expected, actual1, tol<double>::percent());
    MY_CHECK_VECTOR_CLOSE(expected, actual2 -= update, tol<double>::percent());
  }
  {
    // self
    full_vector actual {{0.0, 1.0, 2.0, 3.0, 4.0}};
    const full_vector expected {actual.size()};
    actual -= actual;
    MY_CHECK_VECTOR_EQUAL(expected, actual);
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the `+=` and `-=` operator.
 *
 * This test checks that:
 *
 * - attempts to add or subtract vectors of incompatible dimensions are
 *   rejected
 * - the vector is not changed in this case
 *
 */
BOOST_AUTO_TEST_CASE(operator_plusminus_assign__invalid)
{
  typedef double value_type;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using index_type = full_vector::index_type;
  const std::function<void(full_vector&, const full_vector&)> operations[] = {
    [](full_vector& lhs, const full_vector& rhs){lhs += rhs;},
    [](full_vector& lhs, const full_vector& rhs){lhs -= rhs;},
  };
  const full_vector before {{0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0}};
  const index_type bad_update_sizes[] = {before.size() + 1, before.size() - 1};
  for (const auto operation : operations)
    {
      for (const auto& bad_size : bad_update_sizes)
        {
          full_vector actual {before};
          full_vector update {bad_size};
          for (index_type i = 0; i < update.size(); ++i)
            update.set(random_real<value_type>(), i);
          try
            {
              operation(actual, update);
              BOOST_ERROR("no exception thrown");
            }
          catch (const std::invalid_argument&) { /* good */ }
          MY_CHECK_VECTOR_EQUAL(before, actual);
        }
    }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the `*=` operator.
 *
 * This test checks that:
 *
 * - the scaling is applied correctly
 * - the result is exact
 * - the return value is a reference to the updated matrix
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_assign_m)
{
  typedef double value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  {
    // empty matrix
    const full_matrix expected {};
    full_matrix actual {};
    actual *= 12.0;
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(&actual, &(actual *= 12.0));
  }
  {
    // zero matrix
    const full_matrix expected {4, 5};
    full_matrix actual {4, 5};
    actual *= 13.0;
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    // exact results
    full_matrix actual {
      {1.1, 1.2, 1.3, 1.4},
      {2.1, 2.2, 2.3, 2.4},
      {3.1, 3.2, 3.3, 3.4},
    };
    const full_matrix expected {
      {1.1 * 0.1, 1.2 * 0.1, 1.3 * 0.1, 1.4 * 0.1},
      {2.1 * 0.1, 2.2 * 0.1, 2.3 * 0.1, 2.4 * 0.1},
      {3.1 * 0.1, 3.2 * 0.1, 3.3 * 0.1, 3.4 * 0.1},
    };
    actual *= 0.1;
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    // scaling with 0 and 1
    const full_matrix before {
      {1.1, 1.2, 1.3, 1.4},
      {2.1, 2.2, 2.3, 2.4},
      {3.1, 3.2, 3.3, 3.4},
    };
    const full_matrix zero {before.rows(), before.cols()};
    full_matrix actual0 {before};
    full_matrix actual1 {before};
    actual0 *= 0.0;
    actual1 *= 1.0;
    MY_CHECK_MATRIX_EQUAL(zero, actual0);
    MY_CHECK_MATRIX_EQUAL(before, actual1);
  }
  {
    // random data
    // Test data generated with GNU Octave.
    constexpr value_type factor {0.4998539773948992};
    const full_matrix before {
      {0.3397355641975351,  0.2959417616176837, 0.2892264224714829},
      {0.3373258822617787,  0.1682781792689167, 0.3607320401270976},
      {0.02388709843775676, 0.7765971606921062, 0.08607467654360905},
      {0.8339620523225885,  0.5258262028974707, 0.732062766634873},
    };
    const full_matrix expected {
      {0.1698181730266380,  0.1479276666218523,  0.1445709776400682},
      {0.1686136839267935,  0.08411451721633988, 0.1803133450313061},
      {0.01194006116253620, 0.3881851796055349,  0.04302476942330242},
      {0.4168592488498589,  0.2628363189367580,  0.3659244856051552},
    };
    full_matrix actual1 {before};
    full_matrix actual2 {before};
    actual1 *= factor;
    MY_CHECK_MATRIX_CLOSE(expected, actual1, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, actual2 *= factor, tol<double>::percent());
  }
}


/**
 * @brief
 *         Unit test for the `/=` operator.
 *
 * This test checks that:
 *
 * - the scaling is applied correctly
 * - <del>the result is exact</del>
 * - the return value is a reference to the updated matrix
 *
 */
BOOST_AUTO_TEST_CASE(operator_div_assign_m)
{
  typedef double value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  {
    // empty matrix
    const full_matrix expected {};
    full_matrix actual {};
    actual /= 12.0;
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(&actual, &(actual /= 12.0));
  }
  {
    // zero matrix
    const full_matrix expected {4, 5};
    full_matrix actual {4, 5};
    actual /= 13.0;
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    // exact results
    full_matrix actual {
      {1.1, 1.2, 1.3, 1.4},
      {2.1, 2.2, 2.3, 2.4},
      {3.1, 3.2, 3.3, 3.4},
    };
    const full_matrix expected {
      {1.1 / 0.1, 1.2 / 0.1, 1.3 / 0.1, 1.4 / 0.1},
      {2.1 / 0.1, 2.2 / 0.1, 2.3 / 0.1, 2.4 / 0.1},
      {3.1 / 0.1, 3.2 / 0.1, 3.3 / 0.1, 3.4 / 0.1},
    };
    actual /= 0.1;
    // It turns out that the following check no longer holds after we
    // introduced Eigen's algorithms.  I <del>suspect</del> <ins>have ben able
    // to show</ins> that Eigen implements 'M /= f' as 'M *= 1.0 / f' which
    // <del>might introduce</del> <ins>introduces</ins> the error.
    //MY_CHECK_MATRIX_EQUAL(expected, actual);
    MY_CHECK_MATRIX_CLOSE(expected, actual, tol<double>::percent());
  }
  {
    // dividing by 1
    const full_matrix before {
      {1.1, 1.2, 1.3, 1.4},
      {2.1, 2.2, 2.3, 2.4},
      {3.1, 3.2, 3.3, 3.4},
    };
    full_matrix actual {before};
    actual /= 1.0;
    MY_CHECK_MATRIX_EQUAL(before, actual);
  }
  {
    // random data
    // Test data generated with GNU Octave.
    constexpr value_type factor {0.5366288392947208};
    const full_matrix before {
      {0.7564942242836580, 0.8536163602205794,  0.5605941403549034},
      {0.3757224443064668, 0.3205419557907945,  0.3739718061596721},
      {0.1982558828332650, 0.6119583837085013,  0.6102204523412891},
      {0.7178932726821804, 0.07305763322972973, 0.6968313938629781},
    };
    const full_matrix expected {
      {1.409715931924012,  1.590701612948100,   1.044658988308716},
      {0.7001532843450425, 0.5973252503761736,  0.6968909957414418},
      {0.3694469404473829, 1.140375505186759,   1.137136895481220},
      {1.337783622709676,  0.1361418318958551,  1.298535119317866},
    };
    full_matrix actual1 {before};
    full_matrix actual2 {before};
    actual1 /= factor;
    MY_CHECK_MATRIX_CLOSE(expected, actual1, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, actual2 /= factor, tol<double>::percent());
  }
}


/**
 * @brief
 *         Unit test for the `+=` operator.
 *
 * This test checks that:
 *
 * - the adding is applied correctly
 * - the result is exact
 * - the return value is a reference to the updated matrix
 * - a matrix can be added to itself
 *
 */
BOOST_AUTO_TEST_CASE(operator_plus_assign_mm)
{
  typedef double value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using index_type = full_matrix::index_type;
  {
    // empty matrices
    const full_matrix update {};
    const full_matrix expected {};
    full_matrix actual {};
    actual += update;
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(&actual, &(actual += update));
  }
  {
    // zero matrices
    const index_type rows = random_int<index_type>(1, 32);
    const index_type cols = random_int<index_type>(1, 32);
    const full_matrix update {rows, cols};
    const full_matrix expected {rows, cols};
    full_matrix actual {rows, cols};
    actual += update;
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    // exact results
    const full_matrix before {
      {1.0, 2.0, 3.0},
      {4.0, 5.0, 6.0},
      {7.0, 8.0, 9.0},
    };
    const full_matrix update {
      {0.1, 0.2, 0.3},
      {0.4, 0.5, 0.6},
      {0.7, 0.8, 0.9},
    };
    const full_matrix expected {
      {1.0 + 0.1, 2.0 + 0.2, 3.0 + 0.3},
      {4.0 + 0.4, 5.0 + 0.5, 6.0 + 0.6},
      {7.0 + 0.7, 8.0 + 0.8, 9.0 + 0.9},
    };
    full_matrix actual1 {before};
    full_matrix actual2 {before};
    actual1 += update;
    MY_CHECK_MATRIX_EQUAL(expected, actual1);
    MY_CHECK_MATRIX_EQUAL(expected, actual2 += update);
  }
  {
    // random
    // Test data generated with GNU Octave.
    const full_matrix expected {
      {1.710006057554597,   0.9061378994738720,  1.213569170808539},
      {1.571900505877766,   0.9367484329078040,  1.522974066558556},
      {0.8290206952845508,  0.5327343044429286,  1.642807700915533},
      {0.9953755570609418,  1.731535836931267,   1.566717459614767},
      {0.1257118575156025,  0.5417833956465075,  1.428365925192419},
    };
    const full_matrix update {
      {0.8663021925859000,  0.6416818807733743,  0.7755034733306595},
      {0.9449511534281376,  0.07174614707479503, 0.6032668996750610},
      {0.3450825727339813,  0.1617432531005802,  0.8398919544465500},
      {0.04243646567256236, 0.8208983413761120,  0.5930314465482385},
      {0.07845626476359883, 0.2091952044170396,  0.7632218839364315},
    };
    full_matrix actual {
      {0.8437038649686965,  0.2644560187004978,  0.4380656974778794},
      {0.6269493524496281,  0.8650022858330090,  0.9197071668834949},
      {0.4839381225505695,  0.3709910513423484,  0.8029157464689829},
      {0.9529390913883795,  0.9106374955551549,  0.9736860130665288},
      {0.04725559275200368, 0.3325881912294679,  0.665144041255988},
    };
    full_matrix actual1 {actual};
    full_matrix actual2 {actual};
    actual1 += update;
    MY_CHECK_MATRIX_CLOSE(expected, actual1, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, actual2 += update, tol<double>::percent());
  }
  {
    // self
    const full_matrix expected {
      {11.0, 12.0, 13.0},
      {21.0, 22.0, 23.0},
    };
    full_matrix actual {
      { 5.5,  6.0,  6.5},
      {10.5, 11.0, 11.5},
    };
    actual += actual;
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
}


/**
 * @brief
 *         Unit test for the `-=` operator.
 *
 * This test checks that:
 *
 * - the subtracting is applied correctly
 * - the result is exact
 * - the return value is a reference to the updated matrix
 * - a matrix can be subtracted from itself
 *
 */
BOOST_AUTO_TEST_CASE(operator_minus_assign_mm)
{
  typedef double value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using index_type = full_matrix::index_type;
  {
    // empty matrices
    const full_matrix update {};
    const full_matrix expected {};
    full_matrix actual {};
    actual -= update;
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(&actual, &(actual -= update));
  }
  {
    // zero matrices
    const index_type rows = random_int<index_type>(1, 32);
    const index_type cols = random_int<index_type>(1, 32);
    const full_matrix update {rows, cols};
    const full_matrix expected {rows, cols};
    full_matrix actual {rows, cols};
    actual -= update;
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    // exact results
    const full_matrix before {
      {1.0, 2.0, 3.0},
      {4.0, 5.0, 6.0},
      {7.0, 8.0, 9.0},
    };
    const full_matrix update {
      {0.1, 0.2, 0.3},
      {0.4, 0.5, 0.6},
      {0.7, 0.8, 0.9},
    };
    const full_matrix expected {
      {1.0 - 0.1, 2.0 - 0.2, 3.0 - 0.3},
      {4.0 - 0.4, 5.0 - 0.5, 6.0 - 0.6},
      {7.0 - 0.7, 8.0 - 0.8, 9.0 - 0.9},
    };
    full_matrix actual1 {before};
    full_matrix actual2 {before};
    actual1 -= update;
    MY_CHECK_MATRIX_EQUAL(expected, actual1);
    MY_CHECK_MATRIX_EQUAL(expected, actual2 -= update);
  }
  {
    // random
    // Test data generated with GNU Octave.
    const full_matrix expected {
      {-0.4651792298856178,  -0.08193638480060429, -0.2662732587445132},
      {-0.2218196482338235,   0.4025064637266865,  -0.04058903145222992},
      {-0.006276316772615864, 0.6065634258611056,  -0.2560917064453105},
      { 0.4063595657286777,  -0.2591283245400803,  -0.2557432387622197},
      { 0.1110657270546720,   0.7664590747541394,  -0.2439677990367488},
    };
    const full_matrix update {
      { 0.8402999461776504,   0.1294126239121579,   0.7102250916812063},
      { 0.3728408276666391,   0.2964590176766546,   0.5789451567666076},
      { 0.07060754387942363,  0.04170772505693769,  0.7206316159153845},
      { 0.01811640551236842,  0.9912739496066957,   0.4021996555366283},
      { 0.4295292856380776,   0.01608571522146589,  0.5294427348834115},
    };
    full_matrix actual {
      { 0.3751207162920325,   0.04747623911155356,  0.4439518329366931},
      { 0.1510211794328156,   0.6989654814033410,   0.5383561253143777},
      { 0.06433122710680776,  0.6482711509180432,   0.4645399094700740},
      { 0.4244759712410461,   0.7321456250666154,   0.1464564167744086},
      { 0.5405950126927497,   0.7825447899756053,   0.2854749358466628},
    };
    full_matrix actual1 {actual};
    full_matrix actual2 {actual};
    actual1 -= update;
    MY_CHECK_MATRIX_CLOSE(expected, actual1, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, actual2 -= update, tol<double>::percent());
  }
  {
    // self
    full_matrix actual {
      {1.1, 1.2, 1.3, 1.4, 1.5},
      {2.1, 2.2, 2.3, 2.4, 2.5},
      {3.1, 3.2, 3.3, 3.4, 3.5},
    };
    const full_matrix expected {actual.rows(), actual.cols()};
    actual -= actual;
    MY_CHECK_MATRIX_CLOSE(expected, actual, 0.0);
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the `+=` and `-=` operator.
 *
 * This test checks that:
 *
 * - attempts to add or subtract matrices of incompatible dimensions are
 *   rejected
 * - the matrix is not changed in this case
 *
 */
BOOST_AUTO_TEST_CASE(operator_plusminus_assign_mm__invalid)
{
  typedef double value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using index_type = full_matrix::index_type;
  const std::function<void(full_matrix&, const full_matrix&)> operations[] = {
    [](full_matrix& lhs, const full_matrix& rhs){lhs += rhs;},
    [](full_matrix& lhs, const full_matrix& rhs){lhs -= rhs;},
  };
  const full_matrix before {
    {11.0, 12.0, 13.0, 14.0, 15.0},
    {21.0, 22.0, 23.0, 24.0, 25.0},
  };
  const std::pair<index_type, index_type> bad_update_sizes[] = {
    {before.rows(), before.cols() + 1},
    {before.rows(), before.cols() - 1},
    {before.rows() + 1, before.cols()},
    {before.rows() - 1, before.cols()},
  };
  for (const auto operation : operations)
    {
      for (const auto& bad_size : bad_update_sizes)
        {
          full_matrix actual {before};
          full_matrix update {bad_size.first, bad_size.second};
          for (index_type j = 0; j < update.cols(); ++j)
            for (index_type i = 0; i < update.rows(); ++i)
              update.set(random_real<value_type>(), i, j);
          try
            {
              operation(actual, update);
              BOOST_ERROR("no exception thrown");
            }
          catch (const std::invalid_argument&) { /* good */ }
          MY_CHECK_MATRIX_EQUAL(before, actual);
        }
    }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the operator `+` overloaded for two `HMatrix`es.
 *
 * This test checks that:
 *
 * - the result is correct
 * - (0, 0) matrices can be added
 *
 */
BOOST_AUTO_TEST_CASE(operator_plus_mm)
{
  using value_type = double;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using matrix_view_cv = bsc_klammler::ConstMatrixView<value_type>;
  {
    const full_matrix empty {};
    BOOST_CHECK_EQUAL(0, (empty + empty).size());
  }
  {
    const full_matrix matrix1 {{11.0, 12.0}, {21.0, 22.0}};
    const full_matrix matrix2 {{0.5, 0.6}, {0.7, 0.8}};
    const full_matrix expected {{11.5, 12.6}, {21.7, 22.8}};
    const full_matrix actual {matrix1 + matrix2};
    MY_CHECK_MATRIX_CLOSE(expected, actual, tol<double>::percent());
  }
  {
    // Random matrix, generated with GNU Octave.
    const full_matrix matrix1 {
      {0.07943371244121326, 0.8286897199923472,  0.4038393875317880},
      {0.5964088678020341,  0.8950278345361836,  0.4865693737890885},
      {0.7785358745116784,  0.4436007271756989,  0.9002337806304550},
      {0.5701099422872050,  0.08330009026496625, 0.9475011594574079},
    };
    // Random matrix, generated with GNU Octave.
    const full_matrix matrix2 {
      {0.4508024831256598,  0.7078124746008697,  0.3039276033609152},
      {0.2540885338894165,  0.5310967524411135,  0.3477116554933888},
      {0.2244859427637031,  0.7597901543701526,  0.3997012760438999},
      {0.9095812674928280,  0.2159184424183614,  0.3041654521051699},
    };
    // Sum of the above two matrices, computed with GNU Octave.
    const full_matrix expected {
      {0.5302361955668731,  1.536502194593217,   0.7077669908927032},
      {0.8504974016914506,  1.426124586977297,   0.8342810292824774},
      {1.003021817275382,   1.203390881545852,   1.299935056674355},
      {1.479691209780033,   0.2992185326833277,  1.251666611562578},
    };
    const matrix_view_cv view1 {&matrix1, 0, 0, matrix1.rows(), matrix1.cols()};
    const matrix_view_cv view2 {&matrix2, 0, 0, matrix2.rows(), matrix2.cols()};
    MY_CHECK_MATRIX_CLOSE(expected, matrix1 + matrix2, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, matrix1 + view2, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, view1 + matrix2, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, view1 + view2, tol<double>::percent());
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the operator `+` overloaded for two
 *         `HMatrix`es.
 *
 * This test checks that:
 *
 * - attempting to add matrices of different dimensions is rejected
 *
 */
BOOST_AUTO_TEST_CASE(operator_plus_mm__invalid)
{
  using value_type = double;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  const auto matrices = get_mismatched_matrices_<full_matrix>();
  for (const auto& iter : matrices)
    {
      try
        {
          const auto sum = iter.first + iter.second;
          BOOST_ERROR("no exception thrown");
        }
      catch (const std::invalid_argument&) { /* good */ }
    }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the operator `-` overloaded for two `HMatrix`es.
 *
 * This test checks that:
 *
 * - the result is correct
 * - (0, 0) matrices can be subtracted
 *
 */
BOOST_AUTO_TEST_CASE(operator_minus_mm)
{
  using value_type = double;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using matrix_view_cv = bsc_klammler::ConstMatrixView<value_type>;
  {
    const full_matrix empty {};
    BOOST_CHECK_EQUAL(0, (empty - empty).size());
  }
  {
    const full_matrix matrix1 {{11.0, 12.0}, {21.0, 22.0}};
    const full_matrix matrix2 {{0.5, 0.6}, {0.7, 0.8}};
    const full_matrix expected {{10.5, 11.4}, {20.3, 21.2}};
    const full_matrix actual {matrix1 - matrix2};
    MY_CHECK_MATRIX_CLOSE(expected, actual, tol<double>::percent());
  }
  {
    // Random matrix, generated with GNU Octave.
    const full_matrix matrix1 {
      {0.07943371244121326, 0.8286897199923472,  0.4038393875317880},
      {0.5964088678020341,  0.8950278345361836,  0.4865693737890885},
      {0.7785358745116784,  0.4436007271756989,  0.9002337806304550},
      {0.5701099422872050,  0.08330009026496625, 0.9475011594574079},
    };
    // Random matrix, generated with GNU Octave.
    const full_matrix matrix2 {
      {0.4508024831256598,  0.7078124746008697,  0.3039276033609152},
      {0.2540885338894165,  0.5310967524411135,  0.3477116554933888},
      {0.2244859427637031,  0.7597901543701526,  0.3997012760438999},
      {0.9095812674928280,  0.2159184424183614,  0.3041654521051699},
    };
    // Difference of the above two matrices, computed with GNU Octave.
    const full_matrix expected {
      {-0.3713687706844465,  0.1208772453914775, 0.09991178417087276},
      { 0.3423203339126176,  0.3639310820950701, 0.1388577182956997},
      { 0.5540499317479752, -0.3161894271944537, 0.5005325045865552},
      {-0.3394713252056230, -0.1326183521533951, 0.6433357073522379},
    };
    const matrix_view_cv view1 {&matrix1, 0, 0, matrix1.rows(), matrix1.cols()};
    const matrix_view_cv view2 {&matrix2, 0, 0, matrix2.rows(), matrix2.cols()};
    MY_CHECK_MATRIX_CLOSE(expected, matrix1 - matrix2, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, matrix1 - view2, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, view1 - matrix2, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, view1 - view2, tol<double>::percent());
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the operator `-` overloaded for two
 *         `HMatrix`es.
 *
 * This test checks that:
 *
 * - attempting to subtract matrices of different dimensions is rejected
 *
 */
BOOST_AUTO_TEST_CASE(operator_minus_mm__invalid)
{
  using value_type = double;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  const auto matrices = get_mismatched_matrices_<full_matrix>();
  for (const auto& iter : matrices)
    {
      try
        {
          const auto diff = iter.first - iter.second;
          BOOST_ERROR("no exception thrown");
        }
      catch (const std::invalid_argument&) { /* good */ }
    }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the operator `*` overloaded for two `HMatrix`es.
 *
 * This test checks that:
 *
 * - the result is correct
 * - (0, 0) matrices can be multiplied
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_mm)
{
  using value_type = double;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using matrix_view_cv = bsc_klammler::ConstMatrixView<value_type>;
  {
    const full_matrix empty {};
    BOOST_CHECK_EQUAL(0, (empty * empty).size());
  }
  {
    const full_matrix matrix1 {{11.0, 12.0}, {21.0, 22.0}};
    const full_matrix matrix2 {{0.5, 0.0}, {0.0, 0.5}};
    const full_matrix expected {{5.5, 6.0}, {10.5, 11.0}};
    const full_matrix actual {matrix1 * matrix2};
    MY_CHECK_MATRIX_CLOSE(expected, actual, tol<double>::percent());
  }
  {
    // Random matrix, generated with GNU Octave.
    const full_matrix matrix1 {
      {0.9444420539856139, 0.3557190220910920, 0.4554691796098840},
      {0.5243277244327273, 0.9871712704253783, 0.06242775310557817},
      {0.6162568177903549, 0.9376953070995310, 0.3693892842963457},
      {0.9151796227879109, 0.8411497683235566, 0.3201960112979731},
    };
    // Random matrix, generated with GNU Octave.
    const full_matrix matrix2 {
      {0.2268683000377977, 0.2650972850124373},
      {0.6491363259230477, 0.2904956860551989},
      {0.4206032706325999, 0.1431600512773268},
    };
    // Product of the above two matrices, computed with GNU Octave.
    const full_matrix expected {
      {0.6367459289493373, 0.4189088568365852},
      {0.7860193881799079, 0.4347040119957908},
      {0.9038675642363524, 0.4886462396944059},
      {0.8883216050071431, 0.5328012897203183},
    };
    const matrix_view_cv view1 {&matrix1, 0, 0, matrix1.rows(), matrix1.cols()};
    const matrix_view_cv view2 {&matrix2, 0, 0, matrix2.rows(), matrix2.cols()};
    MY_CHECK_MATRIX_CLOSE(expected, matrix1 * matrix2, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, matrix1 * view2, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, view1 * matrix2, tol<double>::percent());
    MY_CHECK_MATRIX_CLOSE(expected, view1 * view2, tol<double>::percent());
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the operator `*` overloaded for two
 *         `HMatrix`es.
 *
 * This test checks that:
 *
 * - attempting to multiply matrices of incompatible dimensions is rejected
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_mm__invalid)
{
  using value_type = double;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using index_type = full_matrix::index_type;
  const index_type rows1 = random_int<index_type>(1, 32);
  const index_type cols1 = random_int<index_type>(1, 32);
  const index_type rows2 = other_random_int_<index_type>(1, 32, cols1);
  const index_type cols2 = random_int<index_type>(1, 32);
  const full_matrix matrix1 {rows1, cols1};
  const full_matrix matrix2 {rows2, cols2};
  try
    {
      const auto prod = matrix1 * matrix2;
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the operator `*` overloaded for two
 *         `ArrangedMatrix`es.
 *
 * This test checks that for `HMatrix * ArrangedMatrix`,
 * `ArrangedMatrix * HMatrix` and `ArrangedMatrix * ArrangedMatrix`:
 *
 * - empty `ArrangedMatrix`es work
 * - shortcuts are taken if possible
 * - if no shortcut is possible, the result is still correct
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_mm__arranged)
{
  typedef int value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  {
    const full_matrix expected {3, 7};
    const full_matrix left {3, 3};
    const arranged_matrix right {3, 7};
    const arranged_matrix actual {left * right};
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(0, actual.tiles().size());
  }
  {
    const full_matrix expected {
      {373, 396,   0, 442, 465},
      {693, 736,   0, 822, 865},
    };
    const full_matrix left {{11, 12}, {21, 22}};
    arranged_matrix right {2, 5};
    right.addTile(full_matrix {{11, 12}, {21, 22}}, 0, 0);
    right.addTile(full_matrix {{14}, {24}}, 0, 3);
    right.addTile(full_matrix {{15}, {25}}, 0, 4);
    assert(right.isRowLinear());
    assert(!right.isColumnLinear());
    const arranged_matrix actual {left * right};
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(3, actual.tiles().size());
  }
  {
    const full_matrix expected {
      {373, 396,   0, 442, 465},
      {693, 736,   0, 822, 865},
    };
    const full_matrix left {{11, 12}, {21, 22}};
    arranged_matrix right {2, 5};
    right.addTile(full_matrix {{11, 12}, {21, 22}}, 0, 0);
    right.addTile(full_matrix {{14, 15}}, 0, 3);
    right.addTile(full_matrix {{24, 25}}, 1, 3);
    assert(!right.isRowLinear());
    assert(!right.isColumnLinear());
    const arranged_matrix actual {left * right};
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {7, 7};
    const arranged_matrix left {7, 3};
    const full_matrix right {3, 7};
    const arranged_matrix actual {left * right};
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(0, actual.tiles().size());
  }
  {
    const full_matrix expected {
      { 373,  396},
      { 693,  736},
      {   0,    0},
      {1333, 1416},
      {1653, 1756},
    };
    const full_matrix right {{11, 12}, {21, 22}};
    arranged_matrix left {5, 2};
    left.addTile(full_matrix {{11, 12}, {21, 22}}, 0, 0);
    left.addTile(full_matrix {{41, 42}}, 3, 0);
    left.addTile(full_matrix {{51, 52}}, 4, 0);
    assert(!left.isRowLinear());
    assert(left.isColumnLinear());
    const arranged_matrix actual {left * right};
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(3, actual.tiles().size());
  }
  {
    const full_matrix expected {
      { 373,  396},
      { 693,  736},
      {   0,    0},
      {1333, 1416},
      {1653, 1756},
    };
    const full_matrix right {{11, 12}, {21, 22}};
    arranged_matrix left {5, 2};
    left.addTile(full_matrix {{11, 12}, {21, 22}}, 0, 0);
    left.addTile(full_matrix {{41}, {51}}, 3, 0);
    left.addTile(full_matrix {{42}, {52}}, 3, 1);
    assert(!left.isRowLinear());
    assert(!left.isColumnLinear());
    const arranged_matrix actual {left * right};
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {3, 3};
    const arranged_matrix left {3, 7};
    const arranged_matrix right {7, 3};
    const arranged_matrix actual {left * right};
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK_EQUAL(0, actual.tiles().size());
  }
  {
    const full_matrix expected {
      { 373,  396,  419},
      { 693,  736,  779},
      {1013, 1076, 1139},
    };
    arranged_matrix left {3, 2};
    arranged_matrix right {2, 3};
    left.addTile(full_matrix {{11, 12}, {21, 22}}, 0, 0);
    left.addTile(full_matrix {{31, 32}}, 2, 0);
    right.addTile(full_matrix {{11, 12}, {21, 22}}, 0, 0);
    right.addTile(full_matrix {{13}, {23}}, 0, 2);
    assert(left.isColumnLinear());
    assert(right.isRowLinear());
    const arranged_matrix actual {left * right};
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {
      { 776,  812},
      {1406, 1472},
    };
    arranged_matrix left {2, 3};
    arranged_matrix right {3, 2};
    left.addTile(full_matrix {{11, 12}, {21, 22}}, 0, 0);
    left.addTile(full_matrix {{13}, {23}}, 0, 2);
    right.addTile(full_matrix {{11, 12}, {21, 22}}, 0, 0);
    right.addTile(full_matrix {{31, 32}}, 2, 0);
    assert(!left.isColumnLinear());
    assert(!right.isRowLinear());
    const arranged_matrix actual {left * right};
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the operator `*` overloaded for two
 *         `HMatrix`es.
 *
 * This test checks that for `HMatrix * ArrangedMatrix`,
 * `ArrangedMatrix * HMatrix` and `ArrangedMatrix * ArrangedMatrix`:
 *
 * - attempting to multiply matrices of incompatible dimensions is rejected
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_mm__arranged__invlaid)
{
  typedef int value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  try
    {
      const arranged_matrix p {full_matrix {1, 2} * arranged_matrix {3, 1}};
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
  try
    {
      const arranged_matrix p {arranged_matrix {3, 1} * full_matrix {2, 2}};
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
  try
    {
      const arranged_matrix p {
        arranged_matrix {14, 15} * arranged_matrix {14, 15}
      };
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the operator `*` overloaded for two `HVector`s.
 *
 * This test checks that:
 *
 * - the result is correct
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_vv)
{
  using value_type = double;
  using full_vector = bsc_klammler::FullVector<value_type>;
  {
    const full_vector vector1 {{2.0}};
    const full_vector vector2 {{0.5}};
    BOOST_CHECK_CLOSE(1.0, vector1 * vector2, tol<double>::percent());
  }
  {
    const full_vector vector1 {{1.2, 2.3, 3.4, 4.5}};
    const full_vector vector2 {{5.6, 6.7, 7.8, 8.9}};
    const value_type expected {1.2 * 5.6 + 2.3 * 6.7 + 3.4 * 7.8 + 4.5 * 8.9};
    const value_type actual {vector1 * vector2};
    BOOST_CHECK_CLOSE(expected, actual, tol<double>::percent());
  }
  {
    const full_vector vector1 {{
        0.8905628690923817,
        0.5165147554296649,
        0.8883561833510195,
        0.4928565328823926,
        0.7945207404908715,
        0.9865217004930557,
        0.6922526352146453,
        0.9422605099811301,
        0.03425942602285874,
        0.8957264547115156,
    }};
    const full_vector vector2 {{
        0.3130117721176491,
        0.6583639991876025,
        0.5506243921368635,
        0.9454495277046135,
        0.7051848377421176,
        0.5287465907482302,
        0.6846285997740119,
        0.2600720707854976,
        0.8722816789070551,
        0.1132148445937438,
    }};
    const value_type expected {3.506121902265332};
    const value_type actual {vector1 * vector2};
    BOOST_CHECK_CLOSE(expected, actual, tol<double>::percent());
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the operator `*` overloaded for two
 *         `HVector`s.
 *
 * This test checks that:
 *
 * - attempting to compute the dot product of two vectors of different
 *   dimensions is rejected
 * - attempting to compute the dot product of two empty vectors is rejected
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_vv__invalid)
{
  using value_type = double;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using index_type = full_vector::index_type;
  {
    const full_vector empty {};
    try
      {
        const auto dot = empty * empty;
        BOOST_ERROR("no exception thrown, dot == " << dot);
      }
    catch (const std::invalid_argument&) { /* good */ }
  }
  {
    const index_type size1 = random_int<index_type>(1, 32);
    const index_type size2 = other_random_int_<index_type>(1, 32, size1);
    const full_vector vector1 {size1};
    const full_vector vector2 {size2};
    try
      {
        const auto dot = vector1 * vector2;
        BOOST_ERROR("no exception thrown, dot == " << dot);
      }
    catch (const std::invalid_argument&) { /* good */ }
  }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the operator `*` overloaded for a `HMatrix` and a
 *         `HVector`.
 *
 * This test checks that:
 *
 * - the result is correct
 * - an empty matrix can be multiplied with an empty vector
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_mv)
{
  using value_type = double;
  using matrix_type = bsc_klammler::FullMatrix<value_type>;
  using vector_type = bsc_klammler::FullVector<value_type>;
  {
    const matrix_type matrix {};
    const vector_type vector {};
    BOOST_CHECK_EQUAL(0, (matrix * vector).size());
  }
  {
    const matrix_type matrix {
      {1.0, 2.0, 3.0},
      {4.0, 5.0, 6.0},
      {7.0, 8.0, 9.0},
    };
    const vector_type vector {{1.0, 2.0, 3.0}};
    const vector_type expected {{14.0, 32.0, 50.0}};
    const vector_type actual {matrix * vector};
    MY_CHECK_VECTOR_CLOSE(expected, actual, tol<double>::percent());
  }
  {
    // Random matrix, generated with GNU Octave.
    const matrix_type matrix {
      {0.2403699013437379, 0.07177094726176851, 0.6167800383192727},
      {0.5479979770035311, 0.4202169951253534,  0.5531232150106514},
      {0.2952550218961570, 0.5221371615937695,  0.9614480920819565},
      {0.5638013085929399, 0.3183481826742541,  0.9056899298929952},
      {0.1123633978221041, 0.7411830792179338,  0.8739439005986184},
    };
    // Random vector, generated with GNU Octave.
    const vector_type vector {{
        0.8334074538190513,
        0.8856315176396757,
        0.2925603839070509,
    }};
    // Product of the above matrix and vector, computed with GNU Octave.
    const vector_type expected = {{
        0.4443340851963903,
        0.9906849539747734,
        0.9897704857903260,
        1.016784390804168,
        1.005740951650361,
    }};
    const vector_type actual {matrix * vector};
    MY_CHECK_VECTOR_CLOSE(expected, actual, tol<double>::percent());
  }
}


/**
 * @brief
 *         Unit test that exercises the remaining parts of the operator `*`
 *         function overloaded for a `HMatrix` and a `HVector` that are left
 *         out by the general unit test.
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_mv__extra)
{
  typedef int value_type;
  using block_matrix = bsc_klammler::BlockMatrix<value_type>;
  using const_matrix_view = bsc_klammler::ConstMatrixView<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using index_type = h_matrix::index_type;
  using mock_matrix = bsc_klammler::test::MockHMatrix<value_type>;
  using mock_vector = bsc_klammler::test::MockHVector<value_type>;
  const full_matrix backbone {
    {11, 12, -1, -1, 51, 61},
    {21, 22, -1, -1, 52, 62},
    {-1, -1, 33, 34, 53, 63},
    {-1, -1, 43, 44, 54, 64},
    {15, 25, 35, 45, 55, 56},
    {16, 26, 36, 46, 65, 66},
  };
  const mock_matrix mock_backbone {};
  mock_backbone.mockRows([&backbone]()->index_type{ return backbone.rows(); });
  mock_backbone.mockCols([&backbone]()->index_type{ return backbone.cols(); });
  mock_backbone.mockGet([&backbone](index_type i, index_type j)->value_type{
      return backbone.get(i, j);
  });
  const full_vector vec {{1, 2, 3, 4, 5, 6}};
  const mock_vector mock_vec {};
  mock_vec.mockSize([&vec]()->index_type{ return vec.size(); });
  mock_vec.mockGet([&vec](index_type i)->value_type{ return vec.get(i); });
  const block_matrix matrix {
    std::unique_ptr<h_matrix> {new const_matrix_view {
      &backbone, 0, 0, 2, 2, false
    }},
    std::unique_ptr<h_matrix> {new const_matrix_view {
      &backbone, 2, 2, 2, 2, false
    }},
    std::unique_ptr<h_matrix> {new const_matrix_view {
      &mock_backbone, 4, 4, 2, 2, false
    }},
    std::unique_ptr<h_matrix> {new const_matrix_view {
      &mock_backbone, 4, 0, 2, 2, true
    }},
    std::unique_ptr<h_matrix> {new const_matrix_view {
      &mock_backbone, 4, 2, 2, 2, true
    }},
    std::unique_ptr<h_matrix> {new const_matrix_view {
      &backbone, 0, 4, 2, 2, true
    }},
    std::unique_ptr<h_matrix> {new const_matrix_view {
      &backbone, 2, 4, 2, 2, true
    }},
  };
  const full_vector expected {{206, 346, 626, 806, 1141, 1351}};
  MY_CHECK_VECTOR_EQUAL(expected, matrix * vec);
  MY_CHECK_VECTOR_EQUAL(expected, matrix * mock_vec);
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the operator `*` overloaded for a
 *         `HMatrix` and a `HVector`.
 *
 * This test checks that:
 *
 * - attempting to compute a product with mismatched dimensions is rejected
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_mv__invalid)
{
  using value_type = double;
  using matrix_type = bsc_klammler::FullMatrix<value_type>;
  using vector_type = bsc_klammler::FullVector<value_type>;
  using index_type = matrix_type::index_type;
  const index_type rows = random_int<index_type>(1, 32);
  const index_type cols = random_int<index_type>(1, 32);
  const index_type size = other_random_int_<index_type>(1, 32, cols);
  const matrix_type matrix {rows, cols};
  const vector_type vector {size};
  try
    {
      const auto prod = matrix * vector;
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the operator `*` overloaded for a `HVector` and a
 *         `HMatrix`.
 *
 * This test checks that:
 *
 * - the result is correct
 * - an empty vector can be multiplied with an empty matrix
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_vm)
{
  using value_type = double;
  using matrix_type = bsc_klammler::FullMatrix<value_type>;
  using vector_type = bsc_klammler::FullVector<value_type>;
  {
    const vector_type vector {};
    const matrix_type matrix {};
    BOOST_CHECK_EQUAL(0, (vector * matrix).size());
  }
  {
    const vector_type vector {{1.0, 2.0, 3.0}};
    const matrix_type matrix {
      {1.0, 2.0, 3.0},
      {4.0, 5.0, 6.0},
      {7.0, 8.0, 9.0},
    };
    const vector_type expected {{30.0, 36.0, 42.0}};
    const vector_type actual {vector * matrix};
    MY_CHECK_VECTOR_CLOSE(expected, actual, tol<double>::percent());
  }
  {
    // Random vector, generated with GNU Octave.
    const vector_type vector {{
        0.1635823465352594,
        0.5704725464779090,
        0.1546563724226350,
        0.5956026271251222,
        0.3146051789529300,
        0.6497965433127426,
        0.5379935474857609,
    }};
    // Random matrix, generated with GNU Octave.
    const matrix_type matrix {
      {0.6172542339957553,  0.0002992821751175878, 0.7358363246281153},
      {0.5472914119810783,  0.6536650030572194,    0.2851271904186527},
      {0.4981853309774003,  0.7156049525437944,    0.04529297109332546},
      {0.7168472108456131,  0.9845718004811624,    0.1604047347892455},
      {0.4817982335832118,  0.5541451106562925,    0.4124276184905714},
      {0.8849520339425724,  0.5288641454402112,    0.3816012885517961},
      {0.08450320907087054, 0.3934975016489369,    0.1927629112881663},
    };
    // Product of the above vector and matrix, computed with GNU Octave.
    const vector_type expected {{
        1.689267412945369,
        1.799723445257918,
        0.8669896604986947,
    }};
    const vector_type actual {vector * matrix};
    MY_CHECK_VECTOR_CLOSE(expected, actual, tol<double>::percent());
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the operator `*` overloaded for a
 *         `HVector` and a `HMatrix`.
 *
 * This test checks that:
 *
 * - attempting to compute a product with mismatched dimensions is rejected
 *
 */
BOOST_AUTO_TEST_CASE(operator_mul_vm__invalid)
{
  using value_type = double;
  using vector_type = bsc_klammler::FullVector<value_type>;
  using matrix_type = bsc_klammler::FullMatrix<value_type>;
  using index_type = matrix_type::index_type;
  const index_type rows = random_int<index_type>(1, 32);
  const index_type cols = random_int<index_type>(1, 32);
  const index_type size = other_random_int_<index_type>(1, 32, rows);
  const vector_type vector {size};
  const matrix_type matrix {rows, cols};
  try
    {
      const auto prod = vector * matrix;
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the operator `+` overloaded for two `HVector`s.
 *
 * This test checks that:
 *
 * - the result is correct
 * - empty vectors can be added
 *
 */
BOOST_AUTO_TEST_CASE(operator_plus_vv)
{
  using value_type = double;
  using vector_type = bsc_klammler::FullVector<value_type>;
  {
    const vector_type empty {};
    BOOST_CHECK_EQUAL(0, (empty + empty).size());
  }
  {
    const vector_type vector1 {{1.0, 2.0, 3.0, 4.0}};
    const vector_type vector2 {{5.0, 6.0, 7.0, 8.0}};
    const vector_type expected {{6.0, 8.0, 10.0, 12.0}};
    const vector_type actual {vector1 + vector2};
    MY_CHECK_VECTOR_CLOSE(expected, actual, tol<double>::percent());
  }
  {
    // Random vector, generated with GNU Octave.
    const vector_type vector1 {{
        0.6782908988782959,
        0.3982570939339531,
        0.9388866263828167,
        0.6772238775566067,
        0.2494187519829758,
        0.7642862056060350,
        0.4978998361089790,
        0.8377076662831661,
        0.1661900071538856,
        0.6964417682201772,
    }};
    // Random vector, generated with GNU Octave.
    const vector_type vector2 {{
        0.002638167374943201,
        0.9027842686339620,
        0.1596856823295389,
        0.3380506786953427,
        0.5199019975873190,
        0.2815494688573520,
        0.5430796242872542,
        0.3570758530117285,
        0.0358149374391236,
        0.3913622655474543,
    }};
    // Sum of the above two vectors, computed with GNU Octave.
    const vector_type expected {{
        0.6809290662532390,
        1.301041362567915,
        1.098572308712356,
        1.015274556251949,
        0.7693207495702947,
        1.045835674463387,
        1.040979460396233,
        1.194783519294895,
        0.2020049445930092,
        1.087804033767632,
    }};
    const vector_type actual {vector1 + vector2};
    MY_CHECK_VECTOR_CLOSE(expected, actual, tol<double>::percent());
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the operator `+` overloaded for two
 *         `HVector`s.
 *
 * This test checks that:
 *
 * - attempting to add two vectors of different dimensions is rejected
 *
 */
BOOST_AUTO_TEST_CASE(operator_plus_vv__invalid)
{
  using value_type = double;
  using vector_type = bsc_klammler::FullVector<value_type>;
  using index_type = vector_type::index_type;
  const index_type size1 = random_int<index_type>(0, 32);
  const index_type size2 = other_random_int_<index_type>(0, 32, size1);
  const vector_type vector1 {size1};
  const vector_type vector2 {size2};
  try
    {
      const auto sum = vector1 + vector2;
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the operator `-` overloaded for two `HVector`s.
 *
 * This test checks that:
 *
 * - the result is correct
 * - empty vectors can be subtracted
 *
 */
BOOST_AUTO_TEST_CASE(operator_minus_vv)
{
  using value_type = double;
  using vector_type = bsc_klammler::FullVector<value_type>;
  {
    const vector_type empty {};
    BOOST_CHECK_EQUAL(0, (empty + empty).size());
  }
  {
    const vector_type vector1 {{1.0, 3.0, 5.0, 0.5}};
    const vector_type vector2 {{2.0, 3.0, 4.0, -0.5}};
    const vector_type expected {{-1.0, 0.0, 1.0, 1.0}};
    const vector_type actual {vector1 - vector2};
    MY_CHECK_VECTOR_CLOSE(expected, actual, tol<double>::percent());
  }
  {
    // Random vector, generated with GNU Octave.
    const vector_type vector1 {{
        0.3846626782833129,
        0.6577148470146013,
        0.6595332924594182,
        0.7999210291546774,
        0.4124100365024284,
        0.7050567881616351,
        0.8039279316078710,
        0.1511078983284956,
        0.8022533455961871,
    }};
    // Random vector, generated with GNU Octave.
    const vector_type vector2 {{
        0.4979424692035047,
        0.6193334947814417,
        0.8059449375309673,
        0.2749964977769948,
        0.1262716593140454,
        0.5254634300718004,
        0.5024535150194639,
        0.2543751880470881,
        0.3050418996149224,
    }};
    // Difference of the above two vectors, computed with GNU Octave.
    const vector_type expected {{
        -0.1132797909201918,
         0.03838135223315953,
        -0.1464116450715491,
         0.5249245313776827,
         0.2861383771883830,
         0.1795933580898347,
         0.3014744165884071,
        -0.1032672897185924,
         0.4972114459812647,
    }};
    const vector_type actual {vector1 - vector2};
    MY_CHECK_VECTOR_CLOSE(expected, actual, tol<double>::percent());
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the operator `-` overloaded for two
 *         `HVector`s.
 *
 * This test checks that:
 *
 * - attempting to subtract two vectors of different dimensions is rejected
 *
 */
BOOST_AUTO_TEST_CASE(operator_minus_vv__invalid)
{
  using value_type = double;
  using vector_type = bsc_klammler::FullVector<value_type>;
  using index_type = vector_type::index_type;
  const index_type size1 = random_int<index_type>(0, 32);
  const index_type size2 = other_random_int_<index_type>(0, 32, size1);
  const vector_type vector1 {size1};
  const vector_type vector2 {size2};
  try
    {
      const auto diff = vector1 - vector2;
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the `invert` free function.
 *
 * This test checks that:
 *
 * - empty matrices are &ldquo;inverted&rdquo; correctly
 * - matrices are inverted correctly
 * - the correct inverse is computed for any `HMatrix` type (at least for
 *   `MatrixView`s),
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(invert, value_type, mpl_reals)
{
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  using mock_matrix = bsc_klammler::test::MockHMatrix<value_type>;
  using index_type = typename h_matrix::index_type;
  {
    const full_matrix matrix {};
    const full_matrix inverse {bsc_klammler::invert(matrix)};
    BOOST_CHECK_EQUAL(0, inverse.size());
  }
  {
    const full_matrix matrix {{4.0}};
    const full_matrix inverse {bsc_klammler::invert(matrix)};
    BOOST_CHECK_EQUAL(1, inverse.size());
    BOOST_CHECK_CLOSE(0.25, inverse.get(0, 0), tol<value_type>::percent());
  }
  {
    const full_matrix matrix {{1.0, 0, 0}, {0.0, 1.0, 0.0}, {0.0, 0.0, 1.0}};
    const full_matrix inverse {bsc_klammler::invert(matrix)};
    MY_CHECK_MATRIX_CLOSE(matrix, inverse, tol<value_type>::percent());
  }
  {
    // Random matrix with condition number 10.503, generated with GNU Octave.
    full_matrix backbone {
      {
        0.5434297325922491,
        0.4247743967214864,
        0.1534311360397563,
        0.5251386952454549,
        0.2611721983995108,
      },
      {
        0.5058471009838904,
        0.6690233032225377,
        0.5778808207388868,
        0.5707808705185264,
        0.2599630605680950,
      },
      {
        0.1922273745205643,
        0.5755590024465010,
        0.2978204392373229,
        0.9068674926983344,
        0.8329299193263074,
      },
      {
        0.7158982373884920,
        0.5368361096301441,
        0.7699082770357367,
        0.8243317688221784,
        0.8266542147492641,
      },
      {
        0.5399699875649241,
        0.1128289652958386,
        0.6321892342705426,
        0.8320520152503587,
        0.1558439305251980,
      },
    };
    const mock_matrix mock {};
    mock.mockRows([&backbone]()->index_type{ return backbone.rows(); });
    mock.mockCols([&backbone]()->index_type{ return backbone.cols(); });
    mock.mockGet([&backbone](index_type i, index_type j)->value_type{
        return backbone.get(i, j);
    });
    // Inverse computed with GNU Octave.
    const full_matrix expected {
      {
         2.112768667013363,
        -0.9813057588004133,
        -1.316962174791721,
         1.010699738031440,
        -0.2262211253811485,
      },
      {
         0.1096775397315655,
         1.941778755414256,
         0.3486764683920800,
        -0.8125690298025617,
        -0.9762646596848246,
      },
      {
        -2.272145014690966,
         1.180005914231845,
        -0.4343482653308656,
         0.7100474553804735,
         0.3945043181936764,
      },
      {
         0.3689873604928420,
        -0.2953503150954534,
         1.094917836117797,
        -1.380697029666307,
         1.346091086013829,
      },
      {
        -0.1527010645171857,
        -1.215661330309513,
         0.2267714184688434,
         1.577611562606934,
        -0.8798265902010788,
      },
    };
    const full_matrix actual {bsc_klammler::invert(mock)};
    MY_CHECK_MATRIX_CLOSE(expected, actual, tol<value_type>::percent());
  }
  {
    for (const bool spd : {false, true})
      {
        const full_matrix small {
          {4, 1},
          {static_cast<value_type>(spd ? 1 : 2), 5},
        };
        std::unique_ptr<h_matrix> large {new full_matrix {small}};
        for (int i = 1; i < (RUNNING_ON_VALGRIND ? 3 : 7); ++i)
          {
            arranged_matrix tmp {2 * large->rows(), 2 * large->cols()};
            std::unique_ptr<h_matrix> nw {large->clone()};
            std::unique_ptr<h_matrix> ne {large->clone()};
            std::unique_ptr<h_matrix> sw {large->clone()};
            std::unique_ptr<h_matrix> se {large->clone()};
            *nw *= small.get(0, 0);
            *ne *= small.get(0, 1);
            *sw *= small.get(1, 0);
            *se *= small.get(1, 1);
            tmp.addTile(std::move(nw), 0, 0);
            tmp.addTile(std::move(ne), 0, large->cols());
            tmp.addTile(std::move(sw), large->rows(), 0);
            tmp.addTile(std::move(se), large->rows(), large->cols());
            large.reset(new arranged_matrix {std::move(tmp)});
            assert(tmp.size() == 0);
          }
        full_matrix identity {large->rows(), large->cols()};
        for (index_type i = 0; i < identity.size(); ++i)
          identity.set(static_cast<value_type>(1), i, i);
        const full_matrix error {
          *large * bsc_klammler::invert(*large, spd) - identity
        };
        for (index_type j = 0; j < error.cols(); ++j)
          for (index_type i = 0; i < error.rows(); ++i)
            BOOST_CHECK(std::abs(error.get(i, j))
                        < 10.0 * tol<value_type>::value);
      }
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the `invert` free function.
 *
 * This test checks that:
 *
 * - <del>inverting an empty matrix is rejected</del>
 * - inverting non-square matrices is rejcted
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(invert__invalid, value_type, mpl_reals)
{
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using index_type = typename full_matrix::index_type;
  const std::pair<index_type, index_type> sizes[] = {
    {0, 1},
    {1, 0},
    {1, 2},
    {2, 1},
    {1, 1 + random_int<index_type>(1, 32)},
    {1 + random_int<index_type>(1, 32), 1},
  };
  for (const auto& iter : sizes)
    {
      const full_matrix rectangular {iter.first, iter.second};
      try
        {
          bsc_klammler::invert(rectangular);
          BOOST_ERROR("no exception thrown");
        }
      catch (const std::invalid_argument&) { /* good */ }
    }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the `norm` free function.
 *
 * This test checks that:
 *
 * - the correct norm is computed for any `HVector` type
 *
 */
BOOST_AUTO_TEST_CASE(norm)
{
  {
    const bsc_klammler::FullVector<int> vec {{1, 2, 2}};
    BOOST_CHECK_CLOSE(3.0, bsc_klammler::norm(vec), tol<double>::percent());
  }
  {
    // Random vector generated with GNU AWK.
    const bsc_klammler::FullVector<double> vec {
      {0.238, 0.291, 0.846, 0.152, 0.586, 0.193, 0.811, 0.174, 0.485, 0.152}
    };
    // 2-norm computed with GNU AWK.
    const double expected {1.4856365639011446};
    const double actual {bsc_klammler::norm(vec)};
    BOOST_CHECK_CLOSE(expected, actual, tol<double>::percent());
  }
  {
    // Normalized random vector generated with GNU AWK and normalized with GNU
    // Octave.
    const bsc_klammler::FullVector<double> vec {{
       0.1602006882322780,
       0.1958756314100542,
      -0.5694528665735595,
       0.1023130445853204,
       0.3944437113618273,
       0.1299106421379397,
      -0.5458939418335187,
       0.1171215115647747,
       0.3264593856834236,
       0.1023130445853204,
    }};
    BOOST_CHECK_CLOSE(1.0, bsc_klammler::norm(vec), tol<double>::percent());
  }
  {
    typedef double value_type;
    using index_type = bsc_klammler::HVector<value_type>::index_type;
    const index_type size = random_int<index_type>(1, 128);
    bsc_klammler::test::MockHVector<double> mock {};
    mock.mockSize([size](){ return size; });
    mock.mockGet([](const index_type)->value_type{ return 1.0; });
    BOOST_CHECK_CLOSE(std::sqrt(size),
                      bsc_klammler::norm(mock),
                      tol<double>::percent());
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the `norm` free function.
 *
 * This test checks that:
 *
 * - computing the norm of a 0 vector is rejected
 *
 */
BOOST_AUTO_TEST_CASE(norm__invalid)
{
  const bsc_klammler::FullVector<double> v {};
  try
    {
      bsc_klammler::norm(v);
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the `is_zero` free function.
 *
 * This test checks that:
 *
 * - the check is correct for empty, singleton, square and non-square matrices
 *
 */
BOOST_AUTO_TEST_CASE(is_zero)
{
  using value_type = Bit;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  {
    const full_matrix matrix {};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_zero(matrix));
  }
  {
    const full_matrix matrix {{0}};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_zero(matrix));
  }
  {
    const full_matrix matrix {{1}};
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_zero(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0},
      {0, 0, 0},
      {0, 0, 0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_zero(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0},
      {0, 0, 1},
      {0, 0, 0},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_zero(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0},
      {0, 0, 0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_zero(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0},
      {0, 0, 1},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_zero(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0},
      {0, 0},
      {0, 0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_zero(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0},
      {0, 0},
      {1, 0},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_zero(matrix));
  }
}


/**
 * @brief
 *         Unit test for the `is_lower_triangular` free function.
 *
 * This test checks that:
 *
 * - strict and non-strict checks for lower triangularity are correct for
 *   empty, singleton, square and non-square matrices
 *
 */
BOOST_AUTO_TEST_CASE(is_lower_triangular)
{
  using value_type = Bit;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  {
    const full_matrix matrix {};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix, false));
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix, true));
  }
  {
    const full_matrix matrix {{0}};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix, false));
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix, true));
  }
  {
    const full_matrix matrix {{1}};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_lower_triangular(matrix, true));
  }
  {
    const full_matrix matrix {
      {1, 0, 0, 0},
      {1, 1, 0, 0},
      {1, 1, 1, 0},
      {1, 1, 1, 1},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_lower_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0, 0},
      {1, 0, 0, 0},
      {1, 1, 0, 0},
      {1, 1, 1, 0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix, false));
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0, 0},
      {0, 0, 1, 0},
      {0, 0, 0, 0},
      {0, 0, 0, 0},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_lower_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_lower_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {1, 0, 0, 0, 0},
      {1, 1, 0, 0, 0},
      {1, 1, 1, 0, 0},
      {1, 1, 1, 1, 0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_lower_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0, 0, 0},
      {1, 0, 0, 0, 0},
      {1, 1, 0, 0, 0},
      {1, 1, 1, 0, 0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix, false));
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0, 0, 0},
      {0, 0, 0, 0, 1},
      {0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_lower_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_lower_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0, 0},
      {1, 0, 0, 0},
      {1, 1, 0, 0},
      {1, 1, 1, 0},
      {1, 1, 1, 1},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_lower_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0, 0},
      {0, 0, 0, 0},
      {1, 0, 0, 0},
      {1, 1, 0, 0},
      {1, 1, 1, 0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix, false));
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_lower_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0, 1},
      {0, 0, 0, 0},
      {0, 0, 0, 0},
      {0, 0, 0, 0},
      {0, 0, 0, 0},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_lower_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_lower_triangular(matrix));
  }
}


/**
 * @brief
 *         Unit test for the `is_upper_triangular` free function.
 *
 * This test checks that:
 *
 * - strict and non-strict checks for upper triangularity are correct for
 *   empty, singleton, square and non-square matrices
 *
 */
BOOST_AUTO_TEST_CASE(is_upper_triangular)
{
  using value_type = Bit;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  {
    const full_matrix matrix {};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix, false));
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix, true));
  }
  {
    const full_matrix matrix {{0}};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix, false));
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix, true));
  }
  {
    const full_matrix matrix {{1}};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_upper_triangular(matrix, true));
  }
  {
    const full_matrix matrix {
      {1, 1, 1, 1},
      {0, 1, 1, 1},
      {0, 0, 1, 1},
      {0, 0, 0, 1},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_upper_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 1, 1, 1},
      {0, 0, 1, 1},
      {0, 0, 0, 1},
      {0, 0, 0, 0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix, false));
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0, 0},
      {0, 0, 0, 0},
      {0, 0, 0, 0},
      {1, 0, 0, 0},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_upper_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_upper_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 1, 1, 1, 1},
      {0, 0, 1, 1, 1},
      {0, 0, 0, 1, 1},
      {0, 0, 0, 0, 1},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_upper_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 1, 1, 1},
      {0, 0, 0, 1, 1},
      {0, 0, 0, 0, 1},
      {0, 0, 0, 0, 0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix, false));
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {1, 0, 0, 0, 0},
      {0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0},
      {0, 0, 0, 0, 0},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_upper_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_upper_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {1, 1, 1, 1},
      {0, 1, 1, 1},
      {0, 0, 1, 1},
      {0, 0, 0, 1},
      {0, 0, 0, 0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_upper_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 1, 1, 1},
      {0, 0, 1, 1},
      {0, 0, 0, 1},
      {0, 0, 0, 0},
      {0, 0, 0, 0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix, false));
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_upper_triangular(matrix));
  }
  {
    const full_matrix matrix {
      {0, 0, 0, 0},
      {0, 0, 0, 0},
      {0, 0, 0, 0},
      {0, 0, 0, 0},
      {1, 0, 0, 0},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_upper_triangular(matrix, false));
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_upper_triangular(matrix));
  }
}


/**
 * @brief
 *         Unit test for the `is_symmetric` free function.
 *
 * This test checks that:
 *
 * - the result is correct
 *
 */
BOOST_AUTO_TEST_CASE(is_symmetric)
{
  using value_type = double;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  {
    const full_matrix matrix {};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_symmetric(matrix));
  }
  {
    const full_matrix matrix {{1.0}};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_symmetric(matrix));
  }
  {
    const full_matrix matrix {42, 42};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_symmetric(matrix));
  }
  {
    const full_matrix matrix {
      {11.0, 12.0, 13.0, 14.0},
      {12.0, 22.0, 23.0, 24.0},
      {13.0, 23.0, 33.0, 34.0},
      {14.0, 24.0, 34.0, 44.0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_symmetric(matrix));
  }
  {
    const full_matrix matrix {
      {11.0, 12.0, 13.0},
      {21.0, 22.0, 23.0},
      {13.0, 23.0, 33.0},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_symmetric(matrix));
  }
  {
    constexpr value_type epsilon {std::numeric_limits<value_type>::epsilon()};
    const full_matrix matrix = {
      {1.0, 1.0 + epsilon},
      {1.0, 1.0},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_symmetric(matrix));
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the `is_symmetric` free function.
 *
 * This test checks that:
 *
 * - non-square matrices are rejected
 *
 */
BOOST_AUTO_TEST_CASE(is_symmetric__invalid)
{
  using value_type = double;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using index_type = full_matrix::index_type;
  const index_type rows = random_int<index_type>(1, 32);
  const index_type cols = other_random_int_<index_type>(1, 32, rows);
  const full_matrix rectangular {rows, cols};
  try
  {
    bsc_klammler::is_symmetric(rectangular);
    BOOST_ERROR("no exception thrown");
  }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the `is_positive_definite` free function.
 *
 * This test checks that:
 *
 * - the result is correct
 * - any `HMatrix` is accepted
 *
 */
BOOST_AUTO_TEST_CASE(is_positive_definite)
{
  typedef double value_type;
  using index_type = bsc_klammler::HMatrix<value_type>::index_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using mock_matrix = bsc_klammler::test::MockHMatrix<value_type>;
  {
    const full_matrix matrix {};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_positive_definite(matrix));
  }
  {
    const full_matrix matrix {{1.0}};
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_positive_definite(matrix));
  }
  {
    const full_matrix matrix {42, 42};
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_positive_definite(matrix));
  }
  {
    const full_matrix matrix {
      {4.0, 1.2, 1.3, 1.4},
      {2.1, 5.0, 2.3, 2.4},
      {3.1, 3.2, 6.0, 3.4},
      {4.1, 4.2, 4.3, 7.0},
    };
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_positive_definite(matrix));
  }
  {
    const full_matrix matrix {
      {11.0, 12.0, 13.0, 14.0},
      {21.0, 22.0, 23.0, 24.0},
      {31.0, 32.0, 33.0, 34.0},
      {41.0, 42.0, 43.0, 44.0},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_positive_definite(matrix));
  }
  {
    constexpr value_type epsilon {std::numeric_limits<value_type>::epsilon()};
    const full_matrix matrix {
      {1.0, 1.0 + epsilon},
      {1.0, 1.0},
    };
    BOOST_CHECK_EQUAL(false, bsc_klammler::is_positive_definite(matrix));
  }
  {
    const full_matrix matrix {
      {1.0, 0.0},
      {0.0, 2.0},
    };
    const mock_matrix mock {};
    mock.mockRows([&matrix]()->index_type{ return matrix.rows(); });
    mock.mockCols([&matrix]()->index_type{ return matrix.cols(); });
    mock.mockGet([&matrix](index_type i, index_type j)->value_type{
        return matrix.get(i, j);
    });
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_positive_definite(matrix));
    BOOST_CHECK_EQUAL(true, bsc_klammler::is_positive_definite(mock));
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the `is_positive_definite` free
 *         function.
 *
 * This test checks that:
 *
 * - non-square matrices are rejected
 *
 */
BOOST_AUTO_TEST_CASE(is_positive_definite__invalid)
{
  using value_type = double;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using index_type = full_matrix::index_type;
  const index_type rows = random_int<index_type>(1, 32);
  const index_type cols = other_random_int_<index_type>(1, 32, rows);
  const full_matrix rectangular {rows, cols};
  try
    {
      bsc_klammler::is_positive_definite(rectangular);
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the `mul_div_to` helper function template overloaded
 *         for `HVector`s.
 *
 * This test checks that:
 *
 * - all kinds of `HVector`s are scaled correctly
 * - `FullVector`s are scaled efficiently
 * - integer division is applied correctly
 *
 */
BOOST_AUTO_TEST_CASE(mul_div_to_v)
{
  using bsc_klammler::InefficientAlgorithm;
  {
    typedef double value_type;
    using full_vector = bsc_klammler::FullVector<value_type>;
    const PolicyGuard pg {InefficientAlgorithm::Policies::THROW};
    const full_vector before {{0.0, 1.0, 2.0, 3.0, 4.0}};
    const full_vector expected_twice {{0.0, 2.0, 4.0, 6.0, 8.0}};
    const full_vector expected_half {{0.0, 0.5, 1.0, 1.5, 2.0}};
    full_vector actual_twice {before};
    full_vector actual_half {before};
    bsc_klammler::algorithms_hpp::mul_div_to<'*'>(actual_twice, 2.0);
    bsc_klammler::algorithms_hpp::mul_div_to<'/'>(actual_half, 2.0);
    MY_CHECK_VECTOR_EQUAL(expected_twice, actual_twice);
    MY_CHECK_VECTOR_EQUAL(expected_half, actual_half);
  }
  {
    typedef int value_type;
    using full_vector = bsc_klammler::FullVector<value_type>;
    const PolicyGuard pg {InefficientAlgorithm::Policies::THROW};
    const full_vector before {{0, 1, 2, 3, 4}};
    const full_vector expected_twice {{0, 2, 4, 6, 8}};
    const full_vector expected_half {{0, 0, 1, 1, 2}};
    full_vector actual_twice {before};
    full_vector actual_half {before};
    bsc_klammler::algorithms_hpp::mul_div_to<'*'>(actual_twice, 2);
    bsc_klammler::algorithms_hpp::mul_div_to<'/'>(actual_half, 2);
    MY_CHECK_VECTOR_EQUAL(expected_twice, actual_twice);
    MY_CHECK_VECTOR_EQUAL(expected_half, actual_half);
  }
  {
    typedef int value_type;
    using index_type = bsc_klammler::HVector<value_type>::index_type;
    using full_vector = bsc_klammler::FullVector<value_type>;
    using mock_vector = bsc_klammler::test::MockHVector<value_type>;
    const PolicyGuard pg {InefficientAlgorithm::Policies::IGNORE};
    const full_vector before {{0, 1, 2, 3, 4}};
    const full_vector expected_twice {{0, 2, 4, 6, 8}};
    const full_vector expected_half {{0, 0, 1, 1, 2}};
    full_vector actual_twice {before};
    full_vector actual_half {before};
    mock_vector mock_twice {};
    mock_vector mock_half {};
    mock_twice.mockSize([&actual_twice]()->index_type{
        return actual_twice.size();
    });
    mock_half.mockSize([&actual_half]()->index_type{
        return actual_half.size();
    });
    mock_twice.mockGet([&actual_twice](index_type i)->value_type{
        return actual_twice.get(i);
    });
    mock_half.mockGet([&actual_half](index_type i)->value_type{
        return actual_half.get(i);
    });
    mock_twice.mockSet([&actual_twice](value_type v, index_type i)->void{
        return actual_twice.set(v, i);
    });
    mock_half.mockSet([&actual_half](value_type v, index_type i)->void{
        return actual_half.set(v, i);
    });
    bsc_klammler::algorithms_hpp::mul_div_to<'*'>(mock_twice, 2);
    bsc_klammler::algorithms_hpp::mul_div_to<'/'>(mock_half, 2);
    MY_CHECK_VECTOR_EQUAL(expected_twice, actual_twice);
    MY_CHECK_VECTOR_EQUAL(expected_half, actual_half);
  }
}


/**
 * @brief
 *         Unit test for the `plus_minus_to` helper function template
 *         overloaded for `HVector`s.
 *
 * This test checks that:
 *
 * - all kinds of `HVector`s are added / subtracted correctly
 * - `FullVector`s are added / subtracted efficiently
 *
 */
BOOST_AUTO_TEST_CASE(plus_minus_to_v)
{
  using bsc_klammler::InefficientAlgorithm;
  typedef double value_type;
  using index_type = bsc_klammler::HVector<value_type>::index_type;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using mock_vector = bsc_klammler::test::MockHVector<value_type>;
  const full_vector before        {{ 0.0,  1.0,  2.0,  3.0,  4.0}};
  const full_vector delta         {{ 2.0,  1.5,  1.0,  0.5,  0.0}};
  const full_vector expected_more {{ 2.0,  2.5,  3.0,  3.5,  4.0}};
  const full_vector expected_less {{-2.0, -0.5,  1.0,  2.5,  4.0}};
  {
    // full +/- full
    const PolicyGuard pg {InefficientAlgorithm::Policies::THROW};
    full_vector actual_more {before};
    full_vector actual_less {before};
    bsc_klammler::algorithms_hpp::plus_minus_to<'+'>(actual_more, delta);
    bsc_klammler::algorithms_hpp::plus_minus_to<'-'>(actual_less, delta);
    MY_CHECK_VECTOR_EQUAL(expected_more, actual_more);
    MY_CHECK_VECTOR_EQUAL(expected_less, actual_less);
  }
  {
    const PolicyGuard pg {InefficientAlgorithm::Policies::IGNORE};
    full_vector actual {};
    mock_vector mock_delta {};
    mock_vector mock_actual {};
    mock_delta.mockSize([&delta]()->index_type{
        return delta.size();
    });
    mock_delta.mockGet([&delta](index_type i)->value_type{
        return delta.get(i);
    });
    mock_actual.mockSize([&actual]()->index_type{
        return actual.size();
    });
    mock_actual.mockGet([&actual](index_type i)->value_type{
        return actual.get(i);
    });
    mock_actual.mockSet([&actual](value_type v, index_type i)->void{
        return actual.set(v, i);
    });
    // full + ?
    actual = before;
    bsc_klammler::algorithms_hpp::plus_minus_to<'+'>(actual, mock_delta);
    MY_CHECK_VECTOR_EQUAL(expected_more, actual);
    // full - ?
    actual = before;
    bsc_klammler::algorithms_hpp::plus_minus_to<'-'>(actual, mock_delta);
    MY_CHECK_VECTOR_EQUAL(expected_less, actual);
    // ? + full
    actual = before;
    bsc_klammler::algorithms_hpp::plus_minus_to<'+'>(mock_actual, delta);
    MY_CHECK_VECTOR_EQUAL(expected_more, actual);
    // full - ?
    actual = before;
    bsc_klammler::algorithms_hpp::plus_minus_to<'-'>(mock_actual, delta);
    MY_CHECK_VECTOR_EQUAL(expected_less, actual);
    // ? + ?
    actual = before;
    bsc_klammler::algorithms_hpp::plus_minus_to<'+'>(mock_actual, mock_delta);
    MY_CHECK_VECTOR_EQUAL(expected_more, actual);
    // ? - ?
    actual = before;
    bsc_klammler::algorithms_hpp::plus_minus_to<'-'>(mock_actual, mock_delta);
    MY_CHECK_VECTOR_EQUAL(expected_less, actual);
  }
}


/**
 * @brief
 *         Unit test for the `mul_div_to` helper function template overloaded
 *         for `HMatrix`es.
 *
 * This test checks that:
 *
 * - all kinds of `HMatrix`es are scaled correctly
 * - `FullMatrix`s are scaled efficiently
 * - integer division is applied correctly
 *
 */
BOOST_AUTO_TEST_CASE(mul_div_to_m)
{
  using bsc_klammler::InefficientAlgorithm;
  typedef int value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  using matrix_view = bsc_klammler::MutableMatrixView<value_type>;
  using mock_matrix = bsc_klammler::test::MockHMatrix<value_type>;
  using index_type = h_matrix::index_type;
  const PolicyGuard pg {InefficientAlgorithm::Policies::IGNORE};
  const full_matrix expected_before {
    { 11, 12,  0,  0,  0,  0, 11, 12},
    { 21, 22,  0,  0,  0,  0, 21, 22},
    { 31, 32,  0, 11, 12, 13, 11, 12},
    { 41, 42,  0, 21, 22, 23, 21, 22},
  };
  const full_matrix expected_twice {
    { 22, 24,  0,  0,  0,  0, 22, 24},
    { 42, 44,  0,  0,  0,  0, 42, 44},
    { 62, 64,  0, 22, 24, 26, 22, 24},
    { 82, 84,  0, 42, 44, 46, 42, 44},
  };
  const full_matrix expected_half {
    {  5,  6,  0,  0,  0,  0,  5,  6},
    { 10, 11,  0,  0,  0,  0, 10, 11},
    { 15, 16,  0,  5,  6,  6,  5,  6},
    { 20, 21,  0, 10, 11, 11, 10, 11},
  };
  arranged_matrix actual {4, 8};
  full_matrix tile1 {
    {-1, -1, -1, -1},
    {-1, -1, 11, 12},
    {-1, -1, 21, 22},
    {-1, -1, 31, 32},
    {-1, -1, 41, 42},
    {-1, -1, -1, -1},
  };
  full_matrix tile2 {
    {-1, -1, -1, -1},
    {-1, -1, -1, -1},
    {-1, 11, 21, -1},
    {-1, 12, 22, -1},
    {-1, 13, 23, -1},
    {-1, -1, -1, -1},
  };
  full_matrix tile3 {
    {-1, -1, -1, -1},
    {-1, -1, -1, -1},
    {-1, -1, -1, -1},
    {-1, -1, 11, 12},
    {-1, -1, 21, 22},
  };
  full_matrix tile4 {
    {-1, -1, -1, -1},
    {-1, -1, 11, 21},
    {-1, -1, 12, 22},
  };
  mock_matrix tile3_mock {};
  mock_matrix tile4_mock {};
  tile3_mock.mockRows([&tile3]()->index_type{
      return tile3.rows();
  });
  tile3_mock.mockCols([&tile3]()->index_type{
      return tile3.cols();
  });
  tile3_mock.mockGet([&tile3](index_type i, index_type j)->value_type{
      return tile3.get(i, j);
  });
  tile3_mock.mockSet([&tile3](value_type v, index_type i, index_type j)->void{
      return tile3.set(v, i, j);
  });
  tile4_mock.mockRows([&tile4]()->index_type{
      return tile4.rows();
  });
  tile4_mock.mockCols([&tile4]()->index_type{
      return tile4.cols();
  });
  tile4_mock.mockGet([&tile4](index_type i, index_type j)->value_type{
      return tile4.get(i, j);
  });
  tile4_mock.mockSet([&tile4](value_type v, index_type i, index_type j)->void{
      return tile4.set(v, i, j);
  });
  std::unique_ptr<h_matrix> tile1_view_uptr {new matrix_view {
      &tile1, 1, 2, 4, 2, false
  }};
  std::unique_ptr<h_matrix> tile2_view_uptr {new matrix_view {
      &tile2, 1, 2, 2, 3, true
  }};
  std::unique_ptr<h_matrix> tile3_view_uptr {new matrix_view {
      &tile3_mock, 3, 2, 2, 2, false
  }};
  std::unique_ptr<h_matrix> tile4_view_uptr {new matrix_view {
      &tile4_mock, 2, 1, 2, 2, true
  }};
  actual.addTile(std::move(tile1_view_uptr), 0, 0);
  actual.addTile(std::move(tile2_view_uptr), 2, 3);
  {
    // Only 'FullMatrix'es so far.  A factor 1 does not do any damage.
    const PolicyGuard pg {InefficientAlgorithm::Policies::THROW};
    bsc_klammler::algorithms_hpp::mul_div_to<'*'>(actual, 1);
    bsc_klammler::algorithms_hpp::mul_div_to<'/'>(actual, 1);
  }
  actual.addTile(std::move(tile3_view_uptr), 0, 6);
  actual.addTile(std::move(tile4_view_uptr), 2, 6);
  MY_CHECK_MATRIX_EQUAL(expected_before, actual);
  bsc_klammler::algorithms_hpp::mul_div_to<'*'>(actual, 2);
  MY_CHECK_MATRIX_EQUAL(expected_twice, actual);
  bsc_klammler::algorithms_hpp::mul_div_to<'/'>(actual, 4);
  MY_CHECK_MATRIX_EQUAL(expected_half, actual);
}


/**
 * @brief
 *         Unit test for the `plus_minus_to` helper function template
 *         overloaded for `HMatrix`es.
 *
 * This test checks that:
 *
 * - *nothing*
 *
 */
BOOST_AUTO_TEST_CASE(plus_minus_to_m)
{
  // This is a placeholder.  I don't think we need this test right now since
  // the function is rather trivial and well covered by the other unit tests.
}


#if BSC_KLAMMLER_CHCEK_ARGS
/**
 * @brief
 *         Unit test for the `plus_minus_to` helper function template
 *         overloaded for `HMatrix`es.
 *
 * This test checks that:
 *
 * - attempts to destroy the structure of the augend fail loudly
 *
 */
BOOST_AUTO_TEST_CASE(plus_minus_to_m__invalid)
{
  typedef int value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using block_matrix = bsc_klammler::BlockMatrix<value_type>;
  block_matrix block {
    full_matrix {3, 3},
    full_matrix {4, 4},
    full_matrix {2, 2},
    full_matrix {2, 3},
    full_matrix {2, 4},
    full_matrix {3, 2},
    full_matrix {4, 2},
  };
  full_matrix full {block.rows(), block.cols()};
  bsc_klammler::algorithms_hpp::plus_minus_to<'+'>(full, block);
  try
    {
      bsc_klammler::algorithms_hpp::plus_minus_to<'+'>(block, full);
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHCEK_ARGS


/**
 * @brief
 *         Unit test that exercises the remaining parts of the `take_delta`
 *         function that are left out by the general unit test.
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(take_delta, operation, mpl_plus_minus)
{
  typedef int value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using block_matrix = bsc_klammler::BlockMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  using mock_matrix = bsc_klammler::test::MockHMatrix<value_type>;
  using const_matrix_view = bsc_klammler::ConstMatrixView<value_type>;
  using mutable_matrix_view = bsc_klammler::MutableMatrixView<value_type>;
  using operator_t = typename std::conditional<operation::value == '+',
                                               std::plus<value_type>,
                                               std::minus<value_type>>::type;
  using index_type = h_matrix::index_type;
  const operator_t op {};
  const auto mkrndmatrix = [](index_type r, index_type c)->full_matrix{
    full_matrix matrix {r, c};
    for (index_type j = 0; j < matrix.cols(); ++j)
      for (index_type i = 0; i < matrix.rows(); ++i)
        matrix.set(random_int<value_type>(), i, j);
    return matrix;
  };
  {
    // FullMatrix + HMatrix
    const index_type na = random_int<index_type>(0, 32);
    const index_type nb = random_int<index_type>(0, 32);
    const index_type nc = random_int<index_type>(0, 32);
    full_matrix augend {mkrndmatrix(na + nb + nc, na + nb + nc)};
    const block_matrix addend {
      mkrndmatrix(na, na),
      mkrndmatrix(nb, nb),
      mkrndmatrix(nc, nc),
      mkrndmatrix(nc, na),
      mkrndmatrix(nc, nb),
      mkrndmatrix(na, nc),
      mkrndmatrix(nb, nc),
    };
    full_matrix expected {augend.rows(), augend.cols()};
    for (index_type j = 0; j < expected.cols(); ++j)
      for (index_type i = 0; i < expected.rows(); ++i)
        expected.set(op(augend.get(i, j), addend.get(i, j)), i, j);
    switch (operation::value)
      {
      case '+': augend += addend; break;
      case '-': augend -= addend; break;
      default: assert(false);
      }
    MY_CHECK_MATRIX_EQUAL(expected, augend);
  }
  {
    // HMatrix + HMatrix
    const index_type na = random_int<index_type>(0, 32);
    const index_type nb = random_int<index_type>(0, 32);
    const index_type nc = random_int<index_type>(0, 32);
    block_matrix augend {
      mkrndmatrix(na, na),
      mkrndmatrix(nb, nb),
      mkrndmatrix(nc, nc),
      mkrndmatrix(nc, na),
      mkrndmatrix(nc, nb),
      mkrndmatrix(na, nc),
      mkrndmatrix(nb, nc),
    };
    const block_matrix addend {
      mkrndmatrix(na, na),
      mkrndmatrix(nb, nb),
      mkrndmatrix(nc, nc),
      mkrndmatrix(nc, na),
      mkrndmatrix(nc, nb),
      mkrndmatrix(na, nc),
      mkrndmatrix(nb, nc),
    };
    full_matrix expected {augend.rows(), augend.cols()};
    for (index_type j = 0; j < expected.cols(); ++j)
      for (index_type i = 0; i < expected.rows(); ++i)
        expected.set(op(augend.get(i, j), addend.get(i, j)), i, j);
    switch (operation::value)
      {
      case '+': augend += addend; break;
      case '-': augend -= addend; break;
      default: assert(false);
      }
    MY_CHECK_MATRIX_EQUAL(expected, augend);
  }
  {
    // ArrangedMatrix + HMatrix
    //
    // 00000     11006
    // 00000     11006
    // 44110  +  00227
    // 44110     00227
    // 33332     44553
    //
    const full_matrix view_a {mkrndmatrix(10, 10)};
    const full_matrix view_b {mkrndmatrix(10, 10)};
    const full_matrix view_c {mkrndmatrix(10, 10)};
    const full_matrix view_u {mkrndmatrix(10, 10)};
    const full_matrix view_v {mkrndmatrix(10, 10)};
    const full_matrix view_x {mkrndmatrix(10, 10)};
    const full_matrix view_y {mkrndmatrix(10, 10)};
    const block_matrix addend {
      std::unique_ptr<h_matrix> {new const_matrix_view {
          &view_a, 2, 4, 2, 2, false}
      },
      std::unique_ptr<h_matrix> {new const_matrix_view {
          &view_b, 3, 5, 2, 2, true}
      },
      std::unique_ptr<h_matrix> {new const_matrix_view {
          &view_c, 9, 9, 1, 1, true}
      },
      std::unique_ptr<h_matrix> {new const_matrix_view {
          &view_u, 3, 4, 1, 2, false}
      },
      std::unique_ptr<h_matrix> {new const_matrix_view {
          &view_v, 3, 4, 1, 2, true}
      },
      std::unique_ptr<h_matrix> {new const_matrix_view {
          &view_x, 1, 5, 2, 1, false}
      },
      std::unique_ptr<h_matrix> {new const_matrix_view {
          &view_y, 1, 5, 2, 1, true}
      },
    };
    arranged_matrix augend {addend.rows(), addend.cols()};
    augend.addTile(mkrndmatrix(2, 2), 2, 2);
    augend.addTile(mkrndmatrix(1, 1), 4, 4);
    augend.addTile(mkrndmatrix(1, 4), 4, 0);
    augend.addTile(mkrndmatrix(2, 2), 2, 0);
    full_matrix expected {augend.rows(), augend.cols()};
    for (index_type j = 0; j < expected.cols(); ++j)
      for (index_type i = 0; i < expected.rows(); ++i)
        expected.set(op(augend.get(i, j), addend.get(i, j)), i, j);
    switch (operation::value)
      {
      case '+': augend += addend; break;
      case '-': augend -= addend; break;
      default: assert(false);
      }
    MY_CHECK_MATRIX_EQUAL(expected, augend);
  }
  {
    // ArrangedMatrix + ArrangedMatrix (fast_take_delta)
    //
    // 00000000000     00000000000000m
    // 00000000000  +  fffFmmmMMMMffFF
    // 00000000000     fffFmmmMMMMffFF
    //
    arranged_matrix augend {3, 15};
    arranged_matrix addend {augend.rows(), augend.cols()};
    const full_matrix tile1 {mkrndmatrix(10, 10)};
    const full_matrix tile2 {mkrndmatrix(10, 10)};
    const full_matrix tile3 = {mkrndmatrix(10, 10)};
    const full_matrix tile4 = {mkrndmatrix(10, 10)};
    const mock_matrix mock5 {};
    const mock_matrix mock6 {};
    const mock_matrix mock7 {};
    mock5.mockRows([]()->index_type{ return 10; });
    mock5.mockCols([]()->index_type{ return 10; });
    mock5.mockGet([](index_type i, index_type j)->value_type{
        return 10 * i + j; });
    mock6.mockRows([]()->index_type{ return 10; });
    mock6.mockCols([]()->index_type{ return 10; });
    mock6.mockGet([](index_type i, index_type j)->value_type{
        return 10 * i + j; });
    mock7.mockRows([]()->index_type{ return 1; });
    mock7.mockCols([]()->index_type{ return 1; });
    mock7.mockGet([](index_type, index_type)->value_type{ return -1; });
    addend.addTile(std::unique_ptr<h_matrix> {
        new const_matrix_view {&tile1, 2, 1, 2, 3, false}}, 1, 0);
    addend.addTile(std::unique_ptr<h_matrix> {
        new const_matrix_view {&tile2, 3, 1, 2, 1, true}}, 1, 3);
    addend.addTile(std::unique_ptr<h_matrix> {
        new const_matrix_view {&tile3, 0, 0, 2, 2, false}}, 1, 11);
    addend.addTile(std::unique_ptr<h_matrix> {
        new const_matrix_view {&tile4, 0, 0, 2, 2, true}}, 1, 13);
    addend.addTile(std::unique_ptr<h_matrix> {
        new const_matrix_view {&mock5, 4, 5, 2, 3, false}}, 1, 4);
    addend.addTile(std::unique_ptr<h_matrix> {
        new const_matrix_view {&mock6, 3, 4, 2, 4, true}}, 1, 7);
    addend.addTile(std::unique_ptr<h_matrix> {
        new const_matrix_view {&mock7, 0, 0, 1, 1, false}}, 0, 14);
    full_matrix expected {augend.rows(), augend.cols()};
    for (index_type j = 0; j < expected.cols(); ++j)
      for (index_type i = 0; i < expected.rows(); ++i)
        expected.set(op(augend.get(i, j), addend.get(i, j)), i, j);
    switch (operation::value)
      {
      case '+': augend += addend; break;
      case '-': augend -= addend; break;
      default: assert(false);
      }
    MY_CHECK_MATRIX_EQUAL(expected, augend);
  }
  {
    // ArrangedMatrix + ArrangedMatrix
    // (dumb_take_delta: fbe.block_count != 1)
    //
    // ffFF0000000     fff0
    // ffFF0000000  +  fff0
    // ffFF0000000     0FFF
    // ffFF0000000     0FFF
    //
    arranged_matrix augend {4, 4};
    arranged_matrix addend {augend.rows(), augend.cols()};
    full_matrix aug_tile1 {mkrndmatrix(10, 10)};
    full_matrix aug_tile2 {mkrndmatrix(10, 10)};
    const full_matrix add_tile1 {mkrndmatrix(10, 10)};
    const full_matrix add_tile2 {mkrndmatrix(10, 10)};
    augend.addTile(std::unique_ptr<h_matrix> {
        new mutable_matrix_view {&aug_tile1, 3, 5, 4, 2, false}}, 0, 0);
    augend.addTile(std::unique_ptr<h_matrix> {
        new mutable_matrix_view {&aug_tile2, 2, 5, 4, 2, true}}, 0, 2);
    addend.addTile(std::unique_ptr<h_matrix> {
        new const_matrix_view {&add_tile1, 1, 3, 2, 3, false}}, 0, 0);
    addend.addTile(std::unique_ptr<h_matrix> {
        new const_matrix_view {&add_tile2, 5, 1, 2, 3, true}}, 2, 1);
    full_matrix expected {augend.rows(), augend.cols()};
    for (index_type j = 0; j < expected.cols(); ++j)
      for (index_type i = 0; i < expected.rows(); ++i)
        expected.set(op(augend.get(i, j), addend.get(i, j)), i, j);
    switch (operation::value)
      {
      case '+': augend += addend; break;
      case '-': augend -= addend; break;
      default: assert(false);
      }
    MY_CHECK_MATRIX_EQUAL(expected, augend);
  }
  {
    // ArrangedMatrix + ArrangedMatrix
    // (dumb_take_delta: full_aug_ptr == nullptr)
    //
    //
    full_matrix aug_data {mkrndmatrix(10, 10)};
    const full_matrix add_data {mkrndmatrix(10, 10)};
    mock_matrix aug_mock {};
    aug_mock.mockRows([&aug_data]()->index_type{ return aug_data.rows(); });
    aug_mock.mockCols([&aug_data]()->index_type{ return aug_data.cols(); });
    aug_mock.mockGet([&aug_data](index_type i, index_type j)->index_type{
        return aug_data.get(i, j);
    });
    aug_mock.mockSet([&aug_data](value_type v, index_type i, index_type j)->void{
        return aug_data.set(v, i, j);
    });
    arranged_matrix augend {30, 50};
    arranged_matrix addend {30, 50};
    augend.addTile(std::unique_ptr<h_matrix> {new mutable_matrix_view {
          &aug_mock, 2, 4, 3, 5, false}}, 7, 12);
    addend.addTile(std::unique_ptr<h_matrix> {new const_matrix_view {
          &add_data, 1, 3, 3, 5, true}}, 7, 12);
    full_matrix expected {augend.rows(), augend.cols()};
    for (index_type j = 0; j < expected.cols(); ++j)
      for (index_type i = 0; i < expected.rows(); ++i)
        expected.set(op(augend.get(i, j), addend.get(i, j)), i, j);
    switch (operation::value)
      {
      case '+': augend += addend; break;
      case '-': augend -= addend; break;
      default: assert(false);
      }
    MY_CHECK_MATRIX_EQUAL(expected, augend);
  }
  {
    // ArrangedMatrix + ArrangedMatrix
    // (dumb_take_delta: full_add_ptr == nullptr)
    //
    //
    full_matrix aug_data {mkrndmatrix(10, 10)};
    const full_matrix add_data {mkrndmatrix(10, 10)};
    const mock_matrix add_mock {};
    add_mock.mockRows([&add_data]()->index_type{ return add_data.rows(); });
    add_mock.mockCols([&add_data]()->index_type{ return add_data.cols(); });
    add_mock.mockGet([&add_data](index_type i, index_type j)->index_type{
        return add_data.get(i, j);
    });
    arranged_matrix augend {30, 50};
    arranged_matrix addend {30, 50};
    augend.addTile(std::unique_ptr<h_matrix> {new mutable_matrix_view {
          &aug_data, 2, 4, 3, 5, false}}, 7, 12);
    addend.addTile(std::unique_ptr<h_matrix> {new const_matrix_view {
          &add_mock, 1, 3, 3, 5, true}}, 7, 12);
    full_matrix expected {augend.rows(), augend.cols()};
    for (index_type j = 0; j < expected.cols(); ++j)
      for (index_type i = 0; i < expected.rows(); ++i)
        expected.set(op(augend.get(i, j), addend.get(i, j)), i, j);
    switch (operation::value)
      {
      case '+': augend += addend; break;
      case '-': augend -= addend; break;
      default: assert(false);
      }
    MY_CHECK_MATRIX_EQUAL(expected, augend);
  }
}


/**
 * @brief
 *         Unit test that exercises the remaining parts of the `poly_prod`
 *         function that are left out by the general unit test.
 *
 */
BOOST_AUTO_TEST_CASE(poly_prod)
{
  typedef int value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  //using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  using mock_matrix = bsc_klammler::test::MockHMatrix<value_type>;
  using const_matrix_view = bsc_klammler::ConstMatrixView<value_type>;
  using index_type = h_matrix::index_type;
  constexpr index_type left_off_rows {1};
  constexpr index_type left_off_cols {2};
  constexpr index_type left_rows {3};
  constexpr index_type left_cols {2};
  constexpr index_type right_off_rows {3};
  constexpr index_type right_off_cols {4};
  constexpr index_type right_rows {2};
  constexpr index_type right_cols {4};
  const auto mkviewl = [](const h_matrix& matrix)->const_matrix_view{
    assert(matrix.rows() == left_off_rows + left_rows);
    assert(matrix.cols() == left_off_cols + left_cols);
    return const_matrix_view {
      &matrix, left_off_rows, left_off_cols, left_rows, left_cols, false
    };
  };
  const auto mkviewr = [](const h_matrix& matrix)->const_matrix_view{
    assert(matrix.rows() == right_off_rows + right_rows);
    assert(matrix.cols() == right_off_cols + right_cols);
    return const_matrix_view {
      &matrix, right_off_rows, right_off_cols, right_rows, right_cols, false
    };
  };
  const auto mktransl = [](const h_matrix& matrix)->const_matrix_view{
    assert(matrix.rows() == left_off_cols + left_cols);
    assert(matrix.cols() == left_off_rows + left_rows);
    return const_matrix_view {
      &matrix, left_off_rows, left_off_cols, left_rows, left_cols, true
    };
  };
  const auto mktransr = [](const h_matrix& matrix)->const_matrix_view{
    assert(matrix.rows() == right_off_cols + right_cols);
    assert(matrix.cols() == right_off_rows + right_rows);
    return const_matrix_view {
      &matrix, right_off_rows, right_off_cols, right_rows, right_cols, true
    };
  };
  const full_matrix left = {
    {-1, -1, -1, -1},
    {-1, -1, 11, 12},
    {-1, -1, 21, 22},
    {-1, -1, 31, 32},
  };
  const full_matrix left_trans = {
    {-1, -1, -1, -1},
    {-1, -1, -1, -1},
    {-1, 11, 21, 31},
    {-1, 12, 22, 32},
  };
  const full_matrix right = {
    {-1, -1, -1, -1, -1, -1, -1, -1},
    {-1, -1, -1, -1, -1, -1, -1, -1},
    {-1, -1, -1, -1, -1, -1, -1, -1},
    {-1, -1, -1, -1, 11, 12, 13, 14},
    {-1, -1, -1, -1, 21, 22, 23, 24},
  };
  const full_matrix right_trans = {
    {-1, -1, -1, -1, -1},
    {-1, -1, -1, -1, -1},
    {-1, -1, -1, -1, -1},
    {-1, -1, -1, -1, -1},
    {-1, -1, -1, 11, 21},
    {-1, -1, -1, 12, 22},
    {-1, -1, -1, 13, 23},
    {-1, -1, -1, 14, 24},
  };
  const full_matrix expected = {
    { 373,  396,  419,  442},
    { 693,  736,  779,  822},
    {1013, 1076, 1139, 1202},
  };
  const mock_matrix mock_left {};
  const mock_matrix mock_left_trans {};
  const mock_matrix mock_right {};
  const mock_matrix mock_right_trans {};
  mock_left.mockRows([&left]()->index_type{
      return left.rows();
    });
  mock_left.mockCols([&left]()->index_type{
      return left.cols();
    });
  mock_left.mockGet([&left](index_type i, index_type j)->value_type{
      return left.get(i, j);
    });
  mock_left_trans.mockRows([&left_trans]()->index_type{
      return left_trans.rows();
    });
  mock_left_trans.mockCols([&left_trans]()->index_type{
      return left_trans.cols();
    });
  mock_left_trans.mockGet([&left_trans](index_type i, index_type j)->value_type{
      return left_trans.get(i, j);
    });
  mock_right.mockRows([&right]()->index_type{
      return right.rows();
    });
  mock_right.mockCols([&right]()->index_type{
      return right.cols();
    });
  mock_right.mockGet([&right](index_type i, index_type j)->value_type{
      return right.get(i, j);
    });
  mock_right_trans.mockRows([&right_trans]()->index_type{
      return right_trans.rows();
    });
  mock_right_trans.mockCols([&right_trans]()->index_type{
      return right_trans.cols();
    });
  mock_right_trans.mockGet([&right_trans](index_type i, index_type j)->value_type{
      return right_trans.get(i, j);
    });
  MY_CHECK_MATRIX_EQUAL(
      expected, mkviewl(left) * mkviewr(right));
  MY_CHECK_MATRIX_EQUAL(
      expected, mkviewl(left) * mktransr(right_trans));
  MY_CHECK_MATRIX_EQUAL(
      expected, mkviewl(left) * mkviewr(mock_right));
  MY_CHECK_MATRIX_EQUAL(
      expected, mkviewl(left) * mktransr(mock_right_trans));
  MY_CHECK_MATRIX_EQUAL(
      expected, mktransl(left_trans) * mkviewr(right));
  MY_CHECK_MATRIX_EQUAL(
      expected, mktransl(left_trans) * mktransr(right_trans));
  MY_CHECK_MATRIX_EQUAL(
      expected, mktransl(left_trans) * mkviewr(mock_right));
  MY_CHECK_MATRIX_EQUAL(
      expected, mktransl(left_trans) * mktransr(mock_right_trans));
  MY_CHECK_MATRIX_EQUAL(
      expected, mkviewl(mock_left) * mkviewr(right));
  MY_CHECK_MATRIX_EQUAL(
      expected, mkviewl(mock_left) * mktransr(right_trans));
  MY_CHECK_MATRIX_EQUAL(
      expected, mkviewl(mock_left) * mkviewr(mock_right));
  MY_CHECK_MATRIX_EQUAL(
      expected, mkviewl(mock_left) * mktransr(mock_right_trans));
  MY_CHECK_MATRIX_EQUAL(
      expected, mktransl(mock_left_trans) * mkviewr(right));
  MY_CHECK_MATRIX_EQUAL(
      expected, mktransl(mock_left_trans) * mktransr(right_trans));
  MY_CHECK_MATRIX_EQUAL(
      expected, mktransl(mock_left_trans) * mkviewr(mock_right));
  MY_CHECK_MATRIX_EQUAL(
      expected, mktransl(mock_left_trans) * mktransr(mock_right_trans));
}

// BOOST_AUTO_TEST_CASE(parallel)
// {
//   using full_matrix = bsc_klammler::FullMatrix<double>;
//   const full_matrix matrix1 {1000, 500};
//   const full_matrix matrix2 {500, 1500};
//   std::string ignore {};
//   {
//     std::cout << "Hit RET to start parallel multiplication." << std::flush;
//     std::getline(std::cin, ignore);
//     const full_matrix prod {matrix1 * matrix2};
//   }
//   {
//     std::cout << "Hit RET to start sequential multiplication." << std::flush;
//     std::getline(std::cin, ignore);
// #ifdef _OPENMP
// #pragma omp parallel for
// #endif
//     for (int i = 0; i < 1; ++i)
//       {
//         const full_matrix prod {matrix1 * matrix2};
//       }
//   }
// }
