/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_arranged_matrix.cpp
 *
 * @brief
 *         Unit tests for the `ArrangedMatrix` template class.
 *
 */

#include "../global.hpp"           // (package configuration)
#include "../arranged_matrix.hpp"  // bsc_klammler::ArrangedMatrix

#include <array>                   // std::array
#include <climits>                 // CHAR_BIT
#include <stdexcept>               // std::invalid_argument
#include <typeinfo>                // typeid
#include <utility>                 // std::move, std::swap, std::pair

#include "../full_matrix.hpp"      // bsc_klammler::FullMatrix
#include "../matrix_view.hpp"      // bsc_klammler::MatrixView
#include "../random.hpp"           // bsc_klammler::random_real
#include "foo.hpp"                 // bsc_klammler::test::Foo
#include "util.hpp"                // MY_CHECK_MATRIX_EQUAL, ...
#include "visitors.hpp"            // bsc_klammler::test::Extractor, ...

#define BOOST_TEST_MODULE arranged_matrix
#include "module.hpp"

using bsc_klammler::test::Foo;
using bsc_klammler::random_int;
using bsc_klammler::random_real;


namespace /* anonymous */
{
  struct Bar
  {
    static unsigned object_count;
    Bar() noexcept { Bar::object_count += 1; }
    ~Bar() noexcept { Bar::object_count -= 1; }
  };
  unsigned Bar::object_count = 0;
}


/**
 * @brief
 *         Unit test for the member typdefs.
 *
 * This test checks that:
 *
 * - the member types `value_type`, `index_type` and `super_type` are defined
 *   correctly
 *
 */
BOOST_AUTO_TEST_CASE(member_typedefs)
{
  // There is no overload for 'operator<<(std::ostream&, std::type_info)' so we
  // cannot use 'BOOST_CHECK_EQUAL'.
  struct Baz {};
  BOOST_CHECK(typeid(bsc_klammler::HMatrix<Baz>) ==
              typeid(bsc_klammler::ArrangedMatrix<Baz>::super_type));
  BOOST_CHECK(typeid(Baz) ==
              typeid(bsc_klammler::ArrangedMatrix<Baz>::value_type));
  for (typename bsc_klammler::ArrangedMatrix<Baz>::index_type i = 0; i < 10; ++i)
    continue;
}


/**
 * @brief
 *         Unit test for the default constructor.
 *
 * This test checks that:
 *
 * - the default constructor creates an empty `ArrangedMatrix`
 *
 */
BOOST_AUTO_TEST_CASE(ctor_default)
{
  bsc_klammler::ArrangedMatrix<double> matrix {};
  BOOST_CHECK_EQUAL(0, matrix.rows());
  BOOST_CHECK_EQUAL(0, matrix.cols());
  BOOST_CHECK_EQUAL(true, matrix.isRowLinear());
  BOOST_CHECK_EQUAL(true, matrix.isColumnLinear());
  BOOST_CHECK_EQUAL(true, matrix.isDiagonal());
}


/**
 * @brief
 *         Unit test for the constructor taking an explicit size.
 *
 * This test checks that:
 *
 * - the default constructor creates an empty `ArrangedMatrix` of correct size
 *
 */
BOOST_AUTO_TEST_CASE(ctor)
{
  typedef double value_type;
  using index_type = bsc_klammler::HMatrix<value_type>::index_type;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  const index_type rows {random_int<index_type>(1, 128)};
  const index_type cols {random_int<index_type>(1, 128)};
  const arranged_matrix matrix {rows, cols};
  BOOST_CHECK_EQUAL(rows, matrix.rows());
  BOOST_CHECK_EQUAL(cols, matrix.cols());
  BOOST_CHECK_EQUAL(true, matrix.isRowLinear());
  BOOST_CHECK_EQUAL(true, matrix.isColumnLinear());
  BOOST_CHECK_EQUAL(rows == cols, matrix.isDiagonal());
  for (index_type j = 0; j < matrix.cols(); ++j)
    {
      for (index_type i = 0; i < matrix.rows(); ++i)
        {
          BOOST_CHECK_EQUAL(0.0, matrix.get(i, j));
        }
    }
}


/**
 * @brief
 *         Unit tests for copy and move constructors and assignment operators.
 *
 * This test checks that:
 *
 * - the copy constructor has *copy* semantics
 * - the move constructor has *move* semantics
 * - the copy-assignment operator has *copy* semantics
 * - the move-assignment operator has *move* semantics
 * - the `clone` friend function has *copy* semantics
 *
 */
BOOST_AUTO_TEST_CASE(copy_move)
{
  typedef Foo value_type;
  using index_type = bsc_klammler::HMatrix<value_type>::index_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  // These 'check_*' lambdas return 0 if the predicate is true and a non-zero
  // error code otherwise.
  const auto check_is_correct_default = [](const arranged_matrix& matrix)->int{
    if (matrix.rows() != 0 || matrix.cols() != 0) { return 1; }
    if (!matrix.isRowLinear() || !matrix.isColumnLinear()) { return 2; }
    if (!matrix.isDiagonal()) { return 3; }
    return 0;
  };
  const auto check_are_copies = [](h_matrix& first,
                                   h_matrix& second)->int{
    if (&first == &second) { return 1; }
    if (first.rows() != second.rows()) { return 2; }
    if (first.cols() != second.cols()) { return 3; }
    for (index_type j = 0; j < first.cols(); ++j)
      {
        for (index_type i = 0; i < first.rows(); ++i)
          {
            const value_type v1 {first.get(i, j)};
            const value_type v2 {second.get(i, j)};
            if (v1 != v2) { return 4; }
            if (v1 != h_matrix::zero())
              {
                {
                  first.set(~v1, i, j);
                  if (second.get(i, j) != v2) { return 5; }
                  first.set(v1, i, j);
                }
                {
                  second.set(~v2, i, j);
                  if (first.get(i, j) != v1) { return 6; }
                  second.set(v2, i, j);
                }
              }
          }
      }
    return 0;
  };
  Foo::activate();
  const std::array<full_matrix, 6> expectations = {
    full_matrix {},
    full_matrix {{11}},
    full_matrix {12, 34},
    full_matrix {
      {11, 12,  0, 11,  0,  0},
      {21, 22,  0, 21,  0,  0},
    },
    full_matrix {
      { 0,  0},
      {11, 12},
      {11, 12},
      {21, 22},
    },
    full_matrix {
      {0,  0,  0,  0,  0,  0,  0,  0,  0, 11},
      {0,  0,  0,  0,  0,  0, 11, 12, 13, 21},
      {0,  0,  0,  0,  0,  0, 21, 22, 23, 31},
      {0, 11,  0,  0,  0,  0, 31, 32, 33, 41},
    },
  };
  const std::array<bool, 6> are_row_linear = {
    true, true, true, true, false, false
  };
  const std::array<bool, 6> are_col_linear = {
    true, true, true, false, true, false
  };
  const std::array<bool, 6> are_diagonal = {
    true, true, false, false, false, false
  };
  const std::array<std::function<void(arranged_matrix&)>, 6> arrangings = {
    [](arranged_matrix& arr BSC_KLAMMLER_ATTRIBUTE_UNUSED)->void{
      // do nothing
    },
    [](arranged_matrix& arr)->void{
      arr.addTile(full_matrix {{11}}, 0, 0);
    },
    [](arranged_matrix& arr BSC_KLAMMLER_ATTRIBUTE_UNUSED)->void{
      // do nothing
    },
    [](arranged_matrix& arr)->void{
      arr.addTile(full_matrix {{11, 12}, {21, 22}}, 0, 0);
      arr.addTile(full_matrix {{11}, {21}}, 0, 3);
    },
    [](arranged_matrix& arr)->void{
      arr.addTile(full_matrix {{11, 12}}, 1, 0);
      arr.addTile(full_matrix {{11, 12}, {21, 22}}, 2, 0);
    },
    [](arranged_matrix& arr)->void{
      arr.addTile(full_matrix {{11}}, 3, 1);
      arr.addTile(full_matrix {{11}, {21}, {31}, {41}}, 0, 9);
      arr.addTile(full_matrix {{11, 12, 13}, {21, 22, 23}, {31, 32, 33}}, 1, 6);
    },
  };
  // copy c'tor
  for (std::size_t i = 0; i < 6; ++i)
    {
      arranged_matrix first {
        expectations.at(i).rows(), expectations.at(i).cols()
      };
      arrangings.at(i)(first);
      MY_CHECK_MATRIX_EQUAL(expectations.at(i), first);
      BOOST_CHECK_EQUAL(are_row_linear.at(i), first.isRowLinear());
      BOOST_CHECK_EQUAL(are_col_linear.at(i), first.isColumnLinear());
      BOOST_CHECK_EQUAL(are_diagonal.at(i), first.isDiagonal());
      // We should really use a const-reference to do the check.
      arranged_matrix second {const_cast<const arranged_matrix&>(first)};
      BOOST_CHECK_EQUAL(0, check_are_copies(first, second));
      BOOST_CHECK_EQUAL(are_row_linear.at(i), second.isRowLinear());
      BOOST_CHECK_EQUAL(are_col_linear.at(i), second.isColumnLinear());
      BOOST_CHECK_EQUAL(are_diagonal.at(i), second.isDiagonal());
    }
  // move c'tor
  for (std::size_t i = 0; i < 6; ++i)
    {
      arranged_matrix first {
        expectations.at(i).rows(), expectations.at(i).cols()
      };
      arrangings.at(i)(first);
      MY_CHECK_MATRIX_EQUAL(expectations.at(i), first);
      BOOST_CHECK_EQUAL(are_row_linear.at(i), first.isRowLinear());
      BOOST_CHECK_EQUAL(are_col_linear.at(i), first.isColumnLinear());
      BOOST_CHECK_EQUAL(are_diagonal.at(i), first.isDiagonal());
      Foo::deactivate();
      arranged_matrix second {std::move(first)};
      Foo::activate();
      BOOST_CHECK_EQUAL(0, check_is_correct_default(first));
      MY_CHECK_MATRIX_EQUAL(expectations.at(i), second);
      BOOST_CHECK_EQUAL(are_row_linear.at(i), second.isRowLinear());
      BOOST_CHECK_EQUAL(are_col_linear.at(i), second.isColumnLinear());
      BOOST_CHECK_EQUAL(are_diagonal.at(i), second.isDiagonal());
    }
  // copy-assignment op'tor
  for (std::size_t i = 0; i < 6; ++i)
    {
      arranged_matrix first {
        expectations.at(i).rows(), expectations.at(i).cols()
      };
      arrangings.at(i)(first);
      MY_CHECK_MATRIX_EQUAL(expectations.at(i), first);
      BOOST_CHECK_EQUAL(are_row_linear.at(i), first.isRowLinear());
      BOOST_CHECK_EQUAL(are_col_linear.at(i), first.isColumnLinear());
      BOOST_CHECK_EQUAL(are_diagonal.at(i), first.isDiagonal());
      arranged_matrix second {};
      BOOST_CHECK_EQUAL(0, check_is_correct_default(second));
      // We should really use a const-reference to do the check.
      second = const_cast<const arranged_matrix&>(first);
      MY_CHECK_MATRIX_EQUAL(expectations.at(i), first);
      MY_CHECK_MATRIX_EQUAL(expectations.at(i), second);
      BOOST_CHECK_EQUAL(0, check_are_copies(first, second));
    }
  // move-assignment op'tor
  for (std::size_t i = 0; i < 6; ++i)
    {
      arranged_matrix first {
        expectations.at(i).rows(), expectations.at(i).cols()
      };
      arrangings.at(i)(first);
      MY_CHECK_MATRIX_EQUAL(expectations.at(i), first);
      BOOST_CHECK_EQUAL(are_row_linear.at(i), first.isRowLinear());
      BOOST_CHECK_EQUAL(are_col_linear.at(i), first.isColumnLinear());
      BOOST_CHECK_EQUAL(are_diagonal.at(i), first.isDiagonal());
      arranged_matrix second {};
      BOOST_CHECK_EQUAL(0, check_is_correct_default(second));
      Foo::deactivate();
      second = std::move(first);
      Foo::activate();
      BOOST_CHECK_EQUAL(0, check_is_correct_default(first));
      MY_CHECK_MATRIX_EQUAL(expectations.at(i), second);
    }
  // clone()
  for (std::size_t i = 0; i < 6; ++i)
    {
      arranged_matrix first {
        expectations.at(i).rows(), expectations.at(i).cols()
      };
      arrangings.at(i)(first);
      MY_CHECK_MATRIX_EQUAL(expectations.at(i), first);
      BOOST_CHECK_EQUAL(are_row_linear.at(i), first.isRowLinear());
      BOOST_CHECK_EQUAL(are_col_linear.at(i), first.isColumnLinear());
      BOOST_CHECK_EQUAL(are_diagonal.at(i), first.isDiagonal());
      // We should really use a const-reference to do the check.
      std::unique_ptr<h_matrix> second_uptr =
        const_cast<const arranged_matrix *>(&first)->clone();
      BOOST_CHECK_EQUAL(0, check_are_copies(first, *second_uptr));
      BOOST_CHECK(dynamic_cast<arranged_matrix *>(second_uptr.get()));
      BOOST_CHECK_EQUAL(are_row_linear.at(i),
                        dynamic_cast<arranged_matrix *>(second_uptr.get())
                        ->isRowLinear());
      BOOST_CHECK_EQUAL(are_col_linear.at(i),
                        dynamic_cast<arranged_matrix *>(second_uptr.get())
                        ->isColumnLinear());
      BOOST_CHECK_EQUAL(are_diagonal.at(i),
                        dynamic_cast<arranged_matrix *>(second_uptr.get())
                        ->isDiagonal());
    }
}


/**
 * @brief
 *         Unit tests for the `addTile` function members.
 *
 * This test checks that:
 *
 * - the functions add `Tile`s correctly
 * - the `isRowLinear`, `isColumnLinear` and `isDiagonal` function members
 *   return the correct status hereafter
 *
 */
BOOST_AUTO_TEST_CASE(add_tile__general)
{
  typedef int value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using index_type = h_matrix::index_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  {
    // We only make one test for the version taking a pointer since we know
    // that the reference version will call it anyway and the test can be
    // written cleaner with it.
    const full_matrix expected {
      {11, 12, 13, 14},
      {21, 22, 23, 24},
      {31, 32, 33, 34},
    };
    arranged_matrix actual1 {};
    arranged_matrix actual2 {3, 4};
    for (index_type j = 0; j < expected.cols(); ++j)
      {
        for (index_type i = 0; i < expected.rows(); ++i)
          {
            full_matrix full {{expected.get(i, j)}};
            actual1.addTile(full.clone(), i, j, true);
            actual2.addTile(full.clone(), i, j, false);
          }
      }
    MY_CHECK_MATRIX_EQUAL(expected, actual1);
    MY_CHECK_MATRIX_EQUAL(expected, actual2);
    BOOST_CHECK(!actual1.isRowLinear());
    BOOST_CHECK(!actual2.isRowLinear());
    BOOST_CHECK(!actual1.isColumnLinear());
    BOOST_CHECK(!actual2.isColumnLinear());
    BOOST_CHECK(!actual1.isDiagonal());
    BOOST_CHECK(!actual2.isDiagonal());
  }
  {
    const full_matrix expected {
      {11, 12,  0,  0, 11},
      {21, 22,  0,  0, 21},
    };
    arranged_matrix actual {2, 5};
    actual.addTile(full_matrix {{11, 12}, {21, 22}}, 0, 0);
    actual.addTile(full_matrix {{11}, {21}}, 0, 4);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK(actual.isRowLinear());
    BOOST_CHECK(!actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
  }
  {
    const full_matrix expected {
      { 0},
      { 0},
      {11},
    };
    arranged_matrix actual {3, 1};
    actual.addTile(full_matrix {{11}}, 2, 0);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK(!actual.isRowLinear());
    BOOST_CHECK(actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
  }
  {
    const full_matrix expected = {
      {11, 12,  0,  0},
      {21, 22,  0,  0},
      { 0,  0,  0,  0},
      { 0,  0,  0, 11},
    };
    arranged_matrix actual {4, 4};
    actual.addTile(full_matrix {{11, 12}, {21, 22}}, 0, 0);
    actual.addTile(full_matrix {{11}}, 3, 3);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK(!actual.isRowLinear());
    BOOST_CHECK(!actual.isColumnLinear());
    BOOST_CHECK(actual.isDiagonal());
  }
  {
    const full_matrix expected {
      { 0,  0,  0},
      { 0,  0,  0},
      {11, 12, 13},
      {21, 22, 23},
      { 0,  0,  0},
      {11, 12, 13},
      { 0,  0,  0},
      {11, 12, 13},
      {21, 22, 23},
      {31, 32, 33},
      {41, 42, 43},
    };
    arranged_matrix actual {11, 3};
    BOOST_CHECK(actual.isRowLinear());
    BOOST_CHECK(actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    actual.addTile(
        full_matrix {{11, 12, 13}, {21, 22, 23}},
        2,
        0);
    BOOST_CHECK(!actual.isRowLinear());
    BOOST_CHECK(actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    actual.addTile(
        full_matrix {{11, 12, 13}},
        5,
        0);
    BOOST_CHECK(!actual.isRowLinear());
    BOOST_CHECK(actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    actual.addTile(
        full_matrix {{11, 12, 13}, {21, 22, 23}, {31, 32, 33}, {41, 42, 43}},
        7,
        0);
    BOOST_CHECK(!actual.isRowLinear());
    BOOST_CHECK(actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {
      {11, 12,  0,  0,  0,  0, 11, 12, 13, 14, 15,  0,  0,  0, 11,  0,  0,  0},
      {21, 22,  0,  0,  0,  0, 21, 22, 23, 24, 25,  0,  0,  0, 21,  0,  0,  0},
      {31, 32,  0,  0,  0,  0, 31, 32, 33, 34, 35,  0,  0,  0, 31,  0,  0,  0},
      {41, 42,  0,  0,  0,  0, 41, 42, 43, 44, 45,  0,  0,  0, 41,  0,  0,  0},
    };
    arranged_matrix actual {4, 18};
    BOOST_CHECK(actual.isRowLinear());
    BOOST_CHECK(actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    actual.addTile(
        full_matrix {{11, 12}, {21, 22}, {31, 32}, {41, 42}},
        0,
        0);
    BOOST_CHECK(actual.isRowLinear());
    BOOST_CHECK(!actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    actual.addTile(full_matrix {
        {11, 12, 13, 14, 15},
        {21, 22, 23, 24, 25},
        {31, 32, 33, 34, 35},
        {41, 42, 43, 44, 45}
      }, 0, 6);
    BOOST_CHECK(actual.isRowLinear());
    BOOST_CHECK(!actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    actual.addTile(
        full_matrix {{11}, {21}, {31}, {41}},
        0,
        14);
    BOOST_CHECK(actual.isRowLinear());
    BOOST_CHECK(!actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {
      {11, 12, 13},
      {21, 22, 23},
      {31, 32, 33},
    };
    arranged_matrix actual {3, 3};
    actual.addTile(full_matrix {{11, 12, 13}, {21, 22, 23}, {31, 32, 33}}, 0, 0);
    BOOST_CHECK(actual.isRowLinear());
    BOOST_CHECK(actual.isColumnLinear());
    BOOST_CHECK(actual.isDiagonal());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {
      {11, 12,  0, 11},
      {21, 22,  0, 11},
    };
    arranged_matrix actual {2, 4};
    BOOST_CHECK(actual.isRowLinear());
    BOOST_CHECK(actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    actual.addTile(full_matrix {{11, 12}, {21, 22}}, 0, 0);
    BOOST_CHECK(actual.isRowLinear());
    BOOST_CHECK(!actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    actual.addTile(full_matrix {{11}}, 0, 3);
    BOOST_CHECK(!actual.isRowLinear());
    BOOST_CHECK(!actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    actual.addTile(full_matrix {{11}}, 1, 3);
    BOOST_CHECK(!actual.isRowLinear());
    BOOST_CHECK(!actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {
      {11, 12},
      {11, 11},
    };
    arranged_matrix actual {2, 2};
    BOOST_CHECK(actual.isRowLinear());
    BOOST_CHECK(actual.isColumnLinear());
    BOOST_CHECK(actual.isDiagonal());
    actual.addTile(full_matrix {{11, 12}}, 0, 0);
    BOOST_CHECK(!actual.isRowLinear());
    BOOST_CHECK(actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    actual.addTile(full_matrix {{11}}, 1, 0);
    BOOST_CHECK(!actual.isRowLinear());
    BOOST_CHECK(!actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    actual.addTile(full_matrix {{11}}, 1, 1);
    BOOST_CHECK(!actual.isRowLinear());
    BOOST_CHECK(!actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {
      { 0,  0,  0,  0,  0,  0,  0,  0, 11,  0,  0,  0,  0,  0,  0,  0},
      { 0, 11, 12, 13, 14,  0,  0,  0, 21,  0,  0,  0,  0,  0,  0,  0},
      { 0, 21, 22, 23, 24,  0,  0,  0, 31,  0,  0, 11, 12,  0,  0,  0},
      { 0, 31, 32, 33, 34,  0,  0,  0, 41,  0,  0, 21, 22,  0,  0,  0},
      { 0,  0,  0,  0,  0,  0,  0,  0, 51,  0,  0,  0,  0,  0,  0,  0},
      { 0,  0,  0,  0,  0,  0,  0,  0, 61,  0,  0,  0,  0,  0,  0,  0},
      { 0,  0,  0,  0,  0,  0,  0,  0, 71,  0,  0,  0,  0,  0, 11, 11},
      { 0,  0,  0,  0,  0,  0,  0,  0, 81,  0,  0,  0,  0,  0, 11, 11},
    };
    arranged_matrix actual {4, 5};
    actual.addTile(
        full_matrix {{11, 12, 13, 14}, {21, 22, 23, 24}, {31, 32, 33, 34}},
        1,
        1,
        false);
    actual.addTile(
        full_matrix {{11, 12}, {21, 22}},
        2,
        11,
        true);
    actual.addTile(
        full_matrix {{11}, {21}, {31}, {41}, {51}, {61}, {71}, {81}},
        0,
        8,
        true);
    actual.addTile(full_matrix {{11}}, 6, 15, true);
    actual.addTile(full_matrix {{11}}, 6, 14, false);
    actual.addTile(full_matrix {{11}}, 7, 15, true);
    actual.addTile(full_matrix {{11}}, 7, 14, false);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK(!actual.isRowLinear());
    BOOST_CHECK(!actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
  }
  {
    const full_matrix expected {
      { 0,  0,  0,},
      { 0,  0,  0,},
    };
    arranged_matrix actual {};
    actual.addTile(full_matrix {}, 2, 3, true);
    actual.addTile(std::unique_ptr<h_matrix> {new full_matrix {}}, 2, 3, true);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    BOOST_CHECK(actual.isRowLinear());
    BOOST_CHECK(actual.isColumnLinear());
    BOOST_CHECK(!actual.isDiagonal());
  }
}


/**
 * @brief
 *         Unit test for the `addTile` function members.
 *
 * This test checks that:
 *
 * - no extranous copies of the tile data are made
 *
 */
BOOST_AUTO_TEST_CASE(add_tile__no_copies)
{
  typedef Bar value_type;
  using index_type = bsc_klammler::HMatrix<value_type>::index_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  unsigned bar_expect {0};
  Bar::object_count = 0;
  {
    arranged_matrix arrangement {};
    BOOST_CHECK_EQUAL(0, Bar::object_count);
    {
      const index_type rows {random_int<index_type>(1, 32)};
      const index_type cols {random_int<index_type>(1, 32)};
      const index_type size {rows * cols};
      // Create 'Bar's locally on the stack.
      const full_matrix full {rows, cols};
      bar_expect += size;
      BOOST_CHECK_EQUAL(bar_expect, Bar::object_count);
      // 'addTile' must 'clone()' the reference onto the free store.
      arrangement.addTile(full,
                          random_int<index_type>(0, 1024),
                          random_int<index_type>(0, 1024),
                          true);
      bar_expect += size;
      BOOST_CHECK_EQUAL(bar_expect, Bar::object_count);
      // 'full' goes out of scope.
      bar_expect -= size;
    }
    BOOST_CHECK_EQUAL(bar_expect, Bar::object_count);
    {
      const index_type rows {random_int<index_type>(1, 32)};
      const index_type cols {random_int<index_type>(1, 32)};
      const index_type size {rows * cols};
      // Create 'Bar's dynamically on the heap.
      std::unique_ptr<h_matrix> full_uptr {new full_matrix {rows, cols}};
      bar_expect += size;
      BOOST_CHECK_EQUAL(bar_expect, Bar::object_count);
      // 'addTile' must not 'clone()' the pointer.
      arrangement.addTile(std::move(full_uptr),
                          random_int<index_type>(1025, 2048),
                          random_int<index_type>(1025, 2048),
                          true);
      bar_expect += 0;
      BOOST_CHECK_EQUAL(bar_expect, Bar::object_count);
      // Nothing goes out of scope.
      bar_expect -= 0;
    }
    BOOST_CHECK_EQUAL(bar_expect, Bar::object_count);
    // 'arrangement' goes out of scope.
  }
  BOOST_CHECK_EQUAL(0, Bar::object_count);
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checing unit test for the `addTile` function members.
 *
 * This test checks that:
 *
 * - attempting to add a `nullptr` is rejcted
 * - attempting to add tiles that would not fit is rejected unless resizing is
 *   explicitely requested
 * - attempting to add tiles that would overlap is rejected
 * - no memory is leaked if a tile is rejected
 *
 */
BOOST_AUTO_TEST_CASE(add_tile__invalid)
{
  typedef Bar value_type;
  using index_type = bsc_klammler::HMatrix<value_type>::index_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  {
    arranged_matrix frame {10, 10};
    for (const auto resize : {false, true})
      {
        try
          {
            frame.addTile(std::unique_ptr<h_matrix> {}, 1, 1, resize);
            BOOST_ERROR("no exception thrown");
          }
        catch (const std::invalid_argument&) { /* good */ }
      }
  }
  Bar::object_count = 0;
  {
    arranged_matrix frame {};
    full_matrix tile {};
    try
      {
        frame.addTile(tile, 1, 0);
        BOOST_ERROR("no exception thrown");
      }
    catch (const std::invalid_argument&) { /* good */ }
  }
  BOOST_CHECK_EQUAL(0, Bar::object_count);
  {
    arranged_matrix frame {};
    std::unique_ptr<h_matrix> tile_uptr {new full_matrix {1, 1}};
    try
      {
        frame.addTile(std::move(tile_uptr), 0, 0);
        BOOST_ERROR("no exception thrown");
      }
    catch (const std::invalid_argument&) { /* good */ }
    BOOST_CHECK_EQUAL(0, Bar::object_count);
  }
  BOOST_CHECK_EQUAL(0, Bar::object_count);
  {
    arranged_matrix frame {19, 91};
    full_matrix tile {10, 10};
    try
      {
        frame.addTile(tile, 10, 20);
        BOOST_ERROR("no exception thrown");
      }
    catch (const std::invalid_argument&) { /* good */ }
    BOOST_CHECK_EQUAL(100, Bar::object_count);
    try
      {
        frame.addTile(tile, 1, 82);
        BOOST_ERROR("no exception thrown");
      }
    catch (const std::invalid_argument&) { /* good */ }
    BOOST_CHECK_EQUAL(100, Bar::object_count);
    try
      {
        frame.addTile(tile, 200, 200);
        BOOST_ERROR("no exception thrown");
      }
    catch (const std::invalid_argument&) { /* good */ }
  }
  BOOST_CHECK_EQUAL(0, Bar::object_count);
  {
    arranged_matrix frame {10, 10};
    full_matrix tile1 {2, 3};
    full_matrix tile2 {2, 2};
    const std::pair<index_type, index_type> bad_offsets[] = {
      {5, 5},
      {4, 4},
      {4, 6},
      {5, 7},
      {6, 7},
      {6, 5},
      {5, 4},
    };
    frame.addTile(tile1, 5, 5);
    for (const auto& iter : bad_offsets)
      {
        try
          {
            frame.addTile(tile2, iter.first, iter.second);
            BOOST_ERROR("no exception thrown");
          }
        catch (const std::invalid_argument&) { /* good */ }
      }
  }
  BOOST_CHECK_EQUAL(0, Bar::object_count);
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit tests for the `get` and `set` function members.
 *
 * This test checks that:
 *
 * - the functions are correct
 *
 */
BOOST_AUTO_TEST_CASE(accessors)
{
  typedef unsigned value_type;
  using index_type = bsc_klammler::HMatrix<value_type>::index_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  const index_type rows {random_int<index_type>(16, 32) << 1};
  const index_type cols {random_int<index_type>(16, 32) << 1};
  const auto mktag = [](const index_type row,
                        const index_type col)->value_type{
    constexpr auto shift = CHAR_BIT * sizeof(value_type) / 2;
    return static_cast<value_type>((row << shift) | col);
  };
  arranged_matrix arrangement {rows, cols};
  for (index_type j = 0; j < cols; j += 2)
    for (index_type i = 0; i < rows; i += 2)
      arrangement.addTile(full_matrix {2, 2}, i, j);
  for (index_type j = 0; j < cols; ++j)
    for (index_type i = 0; i < rows; ++i)
      arrangement.set(mktag(i, j), i, j);
  for (index_type j = 0; j < cols; ++j)
    for (index_type i = 0; i < rows; ++i)
      BOOST_CHECK_EQUAL(mktag(i, j), arrangement.get(i, j));
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit tests for the `get` and `set` function members.
 *
 * This test checks that:
 *
 * - `get` and `set` properly detect out-of-range indices
 * - attempting to `set()` a value outside a tile is rejected
 *
 */
BOOST_AUTO_TEST_CASE(accessors__invalid)
{
  typedef double value_type;
  using index_type = bsc_klammler::HMatrix<value_type>::index_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  arranged_matrix frame {30, 20};
    // These indices fail the entire frame.
  const std::pair<index_type, index_type> bad_indices_frame[] = {
    {0, frame.cols()},
    {frame.rows(), 0},
    {frame.rows(), frame.cols()},
  };
  const std::pair<index_type, index_type> bad_indices_tile[] = {
    // These indices hit the entire frame but fail the tiles.
    {0, 0},
    {2, 0},
    {2, 5},
    {4, 2},
    {9, 9},
  };
  {
    const full_matrix tile1 {3, 4};
    const full_matrix tile2 {1, 1};
    frame.addTile(tile1, 1, 1);
    frame.addTile(tile2, 8, 8);
  }
  for (const auto iter : bad_indices_frame)
    {
      try
        {
          frame.get(iter.first, iter.second);
          BOOST_ERROR("getter threw no exception for out-of-frame index");
        }
      catch (const std::invalid_argument&) { /* good */ }
      try
        {
          frame.set(0.0, iter.first, iter.second);
          BOOST_ERROR("setter threw no exception for out-of-frame index");
        }
      catch (const std::invalid_argument&) { /* good */ }
    }
  for (const auto iter : bad_indices_tile)
    {
      BOOST_CHECK_EQUAL(0.0, frame.get(iter.first, iter.second));
      try
        {
          frame.set(0.0, iter.first, iter.second);
          BOOST_ERROR("setter threw no exception for out-of-tile index");
        }
      catch (const std::invalid_argument&) { /* good */ }
    }
}
#endif  // BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the `swap` friend function.
 *
 * This test checks that:
 *
 * - `swap` swaps data and makes no copies
 *
 */
BOOST_AUTO_TEST_CASE(swapping)
{
  typedef int value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  const full_matrix expected1 {
    {11, 12, 13},
    {21, 22, 23},
    { 0,  0,  0},
    {11, 12, 13},
  };
  const full_matrix expected2 {
    {11, 12,  0, 11,  0},
    {21, 22,  0, 12,  0},
  };
  h_matrix * tile1a_ptr {new full_matrix {{11, 12, 13}, {21, 22, 23}}};
  h_matrix * tile1b_ptr {new full_matrix {{11, 12, 13}}};
  h_matrix * tile2a_ptr {new full_matrix {{11, 12}, {21, 22}}};
  h_matrix * tile2b_ptr {new full_matrix {{11}, {12}}};
  arranged_matrix actual1 {expected1.rows(), expected1.cols()};
  arranged_matrix actual2 {expected2.rows(), expected2.cols()};
  actual1.addTile(std::unique_ptr<h_matrix> {tile1a_ptr}, 0, 0);
  actual1.addTile(std::unique_ptr<h_matrix> {tile1b_ptr}, 3, 0);
  actual2.addTile(std::unique_ptr<h_matrix> {tile2a_ptr}, 0, 0);
  actual2.addTile(std::unique_ptr<h_matrix> {tile2b_ptr}, 0, 3);
  MY_CHECK_MATRIX_EQUAL(expected1, actual1);
  MY_CHECK_MATRIX_EQUAL(expected2, actual2);
  BOOST_CHECK(!actual1.isRowLinear() && actual1.isColumnLinear());
  BOOST_CHECK(actual2.isRowLinear() && !actual2.isColumnLinear());
  swap(actual1, actual2);
  MY_CHECK_MATRIX_EQUAL(expected2, actual1);
  MY_CHECK_MATRIX_EQUAL(expected1, actual2);
  BOOST_CHECK(actual1.isRowLinear() && !actual1.isColumnLinear());
  BOOST_CHECK(!actual2.isRowLinear() && actual2.isColumnLinear());
  actual1.set(-3, 0, 0);
  actual1.set(-4, 0, 3);
  actual2.set(-1, 0, 0);
  actual2.set(-2, 3, 0);
  BOOST_CHECK_EQUAL(-1, tile1a_ptr->get(0, 0));
  BOOST_CHECK_EQUAL(-2, tile1b_ptr->get(0, 0));
  BOOST_CHECK_EQUAL(-3, tile2a_ptr->get(0, 0));
  BOOST_CHECK_EQUAL(-4, tile2b_ptr->get(0, 0));
}


/**
 * @brief
 *         Test that `MatrixView`s can be tiles in an `ArrangedMatrix`.
 *
 */
BOOST_AUTO_TEST_CASE(views)
{
  typedef int value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using const_matrix_view = bsc_klammler::ConstMatrixView<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  const full_matrix expected {
    {11,  0,  0,  0,  11, 12, 13, 14,  0,  0},
    { 0,  0,  0,  0,  21, 22, 23, 24,  0,  0},
    { 0, 11, 12, 13,  31, 32, 33, 34,  0,  0},
    { 0, 21, 22, 23,   0,  0,  0,  0,  0,  0},
  };
  arranged_matrix actual {};
  std::unique_ptr<h_matrix> view0_uptr {
    new const_matrix_view {}
  };
  std::unique_ptr<h_matrix> view1_uptr {
    new const_matrix_view {&expected, 0, 0, 1, 1}
  };
  std::unique_ptr<h_matrix> view2_uptr {
    new const_matrix_view {&expected, 2, 1, 2, 3}
  };
  std::unique_ptr<h_matrix> view3_uptr {
    new const_matrix_view {&expected, 0, 4, 3, 4}
  };
  actual.addTile(std::move(view0_uptr), 4, 10, true);
  actual.addTile(std::move(view1_uptr), 0,  0);
  actual.addTile(std::move(view2_uptr), 2,  1);
  actual.addTile(std::move(view3_uptr), 0,  4);
  MY_CHECK_MATRIX_EQUAL(expected, actual);
}


/**
 * @brief
 *         Unit test for the passive `passOverRegion` member function.
 *
 * This test checks that:
 *
 * - the correct region is shown to the visitor
 * - zero-blocks are not shown to the visitor
 * - transposition is applied correctly
 * - each element is visited exactly once
 * - the visited matrix is `const` but the visitor may mutate
 *
 */
BOOST_AUTO_TEST_CASE(pass_over_region__passive)
{
  typedef std::string value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  using extractor = bsc_klammler::test::Extractor<value_type>;
  // Assemble the matrix in a function so we can make it 'const' and don't need
  // to carry the building blocks around in our continued scope.
  const auto mkmatrix = []()->arranged_matrix{
    const full_matrix block_a {
      {"a11", "a12"},
      {"a21", "a22"},
    };
    const full_matrix block_b {
      {"b11", "b12", "b13"},
      {"b21", "b22", "b23"},
    };
    const full_matrix block_c {
      {"c11"},
      {"c21"},
    };
    const full_matrix block_d {};
    arranged_matrix matrix {5, 7};
    matrix.addTile(block_a, 1, 1);
    matrix.addTile(block_b, 3, 1);
    matrix.addTile(block_c, 0, 6);
    matrix.addTile(block_d, 2, 2);
    return matrix;
  };
  const arranged_matrix matrix {mkmatrix()};
  // We extract into a matrix filled with non-zero entries so we can tell
  // whether the visitor was passed over zero-blocks, which it shouldn't.
  const full_matrix qmarks {
    {"?", "?", "?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?", "?", "?"},
  };
  const full_matrix trans_qmarks {
    {"?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?"},
  };
  assert(matrix.rows() == qmarks.rows());
  assert(matrix.cols() == qmarks.cols());
  {
    const full_matrix expected {
      {  "?",   "?",   "?",   "?",   "?",   "?", "c11"},
      {  "?", "a11", "a12",   "?",   "?",   "?", "c21"},
      {  "?", "a21", "a22",   "?",   "?",   "?",   "?"},
      {  "?", "b11", "b12", "b13",   "?",   "?",   "?"},
      {  "?", "b21", "b22", "b23",   "?",   "?",   "?"},
    };
    full_matrix actual {qmarks};
    extractor pv {actual};
    matrix.passOverRegion(0, 0, false, pv, 0, 0, matrix.rows(), matrix.cols());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    std::cout << "\n\nexpected:\n\n" << expected;
    std::cout << "\n\nactual:\n\n" << actual;
  }
  {
    const full_matrix expected = {
      {  "?",   "?",   "?",   "?",   "?"},
      {  "?", "a11", "a21", "b11", "b21"},
      {  "?", "a12", "a22", "b12", "b22"},
      {  "?",   "?",   "?", "b13", "b23"},
      {  "?",   "?",   "?",   "?",   "?"},
      {  "?",   "?",   "?",   "?",   "?"},
      {"c11", "c21",   "?",   "?",   "?"},
    };
    full_matrix actual {trans_qmarks};
    extractor pv {actual};
    matrix.passOverRegion(0, 0, true, pv, 0, 0, matrix.cols(), matrix.rows());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    std::cout << "\n\nexpected:\n\n" << expected;
    std::cout << "\n\nactual:\n\n" << actual;
  }
  {
    const full_matrix expected {
      {  "?",   "?",   "?",   "?",   "?",   "?",   "?"},
      {  "?",   "?",   "?",   "?",   "?",   "?",   "?"},
      {  "?",   "?",   "?",   "?",   "?",   "?",   "?"},
      {  "?", "a11", "a12",   "?",   "?",   "?", "c21"},
      {  "?",   "?",   "?",   "?",   "?",   "?",   "?"},
    };
    full_matrix actual {qmarks};
    extractor pv {actual};
    matrix.passOverRegion(3, 0, false, pv, 1, 0, 2, 7);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    std::cout << "\n\nexpected:\n\n" << expected;
    std::cout << "\n\nactual:\n\n" << actual;
  }
  {
    const full_matrix expected = {
      {  "?",   "?",   "?",   "?",   "?"},
      {  "?",   "?", "a12",   "?",   "?"},
      {  "?",   "?",   "?",   "?",   "?"},
      {  "?",   "?",   "?",   "?",   "?"},
      {  "?",   "?",   "?",   "?",   "?"},
      {  "?",   "?", "c21",   "?",   "?"},
      {  "?",   "?",   "?",   "?",   "?"},
    };
    full_matrix actual {trans_qmarks};
    extractor pv {actual};
    matrix.passOverRegion(1, 2, true, pv, 2, 1, 7, 2);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    std::cout << "\n\nexpected:\n\n" << expected;
    std::cout << "\n\nactual:\n\n" << actual;
  }
  {
    const full_matrix expected {qmarks};
    full_matrix actual {qmarks};
    extractor pv {actual};
    matrix.passOverRegion(1, 1, false, pv, 2, 3, 2, 5);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const arranged_matrix empty {};
    const full_matrix expected {};
    full_matrix actual {};
    extractor pv {actual};
    matrix.passOverRegion(0, 0, false, pv, 0, 0, 0, 0);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
}


/**
 * @brief
 *         Unit test for the active `passOverRegion` member function.
 *
 * This test checks that:
 *
 * - the correct region is shown to the visitor
 * - zero-blocks are not shown to the visitor
 * - transposition is applied correctly
 * - each element is visited exactly once
 * - the visited matrix and the visitor may mutate
 *
 */
BOOST_AUTO_TEST_CASE(pass_over_region__active)
{
  typedef std::string value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  using inserter = bsc_klammler::test::Inserter<value_type>;
  // Assemble the matrix in a function so we can make it 'const' and don't need
  // to carry the building blocks around in our continued scope.
  const auto mkmatrix = []()->arranged_matrix{
    arranged_matrix matrix {5, 7};
    matrix.addTile(full_matrix {2, 2}, 1, 1);
    matrix.addTile(full_matrix {2, 3}, 3, 1);
    matrix.addTile(full_matrix {2, 1}, 0, 6);
    matrix.addTile(full_matrix {0, 0}, 2, 2);
    return matrix;
  };
  const arranged_matrix matrix {mkmatrix()};
  const full_matrix data = {
    {  "?",   "?",   "?",   "?",   "?",   "?", "c11"},
    {  "?", "a11", "a12",   "?",   "?",   "?", "c21"},
    {  "?", "a21", "a22",   "?",   "?",   "?",   "?"},
    {  "?", "b11", "b12", "b13",   "?",   "?",   "?"},
    {  "?", "b21", "b22", "b23",   "?",   "?",   "?"},
  };
  const full_matrix trans_data = {
    {  "?",   "?",   "?",   "?",   "?"},
    {  "?", "a11", "a21", "b11", "b21"},
    {  "?", "a12", "a22", "b12", "b22"},
    {  "?",   "?",   "?", "b13", "b23"},
    {  "?",   "?",   "?",   "?",   "?"},
    {  "?",   "?",   "?",   "?",   "?"},
    {"c11", "c21",   "?",   "?",   "?"},
  };
  assert(matrix.rows() == data.rows());
  assert(matrix.cols() == data.cols());
  // If the visitor is passed over a zero-block, the attempt to 'set()' values
  // therein will trigger an exception.
  {
    const full_matrix expected {
      {   "",    "",    "",    "",    "",    "", "c11"},
      {   "", "a11", "a12",    "",    "",    "", "c21"},
      {   "", "a21", "a22",    "",    "",    "",    ""},
      {   "", "b11", "b12", "b13",    "",    "",    ""},
      {   "", "b21", "b22", "b23",    "",    "",    ""},
    };
    arranged_matrix actual {matrix};
    inserter av {data};
    actual.passOverRegion(0, 0, false, av, 0, 0, matrix.rows(), matrix.cols());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    std::cout << "\n\nexpected:\n\n" << expected;
    std::cout << "\n\nactual:\n\n" << actual;
  }
  {
    const full_matrix expected = {
      {   "",    "",    "",    "",    "",    "", "c11"},
      {   "", "a11", "a12",    "",    "",    "", "c21"},
      {   "", "a21", "a22",    "",    "",    "",    ""},
      {   "", "b11", "b12", "b13",    "",    "",    ""},
      {   "", "b21", "b22", "b23",    "",    "",    ""},
    };
    arranged_matrix actual {matrix};
    inserter av {trans_data};
    actual.passOverRegion(0, 0, true, av, 0, 0, matrix.cols(), matrix.rows());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    std::cout << "\n\nexpected:\n\n" << expected;
    std::cout << "\n\nactual:\n\n" << actual;
  }
  {
    const full_matrix expected = {
      {   "",    "",    "",    "",    "",    "",    ""},
      {   "",    "", "a11",    "",    "",    "",    ""},
      {   "",    "", "a21",    "",    "",    "",    ""},
      {   "",    "", "b11",    "",    "",    "",    ""},
      {   "",    "",    "",    "",    "",    "",    ""},
    };
    arranged_matrix actual {matrix};
    inserter av {data};
    actual.passOverRegion(1, 1, false, av, 1, 2, 4, 3);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    std::cout << "\n\nexpected:\n\n" << expected;
    std::cout << "\n\nactual:\n\n" << actual;
  }
  {
    const full_matrix expected = {
      {   "",    "",    "",    "",    "",    "",    ""},
      {   "",    "",    "",    "",    "",    "",    ""},
      {   "",    "", "c11",    "",    "",    "",    ""},
      {   "",    "", "c21",    "",    "",    "",    ""},
      {   "",    "",    "",    "",    "",    "",    ""},
    };
    arranged_matrix actual {matrix};
    inserter av {trans_data};
    actual.passOverRegion(6, 0, true, av, 2, 2, 3, 4);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
    std::cout << "\n\nexpected:\n\n" << expected;
    std::cout << "\n\nactual:\n\n" << actual;
  }
  {
    const full_matrix expected {data.rows(), data.cols()};
    arranged_matrix actual {matrix};
    inserter av {data};
    actual.passOverRegion(2, 3, false, av, 2, 3, 2, 3);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix empty {};
    const full_matrix expected {};
    arranged_matrix actual {};
    inserter av {empty};
    actual.passOverRegion(0, 0, false, av, 0, 0, 0, 0);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
}
