/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_block_matrix.cpp
 *
 * @brief
 *         Unit tests for the `BlockMatrix` template class.
 *
 */

#include "../global.hpp"        // (package configuration)
#include "../block_matrix.hpp"  // bsc_klammler::BlockMatrix

#include <array>                // std::array
#include <cassert>              // assert
#include <memory>               // std::unique_ptr
#include <stdexcept>            // std::invalid_argument, std::runtime_error
#include <string>               // std::string
#include <tuple>                // std::tuple
#include <utility>              // std::move

#include "../full_matrix.hpp"   // bsc_klammler::FullMatrix
#include "../random.hpp"        // bsc_klammler::random_real
#include "foo.hpp"              // bsc_klammler::test::Foo
#include "mock_h_matrix.hpp"    // bsc_klammler::test::MockHMatrix
#include "util.hpp"             // MY_CHECK_MATRIX_EQUAL
#include "visitors.hpp"         // bsc_klammler::test::Extractor, ...

#define BOOST_TEST_MODULE block_matrix
#include "module.hpp"

using bsc_klammler::random_int;
using bsc_klammler::random_real;
using bsc_klammler::test::Foo;


/**
 * @brief
 *         Unit test for the constructor.
 *
 * This test checks that:
 *
 * - the constructor and the overload taking references work in the most
 *   trivial case
 * - the overload taking references clones the objects
 * - the overload taking pointers does not clone the objcts and takes ownership
 *
 */
BOOST_AUTO_TEST_CASE(ctor__1)
{
  // This is the simple test matrix we use:
  //
  //   0     3       7    9
  //   +-----+-------+----+
  // 0 |a a a         x x |
  //   |a a a         x x |
  //   |a a a         x x |
  // 3 +      b b b b y y +
  //   |      b b b b y y |
  //   |      b b b b y y |
  //   |      b b b b y y |
  // 7 +u u u v v v v c c +
  //   |u u u v v v v c c |
  // 9 +------+-------+---+
  //
  typedef char value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using block_matrix = bsc_klammler::BlockMatrix<value_type>;
  using index_type = h_matrix::index_type;
  std::unique_ptr<block_matrix> matrix1_uptr {};
  std::unique_ptr<block_matrix> matrix2_uptr {};
  {
    const full_matrix block_a {
      {'a', 'a', 'a'},
      {'a', 'a', 'a'},
      {'a', 'a', 'a'},
    };
    const full_matrix block_b {
      {'b', 'b', 'b', 'b'},
      {'b', 'b', 'b', 'b'},
      {'b', 'b', 'b', 'b'},
      {'b', 'b', 'b', 'b'},
    };
    const full_matrix block_c {
      {'c', 'c'},
      {'c', 'c'},
    };
    const full_matrix block_u {
      {'u', 'u', 'u'},
      {'u', 'u', 'u'},
    };
    const full_matrix block_v {
      {'v', 'v', 'v', 'v'},
      {'v', 'v', 'v', 'v'},
    };
    const full_matrix block_x {
      {'x', 'x'},
      {'x', 'x'},
      {'x', 'x'},
    };
    const full_matrix block_y {
      {'y', 'y'},
      {'y', 'y'},
      {'y', 'y'},
      {'y', 'y'},
    };
    // This is terrible code, but we don't want beautiful exception safe unit
    // test but rather check the library code thoroughly.
    h_matrix *const block_a_ptr {block_a.clone().release()};
    h_matrix *const block_b_ptr {block_b.clone().release()};
    h_matrix *const block_c_ptr {block_c.clone().release()};
    h_matrix *const block_u_ptr {block_u.clone().release()};
    h_matrix *const block_v_ptr {block_v.clone().release()};
    h_matrix *const block_x_ptr {block_x.clone().release()};
    h_matrix *const block_y_ptr {block_y.clone().release()};
    matrix1_uptr.reset(new block_matrix {
        block_a,
        block_b,
        block_c,
        block_u,
        block_v,
        block_x,
        block_y
    });
    matrix2_uptr.reset(new block_matrix {
        std::unique_ptr<h_matrix> {block_a_ptr},
        std::unique_ptr<h_matrix> {block_b_ptr},
        std::unique_ptr<h_matrix> {block_c_ptr},
        std::unique_ptr<h_matrix> {block_u_ptr},
        std::unique_ptr<h_matrix> {block_v_ptr},
        std::unique_ptr<h_matrix> {block_x_ptr},
        std::unique_ptr<h_matrix> {block_y_ptr}
    });
    BOOST_CHECK(matrix1_uptr->getBlockA() != &block_a);
    BOOST_CHECK(matrix1_uptr->getBlockB() != &block_b);
    BOOST_CHECK(matrix1_uptr->getBlockC() != &block_c);
    BOOST_CHECK(matrix1_uptr->getBlockU() != &block_u);
    BOOST_CHECK(matrix1_uptr->getBlockV() != &block_v);
    BOOST_CHECK(matrix1_uptr->getBlockX() != &block_x);
    BOOST_CHECK(matrix1_uptr->getBlockY() != &block_y);
    BOOST_CHECK(matrix2_uptr->getBlockA() == block_a_ptr);
    BOOST_CHECK(matrix2_uptr->getBlockB() == block_b_ptr);
    BOOST_CHECK(matrix2_uptr->getBlockC() == block_c_ptr);
    BOOST_CHECK(matrix2_uptr->getBlockU() == block_u_ptr);
    BOOST_CHECK(matrix2_uptr->getBlockV() == block_v_ptr);
    BOOST_CHECK(matrix2_uptr->getBlockX() == block_x_ptr);
    BOOST_CHECK(matrix2_uptr->getBlockY() == block_y_ptr);
    // The building blocks go out of scope here.
  }
  BOOST_CHECK_EQUAL(9, matrix1_uptr->rows());
  BOOST_CHECK_EQUAL(9, matrix2_uptr->rows());
  BOOST_CHECK_EQUAL(9, matrix1_uptr->cols());
  BOOST_CHECK_EQUAL(9, matrix2_uptr->cols());
  for (index_type j = 0; j < 3; ++j)
    {
      for (index_type i = 0; i < 3; ++i)
        {
          BOOST_CHECK_EQUAL('a', matrix1_uptr->get(i, j));
          BOOST_CHECK_EQUAL('a', matrix2_uptr->get(i, j));
        }
      for (index_type i = 3; i < 7; ++i)
        {
          BOOST_CHECK_EQUAL(0, matrix1_uptr->get(i, j));
          BOOST_CHECK_EQUAL(0, matrix2_uptr->get(i, j));
        }
      for (index_type i = 7; i < 9; ++i)
        {
          BOOST_CHECK_EQUAL('u', matrix1_uptr->get(i, j));
          BOOST_CHECK_EQUAL('u', matrix2_uptr->get(i, j));
        }
    }
  for (index_type j = 3; j < 7; ++j)
    {
      for (index_type i = 0; i < 3; ++i)
        {
          BOOST_CHECK_EQUAL(0, matrix1_uptr->get(i, j));
          BOOST_CHECK_EQUAL(0, matrix2_uptr->get(i, j));
        }
      for (index_type i = 3; i < 7; ++i)
        {
          BOOST_CHECK_EQUAL('b', matrix1_uptr->get(i, j));
          BOOST_CHECK_EQUAL('b', matrix2_uptr->get(i, j));
        }
      for (index_type i = 7; i < 9; ++i)
        {
          BOOST_CHECK_EQUAL('v', matrix1_uptr->get(i, j));
          BOOST_CHECK_EQUAL('v', matrix2_uptr->get(i, j));
        }
    }
  for (index_type j = 7; j < 9; ++j)
    {
      for (index_type i = 0; i < 3; ++i)
        {
          BOOST_CHECK_EQUAL('x', matrix1_uptr->get(i, j));
          BOOST_CHECK_EQUAL('x', matrix2_uptr->get(i, j));
        }
      for (index_type i = 3; i < 7; ++i)
        {
          BOOST_CHECK_EQUAL('y', matrix1_uptr->get(i, j));
          BOOST_CHECK_EQUAL('y', matrix2_uptr->get(i, j));
        }
      for (index_type i = 7; i < 9; ++i)
        {
          BOOST_CHECK_EQUAL('c', matrix1_uptr->get(i, j));
          BOOST_CHECK_EQUAL('c', matrix2_uptr->get(i, j));
        }
    }
}


/**
 * @brief
 *         Unit test for the constructor.
 *
 * This test checks that:
 *
 * - the constructed `BlockMatrix` contains the original pointers at the
 *   correct positions and reads and writes through
 *
 */
BOOST_AUTO_TEST_CASE(ctor__2)
{
  using h_matrix = bsc_klammler::HMatrix<double>;
  using full_matrix = bsc_klammler::FullMatrix<double>;
  using block_matrix = bsc_klammler::BlockMatrix<double>;
  using index_type = full_matrix::index_type;
  using param_tuple = std::tuple<full_matrix *, index_type, index_type>;
  const index_type n_a {random_int<index_type>(1, 32)};
  const index_type n_b {random_int<index_type>(1, 32)};
  const index_type n_c {random_int<index_type>(1, 32)};
  // Preapare an array of tuples that store at position:
  //
  // 0. a pointer to the matrix block
  // 1. the row index of north-west corner inside the block matrix
  // 2. the column index of north-west corner inside the block matrix
  //
  // The order inside the array is -- as always -- A, B, C, U, V, X, Y.
  std::array<param_tuple, 7> blocks = {
    param_tuple {new full_matrix {n_a, n_a}, 0,         0},          // A
    param_tuple {new full_matrix {n_b, n_b}, n_a,       n_a},        // B
    param_tuple {new full_matrix {n_c, n_c}, n_a + n_b, n_a + n_b},  // C
    param_tuple {new full_matrix {n_c, n_a}, n_a + n_b, 0},          // U
    param_tuple {new full_matrix {n_c, n_b}, n_a + n_b, n_a},        // V
    param_tuple {new full_matrix {n_a, n_c}, 0,         n_a + n_b},  // X
    param_tuple {new full_matrix {n_b, n_c}, n_a,       n_a + n_b},  // Y
  };
  block_matrix matrix {
    std::unique_ptr<h_matrix> {std::get<0>(blocks.at(0))},
    std::unique_ptr<h_matrix> {std::get<0>(blocks.at(1))},
    std::unique_ptr<h_matrix> {std::get<0>(blocks.at(2))},
    std::unique_ptr<h_matrix> {std::get<0>(blocks.at(3))},
    std::unique_ptr<h_matrix> {std::get<0>(blocks.at(4))},
    std::unique_ptr<h_matrix> {std::get<0>(blocks.at(5))},
    std::unique_ptr<h_matrix> {std::get<0>(blocks.at(6))}
  };
  for (const auto& iter : blocks)
    {
      const index_type n {std::get<0>(iter)->rows()};
      const index_type m {std::get<0>(iter)->cols()};
      const index_type i0 {std::get<1>(iter)};
      const index_type j0 {std::get<2>(iter)};
      // Write into the matrix block, read from the block matrix and vice
      // versa.
      for (index_type j = 0; j < m; ++j)
        {
          for (index_type i = 0; i < n; ++i)
            {
              const double val1 {random_real<double>()};
              const double val2 {random_real<double>()};
              std::get<0>(iter)->set(val1, i, j);
              BOOST_CHECK_EQUAL(val1, std::get<0>(iter)->get(i, j));
              BOOST_CHECK_EQUAL(val1, matrix.get(i0 + i, j0 + j));
              matrix.set(val2, i0 + i, j0 + j);
              BOOST_CHECK_EQUAL(val2, matrix.get(i0 + i, j0 + j));
              BOOST_CHECK_EQUAL(val2, std::get<0>(iter)->get(i, j));
            }
        }
      // Check that the rest of the matrix is unaffected.
      for (index_type j = 0; j < n_a + n_b + n_c; ++j)
        {
          for (index_type i = 0; i < n_a + n_b + n_c; ++i)
            {
              if ((j >= j0 && j < j0 + m) && (i >= i0 && i < i0 + n))
                {
                  // inside block
                  continue;
                }
              BOOST_CHECK_EQUAL(0.0, matrix.get(i, j));
            }
        }
      // Zero-out the matrix block again.  We don't do this earlier (though it
      // would conveniently fit into one of the other loops) because we want to
      // spot the case that a write accidentially modifies more than one cell.
      // If we zero-out too eraly, we might hide this bug.
      for (index_type j = 0; j < m; ++j)
        {
          for (index_type i = 0; i < n; ++i)
            {
              std::get<0>(iter)->set(0.0, i, j);
            }
        }
    }
  // Don't delete the pointers to the matrix blocks.  They are now exclusively
  // owned by 'matrix' which must destroy them as it goes out of scope.
  // Valgrind can check this.
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the constructor.
 *
 * This test checks that:
 *
 * - `nullptr`s are rejected
 * - matrices of incompatible dimensions are rejected
 * - no memory is leaked if the arguments are rejected
 *
 */
BOOST_AUTO_TEST_CASE(ctor__invalid)
{
  // TODO: This test is so terribly ugly!  Write a cleaner one.
  typedef double value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using block_matrix = bsc_klammler::BlockMatrix<value_type>;
  using index_type = bsc_klammler::HMatrix<value_type>::index_type;
  const index_type n_a {random_int<index_type>(2, 32)};
  const index_type n_b {random_int<index_type>(2, 32)};
  const index_type n_c {random_int<index_type>(2, 32)};
  for (int bad_dim = 0; bad_dim < 13; ++bad_dim)
    {
      const signed deltas[] = {-1, 0, +1};  // 'delta = 0' means pass 'nullptr'
      for (const auto delta : deltas)
        {
          index_type dims[7][2];
          full_matrix * blocks[7];
          dims[0][0] = n_a; dims[0][1] = n_a;  // A
          dims[1][0] = n_b; dims[1][1] = n_b;  // B
          dims[2][0] = n_c; dims[2][1] = n_c;  // C
          dims[3][0] = n_c; dims[3][1] = n_a;  // U
          dims[4][0] = n_c; dims[4][1] = n_b;  // V
          dims[5][0] = n_a; dims[5][1] = n_c;  // X
          dims[6][0] = n_b; dims[6][1] = n_c;  // Y
          dims[bad_dim / 2][bad_dim % 2] += delta;
          for (int i = 0; i < 7; ++i)
            {
              blocks[i] = nullptr;
              if (delta || bad_dim % 7 != i)
                {
                  blocks[i] = new full_matrix {dims[i][0], dims[i][1]};
                }
            }
          if (delta)
            {
              // If 'delta == 0', one block is a 'nullptr' so we must not
              // dereference it.
              try
                {
                  block_matrix matrix {
                    *blocks[0],
                    *blocks[1],
                    *blocks[2],
                    *blocks[3],
                    *blocks[4],
                    *blocks[5],
                    *blocks[6],
                      };
                  BOOST_ERROR("no exception thrown");
                }
              catch (std::invalid_argument&) { /* good */ }
            }
          try
            {
              block_matrix matrix {
                std::unique_ptr<h_matrix> {blocks[0]},
                std::unique_ptr<h_matrix> {blocks[1]},
                std::unique_ptr<h_matrix> {blocks[2]},
                std::unique_ptr<h_matrix> {blocks[3]},
                std::unique_ptr<h_matrix> {blocks[4]},
                std::unique_ptr<h_matrix> {blocks[5]},
                std::unique_ptr<h_matrix> {blocks[6]},
              };
              BOOST_ERROR("no exception thrown");
            }
          catch (std::invalid_argument&) { /* good */ }
        }
    }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the copy constructor.
 *
 * This test checks that:
 *
 * - the constructor works
 * - created copies are deep
 *
 */
BOOST_AUTO_TEST_CASE(ctor_copy)
{
  using full_matrix = bsc_klammler::FullMatrix<double>;
  using block_matrix = bsc_klammler::BlockMatrix<double>;
  using index_type = bsc_klammler::HMatrix<double>::index_type;
  const auto random_block = [](const index_type rows,
                               const index_type cols)->full_matrix{
    full_matrix matrix {rows, cols};
    for (index_type j = 0; j < matrix.cols(); ++j)
      {
        for (index_type i = 0; i < matrix.rows(); ++i)
          {
            const double value {random_real<double>()};
            matrix.set(value, i, j);
          }
      }
    return matrix;
  };
  const index_type n_a {random_int<index_type>(1, 32)};
  const index_type n_b {random_int<index_type>(1, 32)};
  const index_type n_c {random_int<index_type>(1, 32)};
  const full_matrix block_a {random_block(n_a, n_a)};
  const full_matrix block_b {random_block(n_b, n_b)};
  const full_matrix block_c {random_block(n_c, n_c)};
  const full_matrix block_u {random_block(n_c, n_a)};
  const full_matrix block_v {random_block(n_c, n_b)};
  const full_matrix block_x {random_block(n_a, n_c)};
  const full_matrix block_y {random_block(n_b, n_c)};
  const block_matrix matrix1 {
    block_a,
    block_b,
    block_c,
    block_u,
    block_v,
    block_x,
    block_y
  };
  block_matrix matrix2 {matrix1};
  BOOST_CHECK_EQUAL(matrix1.rows(), matrix2.rows());
  BOOST_CHECK_EQUAL(matrix1.cols(), matrix2.cols());
  for (index_type j = 0; j < matrix1.cols(); ++j)
    {
      for (index_type i = 0; i < matrix1.rows(); ++i)
        {
          BOOST_CHECK_EQUAL(matrix1.get(i, j), matrix2.get(i, j));
        }
    }
  for (index_type j = 0; j < n_a; ++j)
    {
      for (index_type i = 0; i < n_a; ++i)
        {
          matrix2.set(-1.0, i, j);
          BOOST_CHECK_EQUAL(block_a.get(i, j), matrix1.get(i, j));
        }
      for (index_type i = 0; i < n_c; ++i)
        {
          matrix2.set(-1.0, n_a + n_b + i, j);
          BOOST_CHECK_EQUAL(block_u.get(i, j), matrix1.get(n_a + n_b + i, j));
        }
    }
  for (index_type j = 0; j < n_b; ++j)
    {
      for (index_type i = 0; i < n_b; ++i)
        {
          matrix2.set(-1.0, n_a + i, n_a + j);
          BOOST_CHECK_EQUAL(block_b.get(i, j), matrix1.get(n_a + i, n_a + j));
        }
      for (index_type i = 0; i < n_c; ++i)
        {
          matrix2.set(-1.0, n_a + n_b + i, n_a + j);
          BOOST_CHECK_EQUAL(block_v.get(i, j),
                            matrix1.get(n_a + n_b + i, n_a + j));
        }
    }
  for (index_type j = 0; j < n_c; ++j)
    {
      for (index_type i = 0; i < n_a; ++i)
        {
          matrix2.set(-1.0, i, n_a + n_b + j);
          BOOST_CHECK_EQUAL(block_x.get(i, j), matrix1.get(i, n_a + n_b + j));
        }
      for (index_type i = 0; i < n_b; ++i)
        {
          matrix2.set(-1.0, n_a + i, n_a + n_b + j);
          BOOST_CHECK_EQUAL(block_y.get(i, j),
                            matrix1.get(n_a + i, n_a + n_b + j));
        }
      for (index_type i = 0; i < n_c; ++i)
        {
          matrix2.set(-1.0, n_a + n_b + i, n_a + n_b + j);
          BOOST_CHECK_EQUAL(block_c.get(i, j),
                            matrix1.get(n_a + n_b + i, n_a + n_b + j));
        }
    }
}


/**
 * @brief
 *         Unit test for the move constructor.
 *
 * This test checks that:
 *
 * - the constructor works
 * - no intermediate copies are created
 * - no element-wise move is performed
 *
 */
BOOST_AUTO_TEST_CASE(ctor_move)
{
  Foo::activate();
  const bsc_klammler::FullMatrix<Foo> block_a {{'a'}};
  const bsc_klammler::FullMatrix<Foo> block_b {{'b'}};
  const bsc_klammler::FullMatrix<Foo> block_c {{'c'}};
  const bsc_klammler::FullMatrix<Foo> block_u {{'u'}};
  const bsc_klammler::FullMatrix<Foo> block_v {{'v'}};
  const bsc_klammler::FullMatrix<Foo> block_x {{'x'}};
  const bsc_klammler::FullMatrix<Foo> block_y {{'y'}};
  bsc_klammler::BlockMatrix<Foo> matrix1 {
    block_a, block_b, block_c, block_u, block_v, block_x, block_y
  };
  Foo::deactivate();
  bsc_klammler::BlockMatrix<Foo> matrix2 {std::move(matrix1)};
  Foo::activate();
  BOOST_CHECK_EQUAL(0, matrix1.rows());
  BOOST_CHECK_EQUAL(0, matrix1.cols());
  BOOST_CHECK_EQUAL(3, matrix2.rows());
  BOOST_CHECK_EQUAL(3, matrix2.cols());
  BOOST_CHECK_EQUAL('a', matrix2.get(0, 0).value());
  BOOST_CHECK_EQUAL(0,   matrix2.get(1, 0).value());
  BOOST_CHECK_EQUAL('u', matrix2.get(2, 0).value());
  BOOST_CHECK_EQUAL(0,   matrix2.get(0, 1).value());
  BOOST_CHECK_EQUAL('b', matrix2.get(1, 1).value());
  BOOST_CHECK_EQUAL('v', matrix2.get(2, 1).value());
  BOOST_CHECK_EQUAL('x', matrix2.get(0, 2).value());
  BOOST_CHECK_EQUAL('y', matrix2.get(1, 2).value());
  BOOST_CHECK_EQUAL('c', matrix2.get(2, 2).value());
}


/**
 * @brief
 *         Unit test for the copy-assignment operator.
 *
 * This test checks that:
 *
 * - the copy-assignment operator works
 * - created copies are deep
 *
 */
BOOST_AUTO_TEST_CASE(assign_copy)
{
  {
    const bsc_klammler::BlockMatrix<int> matrix1 {};
    bsc_klammler::BlockMatrix<int> matrix2 {};
    matrix2 = matrix1;
    BOOST_CHECK_EQUAL(0, matrix1.rows());
    BOOST_CHECK_EQUAL(0, matrix2.rows());
    BOOST_CHECK_EQUAL(0, matrix1.cols());
    BOOST_CHECK_EQUAL(0, matrix2.cols());
  }
  {
    using full_matrix = bsc_klammler::FullMatrix<char>;
    using block_matrix = bsc_klammler::BlockMatrix<char>;
    const full_matrix block_a {{'a'}};
    const full_matrix block_b {{'b'}};
    const full_matrix block_c {{'c'}};
    const full_matrix block_u {{'u'}};
    const full_matrix block_v {{'v'}};
    const full_matrix block_x {{'x'}};
    const full_matrix block_y {{'y'}};
    const block_matrix matrix1 {
      block_a, block_b, block_c, block_u, block_v, block_x, block_y
    };
    block_matrix matrix2 {
      block_x, block_x, block_x, block_x, block_x, block_x, block_x
    };
    matrix2 = matrix1;
    BOOST_CHECK_EQUAL(matrix1.rows(), matrix2.rows());
    BOOST_CHECK_EQUAL(matrix1.cols(), matrix2.cols());
    matrix2.set('A', 0, 0);
    matrix2.set('B', 1, 1);
    matrix2.set('C', 2, 2);
    matrix2.set('U', 2, 0);
    matrix2.set('V', 2, 1);
    matrix2.set('X', 0, 2);
    matrix2.set('Y', 1, 2);
    BOOST_CHECK_EQUAL('a', matrix1.get(0, 0));
    BOOST_CHECK_EQUAL('A', matrix2.get(0, 0));
    BOOST_CHECK_EQUAL(0,   matrix1.get(1, 0));
    BOOST_CHECK_EQUAL(0,   matrix2.get(1, 0));
    BOOST_CHECK_EQUAL('u', matrix1.get(2, 0));
    BOOST_CHECK_EQUAL('U', matrix2.get(2, 0));
    BOOST_CHECK_EQUAL(0,   matrix1.get(0, 1));
    BOOST_CHECK_EQUAL(0,   matrix2.get(0, 1));
    BOOST_CHECK_EQUAL('b', matrix1.get(1, 1));
    BOOST_CHECK_EQUAL('B', matrix2.get(1, 1));
    BOOST_CHECK_EQUAL('v', matrix1.get(2, 1));
    BOOST_CHECK_EQUAL('V', matrix2.get(2, 1));
    BOOST_CHECK_EQUAL('x', matrix1.get(0, 2));
    BOOST_CHECK_EQUAL('X', matrix2.get(0, 2));
    BOOST_CHECK_EQUAL('y', matrix1.get(1, 2));
    BOOST_CHECK_EQUAL('Y', matrix2.get(1, 2));
    BOOST_CHECK_EQUAL('c', matrix1.get(2, 2));
    BOOST_CHECK_EQUAL('C', matrix2.get(2, 2));
  }
}


/**
 * @brief
 *         Unit test for the copy-assignment operator's exception safety.
 *
 * If run through *Valgrind*, this test checks that:
 *
 * - the constructor performs correctly even in case an exception is thrown
 *   while copying the matrix blocks
 *
 */
BOOST_AUTO_TEST_CASE(assign_copy__except)
{
  typedef int value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using block_matrix = bsc_klammler::BlockMatrix<value_type>;
  using mockup = bsc_klammler::test::MockHMatrix<value_type>;
  using index_type = h_matrix::index_type;
  class MyFailure : public std::exception {};
  std::array<mockup *, 7> blocks;  // use of raw pointers is intentional
  for (std::size_t i = 0; i < blocks.size(); ++i)
    {
      mockup * mock_ptr = new mockup {};
      mock_ptr->mockRows([]()->index_type{ return 0; });
      mock_ptr->mockCols([]()->index_type{ return 0; });
      if (i == 5)
        {
          mock_ptr->mockClone([]()->std::unique_ptr<h_matrix>{
              throw MyFailure {};
            });
        }
      else
        {
          mock_ptr->mockClone([]()->std::unique_ptr<h_matrix>{
              return std::unique_ptr<h_matrix> {new full_matrix {}};
            });
        }
      blocks.at(i) = mock_ptr;
    }
  const block_matrix matrix1 {
    std::unique_ptr<h_matrix> {blocks.at(0)},
    std::unique_ptr<h_matrix> {blocks.at(1)},
    std::unique_ptr<h_matrix> {blocks.at(2)},
    std::unique_ptr<h_matrix> {blocks.at(3)},
    std::unique_ptr<h_matrix> {blocks.at(4)},
    std::unique_ptr<h_matrix> {blocks.at(5)},
    std::unique_ptr<h_matrix> {blocks.at(6)}
  };
  try
    {
      block_matrix matrix2 {};
      matrix2 = matrix1;
      BOOST_ERROR("no exception was thrown");
    }
  catch (const MyFailure&) { /* okay */ }
  BOOST_WARN(RUNNING_ON_VALGRIND);
}


/**
 * @brief
 *         Unit test for the move-assignment operator.
 *
 * This test checks that:
 *
 * - the move-assignment operator works
 *
 */
BOOST_AUTO_TEST_CASE(assign_move)
{
  using full_matrix = bsc_klammler::FullMatrix<int>;
  using block_matrix = bsc_klammler::BlockMatrix<int>;
  {
    block_matrix matrix1 {};
    block_matrix matrix2 {
      full_matrix {{1}},
      full_matrix {{2}},
      full_matrix {{3}},
      full_matrix {{4}},
      full_matrix {{5}},
      full_matrix {{6}},
      full_matrix {{7}},
    };
    matrix2 = std::move(matrix1);
    BOOST_CHECK_EQUAL(0, matrix1.rows());
    BOOST_CHECK_EQUAL(0, matrix1.cols());
    BOOST_CHECK_EQUAL(0, matrix2.rows());
    BOOST_CHECK_EQUAL(0, matrix2.cols());
  }
  {
    block_matrix matrix1 {
      full_matrix {{'a'}},
      full_matrix {{'b'}},
      full_matrix {{'c'}},
      full_matrix {{'u'}},
      full_matrix {{'v'}},
      full_matrix {{'x'}},
      full_matrix {{'y'}},
    };
    block_matrix matrix2 {};
    matrix2 = std::move(matrix1);
    BOOST_CHECK_EQUAL(0, matrix1.rows());
    BOOST_CHECK_EQUAL(0, matrix1.cols());
    BOOST_CHECK_EQUAL(3, matrix2.rows());
    BOOST_CHECK_EQUAL(3, matrix2.cols());
    BOOST_CHECK_EQUAL('a', matrix2.get(0, 0));
    BOOST_CHECK_EQUAL(0,   matrix2.get(1, 0));
    BOOST_CHECK_EQUAL('u', matrix2.get(2, 0));
    BOOST_CHECK_EQUAL(0,   matrix2.get(0, 1));
    BOOST_CHECK_EQUAL('b', matrix2.get(1, 1));
    BOOST_CHECK_EQUAL('v', matrix2.get(2, 1));
    BOOST_CHECK_EQUAL('x', matrix2.get(0, 2));
    BOOST_CHECK_EQUAL('y', matrix2.get(1, 2));
    BOOST_CHECK_EQUAL('c', matrix2.get(2, 2));
  }
}


/**
 * @brief
 *         Unit test for the accessor member functions.
 *
 * This test checks that:
 *
 * - `get` and `set` work for recursive matrices
 *
 */
BOOST_AUTO_TEST_CASE(accessors)
{
  using index_type = bsc_klammler::HMatrix<int>::index_type;
  const char expected[6][6] = {
    {'a',  0,  'x',  0,  'X', 'X'},
    { 0,  'b', 'y',  0,  'X', 'X'},
    {'u', 'v', 'c',  0,  'X', 'X'},
    { 0,   0,   0,  'B', 'Y', 'Y'},
    {'U', 'U', 'U', 'V', 'C', 'C'},
    {'U', 'U', 'U', 'V', 'C', 'C'},
  };
  const bsc_klammler::BlockMatrix<int> block_a {
    bsc_klammler::FullMatrix<int> {{'a'}},
    bsc_klammler::FullMatrix<int> {{'b'}},
    bsc_klammler::FullMatrix<int> {{'c'}},
    bsc_klammler::FullMatrix<int> {{'u'}},
    bsc_klammler::FullMatrix<int> {{'v'}},
    bsc_klammler::FullMatrix<int> {{'x'}},
    bsc_klammler::FullMatrix<int> {{'y'}},
  };
  const bsc_klammler::FullMatrix<int> block_b {{'B'}};
  const bsc_klammler::FullMatrix<int> block_c {
    {'C', 'C'},
    {'C', 'C'},
  };
  const bsc_klammler::FullMatrix<int> block_u {
    {'U', 'U', 'U'},
    {'U', 'U', 'U'},
  };
  const bsc_klammler::FullMatrix<int> block_v {
    {'V'},
    {'V'},
  };
  const bsc_klammler::FullMatrix<int> block_x {
    {'X', 'X'},
    {'X', 'X'},
    {'X', 'X'},
  };
  const bsc_klammler::FullMatrix<int> block_y {{'Y', 'Y'}};
  // Combine to new martrix:
  BOOST_CHECK(block_a.rows() == 3 && block_a.cols() == 3);
  BOOST_CHECK(block_b.rows() == 1 && block_b.cols() == 1);
  BOOST_CHECK(block_c.rows() == 2 && block_c.cols() == 2);
  BOOST_CHECK(block_u.rows() == 2 && block_u.cols() == 3);
  BOOST_CHECK(block_v.rows() == 2 && block_v.cols() == 1);
  BOOST_CHECK(block_x.rows() == 3 && block_x.cols() == 2);
  BOOST_CHECK(block_y.rows() == 1 && block_y.cols() == 2);
  bsc_klammler::BlockMatrix<int> matrix {
    block_a, block_b, block_c, block_u, block_v, block_x, block_y
  };
  BOOST_CHECK_EQUAL(6, matrix.rows());
  BOOST_CHECK_EQUAL(6, matrix.cols());
  for (index_type j = 0; j < 6; ++j)
    {
      for (index_type i = 0; i < 6; ++i)
        {
          const char digit {random_int<char>('0', '9')};
          BOOST_CHECK_EQUAL(expected[i][j], matrix.get(i, j));
          if (expected[i][j])
            {
              // not in zero block
              matrix.set(digit, i, j);
              BOOST_CHECK_EQUAL(digit, matrix.get(i, j));
            }
          else if (BSC_KLAMMLER_CHECK_ARGS)
            {
              try
                {
                  matrix.set(digit, i, j);
                  BOOST_ERROR("no exception thrown");
                }
              catch (const std::invalid_argument&) { /* good */ }
            }
        }
    }
}


/**
 * @brief
 *         Unit test for the `swap` friend function.
 *
 * This test checks that:
 *
 * - `get` and `set` behave well after a `swap()`
 * - no elements are copied or moved
 *
 */
BOOST_AUTO_TEST_CASE(swapping)
{
  typedef int value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using block_matrix = bsc_klammler::BlockMatrix<value_type>;
  using mockup = bsc_klammler::test::MockHMatrix<value_type>;
  using index_type = h_matrix::index_type;
  std::array<mockup *, 7> first_blocks;
  std::array<mockup *, 7> second_blocks;
  for (std::size_t i = 0; i < first_blocks.size(); ++i)
    {
      first_blocks.at(i) = new mockup {};
      first_blocks.at(i)->mockRows([]()->index_type{ return 1; });
      first_blocks.at(i)->mockCols([]()->index_type{ return 1; });
    }
  for (std::size_t i = 0; i < second_blocks.size(); ++i)
    {
      second_blocks.at(i) = new mockup {};
      second_blocks.at(i)->mockRows([]()->index_type{ return 2; });
      second_blocks.at(i)->mockCols([]()->index_type{ return 2; });
    }
  block_matrix first {
    std::unique_ptr<h_matrix> {first_blocks.at(0)},
    std::unique_ptr<h_matrix> {first_blocks.at(1)},
    std::unique_ptr<h_matrix> {first_blocks.at(2)},
    std::unique_ptr<h_matrix> {first_blocks.at(3)},
    std::unique_ptr<h_matrix> {first_blocks.at(4)},
    std::unique_ptr<h_matrix> {first_blocks.at(5)},
    std::unique_ptr<h_matrix> {first_blocks.at(6)},
  };
  block_matrix second {
    std::unique_ptr<h_matrix> {second_blocks.at(0)},
    std::unique_ptr<h_matrix> {second_blocks.at(1)},
    std::unique_ptr<h_matrix> {second_blocks.at(2)},
    std::unique_ptr<h_matrix> {second_blocks.at(3)},
    std::unique_ptr<h_matrix> {second_blocks.at(4)},
    std::unique_ptr<h_matrix> {second_blocks.at(5)},
    std::unique_ptr<h_matrix> {second_blocks.at(6)},
  };
  {
    BOOST_REQUIRE_EQUAL(first_blocks.at(0), first.getBlockA());
    BOOST_REQUIRE_EQUAL(first_blocks.at(1), first.getBlockB());
    BOOST_REQUIRE_EQUAL(first_blocks.at(2), first.getBlockC());
    BOOST_REQUIRE_EQUAL(first_blocks.at(3), first.getBlockU());
    BOOST_REQUIRE_EQUAL(first_blocks.at(4), first.getBlockV());
    BOOST_REQUIRE_EQUAL(first_blocks.at(5), first.getBlockX());
    BOOST_REQUIRE_EQUAL(first_blocks.at(6), first.getBlockY());
  }
  {
    BOOST_REQUIRE_EQUAL(second_blocks.at(0), second.getBlockA());
    BOOST_REQUIRE_EQUAL(second_blocks.at(1), second.getBlockB());
    BOOST_REQUIRE_EQUAL(second_blocks.at(2), second.getBlockC());
    BOOST_REQUIRE_EQUAL(second_blocks.at(3), second.getBlockU());
    BOOST_REQUIRE_EQUAL(second_blocks.at(4), second.getBlockV());
    BOOST_REQUIRE_EQUAL(second_blocks.at(5), second.getBlockX());
    BOOST_REQUIRE_EQUAL(second_blocks.at(6), second.getBlockY());
  }
  swap(first, second);
  {
    BOOST_CHECK_EQUAL(second_blocks.at(0), first.getBlockA());
    BOOST_CHECK_EQUAL(second_blocks.at(1), first.getBlockB());
    BOOST_CHECK_EQUAL(second_blocks.at(2), first.getBlockC());
    BOOST_CHECK_EQUAL(second_blocks.at(3), first.getBlockU());
    BOOST_CHECK_EQUAL(second_blocks.at(4), first.getBlockV());
    BOOST_CHECK_EQUAL(second_blocks.at(5), first.getBlockX());
    BOOST_CHECK_EQUAL(second_blocks.at(6), first.getBlockY());
  }
  {
    BOOST_CHECK_EQUAL(first_blocks.at(0), second.getBlockA());
    BOOST_CHECK_EQUAL(first_blocks.at(1), second.getBlockB());
    BOOST_CHECK_EQUAL(first_blocks.at(2), second.getBlockC());
    BOOST_CHECK_EQUAL(first_blocks.at(3), second.getBlockU());
    BOOST_CHECK_EQUAL(first_blocks.at(4), second.getBlockV());
    BOOST_CHECK_EQUAL(first_blocks.at(5), second.getBlockX());
    BOOST_CHECK_EQUAL(first_blocks.at(6), second.getBlockY());
  }
}


/**
 * @brief
 *         Smaller unit tests for the constructors and assignment operators in
 *         case of symmetric matrices.
 *
 * Now we don't have symmetry information stored in the matrix any more, this
 * test has become rather useless, but we keep it as another (and more compact)
 * opportunity to exercise the various constructors and assignment operators.
 *
 * This test checks that:
 *
 * - symmetric and non-symmetric `BlockMatrix`es are constructed correctly
 * - symmetric and non-symmetric `BlockMatrix`es are copied correctly
 * - symmetric and non-symmetric `BlockMatrix`es are moved correctly
 * - a symmetric `BlockMatrix` can be (copy / move) assigned to a non-symmetric
 *   one and vice versa
 * - symmetric and non-symmetric `BlockMatrix`es can be `%swap()`ped
 *
 */
BOOST_AUTO_TEST_CASE(symmetric)
{
  typedef int value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using block_matrix = bsc_klammler::BlockMatrix<value_type>;
  const full_matrix expected_sym {
    {11,  0,  0,  0, 15, 16},
    { 0, 22, 23, 24, 25, 26},
    { 0, 23, 33, 34, 35, 36},
    { 0, 24, 34, 44, 45, 46},
    {15, 25, 35, 45, 55, 56},
    {16, 26, 36, 46, 56, 66},
  };
  const full_matrix expected_non {
    {11,  0,  0,  0, 15, 16},
    { 0, 22, 23, 24, 25, 26},
    { 0, 32, 33, 34, 35, 36},
    { 0, 42, 43, 44, 45, 46},
    {51, 52, 53, 54, 55, 56},
    {61, 62, 63, 64, 65, 66},
  };
#define SYM_BLOCK_MATRIX_INIT                                                 \
    full_matrix {{11}},                                                       \
    full_matrix {{22, 23, 24}, {23, 33, 34}, {24, 34, 44}},                   \
    full_matrix {{55, 56}, {56, 66}},                                         \
    full_matrix {{15}, {16}},                                                 \
    full_matrix {{25, 35, 45}, {26, 36, 46}},                                 \
    full_matrix {{15, 16}},                                                   \
    full_matrix {{25, 26}, {35, 36}, {45, 46}},
#define NON_BLOCK_MATRIX_INIT                                                 \
    full_matrix {{11}},                                                       \
    full_matrix {{22, 23, 24}, {32, 33, 34}, {42, 43, 44}},                   \
    full_matrix {{55, 56}, {65, 66}},                                         \
    full_matrix {{51}, {61}},                                                 \
    full_matrix {{52, 53, 54}, {62, 63, 64}},                                 \
    full_matrix {{15, 16}},                                                   \
    full_matrix {{25, 26}, {35, 36}, {45, 46}},
#define X_SYM_BLOCK_MATRIX_INIT                                               \
    full_matrix {{1}}, full_matrix {{2}}, full_matrix {{3}},                  \
    full_matrix {{4}}, full_matrix {{5}},                                     \
    full_matrix {{4}}, full_matrix {{5}},
#define X_NON_BLOCK_MATRIX_INIT                                               \
    full_matrix {{1}}, full_matrix {{2}}, full_matrix {{3}},                  \
    full_matrix {{4}}, full_matrix {{5}},                                     \
    full_matrix {{6}}, full_matrix {{7}},
  {
    // construction
    const block_matrix actual_sym {SYM_BLOCK_MATRIX_INIT};
    const block_matrix actual_non {NON_BLOCK_MATRIX_INIT};
    MY_CHECK_MATRIX_EQUAL(expected_sym, actual_sym);
    MY_CHECK_MATRIX_EQUAL(expected_non, actual_non);
  }
  {
    // copy construction
    const block_matrix proto_sym {SYM_BLOCK_MATRIX_INIT};
    const block_matrix proto_non {NON_BLOCK_MATRIX_INIT};
    const block_matrix actual_sym {proto_sym};
    const block_matrix actual_non {proto_non};
    MY_CHECK_MATRIX_EQUAL(expected_sym, actual_sym);
    MY_CHECK_MATRIX_EQUAL(expected_non, actual_non);
  }
  {
    // copy assignment
    const block_matrix proto_sym {SYM_BLOCK_MATRIX_INIT};
    const block_matrix proto_non {NON_BLOCK_MATRIX_INIT};
    block_matrix actual_sym_was_sym {X_SYM_BLOCK_MATRIX_INIT};
    block_matrix actual_sym_was_non {X_NON_BLOCK_MATRIX_INIT};
    block_matrix actual_non_was_sym {X_SYM_BLOCK_MATRIX_INIT};
    block_matrix actual_non_was_non {X_NON_BLOCK_MATRIX_INIT};
    actual_sym_was_sym = proto_sym;
    actual_sym_was_non = proto_sym;
    actual_non_was_sym = proto_non;
    actual_non_was_non = proto_non;
    MY_CHECK_MATRIX_EQUAL(expected_sym, actual_sym_was_sym);
    MY_CHECK_MATRIX_EQUAL(expected_sym, actual_sym_was_non);
    MY_CHECK_MATRIX_EQUAL(expected_non, actual_non_was_sym);
    MY_CHECK_MATRIX_EQUAL(expected_non, actual_non_was_non);
  }
  {
    // move construction
    block_matrix proto_sym {SYM_BLOCK_MATRIX_INIT};
    block_matrix proto_non {NON_BLOCK_MATRIX_INIT};
    const block_matrix actual_sym {std::move(proto_sym)};
    const block_matrix actual_non {std::move(proto_non)};
    MY_CHECK_MATRIX_EQUAL(expected_sym, actual_sym);
    MY_CHECK_MATRIX_EQUAL(expected_non, actual_non);
    BOOST_CHECK_EQUAL(0, proto_sym.size());
    BOOST_CHECK_EQUAL(0, proto_non.size());
  }
  {
    // move assignment
    block_matrix proto_sym_1 {SYM_BLOCK_MATRIX_INIT};
    block_matrix proto_sym_2 {SYM_BLOCK_MATRIX_INIT};
    block_matrix proto_non_1 {NON_BLOCK_MATRIX_INIT};
    block_matrix proto_non_2 {NON_BLOCK_MATRIX_INIT};
    block_matrix actual_sym_was_sym {X_SYM_BLOCK_MATRIX_INIT};
    block_matrix actual_sym_was_non {X_NON_BLOCK_MATRIX_INIT};
    block_matrix actual_non_was_sym {X_SYM_BLOCK_MATRIX_INIT};
    block_matrix actual_non_was_non {X_NON_BLOCK_MATRIX_INIT};
    actual_sym_was_sym = std::move(proto_sym_1);
    actual_sym_was_non = std::move(proto_sym_2);
    actual_non_was_sym = std::move(proto_non_1);
    actual_non_was_non = std::move(proto_non_2);
    MY_CHECK_MATRIX_EQUAL(expected_sym, actual_sym_was_sym);
    MY_CHECK_MATRIX_EQUAL(expected_sym, actual_sym_was_non);
    MY_CHECK_MATRIX_EQUAL(expected_non, actual_non_was_sym);
    MY_CHECK_MATRIX_EQUAL(expected_non, actual_non_was_non);
    BOOST_CHECK_EQUAL(0, proto_sym_1.size());
    BOOST_CHECK_EQUAL(0, proto_sym_2.size());
    BOOST_CHECK_EQUAL(0, proto_non_1.size());
    BOOST_CHECK_EQUAL(0, proto_non_2.size());
  }
  {
    // swapping
    using std::swap;
    block_matrix sym_2_non {SYM_BLOCK_MATRIX_INIT};
    block_matrix non_2_sym {NON_BLOCK_MATRIX_INIT};
    MY_CHECK_MATRIX_EQUAL(expected_sym, sym_2_non);
    MY_CHECK_MATRIX_EQUAL(expected_non, non_2_sym);
    swap(sym_2_non, non_2_sym);
    MY_CHECK_MATRIX_EQUAL(expected_non, sym_2_non);
    MY_CHECK_MATRIX_EQUAL(expected_sym, non_2_sym);
  }
#undef SYM_BLOCK_MATRIX_INIT
#undef NON_BLOCK_MATRIX_INIT
#undef X_SYM_BLOCK_MATRIX_INIT
#undef X_NON_BLOCK_MATRIX_INIT
}


/**
 * @brief
 *         Unit test for the passive `passOverRegion` member function.
 *
 * This test checks that:
 *
 * - the correct region is shown to the visitor
 * - zero-blocks are not shown to the visitor
 * - transposition is applied correctly
 * - each element is visited exactly once
 * - the visited matrix is `const` but the visitor may mutate
 *
 */
BOOST_AUTO_TEST_CASE(pass_over_region__passive)
{
  typedef std::string value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using block_matrix = bsc_klammler::BlockMatrix<value_type>;
  using extractor = bsc_klammler::test::Extractor<value_type>;
  const full_matrix block_a {
    {"a11", "a12"},
    {"a21", "a22"},
  };
  const full_matrix block_b {
    {"b11", "b12", "b13"},
    {"b21", "b22", "b23"},
    {"b31", "b32", "b33"},
  };
  const full_matrix block_c {
    {"c11", "c12"},
    {"c21", "c22"},
  };
  const full_matrix block_u {
    {"u11", "u12"},
    {"u21", "u22"},
  };
  const full_matrix block_v {
    {"v11", "v12", "v13"},
    {"v21", "v22", "v23"},
  };
  const full_matrix block_x {
    {"x11", "x12"},
    {"x21", "x22"},
  };
  const full_matrix block_y {
    {"y11", "y12"},
    {"y21", "y22"},
    {"y31", "y32"},
  };
  const block_matrix matrix {
    block_a,
    block_b,
    block_c,
    block_u,
    block_v,
    block_x,
    block_y,
  };
  // We extract into a matrix filled with non-zero entries so we can tell
  // whether the visitor was passed over zero-blocks, which it shouldn't.
  const full_matrix qmarks {
    {"?", "?", "?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?", "?", "?"},
    {"?", "?", "?", "?", "?", "?", "?"},
  };
  assert(matrix.rows() == qmarks.rows());
  assert(matrix.cols() == qmarks.cols());
  {
    const full_matrix expected {
      {"a11", "a12",   "?",   "?",   "?", "x11", "x12"},
      {"a21", "a22",   "?",   "?",   "?", "x21", "x22"},
      {  "?",   "?", "b11", "b12", "b13", "y11", "y12"},
      {  "?",   "?", "b21", "b22", "b23", "y21", "y22"},
      {  "?",   "?", "b31", "b32", "b33", "y31", "y32"},
      {"u11", "u12", "v11", "v12", "v13", "c11", "c12"},
      {"u21", "u22", "v21", "v22", "v23", "c21", "c22"},
    };
    full_matrix actual {qmarks};
    extractor pv {actual};
    matrix.passOverRegion(0, 0, false, pv, 0, 0, matrix.rows(), matrix.cols());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {
      {"a11", "a21",   "?",   "?",   "?", "u11", "u21"},
      {"a12", "a22",   "?",   "?",   "?", "u12", "u22"},
      {  "?",   "?", "b11", "b21", "b31", "v11", "v21"},
      {  "?",   "?", "b12", "b22", "b32", "v12", "v22"},
      {  "?",   "?", "b13", "b23", "b33", "v13", "v23"},
      {"x11", "x21", "y11", "y21", "y31", "c11", "c21"},
      {"x12", "x22", "y12", "y22", "y32", "c12", "c22"},
    };
    full_matrix actual {qmarks};
    extractor pv {actual};
    matrix.passOverRegion(0, 0, true, pv, 0, 0, matrix.rows(), matrix.cols());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {
      {  "?",   "?",   "?",   "?",   "?",   "?",   "?"},
      {  "?", "a22",   "?",   "?",   "?",   "?",   "?"},
      {  "?",   "?", "b11",   "?",   "?",   "?",   "?"},
      {  "?",   "?", "b21",   "?",   "?",   "?",   "?"},
      {  "?",   "?", "b31",   "?",   "?",   "?",   "?"},
      {  "?", "u12", "v11",   "?",   "?",   "?",   "?"},
      {  "?", "u22", "v21",   "?",   "?",   "?",   "?"},
    };
    full_matrix actual {qmarks};
    extractor pv {actual};
    matrix.passOverRegion(1, 1, false, pv, 1, 1, 7, 3);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {
      {  "?",   "?",   "?",   "?",   "?",   "?",   "?"},
      {  "?",   "?",   "?",   "?",   "?",   "?",   "?"},
      {  "?",   "?",   "?",   "?",   "?",   "?",   "?"},
      {  "?", "v12", "v22",   "?",   "?",   "?",   "?"},
      {  "?", "v13", "v23",   "?",   "?",   "?",   "?"},
      {  "?", "c11", "c21",   "?",   "?",   "?",   "?"},
      {  "?",   "?",   "?",   "?",   "?",   "?",   "?"},
    };
    full_matrix actual {qmarks};
    extractor pv {actual};
    matrix.passOverRegion(3, 1, true, pv, 3, 5, 6, 7);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {qmarks};
    full_matrix actual {qmarks};
    extractor pv {actual};
    matrix.passOverRegion(1, 1, false, pv, 0, 5, 7, 5);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const block_matrix empty {};
    const full_matrix expected {};
    full_matrix actual {};
    extractor pv {actual};
    matrix.passOverRegion(0, 0, false, pv, 0, 0, 0, 0);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
}


/**
 * @brief
 *         Unit test for the active `passOverRegion` member function.
 *
 * This test checks that:
 *
 * - the correct region is shown to the visitor
 * - zero-blocks are not shown to the visitor
 * - transposition is applied correctly
 * - each element is visited exactly once
 * - the visited matrix and the visitor may mutate
 *
 */
BOOST_AUTO_TEST_CASE(pass_over_region__active)
{
  typedef std::string value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using block_matrix = bsc_klammler::BlockMatrix<value_type>;
  using inserter = bsc_klammler::test::Inserter<value_type>;
  const full_matrix data {
    {"a11", "a12",   "0",   "0",   "0", "x11", "x12"},
    {"a21", "a22",   "0",   "0",   "0", "x21", "x22"},
    {  "0",   "0", "b11", "b12", "b13", "y11", "y12"},
    {  "0",   "0", "b21", "b22", "b23", "y21", "y22"},
    {  "0",   "0", "b31", "b32", "b33", "y31", "y32"},
    {"u11", "u12", "v11", "v12", "v13", "c11", "c12"},
    {"u21", "u22", "v21", "v22", "v23", "c21", "c22"},
  };
  const full_matrix block_a {2, 2};
  const full_matrix block_b {3, 3};
  const full_matrix block_c {2, 2};
  const full_matrix block_u {2, 2};
  const full_matrix block_v {2, 3};
  const full_matrix block_x {2, 2};
  const full_matrix block_y {3, 2};
  const block_matrix matrix {
    block_a,
    block_b,
    block_c,
    block_u,
    block_v,
    block_x,
    block_y,
  };
  assert(matrix.rows() == data.rows());
  assert(matrix.cols() == data.cols());
  // If the visitor is passed over a zero-block, the attempt to 'set()' values
  // therein will trigger an exception.
  {
    const full_matrix expected {
      {"a11", "a12",    "",    "",    "", "x11", "x12"},
      {"a21", "a22",    "",    "",    "", "x21", "x22"},
      {   "",    "", "b11", "b12", "b13", "y11", "y12"},
      {   "",    "", "b21", "b22", "b23", "y21", "y22"},
      {   "",    "", "b31", "b32", "b33", "y31", "y32"},
      {"u11", "u12", "v11", "v12", "v13", "c11", "c12"},
      {"u21", "u22", "v21", "v22", "v23", "c21", "c22"},
    };
    block_matrix actual {matrix};
    inserter av {data};
    actual.passOverRegion(0, 0, false, av, 0, 0, matrix.rows(), matrix.cols());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {
      {"a11", "a21",    "",    "",    "", "u11", "u21"},
      {"a12", "a22",    "",    "",    "", "u12", "u22"},
      {   "",    "", "b11", "b21", "b31", "v11", "v21"},
      {   "",    "", "b12", "b22", "b32", "v12", "v22"},
      {   "",    "", "b13", "b23", "b33", "v13", "v23"},
      {"x11", "x21", "y11", "y21", "y31", "c11", "c21"},
      {"x12", "x22", "y12", "y22", "y32", "c12", "c22"},
    };
    block_matrix actual {matrix};
    inserter av {data};
    actual.passOverRegion(0, 0, true, av, 0, 0, matrix.cols(), matrix.rows());
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {
      {   "",    "",    "",    "",    "",    "",    ""},
      {   "", "a22",    "",    "",    "",    "",    ""},
      {   "",    "", "b11",    "",    "",    "",    ""},
      {   "",    "", "b21",    "",    "",    "",    ""},
      {   "",    "", "b31",    "",    "",    "",    ""},
      {   "", "u12", "v11",    "",    "",    "",    ""},
      {   "", "u22", "v21",    "",    "",    "",    ""},
    };
    block_matrix actual {matrix};
    inserter av {data};
    actual.passOverRegion(1, 1, false, av, 1, 1, 7, 3);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {
      {   "",    "",    "",    "",    "",    "",    ""},
      {   "",    "",    "",    "",    "",    "",    ""},
      {   "",    "",    "",    "",    "",    "",    ""},
      {   "",    "",    "",    "",    "", "a11", "a21"},
      {   "",    "",    "",    "",    "", "a12", "a22"},
      {   "",    "",    "",    "",    "",    "",    ""},
      {   "",    "",    "",    "",    "",    "",    ""},
    };
    block_matrix actual {matrix};
    inserter av {data};
    actual.passOverRegion(0, 0, true, av, 5, 3, 7, 5);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix expected {data.rows(), data.cols()};
    block_matrix actual {matrix};
    inserter av {data};
    actual.passOverRegion(1, 1, false, av, 0, 5, 7, 5);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    const full_matrix empty {};
    const full_matrix expected {};
    block_matrix actual {};
    inserter av {empty};
    actual.passOverRegion(0, 0, false, av, 0, 0, 0, 0);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
}
