/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_full_matrix.cpp
 *
 * @brief
 *         Unit tests for the `FullMatrix` template class.
 *
 */

#include "../global.hpp"           // (package configuration)
#include "../full_matrix.hpp"      // bsc_klammler::FullMatrix

#include <complex>                 // std::complex
#include <memory>                  // std::unique_ptr
#include <stdexcept>               // std::invalid_argument
#include <string>                  // std::string
#include <typeinfo>                // typeid
#include <vector>                  // std::vector

#include <boost/mpl/list.hpp>      // boost::mpl::list

#include "../arranged_matrix.hpp"  // bsc_klammler::ArrangedMatrix
#include "../full_matrix.hpp"      // bsc_klammler::FullMatrix
#include "../matrix_view.hpp"      // bsc_klammler::MatrixView
#include "../random.hpp"           // bsc_klammler::random_real
#include "foo.hpp"                 // bsc_klammler::test::Foo
#include "mock_h_matrix.hpp"       // bsc_klammler::test::MockHMatrix
#include "util.hpp"                // MY_CHECK_MATRIX_EQUAL
#include "visitors.hpp"            // bsc_klammler::test::Extractor, ...

#define BOOST_TEST_MODULE full_matrix
#include "module.hpp"

using bsc_klammler::random_int;
using bsc_klammler::random_real;
using bsc_klammler::test::Foo;


/** @brief List of `Extractor` and `Inserter` for `double`s. */
using mpl_double_visitors = boost::mpl::list<bsc_klammler::test::Extractor<double>,
                                             bsc_klammler::test::Inserter<double>>;


/**
 * @brief
 *         Unit test for the member typdefs.
 *
 * This test checks that:
 *
 * - the member types `value_type`, `index_type` and `super_type` are defined
 *   correctly
 *
 */
BOOST_AUTO_TEST_CASE(member_typedefs)
{
  // There is no overload for 'operator<<(std::ostream&, std::type_info)' so we
  // cannot use 'BOOST_CHECK_EQUAL'.
  struct Bar {};
  BOOST_CHECK(typeid(bsc_klammler::HMatrix<Bar>) ==
              typeid(bsc_klammler::FullMatrix<Bar>::super_type));
  BOOST_CHECK(typeid(Bar) ==
              typeid(bsc_klammler::FullMatrix<Bar>::value_type));
  for (typename bsc_klammler::FullMatrix<Bar>::index_type i = 0; i < 10; ++i)
    continue;
}


/**
 * @brief
 *         Unit test for the default constructor.
 *
 * This test checks that:
 *
 * - the dimensions of the created matrix are correct (0, 0).
 *
 */
BOOST_AUTO_TEST_CASE(ctor_default)
{
  const bsc_klammler::FullMatrix<double> matrix {};
  BOOST_CHECK_EQUAL(0, matrix.rows());
  BOOST_CHECK_EQUAL(0, matrix.cols());
}


namespace /* anonymous */
{
  // Helper template to implement checks for the 'construction' test case with
  // static polymorphism.
  template<typename T>
  static void
  construction_check_(const typename bsc_klammler::FullMatrix<T>::index_type rows,
                      const typename bsc_klammler::FullMatrix<T>::index_type cols)
  {
    using index_type = typename bsc_klammler::FullMatrix<T>::index_type;
    const bsc_klammler::FullMatrix<T> matrix {rows, cols};
    BOOST_CHECK_EQUAL(rows, matrix.rows());
    BOOST_CHECK_EQUAL(cols, matrix.cols());
    for (index_type j = 0; j < cols; ++j)
      {
        for (index_type i = 0; i < rows; ++i)
          {
            BOOST_CHECK_EQUAL(bsc_klammler::FullMatrix<T>::zero(),
                              matrix.get(i, j));
          }
      }
  }
}  // namespace /* anonymous */

/**
 * @brief
 *         Unit test for the constructor taking two dimensions as arguments.
 *
 * This test checks that:
 *
 * - the dimensions of the created matrix are correct
 * - (0, 0) matrices are supported
 * - the elements are value-initialized
 *
 * This is also a unit test for the `rows` and `cols` function.
 *
 */
BOOST_AUTO_TEST_CASE(ctor)
{
  using index_type = typename bsc_klammler::FullMatrix<int>::index_type;
  const index_type min {1};
  const index_type max {128};
  for (const auto empty : {true, false})
    {
      const auto d = [empty]()->index_type{
        return empty ? 0 : random_int(min, max);
      };
      const index_type n {d()};
      const index_type m {d()};
      construction_check_<bool>(n, m);
      construction_check_<char>(n, m);
      construction_check_<signed>(n, m);
      construction_check_<unsigned>(n, m);
      construction_check_<float>(n, m);
      construction_check_<double>(n, m);
      construction_check_<std::complex<double>>(n, m);
      construction_check_<std::string>(n, m);
    }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the constructor taking two dimensions
 *         as arguments.
 *
 * This test checks that:
 *
 * - <del>ill-formed matrices (<var>n</var> = 0, m &gt; 0) or
 *   (<var>n</var> &gt; 0, <var>m</var> = 0) are rejected</del>
 * - dimensions that would (likely) overflow are rejected
 *
 */
BOOST_AUTO_TEST_CASE(ctor__invalid)
{
  using index_type = typename bsc_klammler::FullMatrix<int>::index_type;
  const index_type min {1};
  const index_type max {1 << 20};  // should never be allocated...
  const index_type sizes[][2] = {
    /* {0, 1}, */
    /* {1, 0}, */
    /* {random_int(min, max), 0}, */
    /* {0, random_int(min, max)}, */
    {random_int(min, max), static_cast<index_type>(-1)},
    {static_cast<index_type>(-1), random_int(min, max)},
    {static_cast<index_type>(-1), static_cast<index_type>(-1)},
  };
  for (const auto& iter : sizes)
    {
      try
        {
          bsc_klammler::FullMatrix<double> bad_matrix {iter[0], iter[1]};
          BOOST_ERROR("exception not thrown");
        }
      catch (const std::invalid_argument&) { /* good */ }
    }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


#if !(defined BSC_KLAMMLER_EIGEN_BUG_CTOR_NOT_EXCEPTION_SAFE)
/**
 * @brief
 *         Unit test for the exception safety of the constructor taking two
 *         dimensions as arguments.
 *
 * If run through *Valgrind*, this test checks that:
 *
 * - the constructor performs correctly even in case an exception is thrown
 *   while value-initializing the matrix elements
 *
 */
BOOST_AUTO_TEST_CASE(ctor__except)
{
  Foo::deactivate();
  try
    {
      const bsc_klammler::FullMatrix<Foo> matrix {14, 92};
      BOOST_ERROR("no exception was thrown");
    }
  catch (const Foo::NotActive&) { /* okay */ }
  BOOST_WARN(RUNNING_ON_VALGRIND);
}
#endif  // if !(defined BSC_KLAMMLER_EIGEN_BUG_CTOR_NOT_EXCEPTION_SAFE)


/**
 * @brief
 *         Unit test for the initializer-list constructor.
 *
 * This test checks that:
 *
 * - initializer lists produce corrent matrices
 *
 */
BOOST_AUTO_TEST_CASE(ctor_initializer_list)
{
  using index_type = typename bsc_klammler::FullMatrix<double>::index_type;
  {
    // copied from the DocString
    const bsc_klammler::FullMatrix<double> matrix {
      {11.0, 12.0, 13.0, 14.0},
      {21.0, 22.0, 23.0, 24.0},
      {31.0, 32.0, 33.0, 34.0},
    };
    BOOST_CHECK_EQUAL(3, matrix.rows());
    BOOST_CHECK_EQUAL(4, matrix.cols());
    for (index_type j = 0; j < 4; ++j)
      {
        for (index_type i = 0; i < 3; ++i)
          {
            const double expected {10.0 * (i + 1.0) + (j + 1.0)};
            const double actual {matrix.get(i, j)};
            BOOST_CHECK_EQUAL(expected, actual);
          }
      }
  }
  {
    // (0, 0) matrix
    const std::initializer_list<std::initializer_list<double>> init {};
    const bsc_klammler::FullMatrix<double> matrix {init};
    BOOST_CHECK_EQUAL(0, matrix.rows());
    BOOST_CHECK_EQUAL(0, matrix.cols());
  }
  {
    // (1, 1) matrix
    const bsc_klammler::FullMatrix<double> matrix {{1.0}};
    BOOST_CHECK_EQUAL(1, matrix.rows());
    BOOST_CHECK_EQUAL(1, matrix.cols());
    BOOST_CHECK_EQUAL(1.0, matrix.get(0, 0));
  }
  {
    // (n, 1) matrix
    const bsc_klammler::FullMatrix<double> matrix {{1.0}, {2.0}, {3.0}};
    BOOST_CHECK_EQUAL(3, matrix.rows());
    BOOST_CHECK_EQUAL(1, matrix.cols());
    for (index_type i = 0; i < 3; ++i)
      {
        BOOST_CHECK_EQUAL(static_cast<double>(i + 1), matrix.get(i, 0));
      }
  }
  {
    // (1, m) matrix
    const bsc_klammler::FullMatrix<double> matrix {{1.0, 2.0, 3.0}};
    BOOST_CHECK_EQUAL(1, matrix.rows());
    BOOST_CHECK_EQUAL(3, matrix.cols());
    for (index_type j = 0; j < 3; ++j)
      {
        BOOST_CHECK_EQUAL(static_cast<double>(j + 1), matrix.get(0, j));
      }
  }
  {
    // (1, 0) matrix
    const bsc_klammler::FullMatrix<double> matrix {{}};
    BOOST_CHECK_EQUAL(1, matrix.rows());
    BOOST_CHECK_EQUAL(0, matrix.cols());
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the initializer-list constructor.
 *
 * This test checks that:
 *
 * - jagged initializer lists are rejected
 *
 */
BOOST_AUTO_TEST_CASE(ctor_initializer_list__invalid)
{
  try
    {
      // copied from the DocString
      bsc_klammler::FullMatrix<double> bad_matrix {{1.0}, {2.0, 3.0}};
      BOOST_ERROR("expected error: jagged matrix ill-defined");
    }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the initializer-list constructor's exception safety.
 *
 * If run through *Valgrind*, this test checks that:
 *
 * - the constructor performs correctly even in case an exception is thrown
 *   while copying the matrix elements
 *
 */
BOOST_AUTO_TEST_CASE(ctor_initializer_list__except)
{
  Foo::deactivate();
  try
    {
      const bsc_klammler::FullMatrix<Foo> matrix {
        {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15}
      };
      BOOST_ERROR("no exception was thrown");
    }
  catch (const Foo::NotActive&) { /* okay */ }
  BOOST_WARN(RUNNING_ON_VALGRIND);
}


/**
 * @brief
 *         Unit test for the copy constructor.
 *
 * This test checks that:
 *
 * - the copy constructor works
 * - created copies are deep
 *
 */
BOOST_AUTO_TEST_CASE(ctor_copy)
{
  {
    const bsc_klammler::FullMatrix<int> matrix1 {};
    const bsc_klammler::FullMatrix<int> matrix2 {matrix1};
    BOOST_CHECK_EQUAL(0, matrix1.rows());
    BOOST_CHECK_EQUAL(0, matrix2.rows());
    BOOST_CHECK_EQUAL(0, matrix1.cols());
    BOOST_CHECK_EQUAL(0, matrix2.cols());
  }
  {
    const bsc_klammler::FullMatrix<char> matrix1 {{'A', 'B'}, {'C', 'D'}};
    const bsc_klammler::FullMatrix<char> matrix2 {matrix1};
    BOOST_CHECK_EQUAL(2, matrix1.rows());
    BOOST_CHECK_EQUAL(2, matrix2.rows());
    BOOST_CHECK_EQUAL(2, matrix1.cols());
    BOOST_CHECK_EQUAL(2, matrix2.cols());
    BOOST_CHECK_EQUAL('A', matrix2.get(0, 0));
    BOOST_CHECK_EQUAL('B', matrix2.get(0, 1));
    BOOST_CHECK_EQUAL('C', matrix2.get(1, 0));
    BOOST_CHECK_EQUAL('D', matrix2.get(1, 1));
  }
  {
    using index_type = typename bsc_klammler::FullMatrix<double>::index_type;
    const index_type rows {random_int<index_type>(1, 128)};
    const index_type cols {random_int<index_type>(1, 128)};
    std::vector<double> values {};  // column-major
    index_type k;  // use as iteration index over 'values'
    // 0. create reference vector with random elements
    for (k = 0; k < rows * cols; ++k)
      values.push_back(random_real<double>());
    // 1. create matrix M1 filled with these elements
    bsc_klammler::FullMatrix<double> matrix1 {rows, cols};
    k = 0;
    for (index_type j = 0; j < cols; ++j)
      {
        for (index_type i = 0; i < rows; ++i)
          {
            matrix1.set(values.at(k++), i, j);
          }
      }
    // 2. create clone M2 of matrix M1
    bsc_klammler::FullMatrix<double> matrix2 {matrix1};
    // 3. check that M1[i][j] == M2[i][j] == values[...]
    k = 0;
    for (index_type j = 0; j < cols; ++j)
      {
        for (index_type i = 0; i < rows; ++i)
          {
            const double v {values.at(k++)};
            BOOST_CHECK_EQUAL(v, matrix1.get(i, j));
            BOOST_CHECK_EQUAL(v, matrix2.get(i, j));
          }
      }
    // 4. overwrite M1 with zeros
    for (index_type j = 0; j < cols; ++j)
      {
        for (index_type i = 0; i < rows; ++i)
          {
            matrix1.set(0.0, i, j);
          }
      }
    // 5. check that M1[i][j] == 0 and M2[i][j] == values[...]
    k = 0;
    for (index_type j = 0; j < cols; ++j)
      {
        for (index_type i = 0; i < rows; ++i)
          {
            BOOST_CHECK_EQUAL(0.0, matrix1.get(i, j));
            BOOST_CHECK_EQUAL(values.at(k++), matrix2.get(i, j));
          }
      }
  }
  {
    std::unique_ptr<bsc_klammler::FullMatrix<int>> matrix1_uptr {};
    std::unique_ptr<bsc_klammler::FullMatrix<int>> matrix2_uptr {};
    matrix1_uptr.reset(
        new bsc_klammler::FullMatrix<int> {{11, 12}, {21, 22}, {31, 32}});
    matrix2_uptr.reset(
        new bsc_klammler::FullMatrix<int> {*matrix1_uptr.get()});
    matrix1_uptr.reset(nullptr);
    BOOST_CHECK_EQUAL(11, matrix2_uptr.get()->get(0, 0));
    BOOST_CHECK_EQUAL(12, matrix2_uptr.get()->get(0, 1));
    BOOST_CHECK_EQUAL(21, matrix2_uptr.get()->get(1, 0));
    BOOST_CHECK_EQUAL(22, matrix2_uptr.get()->get(1, 1));
    BOOST_CHECK_EQUAL(31, matrix2_uptr.get()->get(2, 0));
    BOOST_CHECK_EQUAL(32, matrix2_uptr.get()->get(2, 1));
  }
}


#if !(defined BSC_KLAMMLER_EIGEN_BUG_CTOR_NOT_EXCEPTION_SAFE)
/**
 * @brief
 *         Unit test for the copy constructor's exception safety.
 *
 * If run through *Valgrind*, this test checks that:
 *
 * - the constructor performs correctly even in case an exception is thrown
 *   while copying the matrix elements
 *
 */
BOOST_AUTO_TEST_CASE(ctor_copy__except)
{
  Foo::activate();
  const bsc_klammler::FullMatrix<Foo> matrix1 {14, 92};
  Foo::deactivate();
  try
    {
      bsc_klammler::FullMatrix<Foo> matrix2 {matrix1};
      BOOST_ERROR("no exception was thrown");
    }
  catch (const Foo::NotActive&) { /* okay */ }
  BOOST_WARN(RUNNING_ON_VALGRIND);
}
#endif  // !(defined BSC_KLAMMLER_EIGEN_BUG_CTOR_NOT_EXCEPTION_SAFE)


/**
 * @brief
 *         Unit test for the move constructor.
 *
 * This test checks that:
 *
 * - the move constructor works
 * - no copies are created
 * - no member-wise move is performed
 *
 */
BOOST_AUTO_TEST_CASE(ctor_move)
{
  using index_type = typename bsc_klammler::FullMatrix<Foo>::index_type;
  const index_type rows {random_int<index_type>(1, 128)};
  const index_type cols {random_int<index_type>(1, 128)};
  static int random_value {random_int<int>()};
  Foo::activate();
  bsc_klammler::FullMatrix<Foo> matrix1 {rows, cols};
  for (index_type j = 0; j < matrix1.cols(); ++j)
    {
      for (index_type i = 0; i < matrix1.rows(); ++i)
        {
          matrix1.set(random_value, i, j);
        }
    }
  Foo::deactivate();
  bsc_klammler::FullMatrix<Foo> matrix2 {std::move(matrix1)};
  Foo::activate();
  BOOST_CHECK_EQUAL(0, matrix1.rows());
  BOOST_CHECK_EQUAL(0, matrix1.cols());
  BOOST_CHECK_EQUAL(rows, matrix2.rows());
  BOOST_CHECK_EQUAL(cols, matrix2.cols());
  for (index_type j = 0; j < matrix2.cols(); ++j)
    {
      for (index_type i = 0; i < matrix2.rows(); ++i)
        {
          BOOST_CHECK_EQUAL(random_value, matrix2.get(i, j));
        }
    }
}


/**
 * @brief
 *         Unit test for the polymorphic copy constructor.
 *
 * This test checks that:
 *
 * - the constructor works
 *
 */
BOOST_AUTO_TEST_CASE(ctor_copy_poly)
{
  using index_type = bsc_klammler::HMatrix<int>::index_type;
  struct value_type
  {
    // Value types must be default constructible.
    index_type row;
    index_type col;

    value_type() noexcept : row {}, col {}
    {
    }

    value_type(const index_type r, const index_type c) noexcept :
      row {r}, col {c}
    {
    }
  };
  using mockup = bsc_klammler::test::MockHMatrix<value_type>;
  const index_type rows {random_int<index_type>(1, 32)};
  const index_type cols {random_int<index_type>(1, 32)};
  std::unique_ptr<mockup> mock_uptr {new mockup {}};
  mock_uptr->mockRows([rows]()->index_type{ return rows; });
  mock_uptr->mockCols([cols]()->index_type{ return cols; });
  mock_uptr->mockGet([](const index_type i, const index_type j)->value_type{
      return value_type {i, j};
  });
  const bsc_klammler::FullMatrix<value_type> clone {*mock_uptr};
  mock_uptr.reset(nullptr);
  BOOST_CHECK_EQUAL(rows, clone.rows());
  BOOST_CHECK_EQUAL(cols, clone.cols());
  for (index_type j = 0; j < clone.cols(); ++j)
    {
      for (index_type i = 0; i < clone.rows(); ++i)
        {
          BOOST_CHECK_EQUAL(i, clone.get(i, j).row);
          BOOST_CHECK_EQUAL(j, clone.get(i, j).col);
        }
    }
}


#if !(defined BSC_KLAMMLER_EIGEN_BUG_CTOR_NOT_EXCEPTION_SAFE)
/**
 * @brief
 *         Unit test for the exception safety of the polymorphic copy
 *         constructor.
 *
 * If run through *Valgrind*, this test checks that:
 *
 * - the constructor performs correctly even in case an exception is thrown
 *
 */
BOOST_AUTO_TEST_CASE(ctor_copy_poly__except)
{
  using index_type = bsc_klammler::FullMatrix<Foo>::index_type;
  const index_type rows {random_int<index_type>(1, 128)};
  const index_type cols {random_int<index_type>(1, 128)};
  Foo::activate();
  const bsc_klammler::FullMatrix<Foo> matrix {rows, cols};
  const bsc_klammler::HMatrix<Foo>& polyref {matrix};
  try
    {
      Foo::deactivate();
      const bsc_klammler::FullMatrix<Foo> clone {polyref};
      BOOST_ERROR("no exception was thrown");
    }
  catch (const Foo::NotActive&) { /* okay */ }
  BOOST_WARN(RUNNING_ON_VALGRIND);
}
#endif  // !(defined BSC_KLAMMLER_EIGEN_BUG_CTOR_NOT_EXCEPTION_SAFE)


/**
 * @brief
 *         Unit test for the `clone` function member.
 *
 * This test checks that:
 *
 * - the `clone` function works
 * - the clone is allocated on the free store
 * - the clone is deep
 * - the clone is polymorphic
 *
 */
BOOST_AUTO_TEST_CASE(cloning)  // 'clone' clashes with 'clone()' in 'sched.h'
{
  using full_matrix = bsc_klammler::FullMatrix<double>;
  using h_matrix = full_matrix::super_type;
  using index_type = h_matrix::index_type;
  using value_type = h_matrix::value_type;
  std::unique_ptr<h_matrix> matrix_uptr {};
  {
    const full_matrix ephermal {
      {11.0, 12.0, 13.0, 14.0},
      {21.0, 22.0, 23.0, 24.0},
      {31.0, 32.0, 33.0, 34.0},
    };
    matrix_uptr = ephermal.clone();
  }
  BOOST_CHECK(dynamic_cast<full_matrix *>(matrix_uptr.get()));
  BOOST_CHECK_EQUAL(3, matrix_uptr->rows());
  BOOST_CHECK_EQUAL(4, matrix_uptr->cols());
  for (index_type j = 0; j < matrix_uptr->cols(); ++j)
    {
      for (index_type i = 0; i < matrix_uptr->rows(); ++i)
        {
          const value_type v {10.0 * (i + 1.0) + (j + 1.0)};
          BOOST_CHECK_EQUAL(v, matrix_uptr->get(i, j));
        }
    }
}


#if !(defined BSC_KLAMMLER_EIGEN_BUG_CTOR_NOT_EXCEPTION_SAFE)
/**
 * @brief
 *         Unit test for the exception safety of the `clone` member function.
 *
 * If run through *Valgrind*, this test checks that:
 *
 * - the function performs correctly even in case an exception is thrown
 *
 */
BOOST_AUTO_TEST_CASE(cloning__except)
{
  using index_type = bsc_klammler::FullMatrix<Foo>::index_type;
  const index_type rows {random_int<index_type>(1, 128)};
  const index_type cols {random_int<index_type>(1, 128)};
  Foo::activate();
  const bsc_klammler::FullMatrix<Foo> matrix {rows, cols};
  try
    {
      Foo::deactivate();
      matrix.clone();
      BOOST_ERROR("no exception was thrown");
    }
  catch (const Foo::NotActive&) { /* okay */ }
  BOOST_WARN(RUNNING_ON_VALGRIND);
}
#endif  // !(defined BSC_KLAMMLER_EIGEN_BUG_CTOR_NOT_EXCEPTION_SAFE)


/**
 * @brief
 *         Unit test for the `set` and `get` function members.
 *
 * This test checks that:
 *
 * - a value that was set via `set` can be accessed via `get`
 * - `get` does not mutate other memory
 *
 */
BOOST_AUTO_TEST_CASE(accessors)
{
  using index_type = typename bsc_klammler::FullMatrix<double>::index_type;
  const index_type rows {random_int<index_type>(1, 128)};
  const index_type cols {random_int<index_type>(1, 128)};
  bsc_klammler::FullMatrix<double> matrix {rows, cols};
  for (int k = 0; k < 10; ++k)
    {
      const index_type ik {random_int<index_type>(0, rows - 1)};
      const index_type jk {random_int<index_type>(0, cols - 1)};
      const double val {random_real<double>()};
      matrix.set(val, ik, jk);
      for (index_type j = 0; j < cols; ++j)
        {
          for (index_type i = 0; i < rows; ++i)
            {
              const double expected {(i == ik && j == jk) ? val : 0.0};
              const double actual {matrix.get(i, j)};
              BOOST_CHECK_EQUAL(expected, actual);
            }
        }
      matrix.set(0.0, ik, jk);
    }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the `set` and `get` function members.
 *
 * This test checks that:
 *
 * - invalid indices are rejected
 *
 */
BOOST_AUTO_TEST_CASE(accessors__invalid)
{
  using index_type = typename bsc_klammler::FullMatrix<double>::index_type;
  const index_type n {random_int<index_type>(1, 128)};
  const index_type m {random_int<index_type>(1, 128)};
  const index_type indices[][2] = {
    {0, m}, {n, 0}, {n, m}, {0, m + 1}, {n + 1, 0}, {n, m + 1}, {n + 1, m},
  };
  bsc_klammler::FullMatrix<double> matrix {n, m};
  for (const auto& iter : indices)
    {
      const index_type i {iter[0]};
      const index_type j {iter[1]};
      try
        {
          matrix.get(i, j);
          BOOST_ERROR("exception not thrown");
        }
      catch (const std::invalid_argument&) { /* good */ }
      try
        {
          matrix.set(0.0, i, j);
          BOOST_ERROR("exception not thrown");
        }
      catch (const std::invalid_argument&) { /* good */ }
    }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the copy-assignment operator.
 *
 * This test checks that:
 *
 * - the copy-assignment operator works
 * - created copies are deep
 *
 */
BOOST_AUTO_TEST_CASE(assign_copy)
{
  {
    const bsc_klammler::FullMatrix<int> matrix1 {};
    bsc_klammler::FullMatrix<int> matrix2 {};
    matrix2 = matrix1;
    BOOST_CHECK_EQUAL(0, matrix1.rows());
    BOOST_CHECK_EQUAL(0, matrix2.rows());
    BOOST_CHECK_EQUAL(0, matrix1.cols());
    BOOST_CHECK_EQUAL(0, matrix2.cols());
  }
  {
    const bsc_klammler::FullMatrix<char> matrix1 {{'A', 'B'}, {'C', 'D'}};
    bsc_klammler::FullMatrix<char> matrix2 {{'.', '.', '.', '.', '.'}};
    matrix2 = matrix1;
    BOOST_CHECK_EQUAL(2, matrix1.rows());
    BOOST_CHECK_EQUAL(2, matrix2.rows());
    BOOST_CHECK_EQUAL(2, matrix1.cols());
    BOOST_CHECK_EQUAL(2, matrix2.cols());
    BOOST_CHECK_EQUAL('A', matrix2.get(0, 0));
    BOOST_CHECK_EQUAL('B', matrix2.get(0, 1));
    BOOST_CHECK_EQUAL('C', matrix2.get(1, 0));
    BOOST_CHECK_EQUAL('D', matrix2.get(1, 1));
  }
  {
    using index_type = typename bsc_klammler::FullMatrix<double>::index_type;
    const index_type rows {random_int<index_type>(1, 128)};
    const index_type cols {random_int<index_type>(1, 128)};
    std::vector<double> values {};  // column-major
    index_type k;  // use as iteration index over 'values'
    bsc_klammler::FullMatrix<double> matrix1 {rows, cols};
    bsc_klammler::FullMatrix<double> matrix2 {};
    // 0. create reference vector with random elements
    for (k = 0; k < rows * cols; ++k)
      values.push_back(random_real<double>());
    // 1. filled matrix M1 with these elements
    k = 0;
    for (index_type j = 0; j < cols; ++j)
      {
        for (index_type i = 0; i < rows; ++i)
          {
            matrix1.set(values.at(k++), i, j);
          }
      }
    // 2. (re-)assign M1 to M2
    matrix2 = matrix1;
    // 3. check that M1[i][j] == M2[i][j] == values[...]
    k = 0;
    for (index_type j = 0; j < cols; ++j)
      {
        for (index_type i = 0; i < rows; ++i)
          {
            const double v {values.at(k++)};
            BOOST_CHECK_EQUAL(v, matrix1.get(i, j));
            BOOST_CHECK_EQUAL(v, matrix2.get(i, j));
          }
      }
    // 4. overwrite M1 with zeros
    for (index_type j = 0; j < cols; ++j)
      {
        for (index_type i = 0; i < rows; ++i)
          {
            matrix1.set(0.0, i, j);
          }
      }
    // 5. check that M1[i][j] == 0 and M2[i][j] == values[...]
    k = 0;
    for (index_type j = 0; j < cols; ++j)
      {
        for (index_type i = 0; i < rows; ++i)
          {
            BOOST_CHECK_EQUAL(0.0, matrix1.get(i, j));
            BOOST_CHECK_EQUAL(values.at(k++), matrix2.get(i, j));
          }
      }
  }
  {
    bsc_klammler::FullMatrix<int> matrix {};
    {
      bsc_klammler::FullMatrix<int> ephermal {{11, 12}, {21, 22}, {31, 32}};
      matrix = ephermal;
    }
    BOOST_CHECK_EQUAL(11, matrix.get(0, 0));
    BOOST_CHECK_EQUAL(12, matrix.get(0, 1));
    BOOST_CHECK_EQUAL(21, matrix.get(1, 0));
    BOOST_CHECK_EQUAL(22, matrix.get(1, 1));
    BOOST_CHECK_EQUAL(31, matrix.get(2, 0));
    BOOST_CHECK_EQUAL(32, matrix.get(2, 1));
  }
}


#if !(defined BSC_KLAMMLER_EIGEN_BUG_CTOR_NOT_EXCEPTION_SAFE)
/**
 * @brief
 *         Unit test for the copy-assignment operator's exception safety.
 *
 * If run through *Valgrind*, this test checks that:
 *
 * - the constructor performs correctly even in case an exception is thrown
 *   while copying the matrix elements
 *
 */
BOOST_AUTO_TEST_CASE(assign_copy__except)
{
  Foo::activate();
  const bsc_klammler::FullMatrix<Foo> matrix1 {14, 92};
  try
    {
      bsc_klammler::FullMatrix<Foo> matrix2 {};
      Foo::deactivate();
      matrix2 = matrix1;
      BOOST_ERROR("no exception was thrown");
    }
  catch (const Foo::NotActive&) { /* okay */ }
  BOOST_WARN(RUNNING_ON_VALGRIND);
}
#endif  // !(defined BSC_KLAMMLER_EIGEN_BUG_CTOR_NOT_EXCEPTION_SAFE)


/**
 * @brief
 *         Unit test for the move-assignment operator.
 *
 * This test checks that:
 *
 * - the move-assignment operator works
 *
 */
BOOST_AUTO_TEST_CASE(assign_move)
{
  {
    bsc_klammler::FullMatrix<int> matrix1 {};
    bsc_klammler::FullMatrix<int> matrix2 {{0, 1, 2, 3, 4, 5, 6, 7}};
    matrix2 = std::move(matrix1);
    BOOST_CHECK_EQUAL(0, matrix1.rows());
    BOOST_CHECK_EQUAL(0, matrix1.cols());
    BOOST_CHECK_EQUAL(0, matrix2.rows());
    BOOST_CHECK_EQUAL(0, matrix2.cols());
  }
  {
    bsc_klammler::FullMatrix<int> matrix1 {{0, 1, 2, 3}};
    bsc_klammler::FullMatrix<int> matrix2 {};
    matrix2 = std::move(matrix1);
    BOOST_CHECK_EQUAL(0, matrix1.rows());
    BOOST_CHECK_EQUAL(0, matrix1.cols());
    BOOST_CHECK_EQUAL(1, matrix2.rows());
    BOOST_CHECK_EQUAL(4, matrix2.cols());
    BOOST_CHECK_EQUAL(0, matrix2.get(0, 0));
    BOOST_CHECK_EQUAL(1, matrix2.get(0, 1));
    BOOST_CHECK_EQUAL(2, matrix2.get(0, 2));
    BOOST_CHECK_EQUAL(3, matrix2.get(0, 3));
  }
}


/**
 * @brief
 *         Unit test for the `swap` friend function.
 *
 * This test checks that:
 *
 * - `get` and `set` behave well after a `swap()`
 * - no elements are copied or moved
 *
 */
BOOST_AUTO_TEST_CASE(swapping)
{
  using index_type = typename bsc_klammler::FullMatrix<Foo>::index_type;
  Foo::activate();
  const index_type rows1 {random_int<index_type>(1, 128)};
  const index_type cols1 {random_int<index_type>(1, 128)};
  const index_type rows2 {random_int<index_type>(1, 128)};
  const index_type cols2 {random_int<index_type>(1, 128)};
  bsc_klammler::FullMatrix<Foo> matrix1 {rows1, cols1};
  bsc_klammler::FullMatrix<Foo> matrix2 {rows2, cols2};
  std::vector<int> values1 {};
  std::vector<int> values2 {};
  for (index_type i = 0; i < rows1 * cols1; ++i)
    values1.push_back(random_int<int>());
  for (index_type i = 0; i < rows2 * cols2; ++i)
    values2.push_back(random_int<int>());
  BOOST_CHECK_EQUAL(rows1, matrix1.rows());
  BOOST_CHECK_EQUAL(cols1, matrix1.cols());
  BOOST_CHECK_EQUAL(rows2, matrix2.rows());
  BOOST_CHECK_EQUAL(cols2, matrix2.cols());
  {
    index_type k {0};
    for (index_type j = 0; j < matrix1.cols(); ++j)
      for (index_type i = 0; i < matrix1.rows(); ++i)
        matrix1.set(values1.at(k++), i, j);
  }
  {
    index_type k {0};
    for (index_type j = 0; j < matrix2.cols(); ++j)
      for (index_type i = 0; i < matrix2.rows(); ++i)
        matrix2.set(values2.at(k++), i, j);
  }
  Foo::deactivate();
  swap(matrix1, matrix2);
  Foo::activate();
  BOOST_CHECK_EQUAL(rows2, matrix1.rows());
  BOOST_CHECK_EQUAL(cols2, matrix1.cols());
  BOOST_CHECK_EQUAL(rows1, matrix2.rows());
  BOOST_CHECK_EQUAL(cols1, matrix2.cols());
  {
    index_type k {0};
    for (index_type j = 0; j < matrix1.cols(); ++j)
      for (index_type i = 0; i < matrix1.rows(); ++i)
        BOOST_CHECK_EQUAL(values2.at(k++), matrix1.get(i, j));
  }
  {
    index_type k {0};
    for (index_type j = 0; j < matrix2.cols(); ++j)
      for (index_type i = 0; i < matrix2.rows(); ++i)
        BOOST_CHECK_EQUAL(values1.at(k++), matrix2.get(i, j));
  }
}


/**
 * @brief
 *         Unit test for the passive `passOverRegion` member function.
 *
 * This test checks that:
 *
 * - the correct region is shown to the visitor
 * - transposition is applied correctly
 * - each element is visited exactly once
 * - the visited matrix is `const` but the visiter may mutate
 *
 */
BOOST_AUTO_TEST_CASE(pass_over_region__passive)
{
  typedef int value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using extractor = bsc_klammler::test::Extractor<value_type>;
  const full_matrix data {
    {11, 12, 13, 14, 15, 16},
    {21, 22, 23, 24, 25, 26},
    {31, 32, 33, 34, 35, 36},
    {41, 42, 43, 44, 45, 46},
    {51, 52, 53, 54, 55, 56},
  };
  const full_matrix data_trans {
    {11, 21, 31, 41, 51},
    {12, 22, 32, 42, 52},
    {13, 23, 33, 43, 53},
    {14, 24, 34, 44, 54},
    {15, 25, 35, 45, 55},
    {16, 26, 36, 46, 56},
  };
  const full_matrix expected {
    { 0,  0,  0,  0,  0,  0,  0},
    { 0,  0,  0,  0,  0,  0,  0},
    { 0,  0,  0,  0,  0,  0,  0},
    { 0,  0,  0, 23, 24, 25, 26},
    { 0,  0,  0, 33, 34, 35, 36},
    { 0,  0,  0, 43, 44, 45, 46},
    { 0,  0,  0,  0,  0,  0,  0},
  };
  {
    full_matrix actual {expected.rows(), expected.cols()};
    extractor pv {actual};
    BOOST_CHECKPOINT("passive not transposed");
    data.passOverRegion(3, 3, false, pv, 1, 2, 4, 6);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    full_matrix actual {expected.rows(), expected.cols()};
    extractor pv {actual};
    BOOST_CHECKPOINT("passive transposed");
    data_trans.passOverRegion(3, 3, true, pv, 1, 2, 4, 6);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
}


/**
 * @brief
 *         Unit test for the active `passOverRegion` member function.
 *
 * This test checks that:
 *
 * - the correct region is shown to the visitor
 * - each element is visited exactly once
 * - the visited matrix is not `const` and the visiter may mutate
 *
 */
BOOST_AUTO_TEST_CASE(pass_over_region__active)
{
  typedef int value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using inserter = bsc_klammler::test::Inserter<value_type>;
  const full_matrix data {
    {11, 12, 13, 14, 15, 16, 17},
    {21, 22, 23, 24, 25, 26, 27},
    {31, 32, 33, 34, 35, 36, 37},
    {41, 42, 43, 44, 45, 46, 47},
    {51, 52, 53, 54, 55, 56, 57},
    {61, 62, 63, 64, 65, 66, 67},
    {71, 72, 73, 74, 75, 76, 77},
  };
  const full_matrix expected {
    { 0,  0,  0,  0,  0,  0},
    { 0,  0, 44, 45, 46, 47},
    { 0,  0, 54, 55, 56, 57},
    { 0,  0, 64, 65, 66, 67},
    { 0,  0,  0,  0,  0,  0},
  };
  const full_matrix trans_expected = {
    { 0,  0,  0,  0,  0},
    { 0,  0,  0,  0,  0},
    { 0, 44, 54, 64,  0},
    { 0, 45, 55, 65,  0},
    { 0, 46, 56, 66,  0},
    { 0, 47, 57, 67,  0},
  };
  {
    full_matrix actual {expected.rows(), expected.cols()};
    inserter av {data};
    BOOST_CHECKPOINT("active not transposed");
    actual.passOverRegion(3, 3, false, av, 1, 2, 4, 6);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    full_matrix trans_actual {trans_expected.rows(), trans_expected.cols()};
    inserter av {data};
    BOOST_CHECKPOINT("active transposed");
    trans_actual.passOverRegion(3, 3, true, av, 1, 2, 4, 6);
    MY_CHECK_MATRIX_EQUAL(trans_expected, trans_actual);
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the active `passOverRegion` member
 *         function.
 *
 * This test checks that:
 *
 * - invalid regions are rejected
 * - rejection happens before the visitor is invoked
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(pass_over_region__invalid,
                              visitor_type,
                              mpl_double_visitors)
{
  using value_type = typename visitor_type::value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  full_matrix matrix {};
  full_matrix dummy {};
  visitor_type vtor {dummy};
  try
    {
      matrix.passOverRegion(0, 0, false, vtor, 0, 0, 0, 1);
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument& e) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the `getRawMatrix` member function and the associated
 *         copy and move constructors.
 *
 * This test checks that:
 *
 * - the matrix returned by `getRawMatrix` has the same data as the
 *   `FullMatrix`
 * - changes to the `FullMatrix` are visible in the raw matrix
 * - the copy constructor works correctly and has *copy* semantics
 * - the move constructor works correctly and has *move* semantics
 *
 */
BOOST_AUTO_TEST_CASE(raw_data)
{
  typedef Foo value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using raw_matrix = full_matrix::raw_matrix_type;
  Foo::activate();
  {
    // empty matrix
    const full_matrix full {};
    const raw_matrix& raw {full.getRawMatrix()};
    BOOST_CHECK_EQUAL(0, raw.rows());
    BOOST_CHECK_EQUAL(0, raw.cols());
  }
  {
    // same data
    const full_matrix full {{11}, {21}};
    const raw_matrix& raw {full.getRawMatrix()};
    BOOST_CHECK_EQUAL(2, raw.rows());
    BOOST_CHECK_EQUAL(1, raw.cols());
    BOOST_CHECK_EQUAL(full.get(0, 0), raw(0, 0));
    BOOST_CHECK_EQUAL(full.get(1, 0), raw(1, 0));
  }
  {
    // write-through
    const auto check = [](full_matrix& fm, const raw_matrix& rm)->bool{
      if (rm(0, 0) != 0) { return false; }
      fm.set(2, 0, 0);
      if (rm(0, 0) != 2) { return false; }
      return true;
    };
    full_matrix full {1, 1};
    BOOST_CHECK(check(full, full.getRawMatrix()));
  }
  {
    // copy c'tor
    const full_matrix first {
      {11, 12, 13, 14},
      {21, 22, 23, 24},
      {31, 32, 33, 34},
    };
    full_matrix second {first.getRawMatrix()};
    MY_CHECK_MATRIX_EQUAL(first, second);
    second.set(-1, 1, 1);
    BOOST_CHECK_EQUAL(22, first.get(1, 1));
    BOOST_CHECK_EQUAL(-1, second.get(1, 1));
  }
  {
    // move c'tor
    const full_matrix first {
      {11, 12, 13, 14},
      {21, 22, 23, 24},
      {31, 32, 33, 34},
    };
    raw_matrix raw {first.getRawMatrix()};  // make a copy
    Foo::deactivate();
    const full_matrix second {std::move(raw)};
    Foo::activate();
    MY_CHECK_MATRIX_EQUAL(first, second);
    BOOST_CHECK_EQUAL(0, raw.rows());
    BOOST_CHECK_EQUAL(0, raw.cols());
  }
}


/**
 * @brief
 *         Additional unit test for the polymorphic copy constructor that
 *         exercises the visitor more thoroughly.
 *
 * This test breaks modularity of this file (we need other `HMatrix` classes)
 * but I se no other way to test this.
 *
 */
BOOST_AUTO_TEST_CASE(full_matrix_ctor_copy_poly)
{
  typedef int value_type;
  using arranged_matrix = bsc_klammler::ArrangedMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using matrix_view = bsc_klammler::ConstMatrixView<value_type>;
  using mock_matrix = bsc_klammler::test::MockHMatrix<value_type>;
  using index_type = h_matrix::index_type;
  const full_matrix expected = {
    { 0,  0,  0,  0,  0,  0,  0,  0,  0, 11, 12,  0},
    { 0,  0,  0,  0,  0,  0,  0,  0,  0, 21, 22,  0},
    { 0, 11, 12, 13,  0,  0, 11, 12, 13, 31, 32,  0},
    { 0, 21, 22, 23,  0,  0, 21, 22, 23,  0,  0,  0},
    { 0,  0,  0,  0,  0,  0, 31, 32, 33,  0,  0,  0},
    { 0,  0, 11, 12,  0,  0, 41, 42, 43,  0,  0,  0},
    { 0,  0, 21, 22,  0,  0,  0,  0,  0,  0,  0, 11},
    { 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 12},
  };
  arranged_matrix proto {expected.rows(), expected.cols()};
  const full_matrix tile1 = {
    {11, 12, 13},
    {21, 22, 23},
  };
  proto.addTile(tile1, 2, 1);
  const full_matrix tile2 = {
    {11, 21, 31, 41},
    {12, 22, 32, 42},
    {13, 23, 33, 43},
  };
  proto.addTile(std::unique_ptr<h_matrix> {new matrix_view {
        matrix_view::makeTransposedView(tile2)}}, 2, 6);
  mock_matrix * tile3_ptr {new mock_matrix {}};
  tile3_ptr->mockRows([]()->index_type{ return 2; });
  tile3_ptr->mockCols([]()->index_type{ return 2; });
  tile3_ptr->mockGet([](index_type i, index_type j)->index_type{
      return 10 * (i + 1) + j + 1;});
  proto.addTile(std::unique_ptr<h_matrix> {tile3_ptr}, 5, 2);
  tile3_ptr = nullptr;
  const full_matrix tile4 = {
    {-1, -1, -1, -1},
    {-1, -1, -1, -1},
    {-1, 11, 21, 31},
    {-1, 12, 22, 32},
    {-1, 13, 23, 33},
  };
  proto.addTile(std::unique_ptr<h_matrix> {
      new matrix_view {&tile4, 2, 1, 2, 1}}, 6, 11);
  mock_matrix tile5 {};
  tile5.mockRows([]()->index_type{ return 7; });
  tile5.mockCols([]()->index_type{ return 8; });
  tile5.mockGet([](index_type i, index_type j)->index_type{
      return 10 * (j + 0) + i - 1;});
  proto.addTile(std::unique_ptr<h_matrix> {new matrix_view {
        &tile5, 1, 2, 3, 2, true}}, 0, 9);
  MY_CHECK_MATRIX_EQUAL(expected, proto);
  const full_matrix actual {proto};
  MY_CHECK_MATRIX_EQUAL(expected, actual);
  std::cout << "\nproto:\n\n" << proto;
  std::cout << "\nexpected:\n\n" << expected;
  std::cout << "\nactual:\n\n" << actual;
}
