/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_full_vector.cpp
 *
 * @brief
 *         Unit tests for the `FullVector` template class.
 *
 */

#include "../global.hpp"       // (package configuration)
#include "../full_vector.hpp"  // bsc_klammler::FullVector

#include <array>               // std::array
#include <complex>             // std::complex
#include <string>              // std::string
#include <typeinfo>            // typeid
#include <utility>             // std::move, std::swap, std::pair

#include "../random.hpp"       // bsc_klammler::random_real
#include "foo.hpp"             // bsc_klammler::test::Foo
#include "mock_h_vector.hpp"   // bsc_klammler::test::MockHVector
#include "util.hpp"            // MY_CHECK_VECTOR_EQUAL

#define BOOST_TEST_MODULE full_vector
#include "module.hpp"

using bsc_klammler::random_int;
using bsc_klammler::random_real;
using bsc_klammler::test::Foo;


/**
 * @brief
 *         Unit test for the member typdefs.
 *
 * This test checks that:
 *
 * - the member types `value_type`, `index_type` and `super_type` are defined
 *   correctly
 *
 */
BOOST_AUTO_TEST_CASE(member_typedefs)
{
  // There is no overload for 'operator<<(std::ostream&, std::type_info)' so we
  // cannot use 'BOOST_CHECK_EQUAL'.
  struct Bar {};
  BOOST_CHECK(typeid(bsc_klammler::HVector<Bar>) ==
              typeid(bsc_klammler::FullVector<Bar>::super_type));
  BOOST_CHECK(typeid(Bar) ==
              typeid(bsc_klammler::FullVector<Bar>::value_type));
  for (typename bsc_klammler::FullVector<Bar>::index_type i = 0; i < 10; ++i)
    continue;
}


namespace /* anonymous */
{
  // Helper template to implement checks for the 'construction' test case with
  // static polymorphism.
  template<typename T>
  static void
  construction_check_(const typename bsc_klammler::FullVector<T>::index_type size)
  {
    using index_type = typename bsc_klammler::FullVector<T>::index_type;
    const bsc_klammler::FullVector<T> vector {size};
    BOOST_CHECK_EQUAL(size, vector.size());
    for (index_type i = 0; i < size; ++i)
      {
        BOOST_CHECK_EQUAL(bsc_klammler::FullVector<T>::zero(), vector.get(i));
      }
  }
}  // namespace /* anonymous */


/**
 * @brief
 *         Unit test for the default constructor.
 *
 * This test checks that:
 *
 * - the constructor creates an empty vector
 *
 */
BOOST_AUTO_TEST_CASE(ctor_default)
{
  {
    const bsc_klammler::FullVector<double> vec {};
    BOOST_CHECK_EQUAL(0, vec.size());
  }
  {
    const bsc_klammler::FullVector<int> vec {};
    BOOST_CHECK_EQUAL(0, vec.size());
  }
  {
    using index_type = bsc_klammler::FullVector<double>::index_type;
    const bsc_klammler::FullVector<index_type> vec {};
    BOOST_CHECK_EQUAL(0, vec.size());
  }
}


/**
 * @brief
 *         Unit test for the constructor taking the dimension as argument.
 *
 * This test checks that:
 *
 * - the dimension of the created vector are correct
 * - (0) vectors are supported
 * - the elements are value-initialized
 *
 * This is also a unit test for the `size` member function.
 *
 */
BOOST_AUTO_TEST_CASE(ctor)
{
  using index_type = typename bsc_klammler::FullVector<int>::index_type;
  const index_type min {1};
  const index_type max {128};
  for (const auto empty : {true, false})
    {
      const auto d = [empty]()->index_type{
        return empty ? 0 : random_int(min, max);
      };
      const index_type n {d()};
      construction_check_<bool>(n);
      construction_check_<char>(n);
      construction_check_<signed>(n);
      construction_check_<unsigned>(n);
      construction_check_<float>(n);
      construction_check_<double>(n);
      construction_check_<std::complex<double>>(n);
      construction_check_<std::string>(n);
    }
}


/**
 * @brief
 *         Unit test for the initializer-list constructor.
 *
 * This test checks that:
 *
 * - the constructor creates the correct `FullVector`
 *
 */
BOOST_AUTO_TEST_CASE(ctor_initializer_list)
{
  {
    using full_vector = bsc_klammler::FullVector<char>;
    const full_vector vec {{'g', 'a', 'm', 'm', 'a'}};
    BOOST_CHECK_EQUAL(5, vec.size());
    BOOST_CHECK_EQUAL('g', vec.get(0));
    BOOST_CHECK_EQUAL('a', vec.get(1));
    BOOST_CHECK_EQUAL('m', vec.get(2));
    BOOST_CHECK_EQUAL('m', vec.get(3));
    BOOST_CHECK_EQUAL('a', vec.get(4));
  }
  {
    using full_vector = bsc_klammler::FullVector<int>;
    const full_vector vec {{2, 4, 7, 7, 3}};
    BOOST_CHECK_EQUAL(5, vec.size());
    BOOST_CHECK_EQUAL(2, vec.get(0));
    BOOST_CHECK_EQUAL(4, vec.get(1));
    BOOST_CHECK_EQUAL(7, vec.get(2));
    BOOST_CHECK_EQUAL(7, vec.get(3));
    BOOST_CHECK_EQUAL(3, vec.get(4));
  }
  {
    using full_vector = bsc_klammler::FullVector<double>;
    const full_vector vec {{0.2, 0.7, 0.1, 0.8, 0.2, 0.8}};
    BOOST_CHECK_EQUAL(6, vec.size());
    BOOST_CHECK_EQUAL(0.2, vec.get(0));
    BOOST_CHECK_EQUAL(0.7, vec.get(1));
    BOOST_CHECK_EQUAL(0.1, vec.get(2));
    BOOST_CHECK_EQUAL(0.8, vec.get(3));
    BOOST_CHECK_EQUAL(0.2, vec.get(4));
    BOOST_CHECK_EQUAL(0.8, vec.get(5));
  }
  {
    using full_vector = bsc_klammler::FullVector<int>;
    const full_vector v1 {};     // default c'tor
    const full_vector v2 {{}};   // empty initializer list
    const full_vector v3 {{3}};  // values == {3}
    const full_vector v4 {3};    // size == 3
    BOOST_CHECK_EQUAL(0, v1.size());
    BOOST_CHECK_EQUAL(0, v2.size());
    BOOST_CHECK_EQUAL(1, v3.size());
    BOOST_CHECK_EQUAL(3, v4.size());
    BOOST_CHECK_EQUAL(3, v3.get(0));
    BOOST_CHECK_EQUAL(0, v4.get(0));
    BOOST_CHECK_EQUAL(0, v4.get(1));
    BOOST_CHECK_EQUAL(0, v4.get(2));
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
BOOST_AUTO_TEST_CASE(ctor_initializer_list__invalid)
{
  try
    {
      const bsc_klammler::FullVector<char> bad {
        {'n', 'o', 't'},
        {'a'},
        {'v', 'e', 'c', 't', 'o', 'r'},
      };
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for copy and move construction and assignment.
 *
 * This test checks that:
 *
 * - the copy constructor has *copy* semantics
 * - the move constructor has *move* semantics
 * - the copy-assignment operator has *copy* semantics
 * - the move-assignment operator has *move* semantics
 * - the `clone` friend function has *copy* semantics
 *
 */
BOOST_AUTO_TEST_CASE(copy_move)
{
#define VALUES_10  84, 9, 30, 77, 51, 11, 34, 35, 29, 54
  typedef Foo value_type;
  using h_vector = bsc_klammler::HVector<value_type>;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using index_type = full_vector::index_type;
  const auto check_are_copies = [](h_vector& first,
                                   h_vector& second)->int{
    if (first.size() != second.size()) { return 1; }
    for (index_type i = 0; i < first.size(); ++i)
      {
        const value_type v1 {first.get(i)};
        const value_type v2 {second.get(i)};
        if (v1 != v2) { return 2; }
        {
          first.set(~v1, i);
          if (second.get(i) != v2) { return 3; }
          first.set(v1, i);
        }
        {
          second.set(~v2, i);
          if (first.get(i) != v1) { return 4; }
          second.set(v2, i);
        }
      }
    return 0;
  };
  Foo::activate();
  const full_vector values10 {{VALUES_10}};
  {
    // copy c'tor
    full_vector vec1 {{VALUES_10}};
    full_vector vec2 {const_cast<const full_vector&>(vec1)};
    MY_CHECK_VECTOR_EQUAL(values10, vec1);
    MY_CHECK_VECTOR_EQUAL(values10, vec2);
    BOOST_CHECK_EQUAL(0, check_are_copies(vec1, vec2));
  }
  {
    // move c'tor
    full_vector vec1 {{VALUES_10}};
    Foo::deactivate();
    full_vector vec2 {std::move(vec1)};
    Foo::activate();
    BOOST_CHECK_EQUAL(0, vec1.size());
    MY_CHECK_VECTOR_EQUAL(values10, vec2);
  }
  {
    // copy-assignment op'tor
    full_vector vec1 {{VALUES_10}};
    full_vector vec2 {};
    vec2 = const_cast<const full_vector&>(vec1);
    MY_CHECK_VECTOR_EQUAL(values10, vec1);
    MY_CHECK_VECTOR_EQUAL(values10, vec2);
    BOOST_CHECK_EQUAL(0, check_are_copies(vec1, vec2));
  }
  {
    // move-assignment op'tor
    full_vector vec1 {{VALUES_10}};
    full_vector vec2 {};
    Foo::deactivate();
    vec2 = std::move(vec1);
    Foo::activate();
    BOOST_CHECK_EQUAL(0, vec1.size());
    MY_CHECK_VECTOR_EQUAL(values10, vec2);
  }
  {
    // polymorphic copy c'tor
    const bsc_klammler::test::MockHVector<value_type> mock {};
    mock.mockSize([]()->index_type{ return 10; });
    mock.mockGet([](const index_type i)->value_type{
        const std::array<value_type, 10> values = {VALUES_10};
        return values.at(i);
    });
    const full_vector copy {mock};
    MY_CHECK_VECTOR_EQUAL(values10, mock);
    MY_CHECK_VECTOR_EQUAL(values10, copy);
  }
  {
    // clone()
    full_vector vec1 {{VALUES_10}};
    std::unique_ptr<h_vector> vec2_uptr {vec1.clone()};
    MY_CHECK_VECTOR_EQUAL(values10, vec1);
    MY_CHECK_VECTOR_EQUAL(values10, *vec2_uptr);
    BOOST_CHECK_EQUAL(0, check_are_copies(vec1, *vec2_uptr));
  }
#undef VALUES_10
}


/**
 * @brief
 *         Unit test for the `getRawVector` member function and the associated
 *         copy and move constructors.
 *
 * This test checks that:
 *
 * - the vector returned by `getRawVector` has the same data as the
 *   `FullVector`
 * - changes to the `FullVector` are visible in the raw vector
 * - the copy constructor works correctly and has *copy* semantics
 * - the move constructor works correctly and has *move* semantics
 *
 */
BOOST_AUTO_TEST_CASE(raw_data)
{
  typedef Foo value_type;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using raw_vector = full_vector::raw_vector_type;
  Foo::activate();
  {
    // empty vector
    const full_vector full {};
    const raw_vector& raw {full.getRawVector()};
    BOOST_CHECK_EQUAL(0, raw.size());
  }
  {
    // same data
    const full_vector full {{11, 12, 13}};
    const raw_vector& raw {full.getRawVector()};
    BOOST_CHECK_EQUAL(3, raw.size());
    BOOST_CHECK_EQUAL(full.get(0), raw(0));
    BOOST_CHECK_EQUAL(full.get(1), raw(1));
    BOOST_CHECK_EQUAL(full.get(2), raw(2));
  }
  {
    // write-through
    const auto check = [](full_vector& fv, const raw_vector& rv)->bool{
      if (rv(0) != 0) { return false; }
      fv.set(2, 0);
      if (rv(0) != 2) { return false; }
      return true;
    };
    full_vector full {3};
    BOOST_CHECK(check(full, full.getRawVector()));
  }
  {
    // copy c'tor
    const full_vector first {{11, 12, 13, 14, 15, 16, 17}};
    full_vector second {first.getRawVector()};
    MY_CHECK_VECTOR_EQUAL(first, second);
    second.set(-1, 1);
    BOOST_CHECK_EQUAL(12, first.get(1));
    BOOST_CHECK_EQUAL(-1, second.get(1));
  }
  {
    // move c'tor
    const full_vector first {{11, 12, 13, 14, 15, 16, 17}};
    raw_vector raw {first.getRawVector()};  // make a copy
    Foo::deactivate();
    const full_vector second {std::move(raw)};
    Foo::activate();
    MY_CHECK_VECTOR_EQUAL(first, second);
    BOOST_CHECK_EQUAL(0, raw.size());
  }
}
