/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_h_matrix.cpp
 *
 * @brief
 *         Unit tests for the `h_matrix.hpp` header.
 *
 */

#include "../global.hpp"       // (package configuration)
#include "../h_matrix.hpp"     // bsc_klammler::HMatrix

#include <limits>              // std::numeric_limits
#include <string>              // std::string
#include <typeinfo>            // typeid
#include <utility>             // std::pair

#include "../random.hpp"       // bsc_klammler::random_real
#include "mock_h_matrix.hpp"   // bsc_klammler::test::MockHMatrix
#include "util.hpp"            // MY_CHECK_MATRIX_EQUAL

#define BOOST_TEST_MODULE h_matrix
#include "module.hpp"

#ifndef TOL_DBL
// We only do single +/- operations here, so a very small tolerance is
// warranted.
#define TOL_DBL 1.0E-15
#endif

/** @brief Tolerance for considering two `double`s close (in percent). */
static constexpr double tol_double_percent {100 * TOL_DBL};

using bsc_klammler::random_int;
using bsc_klammler::random_real;


/**
 * @brief
 *         Unit test for the member typdefs.
 *
 * This test checks that:
 *
 * - the member types `value_type` and `index_type` are defined correctly
 *
 */
BOOST_AUTO_TEST_CASE(member_typedefs)
{
  // There is no overload for 'operator<<(std::ostream&, std::type_info)' so we
  // cannot use 'BOOST_CHECK_EQUAL'.
  struct Bar {};
  BOOST_CHECK(typeid(Bar) ==
              typeid(bsc_klammler::HMatrix<Bar>::value_type));
  for (typename bsc_klammler::HMatrix<Bar>::index_type i = 0; i < 10; ++i)
    continue;
}


/**
 * @brief
 *         Unit test for the `zero` static member function.
 *
 * This test checks that:
 *
 * - `zero()` returns zero
 * - the returned type is correct
 *
 */
BOOST_AUTO_TEST_CASE(zero)
{
  struct Bar
  {
    int bar;
    constexpr Bar() noexcept : bar {-41} {}
  };
  BOOST_CHECK(typeid(bool) == typeid(bsc_klammler::HMatrix<bool>::zero()));
  BOOST_CHECK_EQUAL(false, bsc_klammler::HMatrix<bool>::zero());
  BOOST_CHECK(typeid(int) == typeid(bsc_klammler::HMatrix<int>::zero()));
  BOOST_CHECK_EQUAL(0, bsc_klammler::HMatrix<int>::zero());
  BOOST_CHECK(typeid(char) == typeid(bsc_klammler::HMatrix<char>::zero()));
  BOOST_CHECK_EQUAL('\0', bsc_klammler::HMatrix<char>::zero());
  BOOST_CHECK(typeid(float) == typeid(bsc_klammler::HMatrix<float>::zero()));
  BOOST_CHECK_EQUAL(0.0f, bsc_klammler::HMatrix<float>::zero());
  BOOST_CHECK(typeid(double) == typeid(bsc_klammler::HMatrix<double>::zero()));
  BOOST_CHECK_EQUAL(0.0, bsc_klammler::HMatrix<double>::zero());
  BOOST_CHECK(typeid(Bar) == typeid(bsc_klammler::HMatrix<Bar>::zero()));
  BOOST_CHECK_EQUAL(-41, bsc_klammler::HMatrix<Bar>::zero().bar);
}


/**
 * @brief
 *         Unit test for the `throwMismatchedDimensions` static member
 *         function.
 *
 * This test checks that:
 *
 * - an appropriate exception is thrown
 *
 */
BOOST_AUTO_TEST_CASE(throw_mismatched_dimensions)
{
  using mockup = bsc_klammler::test::MockHMatrix<double>;
  using index_type = mockup::index_type;
  const mockup first {};
  const mockup second {};
  first.mockRows([]()->index_type{return 14;});
  first.mockCols([]()->index_type{return 92;});
  second.mockRows([]()->index_type{return 7;});
  second.mockCols([]()->index_type{return 8;});
  try
    {
      mockup::throwMismatchedDimensions("pull", first, "out of", second);
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument& e)
    {
      const std::string expected {
        "cannot pull a (14, 92) out of a (7, 8) matrix"
      };
      const std::string actual {e.what()};
      BOOST_CHECK_EQUAL(expected, actual);
    }
}


/**
 * @brief
 *         Unit test for the `checkIndex` member function.
 *
 * This test checks that:
 *
 * - the check is correct
 * - an appropriate exception is thrown
 *
 */
BOOST_AUTO_TEST_CASE(check_index)
{
  using mockup = bsc_klammler::test::MockHMatrix<double>;
  using index_type = mockup::index_type;
  {
    const index_type rows {random_int<index_type>(2, 8)};
    const index_type cols {random_int<index_type>(2, 8)};
    const mockup mock {};
    mock.mockRows([rows]()->index_type{return rows;});
    mock.mockCols([cols]()->index_type{return cols;});
    for (index_type j = 0; j < cols; ++j)
      for (index_type i = 0; i < rows; ++i)
        mock.checkIndex(i, j);
    for (index_type i = 0; i <= rows; ++i)
      {
        try
          {
            mock.checkIndex(i, cols + random_int<index_type>(0, 3));
            BOOST_ERROR("no exception thrown");
          }
        catch (const std::invalid_argument&) { /* good */ }
      }
    for (index_type j = 0; j <= cols; ++j)
      {
        try
          {
            mock.checkIndex(rows + random_int<index_type>(0, 3), j);
            BOOST_ERROR("no exception thrown");
          }
        catch (const std::invalid_argument&) { /* good */ }
      }
  }
  {
    const mockup mock {};
    mock.mockRows([]()->index_type{return 17;});
    mock.mockCols([]()->index_type{return 23;});
    try
      {
        mock.checkIndex(10, 100);
        BOOST_ERROR("no exception thrown");
      }
    catch (const std::invalid_argument& e)
      {
        const std::string expected {
          "invalid index (10, 100) for a (17, 23) matrix"
        };
        const std::string actual {e.what()};
        BOOST_CHECK_EQUAL(expected, actual);
      }
  }
}


/**
 * @brief
 *         Unit test for the `ctorCheckSize` static member function.
 *
 * This test checks that:
 *
 * - obviously bad dimensions are rejected
 *
 */
BOOST_AUTO_TEST_CASE(ctor_check_size)
{
  using mockup = bsc_klammler::test::MockHMatrix<double>;
  using index_type = mockup::index_type;
  constexpr index_type good_large_size {
    std::numeric_limits<index_type>::max() >> 2
  };
  constexpr index_type bad_large_size {
    std::numeric_limits<index_type>::max()
  };
  const std::pair<index_type, index_type> good_dimensions[] = {
    {0, 0},
    {1, 0},
    {0, 1},
    {1, 1},
    {1, good_large_size},
    {good_large_size, 1},
    {good_large_size, good_large_size},
    {
      random_int<index_type>(1, good_large_size),
      random_int<index_type>(1, good_large_size),
    },
  };
  const std::pair<index_type, index_type> bad_dimensions[] = {
    {0, -1},
    {-1, 0},
    {1, -1},
    {-1, 1},
    {-1, -1},
    {bad_large_size, 1},
    {1, bad_large_size},
    {bad_large_size, bad_large_size},
  };
  for (const auto iter : good_dimensions)
    {
      mockup::ctorCheckSize(iter.first, iter.second);
    }
  for (const auto iter : bad_dimensions)
    {
      try
        {
          mockup::ctorCheckSize(iter.first, iter.second);
          BOOST_ERROR("no exception thrown");
        }
      catch (const std::invalid_argument&) { /* good */ }
    }
}


/**
 * @brief
 *         Unit test for the `size` member function.
 *
 * This test checks that:
 *
 * - the function returns the correct size for square matrices
 *
 */
BOOST_AUTO_TEST_CASE(size)
{
  // TODO: How can we check the non-square case?  Is there a chance to mock
  //       the 'assert' macro?
  using mockup = bsc_klammler::test::MockHMatrix<double>;
  using index_type = mockup::index_type;
  constexpr index_type size {56};
  unsigned rows_cnt {};
  unsigned cols_cnt {};
  const mockup mock {};
  mock.mockRows([&rows_cnt]()->index_type{++rows_cnt; return size;});
  mock.mockCols([&cols_cnt]()->index_type{++cols_cnt; return size;});
  for (unsigned i = 0; i < 10; ++i)
    {
      const unsigned rows_cnt_before {rows_cnt};
      const unsigned cols_cnt_before {cols_cnt};
      BOOST_CHECK_EQUAL(size, mock.size());
      if (BSC_KLAMMLER_DEBUG)
        {
          BOOST_CHECK_EQUAL(rows_cnt_before + 1, rows_cnt);
          BOOST_CHECK_EQUAL(cols_cnt_before + 1, cols_cnt);
        }
      else
        {
          if (rows_cnt == rows_cnt_before)
            {
              BOOST_CHECK_EQUAL(cols_cnt_before + 1, cols_cnt);
            }
          else if (cols_cnt == cols_cnt_before)
            {
              BOOST_CHECK_EQUAL(rows_cnt_before + 1, rows_cnt);
            }
          else
            {
              BOOST_ERROR("rows_cnt_before = " << rows_cnt_before << ", "
                          << "rows_cnt = " << rows_cnt << ", "
                          << "cols_cnt_before = " << cols_cnt_before << ", "
                          << "cols_cnt = " << cols_cnt);
            }
        }
    }
}


/**
 * @brief
 *         Unit test for the `checkRegion` member function.
 *
 * This test checks that:
 *
 * - valid regions are accepted
 * - invalid regions are rejected
 *
 */
BOOST_AUTO_TEST_CASE(check_region)
{
  typedef int value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using index_type = h_matrix::index_type;
  using mock_matrix = typename bsc_klammler::test::MockHMatrix<value_type>;
  struct parameters
  {
    index_type rows;
    index_type cols;
    index_type north;
    index_type west;
    index_type south;
    index_type east;
  };
  const index_type rows = random_int<index_type>(2, 32);
  const index_type cols = random_int<index_type>(2, 32);
  const parameters valid_parameters[] = {
    parameters {0, 0, 0, 0, 0, 0},
    parameters {rows, cols, 0, 0, 0, 0},
    parameters {rows, cols, 0, 0, 0, cols},
    parameters {rows, cols, 0, 0, rows, 0},
    parameters {rows, cols, 0, 0, rows, cols},
    parameters {rows, cols, rows / 2, cols / 2, rows / 2, cols / 2},
  };
  const parameters invalid_parameters[] = {
    parameters {0, 0, 0, 0, 0, 1},
    parameters {0, 0, 0, 0, 1, 0},
    parameters {0, 0, 0, 0, 1, 1},
    parameters {0, 0, 0, 1, 0, 1},
    parameters {0, 0, 1, 0, 1, 0},
    parameters {0, 0, 1, 1, 1, 1},
    parameters {0, 0, 1, 1, 2, 2},
    parameters {rows, cols, 0, 1, 0, 0},
    parameters {rows, cols, 1, 0, 0, 0},
    parameters {rows, cols, 1, 1, 0, 0},
    parameters {rows, cols, 0, 0, rows, cols + 1},
    parameters {rows, cols, 0, 0, rows + 1, cols},
    parameters {rows, cols, 0, 0, rows + 1, cols + 1},
    parameters {rows, cols, -1, 0, 0, 0},
    parameters {rows, cols, 0, -1, 0, 0},
    parameters {rows, cols, -1, -1, 0, 0},
    parameters {rows, cols, rows, 0, 0, 0},
    parameters {rows, cols, 0, cols, 0, 0},
    parameters {rows, cols, rows, cols, 0, 0},
    parameters {rows, cols, rows, 0, 1, 1},
    parameters {rows, cols, 0, cols, 1, 1},
  };
  for (const auto iter : valid_parameters)
    {
      for (const bool trans : {false, true})
        {
          const index_type north {!trans ? iter.north : iter.west};
          const index_type west  {!trans ? iter.west  : iter.north};
          const index_type south {!trans ? iter.south : iter.east};
          const index_type east  {!trans ? iter.east  : iter.south};
          mock_matrix matrix {};
          matrix.mockRows([&]()->index_type{ return iter.rows; });
          matrix.mockCols([&]()->index_type{ return iter.cols; });
          matrix.checkRegion(north, west, south, east, trans);
        }
    }
  for (const auto iter : invalid_parameters)
    {
      for (const bool trans : {false, true})
        {
          const index_type north {!trans ? iter.north : iter.west};
          const index_type west  {!trans ? iter.west  : iter.north};
          const index_type south {!trans ? iter.south : iter.east};
          const index_type east  {!trans ? iter.east  : iter.south};
          mock_matrix matrix {};
          matrix.mockRows([&]()->index_type{ return iter.rows; });
          matrix.mockCols([&]()->index_type{ return iter.cols; });
          try
            {
              matrix.checkRegion(north, west, south, east, trans);
              BOOST_ERROR("no exception thrown ("
                          << "rows = " << iter.rows << ", "
                          << "cols = " << iter.cols << ", "
                          << "north = " << iter.north << ", "
                          << "west = " << iter.west << ", "
                          << "south = " << iter.south << ", "
                          << "east = " << iter.east << ", "
                          << "trans = " << trans << ")");
            }
          catch (const std::invalid_argument&) { /* good */ }
        }
    }
}
