/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_h_vector.cpp
 *
 * @brief
 *         Unit tests for the `h_vector.hpp` header.
 *
 */

#include "../global.hpp"       // (package configuration)
#include "../h_vector.hpp"     // bsc_klammler::HVector

#include <limits>              // std::numeric_limits
#include <string>              // std::string
#include <typeinfo>            // typeid

#include "../random.hpp"       // bsc_klammler::random_real
#include "mock_h_vector.hpp"   // bsc_klammler::test::MockHVector
#include "util.hpp"            // MY_CHECK_VECTOR_CLOSE, ...

#define BOOST_TEST_MODULE h_vector
#include "module.hpp"

#ifndef TOL_DBL
// We only do single +/- operations here, so a very small tolerance is
// warranted.  (Actually, I'm surprised it needs to be that large...)
#define TOL_DBL 1.0E-13
#endif

/** @brief Tolerance for considering two `double`s close (in percent). */
static constexpr double tol_double_percent {100 * TOL_DBL};

using bsc_klammler::random_int;
using bsc_klammler::random_real;


/**
 * @brief
 *         Unit test for the member typdefs.
 *
 * This test checks that:
 *
 * - the member types `value_type` and `index_type` are defined correctly
 *
 */
BOOST_AUTO_TEST_CASE(member_typedefs)
{
  // There is no overload for 'operator<<(std::ostream&, std::type_info)' so we
  // cannot use 'BOOST_CHECK_EQUAL'.
  struct Bar {};
  BOOST_CHECK(typeid(Bar) ==
              typeid(bsc_klammler::HVector<Bar>::value_type));
  for (typename bsc_klammler::HVector<Bar>::index_type i = 0; i < 10; ++i)
    continue;
}


/**
 * @brief
 *         Unit test for the `zero` static member function.
 *
 * This test checks that:
 *
 * - `zero()` returns zero
 * - the returned type is correct
 *
 */
BOOST_AUTO_TEST_CASE(zero)
{
  struct Bar
  {
    int bar;
    constexpr Bar() noexcept : bar {-41} {}
  };
  BOOST_CHECK(typeid(bool) == typeid(bsc_klammler::HVector<bool>::zero()));
  BOOST_CHECK_EQUAL(false, bsc_klammler::HVector<bool>::zero());
  BOOST_CHECK(typeid(int) == typeid(bsc_klammler::HVector<int>::zero()));
  BOOST_CHECK_EQUAL(0, bsc_klammler::HVector<int>::zero());
  BOOST_CHECK(typeid(char) == typeid(bsc_klammler::HVector<char>::zero()));
  BOOST_CHECK_EQUAL('\0', bsc_klammler::HVector<char>::zero());
  BOOST_CHECK(typeid(float) == typeid(bsc_klammler::HVector<float>::zero()));
  BOOST_CHECK_EQUAL(0.0f, bsc_klammler::HVector<float>::zero());
  BOOST_CHECK(typeid(double) == typeid(bsc_klammler::HVector<double>::zero()));
  BOOST_CHECK_EQUAL(0.0, bsc_klammler::HVector<double>::zero());
  BOOST_CHECK(typeid(Bar) == typeid(bsc_klammler::HVector<Bar>::zero()));
  BOOST_CHECK_EQUAL(-41, bsc_klammler::HVector<Bar>::zero().bar);
}


/**
 * @brief
 *         Unit test for the `throwMismatchedDimensions` static member
 *         function.
 *
 * This test checks that:
 *
 * - an appropriate exception is thrown
 *
 */
BOOST_AUTO_TEST_CASE(throw_mismatched_dimensions)
{
  using mockup = bsc_klammler::test::MockHVector<double>;
  using index_type = mockup::index_type;
  const mockup first {};
  const mockup second {};
  first.mockSize([]()->index_type{return 41;});
  second.mockSize([]()->index_type{return 14;});
  try
    {
      mockup::throwMismatchedDimensions("really stuff", first, "into", second);
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument& e)
    {
      const std::string expected {
        "cannot really stuff a 41 into a 14 vector"
      };
      const std::string actual {e.what()};
      BOOST_CHECK_EQUAL(expected, actual);
    }
}


/**
 * @brief
 *         Unit test for the `checkIndex` member function.
 *
 * This test checks that:
 *
 * - the check is correct
 * - an appropriate exception is thrown
 *
 */
BOOST_AUTO_TEST_CASE(check_index)
{
  using mockup = bsc_klammler::test::MockHVector<double>;
  using index_type = mockup::index_type;
  {
    const index_type size {random_int<index_type>(2, 8)};
    const mockup mock {};
    mock.mockSize([size]()->index_type{return size;});
    for (index_type i = 0; i < size; ++i)
      mock.checkIndex(i);
    try
      {
        mock.checkIndex(size + random_int<index_type>(1, 3));
        BOOST_ERROR("no exception thrown");
      }
    catch (const std::invalid_argument&) { /* good */ }
  }
  {
    const mockup mock {};
    mock.mockSize([]()->index_type{return 17;});
    try
      {
        mock.checkIndex(100);
        BOOST_ERROR("no exception thrown");
      }
    catch (const std::invalid_argument& e)
      {
        const std::string expected {
          "invalid index 100 for a 17 vector"
        };
        const std::string actual {e.what()};
        BOOST_CHECK_EQUAL(expected, actual);
      }
  }
}


/**
 * @brief
 *         Unit test for the `ctorCheckSize` static member function.
 *
 * This test checks that:
 *
 * - obviously bad dimensions are rejected
 *
 */
BOOST_AUTO_TEST_CASE(ctor_check_size)
{
  using mockup = bsc_klammler::test::MockHVector<double>;
  using index_type = mockup::index_type;
  constexpr index_type good_large_size {
    std::numeric_limits<index_type>::max() >> 2
  };
  constexpr index_type bad_large_size {
    std::numeric_limits<index_type>::max()
  };
  const index_type good_dimensions[] = {0, 1, good_large_size};
  const index_type bad_dimensions[] = {bad_large_size};
  for (const auto iter : good_dimensions)
    {
      mockup::ctorCheckSize(iter);
    }
  for (const auto iter : bad_dimensions)
    {
      try
        {
          mockup::ctorCheckSize(iter);
          BOOST_ERROR("no exception thrown");
        }
      catch (const std::invalid_argument&) { /* good */ }
    }
}
