/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_inefficient_algorithm.cpp
 *
 * @brief
 *         Unit tests for the `inefficient_algorithm.hpp` header.
 *
 */

#include "../global.hpp"                 // (package configuration)
#include "../inefficient_algorithm.hpp"  // BSC_KLAMMLER_INEFFICIENT_ALGORITHM

#include <iostream>                      // std::cout, std::cerr

#define BOOST_TEST_MODULE inefficient_algorithm
#include "module.hpp"


namespace /* anonymous */
{
  /**
   * @brief
   *         RAII-style scope guard for reliably restoring the global policy.
   *
   */
  class PolicyGuard
  {
  private:

    /** @brief Global policy to restore. */
    const bsc_klammler::InefficientAlgorithm::Policies oldPolicy_;

  public:

    /**
     * @brief
     *         Constructor: Creates a new scope guard.
     *
     */
    PolicyGuard() :
      oldPolicy_ {bsc_klammler::InefficientAlgorithm::getPolicy()}
    {
      // do nothing
    }

    /**
     * @brief
     *         Destructor: Restores the old global bolicy.
     *
     */
    ~PolicyGuard()
    {
      bsc_klammler::InefficientAlgorithm::setPolicy(this->oldPolicy_);
    }
  };
}  // namespace /* anonymous */


/**
 * @brief
 *         Unit test for the policy management facility.
 *
 * This test checks that:
 *
 * - the default policy is `IGNORE`
 * - getting and setting the policy works
 * - attempting to set invalid policies is rejected
 *
 */
BOOST_AUTO_TEST_CASE(default_policy)
{
  using bsc_klammler::InefficientAlgorithm;
  const PolicyGuard pg {};
  constexpr InefficientAlgorithm::Policies policies[] = {
    InefficientAlgorithm::Policies::IGNORE,
    InefficientAlgorithm::Policies::WARN,
    InefficientAlgorithm::Policies::THROW,
  };
  {
    // default policy
    const auto expected = InefficientAlgorithm::Policies::IGNORE;
    const auto actual = InefficientAlgorithm::getPolicy();
    BOOST_CHECK_EQUAL(expected, actual);
  }
  {
    // setting and getting
    const PolicyGuard pg {};  // this one is rather useless here...
    for (const auto iter : policies)
      {
        InefficientAlgorithm::setPolicy(iter);
        BOOST_CHECK_EQUAL(iter, InefficientAlgorithm::getPolicy());
      }
  }
  if (BSC_KLAMMLER_CHECK_ARGS)
    {
      // invalid setting
      for (const auto iter : policies)
        {
          const PolicyGuard pg {};
          InefficientAlgorithm::setPolicy(iter);
          BOOST_CHECK_EQUAL(iter, InefficientAlgorithm::getPolicy());
          try
            {
              // As insiders, use our knowledge that this is an invalid value.
              constexpr InefficientAlgorithm::Policies invalid {
                static_cast<InefficientAlgorithm::Policies>(0)
              };
              InefficientAlgorithm::setPolicy(invalid);
              BOOST_ERROR("no exception thrown");
            }
          catch (const std::invalid_argument&) { /* good */ }
          BOOST_CHECK_EQUAL(iter, InefficientAlgorithm::getPolicy());
        }
    }
}


/**
 * @brief
 *         Unit test for the policy formatting operator.
 *
 * This test checks that:
 *
 * - all policies are formatted nicely
 * - invalid policies are rejcted
 *
 */
BOOST_AUTO_TEST_CASE(policy_output_operator)
{
  using bsc_klammler::InefficientAlgorithm;
  constexpr InefficientAlgorithm::Policies policies[] = {
    InefficientAlgorithm::Policies::IGNORE,
    InefficientAlgorithm::Policies::WARN,
    InefficientAlgorithm::Policies::THROW,
  };
  for (const auto iter : policies)
    {
      std::ostringstream oss {};
      oss << iter;
      const std::string str {oss.str()};
      BOOST_CHECK(!str.empty());
      // This is another insider:
      BOOST_CHECK_EQUAL(static_cast<int>(iter),
                        static_cast<int>(str.data()[0]) + 'a' - 'A');
    }
  if (BSC_KLAMMLER_CHECK_ARGS)
    {
      // As insiders, use our knowledge that this is an invalid value.
      constexpr InefficientAlgorithm::Policies invalid {
        static_cast<InefficientAlgorithm::Policies>(0)
      };
      std::ostringstream oss {};
      try
        {
          oss << invalid;
          BOOST_ERROR("no exception thrown");
        }
      catch (const std::invalid_argument&) { /* good */ }
      BOOST_CHECK(oss.str().empty());
    }
}


/**
 * @brief
 *         Unit test for the `IGNORE` policy.
 *
 * This test checks that:
 *
 * - the normal control flow is not interrupted
 * - no output is produced
 *
 */
BOOST_AUTO_TEST_CASE(policy_ignore)
{
  // TODO: The stream plumbing should really be done using RAII.
  constexpr auto pol = bsc_klammler::InefficientAlgorithm::Policies::IGNORE;
  const PolicyGuard pg {};
  const auto cout_rdbuf = std::cout.rdbuf(nullptr);
  const auto cerr_rdbuf = std::cerr.rdbuf(nullptr);
  std::ostringstream stdout {};
  std::ostringstream stderr {};
  std::cout.rdbuf(stdout.rdbuf());
  std::cerr.rdbuf(stderr.rdbuf());
  {
    bsc_klammler::InefficientAlgorithm::setPolicy(pol);
    bsc_klammler::InefficientAlgorithm {"ignore.cpp", 42};
  }
  std::cout.rdbuf(cout_rdbuf);
  std::cerr.rdbuf(cerr_rdbuf);
  BOOST_CHECK(stdout.str().empty());
  BOOST_CHECK(stderr.str().empty());
}


/**
 * @brief
 *         Unit test for the `WARN` policy.
 *
 * This test checks that:
 *
 * - the normal control flow is not interrupted
 * - something is printed to the standard error output
 *
 */
BOOST_AUTO_TEST_CASE(policy_warn)
{
  // TODO: The stream plumbing should really be done using RAII.
  constexpr auto pol = bsc_klammler::InefficientAlgorithm::Policies::WARN;
  const PolicyGuard pg {};
  const auto cout_rdbuf = std::cout.rdbuf(nullptr);
  const auto cerr_rdbuf = std::cerr.rdbuf(nullptr);
  std::ostringstream stdout {};
  std::ostringstream stderr {};
  std::cout.rdbuf(stdout.rdbuf());
  std::cerr.rdbuf(stderr.rdbuf());
  {
    bsc_klammler::InefficientAlgorithm::setPolicy(pol);
    bsc_klammler::InefficientAlgorithm {"warn.cpp", 42};
  }
  std::cout.rdbuf(cout_rdbuf);
  std::cerr.rdbuf(cerr_rdbuf);
  BOOST_CHECK(stdout.str().empty());
  BOOST_CHECK(!stderr.str().empty());
}


/**
 * @brief
 *         Unit test for the `BSC_KLAMMLER_INEFFICIENT_ALGORITHM` macro.
 *
 * This test checks that:
 *
 * - the macro has the same effect as instantiating an `InefficientAlgorithm`
 *   object
 *
 */
BOOST_AUTO_TEST_CASE(macro)
{
  const PolicyGuard pg {};
  constexpr auto pol_i = bsc_klammler::InefficientAlgorithm::Policies::IGNORE;
  constexpr auto pol_w = bsc_klammler::InefficientAlgorithm::Policies::WARN;
  constexpr auto pol_t = bsc_klammler::InefficientAlgorithm::Policies::THROW;
  bsc_klammler::InefficientAlgorithm::setPolicy(pol_i);
  BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
  bsc_klammler::InefficientAlgorithm::setPolicy(pol_w);
  BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
  bsc_klammler::InefficientAlgorithm::setPolicy(pol_t);
  try
    {
      BSC_KLAMMLER_INEFFICIENT_ALGORITHM;
      BOOST_ERROR("no exception thrown");
    }
  catch (const bsc_klammler::InefficientAlgorithm::Error& e)
    {
      BOOST_CHECK_EQUAL(__FILE__, e.getFile());
      BOOST_CHECK_EQUAL(__LINE__ - 6, e.getLine());
    }
}
