/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_matrix_anchor.cpp
 *
 * @brief
 *         Unit tests for the `MatrixAnchor` template class.
 *
 */

#include "../global.hpp"         // (package configuration)
#include "../matrix_anchor.hpp"  // bsc_klammler::MatrixAnchor

#include <stdexcept>             // std::invalid_argument
#include <utility>               // std::move

#include <boost/mpl/list.hpp>    // boost::mpl::list

#include "../full_matrix.hpp"    // bsc_klammler::MatrixView
#include "../random.hpp"         // bsc_klammler::random_real

#define BOOST_TEST_MODULE matrix_anchor
#include "module.hpp"

using bsc_klammler::random_int;
using bsc_klammler::random_real;


/** @brief List of immutable and mutable `MatrixAnchor`s for `int`s. */
using mpl_int_anchors = boost::mpl::list<bsc_klammler::MatrixAnchor<int, false>, bsc_klammler::MatrixAnchor<int, true>>;

/**
 * @brief
 *         Unit tests for the constructor and function members.
 *
 * This test checks that:
 *
 * - the constructor works
 * - the `rows`, `cols`, `north`, `south`, `west` and `east` function members
 *   return the correct values
 * - no ownership of the pointer is taken
 * - `MatrixAnchor`s can be copied and moved.
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(general, matrix_anchor, mpl_int_anchors)
{
  using index_type = typename matrix_anchor::index_type;
  using value_type = typename matrix_anchor::value_type;
  using full_type = bsc_klammler::FullMatrix<value_type>;
  {
    full_type matrix {};
    const matrix_anchor anchor {&matrix, 0, 0};
    BOOST_CHECK_EQUAL(&matrix, anchor.getMatrix());
    BOOST_CHECK_EQUAL(0, anchor.rows());
    BOOST_CHECK_EQUAL(0, anchor.cols());
    BOOST_CHECK_EQUAL(0, anchor.north());
    BOOST_CHECK_EQUAL(0, anchor.south());
    BOOST_CHECK_EQUAL(0, anchor.west());
    BOOST_CHECK_EQUAL(0, anchor.east());
  }
  {
    const index_type rows {random_int<index_type>(1, 32)};
    const index_type cols {random_int<index_type>(1, 32)};
    const index_type off_rows {random_int<index_type>(0, 1024)};
    const index_type off_cols {random_int<index_type>(0, 1024)};
    full_type matrix {rows, cols};
    const matrix_anchor anchor {&matrix, off_rows, off_cols};
    const matrix_anchor copy_constructed_anchor {anchor};
    const matrix_anchor copy_assigned_anchor = anchor;
    const matrix_anchor move_constructed_anchor {std::move(anchor)};
    const matrix_anchor move_assigned_anchor = std::move(anchor);
    const matrix_anchor *const anchors[] = {
      &anchor,
      &copy_constructed_anchor,
      &copy_assigned_anchor,
      &move_constructed_anchor,
      &move_assigned_anchor,
    };
    for (const matrix_anchor *const iter : anchors)
      {
        BOOST_CHECK_EQUAL(&matrix, iter->getMatrix());
        BOOST_CHECK_EQUAL(rows, iter->rows());
        BOOST_CHECK_EQUAL(cols, iter->cols());
        BOOST_CHECK_EQUAL(off_rows, iter->north());
        BOOST_CHECK_EQUAL(off_cols, iter->west());
        BOOST_CHECK_EQUAL(off_rows + rows, iter->south());
        BOOST_CHECK_EQUAL(off_cols + cols, iter->east());
      }
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-chcking Unit tests for the constructor.
 *
 * This test checks that:
 *
 * - `nullptr`s are rejected
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(ctor__invalid, matrix_anchor, mpl_int_anchors)
{
  try
    {
      matrix_anchor anchor {nullptr, 0, 0};
      BOOST_ERROR("no exception thrown");
    }
  catch (const std::invalid_argument&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit tests for the `overlap` free function.
 *
 * This test checks that:
 *
 * - the function is found via ADL
 * - the function returns the correct result
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(overlapping, matrix_anchor, mpl_int_anchors)
{
  using index_type = typename matrix_anchor::index_type;
  using value_type = typename matrix_anchor::value_type;
  using full_type = bsc_klammler::FullMatrix<value_type>;
  {
    full_type full1 {1, 1};
    full_type full2 {1, 1};
    const matrix_anchor anchor1 {&full1, 10, 20};
    const matrix_anchor anchor2 {&full1, 30, 40};
    BOOST_CHECK_EQUAL(false, overlap(anchor1, anchor2));
    BOOST_CHECK_EQUAL(false, overlap(anchor2, anchor1));
  }
  {
    full_type full1 {1, 1};
    full_type full2 {1, 1};
    const matrix_anchor anchor1 {&full1, 0, 0};
    const matrix_anchor anchor2 {&full2, 0, 0};
    BOOST_CHECK_EQUAL(true, overlap(anchor1, anchor2));
    BOOST_CHECK_EQUAL(true, overlap(anchor2, anchor1));
  }
  {
    full_type full1 {};
    full_type full2 {};
    const matrix_anchor anchor1 {&full1, 0, 0};
    const matrix_anchor anchor2 {&full2, 0, 0};
    BOOST_CHECK_EQUAL(false, overlap(anchor1, anchor2));
    BOOST_CHECK_EQUAL(false, overlap(anchor2, anchor1));
  }
  {
    full_type full1 {10, 10};
    full_type full2 {};
    const matrix_anchor anchor1 {&full1, 1, 1};
    const matrix_anchor anchor2 {&full2, 3, 3};
    BOOST_CHECK_EQUAL(false, overlap(anchor1, anchor2));
    BOOST_CHECK_EQUAL(false, overlap(anchor2, anchor1));
  }
  {
    full_type full1 {3, 5};
    full_type full2 {2, 2};
    const matrix_anchor anchor1 {&full1, 14, 92};
    const index_type above_and_below[] = {
      anchor1.north() - full2.rows(),
      anchor1.south(),
    };
    const index_type left_and_right[] = {
      anchor1.west() - full2.cols(),
      anchor1.east(),
    };
    const index_type not_above_and_below[] = {
      above_and_below[0] + 1,
      above_and_below[1] - 1,
    };
    const index_type not_left_and_right[] = {
      left_and_right[0] + 1,
      left_and_right[1] - 1,
    };
    for (index_type i : above_and_below)
      {
        for (index_type j = left_and_right[0]; j < left_and_right[1]; ++j)
          {
            const matrix_anchor anchor2 {&full2, i, j};
            BOOST_CHECK_EQUAL(false, overlap(anchor1, anchor2));
            BOOST_CHECK_EQUAL(false, overlap(anchor2, anchor1));
          }
      }
    for (index_type j : left_and_right)
      {
        for (index_type i = above_and_below[0]; i < above_and_below[1]; ++i)
          {
            const matrix_anchor anchor2 {&full2, i, j};
            BOOST_CHECK_EQUAL(false, overlap(anchor1, anchor2));
            BOOST_CHECK_EQUAL(false, overlap(anchor2, anchor1));
          }
      }
    for (index_type j = not_left_and_right[0]; j < not_left_and_right[1]; ++j)
      {
        for (index_type i : not_above_and_below)
          {
            const matrix_anchor anchor2 {&full2, i, j};
            BOOST_CHECK_EQUAL(true, overlap(anchor1, anchor2));
            BOOST_CHECK_EQUAL(true, overlap(anchor2, anchor1));
          }
      }
    for (index_type i = not_above_and_below[0]; i < not_above_and_below[1]; ++i)
      {
        for (index_type j : not_left_and_right)
          {
            const matrix_anchor anchor2 {&full2, i, j};
            BOOST_CHECK_EQUAL(true, overlap(anchor1, anchor2));
            BOOST_CHECK_EQUAL(true, overlap(anchor2, anchor1));
          }
      }
  }
}


/**
 * @brief
 *         Unit test for `const`-correctness.
 *
 * This test checks that:
 *
 * - all `const` member functions work likewise on mutable and immutable
 *   `MatrixAnchor`s
 * - the pointer returned by the `getMatrix` member function has correct
 *   `const`-qualification
 * - mutable and immutable `MatrixAnchor`s can be tested for overlapping in all
 *   combinations
 *
 */
BOOST_AUTO_TEST_CASE(const_correctness)
{
  typedef float value_type;
  using const_anchor = bsc_klammler::MatrixAnchor<value_type, false>;
  using mutable_anchor = bsc_klammler::MatrixAnchor<value_type, true>;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  {
    const auto readwrite = [](h_matrix * matrix_ptr)->value_type{
      matrix_ptr->set(11, 0, 0);
      return matrix_ptr->get(0, 0);
    };
    full_matrix matrix {2, 3};
    /* !const */ mutable_anchor anchor {&matrix, 4, 5};
    BOOST_CHECK_EQUAL(2, anchor.rows());
    BOOST_CHECK_EQUAL(3, anchor.cols());
    BOOST_CHECK_EQUAL(4, anchor.north());
    BOOST_CHECK_EQUAL(5, anchor.west());
    BOOST_CHECK_EQUAL(6, anchor.south());
    BOOST_CHECK_EQUAL(8, anchor.east());
    BOOST_CHECK_EQUAL(&matrix, anchor.getMatrix());
    // Only compiles if the returned pointer is mutable:
    BOOST_CHECK_EQUAL(11, readwrite(anchor.getMatrix()));
  }
  {
    const auto read = [](const h_matrix * matrix_ptr)->value_type{
      return matrix_ptr->get(0, 0);
    };
    const full_matrix matrix {{11, 12, 13}, {21, 22, 23}};
    /* !const */ const_anchor anchor {&matrix, 4, 5};
    BOOST_CHECK_EQUAL(2, anchor.rows());
    BOOST_CHECK_EQUAL(3, anchor.cols());
    BOOST_CHECK_EQUAL(4, anchor.north());
    BOOST_CHECK_EQUAL(5, anchor.west());
    BOOST_CHECK_EQUAL(6, anchor.south());
    BOOST_CHECK_EQUAL(8, anchor.east());
    BOOST_CHECK_EQUAL(&matrix, anchor.getMatrix());
    // Only compiles if the returned pointer is immutable:
    BOOST_CHECK_EQUAL(11, read(anchor.getMatrix()));
  }
  {
    const full_matrix matrix1 {2, 2};
    /* !const */ full_matrix matrix2 {2, 2};
    const const_anchor anchor1 {&matrix1, 10, 10};
    const mutable_anchor anchor2 {&matrix2, 11, 11};
    BOOST_CHECK_EQUAL(true, overlap(anchor1, anchor1));
    BOOST_CHECK_EQUAL(true, overlap(anchor1, anchor2));
    BOOST_CHECK_EQUAL(true, overlap(anchor2, anchor1));
    BOOST_CHECK_EQUAL(true, overlap(anchor2, anchor2));
  }
}
