/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_matrix_view.cpp
 *
 * @brief
 *         Unit tests for the `MatrixView` template class.
 *
 */

#include "../global.hpp"           // (package configuration)
#include "../matrix_view.hpp"      // bsc_klammler::MatrixView

#include <array>                   // std::array
#include <stdexcept>               // std::invalid_argument
#include <tuple>                   // std::tuple
#include <type_traits>             // std::true_type, ...
#include <typeinfo>                // typeid
#include <utility>                 // std::move, std::swap

#include <boost/mpl/list.hpp>      // boost::mpl::list

#include "../full_matrix.hpp"      // bsc_klammler::MatrixView
#include "../not_implemented.hpp"  // bsc_klammler::NotImplementedError
#include "../random.hpp"           // bsc_klammler::random_real
#include "mock_h_matrix.hpp"       // bsc_klammler::test::MockHMatrix
#include "util.hpp"                // MY_CHECK_MATRIX_EQUAL
#include "visitors.hpp"            // bsc_klammler::test::Extractor, ...

#define BOOST_TEST_MODULE matrix_view
#include "module.hpp"

using bsc_klammler::random_int;
using bsc_klammler::random_real;


/** @brief List of `std::false_type` and `std::true_type`. */
using mpl_true_false = boost::mpl::list<std::false_type, std::true_type>;

/** @brief List of `ConstMatrixView` and `MutableMatricView` for `double`s. */
using mpl_double_views = boost::mpl::list<bsc_klammler::ConstMatrixView<double>, bsc_klammler::MutableMatrixView<double>>;

/** @brief List of `ConstMatrixView` and `MutableMatricView` for `int`s. */
using mpl_int_views = boost::mpl::list<bsc_klammler::ConstMatrixView<int>, bsc_klammler::MutableMatrixView<int>>;


/**
 * @brief
 *         Unit test for the `isMutable` static member function.
 *
 * This test checks that:
 *
 * - the function returns the correct value
 * - the function is a `constexpr`
 *
 */
BOOST_AUTO_TEST_CASE(is_mutable)
{
  typedef double value_type;
  using const_matrix_view = bsc_klammler::ConstMatrixView<value_type>;
  using mutable_matrix_view = bsc_klammler::MutableMatrixView<value_type>;
  static_assert(!const_matrix_view::isMutable(),
                "ConstMatrixView::isMutable()");
  static_assert(mutable_matrix_view::isMutable(),
                "!MutableMatrixView::isMutable()");
}


/**
 * @brief
 *         Unit test for the member typdefs.
 *
 * This test checks that:
 *
 * - the member types `value_type`, `index_type` and `super_type` are defined
 *   correctly
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(member_typedefs, M, mpl_true_false)
{
  // There is no overload for 'operator<<(std::ostream&, std::type_info)' so we
  // cannot use 'BOOST_CHECK_EQUAL'.
  using bsc_klammler::HMatrix;
  using bsc_klammler::MatrixView;
  struct Bar {};
  BOOST_CHECK(typeid(HMatrix<Bar>) ==
              typeid(typename MatrixView<Bar, M::value>::super_type));
  BOOST_CHECK(typeid(Bar) ==
              typeid(typename MatrixView<Bar, M::value>::value_type));
  for (typename MatrixView<Bar, M::value>::index_type i = 0; i < 10; ++i)
    continue;
}


/**
 * @brief
 *         Unit test for the default constructor.
 *
 * This test checks that:
 *
 * - the dimensions of the created matrix are correct (0, 0).
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(ctor_default, matrix_view, mpl_double_views)
{
  const matrix_view matrix {};
  BOOST_CHECK_EQUAL(0, matrix.rows());
  BOOST_CHECK_EQUAL(0, matrix.cols());
}


/**
 * @brief
 *         Unit test for the constructor.
 *
 * This test checks that:
 *
 * - the view is correct (without transposing)
 * - no copies are made
 * - no ownership is taken
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(ctor, matrix_view, mpl_int_views)
{
  using bsc_klammler::const_if;
  using index_type = typename matrix_view::index_type;
  using value_type = typename matrix_view::value_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using full_matrix_c = typename const_if<full_matrix, !matrix_view::isMutable()>::type;
  static_assert(std::is_const<full_matrix_c>::value != matrix_view::isMutable(),
                "wrong const-qualifier for full_matrix_c");
  full_matrix_c inner {
    {11, 12, 13, 14, 15, 16, 17, 18},
    {21, 22, 23, 24, 25, 26, 27, 28},
    {31, 32, 33, 34, 35, 36, 37, 38},
    {41, 42, 43, 44, 45, 46, 47, 48},
  };
  for (index_type ro = 0; ro < inner.rows(); ++ro)
    {
      for (index_type co = 0; co < inner.cols(); ++co)
        {
          for (index_type r = 0; r < inner.rows() - ro; ++r)
            {
              for (index_type c = 0; c < inner.cols() - co; ++c)
                {
                  const matrix_view view {&inner, ro, co, r, c};
                  BOOST_CHECK(&inner == view.getViewedMatrix());
                  BOOST_CHECK_EQUAL(r, view.rows());
                  BOOST_CHECK_EQUAL(c, view.cols());
                  for (index_type i = 0; i < r; ++i)
                    {
                      for (index_type j = 0; j < c; ++j)
                        {
                          BOOST_CHECK_EQUAL(inner.get(ro + i, co + j),
                                            view.get(i, j));
                        }
                    }
                }
            }
        }
    }
}


namespace /* anonymous */
{
  /**
   * @brief
   *         Helper function that checks whether a `MatrixView` is &ldquo;filled
   *         with indices&rdquo;.
   *
   * By &ldquo;filled with indices&rdquo;, we mean that the element
   * <code>(<var>i</var>, <var>j</var>)</code> is equal to
   * <code><var>ij</var></code>, using mathematical subscripting.  Eg, element
   * (0, 2) will be checked for equality with 13.
   *
   * @param <T>
   *         *arithmetic* type of matrix entries
   *
   * @param view
   *         `MatrixView` to check
   *
   * @returns
   *         whether the view is &ldquo;filled with indices&rdquo;
   *
   * @throws std::invalid_argument
   *         if one dimension of the view is equal to or greater than 10
   *
   */
  template<typename T, typename = typename std::enable_if<std::is_arithmetic<T>::value>::type>
  bool
  is_filled_with_indices(const bsc_klammler::HMatrix<T>& matrix)
  {
    using index_type = typename bsc_klammler::HMatrix<T>::index_type;
    using value_type = typename bsc_klammler::HMatrix<T>::value_type;
    if (matrix.rows() >= 10 || matrix.cols() >= 10)
      throw std::invalid_argument {"matrix too big"};
    for (index_type j = 0; j < matrix.cols(); ++j)
      {
        for (index_type i = 0; i < matrix.rows(); ++i)
          {
            const value_type expected {
              static_cast<value_type>(10 * (i + 1) + (j + 1))
            };
            const value_type actual {matrix.get(i, j)};
            if (expected != actual) { return false; }
          }
      }
    return true;
  }
}  // namespace /* anonymous */


/**
 * @brief
 *         Unit test for the constructor.
 *
 * This test checks that:
 *
 * - transposed views are correct
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(ctor__trans, matrix_view, mpl_int_views)
{
  using bsc_klammler::const_if;
  using value_type = typename matrix_view::value_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using full_matrix_c = typename const_if<full_matrix, !matrix_view::isMutable()>::type;
  full_matrix_c inner {
    {-1, -1, -1, -1, -1, -1, -1},
    {-1, -1, -1, 11, 21, -1, -1},
    {-1, -1, -1, 12, 22, -1, -1},
    {-1, -1, -1, 13, 23, -1, -1},
  };
  const matrix_view view {&inner, 3, 1, 2, 3, true};
  BOOST_CHECK_EQUAL(2, view.rows());
  BOOST_CHECK_EQUAL(3, view.cols());
  BOOST_CHECK_EQUAL(&inner, view.getViewedMatrix());
  BOOST_CHECK(is_filled_with_indices(view));
}


/**
 * @brief
 *         Unit test for the constructor.
 *
 * This test checks that:
 *
 * - the constructor works correct in various corner cases
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(ctor__corners, matrix_view, mpl_double_views)
{
  using bsc_klammler::const_if;
  using value_type = typename matrix_view::value_type;
  using index_type = typename matrix_view::index_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using full_matrix_c = typename const_if<full_matrix, !matrix_view::isMutable()>::type;
  using matrix_view_c = typename const_if<matrix_view, !matrix_view::isMutable()>::type;
  {
    // Viewing an empty matrix.
    full_matrix_c inner {};
    matrix_view view {&inner, 0, 0, 0, 0};
    BOOST_CHECK(&inner == view.getViewedMatrix());
    BOOST_CHECK_EQUAL(0, view.size());
  }
  {
    // Viewing nothing.
    full_matrix_c inner {
      random_int<index_type>(16, 32), random_int<index_type>(16, 32)
    };
    matrix_view view {&inner, inner.rows(), inner.cols(), 0, 0};
    BOOST_CHECK_EQUAL(0, view.size());
  }
  {
    // Viewing a view.
    matrix_view view1 {};
    matrix_view view2 {&view1, 0, 0, 0, 0};
    BOOST_CHECK_EQUAL(0, view1.size());
    BOOST_CHECK_EQUAL(0, view2.size());
    BOOST_CHECK(nullptr == view1.getViewedMatrix());
    BOOST_CHECK(&view1 == view2.getViewedMatrix());
  }
  {
    // Stacking views beyond any restrictions.
    const index_type size {random_int<index_type>(32, 128)};
    full_matrix_c innermost {size, size};
    std::vector<matrix_view_c *> views {};
    views.push_back(new matrix_view {&innermost, 0, 0, size, size});
    for (index_type k = 0; k < size; ++k)
      {
        matrix_view_c *const view_ptr {views.back()};
        for (index_type j = 0; j < view_ptr->cols(); ++j)
          {
            for (index_type i = 0; i < view_ptr->rows(); ++i)
              {
                if (matrix_view::isMutable())
                  {
                    // We must formally cast away the 'const'-qualifier from
                    // the ' view_ptr' to make the compiler like this.
                    // However, we will only ever enter this section if that
                    // cast is redundant.
                    assert(!(std::is_const<matrix_view_c>::value));
                    matrix_view * v_ptr {const_cast<matrix_view *>(view_ptr)};
                    v_ptr->set(static_cast<value_type>(k), i, j);
                  }
              }
          }
        views.push_back(new matrix_view {
            view_ptr, 0, 0, view_ptr->rows() - 1, view_ptr->cols() - 1
        });
      }
    for (index_type k = 0; k < size; ++k)
      {
        for (index_type l = 0; l < size - k; ++l)
          {
            const value_type expected {
              matrix_view::isMutable()
                ? static_cast<value_type>(k)
                : full_matrix::zero()
            };
            const value_type actual_1 {innermost.get(size - k - 1, l)};
            const value_type actual_2 {innermost.get(l, size - k - 1)};
            BOOST_CHECK_EQUAL(expected, actual_1);
            BOOST_CHECK_EQUAL(expected, actual_2);
          }
      }
    for (auto view_ptr : views) { delete view_ptr; }
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error checking unit test for the constructor.
 *
 * This test checks that the following is rejected:
 *
 * - passing `nullptr`s
 * - <del>creating ill-defined (<var>n</var> &gt; 0, <var>m</var> = 0) and
 *   (<var>n</var> = 0, <var>m</var> &gt; 0) views</del>
 * - creating views that don't fit the viewed matrix
 * - creating empty views outside the original matrix
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(ctor__invalid, matrix_view, mpl_double_views)
{
  using bsc_klammler::const_if;
  using value_type = typename matrix_view::value_type;
  using index_type = typename matrix_view::index_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using full_matrix_c = typename const_if<full_matrix, !matrix_view::isMutable()>::type;
  using parameter_tuple = std::tuple<full_matrix_c *, index_type, index_type, index_type, index_type, bool>;
  const index_type rows {random_int<index_type>(4, 32)};
  const index_type cols {random_int<index_type>(4, 32)};
  full_matrix_c inner {rows, cols};
  const std::array<parameter_tuple, 9> good_params = {
    parameter_tuple {&inner, 0, 0, 0, 0, false},
    parameter_tuple {&inner, 0, 0, 1, 0, false},
    parameter_tuple {&inner, 0, 0, 0, 1, false},
    parameter_tuple {&inner, 0, 0, 1, 0, true},
    parameter_tuple {&inner, 0, 0, 0, 1, true},
    parameter_tuple {&inner, 0, 0, rows, cols, false},
    parameter_tuple {&inner, rows, cols, 0, 0, false},
    parameter_tuple {&inner, 0, 0, cols, rows, true},
    parameter_tuple {&inner, cols, rows, 0, 0, true},
  };
  const std::array<parameter_tuple, 20> bad_params = {
    parameter_tuple {nullptr, 0, 0, 0, 0, false},
    parameter_tuple {nullptr, 0, 0, 0, 0, true},
    parameter_tuple {&inner, 0, 0, rows + 1, 0, false},
    parameter_tuple {&inner, 0, 0, rows + 1, 1, false},
    parameter_tuple {&inner, 0, 0, 0, cols + 1, false},
    parameter_tuple {&inner, 0, 0, 1, cols + 1, false},
    parameter_tuple {&inner, rows + 1, 0, 0, 0, false},
    parameter_tuple {&inner, 0, cols + 1, 0, 0, false},
    parameter_tuple {&inner, rows, 0, 1, 1, false},
    parameter_tuple {&inner, 0, cols, 1, 1, false},
    parameter_tuple {&inner, 0, 0, cols + 1, 0, true},
    parameter_tuple {&inner, 0, 0, cols + 1, 1, true},
    parameter_tuple {&inner, 0, 0, 0, rows + 1, true},
    parameter_tuple {&inner, 0, 0, 1, rows + 1, true},
    parameter_tuple {&inner, cols + 1, 0, 0, 0, true},
    parameter_tuple {&inner, 0, rows + 1, 0, 0, true},
    parameter_tuple {&inner, cols, 0, 1, 1, true},
    parameter_tuple {&inner, 0, rows, 1, 1, true},
    parameter_tuple {&inner, 2 + rows / 2, 2 + cols / 2, 2 + rows / 2, 2 + cols / 2, false},
    parameter_tuple {&inner, 2 + cols / 2, 2 + rows / 2, 2 + cols / 2, 2 + rows / 2, true},
  };
  for (std::size_t i = 0; i < good_params.size(); ++i)
    {
      try
        {
          const matrix_view view {
            std::get<0>(good_params.at(i)),
            std::get<1>(good_params.at(i)),
            std::get<2>(good_params.at(i)),
            std::get<3>(good_params.at(i)),
            std::get<4>(good_params.at(i)),
            std::get<5>(good_params.at(i)),
          };
        }
      catch (const std::invalid_argument& e)
        {
          BOOST_ERROR("false positive (i = " << i << "): " << e.what());
        }
    }
  for (std::size_t i = 0; i < bad_params.size(); ++i)
    {
      try
        {
          const matrix_view view {
            std::get<0>(bad_params.at(i)),
            std::get<1>(bad_params.at(i)),
            std::get<2>(bad_params.at(i)),
            std::get<3>(bad_params.at(i)),
            std::get<4>(bad_params.at(i)),
            std::get<5>(bad_params.at(i)),
          };
          BOOST_ERROR("false negative (i = " << i << ")");
        }
      catch (const std::invalid_argument& e) { /* good */ }
    }
}
#endif  // BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the destructor.
 *
 * This test checks that:
 *
 * - `MatrixView`s can be safely destructed before the viewed matrix
 * - `MatrixView`s can be safely destructed after the viewed matrix
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(dtor, matrix_view, mpl_double_views)
{
  using bsc_klammler::const_if;
  using value_type = typename matrix_view::value_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using full_matrix_c = typename const_if<full_matrix, !matrix_view::isMutable()>::type;
  {
    // before
    full_matrix_c *const inner_ptr {new full_matrix_c {{1.0}}};
    matrix_view *const view_ptr {new matrix_view {inner_ptr, 0, 0, 1, 1}};
    if (matrix_view::isMutable()) { view_ptr->set(2.0, 0, 0); }
    delete view_ptr;
    BOOST_CHECK_EQUAL(1, inner_ptr->size());
    BOOST_CHECK_EQUAL(matrix_view::isMutable() ? 2.0 : 1.0,
                      inner_ptr->get(0, 0));
    delete inner_ptr;
  }
  {
    // after
    full_matrix_c *const inner_ptr {new full_matrix_c {{1.0}}};
    matrix_view *const view_ptr {new matrix_view {inner_ptr, 0, 0, 1, 1}};
    delete inner_ptr;
    delete view_ptr;
  }
}


/**
 * @brief
 *         Unit test for the copy constructor.
 *
 * This test checks that:
 *
 * - copied views are shallow copies that still don't own any data
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(ctor_copy, matrix_view, mpl_int_views)
{
  using bsc_klammler::const_if;
  using value_type = typename matrix_view::value_type;
  using index_type = typename matrix_view::index_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using full_matrix_c = typename const_if<full_matrix, !matrix_view::isMutable()>::type;
  full_matrix_c inner {
    {-1, -1, -1, -1, -1, -1, -1, -1},
    {-1, -1, 11, 12, 13, 14, 15, -1},
    {-1, -1, 21, 22, 23, 24, 25, -1},
    {-1, -1, -1, -1, -1, -1, -1, -1},
  };
  const index_type off_rows {1};
  const index_type off_cols {2};
  const index_type rows {2};
  const index_type cols {5};
  const bool trans {false};
  const matrix_view view1 {&inner, off_rows, off_cols, rows, cols, trans};
  const matrix_view view2 {view1};
  BOOST_CHECK_EQUAL(rows, view1.rows());
  BOOST_CHECK_EQUAL(cols, view1.cols());
  BOOST_CHECK(&inner == view1.getViewedMatrix());
  BOOST_CHECK(is_filled_with_indices(view1));
  BOOST_CHECK_EQUAL(rows, view2.rows());
  BOOST_CHECK_EQUAL(cols, view2.cols());
  BOOST_CHECK(&inner == view2.getViewedMatrix());
  BOOST_CHECK(is_filled_with_indices(view2));
}


/**
 * @brief
 *         Unit test for the move constructor.
 *
 * This test checks that:
 *
 * - moved views are just that
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(ctor_move, matrix_view, mpl_int_views)
{
  using bsc_klammler::const_if;
  using value_type = typename matrix_view::value_type;
  using index_type = typename matrix_view::index_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using full_matrix_c = typename const_if<full_matrix, !matrix_view::isMutable()>::type;
  full_matrix_c inner {
    {-1, -1, -1, -1, -1, -1, -1, -1},
    {-1, -1, -1, -1, -1, -1, -1, -1},
    {-1, -1, -1, 11, -1, -1, -1, -1},
    {-1, -1, -1, -1, -1, -1, -1, -1},
  };
  const index_type off_rows {2};
  const index_type off_cols {3};
  const index_type rows {1};
  const index_type cols {1};
  const bool trans {false};
  matrix_view view1 {&inner, off_rows, off_cols, rows, cols, trans};
  matrix_view view2 {std::move(view1)};
  BOOST_CHECK_EQUAL(0, view1.rows());
  BOOST_CHECK_EQUAL(0, view1.cols());
  BOOST_CHECK(nullptr == view1.getViewedMatrix());
  BOOST_CHECK(is_filled_with_indices(view1));
  BOOST_CHECK_EQUAL(rows, view2.rows());
  BOOST_CHECK_EQUAL(cols, view2.cols());
  BOOST_CHECK(&inner == view2.getViewedMatrix());
  BOOST_CHECK(is_filled_with_indices(view2));
}


/**
 * @brief
 *         Unit test for the copy-assignment operator.
 *
 * This test checks that:
 *
 * - copied views are shallow copies that still don't own any data
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(assign_copy, matrix_view, mpl_int_views)
{
  using bsc_klammler::const_if;
  using value_type = typename matrix_view::value_type;
  using index_type = typename matrix_view::index_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using full_matrix_c = typename const_if<full_matrix, !matrix_view::isMutable()>::type;
  full_matrix_c inner {
    {11, -1, -1, -1, -1, -1, -1, -1},
    {12, -1, -1, -1, -1, -1, -1, -1},
    {13, -1, -1, -1, -1, -1, -1, -1},
    {14, -1, -1, -1, -1, -1, -1, -1},
  };
  const index_type off_rows {0};
  const index_type off_cols {0};
  const index_type rows {1};
  const index_type cols {4};
  const bool trans {true};
  matrix_view view1 {&inner, off_rows, off_cols, rows, cols, trans};
  matrix_view view2 {};
  view2 = view1;
  BOOST_CHECK_EQUAL(rows, view1.rows());
  BOOST_CHECK_EQUAL(cols, view1.cols());
  BOOST_CHECK(&inner == view1.getViewedMatrix());
  BOOST_CHECK(is_filled_with_indices(view1));
  BOOST_CHECK_EQUAL(rows, view2.rows());
  BOOST_CHECK_EQUAL(cols, view2.cols());
  BOOST_CHECK(&inner == view2.getViewedMatrix());
  BOOST_CHECK(is_filled_with_indices(view2));
}


/**
 * @brief
 *         Unit test for the move-assignment operator.
 *
 * This test checks that:
 *
 * - move-assigned views are just that
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(assign_move, matrix_view, mpl_int_views)
{
  using bsc_klammler::const_if;
  using value_type = typename matrix_view::value_type;
  using index_type = typename matrix_view::index_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using full_matrix_c = typename const_if<full_matrix, !matrix_view::isMutable()>::type;
  full_matrix_c inner {
    {-1, -1, -1, -1, -1, -1, 11, 21},
    {-1, -1, -1, -1, -1, -1, 12, 22},
    {-1, -1, -1, -1, -1, -1, -1, -1},
    {-1, -1, -1, -1, -1, -1, -1, -1},
  };
  const index_type off_rows {6};
  const index_type off_cols {0};
  const index_type rows {2};
  const index_type cols {2};
  const bool trans {true};
  matrix_view view1 {&inner, off_rows, off_cols, rows, cols, trans};
  matrix_view view2 {};
  view2 = std::move(view1);
  BOOST_CHECK_EQUAL(0, view1.rows());
  BOOST_CHECK_EQUAL(0, view1.cols());
  BOOST_CHECK(nullptr == view1.getViewedMatrix());
  BOOST_CHECK(is_filled_with_indices(view1));
  BOOST_CHECK_EQUAL(rows, view2.rows());
  BOOST_CHECK_EQUAL(cols, view2.cols());
  BOOST_CHECK(&inner == view2.getViewedMatrix());
  BOOST_CHECK(is_filled_with_indices(view2));
}


/**
 * @brief
 *         Unit test for the amputated `clone` function member.
 *
 * This test checks that:
 *
 * - `MatrixView`s cannot be `clone()`d
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(cloning, matrix_view, mpl_double_views)
{
  using bsc_klammler::const_if;
  using value_type = typename matrix_view::value_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using full_matrix_c = typename const_if<full_matrix, !matrix_view::isMutable()>::type;
  full_matrix_c full {};
  const matrix_view views[] = {
    matrix_view {},  // viewes nothing
    matrix_view {&full, 0, 0, 0, 0},  // viewes something
  };
  for (const matrix_view view : views)
    {
      try
        {
          view.clone();
          BOOST_ERROR("no exception thrown");
        }
      catch (const bsc_klammler::NotImplementedError&) { /* good */ }
    }
}


/**
 * @brief
 *         Unit test for the `swap` friend function.
 *
 * This test checks that:
 *
 * - `swap` works
 *
 */

BOOST_AUTO_TEST_CASE_TEMPLATE(swapping, matrix_view, mpl_int_views)
{
  using std::swap;
  using value_type = typename matrix_view::value_type;
  using index_type = typename matrix_view::index_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  // We tag the matrix element x at (i, j) in matrix p \in {1, 2} as follows:
  //
  // ((x >>  0) & 0xFF) == j
  // ((x >>  8) & 0xFF) == i
  // ((x >> 16) & 0xFF) == p
  const auto mktag = [](const int p,
                        const index_type i,
                        const index_type j)->value_type{
    assert((~0xFF & p) == 0);
    assert((~0xFF & i) == 0);
    assert((~0xFF & j) == 0);
    return static_cast<value_type>((p << 16) | (i << 8) | (j << 0));
  };
  const index_type off_rows_1 {1};
  const index_type off_cols_1 {2};
  const index_type rows_1 {3};
  const index_type cols_1 {4};
  const index_type off_rows_2 {5};
  const index_type off_cols_2 {6};
  const index_type rows_2 {7};
  const index_type cols_2 {8};
  full_matrix inner1 {off_rows_1 + rows_1, off_cols_1 + cols_1};  // !transosed
  full_matrix inner2 {off_cols_2 + cols_2, off_rows_2 + rows_2};  // transposed
  for (index_type j = 0; j < inner1.cols(); ++j)
    for (index_type i = 0; i < inner1.rows(); ++i)
      inner1.set(mktag(1, i, j), i, j);
  for (index_type j = 0; j < inner2.cols(); ++j)
    for (index_type i = 0; i < inner2.rows(); ++i)
      inner2.set(mktag(2, i, j), i, j);
  {
    matrix_view view1 {&inner1, off_rows_1, off_cols_1, rows_1, cols_1, false};
    matrix_view view2 {&inner2, 5, 6, rows_2, cols_2, true};
    BOOST_CHECK_EQUAL(rows_1, view1.rows());
    BOOST_CHECK_EQUAL(cols_1, view1.cols());
    BOOST_CHECK(&inner1 == view1.getViewedMatrix());
    for (index_type j = 0; j < view1.cols(); ++j)
      for (index_type i = 0; i < view1.rows(); ++i)
        BOOST_CHECK_EQUAL(inner1.get(off_rows_1 + i, off_cols_1 + j),
                          view1.get(i, j));
    BOOST_CHECK_EQUAL(rows_2, view2.rows());
    BOOST_CHECK_EQUAL(cols_2, view2.cols());
    BOOST_CHECK(&inner2 == view2.getViewedMatrix());
    for (index_type j = 0; j < view2.cols(); ++j)
      for (index_type i = 0; i < view2.rows(); ++i)
        BOOST_CHECK_EQUAL(inner2.get(6 + j, 5 + i), view2.get(i, j));
    // </before>
    swap(view1, view2);
    // <after>
    BOOST_CHECK_EQUAL(rows_1, view2.rows());
    BOOST_CHECK_EQUAL(cols_1, view2.cols());
    BOOST_CHECK(&inner1 == view2.getViewedMatrix());
    for (index_type j = 0; j < view2.cols(); ++j)
      for (index_type i = 0; i < view2.rows(); ++i)
        BOOST_CHECK_EQUAL(inner1.get(off_rows_1 + i, off_cols_1 + j),
                          view2.get(i, j));
    BOOST_CHECK_EQUAL(rows_2, view1.rows());
    BOOST_CHECK_EQUAL(cols_2, view1.cols());
    BOOST_CHECK(&inner2 == view1.getViewedMatrix());
    for (index_type j = 0; j < view1.cols(); ++j)
      for (index_type i = 0; i < view1.rows(); ++i)
        BOOST_CHECK_EQUAL(inner2.get(6 + j, 5 + i), view1.get(i, j));
  }
}


/**
 * @brief
 *         Unit test for the `set` and `get` member functions.
 *
 * This test checks that:
 *
 * - `set` and `get` do well together
 * - `set` is amputated for `ConstMatrixView`s
 *
 */
BOOST_AUTO_TEST_CASE(accessors)
{
  typedef int value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using mutable_matrix_view = bsc_klammler::MutableMatrixView<value_type>;
  using const_matrix_view = bsc_klammler::ConstMatrixView<value_type>;
  {
    full_matrix inner = {
      {-1, 11, 12},
      {-1, 21, 22},
    };
    mutable_matrix_view view {&inner, 0, 1, 2, 2};
    BOOST_CHECK_EQUAL(11, view.get(0, 0));
    BOOST_CHECK_EQUAL(12, view.get(0, 1));
    BOOST_CHECK_EQUAL(21, view.get(1, 0));
    BOOST_CHECK_EQUAL(22, view.get(1, 1));
    view.set(-11, 0, 0);
    view.set(-12, 0, 1);
    view.set(-21, 1, 0);
    view.set(-22, 1, 1);
    BOOST_CHECK_EQUAL(-11, view.get(0, 0));
    BOOST_CHECK_EQUAL(-12, view.get(0, 1));
    BOOST_CHECK_EQUAL(-21, view.get(1, 0));
    BOOST_CHECK_EQUAL(-22, view.get(1, 1));
  }
  {
    const full_matrix inner = {
      {-1, 11, 12},
      {-1, 21, 22},
    };
    const_matrix_view view {&inner, 0, 1, 2, 2};
    BOOST_CHECK_EQUAL(11, view.get(0, 0));
    BOOST_CHECK_EQUAL(12, view.get(0, 1));
    BOOST_CHECK_EQUAL(21, view.get(1, 0));
    BOOST_CHECK_EQUAL(22, view.get(1, 1));
    try
      {
        view.set(-11, 0, 0);
        BOOST_ERROR("no exception thrown");
      }
    catch (const bsc_klammler::NotImplementedError&) { /* good */ }
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the `set` and `get` member functions.
 *
 * This test checks that:
 *
 * - bad indices are caught
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(accessors__invalid, matrix_view, mpl_int_views)
{
  using bsc_klammler::const_if;
  using value_type = typename matrix_view::value_type;
  using index_type = typename matrix_view::index_type;
  using mockup = typename bsc_klammler::test::MockHMatrix<value_type>;
  using mockup_c = typename const_if<mockup, !matrix_view::isMutable()>::type;
  // Mentally, assume we have this setup:
  //
  //     -1 -1 -1 -1 -1
  //     -1 11 12 -1 -1
  //     -1 21 22 -1 -1
  //     -1 31 32 -1 -1
  //     -1 -1 -1 -1 -1
  //     -1 -1 -1 -1 -1
  constexpr std::pair<index_type, index_type> bad_indices[] = {
    {0, 2},
    {0, 3},
    {0, 4},
    {3, 0},
    {4, 0},
    {5, 0},
    {3, 1},
    {3, 2},
    {14, 92},
  };
  // We use a mockup that will throw a different exception, so we know that
  // exceptions came from the view.
  mockup_c mock {};
  mock.mockRows([]()->index_type{ return 6; });
  mock.mockCols([]()->index_type{ return 5; });
  mock.mockGet([](index_type, index_type)->value_type{ throw "get"; });
  mock.mockSet([](value_type, index_type, index_type)->void{ throw "set"; });
  matrix_view view {&mock, 1, 1, 3, 2, false};
  for (const auto iter : bad_indices)
    {
      const index_type i {iter.first};
      const index_type j {iter.second};
      try
        {
          view.get(i, j);
          BOOST_ERROR("no exception thrown "
                      << "(i = " << i << ", j = " << j << ")");
        }
      catch (std::invalid_argument&) { /* good */ }
      try
        {
          view.set(0, i, j);
          BOOST_ERROR("no exception thrown "
                      << "(i = " << i << ", j = " << j << ")");
        }
      catch (std::invalid_argument&)
        {
          BOOST_CHECK(matrix_view::isMutable());
        }
      catch (bsc_klammler::NotImplementedError&)
        {
          BOOST_CHECK(!matrix_view::isMutable());
        }
    }
}
#endif  // BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Trivial unit test for the passive `passOverRegion` member function.
 *
 * This test checks that:
 *
 * - one can pass over empty regions
 * - one can pass over an empty view
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(pass_over_region__passive__trivial,
                              matrix_view,
                              mpl_double_views)
{
  using value_type = typename matrix_view::value_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using extractor = bsc_klammler::test::Extractor<value_type>;
  {
    const matrix_view view {};
    const full_matrix expected {};
    full_matrix actual {expected.rows(), expected.cols()};
    extractor pv {actual};
    view.passOverRegion(0, 0, false, pv, 0, 0, 0, 0);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    full_matrix inner {};
    const matrix_view view {&inner, 0, 0, 0, 0};
    const full_matrix expected {};
    full_matrix actual {expected.rows(), expected.cols()};
    extractor pv {actual};
    view.passOverRegion(0, 0, false, pv, 0, 0, 0, 0);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
}


/**
 * @brief
 *         Unit test for the passive `passOverRegion` member function.
 *
 * This test checks that:
 *
 * - the correct region is shown to the visitor
 * - transposition is applied correctly
 * - each element is visited exactly once
 * - the visited matrix is `const` but the visitor may mutate
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(pass_over_region__passive,
                              matrix_view,
                              mpl_double_views)
{
  using value_type = typename matrix_view::value_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using extractor = bsc_klammler::test::Extractor<value_type>;
  full_matrix data {
    {11, 12, 13, 14, 15, 16},
    {21, 22, 23, 24, 25, 26},
    {31, 32, 33, 34, 35, 36},
    {41, 42, 43, 44, 45, 46},
    {51, 52, 53, 54, 55, 56},
  };
  full_matrix data_trans {
    {11, 21, 31, 41, 51},
    {12, 22, 32, 42, 52},
    {13, 23, 33, 43, 53},
    {14, 24, 34, 44, 54},
    {15, 25, 35, 45, 55},
    {16, 26, 36, 46, 56},
  };
  const full_matrix expected {
    { 0,  0,  0,  0,  0,  0},
    { 0,  0,  0,  0,  0,  0},
    { 0,  0,  0,  0, 35, 36},
    { 0,  0,  0,  0, 45, 46},
    { 0,  0,  0,  0,  0,  0},
  };
  {
    BOOST_CHECKPOINT("view of data");
    const matrix_view view {&data, 1, 2, 3, 4, false};
    full_matrix actual {expected.rows(), expected.cols()};
    extractor pv {actual};
    view.passOverRegion(2, 4, false, pv, 1, 2, 3, 4);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    BOOST_CHECKPOINT("transposed view of data");
    const matrix_view view {&data, 2, 1, 4, 3, true};
    full_matrix actual {expected.rows(), expected.cols()};
    extractor pv {actual};
    view.passOverRegion(2, 4, true, pv, 1, 2, 3, 4);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    BOOST_CHECKPOINT("view of transposed data");
    const matrix_view view {&data_trans, 2, 1, 4, 3, false};
    full_matrix actual {expected.rows(), expected.cols()};
    extractor pv {actual};
    view.passOverRegion(2, 4, true, pv, 1, 2, 3, 4);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    BOOST_CHECKPOINT("transposed view of transposed data");
    const matrix_view view {&data_trans, 1, 2, 3, 4, true};
    full_matrix actual {expected.rows(), expected.cols()};
    extractor pv {actual};
    view.passOverRegion(2, 4, false, pv, 1, 2, 3, 4);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the passive `passOverRegion` member
 *         function.
 *
 * This test checks that:
 *
 * - invalid regions are rejected
 * - rejection happens before the visitor is invoked
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(pass_over_region__passive__invalid,
                              matrix_view,
                              mpl_double_views)
{
  using value_type = typename matrix_view::value_type;
  using index_type = typename matrix_view::index_type;
  using mock_matrix = typename bsc_klammler::test::MockHMatrix<value_type>;
  using extractor = bsc_klammler::test::Extractor<value_type>;
  mock_matrix inner {};
  mock_matrix dummy {};
  inner.mockRows([]()->index_type{ return 10; });
  inner.mockCols([]()->index_type{ return 10; });
  dummy.mockRows([]()->index_type{ return 10; });
  dummy.mockCols([]()->index_type{ return 10; });
  {
    const matrix_view view {&inner, 2, 3, 4, 5};
    extractor pv {dummy};
    try
      {
        // These indices are only invalid for the view.  The underlying matrix
        // would actually support them.  Still, this must be invalid.
        view.passOverRegion(0, 0, true, pv, 0, 0, 4, 5);
        BOOST_ERROR("no exception thrown");
      }
    catch (const std::invalid_argument& e) { /* good */ }
    // The mockup would have thrown a 'NotImplementedError' if the region would
    // not have been rejected early.
  }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Trivial unit test for the active `passOverRegion` member function.
 *
 * This test checks that:
 *
 * - one can pass over empty regions
 * - one can pass over an empty view
 *
 */
BOOST_AUTO_TEST_CASE(pass_over_region__active__mutable__trivial)
{
  typedef double value_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using matrix_view = typename bsc_klammler::MutableMatrixView<value_type>;
  using inserter = bsc_klammler::test::Inserter<value_type>;
  {
    matrix_view view {};
    const full_matrix expected {};
    full_matrix actual {expected.rows(), expected.cols()};
    inserter av {actual};
    view.passOverRegion(0, 0, false, av, 0, 0, 0, 0);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    full_matrix inner {};
    matrix_view view {&inner, 0, 0, 0, 0};
    const full_matrix expected {};
    full_matrix actual {expected.rows(), expected.cols()};
    inserter av {actual};
    view.passOverRegion(0, 0, false, av, 0, 0, 0, 0);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
}


/**
 * @brief
 *         Unit test for the active `passOverRegion` member function for
 *         `MutableMatrixView`s.
 *
 * This test checks that:
 *
 * - the correct region is shown to the visitor
 * - transposition is applied correctly
 * - each element is visited exactly once
 * - the visited matrix and the visitor may mutate
 *
 */
BOOST_AUTO_TEST_CASE(pass_over_region__active__mutable)
{
  typedef double value_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using matrix_view = typename bsc_klammler::MutableMatrixView<value_type>;
  using inserter = bsc_klammler::test::Inserter<value_type>;
  full_matrix data {
    {11, 12, 13, 14, 15, 16, 17},
    {21, 22, 23, 24, 25, 26, 27},
    {31, 32, 33, 34, 35, 36, 37},
    {41, 42, 43, 44, 45, 46, 47},
    {51, 52, 53, 54, 55, 56, 57},
    {61, 62, 63, 64, 65, 66, 67},
    {71, 72, 73, 74, 75, 76, 77},
  };
  const full_matrix expected {
    { 0,  0,  0,  0,  0,  0},
    { 0,  0,  0,  0,  0,  0},
    { 0,  0,  0,  0, 35, 36},
    { 0,  0,  0,  0, 45, 46},
    { 0,  0,  0,  0,  0,  0},
  };
  const full_matrix trans_expected {
    { 0,  0,  0,  0,  0},
    { 0,  0,  0,  0,  0},
    { 0,  0,  0,  0,  0},
    { 0,  0,  0,  0,  0},
    { 0,  0, 35, 45,  0},
    { 0,  0, 36, 46,  0},
  };
  {
    BOOST_CHECKPOINT("view of data");
    full_matrix actual {expected.rows(), expected.cols()};
    matrix_view view {&actual, 1, 2, 3, 4, false};
    inserter av {data};
    view.passOverRegion(2, 4, false, av, 1, 2, 3, 4);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    BOOST_CHECKPOINT("transposed view of data");
    full_matrix actual {expected.rows(), expected.cols()};
    matrix_view view {&actual, 2, 1, 4, 3, true};
    inserter av {data};
    view.passOverRegion(2, 4, true, av, 1, 2, 3, 4);
    MY_CHECK_MATRIX_EQUAL(expected, actual);
  }
  {
    BOOST_CHECKPOINT("view of transposed data");
    full_matrix trans_actual {trans_expected.rows(), trans_expected.cols()};
    matrix_view view {&trans_actual, 2, 1, 4, 3, false};
    inserter av {data};
    view.passOverRegion(2, 4, true, av, 1, 2, 3, 4);
    MY_CHECK_MATRIX_EQUAL(trans_expected, trans_actual);
  }
  {
    BOOST_CHECKPOINT("transposed view of transposed data");
    full_matrix trans_actual {trans_expected.rows(), trans_expected.cols()};
    matrix_view view {&trans_actual, 1, 2, 3, 4, true};
    inserter av {data};
    view.passOverRegion(2, 4, false, av, 1, 2, 3, 4);
    MY_CHECK_MATRIX_EQUAL(trans_expected, trans_actual);
  }
}


/**
 * @brief
 *         Unit test for the amputated active `passOverRegion` member function
 *         for `ConstMatrixView`s.
 *
 * This test checks that:
 *
 * - the function member is indeed amputated
 *
 */
#if BSC_KLAMMLER_CHECK_ARGS
BOOST_AUTO_TEST_CASE(pass_over_region__active__const)
{
  typedef double value_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using matrix_view = typename bsc_klammler::ConstMatrixView<value_type>;
  using inserter = bsc_klammler::test::Inserter<value_type>;
  const full_matrix data {};
  const full_matrix inner {};
  matrix_view view {&inner, 0, 0, 0, 0};
  inserter av {data};
  try
    {
      view.passOverRegion(0, 0, false, av, 0, 0, 0, 0);
      BOOST_ERROR("no exception thrown");
    }
  catch (const bsc_klammler::NotImplementedError&) { /* good */ }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the active `passOverRegion` member
 *         function.
 *
 * This test checks that:
 *
 * - invalid regions are rejected
 * - rejection happens before the visitor is invoked
 *
 */
BOOST_AUTO_TEST_CASE(pass_over_region__active__invalid)
{
  typedef double value_type;
  using matrix_view = typename bsc_klammler::MutableMatrixView<value_type>;
  using mock_matrix = typename bsc_klammler::test::MockHMatrix<value_type>;
  using index_type = typename matrix_view::index_type;
  using inserter = bsc_klammler::test::Inserter<value_type>;
  mock_matrix inner {};
  mock_matrix dummy {};
  inner.mockRows([]()->index_type{ return 10; });
  inner.mockCols([]()->index_type{ return 10; });
  dummy.mockRows([]()->index_type{ return 10; });
  dummy.mockCols([]()->index_type{ return 10; });
  {
    matrix_view view {&inner, 2, 3, 4, 5};
    inserter av {dummy};
    try
      {
        // These indices are only invalid for the view.  The underlying matrix
        // would actually support them.  Still, this must be invalid.
        view.passOverRegion(0, 0, true, av, 0, 0, 4, 5);
        BOOST_ERROR("no exception thrown");
      }
    catch (const std::invalid_argument& e) { /* good */ }
    // The mockup would have thrown a 'NotImplementedError' if the region would
    // not have been rejected early.
  }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


/**
 * @brief
 *         Unit test for the static `makeTransposedView` member function.
 *
 * This test checks that:
 *
 * - a correct view is created
 *
 */
BOOST_AUTO_TEST_CASE_TEMPLATE(make_transposed_view, matrix_view, mpl_double_views)
{
  using bsc_klammler::const_if;
  using value_type = typename matrix_view::value_type;
  using index_type = typename matrix_view::index_type;
  using full_matrix = typename bsc_klammler::FullMatrix<value_type>;
  using full_matrix_c = typename const_if<full_matrix, !matrix_view::isMutable()>::type;
  {
    // empty
    full_matrix_c inner {};
    matrix_view trans {matrix_view::makeTransposedView(inner)};
    BOOST_CHECK_EQUAL(0, trans.rows());
    BOOST_CHECK_EQUAL(0, trans.cols());
    BOOST_CHECK_EQUAL(&inner, trans.getViewedMatrix());
  }
  {
    // non-empty
    const index_type rows {random_int<index_type>(1, 32)};
    const index_type cols {random_int<index_type>(1, 32)};
    full_matrix_c inner {rows, cols};
    matrix_view trans {matrix_view::makeTransposedView(inner)};
    BOOST_CHECK_EQUAL(cols, trans.rows());
    BOOST_CHECK_EQUAL(rows, trans.cols());
    BOOST_CHECK_EQUAL(&inner, trans.getViewedMatrix());
    for (index_type j = 0; j < cols; ++j)
      {
        for (index_type i = 0; i < rows; ++i)
          {
            const value_type v {random_real<value_type>()};
            BOOST_CHECK_EQUAL(inner.get(i, j), trans.get(j, i));
            if (matrix_view::isMutable())
              {
                // That trick again!
                assert(!std::is_const<full_matrix_c>::value);
                const_cast<full_matrix * >(&inner)->set(v, i, j);
              }
            BOOST_CHECK_EQUAL(inner.get(i, j), trans.get(j, i));
            if (matrix_view::isMutable()) { trans.set(v, j, i); }
            BOOST_CHECK_EQUAL(inner.get(i, j), trans.get(j, i));
          }
      }
  }
}
