/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_preconditioner.cpp
 *
 * @brief
 *         Unit tests for the `preconditioner.hpp` header.
 *
 */

#include "../global.hpp"            // (package configuration)
#include "../preconditioner.hpp"    // bsc_klammler::Preconditioner

#include <chrono>                   // std::chrono::seconds
#include <cmath>                    // std::abs
#include <limits>                   // std::numeric_limits
#include <thread>                   // std::this_thread

#include "../full_matrix.hpp"       // bsc_klammler::FullMatrix
#include "../full_vector.hpp"       // bsc_klammler::FullVector
#include "../random.hpp"            // bsc_klammler::random_real
#include "../timer.hpp"             // bsc_klammler::Timer
#include "mock_preconditioner.hpp"  // bsc_klammler::test::MockPreconditioner

#define BOOST_TEST_MODULE preconditioner
#include "module.hpp"

using bsc_klammler::random_int;
using bsc_klammler::random_real;

#ifndef TOL_TIMER
#define TOL_TIMER 1.0E-2
#endif


namespace /* anonymous */
{

  /** @brief Convenience type alias for  `bsc_klammler::Timer::time_delta_type`. */
  using timer_seconds_t = bsc_klammler::Timer::time_delta_type;

  /** @brief `std::chrono::duration` type that matches `timer_seconds_t`. */
  using chrono_seconds_t = std::chrono::duration<timer_seconds_t>;

  /** @brief Tolerance for timing results (in seconds). */
  static constexpr timer_seconds_t timing_tolerance {TOL_TIMER};

  /**
   * @brief
   *         Suspends excution of the current thread for at least the given
   *         (fractional) number of seconds and returns the exact number of
   *         seconds it suspended.
   *
   * @param seconds
   *         non-negative number of seconds to suspend execution
   *
   * @returns
   *         non-negative number of seconds execution was suspended
   *
   */
  timer_seconds_t
  sleep_and_tell(const timer_seconds_t seconds) noexcept
  {
    if (seconds < 0.0)
      {
        // This should never happen!
        return std::numeric_limits<timer_seconds_t>::quiet_NaN();
      }
    else
      {
        const chrono_seconds_t chrono_seconds {seconds};
        const auto before = std::chrono::steady_clock::now();
        std::this_thread::sleep_for(chrono_seconds);
        const auto after = std::chrono::steady_clock::now();
        const chrono_seconds_t delta {after - before};
        return delta.count();
      }
  }
}  // namespace /* anonymous */


/**
 * @brief
 *         Unit test for the `size` function member.
 *
 * This test checks that:
 *
 * - `size` returns the size of the coefficient matrix
 *
 */
BOOST_AUTO_TEST_CASE(size)
{
  typedef double value_type;
  using mockup = bsc_klammler::test::MockPreconditioner<value_type>;
  using index_type = mockup::index_type;
  const index_type size {random_int<index_type>(1, 32)};
  mockup mock {size};
  BOOST_CHECK_EQUAL(size, mock.size());
}


/**
 * @brief
 *         Unit test for the `init` and `solve` function members.
 *
 * This test checks that:
 *
 * - the timing is applied correctly
 *
 */
BOOST_AUTO_TEST_CASE(timing)
{
  typedef timer_seconds_t value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using h_vector = bsc_klammler::HVector<value_type>;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using mockup = bsc_klammler::test::MockPreconditioner<value_type>;
  value_type expected_init;
  value_type expected_solve;
  const auto sleeping_init = [&expected_init](const h_matrix& c)->void{
    expected_init = sleep_and_tell(c.get(0, 0));
  };
  const auto sleeping_solve = [&expected_solve](const h_vector& r)->full_vector{
    expected_solve = sleep_and_tell(r.get(0));
    return full_vector {r.size()};
  };
  for (int i = 0; i < 10; ++i)
    {
      const value_type interval {random_real<value_type>(0.0, 0.125)};
      const full_matrix coeffs {{interval}};
      const full_vector right {{interval}};
      bsc_klammler::Timer timer_init {};
      bsc_klammler::Timer timer_solve {};
      mockup mock {coeffs.size()};
      mock.mockInit(sleeping_init);
      mock.mockSolve(sleeping_solve);
      static_assert(std::numeric_limits<value_type>::has_quiet_NaN, "no NaN");
      expected_init = std::numeric_limits<value_type>::quiet_NaN();
      expected_solve = std::numeric_limits<value_type>::quiet_NaN();
      mock.init(coeffs, &timer_init);
      mock.solve(right, &timer_solve);
      const value_type actual_init {timer_init.getSeconds()};
      const value_type actual_solve {timer_solve.getSeconds()};
      std::cout << "init:  expected = " << expected_init  << ", actual = " << actual_init  << '\n';
      std::cout << "solve: expected = " << expected_solve << ", actual = " << actual_solve << '\n';
      assert(expected_init >= 0.0 && expected_init < 1.0);
      assert(expected_solve >= 0.0 && expected_solve < 1.0);
      if (!RUNNING_ON_VALGRIND)
        {
          BOOST_CHECK(std::abs(actual_init - expected_init)
                      < timing_tolerance);
          BOOST_CHECK(std::abs(actual_solve - expected_solve)
                      < timing_tolerance);
        }
    }
}


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the `init` and `solve` function
 *         members.
 *
 * This test checks that:
 *
 * - empty and non-square coefficient matrices are rejected
 * - right sides of the wrong size are rejected
 * - `initImpl` and `solveImpl` are not called in this case
 *
 */
BOOST_AUTO_TEST_CASE(invalid_args)
{
  typedef double value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using h_vector = bsc_klammler::HVector<value_type>;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using mockup = bsc_klammler::test::MockPreconditioner<value_type>;
  bool toggle;
  const auto mock_init = [&toggle](const h_matrix&)->void{
    toggle = true;
  };
  const auto mock_solve = [&toggle](const h_vector& r)->full_vector{
    toggle = true;
    return full_vector {r.size()};
  };
  {
    // init
    const full_matrix bad_matrices[] = {
      full_matrix {},
      full_matrix {3, 4},
      full_matrix {3, 4},
      full_matrix {3, 3},
    };
    for (const auto iter : bad_matrices)
      {
        mockup mock {4};
        mock.mockInit(mock_init);
        toggle = false;
        try
          {
            mock.init(iter);
            BOOST_ERROR("no exception thrown");
          }
        catch (std::invalid_argument&) { /* good */ }
        BOOST_CHECK(!toggle);
      }
  }
  {
    // solve
    const full_matrix coeffs {{1.0, 0.0}, {0.0, 1.0}};
    const full_vector bad_vectors[] = {
      full_vector {},
      full_vector {1},
      full_vector {3},
    };
    mockup mock {coeffs.size()};
    mock.mockInit(mock_init);
    mock.mockSolve(mock_solve);
    mock.init(coeffs);
    BOOST_CHECK_EQUAL(coeffs.size(), mock.size());
    for (const auto iter : bad_vectors)
      {
        toggle = false;
        try
          {
            mock.solve(iter);
            BOOST_ERROR("no exception thrown");
          }
        catch (std::invalid_argument&) { /* good */ }
        BOOST_CHECK(!toggle);
      }
  }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS


#if BSC_KLAMMLER_CHECK_ARGS
/**
 * @brief
 *         Error-checking unit test for the `init` and `solve` function
 *         members.
 *
 * This test checks that:
 *
 * - calling `solve` before `init` is rejected
 * - calling `init` more than once is rejected
 *
 */
BOOST_AUTO_TEST_CASE(invalid_state)
{
  typedef double value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using h_vector = bsc_klammler::HVector<value_type>;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using mockup = bsc_klammler::test::MockPreconditioner<value_type>;
  const auto mock_init = [](const h_matrix&)->void{
    // do nothing
  };
  const auto mock_solve = [](const h_vector& r)->full_vector{
    // do nothing
    return full_vector {r.size()};
  };
  {
    // multi-init
    const full_matrix coeffs {
      {1.0, 0.0, 0.0},
      {0.0, 2.0, 0.0},
      {0.0, 0.0, 3.0}
    };
    mockup mock {coeffs.size()};
    mock.mockInit(mock_init);
    mock.mockSolve(mock_solve);
    mock.init(coeffs);
    for (unsigned i = 0; i < 3; ++i)
      {
        try
          {
            mock.init(coeffs);
            BOOST_ERROR("no exception thrown");
          }
        catch (const std::logic_error&) { /* good */ }
      }
  }
  {
    // solve before init
    const full_vector right {1};
    mockup mock {right.size()};
    mock.mockInit(mock_init);
    mock.mockSolve(mock_solve);
    try
      {
        mock.solve(right);
        BOOST_ERROR("no exception thrown");
      }
    catch (const std::logic_error&) { /* good */ }
  }
}
#endif  // if BSC_KLAMMLER_CHECK_ARGS
