/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_random_problems.cpp
 *
 * @brief
 *         Unit tests for the `random_problems.hpp` header.
 *
 */

#include "../global.hpp"           // (package configuration)
#include "../random_problems.hpp"  // (module to be tested)

#include "../algorithms.hpp"       // bsc_klammler::is_symmetric, ...
#include "../random.hpp"           // bsc_klammler::random_real

#define BOOST_TEST_MODULE random_problems
#include "module.hpp"

using bsc_klammler::random_int;
using bsc_klammler::random_real;


/**
 * @brief
 *         Unit test for the `get_random_block_matrix` function.
 *
 * This test checks that:
 *
 * - TODO: ...
 *
 */
BOOST_AUTO_TEST_CASE(get_random_block_matrix)
{
  typedef double value_type;
  using h_matrix = bsc_klammler::HMatrix<value_type>;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using block_matrix = bsc_klammler::BlockMatrix<value_type>;
  using index_type = h_matrix::index_type;
  for (const bool spd : {false, true})
    {
      const unsigned recursions {random_int<unsigned>(1, 3)};
      const index_type min_bs {random_int<unsigned>(2, 10)};
      const index_type max_bs {random_int<unsigned>(min_bs, 20)};
      const block_matrix matrix {
        bsc_klammler::get_random_block_matrix<value_type>(
            recursions, min_bs, max_bs, spd)
      };
      {
        const h_matrix * block_ptr = &matrix;
        for (unsigned r = 0; r < recursions; ++r)
          {
            if (auto a_ptr = dynamic_cast<const block_matrix *>(block_ptr))
              block_ptr = a_ptr->getBlockA();
            else
              BOOST_ERROR("recursion stopped prematurely at level " << r);
          }
        BOOST_CHECK(dynamic_cast<const full_matrix *>(block_ptr));
      }
      if (spd)
        {
          BOOST_CHECK(bsc_klammler::is_symmetric(matrix));
          BOOST_CHECK(bsc_klammler::is_positive_definite(matrix));
        }
      else
        {
          // Ignore the small probability of a matrix that is accidentially spd
          // for now.
          BOOST_CHECK(!bsc_klammler::is_symmetric(matrix));
          BOOST_CHECK(!bsc_klammler::is_positive_definite(matrix));
        }
    }
}


/**
 * @brief
 *         Unit test for the `get_random_vector` function.
 *
 * This test checks that:
 *
 * - tiny vectors of size 0, 1, 2, &hellip; are created correctly
 * - small vectors are created correctly
 * - the vectors are not filled with all-zeros
 *
 */
BOOST_AUTO_TEST_CASE(get_random_vector)
{
  typedef double value_type;
  using full_vector = bsc_klammler::FullVector<value_type>;
  using index_type = full_vector::index_type;
  // Does it work for (very) small sizes?
  for (index_type size = 0; size < 10; ++size)
    {
      const full_vector vector {
        bsc_klammler::get_random_vector<value_type>(size)
      };
      BOOST_CHECK_EQUAL(size, vector.size());
    }
  // Does it produce random values?
  {
    const index_type size {random_int<index_type>(128, 256)};
    const full_vector vector {
      bsc_klammler::get_random_vector<value_type>(size)
        };
    index_type cnt_zero {};
    index_type cnt_nonzero {};
    BOOST_CHECK_EQUAL(size, vector.size());
    for (index_type i = 0; i < size; ++i)
      {
        if (vector.get(i) == full_vector::zero())
          ++cnt_zero;
        else
          ++cnt_nonzero;
      }
    assert(cnt_zero + cnt_nonzero == size);
    // If testing for randomness only were that easy!
    BOOST_CHECK(10 * cnt_zero < cnt_nonzero);
  }
}
