/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_timer.cpp
 *
 * @brief
 *         Unit tests for the `timer.hpp` header.
 *
 */

#include "../global.hpp"  // (package configuration)
#include "../timer.hpp"   // bsc_klammler::Timer

#include <chrono>         // std::chrono::seconds
#include <cmath>          // std::abs
#include <thread>         // std::this_thread
#include <type_traits>    // std::is_copy_constructible, ...

#include "../random.hpp"  // bsc_klammler::random_real

#define BOOST_TEST_MODULE timer
#include "module.hpp"

using bsc_klammler::random_int;
using bsc_klammler::random_real;

#ifndef TOL_TIMER
#define TOL_TIMER 1.0E-2
#endif

/** @brief Tolerance for timing results (in seconds). */
static constexpr bsc_klammler::Timer::time_delta_type timing_tolerance {
  TOL_TIMER
};


/**
 * @brief
 *         Unit test for the `Timer` class.
 *
 * This test checks that:
 *
 * - the timing is correct
 *
 */
BOOST_AUTO_TEST_CASE(timing)
{
  using timer_seconds_type = bsc_klammler::Timer::time_delta_type;
  using chrono_seconds_type = std::chrono::duration<timer_seconds_type>;
  for (int i = 0; i < 3; ++i)
    {
      const timer_seconds_type ts {random_real<timer_seconds_type>(0.0, 0.5)};
      const chrono_seconds_type cs {ts};
      bsc_klammler::Timer timer {};
      timer.start();
      std::this_thread::sleep_for(cs);
      timer.stop();
      const timer_seconds_type actual {timer.getSeconds()};
      if (!RUNNING_ON_VALGRIND)
        {
          const timer_seconds_type expected {ts};
          BOOST_CHECK(std::abs(actual - expected) < timing_tolerance);
        }
      // Did it really stop?
      for (int i = 0; i < 100; ++i)
        {
          constexpr std::chrono::microseconds duration {100};
          std::this_thread::sleep_for(duration);
          BOOST_CHECK_EQUAL(actual, timer.getSeconds());
        }
    }
}


/**
 * @brief
 *         Unit test for the `Timer` class.
 *
 * This test checks that:
 *
 * - the state machine is working as documented
 *
 */
BOOST_AUTO_TEST_CASE(state_machine)
{
  {
    // This is the correct way (redundant but catch easy bug first).
    bsc_klammler::Timer timer {};
    timer.start();
    timer.stop();
    timer.getSeconds();
  }
  {
    // Break out at any part of the life-cycle.
    for (int i = 0; i < 10; ++i)
      {
        bsc_klammler::Timer timer {};
        if (i > 0) { timer.start(); }
        if (i > 1) { timer.stop(); }
        for (int j = 2; j < i; ++j) { timer.getSeconds(); }
      }
  }
#if BSC_KLAMMLER_CHECK_ARGS
  {
    bsc_klammler::Timer timer {};
    try
      {
        timer.stop();
        BOOST_ERROR("no exception thrown");
      }
    catch (const bsc_klammler::Timer::InvalidStateError&) { /* good */ }
    // Make sure the 'Timer' is still good.
    timer.start();
    timer.stop();
    timer.getSeconds();
  }
  {
    bsc_klammler::Timer timer {};
    try
      {
        timer.getSeconds();
        BOOST_ERROR("no exception thrown");
      }
    catch (const bsc_klammler::Timer::InvalidStateError&) { /* good */ }
    // Make sure the 'Timer' is still good.
    timer.start();
    timer.stop();
    timer.getSeconds();
  }
  {
    bsc_klammler::Timer timer {};
    timer.start();
    try
      {
        timer.start();
        BOOST_ERROR("no exception thrown");
      }
    catch (const bsc_klammler::Timer::InvalidStateError&) { /* good */ }
    // Make sure the 'Timer' is still good.
    timer.stop();
    timer.getSeconds();
  }
  {
    bsc_klammler::Timer timer {};
    timer.start();
    timer.stop();
    try
      {
        timer.stop();
        BOOST_ERROR("no exception thrown");
      }
    catch (const bsc_klammler::Timer::InvalidStateError&) { /* good */ }
    // Make sure the 'Timer' is still good.
    timer.getSeconds();
  }
  {
    bsc_klammler::Timer timer {};
    timer.start();
    timer.stop();
    try
      {
        timer.start();
        BOOST_ERROR("no exception thrown");
      }
    catch (const bsc_klammler::Timer::InvalidStateError&) { /* good */ }
    // Make sure the 'Timer' is still good.
    timer.getSeconds();
  }
#endif  // if BSC_KLAMMLER_CHECK_ARGS
}


/**
 * @brief
 *         Unit test for the `Timer::Handle` class.
 *
 * This test checks that:
 *
 * - the handle calls the timer correctly
 * - the class is not copyable
 * - `nullptr`s are handled gracefully
 * - no ownership of the `Timer` is taken (Valgrind can check this)
 *
 */
BOOST_AUTO_TEST_CASE(timer_handle)
{
  using bsc_klammler::Timer;
  using chrono_seconds = std::chrono::duration<Timer::time_delta_type>;
  BOOST_CHECK(!std::is_copy_constructible<Timer::Handle>::value);
  BOOST_CHECK(!std::is_copy_assignable<Timer::Handle>::value);
  {
    const Timer::Handle th {nullptr};  // Yes, it's as simple as that!
  }
  {
    const Timer::time_delta_type intervals[] = {1.0E-3, 1.0E-2, 1.0E-1};
    for (const Timer::time_delta_type iter : intervals)
      {
        const chrono_seconds cs {iter};
        Timer timer1 {};
        Timer timer2 {};
        {
          timer1.start();  // started first
          const Timer::Handle th {&timer2};
          std::this_thread::sleep_for(cs);
          timer1.stop();  // stopped first
        }
        const auto diff1 = timer1.getSeconds() - iter;
        const auto diff2 = timer2.getSeconds() - iter;
        BOOST_CHECK(std::abs(diff2) < 2 * std::abs(diff1));
        if (!RUNNING_ON_VALGRIND)
          {
            BOOST_CHECK(std::abs(diff2) < timing_tolerance);
          }
      }
  }
}
