/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


#include <stdexcept>  // std::logic_error

#include "../full_matrix.hpp"

namespace bsc_klammler
{
  namespace test
  {

    template<typename T>
    class Extractor : public HMatrix<T>::PassiveVisitor
    {
    public:

      using index_type = typename HMatrix<T>::index_type;
      using value_type = typename HMatrix<T>::value_type;

    private:

      HMatrix<value_type> *const dstPtr_;
      FullMatrix<bool> visited_;

    public:

      Extractor(HMatrix<value_type>& dst) :
        dstPtr_ {&dst},
        visited_ {dst.rows(), dst.cols()}
      {
        // do nothing
      }

      virtual void
      visitRegion(const index_type g_north,
                  const index_type g_west,
                  const bool g_trans,
                  const HMatrix<value_type>& matrix,
                  const index_type north,
                  const index_type west,
                  const index_type south,
                  const index_type east) final override
      {
        for (index_type j = west; j < east; ++j)
          {
            for (index_type i = north; i < south; ++i)
              {
                const index_type gi {g_north + i - north};
                const index_type gj {g_west  + j - west};
                if (this->visited_.get(gi, gj))
                  {
                    throw std::logic_error {"already visited (global check)"};
                  }
                else
                  {
                    const index_type ii {g_trans ? j : i};
                    const index_type jj {g_trans ? i : j};
                    const value_type v {matrix.get(ii, jj)};
                    this->dstPtr_->set(v, gi, gj);
                    this->visited_.set(true, gi, gj);
                  }
              }
          }
      }

    private:

      // Delte these so we can be sure that the visitor is never copied.  It
      // must not as we want to carry state with it.

      Extractor(const Extractor&) = delete;
      Extractor(Extractor&&) = delete;
      Extractor& operator=(const Extractor&) = delete;
      Extractor& operator=(Extractor&&) = delete;

    };  // class Extractor

    template<typename T>
    class Inserter : public HMatrix<T>::ActiveVisitor
    {
    public:

      using index_type = typename HMatrix<T>::index_type;
      using value_type = typename HMatrix<T>::value_type;

    private:

      const HMatrix<value_type> *const srcPtr_;
      FullMatrix<bool> visited_;

    public:

      Inserter(const HMatrix<value_type>& src) :
        srcPtr_ {&src},
        visited_ {src.rows(), src.cols()}
      {
        // do nothing
      }

      virtual void
      visitRegion(const index_type g_north,
                  const index_type g_west,
                  const bool g_trans,
                  /* !const */ HMatrix<value_type>& matrix,
                  const index_type north,
                  const index_type west,
                  const index_type south,
                  const index_type east) final override
      {
        for (index_type j = west; j < east; ++j)
          {
            for (index_type i = north; i < south; ++i)
              {
                const index_type gi {g_north + i - north};
                const index_type gj {g_west  + j - west};
                if (this->visited_.get(gi, gj))
                  {
                    throw std::logic_error {"already visited (global check)"};
                  }
                else
                  {
                    const index_type ii {g_trans ? j : i};
                    const index_type jj {g_trans ? i : j};
                    const value_type v {this->srcPtr_->get(gi, gj)};
                    matrix.set(v, ii, jj);
                    this->visited_.set(true, gi, gj);
                  }
              }
          }
      }
    private:

      // Delte these so we can be sure that the visitor is never copied.  It
      // must not as we want to carry state with it.

      Inserter(const Inserter&) = delete;
      Inserter(Inserter&&) = delete;
      Inserter& operator=(const Inserter&) = delete;
      Inserter& operator=(Inserter&&) = delete;

    };  // class Inserter
  }  // namespace test
}  // namespace bsc_klammler
