/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file timer.hpp
 *
 * @brief
 *         Header file providing the `Timer` class.
 *
 */

#ifndef BSC_KLAMMLER_TIMER_HPP
#define BSC_KLAMMLER_TIMER_HPP

#include <cassert>      // std::assert
#include <chrono>       // std::chrono::steady_clock, ...
#include <stdexcept>    // std::logic_error
#include <type_traits>  // std::conditional


namespace bsc_klammler
{


  /**
   * @brief
   *         A simple start-stop timer that is wrapped around
   *         `std::chrono::high_resolution_clock` or -- if that one is not
   *         steady -- `std::chrono::steady_clock`.
   *
   * A `Timer` can only be used to measure a single time span and cannot be
   * reset or restarted.  The `Timer` is a finite state machine with three
   * states and the following state-transition graph:
   *
   *                 start()             stop()
   *     UNDEFINED ----------> STARTED ---------> STOPPED <-+
   *                                                 |      | getSeconds()
   *                                                 +------+
   *
   * The function members may only be called in certain states:
   *
   * <table>
   *   <tr><th>function</th><th>calling allowed in state</th></tr>
   *   <tr><td><code>start</code></td><td><code>UNDEFINED</code></td></tr>
   *   <tr><td><code>stop</code></td><td><code>STARTED</code></td></tr>
   *   <tr><td><code>getSeconds</code></td><td><code>STOPPED</code></td></tr>
   * </table>
   *
   */
  class Timer
  {


    /* ---------------- */
    /*   Member Types   */
    /* ---------------- */

  public:

    /** @brief Floating point type used to measure time in seconds. */
    typedef double time_delta_type;

    /**
     * @brief
     *         Exception indicating that an operation was invalid in the current
     *         state of the `Timer`.
     *
     */
    class InvalidStateError : public std::logic_error
    {
    public:

      /**
       * @brief
       *         Constructor: Creates a new exception.
       *
       * @param msg
       *         informal error message explaining why the operation was
       *         disallowed in the current state
       *
       */
      InvalidStateError(const std::string& msg) : std::logic_error {msg}
      {
        // do nothing
      }
    };  // class InvalidStateError

    /**
     * @brief
     *         Helper class that allows RAII-style use of `Timer`s.
     *
     * The `Timer` (if any) will be started in the constructor and stopped in
     * the destructor.
     *
     */
    class Handle
    {

    private:

      /** @brief Pointer to the `Timer` to start / stop (if any). */
      Timer *const timerPtr_ {};

    public:

      /**
       * @brief
       *         Constructor: Creates a new `Handle` and starts the `Timer` if
       *         any.
       *
       * @param timer_ptr
       *         pointer to the `Timer` to manage or a `nullptr`
       *
       */
      Handle(Timer *const timer_ptr) : timerPtr_ {timer_ptr}
      {
        if (this->timerPtr_ != nullptr) { this->timerPtr_->start(); }
      }

      /**
       * @brief
       *         Destructor: Stopps the `Timer` if any.
       *
       */
      ~Handle()
      {
        if (this->timerPtr_ != nullptr) { this->timerPtr_->stop(); }
      }

    private:

      // These must be deleted to prevent causing invalid states.
      Handle(const Handle&) = delete;
      Handle&operator=(const Handle&) = delete;
    };  // class Handle

  private:

    /** @brief Best available clock. */
    using clock_type = typename std::conditional<
      std::chrono::high_resolution_clock::is_steady,
      std::chrono::high_resolution_clock,
      std::chrono::steady_clock>::type;

    static_assert(clock_type::is_steady, "Timer's clock is not steady");

    /** @brief Timer states. */
    enum class State {UNDEFINED, STARTED, STOPPED};


    /* ---------------- */
    /*   Data Members   */
    /* ---------------- */

  private:

    /** @brief Current timer state. */
    Timer::State state_;

    /** @brief Time point when started (not defined in state `UNDEFINED`). */
    std::chrono::time_point<clock_type> start_;

    /** @brief Time point when stopped (only defined in state `STOPPED`). */
    std::chrono::time_point<clock_type> stop_;


    /* ---------------- */
    /*   Constructors   */
    /* ---------------- */

  public:

    /**
     * @brief
     *         Constructor: Creates a new `Timer` object.
     *
     */
    Timer() noexcept : state_ {Timer::State::UNDEFINED}
    {
      // do nothing
    }


    /* -------------------- */
    /*   Function Members   */
    /* -------------------- */

    /**
     * @brief
     *         Starts the `Timer`.
     *
     * This member function can only be called once.
     *
     * @throws Timer::InvalidStateError
     *         if called more than once
     *
     */
    void
    start()
    {
      if (this->state_ != Timer::State::UNDEFINED)
        throw Timer::InvalidStateError {"timer started multiple times"};
      this->start_ = clock_type::now();
      this->state_ = Timer::State::STARTED;
    }

    /**
     * @brief
     *         Stopps the `Timer`.
     *
     * This member function can only be called once and only after the `Timer`
     * was `start()`ed.
     *
     * @throws Timer::InvalidStateError
     *         if called more than once or before `start()`ing the `Timer`
     *
     */
    void
    stop()
    {
      if (this->state_ != Timer::State::STARTED)
        throw Timer::InvalidStateError {"timer stopped but not started"};
      this->stop_ = clock_type::now();
      this->state_ = Timer::State::STOPPED;
    }

    /**
     * @brief
     *         Returns the seconds elapsed between the `Timer` was `start()`ed
     *         and `stop()`ped.
     *
     * @returns
     *         the seconds elapsed
     *
     * @throws Timer::InvalidStateError
     *         if called before the `Timer` was `start()`ed and `stop()`ped
     *
     */
    time_delta_type
    getSeconds() const
    {
      typedef std::chrono::duration<time_delta_type> delta_type;
      if (this->state_ != Timer::State::STOPPED)
        {
          throw Timer::InvalidStateError {
            "elapsed seconds requested but timer was not started and stopped"
              };
        }
      else
        {
          const delta_type delta {this->stop_ - this->start_};
          const time_delta_type seconds {delta.count()};
          assert(seconds >= 0);
          return seconds;
        }
    }
  };  // class Timer
}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_TIMER_HPP
