/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file utility.hpp
 *
 * @brief
 *         Header file providing various utility stuff.
 *
 */

#ifndef BSC_KLAMMLER_UTILITY_HPP
#define BSC_KLAMMLER_UTILITY_HPP

#include <mutex>        // std::mutex, std::unique_lock
#include <stdexcept>    // std::invalid_argument
#include <thread>       // std::thread::hardware_concurrency
#include <type_traits>  // std::conditional, std::add_const, std::remove_const
#include <utility>      // std::move

#include <Eigen/Core>   // Eigen::setNbThreads

#if BSC_KLAMMLER_PARALLEL >= BSC_KLAMMLER_PARALLEL_DATA && defined(_OPENMP)
#  include <omp.h>
#else
#  define BSC_KLAMMLER_UTILITY_HPP_FAKE_OPENMP
#endif

#if BSC_KLAMMLER_PARALLEL >= BSC_KLAMMLER_PARALLEL_DATA                       \
  && (defined(EIGEN_USE_MKL_ALL) || BSC_KLAMMLER_WITH_NEST)
#  include <mkl_service.h>
#else
#  define BSC_KLAMMLER_UTILITY_HPP_FAKE_MKL
#endif



namespace bsc_klammler
{
  /**
   * @brief
   *         Helper struct to make a type `const`-qualified if and only if a
   *         predicate is true.
   *
   * @param <T>
   *         type to make `const`-qualified
   *
   * @param <P>
   *         predicate that slects whether the type is to be made
   *         `const`-qualified
   *
   */
  template<typename T, bool P>
  struct const_if
  {
    /** @brief Correctly `const`-qualified type `T`. */
    using type = typename std::conditional<P, typename std::add_const<T>::type, typename std::remove_const<T>::type>::type;
  };

  /**
   * @brief
   *         Helper struct to force evaluation of a `constexpr` at compile
   *         time.
   *
   * @param <T>
   *         type to evaluate the expression to
   *
   * @param <V>
   *         expression to evaluate
   *
   */
  template<typename T, T V>
  struct static_eval
  {
    /** @brief Evaluated expression. */
    static constexpr T value = V;
  };

  /**
   * @brief
   *         Exclusive logical or.
   *
   * @param a
   *         first condition
   *
   * @param b
   *         second condition
   *
   * @returns
   *         `true` if and only if one condition is `true` and the other is
   *         `false`
   *
   */
  inline constexpr bool
  lxor(const bool a, const bool b) noexcept
  {
    return (a && !b) || (!a && b);
  }

  /**
   * @brief
   *         Implementation details for file `utility.hpp`.
   *
   * Depending on some compile-time `#define`s, this namespace might or might
   * not contain stubs for library functions.  This means, they can be used in
   * code like
   *
   *     void
   *     f()
   *     {
   *       using namespace utility_hpp;  // bring in stubs
   *       omp_set_num_threads(4);  // might call the actual function or else the stub
   *     }
   *
   * without the need for enclosing all of them in complicated `#ifdef`s.
   *
   */
  namespace utility_hpp
  {
#ifdef BSC_KLAMMLER_UTILITY_HPP_FAKE_OPENMP

    // https://gcc.gnu.org/onlinedocs/libgomp/omp_005fset_005fnum_005fthreads.html
    inline void
    omp_set_num_threads(int) noexcept
    {
      // do nothing
    }

    // https://gcc.gnu.org/onlinedocs/libgomp/omp_005fset_005fdynamic.html
    inline void
    omp_set_dynamic(int) noexcept
    {
      // do nothing
    }

#endif

#ifdef BSC_KLAMMLER_UTILITY_HPP_FAKE_MKL

    // https://software.intel.com/en-us/node/471134
    inline void
    mkl_set_num_threads(int) noexcept
    {
      // do nothing
    }

    // https://software.intel.com/en-us/node/471140
    inline void
    mkl_set_dynamic (int) noexcept
    {
      // do nothing
    }

#endif
  }

  /**
   * @brief
   *         Globally enables / disables data-level parallelism and sets the
   *         number of threads to use for dense linear algebra operations.
   *
   * If called for the first time with <code><var>nthreads</var> == 0</var>,
   * the number of threads is set to the guessed number of CPUs if compiled
   * with data-level parallelization or else to 1.  If the number of threads
   * was alreads set before, it is left unchanged.  In any case, the current
   * (new) number of threads is returned.
   *
   * Examples:
   *
   *     // Query the number of threads.  This performs initialization to a
   *     // default value this is the first call in the program.  Otherwise, it
   *     // is a read-only operation.
   *     unsigned nthreads = enable_data_level_multithreading();
   *
   *     // Force dense linear algebra operations to be done using only on CPU.
   *     enable_data_level_multithreading(1);
   *
   *     // Request up to 4 threads be used for dens linear algebra
   *     // (requires BSC_KLAMMLER_PARALLEL >= BSC_KLAMMLER_PARALLEL_DATA).
   *     enable_data_level_multithreading(4);
   *
   * This function is not gratis so don't call it just for fun.
   *
   * @param nthreads
   *         number of threads to use
   *
   * @returns
   *         number of threads to use
   *
   * @throws std::invalid_argument
   *         if <code><var>nthreads</var> &gt; 1</var> but the code was
   *         compiled with data-level parallelization disabled
   *
   */
  inline unsigned
  enable_data_level_multithreading(const unsigned nthreads = 0)
  {
    using namespace utility_hpp;
    static unsigned global_num_threads {1};
    static std::mutex mutex {};
    const std::unique_lock<std::mutex> lock {mutex};
    if (BSC_KLAMMLER_CHECK_ARGS && BSC_KLAMMLER_PARALLEL < BSC_KLAMMLER_PARALLEL_DATA)
      {
        if (nthreads > 1)
          {
            throw std::invalid_argument {
              "more than 1 CPU requested for dense linear algebra operations"
              " but compiled with data-level parallelization disabled"
            };
          }
      }
    if (nthreads)
      {
        global_num_threads = nthreads;
        if (BSC_KLAMMLER_PARALLEL >= BSC_KLAMMLER_PARALLEL_DATA)
          {
            //omp_set_dynamic(true);
            omp_set_num_threads(global_num_threads);
            //mkl_set_dynamic(true);
            mkl_set_num_threads(global_num_threads);
            Eigen::setNbThreads(global_num_threads);
          }
      }
    return global_num_threads;
  }

  /**
   * @brief
   *     A RAII-styl wrapper for ensuring a `std::thread` is `%join()`ed.
   *
   * Example usage:
   *
   *     void
   *     demo()
   *     {
   *       ThreadGuard tg {std::thread {[]()->void{
   *         std::this_thread::sleep_for(std::chrono::seconds {1});
   *        }}};
   *        // Some other code that might throw or return...
   *        tg.thread().join();  // optional, but preferable
   *        // If we got here, '~ThreadGuard()' will be a no-op, otherwise, if
   *        // the previous line was not executed, it will join the thread.  If
   *        // an exception is thrown while `%join()`ing (in the destructor),
   *        // the program will be terminated.
   *      }
   *
   */
  class ThreadGuard
  {
  private:

    /** @brief The `std::thread` that is managd. */
    std::thread thread_;

  public:

    /**
     * @brief
     *         Constructor: Creates a new `ThreadGuard` that manages a
     *         `std::thread`.
     *
     * @param thread
     *         the `std::thread` to be managed
     *
     */
    ThreadGuard(std::thread&& thread) noexcept : thread_ {std::move(thread)}
    {
      // do nothing
    }

    /**
     * @brief
     *         Destructor: `%join()`s the managed thread if it is still alive.
     *
     * Note that if `%join()` throws, this will terminate the program.  So
     * manually `%join()`ing while the `ThreadGuard` is still alive and only
     * using the destructor as a security fallback mechanism might be
     * desirable.
     *
     */
    ~ThreadGuard() noexcept
    {
      if (this->thread_.joinable())
        this->thread_.join();
    }

    /** @brief Deleted copy constructor. */
    ThreadGuard(const ThreadGuard&) noexcept = delete;

    /** @brief Default move constructor. */
    ThreadGuard(ThreadGuard&&) noexcept = default;

    /** @brief Deleted copy-assignment operator. */
    ThreadGuard&
    operator=(const ThreadGuard&) noexcept = delete;

    /** @brief Default move-assignment operator. */
    ThreadGuard&
    operator=(ThreadGuard&&) noexcept = default;

    /**
     * @brief
     *         Exchanges thread ownership between two `ThreadGuard`s.
     *
     * @param first
     *         first `ThreadGuard` to swap
     *
     * @param second
     *         second `ThreadGuard` to swap
     *
     */
    friend void
    swap(ThreadGuard& first, ThreadGuard& second) noexcept
    {
      using std::swap;
      swap(first.thread_, second.thread_);
    }

    /**
     * @brief
     *         Proveides access to the managed `std::thread`.
     *
     * @returns
     *         a reference to the managed `std::thread`
     *
     */
    std::thread&
    thread() noexcept
    {
      return this->thread_;
    }

    /**
     * @brief
     *         Proveides read-only access to the managed `std::thread`.
     *
     * @returns
     *         a `const` reference to the managed `std::thread`
     *
     */
    const std::thread&
    thread() const noexcept
    {
      return this->thread_;
    }
  };  // class ThreadGuard

}  // namespace bsc_klammler

#endif  // ifndef BSC_KLAMMLER_UTILITY_HPP
