/// Copyright (c) 2014 Moritz Klammler
///
/// Permission is  hereby granted, free  of charge,  to any person  obtaining a
/// copy of this software and  associated documentation files (the "Software"),
/// to deal in  the Software without restriction,  including without limitation
/// the rights  to use, copy,  modify, merge, publish,  distribute, sublicense,
/// and/or  sell copies  of the  Software, and  to permit  persons to  whom the
/// Software is furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this  permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS",  WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING  BUT NOT LIMITED  TO THE WARRANTIES  OF MERCHANTABILITY,
/// FITNESS FOR  A PARTICULAR PURPOSE  AND NONINFRINGEMENT.  IN NO  EVENT SHALL
/// THE AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER  IN AN  ACTION OF CONTRACT,  TORT OR  OTHERWISE, ARISING
/// FROM,  OUT OF  OR IN  CONNECTION  WITH THE  SOFTWARE  OR THE  USE OR  OTHER
/// DEALINGS IN THE SOFTWARE.


/**
 * @file test_blocl_llt_preconditioner.cpp
 *
 * @brief
 *         Unit tests for the `block_llt_preconditioner.hpp` header.
 *
 */

#include "../global.hpp"                    // (package configuration)
#include "../block_llt_preconditioner.hpp"  // bsc_klammler::BlockLLTPreconditioner

#define BOOST_TEST_MODULE block_llt_preconditioner
#include "module.hpp"

template<typename T>
using preconditioner_t = bsc_klammler::BlockLLTPreconditioner<T>;

#include "common_test_preconditioner.hpp"


/**
 * @brief
 *         Proof of concept for solving linear systems for the factors
 *         <var>L</var> and <var>U</var> both, from the left and from the
 *         right.
 *
 */
BOOST_AUTO_TEST_CASE(triangular_substitution)
{
  using full_matrix = bsc_klammler::FullMatrix<double>;
  using raw_matrix = full_matrix::raw_matrix_type;
  using llt_factorization = Eigen::LLT<raw_matrix>;
  constexpr double tol {1.0E-10};
  const full_matrix coeffs {{3.0, 1.0}, {1.0, 4.0}};
  llt_factorization llt {};
  llt.compute(coeffs.getRawMatrix());
  {
    const full_matrix matrix {llt.matrixL()};
    const full_matrix vectors {{1.0, 3.0, 5.0}, {2.0, 4.0, 6.0}};
    const full_matrix solution {llt.matrixL().solve(vectors.getRawMatrix())};
    MY_CHECK_MATRIX_CLOSE(matrix * solution, vectors, tol);
  }
  {
    const full_matrix matrix {llt.matrixU()};
    const full_matrix vectors {{1.0, 2.0}, {3.0, 4.0}, {5.0, 6.0}};
    const full_matrix solution {
      llt.matrixU().transpose()
        .solve(vectors.getRawMatrix().transpose())
        .transpose()
    };
    MY_CHECK_MATRIX_CLOSE(solution * matrix, vectors, tol);
  }
}


/**
 * @brief
 *         Proof of concept for substituting for triangular parts of matrices.
 *
 */
BOOST_AUTO_TEST_CASE(triangular_view)
{
  typedef double value_type;
  using full_matrix = bsc_klammler::FullMatrix<value_type>;
  using full_vector = bsc_klammler::FullVector<value_type>;
  const full_matrix backbone {
    {-1.0, -1.0, -1.0, -1.0, -1.0},
    {-1.0, -1.0, -1.0, -1.0, -1.0},
    {-1.0, 11.0,  1.2,  1.3,  1.4},
    {-1.0,  2.1, 22.0,  2.3,  2.4},
    {-1.0,  3.1,  3.2, 33.0,  3.4},
    {-1.0,  4.1,  4.2,  4.3, 44.0},
    {-1.0, -1.0, -1.0, -1.0, -1.0},
  };
  const full_matrix expected_lower {
    {11.0,  0.0,  0.0,  0.0},
    { 2.1, 22.0,  0.0,  0.0},
    { 3.1,  3.2, 33.0,  0.0},
    { 4.1,  4.2,  4.3, 44.0},
  };
  const full_matrix expected_upper {
    {11.0,  1.2,  1.3,  1.4},
    { 0.0, 22.0,  2.3,  2.4},
    { 0.0,  0.0, 33.0,  3.4},
    { 0.0,  0.0,  0.0, 44.0},
  };
  const full_vector rhs {{1.0, 2.0, 3.0, 4.0}};
  const auto raw_lower = backbone.getRawMatrix().block(2, 1, 4, 4)
    .triangularView<Eigen::Lower>();
  const auto raw_upper = backbone.getRawMatrix().block(2, 1, 4, 4)
    .triangularView<Eigen::Upper>();
  const full_matrix actual_lower {raw_lower};
  const full_matrix actual_upper {raw_upper};
  MY_CHECK_MATRIX_EQUAL(expected_lower, actual_lower);
  MY_CHECK_MATRIX_EQUAL(expected_upper, actual_upper);
  const full_vector actual_lhs_l {raw_lower.solve(rhs.getRawVector())};
  const full_vector actual_lhs_u {raw_upper.solve(rhs.getRawVector())};
  MY_CHECK_VECTOR_CLOSE(rhs, expected_lower * actual_lhs_l, 1.0E-10);
  MY_CHECK_VECTOR_CLOSE(rhs, expected_upper * actual_lhs_u, 1.0E-10);
}
